Package: MachineShop
Type: Package
Title: Machine Learning Models and Tools
Version: 3.9.4
Date: 2026-09-04
Authors@R: person(c("Brian", "J"), "Smith",
                  email = "brian-j-smith@uiowa.edu",
                  role = c("aut", "cre"))
Author: Brian J Smith [aut, cre]
Maintainer: Brian J Smith <brian-j-smith@uiowa.edu>
Description: Meta-package for statistical and machine learning with a unified
    interface for model fitting, prediction, performance assessment, and
    presentation of results.  Approaches for model fitting and prediction of
    numerical, categorical, or censored time-to-event outcomes include
    traditional regression models, regularization methods, tree-based methods,
    support vector machines, neural networks, ensembles, data preprocessing,
    filtering, and model tuning and selection.  Performance metrics are provided
    for model assessment and can be estimated with independent test sets, split
    sampling, cross-validation, or bootstrap resampling.  Resample estimation
    can be executed in parallel for faster processing and nested in cases of
    model tuning and selection.  Modeling results can be summarized with
    descriptive statistics; calibration curves; variable importance; partial
    dependence plots; confusion matrices; and ROC, lift, and other performance
    curves.
Depends: R (>= 4.1.0)
Imports: abind, cli (>= 3.1.0), dials (>= 0.0.4), foreach, ggplot2 (>=
        3.4.0), kernlab, magrittr, Matrix (>= 1.5-0), methods, nnet,
        party, polspline, progress, recipes (>= 1.0.0), rlang, rsample
        (>= 1.1.0), Rsolnp, survival, tibble, utils
Suggests: adabag, BART, bartMachine, C50, censored, cluster,
        doParallel, e1071, earth, elasticnet, generics, gbm, glmnet,
        gridExtra, Hmisc, kableExtra, kknn, knitr, lars, MASS, mboost,
        mda, ParBayesianOptimization, parsnip (>= 1.1.0), partykit,
        pls, pso, randomForest, randomForestSRC, ranger,
        rBayesianOptimization, rmarkdown, rms, rpart, testthat, tree,
        xgboost (>= 3.1.2)
Additional_repositories: https://brian-j-smith.github.io/drat
LazyData: true
License: GPL-3
URL: https://brian-j-smith.github.io/MachineShop/
BugReports: https://github.com/brian-j-smith/MachineShop/issues
VignetteBuilder: knitr
Encoding: UTF-8
Collate: 'classes.R' 'conditions.R' 'MachineShop-package.R'
        'MLControl.R' 'MLInput.R' 'MLMetric.R' 'MLModel.R'
        'MLOptimization.R' 'ML_AdaBagModel.R' 'ML_AdaBoostModel.R'
        'ML_BARTMachineModel.R' 'ML_BARTModel.R' 'ML_BlackBoostModel.R'
        'ML_C50Model.R' 'ML_CForestModel.R' 'ML_CoxModel.R'
        'ML_EarthModel.R' 'ML_FDAModel.R' 'ML_GAMBoostModel.R'
        'ML_GBMModel.R' 'ML_GLMBoostModel.R' 'ML_GLMModel.R'
        'ML_GLMNetModel.R' 'ML_KNNModel.R' 'ML_LARSModel.R'
        'ML_LDAModel.R' 'ML_LMModel.R' 'ML_MDAModel.R' 'ML_NNetModel.R'
        'ML_NaiveBayesModel.R' 'ML_ParsnipModel.R' 'ML_PLSModel.R'
        'ML_POLRModel.R' 'ML_QDAModel.R' 'ML_RFSRCModel.R'
        'ML_RPartModel.R' 'ML_RandomForestModel.R' 'ML_RangerModel.R'
        'ML_SVMModel.R' 'ML_StackedModel.R' 'ML_SuperModel.R'
        'ML_SurvRegModel.R' 'ML_TreeModel.R' 'ML_XGBModel.R'
        'ModelFrame.R' 'ModelRecipe.R' 'ModelSpecification.R'
        'TrainedInputs.R' 'TrainedModels.R' 'TrainingParams.R'
        'append.R' 'calibration.R' 'case_comps.R' 'coerce.R'
        'combine.R' 'confusion.R' 'convert.R' 'data.R' 'dependence.R'
        'diff.R' 'expand.R' 'extract.R' 'fit.R' 'grid.R' 'metricinfo.R'
        'metrics.R' 'metrics_factor.R' 'metrics_numeric.R'
        'modelinfo.R' 'models.R' 'performance.R' 'performance_curve.R'
        'plot.R' 'predict.R' 'print.R' 'recipe_roles.R' 'reexports.R'
        'resample.R' 'response.R' 'rfe.R' 'settings.R' 'step_kmeans.R'
        'step_kmedoids.R' 'step_lincomp.R' 'step_sbf.R' 'step_spca.R'
        'summary.R' 'survival.R' 'utils.R' 'varimp.R'
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-09-04 13:20:49 UTC; bjsmith
Repository: CRAN
Date/Publication: 2026-09-04 15:50:19 UTC

useDynLib(MachineShop, .registration = TRUE, .fixes = "C_")

import(methods)

importClassesFrom(kernlab, ksvm)
importClassesFrom(party, RandomForest)

importFrom(abind, abind, adrop)
importFrom(cli, ansi_nchar, ansi_strtrim, ansi_strwrap, make_ansi_style,
           console_width, pluralize, qty, style_bold, style_reset)
importFrom(dials, parameters)
importFrom(foreach, "%dopar%", foreach, getDoParName, getDoParRegistered,
           getDoParWorkers, registerDoSEQ)
importFrom(ggplot2, aes, after_stat, coord_fixed, coord_flip, facet_wrap,
           geom_abline, geom_bar, geom_boxplot, geom_crossbar, geom_density,
           geom_errorbar, geom_line, geom_path, geom_point, geom_raster,
           geom_ribbon, geom_segment, geom_violin, ggplot, labs, mean_se,
           position_dodge, scale_fill_gradient, stat_summary)
importFrom(graphics, plot)
importFrom(magrittr, "%>%")
importFrom(polspline, hare, phare)
importFrom(progress, progress_bar)
importFrom(recipes, bake, prep, recipe, tidy, tunable)
importFrom(rsample, bootstraps, group_bootstraps, group_initial_split,
           group_vfold_cv, initial_split, vfold_cv)
importFrom(stats, aggregate, approx, coef, complete.cases, contr.helmert, cor,
           cov.wt, delete.response, drop1, formula, get_all_vars, loess, lm,
           median, model.frame, model.matrix, model.matrix.default,
           model.response, model.weights, na.fail, na.omit, na.pass, offset,
           p.adjust, pchisq, pnorm, predict, pt, quantile, reformulate,
           reshape, runif, sd, t.test, terms, time, update, var, vcov, xtabs)
importFrom(survival, concordance, is.Surv, Surv, survfit, survreg,
           survreg.distributions)
importFrom(tibble, as_tibble, as_tibble_col, is_tibble, tibble)
importFrom(utils, capture.output, combn, head, methods, packageVersion, relist,
           stack, str, tail)


export(BinomialVariate,
       DiscreteVariate,
       NegBinomialVariate,
       PoissonVariate,
       role_binom,
       role_case,
       role_pred,
       role_surv,
       SurvEvents,
       SurvProbs)

export(expand_model,
       expand_modelgrid,
       expand_params,
       expand_steps,
       ParameterGrid,
       TuningGrid)

export(MLModel,
       MLModelFunction,
       AdaBagModel,
       AdaBoostModel,
       BARTModel,
       BARTMachineModel,
       BlackBoostModel,
       C50Model,
       CForestModel,
       CoxModel,
       CoxStepAICModel,
       EarthModel,
       FDAModel,
       GAMBoostModel,
       GBMModel,
       GLMBoostModel,
       GLMModel,
       GLMStepAICModel,
       GLMNetModel,
       KNNModel,
       LARSModel,
       LDAModel,
       LMModel,
       MDAModel,
       NaiveBayesModel,
       NNetModel,
       ParsnipModel,
       PDAModel,
       PLSModel,
       POLRModel,
       QDAModel,
       RandomForestModel,
       RangerModel,
       RFSRCModel,
       RFSRCFastModel,
       RPartModel,
       SelectedModel,
       StackedModel,
       SuperModel,
       SurvRegModel,
       SurvRegStepAICModel,
       SVMModel,
       SVMANOVAModel,
       SVMBesselModel,
       SVMLaplaceModel,
       SVMLinearModel,
       SVMPolyModel,
       SVMRadialModel,
       SVMSplineModel,
       SVMTanhModel,
       TreeModel,
       TunedModel,
       XGBModel,
       XGBDARTModel,
       XGBLinearModel,
       XGBTreeModel)

export(ModelFrame,
       ModelSpecification,
       SelectedInput,
       TunedInput)

export(BootControl,
       BootOptimismControl,
       CVControl,
       CVOptimismControl,
       OOBControl,
       SplitControl,
       TrainControl,
       set_monitor,
       set_optim_bayes,
       set_optim_bfgs,
       set_optim_grid,
       set_optim_method,
       set_optim_pso,
       set_optim_sann,
       set_predict,
       set_strata)

export(MLMetric,
       "MLMetric<-",
       accuracy,
       auc,
       brier,
       cindex,
       cross_entropy,
       f_score,
       fnr,
       fpr,
       gini,
       kappa2,
       mae,
       mse,
       msle,
       npv,
       ppr,
       ppv,
       pr_auc,
       precision,
       r2,
       recall,
       rmse,
       rmsle,
       roc_auc,
       roc_index,
       sensitivity,
       specificity,
       tnr,
       tpr,
       weighted_kappa2)

export(.,
       "%>%",
       as.MLInput,
       as.MLModel,
       calibration,
       case_weights,
       confusion,
       ConfusionMatrix,
       dependence,
       fit,
       lift,
       metricinfo,
       modelinfo,
       performance,
       performance_curve,
       resample,
       response,
       rfe,
       settings,
       step_kmeans,
       step_kmedoids,
       step_lincomp,
       step_sbf,
       step_spca,
       tunable.step_kmeans,
       tunable.step_kmedoids,
       tunable.step_lincomp,
       tunable.step_spca,
       unMLModelFit,
       varimp)

exportMethods(predict)

S3method(as.data.frame, BinomialVariate)
S3method(as.data.frame, ModelFrame)
S3method(as.data.frame, PerformanceDiffTest)
S3method(as.data.frame, Resample)
S3method(as.data.frame, SurvMatrix)
S3method(as.data.frame, TabularArray)
S3method(as.double, BinomialVariate)
S3method(as.MLControl, character)
S3method(as.MLControl, default)
S3method(as.MLControl, "function")
S3method(as.MLControl, MLControl)
S3method(as.MLControl, "NULL")
S3method(as.MLMetric, character)
S3method(as.MLMetric, default)
S3method(as.MLMetric, MLMetric)
S3method(as.MLInput, default)
S3method(as.MLInput, formula)
S3method(as.MLInput, matrix)
S3method(as.MLInput, MLInput)
S3method(as.MLInput, MLModelFit)
S3method(as.MLInput, ModelSpecification)
S3method(as.MLInput, recipe)
S3method(as.MLModel, character)
S3method(as.MLModel, default)
S3method(as.MLModel, MLModel)
S3method(as.MLModel, MLModelFit)
S3method(as.MLModel, MLModelFunction)
S3method(as.MLModel, model_spec)
S3method(as.MLModel, ModelSpecification)
S3method(as.MLModel, "NULL")
S3method(as_string, default)
S3method(as_string, character)
S3method(bake, step_kmeans)
S3method(bake, step_kmedoids)
S3method(bake, step_lincomp)
S3method(bake, step_sbf)
S3method(bake, step_spca)
S3method(bake, ModelRecipe)
S3method(bake, SelectedInput)
S3method(bake, TunedInput)
S3method(c, BinomialVariate)
S3method(c, Calibration)
S3method(c, ConfusionList)
S3method(c, ConfusionMatrix)
S3method(c, DiscreteVariate)
S3method(c, LiftCurve)
S3method(c, ListOf)
S3method(c, PerformanceCurve)
S3method(c, Resample)
S3method(c, SurvMatrix)
S3method(diff, MLModel)
S3method(diff, Performance)
S3method(diff, Resample)
S3method(expand_modelgrid, formula)
S3method(expand_modelgrid, matrix)
S3method(expand_modelgrid, MLModel)
S3method(expand_modelgrid, MLModelFunction)
S3method(expand_modelgrid, ModelFrame)
S3method(expand_modelgrid, ModelSpecification)
S3method(expand_modelgrid, recipe)
S3method(fit, formula)
S3method(fit, matrix)
S3method(fit, MLModel)
S3method(fit, MLModelFunction)
S3method(fit, ModelFrame)
S3method(fit, ModelSpecification)
S3method(fit, recipe)
S3method(format, SurvMatrix)
S3method(formula, ModelFrame)
S3method(map, character)
S3method(map, "function")
S3method(model.matrix, ModelDesignTerms)
S3method(model.matrix, ModelFormulaTerms)
S3method(model.matrix, ModelFrame)
S3method(ModelFrame, data.frame)
S3method(ModelFrame, formula)
S3method(ModelFrame, matrix)
S3method(ModelFrame, ModelFrame)
S3method(ModelFrame, ModelTerms)
S3method(ModelFrame, recipe)
S3method(ModelSpecification, default)
S3method(ModelSpecification, formula)
S3method(ModelSpecification, matrix)
S3method(ModelSpecification, ModelFrame)
S3method(ModelSpecification, recipe)
S3method(ParameterGrid, list)
S3method(ParameterGrid, param)
S3method(ParameterGrid, parameters)
S3method(performance, BinomialVariate)
S3method(performance, BootOptimismControl)
S3method(performance, ConfusionList)
S3method(performance, ConfusionMatrix)
S3method(performance, CVOptimismControl)
S3method(performance, factor)
S3method(performance, matrix)
S3method(performance, MLControl)
S3method(performance, MLModel)
S3method(performance, numeric)
S3method(performance, Resample)
S3method(performance, Surv)
S3method(performance, TrainingStep)
S3method(performance_curve, default)
S3method(performance_curve, Resample)
S3method(plot, Calibration)
S3method(plot, ConfusionList)
S3method(plot, ConfusionMatrix)
S3method(plot, LiftCurve)
S3method(plot, MLModel)
S3method(plot, PartialDependence)
S3method(plot, Performance)
S3method(plot, PerformanceCurve)
S3method(plot, Resample)
S3method(plot, TrainingStep)
S3method(plot, VariableImportance)
S3method(predict, MLModelFit)
S3method(prep, ModelRecipe)
S3method(prep, SelectedInput)
S3method(prep, step_kmeans)
S3method(prep, step_kmedoids)
S3method(prep, step_lincomp)
S3method(prep, step_sbf)
S3method(prep, step_spca)
S3method(prep, TunedInput)
S3method(print, BinomialVariate)
S3method(print, BootControl)
S3method(print, Calibration)
S3method(print, ConfusionList)
S3method(print, ConfusionMatrix)
S3method(print, ConfusionSummary)
S3method(print, CVControl)
S3method(print, DiscreteVariate)
S3method(print, EnsembleModel)
S3method(print, ListOf)
S3method(print, MLControl)
S3method(print, MLMetric)
S3method(print, MLModel)
S3method(print, MLModelFit)
S3method(print, MLModelFunction)
S3method(print, MLOptimization)
S3method(print, ModelFrame)
S3method(print, ModelRecipe)
S3method(print, ModelSpecification)
S3method(print, ModelTerms)
S3method(print, NullControl)
S3method(print, OOBControl)
S3method(print, ParameterGrid)
S3method(print, ParsnipModel)
S3method(print, Performance)
S3method(print, PerformanceCurve)
S3method(print, PerformanceDiffTest)
S3method(print, RandomGridSearch)
S3method(print, RecipeGrid)
S3method(print, Resample)
S3method(print, SelectedInput)
S3method(print, SequentialOptimization)
S3method(print, SplitControl)
S3method(print, step_kmeans)
S3method(print, step_kmedoids)
S3method(print, step_lincomp)
S3method(print, step_sbf)
S3method(print, step_spca)
S3method(print, SurvMatrix)
S3method(print, SurvTimes)
S3method(print, TrainingParams)
S3method(print, TrainingStep)
S3method(print, TunedInput)
S3method(print, TunedModel)
S3method(print, TuningGrid)
S3method(print, VariableImportance)
S3method(resample, formula)
S3method(resample, matrix)
S3method(resample, MLModel)
S3method(resample, MLModelFunction)
S3method(resample, ModelFrame)
S3method(resample, ModelSpecification)
S3method(resample, recipe)
S3method(response, formula)
S3method(response, MLModelFit)
S3method(response, ModelFrame)
S3method(response, ModelSpecification)
S3method(response, recipe)
S3method(rfe, formula)
S3method(rfe, matrix)
S3method(rfe, MLModel)
S3method(rfe, MLModelFunction)
S3method(rfe, ModelFrame)
S3method(rfe, ModelSpecification)
S3method(rfe, recipe)
S3method(SelectedInput, formula)
S3method(SelectedInput, list)
S3method(SelectedInput, matrix)
S3method(SelectedInput, ModelFrame)
S3method(SelectedInput, ModelSpecification)
S3method(SelectedInput, recipe)
S3method(SelectedModel, default)
S3method(SelectedModel, list)
S3method(SelectedModel, ModelSpecification)
S3method(set_monitor, list)
S3method(set_monitor, MLControl)
S3method(set_monitor, MLOptimization)
S3method(set_monitor, ModelSpecification)
S3method(set_monitor, SplitControl)
S3method(set_monitor, TrainingParams)
S3method(set_optim_bayes, ModelSpecification)
S3method(set_optim_bfgs, ModelSpecification)
S3method(set_optim_grid, TrainingParams)
S3method(set_optim_grid, ModelSpecification)
S3method(set_optim_grid, TunedInput)
S3method(set_optim_grid, TunedModel)
S3method(set_optim_method, ModelSpecification)
S3method(set_optim_pso, ModelSpecification)
S3method(set_optim_sann, ModelSpecification)
S3method(summary, ConfusionList)
S3method(summary, ConfusionMatrix)
S3method(summary, MLModel)
S3method(summary, MLModelFit)
S3method(summary, Performance)
S3method(summary, PerformanceCurve)
S3method(summary, Resample)
S3method(summary, SurvRegModelFit)
S3method(summary, TrainingStep)
S3method(t.test, PerformanceDiff)
S3method(terms, ModelFrame)
S3method(tidy, step_kmeans)
S3method(tidy, step_kmedoids)
S3method(tidy, step_lincomp)
S3method(tidy, step_sbf)
S3method(tidy, step_spca)
S3method(tunable, step_kmeans)
S3method(tunable, step_kmedoids)
S3method(tunable, step_lincomp)
S3method(tunable, step_spca)
S3method(TunedInput, recipe)
S3method(update, MLModelFit)
S3method(update, ModelFrame)
S3method(update, ModelRecipe)
S3method(update, ModelSpecification)
S3method(update, TunedModelRecipe)
S3method(update, SelectedInput)
S3method(update, VariableImportance)
S3method("[", BinomialVariate)

MachineShop: Machine Learning Models and Tools for R
================

[![Generic
badge](https://img.shields.io/badge/docs-online-green.svg)](https://brian-j-smith.github.io/MachineShop/)
[![CRAN_Status_Badge](http://www.r-pkg.org/badges/version/MachineShop)](https://CRAN.R-project.org/package=MachineShop)

# Description

**MachineShop** is a meta-package for statistical and machine learning
with a unified interface for model fitting, prediction, performance
assessment, and presentation of results. Support is provided for
predictive modeling of numerical, categorical, and censored
time-to-event outcomes and for resample (bootstrap, cross-validation,
and split training-test sets) estimation of model performance. This
vignette introduces the package interface with a survival data analysis
example, followed by supported methods of variable specification;
applications to other response variable types; available performance
metrics, resampling techniques, and graphical and tabular summaries; and
modeling strategies.

# Features

- Unified and concise interface for model fitting, prediction, and
  performance assessment.
- Support for 53+ models from 28 **R** packages, including model
  specifications from the [**parsnip**](https://parsnip.tidymodels.org/)
  package.
- Dynamic model parameters.
- Ensemble modeling with stacked regression and super learners.
- Modeling of response variables types: binary factors, multi-class
  nominal and ordinal factors, numeric vectors and matrices, and
  censored time-to-event survival.
- Model specification with traditional formulas, design matrices, and
  flexible pre-processing [recipes](https://recipes.tidymodels.org/).
- Resample estimation of predictive performance, including
  cross-validation, bootstrap resampling, and split training-test set
  validation.
- Parallel execution of resampling algorithms.
- Choices of performance metrics: accuracy, areas under ROC and
  precision recall curves, Brier score, coefficient of determination
  (R<sup>2</sup>), concordance index, cross entropy, F score, Gini
  coefficient, unweighted and weighted Cohen’s kappa, mean absolute
  error, mean squared error, mean squared log error, positive and
  negative predictive values, precision and recall, and sensitivity and
  specificity.
- Graphical and tabular performance summaries: calibration curves,
  confusion matrices, partial dependence plots, performance curves, lift
  curves, and model-specific and permutation-based variable importance.
- Model tuning over automatically generated grids and with exhaustive
  and random grid searches, Bayesian optimization, particle swarm
  optimization, quasi-Newton BFGS optimization, simulated annealing, and
  support for user-defined optimization functions.
- Model selection and comparisons for any combination of models and
  model parameter values.
- Recursive feature elimination.
- User-definable models and performance metrics.

# Getting Started

## Installation

``` r
# Current release from CRAN
install.packages("MachineShop")

# Development version from GitHub
# install.packages("pak")
pak::pak("brian-j-smith/MachineShop@develop")
```

## Documentation

Once installed, the following **R** commands will load the package and
display its help system documentation. Online documentation and examples
are available at the [MachineShop
website](https://brian-j-smith.github.io/MachineShop/).

``` r
library(MachineShop)

# Package help summary
?MachineShop
```

## ----setup, include=FALSE-----------------------------------------------------
source("setup.R")
rdoc_url <- function(names, ...) names

## ----overview_modelinfo, echo=FALSE, message=FALSE----------------------------
library(MachineShop)
info <- modelinfo()

## ----overview_install, eval = FALSE-------------------------------------------
# # Current release from CRAN
# install.packages("MachineShop")
# 
# # Development version from GitHub
# # install.packages("devtools")
# devtools::install_github("brian-j-smith/MachineShop")
# 
# # Development version with vignettes
# devtools::install_github("brian-j-smith/MachineShop", build_vignettes = TRUE)

## ----overview_docs, eval = FALSE, message = FALSE-----------------------------
# library(MachineShop)
# 
# # Package help summary
# ?MachineShop
# 
# # Vignette
# RShowDoc("UserGuide", package = "MachineShop")

## ----using_example_melanoma---------------------------------------------------
## Analysis libraries and dataset
library(MachineShop)
library(survival)
data(Melanoma, package = "MASS")

## Malignant melanoma analysis dataset
surv_df <- within(Melanoma, {
  y <- Surv(time, status != 2)
  remove(time, status)
})

## ----using_example_summary, echo=FALSE----------------------------------------
surv_stats <- list(
  list("Number of subjects" = ~ length(y)),
  "time" = list("Median (Range)" = ~ median_range(y[, "time"])),
  "status" = list("1 = Dead" = ~ n_perc(y[, "status"] == 1),
                  "0 = Alive" = ~ n_perc(y[, "status"] == 0)),
  "sex" = list("1 = Male" = ~ n_perc(sex == 1),
               "0 = Female" = ~ n_perc(sex == 0)),
  "age" = list("Median (Range)" = ~ median_range(age)),
  "year" = list("Median (Range)" = ~ median_range(year)),
  "thickness" = list("Median (Range)" = ~ median_range(thickness)),
  "ulcer" = list("1 = Presence" = ~ n_perc(ulcer == 1),
                 "0 = Absence" = ~ n_perc(ulcer == 0))
)

summary_kbl(surv_stats, surv_df)

## ----using_example_survfit, echo=FALSE----------------------------------------
library(ggplot2)

col <- "#F8766D"
survfit(y ~ 1, data = surv_df) %>%
  with(data.frame(time, surv, lower, upper, censor = ifelse(n.censor > 0, time, NA))) %>%
  ggplot(aes(x = time, y = surv)) +
  geom_ribbon(aes(ymin = lower, ymax = upper), fill = col, alpha = 0.2) +
  geom_step(color = col) +
  geom_point(aes(x = censor), shape = 3, color = col) +
  coord_cartesian(ylim = c(0, 1)) +
  labs(x = "Follow-Up Time (Days)", y = "Overall Survival Probability",
       title = "Kaplan-Meier survival plot")

## ----using_example_datasets---------------------------------------------------
## Training and test sets
set.seed(123)
train_indices <- sample(nrow(surv_df), nrow(surv_df) * 2 / 3)
surv_train <- surv_df[train_indices, ]
surv_test <- surv_df[-train_indices, ]

## Global formula for the analysis
surv_fo <- y ~ sex + age + year + thickness + ulcer

## ----using_fit_modelinfo------------------------------------------------------
## All available models
modelinfo() %>% names

## ----using_fit_modelinfo_gbmmodel---------------------------------------------
## Model-specific information
modelinfo(GBMModel)

## ----using_fit_gbmmodel-------------------------------------------------------
GBMModel

## ----using_fit_modelinfo_type-------------------------------------------------
## All survival response-specific models
modelinfo(Surv(0)) %>% names

## Identify survival response-specific models
modelinfo(Surv(0), CoxModel, GBMModel, SVMModel) %>% names

## ----using_fit_modelinfo_response---------------------------------------------
## Models for a responses variable
modelinfo(surv_df$y) %>% names

## ----using_fit_function, results="hide"---------------------------------------
## Generalized boosted regression fit

## Model function
surv_fit <- fit(surv_fo, data = surv_train, model = GBMModel)

## Model function name
fit(surv_fo, data = surv_train, model = "GBMModel")

## Model object
fit(surv_fo, data = surv_train, model = GBMModel(n.trees = 100, interaction.depth = 1))

## ----using_fit_dynamic, results="hide"----------------------------------------
## Dynamic model parameter k = log number of observations

## Number of observations: nobs
fit(surv_fo, data = surv_train, model = CoxStepAICModel(k = .(log(nobs))))

## Response variable: y
fit(surv_fo, data = surv_train, model = CoxStepAICModel(k = .(log(length(y)))))

## ----using_predict_function---------------------------------------------------
## Predicted survival means (default: Weibull distribution)
predict(surv_fit, newdata = surv_test)

## Predicted survival means (empirical distribution)
predict(surv_fit, newdata = surv_test, distr = "empirical")

## ----using_predict_function_times---------------------------------------------
## Predict survival probabilities and events at specified follow-up times
surv_times <- 365 * c(5, 10)

predict(surv_fit, newdata = surv_test, times = surv_times, type = "prob")

predict(surv_fit, newdata = surv_test, times = surv_times, cutoff = 0.7)

## ----using_variables_formula--------------------------------------------------
## Datasets
data(Pima.te, package = "MASS")
data(Pima.tr, package = "MASS")

## Formula specification
model_fit <- fit(type ~ ., data = Pima.tr, model = GBMModel)
predict(model_fit, newdata = Pima.te) %>% head

## ----using_variables_formula_RHS----------------------------------------------
settings("RHS.formula")

## ----using_variables_matrix---------------------------------------------------
## Example design matrix and response object
x <- model.matrix(type ~ . - 1, data = Pima.tr)
y <- Pima.tr$type

## Design matrix specification
model_fit <- fit(x, y, model = GBMModel)
predict(model_fit, newdata = Pima.te) %>% head

## ----using_variables_modelframe-----------------------------------------------
## Model frame specification

## Formula
mf <- ModelFrame(type ~ ., data = Pima.tr)
model_fit <- fit(mf, model = GBMModel)
predict(model_fit, newdata = Pima.te) %>% head

## Design matrix
mf <- ModelFrame(x, y)
model_fit <- fit(mf, model = GBMModel)
predict(model_fit, newdata = Pima.te) %>% head

## ----using_variables_modelframe_weights, results="hide"-----------------------
## Model frame specification with case weights
mf <- ModelFrame(ncases / (ncases + ncontrols) ~ agegp + tobgp + alcgp, data = esoph,
                 weights = ncases + ncontrols)
fit(mf, model = GBMModel)

## ----using_variables_recipe---------------------------------------------------
## Recipe specification
library(recipes)

rec <- recipe(type ~ ., data = Pima.tr)
model_fit <- fit(rec, model = GBMModel)
predict(model_fit, newdata = Pima.te) %>% head

## ----using_variables_recipe_weights, results="hide"---------------------------
## Recipe specification with case weights
df <- within(esoph, {
  y <- ncases / (ncases + ncontrols)
  weights <- ncases + ncontrols
  remove(ncases, ncontrols)
})

rec <- recipe(y ~ agegp + tobgp + alcgp + weights, data = df) %>%
  role_case(weight = weights, replace = TRUE) %>%
  step_ordinalscore(agegp, tobgp, alcgp)
fit(rec, model = GBMModel)

## ----using_variables_summary, echo=FALSE--------------------------------------
df <- data.frame(
  "Specification" = c("Traditional Formula", "Design Matrix",
                      "Traditional Formula", "Design Matrix", "Recipe"),
  "Preprocessing" = factor(c("manual", "manual", "manual", "manual",
                             "automatic"), levels = c("manual", "automatic")),
  "In-line Functions" = factor(c("yes", "no", "yes", "no", "no"),
                               levels = c("no", "yes")),
  "Case Weights" = factor(c("equal", "equal", "user", "user", "user"),
                          levels = c("equal", "user")),
  "Resampling Strata" = factor(c("response", "response", "user", "user",
                                 "user"), levels = c("response", "user")),
  "Computational Overhead" = factor(c("medium", "low", "medium", "low", "high"),
                                    levels = c("high", "medium", "low")),
  check.names = FALSE
)

bg_colors <- c("orange", "blue", "green")
df[-1] <- lapply(df[-1], function(x) {
  bg_colors <- if (nlevels(x) == 2) bg_colors[c(1, 3)] else bg_colors
  cell_spec(x, color = "white", background = bg_colors[x])
})

kable(df, align = c("l", rep("c", ncol(df) - 1)), escape = FALSE,
      caption = "Table 2. Characteristics of available variable specification approaches.") %>%
  kable_styling(c("striped", "condensed"), full_width = FALSE,
                position = "center") %>%
  column_spec(1, bold = TRUE) %>%
  kableExtra::group_rows("Model Frame", 3, 4)

## ----using_responses_factor---------------------------------------------------
## Iris flowers species (3-level factor)
model_fit <- fit(Species ~ ., data = iris, model = GBMModel)
predict(model_fit) %>% head
predict(model_fit, type = "prob") %>% head

## ----using_responses_factor_binary--------------------------------------------
## Pima Indians diabetes statuses (binary factor)
data(Pima.te, package = "MASS")
data(Pima.tr, package = "MASS")

model_fit <- fit(type ~ ., data = Pima.tr, model = GBMModel)
predict(model_fit, newdata = Pima.te) %>% head
predict(model_fit, newdata = Pima.te, type = "prob") %>% head

## ----using_responses_ordered--------------------------------------------------
## Iowa City housing prices (ordered factor)
df <- within(ICHomes, {
  sale_amount <- cut(sale_amount, breaks = 3,
                     labels = c("Low", "Medium", "High"),
                     ordered_result = TRUE)
})

model_fit <- fit(sale_amount ~ ., data = df, model = GBMModel)
predict(model_fit) %>% head
predict(model_fit, type = "prob") %>% head

## ----using_responses_numeric--------------------------------------------------
## Iowa City housing prices
model_fit <- fit(sale_amount ~ ., data = ICHomes, model = GBMModel)
predict(model_fit) %>% head
predict(model_fit, type = "numeric") %>% head

## ----using_responses_matrix---------------------------------------------------
## Anscombe's multiple regression models dataset

## Numeric matrix response formula
model_fit <- fit(cbind(y1, y2, y3) ~ x1, data = anscombe, model = LMModel)
predict(model_fit) %>% head

## ----using_responses_matrix_recipe--------------------------------------------
## Numeric matrix response recipe
## Defined in a recipe formula
rec <- recipe(y1 + y2 + y3 ~ x1, data = anscombe)

## Defined within a data frame
df <- within(anscombe, {
  y <- cbind(y1, y2, y3)
  remove(y1, y2, y3)
})
rec <- recipe(y ~ x1, data = df)
model_fit <- fit(rec, model = LMModel)
predict(model_fit) %>% head

## ----using_responses_surv, results="hide"-------------------------------------
## Survival response formula
library(survival)

fit(Surv(time, status) ~ ., data = veteran, model = GBMModel)

## ----using_responses_surv_recipe, results="hide"------------------------------
## Survival response recipe
## Defined in a recipe formula
rec <- recipe(time + status ~ ., data = veteran) %>%
  role_surv(time = time, event = status)

## Defined within a data frame
df <- within(veteran, {
  y <- Surv(time, status)
  remove(time, status)
})
rec <- recipe(y ~ ., data = df)
fit(rec, model = GBMModel)

## ----using_performance_function-----------------------------------------------
## Survival performance metrics

## Observed responses
obs <- response(surv_fit, newdata = surv_test)

## Predicted survival means
pred_means <- predict(surv_fit, newdata = surv_test)
performance(obs, pred_means)

## Predicted survival probabilities
pred_probs <- predict(surv_fit, newdata = surv_test, times = surv_times, type = "prob")
performance(obs, pred_probs)

## Predicted survival events
pred_events <- predict(surv_fit, newdata = surv_test, times = surv_times)
performance(obs, pred_events)

## ----using_performance_function_metrics, eval=FALSE---------------------------
# ## Single metric function
# performance(obs, pred_means, metrics = cindex)
# 
# ## Single metric function name
# performance(obs, pred_means, metrics = "cindex")
# 
# ## List of metric functions
# performance(obs, pred_means, metrics = c(cindex, rmse, rmsle))
# 
# ## Named list of metric functions
# performance(obs, pred_means,
#             metrics = c("CIndex" = cindex, "RMSE" = rmse, "RMSLE" = rmsle))

## ----using_performance_function_cutoff----------------------------------------
## User-specified survival probability metrics
performance(obs, pred_probs, metrics = c(sensitivity, specificity), cutoff = 0.7)

## ----using_metrics_functions--------------------------------------------------
## Metric functions for survival means
cindex(obs, pred_means)

rmse(obs, pred_means)

rmsle(obs, pred_means)

## Metric functions for survival probabilities
sensitivity(obs, pred_probs)

specificity(obs, pred_probs)

## ----using_metrics_metricinfo-------------------------------------------------
## All available metrics
metricinfo() %>% names

## ----using_metrics_metricinfo_cindex------------------------------------------
## Metric-specific information
metricinfo(cindex)

## ----using_metrics_cindex-----------------------------------------------------
cindex

## ----using_metrics_metricinfo_type--------------------------------------------
## Metrics for observed and predicted response variable types
metricinfo(Surv(0)) %>% names

metricinfo(Surv(0), numeric(0)) %>% names

metricinfo(Surv(0), SurvEvents(0)) %>% names

metricinfo(Surv(0), SurvProbs(0)) %>% names

## Identify survival-specific metrics
metricinfo(Surv(0), auc, cross_entropy, cindex) %>% names

## ----using_metrics_metricinfo_response----------------------------------------
## Metrics for observed and predicted responses from model fits
metricinfo(obs, pred_means) %>% names

metricinfo(obs, pred_probs) %>% names

## ----using_metrics_conf, echo=FALSE-------------------------------------------
conf <- matrix(c("True Negative (TN)", "False Positive (FP)",
                 "False Negative (FN)", "True Positive (TP)"),
               2, 2,
               dimnames = list("Predicted Response" = c("Negative", "Positive"),
                               "Observed Response" = c("Negative", "Positive")))
kable(conf,
      caption = "Table 4. Confusion matrix of observed and predicted response classifications.",
      align = c("c", "c")) %>%
  kable_styling(full_width = FALSE, position = "center") %>%
  add_header_above(c("Predicted Response" = 1, "Observed Response" = 2))

## ----using_metrics_conf_surv, echo=FALSE--------------------------------------
conf <- matrix(
  c(
    "$TN = \\Pr(\\hat{S}(t) \\ge \\text{cutoff} \\cap T \\gt t)$",
    "$FP = \\Pr(\\hat{S}(t) \\lt \\text{cutoff} \\cap T \\gt t)$",
    "$FN = \\Pr(\\hat{S}(t) \\ge \\text{cutoff} \\cap T \\le t)$",
    "$TP = \\Pr(\\hat{S}(t) \\lt \\text{cutoff} \\cap T \\le t)$"
  ),
  2, 2,
  dimnames = list(
    "Predicted Response" = c("Non-Event", "Event"),
    "Observed Response" = c("Non-Event", "Event")
  )
)
kable(conf,
      caption = "Table 5. Confusion matrix of observed and predicted survival response classifications.",
      align = c("c", "c"),
      escape = FALSE) %>%
  kable_styling(full_width = FALSE, position = "center") %>%
  add_header_above(c("Predicted Response" = 1, "Observed Response" = 2))

## ----using_resample_control---------------------------------------------------
## Control structures for K-fold cross-validation

## Prediction of survival means
surv_means_control <- CVControl(folds = 5, repeats = 3, seed = 123)

## Prediction of survival probabilities
surv_probs_control <- CVControl(folds = 5, repeats = 3, seed = 123) %>%
  set_predict(times = surv_times)

## User-specification of the default control structure
MachineShop::settings(control = CVControl(folds = 5, seed = 123))

## Package default
# MachineShop::settings(reset = "control")

## ----using_resample_parallel--------------------------------------------------
## Register multiple cores for parallel computations
library(doParallel)
registerDoParallel(cores = 2)

## ----using_resample_function--------------------------------------------------
## Resample estimation for survival means and probabilities
(res_means <- resample(surv_fo, data = surv_train, model = GBMModel,
                       control = surv_means_control))

(res_probs <- resample(surv_fo, data = surv_train, model = GBMModel,
                       control = surv_probs_control))

## ----using_resample_summary---------------------------------------------------
## Summary of survival means metric
summary(res_means)

## Summary of survival probability metrics
summary(res_probs)

## ----using_resample_summary_performance---------------------------------------
## Resample-specific metrics
metricinfo(res_means) %>% names

## User-specified survival means metrics
summary(performance(res_means, metrics = c(cindex, rmse)))

## ----using_resample_summary_stats---------------------------------------------
## User-defined statistics function
percentiles <- function(x) quantile(x, probs = c(0.25, 0.50, 0.75))
summary(res_means, stats = percentiles)

## User-defined list of statistics functions
summary(res_means, stats = c(Mean = mean, Percentile = percentiles))

## ----using_resample_plots-----------------------------------------------------
## Libraries for plot annotation and fomatting
library(ggplot2)
library(gridExtra)

## Individual ggplots
p1 <- plot(res_means)
p2 <- plot(res_means, type = "density")
p3 <- plot(res_means, type = "errorbar")
p4 <- plot(res_means, type = "violin")

## Grid of plots
grid.arrange(p1, p2, p3, p4, nrow = 2)

## ----using_resample_strata, results="hide"------------------------------------
## Model frame with response variable stratification
mf <- ModelFrame(surv_fo, data = surv_train, strata = surv_train$y)
resample(mf, model = GBMModel)

## Recipe with response variable stratification
rec <- recipe(y ~ ., data = surv_train) %>%
  role_case(stratum = y)
resample(rec, model = GBMModel)

## ----using_resample_dynamic, results="hide"-----------------------------------
## Dynamic model parameter k = log number of training set observations
resample(surv_fo, data = surv_train, model = CoxStepAICModel(k = .(log(nobs))))

## ----using_resample_comparisons-----------------------------------------------
## Resample estimation
res1 <- resample(surv_fo, data = surv_train, model = GBMModel(n.trees = 25),
                 control = surv_means_control)
res2 <- resample(surv_fo, data = surv_train, model = GBMModel(n.trees = 50),
                 control = surv_means_control)
res3 <- resample(surv_fo, data = surv_train, model = GBMModel(n.trees = 100),
                 control = surv_means_control)

## Combine resample output for comparison
(res <- c(GBM1 = res1, GBM2 = res2, GBM3 = res3))

summary(res)

plot(res)

## ----using_resample_diff------------------------------------------------------
## Pairwise model comparisons
(res_diff <- diff(res))

summary(res_diff)

plot(res_diff)

## ----using_resample_diff_test-------------------------------------------------
t.test(res_diff)

## ----using_analyses_vi--------------------------------------------------------
## Predictor variable importance
(vi <- varimp(surv_fit, method = "model"))

plot(vi)

## ----using_analysis_vi_info---------------------------------------------------
SVMModel

modelinfo(SVMModel)[[1]]$varimp

## ----using_analysis_vi_permute------------------------------------------------
## Permutation-based variable importance
varimp(surv_fit)

## ----using_analysis_rfe-------------------------------------------------------
## Recursive feature elimination
(surv_rfe <- rfe(surv_fo, data = surv_train, model = GBMModel,
                 control = surv_means_control))
rfe_summary <- summary(surv_rfe)
rfe_summary$terms[rfe_summary$selected]

## ----using_analyses_pd, results = "hide"--------------------------------------
## Partial dependence plots
pd <- dependence(surv_fit, select = c(thickness, age))
plot(pd)

## ----using_analyses_pd_data, results = "hide"---------------------------------
pd <- dependence(surv_fit, data = surv_test, select = thickness, n = 20,
                 intervals = "quantile")
plot(pd)

## ----using_analyses_cal, results="hide"---------------------------------------
## Binned calibration curves
cal <- calibration(res_probs, breaks = 10)
plot(cal, se = TRUE)

## ----using_analyses_cal_smoothed, results="hide"------------------------------
## Smoothed calibration curves
cal <- calibration(res_probs, breaks = NULL)
plot(cal)

## ----using_analyses_conf------------------------------------------------------
## Confusion matrices
(conf <- confusion(res_probs, cutoff = 0.7))

## ----using_analyses_conf_plot, results="hide"---------------------------------
plot(conf)

## ----using_analyses_conf_summary----------------------------------------------
## Summary performance metrics
summary(conf)

## ----using_analyses_conf_performance------------------------------------------
## Confusion matrix-specific metrics
metricinfo(conf) %>% names

## User-specified metrics
performance(conf, metrics = c("Accuracy" = accuracy,
                              "Sensitivity" = sensitivity,
                              "Specificity" = specificity))

## ----using_analyses_roc-------------------------------------------------------
## ROC curves
roc <- performance_curve(res_probs)
plot(roc, diagonal = TRUE)

## ----using_analyses_roc_cutoffs-----------------------------------------------
plot(roc, type = "cutoffs")

## ----using_analyses_roc_auc---------------------------------------------------
auc(roc)

## ----using_analyses_pr--------------------------------------------------------
## Precision recall curves
pr <- performance_curve(res_probs, metrics = c(precision, recall))
plot(pr)

## ----using_analyses_pr_auc----------------------------------------------------
auc(pr)

## ----using_analyses_lift------------------------------------------------------
## Lift curves
lf <- lift(res_probs)
plot(lf, find = 0.75)

## ----using_stategies_TunedInput1, eval=FALSE----------------------------------
# ## Preprocessing recipe with PCA steps
# pca_rec <- recipe(y ~ ., data = surv_train) %>%
#   role_case(stratum = y) %>%
#   step_center(all_predictors()) %>%
#   step_scale(all_predictors()) %>%
#   step_pca(all_predictors(), id = "PCA")
# 
# ## Tuning grid of number of PCA components
# pca_grid <- expand_steps(
#   PCA = list(num_comp = 1:3)
# )
# 
# ## Tuning specification
# tun_rec <- TunedInput(pca_rec, grid = pca_grid)

## ----using_stategies_TunedInput2, eval=FALSE----------------------------------
# ## Input-tuned model fit and final trained model
# model_fit <- fit(tun_rec, model = GBMModel)
# as.MLModel(model_fit)
# #> --- MLModel object ----------------------------------------------------------
# #>
# #> Model name: GBMModel
# #> Label: Trained Generalized Boosted Regression
# #> Package: gbm
# #> Response types: factor, numeric, PoissonVariate, Surv
# #> Case weights support: TRUE
# #> Missing case removal: response
# #> Tuning grid: TRUE
# #> Variable importance: TRUE
# #>
# #> Parameters:
# #> List of 5
# #>  $ n.trees          : num 100
# #>  $ interaction.depth: num 1
# #>  $ n.minobsinnode   : num 10
# #>  $ shrinkage        : num 0.1
# #>  $ bag.fraction     : num 0.5
# #>
# #> === $TrainingStep1 ==========================================================
# #> === TrainingStep object ===
# #>
# #> Optimization method: Grid Search
# #> TunedModelRecipe log:
# #> # A tibble: 3 × 4
# #>   name          selected params$PCA$num_comp metrics$`C-Index`
# #>   <chr>         <lgl>                  <int>             <dbl>
# #> 1 ModelRecipe.1 TRUE                       1             0.740
# #> 2 ModelRecipe.2 FALSE                      2             0.720
# #> 3 ModelRecipe.3 FALSE                      3             0.725
# #>
# #> Selected row: 1
# #> Metric: C-Index = 0.7399641

## ----using_strategies_SelectedInput1, eval=FALSE------------------------------
# ## Preprocessing recipe without PCA steps
# rec1 <- recipe(y ~ sex + age + year + thickness + ulcer, data = surv_train) %>%
#   role_case(stratum = y)
# rec2 <- recipe(y ~ sex + age + year, data = surv_train) %>%
#   role_case(stratum = y)
# 
# ## Selection among recipes with and without PCA steps
# sel_rec <- SelectedInput(
#   rec1,
#   rec2,
#   TunedInput(pca_rec, grid = pca_grid)
# )

## ----using_strategies_SelectedInput2, eval=FALSE------------------------------
# ## Input-selected model fit and model
# model_fit <- fit(sel_rec, model = GBMModel)
# as.MLModel(model_fit)
# #> --- MLModel object ----------------------------------------------------------
# #>
# #> Model name: GBMModel
# #> Label: Trained Generalized Boosted Regression
# #> Package: gbm
# #> Response types: factor, numeric, PoissonVariate, Surv
# #> Case weights support: TRUE
# #> Missing case removal: response
# #> Tuning grid: TRUE
# #> Variable importance: TRUE
# #>
# #> Parameters:
# #> List of 5
# #>  $ n.trees          : num 100
# #>  $ interaction.depth: num 1
# #>  $ n.minobsinnode   : num 10
# #>  $ shrinkage        : num 0.1
# #>  $ bag.fraction     : num 0.5
# #>
# #> === $TrainingStep1 ==========================================================
# #> === TrainingStep object ===
# #>
# #> Optimization method: Grid Search
# #> SelectedModelRecipe log:
# #> # A tibble: 3 × 4
# #>   name             selected params$id  metrics$`C-Index`
# #>   <chr>            <lgl>    <chr>                  <dbl>
# #> 1 ModelRecipe.1    FALSE    input.azpi             0.761
# #> 2 ModelRecipe.2    FALSE    input.aLHo             0.643
# #> 3 TunedModelRecipe TRUE     input.W4KN             0.796
# #>
# #> Selected row: 3
# #> Metric: C-Index = 0.7960841
# #>
# #> === $TrainingStep2 ==========================================================
# #> === TrainingStep object ===
# #>
# #> Optimization method: Grid Search
# #> TunedModelRecipe log:
# #> # A tibble: 3 × 4
# #>   name          selected params$PCA$num_comp metrics$`C-Index`
# #>   <chr>         <lgl>                  <int>             <dbl>
# #> 1 ModelRecipe.1 TRUE                       1             0.740
# #> 2 ModelRecipe.2 FALSE                      2             0.720
# #> 3 ModelRecipe.3 FALSE                      3             0.725
# #>
# #> Selected row: 1
# #> Metric: C-Index = 0.7399641

## ----using_strategies_SelectedInput3, eval=FALSE------------------------------
# ## Traditional formulas
# fo1 <- y ~ sex + age + year + thickness + ulcer
# fo2 <- y ~ sex + age + year
# 
# ## Selection among formulas
# sel_fo <- SelectedInput(fo1, fo2, data = surv_train)
# 
# ## Input-selected model fit and final trained model
# model_fit <- fit(sel_fo, model = GBMModel)
# as.MLModel(model_fit)

## ----using_strategies_SelectedInput4, eval=FALSE------------------------------
# ## Different combinations of inputs and models
# sel_mfo <- SelectedInput(
#   ModelSpecification(fo1, data = surv_train, model = CoxModel),
#   ModelSpecification(fo2, data = surv_train, model = GBMModel)
# )
# 
# ## Input-selected model fit and final trained model
# model_fit <- fit(sel_mfo)
# as.MLModel(model_fit)

## ----using_strategies_tune, eval=FALSE----------------------------------------
# ## Tune over automatic grid of model parameters
# model_fit <- fit(surv_fo, data = surv_train,
#                  model = TunedModel(
#                    GBMModel,
#                    grid = 3,
#                    control = surv_means_control,
#                    metrics = c("CIndex" = cindex, "RMSE" = rmse)
#                  ))
# (trained_model <- as.MLModel(model_fit))
# #> --- MLModel object ----------------------------------------------------------
# #>
# #> Model name: GBMModel
# #> Label: Trained Generalized Boosted Regression
# #> Package: gbm
# #> Response types: factor, numeric, PoissonVariate, Surv
# #> Case weights support: TRUE
# #> Missing case removal: response
# #> Tuning grid: TRUE
# #> Variable importance: TRUE
# #>
# #> Parameters:
# #> List of 5
# #>  $ n.trees          : int 50
# #>  $ interaction.depth: int 1
# #>  $ n.minobsinnode   : num 10
# #>  $ shrinkage        : num 0.1
# #>  $ bag.fraction     : num 0.5
# #>
# #> === $TrainingStep1 ==========================================================
# #> === TrainingStep object ===
# #>
# #> Optimization method: Grid Search
# #> TunedModel log:
# #> # A tibble: 9 × 4
# #>   name       selected params$n.trees $interaction.depth metrics$CIndex  $RMSE
# #>   <chr>      <lgl>             <int>              <int>          <dbl>  <dbl>
# #> 1 GBMModel.1 TRUE                 50                  1          0.765  3869.
# #> 2 GBMModel.2 FALSE                50                  2          0.750  5450.
# #> 3 GBMModel.3 FALSE                50                  3          0.739  7877.
# #> 4 GBMModel.4 FALSE               100                  1          0.746  3919.
# #> 5 GBMModel.5 FALSE               100                  2          0.730  5262.
# #> 6 GBMModel.6 FALSE               100                  3          0.720 10596.
# #> 7 GBMModel.7 FALSE               150                  1          0.726  4167.
# #> 8 GBMModel.8 FALSE               150                  2          0.712  5428.
# #> 9 GBMModel.9 FALSE               150                  3          0.699 16942.
# #>
# #> Selected row: 1
# #> Metric: CIndex = 0.7647633

## ----using_strategies_tune_grid, eval=FALSE-----------------------------------
# ## Tune over randomly sampled grid points
# fit(surv_fo, data = surv_train,
#     model = TunedModel(
#       GBMModel,
#       grid = TuningGrid(size = 100, random = 10),
#       control = surv_means_control
#     ))
# 
# ## Tune over user-specified grid points
# fit(surv_fo, data = surv_train,
#     model = TunedModel(
#       GBMModel,
#       grid = expand_params(n.trees = c(25, 50, 100),
#                            interaction.depth = 1:3),
#       control = surv_means_control
#     ))

## ----using_strategies_tune_summary, eval=FALSE--------------------------------
# summary(trained_model)
# #> --- $TrainingStep1 ----------------------------------------------------------
# #> # A tibble: 9 × 4
# #>   name       selected params$n.trees $interaction.depth metrics$CIndex  $RMSE
# #>   <chr>      <lgl>             <int>              <int>          <dbl>  <dbl>
# #> 1 GBMModel.1 TRUE                 50                  1          0.765  3869.
# #> 2 GBMModel.2 FALSE                50                  2          0.750  5450.
# #> 3 GBMModel.3 FALSE                50                  3          0.739  7877.
# #> 4 GBMModel.4 FALSE               100                  1          0.746  3919.
# #> 5 GBMModel.5 FALSE               100                  2          0.730  5262.
# #> 6 GBMModel.6 FALSE               100                  3          0.720 10596.
# #> 7 GBMModel.7 FALSE               150                  1          0.726  4167.
# #> 8 GBMModel.8 FALSE               150                  2          0.712  5428.
# #> 9 GBMModel.9 FALSE               150                  3          0.699 16942.
# 
# performance(trained_model) %>% lapply(summary)
# #> $TrainingStep1
# #> , , Metric = CIndex
# #>
# #>             Statistic
# #> Model             Mean    Median         SD       Min       Max NA
# #>   GBMModel.1 0.7647633 0.7630332 0.05151443 0.6497797 0.8632075  0
# #>   GBMModel.2 0.7502765 0.7549020 0.04636802 0.6688742 0.8349057  0
# #>   GBMModel.3 0.7388766 0.7500000 0.04594197 0.6621622 0.8197425  0
# #>   GBMModel.4 0.7463392 0.7601810 0.05087334 0.6497797 0.8160377  0
# #>   GBMModel.5 0.7301734 0.7345133 0.04361416 0.6621622 0.7847222  0
# #>   GBMModel.6 0.7199483 0.7122642 0.05168649 0.6216216 0.7889908  0
# #>   GBMModel.7 0.7263693 0.7385321 0.05069831 0.6351351 0.7896996  0
# #>   GBMModel.8 0.7121314 0.7169811 0.05331152 0.6199095 0.7847222  0
# #>   GBMModel.9 0.6994678 0.7156863 0.05314909 0.6081081 0.7639485  0
# #>
# #> , , Metric = RMSE
# #>
# #>             Statistic
# #> Model             Mean    Median        SD      Min       Max NA
# #>   GBMModel.1  3868.629  3321.249  1431.839 2123.875  7401.789  0
# #>   GBMModel.2  5450.093  4741.129  2438.173 2385.426 11695.273  0
# #>   GBMModel.3  7877.487  5827.739  5036.837 4192.463 20009.854  0
# #>   GBMModel.4  3919.126  3535.514  1454.999 2373.413  7558.280  0
# #>   GBMModel.5  5262.122  5052.969  1900.353 2280.193  8212.727  0
# #>   GBMModel.6 10595.888  8308.733  7057.327 4996.420 28610.367  0
# #>   GBMModel.7  4167.128  3789.596  1115.371 2587.440  6701.087  0
# #>   GBMModel.8  5427.518  6129.309  2313.742 2039.894  9212.369  0
# #>   GBMModel.9 16941.696 10865.160 16841.774 4914.743 71150.177  0

## ----using_strategies_tune_plot, eval=FALSE-----------------------------------
# plot(trained_model, type = "line")
# #> $TrainStep1

## ----using_strategies_tune_png, echo=FALSE------------------------------------
knitr::include_graphics("img/using_strategies_tune_plot-1.png")

## ----using_strategies_select, results="hide", eval=FALSE----------------------
# ## Model interface for model selection
# sel_model <- SelectedModel(
#   expand_model(GBMModel, n.trees = c(50, 100), interaction.depth = 1:2),
#   GLMNetModel(lambda = 0.01),
#   CoxModel,
#   SurvRegModel
# )
# 
# ## Fit the selected model
# fit(surv_fo, data = surv_train, model = sel_model)

## ----using_strategies_select_tune, results="hide", eval=FALSE-----------------
# ## Model interface for selection among tuned models
# sel_tun_model <- SelectedModel(
#   TunedModel(GBMModel, control = surv_means_control),
#   TunedModel(GLMNetModel, control = surv_means_control),
#   TunedModel(CoxModel, control = surv_means_control)
# )
# 
# ## Fit the selected tuned model
# fit(surv_fo, data = surv_train, model = sel_tun_model)

## ----using_strategies_ensembles, eval=FALSE-----------------------------------
# ## Stacked regression
# stackedmodel <- StackedModel(CoxModel, CForestModel, GLMBoostModel)
# res_stacked <- resample(surv_fo, data = surv_train, model = stackedmodel)
# summary(res_stacked)
# #>          Statistic
# #> Metric         Mean   Median        SD       Min       Max NA
# #>   C-Index 0.7294869 0.762963 0.1275484 0.5194805 0.8556701  0
# 
# ## Super learner
# supermodel <- SuperModel(CoxModel, CForestModel, GLMBoostModel,
#                          model = GBMModel)
# res_super <- resample(surv_fo, data = surv_train, model = supermodel)
# summary(res_super)
# #>          Statistic
# #> Metric         Mean    Median        SD       Min       Max NA
# #>   C-Index 0.7534803 0.8325472 0.1243104 0.5748899 0.8454545  0

## ----using_strategies_methods, eval=FALSE-------------------------------------
# ## Preprocessing recipe with PCA steps
# pca_rec <- recipe(y ~ ., data = surv_train) %>%
#   role_case(stratum = y) %>%
#   step_center(all_predictors()) %>%
#   step_scale(all_predictors()) %>%
#   step_pca(all_predictors(), id = "PCA")
# 
# ## Tuning grid of number of PCA components
# pca_grid <- expand_steps(
#   PCA = list(num_comp = 1:3)
# )
# 
# ## Input specification
# tun_rec <- TunedInput(pca_rec, grid = pca_grid)
# 
# ## Model specification
# sel_model <- SelectedModel(
#   GBMModel,
#   TunedModel(GBMModel),
#   StackedModel(CoxModel, TunedModel(CForestModel), TunedModel(GBMModel))
# )
# 
# ## Model fit and final trained model
# model_fit <- fit(tun_rec, model = sel_model)
# as.MLModel(model_fit)

## ----using_strategies_dag, echo = FALSE, out.width = "100%"-------------------
knitr::include_graphics("img/FigModelDAG.png")

## ----using_strategies_nestedcv, echo = FALSE, out.width = "100%"--------------
knitr::include_graphics("img/FigNestedCV.png")

## ----using_strategies_methods1, eval=FALSE------------------------------------
# #> === $TrainingStep1 ==========================================================
# #> === TrainingStep object ===
# #>
# #> Optimization method: Grid Search
# #> TunedModelRecipe log:
# #> # A tibble: 3 × 4
# #>   name          selected params$PCA$num_comp metrics$`C-Index`
# #>   <chr>         <lgl>                  <int>             <dbl>
# #> 1 ModelRecipe.1 TRUE                       1             0.741
# #> 2 ModelRecipe.2 FALSE                      2             0.730
# #> 3 ModelRecipe.3 FALSE                      3             0.702
# #>
# #> Selected row: 1
# #> Metric: C-Index = 0.7405534

## ----using_strategies_methods2, eval=FALSE------------------------------------
# #> === $TrainingStep2 ==========================================================
# #> === TrainingStep object ===
# #>
# #> Optimization method: Grid Search
# #> SelectedModel log:
# #> # A tibble: 3 × 4
# #>   name         selected params$id  metrics$`C-Index`
# #>   <chr>        <lgl>    <chr>                  <dbl>
# #> 1 GBMModel     TRUE     model.1mQk             0.740
# #> 2 TunedModel   FALSE    model.hRF5             0.735
# #> 3 StackedModel FALSE    model.CYj0             0.653
# #>
# #> Selected row: 1
# #> Metric: C-Index = 0.7399641

## ----using_strategies_methods0, eval=FALSE------------------------------------
# #> --- MLModel object ----------------------------------------------------------
# #>
# #> Model name: GBMModel
# #> Label: Trained Generalized Boosted Regression
# #> Package: gbm
# #> Response types: factor, numeric, PoissonVariate, Surv
# #> Case weights support: TRUE
# #> Missing case removal: response
# #> Tuning grid: TRUE
# #> Variable importance: TRUE
# #>
# #> Parameters:
# #> List of 5
# #>  $ n.trees          : num 100
# #>  $ interaction.depth: num 1
# #>  $ n.minobsinnode   : num 10
# #>  $ shrinkage        : num 0.1
# #>  $ bag.fraction     : num 0.5

## ----eval=FALSE---------------------------------------------------------------
# ## Generalization performance of the modeling strategy
# resample(tun_rec, model = sel_model)

## ----using_settings-----------------------------------------------------------
## Change settings
presets <- settings(control = "BootControl", grid = 10)

## View one setting
settings("control")

## View multiple settings
settings("control", "grid")

## Restore the previous settings
settings(presets)

## ----using_extensions_mlmodel-------------------------------------------------
## Logistic regression model extension
LogisticModel <- MLModel(
  name = "LogisticModel",
  label = "Logistic Model",
  response_types = "binary",
  weights = TRUE,
  fit = function(formula, data, weights, ...) {
    glm(formula, data = as.data.frame(data), weights = weights,
        family = binomial, ...)
  },
  predict = function(object, newdata, ...) {
    predict(object, newdata = as.data.frame(newdata), type = "response")
  },
  varimp = function(object, ...) {
    pchisq(coef(object)^2 / diag(vcov(object)), 1)
  }
)

## ----using_extensions_mlmetric------------------------------------------------
## F2 score metric extension
f2_score <- MLMetric(
  function(observed, predicted, ...) {
    f_score(observed, predicted, beta = 2, ...)
  },
  name = "f2_score",
  label = "F2 Score",
  maximize = TRUE
)

## ----using_extensions_usage---------------------------------------------------
## Logistic regression analysis
data(Pima.tr, package = "MASS")
res <- resample(type ~ ., data = Pima.tr, model = LogisticModel)
summary(performance(res, metric = f2_score))

## ----reference_models, echo = FALSE-------------------------------------------
library(MachineShop)

info <- modelinfo()
types <- c("binary" = "b", "factor" = "f", "matrix" = "m", "numeric" = "n",
           "ordered" = "o", "Surv" = "S")
x <- lapply(names(info), function(modelname) {
  c(modelname, ifelse(names(types) %in% info[[modelname]]$response_types, types, NA))
})
df <- as.data.frame(do.call(rbind, x), stringsAsFactors = FALSE)
names(df) <- c("Function", names(types))
rdoc_names <- sapply(df$Function, function(name) {
  switch(name,
    "CoxStepAICModel" = "CoxModel",
    "GLMStepAICModel" = "GLMModel",
    "PDAModel" = "FDAModel",
    "SurvRegStepAICModel" = "SurvRegModel",
    "SVMANOVAModel" = "SVMModel",
    "SVMBesselModel" = "SVMModel",
    "SVMLaplaceModel" = "SVMModel",
    "SVMLinearModel" = "SVMModel",
    "SVMPolyModel" = "SVMModel",
    "SVMRadialModel" = "SVMModel",
    "SVMSplineModel" = "SVMModel",
    "SVMTanhModel" = "SVMModel",
    "XGBDARTModel" = "XGBModel",
    "XGBLinearModel" = "XGBModel",
    "XGBTreeModel" = "XGBModel",
    name
  )
})

toString2 <- function(x) toString(na.omit(x))
df_classes <- data.frame(
  Function = rdoc_url(df$Function, rdoc_names),
  Label = sapply(info, getElement, name = "label"),
  Categorical = apply(df[c("binary", "factor", "ordered")], 1, toString2),
  Continuous = apply(df[c("matrix", "numeric")], 1, toString2),
  Survival = apply(df["Surv"], 1, toString2)
)
names(df_classes)[3:5] <- paste0(names(df_classes)[3:5], footnote_marker_number(1:3))

kable(df_classes,
      caption = "Package-supplied model constructor functions and supported response variable types.",
      align = c("l", "l", "c", "c", "c"), row.names = FALSE,
      escape = FALSE) %>%
  kable_styling(c("striped", "condensed"), full_width = FALSE, position = "center") %>%
  add_header_above(c(" " = 1, " " = 1, "Response Variable Types" = 3)) %>%
  footnote(number = c("b = binary factor, f = factor, o = ordered factor",
                      "m = matrix, n = numeric",
                      "S = Surv"))

## ----reference_metrics, echo=FALSE--------------------------------------------
library(MachineShop)

f <- function(x) {
  types <- x$response_types
  
  is_type <- function(observed, predicted) {
    any(types$observed == observed & types$predicted == predicted)
  }
  
  categorical <- if (is_type("factor", "matrix")) "f" else
    if (is_type("factor", "numeric")) "b" else NULL
  if (is_type("ordered", "ordered")) categorical <- c(categorical, "o")

  continuous <- NULL
  if (is_type("matrix", "matrix")) continuous <- "m"
  if (is_type("numeric", "numeric")) continuous <- c(continuous, "n")
  
  survival <- NULL
  if (any(mapply(is_type, "Surv", c("numeric", "SurvEvents", "SurvProbs")))) {
    survival <- "S"
  }

  data.frame(
    Label = x$label,
    Categorical = toString(categorical),
    Continuous = toString(continuous),
    Survival = toString(survival)
  )
}

info <- metricinfo()
df <- cbind("Function" = rdoc_url(names(info), "metrics"),
            do.call(rbind, lapply(info, f)))
names(df)[3:5] <- paste0(names(df)[3:5], footnote_marker_number(1:3))

kable(df, caption = "Package-supplied performance metric functions and supported response variable types.",
      align = c("l", "l", "c", "c", "c"), row.names = FALSE,
      escape = FALSE) %>%
  kable_styling(c("striped", "condensed"), full_width = FALSE, position = "center") %>%
  add_header_above(c(" " = 1, " " = 1, "Response Variable Types" = 3)) %>%
  footnote(number = c("b = binary factor, f = factor, o = ordered factor",
                      "m = matrix, n = numeric",
                      "S = Surv"))

## Variable Importance

test_that("test permuation-based varimp", {
  skip_if_not(TEST_ALL)
  with_parallel({

    library(recipes)

    fo <- sale_amount ~ .
    data <- ICHomes
    model <- SVMRadialModel

    test_varimp <- function(x, ...) {
      set.seed(123)
      vi <- varimp(x, ..., method = "permute", scale = FALSE)
      expect_true(is(vi, "VariableImportance") && !any(is.na(vi)))
      vi
    }

    test_equal <- function(x, y) {
      expect_true(all(x == y))
    }

    test_block <- function(x) {
      test_varimp(x)
      test_varimp(x, metric = mse)
      test_varimp(x, compare = "/")

      vi1 <- test_varimp(x, samples = 10)
      vi2 <- test_varimp(x, samples = 10, stats = c(Mean = mean, SD = sd))
      test_equal(vi1, vi2[, 1])
      test_equal(
        vi1,
        test_varimp(x, samples = 10, na.rm = FALSE)
      )
      test_equal(
        vi2,
        test_varimp(x, samples = 10, stats = c(Mean = mean, SD = sd), na.rm = FALSE)
      )

      test_equal(
        test_varimp(x, samples = 10, prop = 0.25),
        test_varimp(x, samples = 10, size = nrow(ICHomes) * 0.25)
      )
    }

    fo_fit <- fit(fo, data = data, model = model)
    rec <- recipe(fo, data = data) %>%
      step_pca(all_numeric_predictors())
    rec_fit <- fit(rec, model = model)

    test_block(fo_fit)
    test_block(rec_fit)

  })
})

## Trained Models


test_that("grid tuning of models", {
  skip_if_not(TEST_TRAINING)
  with_parallel({

    models <- c(
      "AdaBagModel",
      "AdaBoostModel",
      "BARTMachineModel",
      "BlackBoostModel",
      "C50Model",
      "CForestModel",
      "EarthModel",
      "FDAModel",
      "GAMBoostModel",
      "GBMModel",
      "GLMBoostModel",
      "GLMNetModel",
      "KNNModel",
      "LDAModel",
      "MDAModel",
      "NNetModel",
      "PLSModel",
      "RandomForestModel",
      "RangerModel",
      "RFSRCModel",
      "RPartModel",
      "SVMANOVAModel",
      "SVMBesselModel",
      "SVMLaplaceModel",
      "SVMLinearModel",
      "SVMPolyModel",
      "SVMRadialModel",
      "SVMTanhModel",
      "XGBDARTModel",
      "XGBLinearModel",
      "XGBTreeModel"
    )

    control <- CVControl(folds = 3)

    test_tune_grid1 <- function(model) {
      fit(type ~ ., data = Pima.tr,
          model = TunedModel(model, grid = 1, control = control))
    }

    test_tune_grid2 <- function(model) {
      fit(type ~ ., data = Pima.tr,
          model = TunedModel(model, grid = 3, control = control))
    }

    test_tune_grid3 <- function(model) {
      fit(type ~ ., data = Pima.tr,
          model = TunedModel(model, grid = TuningGrid(size = 100, random = 10),
                             control = control))
    }

    for (model in models) {
      expect_is(test_tune_grid1(model), "MLModelFit")
      expect_is(test_tune_grid2(model), "MLModelFit")
      expect_is(test_tune_grid3(model), "MLModelFit")
    }

  })
})

## Metric and Model Information

test_that("test response types", {
  skip_if_not(TEST_ALL)
  with_parallel({

    modelinfo_names <- function(...) names(modelinfo(...))
    metricinfo_names <- function(...) names(metricinfo(...))


    ## binary response

    library(MASS)

    df <- Pima.tr
    fo <- type ~ .
    model_fit <- fit(fo, df, model = GBMModel)
    obs <- response(model_fit)
    pred_prob <- predict(model_fit, type = "prob")
    res <- resample(fo, df, model = GBMModel)

    expect_setequal(metricinfo_names(obs), metricinfo_names(factor(1:2)))
    expect_setequal(metricinfo_names(res), metricinfo_names(obs, pred_prob))
    expect_setequal(metricinfo_names(res, accuracy, auc, r2),
                    c("accuracy", "auc"))
    expect_setequal(metricinfo_names(res, "accuracy", "auc", "r2"),
                    c("accuracy", "auc"))
    expect_setequal(metricinfo_names(res, accuracy, "auc", r2),
                    c("accuracy", "auc"))

    expect_setequal(modelinfo_names(obs), modelinfo_names(factor(1:2)))
    expect_setequal(modelinfo_names(obs, CoxModel, GBMModel, GLMModel),
                    c("GBMModel", "GLMModel"))
    expect_setequal(modelinfo_names(obs, "CoxModel", "GBMModel", "GLMModel"),
                    c("GBMModel", "GLMModel"))
    expect_setequal(modelinfo_names(obs, CoxModel, "GBMModel", GLMModel),
                    c("GBMModel", "GLMModel"))


    ## categorical  response

    df <- iris
    fo <- Species ~ .
    model_fit <- fit(fo, df, model = GBMModel)
    obs <- response(model_fit)
    pred_prob <- predict(model_fit, type = "prob")
    res <- resample(fo, df, model = GBMModel)

    expect_setequal(metricinfo_names(obs), metricinfo_names(factor(1:3)))
    expect_setequal(metricinfo_names(res), metricinfo_names(obs, pred_prob))
    expect_setequal(metricinfo_names(res, accuracy, brier, r2),
                    c("accuracy", "brier"))
    expect_setequal(metricinfo_names(res, "accuracy", "brier", "r2"),
                    c("accuracy", "brier"))
    expect_setequal(metricinfo_names(res, accuracy, "brier", r2),
                    c("accuracy", "brier"))

    expect_setequal(modelinfo_names(obs), modelinfo_names(factor(1:3)))
    expect_setequal(modelinfo_names(obs, CoxModel, GBMModel, GLMNetModel),
                    c("GBMModel", "GLMNetModel"))
    expect_setequal(modelinfo_names(obs, "CoxModel", "GBMModel", "GLMNetModel"),
                    c("GBMModel", "GLMNetModel"))
    expect_setequal(modelinfo_names(obs, CoxModel, "GBMModel", GLMNetModel),
                    c("GBMModel", "GLMNetModel"))


    ## numeric response

    df <- ICHomes
    fo <- sale_amount ~ .
    model_fit <- fit(fo, df, model = GBMModel)
    obs <- response(model_fit)
    pred <- predict(model_fit)
    res <- resample(fo, df, model = GBMModel)

    expect_setequal(metricinfo_names(obs), metricinfo_names(numeric(0)))
    expect_setequal(metricinfo_names(res), metricinfo_names(obs, pred))
    expect_setequal(metricinfo_names(res, mse, r2, accuracy), c("mse", "r2"))
    expect_setequal(metricinfo_names(res, "mse", "r2", "accuracy"),
                    c("mse", "r2"))
    expect_setequal(metricinfo_names(res, mse, "r2", accuracy), c("mse", "r2"))

    expect_setequal(modelinfo_names(obs), modelinfo_names(numeric(0)))
    expect_setequal(modelinfo_names(obs, CoxModel, GBMModel, GLMModel),
                    c("GBMModel", "GLMModel"))
    expect_setequal(modelinfo_names(obs, "CoxModel", "GBMModel", "GLMModel"),
                    c("GBMModel", "GLMModel"))
    expect_setequal(modelinfo_names(obs, CoxModel, "GBMModel", GLMModel),
                    c("GBMModel", "GLMModel"))


    ## survival response

    library(MASS)

    df <- Melanoma
    fo <- Surv(time, status != 2) ~ sex + age + year + thickness + ulcer
    model_fit <- fit(fo, df, model = GBMModel)
    obs <- response(model_fit)
    pred <- predict(model_fit)
    pred_times <- predict(model_fit, times = 365 * c(2, 5, 10), type = "prob")
    res <- resample(fo, df, model = GBMModel)

    expect_setequal(metricinfo_names(obs, pred),
                    metricinfo_names(Surv(0), numeric(0)))
    expect_setequal(metricinfo_names(obs, pred_times),
                    metricinfo_names(Surv(0), SurvProbs(0)))
    expect_setequal(metricinfo_names(res), metricinfo_names(obs, pred))
    expect_setequal(metricinfo_names(res, mse, r2, accuracy), c("mse", "r2"))
    expect_setequal(metricinfo_names(res, "mse", "r2", "accuracy"),
                    c("mse", "r2"))
    expect_setequal(metricinfo_names(res, mse, "r2", accuracy), c("mse", "r2"))

    expect_setequal(modelinfo_names(obs), modelinfo_names(Surv(0)))
    expect_setequal(modelinfo_names(obs, CoxModel, GBMModel, GLMModel),
                    c("GBMModel", "CoxModel"))
    expect_setequal(modelinfo_names(obs, "CoxModel", "GBMModel", "GLMModel"),
                    c("GBMModel", "CoxModel"))
    expect_setequal(modelinfo_names(obs, CoxModel, "GBMModel", GLMModel),
                    c("GBMModel", "CoxModel"))

  })
})

## Tuning Grids

test_that("ParameterGrid construction and model fitting", {
  skip_if_not(TEST_ALL)
  with_parallel({

    library(dials)

    param_list <- list(
      num.trees = trees(),
      mtry = mtry(),
      replace = new_qual_param("logical", c(TRUE, FALSE),
                               label = c(replace = "replace"))
    )

    test_fit <- function(grid) {
      model_fit <- expect_is(
        fit(sale_amount ~ ., data = ICHomes,
            model = TunedModel(RangerModel, grid = grid)),
        "MLModelFit"
      )
      model_grid <- as.MLModel(model_fit)@steps[[1]]@log$params
      nrow(model_grid) && ncol(model_grid) == nrow(grid)
    }

    expect_error(ParameterGrid(param_list, size = -1))
    expect_error(ParameterGrid(param_list, random = 0))

    grid <- expect_s4_class(ParameterGrid(param_list), "ParameterGrid")
    expect_true(test_fit(grid))

    grid <- expect_s4_class(
      ParameterGrid(param_list, size = 0),
      "ParameterGrid"
    )
    expect_true(test_fit(grid))

    grid <- expect_s4_class(
      ParameterGrid(param_list, size = 2),
      "ParameterGrid"
    )
    expect_true(test_fit(grid))

    grid <- expect_s4_class(
      ParameterGrid(param_list, size = c(3, 0, 2)),
      "ParameterGrid"
    )
    expect_true(test_fit(grid))

    grid <- expect_s4_class(
      ParameterGrid(param_list, size = 2, random = 5),
      "ParameterGrid"
    )
    expect_true(test_fit(grid))

  })
})

## ModelSpecification class

test_that("test ModelSpecification", {
  skip_if_not(TEST_ALL)
  with_parallel({

    library(recipes)

    df <- ICHomes
    fo <- sale_amount ~ .

    test_modelspec <- function(...) {
      ModelSpecification(..., model = "GLMModel") %>%
        expect_is("ModelSpecification") %>%
        fit %>%
        expect_is("MLModelFit")
    }

    test_modelspec(fo, data = df)
    test_modelspec(
      x = model.matrix(fo, df)[, -1],
      y = model.response(model.frame(fo, df))
    )
    test_modelspec(ModelFrame(fo, data = df))
    test_modelspec(recipe(fo, data = df))

    modelspec <- ModelSpecification(
      input = TunedInput(
        recipe(fo, data = df) %>%
          step_pca(all_numeric_predictors(), id = "pca"),
        grid = expand_steps(pca = list(num_comp = 6:7))
      ),
      model = TunedModel(
        GBMModel,
        grid = c(n.trees = 3, interaction.depth = 3)
      )
    )

    expect_true(is(modelspec@grid, "tbl_df"))
    expect_true(nrow(modelspec@grid) == 18)
    expect_true(length(modelspec@grid) == 2)
    expect_true(length(modelspec@grid[[2]]) == 2)

    test_optim <- function(set, ...) {
      modelspec <- set(modelspec, ...)
      model <- as.MLModel(fit(modelspec))
      step <- model@steps[[1]]
      expect_true(step@method == modelspec@params@optim@label &&
        is(step@log$params, "tbl_df") &&
        length(step@log$params) &&
        nrow(step@log$params))
    }

    test_optim(set_optim_bayes, package = "ParBayesianOptimization")
    test_optim(set_optim_bayes, package = "rBayesianOptimization")
    test_optim(set_optim_bfgs)
    test_optim(set_optim_grid)
    test_optim(set_optim_grid, random = 10)
    test_optim(set_optim_pso, each = 3)
    test_optim(set_optim_sann)

    test_set_control <- function(name, ...) {
      params <- list(...)
      modelspec <- do.call(paste0("set_", name), c(list(modelspec), params))
      control_params <- slot(modelspec@params@control, name)
      expect_equal(control_params[names(params)], params[names(params)])
    }

  })
})

#' Parsnip Model
#'
#' Convert a model specification from the \pkg{parsnip} package to one that can
#' be used with the \pkg{MachineShop} package.
#'
#' @param object \link[parsnip:model_spec]{model specification} from the
#'   \pkg{parsnip} package.
#' @param ... tuning parameters with which to update \code{object}.
#'
#' @return \code{ParsnipModel} class object that inherits from \code{MLModel}.
#'
#' @seealso \code{\link{as.MLModel}}, \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package parsnip to run
#'
#' prsp_model <- parsnip::linear_reg(engine = "glmnet")
#'
#' model <- ParsnipModel(prsp_model, penalty = 1, mixture = 1)
#' model
#'
#' model_fit <- fit(sale_amount ~ ., data = ICHomes, model = model)
#' predict(model_fit)
#' }
#'
ParsnipModel <- function(object, ...) {

  throw(check_packages("parsnip"))

  if (missing(object)) {
    object <- parsnip::null_model() %>%
      parsnip::set_engine("parsnip")
    object$mode <- "unknown"
  } else {
    stopifnot(is(object, "model_spec"))
    object <- update(object, ...)
  }

  modes <- list(
    "censored regression" = "Surv",
    "classification" = "factor",
    "regression" = c("matrix", "numeric")
  )
  modes$unknown <- sort(unlist(modes, use.names = FALSE))
  throw(check_assignment(mode, check_match(object$mode, names(modes))))

  mlmodel <- MLModel(

    name = "ParsnipModel",
    label = sprintf(
      "%s(mode = \"%s\", engine = \"%s\")", class1(object), object$mode,
      object$engine
    ),
    packages = parsnip::required_pkgs(object),
    response_types = modes[[object$mode]],
    predictor_encoding = "model.frame",
    params = list(object = object),

    fit = function(formula, data, weights, object, ...) {
      parsnip::fit(object, formula, data = as.data.frame(formula, data = data))
    },

    predict = function(object, newdata, times, .MachineShop, ...) {
      prsp_predict <- function(...) {
        predict(object, new_data = as.data.frame(newdata), ...)
      }
      model <- .MachineShop$model
      mode <- model@params$object$mode
      if (mode == "censored regression") {
        distr <- model@params$dist
        if (is(distr, "quosure")) distr <- rlang::quo_get_expr(distr)
        if (length(times)) {
          pred_list <- prsp_predict(type = "survival", eval_time = times)$.pred
          pred_vector <- do.call(rbind, pred_list)$.pred_survival
          pred <- matrix(pred_vector, ncol = length(times), byrow = TRUE)
          SurvProbs(pred, times = times, distr = distr)
        } else {
          SurvTimes(prsp_predict(type = "time")$.pred_time, distr = distr)
        }
      } else {
        pred <- prsp_predict(type = switch(mode, "classification" = "prob"))
        select <- grepl("^.pred", names(pred))
        names(pred) <- gsub("^.pred_", "", names(pred))
        as.matrix(pred[select])
      }
    }

  )
  new("ParsnipModel", mlmodel)

}

MLModelFunction(ParsnipModel) <- NULL


update.ParsnipModel <- function(object, ...) {
  NextMethod(quote = FALSE)
}

test_that("check ParsnipModel predictions", {

  skip_if_not(TEST_MODEL_FITTING)

  library(censored)
  library(MASS)
  library(survival)

  compare_models <- function(fo, data, model1, model2, ...) {
    fit1 <- MachineShop::fit(fo, data, model = model1)
    fit2 <- MachineShop::fit(fo, data, model = model2)
    for (type in c("response", "raw")) {
      expect_equal(
        S3Part(predict(fit1, type = type, ...), strictS3 = TRUE),
        S3Part(predict(fit2, type = type, ...), strictS3 = TRUE)
      )
    }
  }

  ## numeric response
  compare_models(
    fo = sale_amount ~ .,
    data = ICHomes,
    model1 = ParsnipModel(parsnip::linear_reg()),
    model2 = GLMModel()
  )

  ## multivariate response
  compare_models(
    fo = cbind(mpg, hp, qsec) ~ .,
    data = mtcars,
    model1 = ParsnipModel(parsnip::linear_reg()),
    model2 = GLMModel()
  )

  ## dichotomous response
  compare_models(
    fo = type ~ .,
    data = Pima.tr,
    model1 = ParsnipModel(parsnip::logistic_reg()),
    model2 = GLMModel()
  )

  ## multinomial response
  compare_models(
    fo = Species ~ .,
    data = iris,
    model1 = ParsnipModel(parsnip::multinom_reg()),
    model2 = GLMModel()
  )

  ## survival response
  fo <- Surv(time, status) ~ .
  data <- veteran

  model <- ParsnipModel(parsnip::survival_reg(), dist = "weibull")

  model_fit <- MachineShop::fit(fo, data = data, model = model)
  pred <- predict(model_fit)
  expect_s4_class(pred, "SurvTimes")
  expect_vector(pred, size = nrow(veteran))

  compare_models(
    fo = fo,
    data = data,
    model1 = model,
    model2 = SurvRegModel(),
    times = c(100, 200)
  )

  model <- ParsnipModel(parsnip::proportional_hazards())

  compare_models(
    fo = fo,
    data = data,
    model1 = model,
    model2 = CoxModel(),
    times = c(100, 200)
  )

})

#' Tuning Parameters Grid
#'
#' Defines a tuning grid from a set of parameters.
#'
#' @rdname ParameterGrid
#'
#' @param ... named \code{param} objects as defined in the \pkg{dials} package.
#' @param object list of named \code{param} objects or a
#'   \code{\link[dials]{parameters}} object.  This is a positional argument
#'   that must be given first in calls to its methods.
#' @param size single integer or vector of integers whose positions or names
#'   match the given parameters and which specify the number of values used to
#'   construct the grid.
#' @param random number of unique points to sample at random from the grid
#'   defined by \code{size}, or \code{FALSE} for all points.
#'
#' @return \code{ParameterGrid} class object that inherits from
#' \code{parameters} and \code{TuningGrid}.
#'
#' @seealso \code{\link{TunedModel}}
#'
#' @examples
#' ## GBMModel tuning parameters
#' grid <- ParameterGrid(
#'   n.trees = dials::trees(),
#'   interaction.depth = dials::tree_depth(),
#'   random = 5
#' )
#' TunedModel(GBMModel, grid = grid)
#'
ParameterGrid <- function(...) {
  UseMethod("ParameterGrid")
}


#' @rdname ParameterGrid
#'
ParameterGrid.param <- function(..., size = 3, random = FALSE) {
  ParameterGrid(list(...), size = size, random = random)
}


#' @rdname ParameterGrid
#'
ParameterGrid.list <- function(object, size = 3, random = FALSE, ...) {
  ParameterGrid(parameters(object), size = size, random = random)
}


#' @rdname ParameterGrid
#'
ParameterGrid.parameters <- function(object, size = 3, random = FALSE, ...) {
  grid <- TuningGrid(size = size, random = random)

  size <- grid@size
  if (!is.null(names(size))) {
    if (!all(names(size) %in% object$id)) {
      throw(LocalWarning(
        "Unmatched parameter names in ParameterGrid() argument `size`.\n",
        "x Existing data has ", note_items("parameter{?s}: ", object$id), ".\n",
        "x Assigned data has ", note_items("name{?s}: ", names(size)), "."
      ))
    }
    size <- size[object$id]
    size[is.na(size)] <- 0L
  } else if (length(size) > 1 && length(size) != nrow(object)) {
    throw(LocalError(
      "Length of ParameterGrid() argument `size` must equal 1 ",
      "or the number of parameters.\n",
      "x Existing data has ", nrow(object), " ",
      note_items("parameter{?s}: ", object$id), ".\n",
      "x Assigned data has ", length(size), " ", note_items("size{?s}: ", size),
      "."
    ))
  }
  keep <- size >= 1
  object <- object[keep, ]
  size <- size[keep]

  new("ParameterGrid", object, size = size, random = grid@random)
}


#' Tuning Grid Control
#'
#' Defines control parameters for a tuning grid.
#'
#' @param size single integer or vector of integers whose positions or names
#'   match the parameters in a model's tuning grid and which specify the number
#'   of values used to construct the grid.
#' @param random number of unique points to sample at random from the grid
#'   defined by \code{size}.  If \code{size} is a single unnamed integer, then
#'   \code{random = Inf} will include all values of all grid parameters in the
#'   constructed grid, whereas \code{random = FALSE} will include all values of
#'   default grid parameters.
#'
#' @details
#' Returned \code{TuningGrid} objects may be supplied to
#' \code{\link{TunedModel}} for automated construction of model tuning grids.
#' These grids can be extracted manually and viewed with the
#' \code{\link{expand_modelgrid}} function.
#'
#' @return \code{TuningGrid} class object.
#'
#' @seealso \code{\link{TunedModel}}, \code{\link{expand_modelgrid}}
#'
#' @examples
#' TunedModel(XGBTreeModel, grid = TuningGrid(10, random = 5))
#'
TuningGrid <- function(size = 3, random = FALSE) {
  size <- check_integer(size, bounds = c(0, Inf), size = NA)
  throw(check_assignment(size))

  if (!isFALSE(random)) {
    random <- check_integer(random, bounds = c(1, Inf), size = 1)
    throw(check_assignment(random))
  }

  new("TuningGrid", size = size, random = random)
}


get_grid <- function(object, ...) {
  UseMethod("get_grid")
}


get_grid.default <- function(object, ...) {
  make_grid(object, class1(object))
}


get_grid.EnsembleInputOrModel <- function(object, ...) {
  make_grid(
    object,
    names(object@candidates),
    tibble(id = map("char", slot, object@candidates, "id"))
  )
}


get_grid.ModelSpecification <- function(object, ...) {
  make_grid(object, "ModelSpec", object@grid)
}


get_grid.StackedModel <- function(object, ...) {
  get_grid.default(object, ...)
}


get_grid.TunedModelRecipe <- function(object, ...) {
  make_grid(object, "ModelRecipe", as(object@grid, "tbl_df"))
}


get_grid.TunedModel <- function(object, ...) {
  make_grid(object, object@model@name, expand_modelgrid(object, ...))
}


grid_diff <- function(x, params = NULL, drop = FALSE) {
  res <- x
  for (name in names(x)) {
    node <- x[[name]]
    if (is_tibble(node)) {
      res[[name]] <- grid_diff(node, params[[name]], drop = TRUE)
    } else if (name %in% names(params)) {
      res[[name]] <- NULL
    }
  }
  if (length(res) || !drop) res
}


make_grid <- function(object, name = character(), params = tibble()) {
  if (is_empty(name)) name <- "Model"
  if (is_empty(params)) params <- tibble(.rows = 1)
  name <- make_unique(rep_len(name, nrow(params)))
  if (is_optim_method(object, "RandomGridSearch")) {
    n <- nrow(params)
    size <- get_optim_field(object, "size")
    inds <- sort(sample.int(n, min(max(size, 1), n)))
    params <- params[inds, ]
    name <- name[inds]
  }
  tibble(name = name, params = params)
}


new_gridinfo <- function(
  param = character(), get_values = list(), default = logical()
) {
  if (is_empty(default)) default <- TRUE

  stopifnot(is.character(param))
  stopifnot(is.list(get_values))
  stopifnot(is.logical(default))

  if (!all(map("logi", is.function, get_values))) {
    get_values <- Error("Value must be a list of functions.")
    throw(check_assignment(get_values))
  }

  as_tibble(list(param = param, get_values = get_values, default = default))
}


random_grid <- function(x, size = integer()) {
  stopifnot(is.list(x))

  names(x) <- make_names_along(x, "value")
  grids <- map(function(value, name) {
    do.call(tibble, structure(list(value), names = name))
  }, x, names(x))

  max_size <- prod(map("int", nrow, grids))
  if (is_empty(size)) size <- max_size
  size <- min(size, max_size)

  if (is_empty(grids) || size < 1) {
    grid <- as_tibble(map(function(data) data[NULL, 1, drop = TRUE], grids))
  } else {
    grid <- NULL
    iter <- 0
    while (size(grid, 1) < size && iter < 100) {
      iter <- iter + 1
      grid_sample <- as_tibble(map(function(data) {
        inds <- sample.int(nrow(data), size = size, replace = TRUE)
        data[inds, 1, drop = TRUE]
      }, grids))
      grid <- unique_grid(rbind(grid, grid_sample))
    }

    grid <- head(grid, size)
    cols <- unname(unnest_params(grid))
    sortable_types <- c(
      "character", "complex", "Date", "factor", "logical", "numeric"
    )
    is_sortable <- map("logi", function(col) {
      any(map("logi", is, list(col), sortable_types))
    }, cols)
    if (any(is_sortable)) {
      sort_order <- do.call(order, cols[is_sortable])
      grid <- grid[sort_order, ]
    }
  }

  grid
}


regular_grid <- function(x) {
  stopifnot(is.list(x))

  names(x) <- make_names_along(x, "value")
  grids <- map(function(value, name) {
    do.call(tibble, structure(list(value), names = name))
  }, x, names(x))

  ns <- map("int", nrow, grids)
  n <- prod(ns)

  if (is_empty(grids) || n == 0) {
    repeats <- 0
  } else {
    ns_cumprod <- cumprod(ns)
    grids <- map(function(data, each) {
      inds <- rep(seq_len(nrow(data)), each = each)
      data[inds, ]
    }, grids, n / ns_cumprod)
    repeats <- ns_cumprod / ns
  }
  grids <- map(function(data, times) {
    inds <- rep(seq_len(nrow(data)), times = times)
    data[inds, 1, drop = TRUE]
  }, grids, repeats)

  as_tibble(grids)
}


renest_params <-function(x, template, drop = FALSE) {
  res <- template
  ind <- 1
  for (name in names(template)) {
    node <- template[[name]]
    res[[name]] <- if (is_tibble(node)) {
      size <- length(unnest_params(node))
      renest_params(x[seq(ind, length = size)], node, drop = TRUE)
    } else {
      size <- 1
      x[[ind]]
    }
    ind <- ind + size
  }
  if (length(res) || !drop) res
}


unique_grid <- function(x) {
  if (is_empty(x)) x else x[!duplicated(unnest_params(x)), ]
}


unnest_params <- function(x, compact_names = FALSE) {
  res <- tibble(.rows = nrow(x))
  for (name in names(x)) {
    value <- x[[name]]
    if (is_tibble(value)) {
      value <- unnest_params(value)
      if (compact_names) name <- c(name, character(length(value) - 1))
      name <- paste0(name, "$", names(value))
    }
    res[name] <- value
  }
  res
}

#' Super Learner Model
#'
#' Fit a super learner model to predictions from multiple base learners.
#'
#' @param ... \link[=models]{model} functions, function names, objects; other
#'   objects that can be \link[=as.MLModel]{coerced} to models; or vector of
#'   these to serve as base learners.
#' @param model \link[=models]{model} function, function name, or object
#'   defining the super model; or another object that can be
#'   \link[=as.MLModel]{coerced} to the model.
#' @param control \link[=controls]{control} function, function name, or object
#'   defining the resampling method to be employed for the estimation of base
#'   learner weights.
#' @param all_vars logical indicating whether to include the original
#'   predictor variables in the super model.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric}, \code{ordered},
#'     \code{Surv}}
#' }
#'
#' @return \code{SuperModel} class object that inherits from \code{MLModel}.
#'
#' @references
#' van der Laan, M. J., Polley, E. C., & Hubbard, A. E. (2007). Super learner.
#' \emph{Statistical Applications in Genetics and Molecular Biology},
#' \emph{6}(1).
#'
#' @seealso \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested packages gbm and glmnet to run
#'
#' model <- SuperModel(GBMModel, SVMRadialModel, GLMNetModel(lambda = 0.01))
#' model_fit <- fit(sale_amount ~ ., data = ICHomes, model = model)
#' predict(model_fit, newdata = ICHomes)
#' }
#'
SuperModel <- function(
  ..., model = GBMModel, control = MachineShop::settings("control"),
  all_vars = FALSE
) {

  models <- map(as.MLModel, unlist(list(model, ...)))
  names(models)[1] <- "SuperLearner"
  names(models)[-1] <- make_names_len(length(models) - 1, "BaseLearner")

  slots <- combine_model_slots(models, models[[1]]@response_types)
  new("SuperModel",
    MLModel(
      name = "SuperModel",
      label = "Super Learner",
      response_types = slots$response_types,
      weights = slots$weights,
      params = TrainingParams(
        optim = NullOptimization(),
        control = control,
        options = list(all_vars = all_vars)
      )
    ),
    candidates = ListOf(models)
  )

}

MLModelFunction(SuperModel) <- NULL


.fit.SuperModel <- function(object, input, ...) {
  input_prep <- prep(input)
  mf <- ModelFrame(input_prep, na.rm = FALSE)

  super_learner <- object@candidates[[1]]
  base_learners <- object@candidates[-1]
  control <- object@params@control
  all_vars <- object@params@options$all_vars

  predictors <- list()
  ind <- new_progress_index(max = length(base_learners), name = object@name)
  while (ind < max(ind)) {
    ind <- ind + 1
    res <- resample(
      input, model = base_learners[[ind]], control = control,
      progress_index = ind
    )
    predictors[[ind]] <- res$Predicted
  }

  df <- super_df(res$Observed, predictors, res$Case, if (all_vars) mf)
  super_mf <- ModelFrame(formula(df), df)

  list(
    base_fits = map(
      function(learner) fit(input, model = learner), base_learners
    ),
    super_fit = fit(super_mf, model = super_learner),
    all_vars = all_vars,
    times = control@predict$times
  ) %>% MLModelFit("SuperModelFit", input = input_prep, model = object)
}


.predict.SuperModel <- function(object, model_fit, newdata, times, ...) {
  predictors <- map(function(fit) {
    predict(fit, newdata = newdata, times = model_fit$times, type = "raw")
  }, model_fit$base_fits)

  df <- if (model_fit$all_vars) {
    newdata[["(names)"]] <- rownames(newdata)
    super_df(NA, predictors, newdata[["(names)"]], newdata)
  } else {
    super_df(NA, predictors)
  }

  predict(model_fit$super_fit, newdata = df, times = times, type = "raw")
}


super_df <- function(y, predictors, case_names = NULL, data = NULL) {
  names(predictors) <- make_names_len(length(predictors), "X")
  df <- data.frame(y = y, unnest(as.data.frame(predictors)))

  if (!is.null(data)) {
    df[["(names)"]] <- case_names

    data_predictors <- predictors(data)
    unique_names <- make_unique(c(names(df), names(data_predictors)))
    names(data_predictors) <- tail(unique_names, length(data_predictors))
    data_predictors[["(names)"]] <- data[["(names)"]]

    merge(df, data_predictors, by = "(names)", sort = FALSE)[-1]
  } else {
    df
  }
}

#' K-Means Clustering Variable Reduction
#'
#' Creates a \emph{specification} of a recipe step that will convert numeric
#' variables into one or more by averaging within k-means clusters.
#'
#' @inheritParams step_lincomp
#' @param k number of k-means clusterings of the variables.  The value of
#'   \code{k} is constrained to be between 1 and one less than the number of
#'   original variables.
#' @param center,scale logicals indicating whether to mean center and standard
#'   deviation scale the original variables prior to deriving components, or
#'   functions or names of functions for the centering and scaling.
#' @param algorithm character string specifying the clustering algorithm to use.
#' @param max_iter maximum number of algorithm iterations allowed.
#' @param num_start number of random cluster centers generated for starting the
#'   Hartigan-Wong algorithm.
#' @param x \code{step_kmeans} object.
#'
#' @return Function \code{step_kmeans} creates a new step whose class is of
#' the same name and inherits from \code{\link{step_lincomp}}, adds it to the
#' sequence of existing steps (if any) in the recipe, and returns the updated
#' recipe.  For the \code{tidy} method, a tibble with columns \code{terms}
#' (selectors or variables selected), \code{cluster} assignments, \code{sqdist}
#' (squared distance from cluster centers), and \code{name} of the new variable
#' names.
#'
#' @details
#' K-means clustering partitions variables into k groups such that the sum of
#' squares between the variables and their assigned cluster means is minimized.
#' Variables within each cluster are then averaged to derive a new set of k
#' variables.
#'
#' @references
#' Forgy, E. W. (1965). Cluster analysis of multivariate data: efficiency versus
#' interpretability of classifications. \emph{Biometrics}, \emph{21}, 768-769.
#'
#' Hartigan, J. A., & Wong, M. A. (1979). A K-means clustering algorithm.
#' \emph{Applied Statistics}, \emph{28}, 100-108.
#'
#' Lloyd, S. P. (1982). Least squares quantization in PCM. \emph{IEEE
#' Transactions on Information Theory}, \emph{28}(2), 129-137.
#'
#' MacQueen, J. (1967). Some methods for classification and analysis of
#' multivariate observations. In L. M. Le Cam & J. Neyman (Eds.),
#' \emph{Proceedings of the fifth Berkeley Symposium on Mathematical Statistics
#' and Probability} (vol. 1, pp. 281-297). University of California Press.
#'
#' @seealso \code{\link[stats]{kmeans}}, \code{\link[recipes]{recipe}},
#' \code{\link[recipes]{prep}}, \code{\link[recipes]{bake}}
#'
#' @examples
#' library(recipes)
#'
#' rec <- recipe(rating ~ ., data = attitude)
#' kmeans_rec <- rec %>%
#'   step_kmeans(all_predictors(), k = 3)
#' kmeans_prep <- prep(kmeans_rec, training = attitude)
#' kmeans_data <- bake(kmeans_prep, attitude)
#'
#' pairs(kmeans_data, lower.panel = NULL)
#'
#' tidy(kmeans_rec, number = 1)
#' tidy(kmeans_prep, number = 1)
#'
step_kmeans <- function(
  recipe, ..., k = 5, center = TRUE, scale = TRUE,
  algorithm = c("Hartigan-Wong", "Lloyd", "Forgy", "MacQueen"), max_iter = 10,
  num_start = 1, replace = TRUE, prefix = "KMeans", role = "predictor",
  skip = FALSE, id = recipes::rand_id("kmeans")
) {

  recipes::add_step(recipe, new_step_kmeans(
    terms = recipes::ellipse_check(...),
    k = k,
    center = center,
    scale = scale,
    algorithm = match.arg(algorithm),
    max_iter = max_iter,
    num_start = num_start,
    replace = replace,
    prefix = prefix,
    role = role,
    skip = skip,
    id = id
  ))

}


new_step_kmeans <- function(..., k, algorithm, max_iter, num_start) {

  transform <- function(x, step) {

    res <- stats::kmeans(
      t(x), centers = min(max(step$k, 1), ncol(x) - 1),
      iter.max = step$max_iter, nstart = step$num_start,
      algorithm = step$algorithm
    )

    cluster <- res$cluster
    weights <- Matrix::t(Matrix::fac2sparse(cluster) / res$size[res$size > 0])
    sqdist <- numeric(length(cluster))
    for (i in unique(cluster)) {
      in_cluster <- cluster == i
      x_centered <- x[, in_cluster, drop = FALSE] - res$centers[i, ]
      sqdist[in_cluster] <- colSums(x_centered^2)
    }

    list(weights = weights, cluster = cluster, sqdist = sqdist)

  }

  options <- list(
    k = k,
    algorithm = algorithm,
    max_iter = max_iter,
    num_start = num_start
  )

  object <- new_step_lincomp(
    ..., transform = transform, num_comp = NULL, options = options
  )
  object$res <- tibble(
    terms = recipes::sel2char(object$terms),
    cluster = NA_integer_,
    sqdist = NA_real_,
    name = NA_character_
  )

  structure(object, class = c("step_kmeans", "step"))

}


prep.step_kmeans <- function(x, ...) prep.step_lincomp(x, ...)
bake.step_kmeans <- function(object, ...) bake.step_lincomp(object, ...)
print.step_kmeans <- function(x, ...) print.step_lincomp(x, ...)


#' @rdname step_kmeans
#'
tidy.step_kmeans <- function(x, ...) {
  res <- x$res
  if (is_trained(x)) {
    cluster <- x$res$cluster
    res <- tibble(
      terms = rownames(x$res$weights),
      cluster = cluster,
      sqdist = x$res$sqdist,
      name = colnames(x$res$weights)[cluster]
    )
  }
  res$id <- x$id
  res
}


#' @rdname step_kmeans
#'
tunable.step_kmeans <- function(x, ...) {
  tibble(
    name = "k",
    call_info = list(list(pkg = "dials", fun = "num_comp", range = c(2, 10))),
    source = "MachineShop",
    component = "step_kmeans",
    component_id = x$id
  )
}

#' MLMetric Class Constructor
#'
#' Create a performance metric for use with the \pkg{MachineShop} package.
#'
#' @rdname MLMetric
#'
#' @param object function to compute the metric, defined to accept
#'   \code{observed} and \code{predicted} as the first two arguments and with an
#'   ellipsis (\code{...}) to accommodate others.
#' @param name character name of the object to which the metric is assigned.
#' @param label optional character descriptor for the model.
#' @param maximize logical indicating whether higher values of the metric
#'   correspond to better predictive performance.
#' @param value list of arguments to pass to the \code{MLMetric} constructor.
#'
#' @return \code{MLMetric} class object.
#'
#' @seealso \code{\link{metrics}}
#'
#' @examples
#' f2_score <- MLMetric(
#'   function(observed, predicted, ...) {
#'     f_score(observed, predicted, beta = 2, ...)
#'   },
#'   name = "f2_score",
#'   label = "F Score (beta = 2)",
#'   maximize = TRUE
#' )
#'
MLMetric <- function(object, name = "MLMetric", label = name, maximize = TRUE) {
  new("MLMetric", object, name = name, label = label, maximize = maximize)
}


#' @rdname MLMetric
#'
"MLMetric<-" <- function(object, value) {
  do.call(MLMetric, c(object, value))
}

#' Model Performance Summaries
#'
#' Summary statistics for resampled model performance metrics.
#'
#' @name summary
#' @rdname summary-methods
#'
#' @param object \link{confusion}, \link{lift}, trained model \link{fit},
#'   \link{performance}, \link[=curves]{performance curve}, \link{resample}, or
#'   \link{rfe} result.
#' @param stat function or character string naming a function to compute a
#'   summary statistic at each cutoff value of resampled metrics in
#'   \code{PerformanceCurve}, or \code{NULL} for resample-specific metrics.
#' @param stats function, function name, or vector of these with which to
#'   compute summary statistics.
#' @param na.rm logical indicating whether to exclude missing values.
#' @param .type character string specifying that
#'   \code{\link[=unMLModelFit]{unMLModelFit(object)}} be passed to
#'   \code{\link[base:summary]{summary}} (\code{"default"}),
#'   \code{\link[generics:glance]{glance}}, or
#'   \code{\link[generics:tidy]{tidy}}.
#' @param ... arguments passed to other methods.
#'
#' @return An object of summmary statistics.
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' ## Factor response example
#'
#' fo <- Species ~ .
#' control <- CVControl()
#'
#' gbm_res1 <- resample(fo, iris, GBMModel(n.trees = 25), control)
#' gbm_res2 <- resample(fo, iris, GBMModel(n.trees = 50), control)
#' gbm_res3 <- resample(fo, iris, GBMModel(n.trees = 100), control)
#' summary(gbm_res3)
#'
#' res <- c(GBM1 = gbm_res1, GBM2 = gbm_res2, GBM3 = gbm_res3)
#' summary(res)
#' }
#'
NULL


#' @rdname summary-methods
#'
summary.ConfusionList <- function(object, ...) {
  ListOf(map(function(conf) summary(conf, ...), object))
}


#' @rdname summary-methods
#'
summary.ConfusionMatrix <- function(object, ...) {
  total <- sum(object)
  object <- object / total

  observed <- colSums(object)
  predicted <- rowSums(object)
  agreement <- diag(object)

  perf <- rbind(
    Observed = observed,
    Predicted = predicted,
    Agreement = agreement,
    Sensitivity = agreement / observed,
    Specificity = (1 - observed - predicted + agreement) / (1 - observed),
    PPV = agreement / predicted,
    NPV = (1 - observed - predicted + agreement) / (1 - predicted)
  )

  ConfusionSummary(
    perf, total = total, accuracy = sum(agreement), majority = max(observed),
    kappa2 = 1 - (1 - sum(agreement)) / (1 - sum(observed * predicted))
  )
}


#' @rdname summary-methods
#'
summary.MLModel <- function(
  object, stats = MachineShop::settings("stats.Resample"), na.rm = TRUE, ...
) {
  if (!is_trained(object)) throw(Warning("No training results to summarize."))
  ListOf(map(summary, object@steps))
}


#' @rdname summary-methods
#'
summary.MLModelFit <- function(
  object, .type = c("default", "glance", "tidy"), ...
) {
  switch(match.arg(.type),
    "default" = summary,
    fget(.type, package = "generics")
  )(unMLModelFit(object), ...)
}


#' @rdname summary-methods
#'
summary.Performance <- function(
  object, stats = MachineShop::settings("stats.Resample"), na.rm = TRUE, ...
) {
  stats <- check_stats(stats, convert = TRUE)
  throw(check_assignment(stats))

  f <- function(x) {
    prop_na <- mean(is.na(x))
    if (na.rm) x <- as.numeric(na.omit(x))
    c(stats(x), "NA" = prop_na)
  }

  margins <- 2
  perm <- c(2, 1)
  names <- c("Metric", "Statistic")
  if (ndim(object) == 3) {
    margins <- c(3, margins)
    perm <- c(perm, 3)
    names <- c("Model", "Statistic", "Metric")
  }
  object_summary <- aperm(apply(object, margins, f), perm = perm)
  names(dimnames(object_summary)) <- names
  TabularArray(object_summary)
}


#' @rdname summary-methods
#'
summary.PerformanceCurve <- function(
  object, stat = MachineShop::settings("stat.Curve"), ...
) {
  if (!(is.null(object$Iteration) || is.null(stat))) {

    stat <- check_stat(stat, convert = TRUE)
    throw(check_assignment(stat))

    object_class <- class(object)
    stat_na_omit <- function(x) stat(x[is.finite(x)])

    object_list <- by(object, object$Model, function(curves) {
      cutoffs <- unique(curves$Cutoff)
      curves_split <- split(curves, curves$Iteration)
      x_all <- y_all <- matrix(NA, length(cutoffs), length(curves_split))
      for (j in seq_along(curves_split)) {
        curve <- curves_split[[j]]
        x_all[, j] <- .curve_approx(curve$Cutoff, curve$x, cutoffs)
        y_all[, j] <- .curve_approx(curve$Cutoff, curve$y, cutoffs)
      }
      do.call(object_class, list(
        data.frame(
          Cutoff = cutoffs,
          x = apply(x_all, 1, stat_na_omit),
          y = apply(y_all, 1, stat_na_omit)
        ),
        metrics = object@metrics
      ))
    })

    object <- do.call(c, object_list)
  }

  object
}


.curve_approx <- function(...) {
  tryCatch(
    approx(..., method = "constant", rule = 2, f = 0.5)$y,
    error = function(cond) NA
  )
}


#' @rdname summary-methods
#'
summary.Resample <- function(
  object, stats = MachineShop::settings("stats.Resample"), na.rm = TRUE, ...
) {
  summary(performance(object), stats = stats, na.rm = na.rm)
}


summary.SurvRegModelFit <- function(object, ...) {
  object$fail <- NULL
  class(object) <- setdiff(class(object), c("psm", "rms"))
  NextMethod()
}


#' @rdname summary-methods
#'
summary.TrainingStep <- function(object, ...) {
  object@log
}

#' Models
#'
#' Model constructor functions supplied by \pkg{MachineShop} are summarized in
#' the table below according to the types of response variables with which each
#' can be used.
#' \tabular{lccc}{
#'   \strong{Function} \tab \strong{Categorical} \tab \strong{Continuous}
#'   \tab \strong{Survival} \cr
#'   \code{\link{AdaBagModel}}         \tab f   \tab     \tab   \cr
#'   \code{\link{AdaBoostModel}}       \tab f   \tab     \tab   \cr
#'   \code{\link{BARTModel}}           \tab f   \tab n   \tab S \cr
#'   \code{\link{BARTMachineModel}}    \tab b   \tab n   \tab   \cr
#'   \code{\link{BlackBoostModel}}     \tab b   \tab n   \tab S \cr
#'   \code{\link{C50Model}}            \tab f   \tab     \tab   \cr
#'   \code{\link{CForestModel}}        \tab f   \tab n   \tab S \cr
#'   \code{\link{CoxModel}}            \tab     \tab     \tab S \cr
#'   \code{\link{CoxStepAICModel}}     \tab     \tab     \tab S \cr
#'   \code{\link{EarthModel}}          \tab f   \tab n   \tab   \cr
#'   \code{\link{FDAModel}}            \tab f   \tab     \tab   \cr
#'   \code{\link{GAMBoostModel}}       \tab b   \tab n   \tab S \cr
#'   \code{\link{GBMModel}}            \tab f   \tab n   \tab S \cr
#'   \code{\link{GLMBoostModel}}       \tab b   \tab n   \tab S \cr
#'   \code{\link{GLMModel}}            \tab f   \tab m,n \tab   \cr
#'   \code{\link{GLMStepAICModel}}     \tab b   \tab n   \tab   \cr
#'   \code{\link{GLMNetModel}}         \tab f   \tab m,n \tab S \cr
#'   \code{\link{KNNModel}}            \tab f,o \tab n   \tab   \cr
#'   \code{\link{LARSModel}}           \tab     \tab n   \tab   \cr
#'   \code{\link{LDAModel}}            \tab f   \tab     \tab   \cr
#'   \code{\link{LMModel}}             \tab f   \tab m,n \tab   \cr
#'   \code{\link{MDAModel}}            \tab f   \tab     \tab   \cr
#'   \code{\link{NaiveBayesModel}}     \tab f   \tab     \tab   \cr
#'   \code{\link{NNetModel}}           \tab f   \tab n   \tab   \cr
#'   \code{\link{ParsnipModel}}        \tab f   \tab m,n \tab S \cr
#'   \code{\link{PDAModel}}            \tab f   \tab     \tab   \cr
#'   \code{\link{PLSModel}}            \tab f   \tab n   \tab   \cr
#'   \code{\link{POLRModel}}           \tab o   \tab     \tab   \cr
#'   \code{\link{QDAModel}}            \tab f   \tab     \tab   \cr
#'   \code{\link{RandomForestModel}}   \tab f   \tab n   \tab   \cr
#'   \code{\link{RangerModel}}         \tab f   \tab n   \tab S \cr
#'   \code{\link{RFSRCModel}}          \tab f   \tab m,n \tab S \cr
#'   \code{\link{RFSRCFastModel}}      \tab f   \tab m,n \tab S \cr
#'   \code{\link{RPartModel}}          \tab f   \tab n   \tab S \cr
#'   \code{\link{SurvRegModel}}        \tab     \tab     \tab S \cr
#'   \code{\link{SurvRegStepAICModel}} \tab     \tab     \tab S \cr
#'   \code{\link{SVMModel}}            \tab f   \tab n   \tab   \cr
#'   \code{\link{SVMANOVAModel}}       \tab f   \tab n   \tab   \cr
#'   \code{\link{SVMBesselModel}}      \tab f   \tab n   \tab   \cr
#'   \code{\link{SVMLaplaceModel}}     \tab f   \tab n   \tab   \cr
#'   \code{\link{SVMLinearModel}}      \tab f   \tab n   \tab   \cr
#'   \code{\link{SVMPolyModel}}        \tab f   \tab n   \tab   \cr
#'   \code{\link{SVMRadialModel}}      \tab f   \tab n   \tab   \cr
#'   \code{\link{SVMSplineModel}}      \tab f   \tab n   \tab   \cr
#'   \code{\link{SVMTanhModel}}        \tab f   \tab n   \tab   \cr
#'   \code{\link{TreeModel}}           \tab f   \tab n   \tab   \cr
#'   \code{\link{XGBModel}}            \tab f   \tab n   \tab S \cr
#'   \code{\link{XGBDARTModel}}        \tab f   \tab n   \tab S \cr
#'   \code{\link{XGBLinearModel}}      \tab f   \tab n   \tab S \cr
#'   \code{\link{XGBTreeModel}}        \tab f   \tab n   \tab S \cr
#' }
#' Categorical: b = binary, f = factor, o = ordered\cr
#' Continuous: m = matrix, n = numeric\cr
#' Survival: S = Surv\cr
#' \cr
#' Models may be combined, tuned, or selected with the following meta-model
#' functions.
#' \tabular{ll}{
#'   \code{\link{ModelSpecification}} \tab Model specification \cr
#'   \code{\link{StackedModel}}  \tab Stacked regression \cr
#'   \code{\link{SuperModel}}    \tab Super learner \cr
#'   \code{\link{SelectedModel}} \tab Model selection from a candidate set \cr
#'   \code{\link{TunedModel}}    \tab Model tuning over a parameter grid \cr
#' }
#'
#' @name models
#'
#' @seealso \code{\link{modelinfo}}, \code{\link{fit}}, \code{\link{resample}}
#'
NULL


#' Model Inputs
#'
#' Model inputs are the predictor and response variables whose relationship is
#' determined by a model fit.  Input specifications supported by
#' \pkg{MachineShop} are summarized in the table below.
#' \tabular{ll}{
#'   \code{\link{formula}}         \tab Traditional model formula \cr
#'   \code{\link{matrix}}          \tab Design matrix of predictors \cr
#'   \code{\link{ModelFrame}}      \tab Model frame \cr
#'   \code{\link{ModelSpecification}} \tab Model specification \cr
#'   \code{\link[recipes]{recipe}} \tab Preprocessing recipe roles and steps \cr
#' }
#' Response variable types in the input specifications are defined by the user
#' with the functions and recipe roles:
#' \tabular{ll}{
#'   \strong{Response Functions}
#'     \tab \code{\link{BinomialVariate}} \cr
#'     \tab \code{\link{DiscreteVariate}} \cr
#'     \tab \code{\link{factor}} \cr
#'     \tab \code{\link{matrix}} \cr
#'     \tab \code{\link{NegBinomialVariate}} \cr
#'     \tab \code{\link{numeric}} \cr
#'     \tab \code{\link{ordered}} \cr
#'     \tab \code{\link{PoissonVariate}} \cr
#'     \tab \code{\link[survival]{Surv}} \cr
#'   \strong{Recipe Roles}
#'     \tab \code{\link{role_binom}} \cr
#'     \tab \code{\link{role_surv}} \cr
#' }
#' Inputs may be combined, selected, or tuned with the following meta-input
#' functions.
#' \tabular{ll}{
#'   \code{\link{ModelSpecification}} \tab Model specification \cr
#'   \code{\link{SelectedInput}} \tab Input selection from a candidate set \cr
#'   \code{\link{TunedInput}}    \tab Input tuning over a parameter grid \cr
#' }
#'
#' @name inputs
#'
#' @seealso \code{\link{fit}}, \code{\link{resample}}
#'
NULL

#' C5.0 Decision Trees and Rule-Based Model
#'
#' Fit classification tree models or rule-based models using Quinlan's C5.0
#' algorithm.
#'
#' @param trials integer number of boosting iterations.
#' @param rules logical indicating whether to decompose the tree into a
#'   rule-based model.
#' @param subset logical indicating whether the model should evaluate groups of
#'   discrete predictors for splits.
#' @param bands integer between 2 and 1000 specifying a number of bands into
#'   which to group rules ordered by their affect on the error rate.
#' @param winnow logical indicating use of predictor winnowing (i.e. feature
#'   selection).
#' @param noGlobalPruning logical indicating a final, global pruning step to
#'   simplify the tree.
#' @param CF number in (0, 1) for the confidence factor.
#' @param minCases integer for the smallest number of samples that must be put
#'   in at least two of the splits.
#' @param fuzzyThreshold logical indicating whether to evaluate possible
#'   advanced splits of the data.
#' @param sample value between (0, 0.999) that specifies the random proportion
#'   of data to use in training the model.
#' @param earlyStopping logical indicating whether the internal method for
#'   stopping boosting should be used.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \code{trials}, \code{rules}, \code{winnow}
#'   }
#' }
#'
#' Latter arguments are passed to \code{\link[C50]{C5.0Control}}.
#' Further model details can be found in the source link below.
#'
#' In calls to \code{\link{varimp}} for \code{C50Model}, argument \code{type}
#' may be specified as \code{"usage"} (default) for the percentage of training
#' set samples that fall into all terminal nodes after the split of each
#' predictor or as \code{"splits"} for the percentage of splits associated with
#' each predictor.  Variable importance is automatically scaled to range from 0
#' to 100.  To obtain unscaled importance values, set \code{scale = FALSE}.  See
#' example below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[C50]{C5.0}}, \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package C50 to run
#'
#' model_fit <- fit(Species ~ ., data = iris, model = C50Model)
#' varimp(model_fit, method = "model", type = "splits", scale = FALSE)
#' }
#'
C50Model <- function(
  trials = 1, rules = FALSE, subset = TRUE, bands = 0, winnow = FALSE,
  noGlobalPruning = FALSE, CF = 0.25, minCases = 2, fuzzyThreshold = FALSE,
  sample = 0, earlyStopping = TRUE
) {

  MLModel(

    name = "C50Model",
    label = "C5.0 Classification",
    packages = "C50",
    response_types = "factor",
    weights = TRUE,
    predictor_encoding = "model.frame",
    na.rm = FALSE,
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = c("trials", "rules", "winnow"),
      get_values = c(
        function(n, ...) round_int(c(1, seq_range(0, 10, c(2, 100), n))),
        function(...) c(FALSE, TRUE),
        function(...) c(FALSE, TRUE)
      )
    ),

    fit = function(formula, data, weights, trials, rules, ...) {
      control <- C50::C5.0Control(...)
      eval_fit(
        data,
        formula = C50::C5.0(
          formula, data = data, weights = weights, trials = trials,
          rules = rules, control = control
        ),
        matrix = C50::C5.0(
          x, y, weights = weights, trials = trials, rules = rules,
          control = control
        )
      )
    },

    predict = function(object, newdata, ...) {
      newdata <- as.data.frame(newdata)
      predict(object, newdata = newdata, type = "prob")
    },

    varimp = function(object, type = c("usage", "splits"), ...) {
      type <- match.arg(type)
      structure(C50::C5imp(object, metric = type), metric = type)
    }

  )

}

MLModelFunction(C50Model) <- NULL

#################### Surv Prediction Method ####################


predict.Surv <- function(
  object, ..., times = numeric(), distr = character(), weights = NULL
) {
  distr <- if (is_counting(object)) {
    "empirical"
  } else if (is_empty(distr)) {
    settings(if (length(times)) "distr.SurvProbs" else "distr.SurvMeans")
  } else {
    match.arg(distr, c("empirical", "exponential", "rayleigh", "weibull"))
  }
  weights <- check_weights(weights, object)
  throw(check_assignment(weights))
  SurvPrediction(
    .predict.Surv(object, ..., times = times, distr = distr, weights = weights),
    times = times, distr = distr
  )
}


.predict.Surv <- function(object, ...) {
  UseMethod(".predict.Surv", ..1)
}


.predict.Surv.list <- function(object, x, times, ...) {
  for (i in seq_along(x)) {
    x[[i]] <- predict(object, x[[i]], times = times, ...)
  }
  if (length(times)) do.call(rbind, x) else as.numeric(x)
}


.predict.Surv.matrix <- function(object, x, times, distr, ...) {
  individual_fits <- surv_fit(distr, SurvProbs(x, time(object)))
  if (length(times)) {
    predict(individual_fits, times = times)
  } else {
    mean(individual_fits)
  }
}


.predict.Surv.numeric <- function(object, lp, new_lp, times, distr, ...) {
  risks <- exp(lp)
  new_risks <- exp(new_lp)
  baseline_fit <- surv_fit(distr, object, risks = risks, ...)
  if (length(times)) {
    predict(baseline_fit, times = times, new_risks = new_risks)
  } else {
    mean(baseline_fit, new_risks = new_risks)
  }
}


.predict.Surv.survfit <- function(object, x, times, distr, ...) {
  individual_fit <- surv_fit(distr, x)
  if (length(times)) {
    predict(individual_fit, times = times)
  } else {
    mean(individual_fit, max_time = max(time(object)))
  }
}


#################### Empirical Survival Distribution ####################


EmpiricalSurv <- function(x, ...) {
  UseMethod("EmpiricalSurv")
}


EmpiricalSurv.Surv <- function(
  x, risks = rep(1, length(x)), weights = rep(1, length(x)),
  method = c("efron", "breslow"), ...
) {
  events <- as.integer(x[, "status"])
  data <- na.omit(data.frame(
    x = x,
    event = events,
    risk = risks,
    wt = weights,
    wt_censor = weights * !events,
    wt_event = weights * events,
    wt_risk = weights * risks
  ))
  sums <- cbind(
    rowsum(data[c("wt_censor", "wt_event")], time(data$x)),
    risksum(data[c("wt", "wt_risk")], data$x)
  )

  if (is_empty(method)) method <- settings("method.EmpiricalSurv")
  method <- match.arg(method)
  cumhaz <- cumsum(switch(method,
    "breslow" = sums$wt_event / sums$wt_risk,
    "efron" = {
      data$wt_eventrisk <- data$wt_event * data$risk
      select <- c("event", "wt_eventrisk")
      sums[select] <- rowsum(data[select], time(data$x))
      hazfit_efron(sums$event, sums$wt_event, sums$wt_risk, sums$wt_eventrisk)
    }
  ))

  structure(list(
    n = length(data$x),
    time = sums$stop_time,
    n.risk = sums$wt,
    n.event = sums$wt_event,
    n.censor = sums$wt_censor,
    surv = exp(-cumhaz),
    cumhaz = cumhaz
  ), class = c("EmpiricalSurv", "survfitcox", "survfit"))
}


EmpiricalSurv.survfit <- function(x, ...) {
  if (!is(x, "EmpiricalSurv")) class(x) <- c("EmpiricalSurv", class(x))
  x
}


EmpiricalSurv.SurvProbs <- function(x, ...) {
  x
}


mean.EmpiricalSurv <- function(x, new_risks = numeric(), ...) {
  times <- x$time[x$n.event > 0]
  surv <- predict(x, times = times, new_risks = new_risks)
  surv_mean(times, surv, max(x$time))
}


predict.EmpiricalSurv <- function(object, times, new_risks = numeric(), ...) {
  surv <- NextMethod()
  if (length(new_risks)) t(outer(surv, new_risks, "^")) else rbind(surv)
}


#################### Exponential Distribution ####################


Exponential <- function(...) {
  Weibull(..., shape = 1)
}


#################### Rayleigh Distribution ####################


Rayleigh <- function(...) {
  Weibull(..., shape = 2)
}


#################### Weibull Distribution ####################


Weibull <- function(x, ...) {
  UseMethod("Weibull")
}


Weibull.numeric <- function(x = scale, shape, scale, ...) {
  structure(list(shape = shape, scale = x), class = "Weibull")
}


Weibull.Surv <- function(
  x, risks = rep(1, length(x)), weights = rep(1, length(x)), shape = numeric(),
  ...
) {
  x[which(time(x) <= 0)] <- NA
  weights[which(weights == 0)] <- NA
  data <- as.data.frame(complete_subset(
    x = x, risks = risks, weights = weights
  ))
  shape <- min(shape, Inf)
  nparams <- 1 + is.infinite(shape)
  params <- if (length(event_time(data$x)) >= nparams) {
    regfit <- survreg(
      x ~ offset(-log(risks)), data = data, dist = "weibull", scale = 1 / shape,
      weights = weights
    )
    c(1 / regfit$scale, exp(coef(regfit)[[1]]))
  } else {
    c(NA_real_, NA_real_)
  }
  Weibull(shape = params[1], scale = params[2]^-params[1])
}


Weibull.survfit <- function(x, weights = NULL, ...) {
  weights <- x$n / sum(x$n.event + x$n.censor)
  time_event <- rep(x$time, round(weights * x$n.event))
  time_censor <- rep(x$time, round(weights * x$n.censor))
  time <- c(time_event, time_censor)
  status <- rep(c(1, 0), c(length(time_event), length(time_censor)))
  Weibull(Surv(time, status), ...)
}


Weibull.SurvProbs <- function(x, shape = numeric(), ...) {
  weibullfit <- if (length(shape)) {
    function(df) c(mean(df$y - shape * df$x), shape)
  } else {
    function(df) if (nrow(df) >= 2) {
      coef(lm(y ~ x, data = df))
    } else c(NA_real_, NA_real_)
  }
  coef <- apply(x, 1, function(surv) {
    df <- surv_cases(
      x = log(x@times), y = log(-log(surv)), subset = diff(c(1, surv)) < 0
    )
    weibullfit(df)
  })
  Weibull(shape = coef[2, ], scale = exp(coef[1, ]))
}


mean.Weibull <- function(x, new_risks = numeric(), ...) {
  if (length(new_risks)) x$scale <- new_risks * x$scale
  x$scale^(-1 / x$shape) * gamma(1 + 1 / x$shape)
}


predict.Weibull <- function(object, times, new_risks = numeric(), ...) {
  if (is_empty(new_risks)) new_risks <- 1
  shape <- object$shape
  times_shape <- if (length(shape) == 1) {
    matrix(times^shape, length(new_risks), length(times), byrow = TRUE)
  } else {
    t(outer(times, shape, "^"))
  }
  exp((new_risks * -object$scale) * times_shape)
}


#################### survfit Methods ####################


mean.survfit <- function(x, max_time = max(x$time), ...) {
  is_event <- x$n.event > 0
  surv_mean(x$time[is_event], x$surv[is_event], max_time)
}


predict.survfit <- function(object, times, ...) {
  inds <- findInterval(times, object$time)
  c(1, object$surv)[inds + 1]
}


#################### SurvMatrix Methods ####################


mean.SurvProbs <- function(x, ...) {
  apply(x, 1, function(surv) surv_mean(x@times, surv))
}


predict.SurvProbs <- function(object, times, ...) {
  inds <- findInterval(times, object@times)
  cbind(1, object)[, inds + 1, drop = FALSE]
}


#################### Survival Utility Functions ####################


event_time <- function(x) {
  x <- x[x[, "status"] == 1]
  if (length(x)) {
    sort(unique(time(x)), na.last = TRUE, method = "quick")
  } else numeric()
}


get_surv_distr <- function(distr, observed, predicted) {
  if (is_counting(observed)) {
    "empirical"
  } else {
    if (is(predicted, "SurvProbs")) {
      pred_distr <- predicted@distr
      default_distr <- settings("distr.SurvProbs")
    } else {
      pred_distr <- if (is(predicted, "SurvTimes")) predicted@distr
      default_distr <- settings("distr.SurvMeans")
    }
    match.arg(
      na.omit(c(distr, pred_distr, default_distr))[1],
      c("empirical", names(survreg.distributions))
    )
  }
}


hazfit_efron <- function(event, wt_event, wt_risk, wt_eventrisk) {
  .Call(C_hazfit_efron, as.integer(event), as.double(wt_event),
        as.double(wt_risk), as.double(wt_eventrisk))
}


risksum <- function(x, group) {
  rcumsum <- function(x) rev(cumsum(rev(x)))
  stop <- time(group)
  stop_time <- sort(unique(stop), na.last = TRUE, method = "quick")
  res <- if (is_counting(group)) {
    start <- group[, "start"]
    start_time <- sort(unique(start), na.last = TRUE, method = "quick")
    unobserved <- approx(
      start_time, seq_along(start_time), stop_time, method = "constant", f = 1,
      yright = length(start_time) + 1
    )$y
    map(function(num_stop, num_start) {
      rcumsum(num_stop) - c(rcumsum(num_start), 0)[unobserved]
    }, rowsum(x, stop), rowsum(x, start))
  } else {
    map(function(num) rcumsum(num), rowsum(x, stop))
  }
  cbind(as.data.frame(res), stop_time = stop_time)
}


surv_cases <- function(..., subset = TRUE) {
  df <- data.frame(...)
  is_finite <- Reduce("&", map(is.finite, df))
  df[subset & is_finite, , drop = FALSE]
}


surv_fit <- function(x, ...) {
  f <- switch(x,
    "empirical" = EmpiricalSurv,
    "exponential" = Exponential,
    "rayleigh" = Rayleigh,
    "weibull" = Weibull
  )
  f(...)
}


surv_mean <- function(times, surv, max_time = max(times)) {
  times <- c(times, max_time)
  surv <- cbind(rbind(surv), 0)
  stopifnot(length(times) == ncol(surv))
  -as.numeric(times %*% diff(t(cbind(1, surv))))
}


surv_subset <- function(x, weights, include, time) {
  surv <- 1
  p <- weighted_mean(include, weights)
  x <- x[include]
  weights <- weights[include]
  if (length(x) && any(event_time(x) <= time)) {
    data <- data.frame(event = weights * x[, "status"], total = weights)
    sums <- cbind(
      rowsum(data["event"], time(x)),
      risksum(data["total"], x)
    )
    sums <- sums[sums$stop_time <= time, ]
    surv <- prod(1 - sums$event / sums$total)
  }
  list(surv = surv, p = p)
}


survmetric_mean <- function(x, times) {
  if (length(x) > 1) {
    weights <- diff(c(0, times)) / tail(times, 1)
    c("Mean" = sum(weights * x), x)
  } else {
    unname(x)
  }
}


time.MLModelFit <- function(x, ...) {
  NULL
}


time.StackedModelFit <- function(x, ...) {
  x$times
}


time.Surv <- function(x, ...) {
  throw(check_censoring(x, c("right", "counting")))
  x[, ncol(x) - 1]
}

#' Fast Random Forest Model
#'
#' Fast implementation of random forests or recursive partitioning.
#'
#' @param num.trees number of trees.
#' @param mtry number of variables to possibly split at in each node.
#' @param importance variable importance mode.
#' @param min.node.size minimum node size.
#' @param replace logical indicating whether to sample with replacement.
#' @param sample.fraction fraction of observations to sample.
#' @param splitrule splitting rule.
#' @param num.random.splits number of random splits to consider for each
#'   candidate splitting variable in the \code{"extratrees"} rule.
#' @param alpha significance threshold to allow splitting in the
#'   \code{"maxstat"} rule.
#' @param minprop lower quantile of covariate distribution to be considered for
#'   splitting in the \code{"maxstat"} rule.
#' @param split.select.weights numeric vector with weights between 0 and 1,
#'   representing the probability to select variables for splitting.
#' @param always.split.variables character vector with variable names to be
#'   always selected in addition to the \code{mtry} variables tried for
#'   splitting.
#' @param respect.unordered.factors handling of unordered factor covariates.
#' @param scale.permutation.importance scale permutation importance by
#'   standard error.
#' @param verbose show computation status and estimated runtime.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric}, \code{Surv}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \code{mtry}, \code{min.node.size}*, \code{splitrule}*
#'   }
#' }
#' * excluded from grids by default
#'
#' Default argument values and further model details can be found in the source
#' See Also link below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[ranger]{ranger}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package ranger to run
#'
#' fit(Species ~ ., data = iris, model = RangerModel)
#' }
#'
RangerModel <- function(
  num.trees = 500, mtry = integer(),
  importance = c("impurity", "impurity_corrected", "permutation"),
  min.node.size = integer(), replace = TRUE,
  sample.fraction = if (replace) 1 else 0.632, splitrule = character(),
  num.random.splits = 1, alpha = 0.5, minprop = 0.1,
  split.select.weights = numeric(), always.split.variables = character(),
  respect.unordered.factors = character(), scale.permutation.importance = FALSE,
  verbose = FALSE
) {

  importance <- match.arg(importance)

  MLModel(

    name = "RangerModel",
    label = "Fast Random Forests",
    packages = "ranger",
    response_types = c("factor", "numeric", "Surv"),
    weights = TRUE,
    predictor_encoding = "model.frame",
    na.rm = TRUE,
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = c("mtry", "min.node.size", "splitrule"),
      get_values = c(
        function(n, data, ...) seq_nvars(data, RangerModel, n),
        function(n, data, ...) {
          round_int(seq(1, min(20, nrow(data)), length = n))
        },
        function(n, data, ...) {
          methods <- switch_class(response(data),
            "factor" = c("gini", "extratrees"),
            "numeric" = c("variance", "extratrees", "maxstat"),
            "Surv" = c("logrank", "extratrees", "C", "maxstat")
          )
          head(methods, n)
        }
      ),
      default = c(TRUE, FALSE, FALSE)
    ),

    fit = function(formula, data, weights, ...) {
      ranger::ranger(
        formula, data = as.data.frame(formula, data = data),
        case.weights = weights, probability = is(response(data), "factor"), ...
      )
    },

    predict = function(object, newdata, ...) {
      newdata <- as.data.frame(newdata)
      pred <- predict(object, data = newdata)
      if (!is.null(object$survival)) {
        predict(Surv(pred$unique.death.times), pred$survival, ...)
      } else {
        pred$predictions
      }
    },

    varimp = function(object, ...) {
      structure(ranger::importance(object), metric = object$splitrule)
    }

  )

}

MLModelFunction(RangerModel) <- NULL

#' Gradient Boosting with Additive Models
#'
#' Gradient boosting for optimizing arbitrary loss functions, where
#' component-wise arbitrary base-learners, e.g., smoothing procedures, are
#' utilized as additive base-learners.
#'
#' @param family optional \code{\link[mboost]{Family}} object.  Set
#'   automatically according to the class type of the response variable.
#' @param baselearner character specifying the component-wise
#'   \code{\link[mboost:baselearners]{base learner}} to be used.
#' @param dfbase gobal degrees of freedom for P-spline base learners
#'   (\code{"bbs"}).
#' @param mstop number of initial boosting iterations.
#' @param nu step size or shrinkage parameter between 0 and 1.
#' @param risk method to use in computing the empirical risk for each boosting
#'   iteration.
#' @param stopintern logical inidicating whether the boosting algorithm stops
#'   internally when the out-of-bag risk increases at a subsequent iteration.
#' @param trace logical indicating whether status information is printed during
#'   the fitting process.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{binary factor}, \code{BinomialVariate},
#'     \code{NegBinomialVariate}, \code{numeric}, \code{PoissonVariate},
#'     \code{Surv}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameter:}{
#'     \code{mstop}
#'   }
#' }
#'
#' Default argument values and further model details can be found in the source
#' See Also links below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[mboost]{gamboost}}, \code{\link[mboost]{Family}},
#' \code{\link[mboost]{baselearners}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package mboost to run
#'
#' data(Pima.tr, package = "MASS")
#'
#' fit(type ~ ., data = Pima.tr, model = GAMBoostModel)
#' }
#'
GAMBoostModel <- function(
  family = NULL, baselearner = c("bbs", "bols", "btree", "bss", "bns"),
  dfbase = 4, mstop = 100, nu = 0.1, risk = c("inbag", "oobag", "none"),
  stopintern = FALSE, trace = FALSE
) {

  baselearner <- match.arg(baselearner)
  risk <- match.arg(risk)

  MLModel(

    name = "GAMBoostModel",
    label = "Gradient Boosting with Additive Models",
    packages = "mboost",
    response_types = c(
      "binary", "BinomialVariate", "NegBinomialVariate", "numeric",
      "PoissonVariate", "Surv"
    ),
    weights = TRUE,
    predictor_encoding = "model.frame",
    na.rm = if (baselearner == "btree") "response" else "all",
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = "mstop",
      get_values = c(
        function(n, ...) round_int(seq_range(0, 50, c(1, 1000), n + 1))
      )
    ),

    fit = function(
      formula, data, weights, family = NULL, baselearner, dfbase, ...
    ) {
      attach_objects(list(
        bbs = mboost::bbs,
        bols = mboost::bols,
        btree = mboost::btree
      ), name = "mboost_exports")

      if (is.null(family)) {
        family <- switch_class(response(data),
          "BinomialVariate" = mboost::Binomial(type = "glm"),
          "factor" = mboost::Binomial(),
          "NegBinomialVariate" = mboost::NBinomial(),
          "numeric" = mboost::Gaussian(),
          "PoissonVariate" = mboost::Poisson(),
          "Surv" = mboost::CoxPH()
        )
      }
      mboost::gamboost(
        formula, data = as.data.frame(formula, data = data),
        na.action = na.pass, weights = weights, family = family,
        baselearner = baselearner, dfbase = dfbase,
        control = mboost::boost_control(...)
      )
    },

    predict = function(object, newdata, .MachineShop, ...) {
      weights <- case_weights(.MachineShop$input)
      newdata <- as.data.frame(newdata)
      if (object$family@name == "Cox Partial Likelihood") {
        lp <- drop(predict(object, type = "link"))
        new_lp <- drop(predict(object, newdata = newdata, type = "link"))
        predict(object$response, lp, new_lp, weights = weights, ...)
      } else {
        predict(object, newdata = newdata, type = "response")
      }
    },

    varimp = function(object, ...) {
      structure(mboost::varimp(object), class = "numeric", metric = "risk")
    }

  )

}

MLModelFunction(GAMBoostModel) <- NULL

ModelRecipe <- function(object, ...) {
  UseMethod("ModelRecipe")
}


ModelRecipe.ModelRecipe <- function(object, ...) {
  object
}


ModelRecipe.recipe <- function(object, ...) {
  if (any(map("logi", function(step) isTRUE(step$trained), object$steps))) {
    throw(Error("Recipe must be untrained."))
  }

  cases_name <- "(names)"
  cases_fo <- ~ -`(names)`

  reserved <- intersect(
    c("(groups)", "(names)", "(strata)"), summary(object)$variable
  )
  if (length(reserved)) {
    throw(Error(note_items(
      "Supplied recipe contains reserved variable{?s}: ", reserved, "."
    )))
  }
  cases_info <- data.frame(
    variable = cases_name,
    type = "nominal",
    role = "case_name",
    source = "original"
  )
  object$var_info <- rbind(object$var_info, cases_info)
  object$term_info <- rbind(object$term_info, cases_info)
  object$template[[cases_name]] <- rownames(object$template)

  for (i in seq_along(object$steps)) {
    step_terms <- object$steps[[i]]$terms
    environment(cases_fo) <- environment(step_terms[[1]])
    new_term <- rlang::as_quosure(cases_fo)
    object$steps[[i]]$terms <- c(step_terms, new_term)
  }

  new("ModelRecipe", object)
}


bake.ModelRecipe <- function(object, new_data, newdata = NULL, ...) {
  if (missing(new_data)) {
    new_data <- if (is.null(newdata)) {
      as.data.frame(object)
    } else if (is(newdata, "ModelRecipe")) {
      as.data.frame(newdata)
    } else {
      prep_recipe_data(newdata)
    }
  }
  bake(as(object, "recipe"), new_data)
}


bake.SelectedInput <- function(object, ...) {
  throw(Error("Cannot create a design matrix from a ", class(object), "."))
}


bake.TunedInput <- function(object, ...) {
  throw(Error("Cannot create a design matrix from a ", class(object), "."))
}


prep.ModelFrame <- function(x, ...) x


prep.ModelRecipe <- function(x, retain = TRUE, ...) {
  if (!is_trained(x)) {
    template <- x$template
    x <- new(class(x), prep(as(x, "recipe"), retain = retain))
    x$orig_template <- template
    x$orig_lvls[["(names)"]] <- list(values = NA, ordered = NA)
    x$levels[["(names)"]] <- x$orig_lvls[["(names)"]]
  } else if (!retain) {
    x$template <- x$template[NULL, ]
    x$retained <- FALSE
  } else if (!x$retained) {
    throw(Error(
      "ModelRecipe prep cannot be updated from `retain = FALSE` to `TRUE`."
    ))
  }
  x
}


prep.SelectedInput <- function(x, ...) {
  throw(Error("Cannot train a ", class(x), "."))
}


prep.TunedInput <- function(x, ...) {
  throw(Error("Cannot train a ", class(x), "."))
}


prep_recipe_data <- function(x) {
  if (is.null(x[["(names)"]])) x[["(names)"]] <- rownames(x)
  x[c("(groups)", "(strata)")] <- NULL
  as_tibble(x)
}


update.ModelRecipe <- function(
  object, params = NULL, data = NULL, id = object@id, ...
) {
  stopifnot(!is_trained(object))
  if (is.list(params)) {
    for (i in seq_along(object$steps)) {
      step <- object$steps[[i]]
      step_params <- params[[step$id]]
      if (length(step_params)) {
        object$steps[[i]] <- do.call(update, c(list(step), step_params))
      }
    }
  }
  if (is.data.frame(data)) object$template <- prep_recipe_data(data)
  object@id <- id
  object
}

#' Proportional Hazards Regression Model
#'
#' Fits a Cox proportional hazards regression model. Time dependent variables,
#' time dependent strata, multiple events per subject, and other extensions are
#' incorporated using the counting process formulation of Andersen and Gill.
#'
#' @rdname CoxModel
#'
#' @param ties character string specifying the method for tie handling.
#' @param ... arguments passed to \code{\link[survival]{coxph.control}}.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{Surv}}
#' }
#'
#' Default argument values and further model details can be found in the source
#' See Also links below.
#'
#' In calls to \code{\link{varimp}} for \code{CoxModel} and
#' \code{CoxStepAICModel}, numeric argument \code{base} may be specified for the
#' (negative) logarithmic transformation of p-values [defaul: \code{exp(1)}].
#' Transformed p-values are automatically scaled in the calculation of variable
#' importance to range from 0 to 100.  To obtain unscaled importance values, set
#' \code{scale = FALSE}.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[survival]{coxph}},
#' \code{\link[survival]{coxph.control}}, \code{\link[MASS]{stepAIC}},
#' \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' library(survival)
#'
#' fit(Surv(time, status) ~ ., data = veteran, model = CoxModel)
#'
CoxModel <- function(ties = c("efron", "breslow", "exact"), ...) {

  ties <- match.arg(ties)

  MLModel(

    name = "CoxModel",
    label = "Cox Regression",
    packages = "survival",
    response_types = "Surv",
    weights = TRUE,
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment(), ...),

    fit = function(formula, data, weights, ...) {
      survival::coxph(
        formula, data = as.data.frame(formula, data = data), weights = weights,
        na.action = na.pass, ...
      )
    },

    predict = function(object, newdata, .MachineShop, ...) {
      y <- object$y
      weights <- case_weights(.MachineShop$input)
      newdata <- as.data.frame(newdata)
      lp <- predict(object, type = "lp")
      new_lp <- predict(object, newdata = newdata, type = "lp")
      predict(y, lp, new_lp, weights = weights, ...)
    },

    varimp = function(object, base = exp(1), ...) {
      varimp_pval(object, base = base)
    }

  )

}

MLModelFunction(CoxModel) <- NULL


#' @rdname CoxModel
#'
#' @param direction mode of stepwise search, can be one of \code{"both"}
#'   (default), \code{"backward"}, or \code{"forward"}.
#' @param scope defines the range of models examined in the stepwise search.
#'   This should be a list containing components \code{upper} and \code{lower},
#'   both formulae.
#' @param k multiple of the number of degrees of freedom used for the penalty.
#'   Only \code{k = 2} gives the genuine AIC; \code{k = .(log(nobs))} is
#'   sometimes referred to as BIC or SBC.
#' @param trace if positive, information is printed during the running of
#'   \code{stepAIC}. Larger values may give more information on the fitting
#'   process.
#' @param steps maximum number of steps to be considered.
#'
CoxStepAICModel <- function(
  ties = c("efron", "breslow", "exact"), ...,
  direction = c("both", "backward", "forward"), scope = list(), k = 2,
  trace = FALSE, steps = 1000
) {

  direction <- match.arg(direction)

  params <- new_params(environment())
  stepmodel <- CoxModel(ties = ties, ...)
  params <- params[setdiff(names(params), names(stepmodel@params))]

  MLModel(

    name = "CoxStepAICModel",
    label = "Cox Regression (Stepwise)",
    packages = c(stepmodel@packages, "MASS"),
    response_types = stepmodel@response_types,
    weights = stepmodel@weights,
    predictor_encoding = stepmodel@predictor_encoding,
    na.rm = stepmodel@na.rm,
    params = c(stepmodel@params, params),

    fit = function(
      formula, data, weights, direction, scope = list(), k, trace, steps, ...
    ) {
      data <- as.data.frame(formula, data = data)
      stepargs <- stepAIC_args(formula, direction, scope)
      MASS::stepAIC(
        survival::coxph(
          stepargs$formula, data = data, weights = weights, na.action = na.pass,
          ...
        ),
        direction = direction, scope = stepargs$scope, k = k, trace = trace,
        steps = steps
      )
    },

    predict = stepmodel@predict,

    varimp = stepmodel@varimp

  )

}

MLModelFunction(CoxStepAICModel) <- NULL

#' Conditional Random Forest Model
#'
#' An implementation of the random forest and bagging ensemble algorithms
#' utilizing conditional inference trees as base learners.
#'
#' @param teststat character specifying the type of the test statistic to be
#'   applied.
#' @param testtype character specifying how to compute the distribution of the
#'   test statistic.
#' @param mincriterion  value of the test statistic that must be exceeded in
#'   order to implement a split.
#' @param replace logical indicating whether sampling of observations is done
#'   with or without replacement.
#' @param fraction fraction of number of observations to draw without
#'   replacement (only relevant if \code{replace = FALSE}).
#' @param ntree number of trees to grow in a forest.
#' @param mtry number of input variables randomly sampled as candidates at each
#'   node for random forest like algorithms.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric}, \code{Surv}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameter:}{
#'     \code{mtry}
#'   }
#' }
#'
#' Supplied arguments are passed to \code{\link[party]{cforest_control}}.
#' Further model details can be found in the source link below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[party]{cforest}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' fit(sale_amount ~ ., data = ICHomes, model = CForestModel)
#'
CForestModel <- function(
  teststat = c("quad", "max"),
  testtype = c("Univariate", "Teststatistic", "Bonferroni", "MonteCarlo"),
  mincriterion = 0, ntree = 500, mtry = 5, replace = TRUE, fraction = 0.632
) {

  teststat <- match.arg(teststat)
  testtype <- match.arg(testtype)

  MLModel(

    name = "CForestModel",
    label = "Conditional Random Forests",
    packages = "party",
    response_types = c("factor", "numeric", "Surv"),
    weights = TRUE,
    predictor_encoding = "model.frame",
    na.rm = "response",
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = "mtry",
      get_values = c(
        function(n, data, ...) seq_nvars(data, CForestModel, n)
      )
    ),

    fit = function(formula, data, weights, ...) {
      party::cforest(
        formula, data = as.data.frame(formula, data = data), weights = weights,
        controls = party::cforest_control(...)
      )
    },

    predict = function(object, newdata, .MachineShop, ...) {
      newdata <- as.data.frame(newdata)
      if (object@responses@is_censored) {
        y <- response(.MachineShop$input)
        fits <- predict(object, newdata = newdata, type = "prob")
        predict(y, fits, ...)
      } else {
        predict(object, newdata = newdata, type = "prob") %>%
          unlist %>%
          matrix(nrow = nrow(newdata), byrow = TRUE)
      }
    },

    varimp = function(object, .MachineShop, ...) {
      structure(
        party::varimp(object, ...),
        metric = if (object@responses@is_censored) "brier" else "accuracy"
      )
    }

  )

}

MLModelFunction(CForestModel) <- NULL

#' Multivariate Adaptive Regression Splines Model
#'
#' Build a regression model using the techniques in Friedman's papers
#' "Multivariate Adaptive Regression Splines" and "Fast MARS".
#'
#' @param pmethod pruning method.
#' @param trace level of execution information to display.
#' @param degree maximum degree of interaction.
#' @param nprune maximum number of terms (including intercept) in the pruned
#'   model.
#' @param nfold number of cross-validation folds.
#' @param ncross number of cross-validations if \code{nfold > 1}.
#' @param stratify logical indicating whether to stratify cross-validation
#'   samples by the response levels.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \code{nprune}, \code{degree}*
#'   }
#' }
#' * excluded from grids by default
#'
#' Default argument values and further model details can be found in the source
#' See Also link below.
#'
#' In calls to \code{\link{varimp}} for \code{EarthModel}, argument
#' \code{type} may be specified as \code{"nsubsets"} (default) for the number of
#' model subsets that include each predictor, as \code{"gcv"} for the
#' generalized cross-validation decrease over all subsets that include each
#' predictor, or as \code{"rss"} for the residual sums of squares decrease.
#' Variable importance is automatically scaled to range from 0 to 100.  To
#' obtain unscaled importance values, set \code{scale = FALSE}.  See example
#' below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[earth]{earth}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package earth to run
#'
#' model_fit <- fit(Species ~ ., data = iris, model = EarthModel)
#' varimp(model_fit, method = "model", type = "gcv", scale = FALSE)
#' }
#'
EarthModel <- function(
  pmethod = c("backward", "none", "exhaustive", "forward", "seqrep", "cv"),
  trace = 0, degree = 1, nprune = integer(), nfold = 0, ncross = 1,
  stratify = TRUE
) {

  pmethod <- match.arg(pmethod)

  MLModel(

    name = "EarthModel",
    label = "Multivariate Adaptive Regression Splines",
    packages = "earth",
    response_types = c("factor", "numeric"),
    weights = TRUE,
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = c("nprune", "degree"),
      get_values = c(
        function(n, data, ...) {
          model_fit <- fit(data, model = EarthModel(pmethod = "none"))
          max_terms <- min(2 + 0.75 * nrow(model_fit$dirs), 200)
          round_int(seq(2, max_terms, length = n))
        },
        function(n, ...) head(1:2, n)
      ),
      default = c(TRUE, FALSE)
    ),

    fit = function(formula, data, weights, ...) {
      attach_objects(list(
        contr.earth.response = earth::contr.earth.response
      ), name = "earth_exports")

      glm <- list(
        family = switch_class(response(data),
          "factor" = "binomial",
          "numeric" = "gaussian"
        )
      )
      eval_fit(
        data,
        formula = earth::earth(
          formula, data = data, weights = weights, glm = glm, ...
        ),
        matrix = earth::earth(x, y, weights = weights, glm = glm, ...)
      )
    },

    predict = function(object, newdata, ...) {
      newdata <- as.data.frame(newdata)
      predict(object, newdata = newdata, type = "response")
    },

    varimp = function(object, type = c("nsubsets", "gcv", "rss"), ...) {
      type <- match.arg(type)
      vi <- earth::evimp(object)
      structure(
        vi[, type], names = rownames(vi),
        metric = switch(type, "rss" = "rss/sse", type)
      )
    }

  )

}

MLModelFunction(EarthModel) <- NULL

PartialDependence <- function(object) {
  structure(object, class = c("PartialDependence", "data.frame"))
}


#' Partial Dependence
#'
#' Calculate partial dependence of a response on select predictor variables.
#'
#' @param object model \link{fit} result.
#' @param data \link[=data.frame]{data frame} containing all predictor
#'   variables.  If not specified, the training data will be used by default.
#' @param select expression indicating predictor variables for which to compute
#'   partial dependence (see \code{\link[base]{subset}} for syntax)
#'   [default: all].
#' @param interaction logical indicating whether to calculate dependence on the
#'   interacted predictors.
#' @param n number of predictor values at which to perform calculations.
#' @param intervals character string specifying whether the \code{n} values are
#'   spaced uniformly (\code{"uniform"}) or according to variable quantiles
#'   (\code{"quantile"}).
#' @param distr,method arguments passed to \code{\link{predict}}.
#' @param stats function, function name, or vector of these with which to
#'   compute response variable summary statistics over non-selected predictor
#'   variables.
#' @param na.rm logical indicating whether to exclude missing predicted response
#'   values from the calculation of summary statistics.
#'
#' @return \code{PartialDependence} class object that inherits from
#' \code{data.frame}.
#'
#' @seealso \code{\link{plot}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' gbm_fit <- fit(Species ~ ., data = iris, model = GBMModel)
#' (pd <- dependence(gbm_fit, select = c(Petal.Length, Petal.Width)))
#' plot(pd)
#' }
#'
dependence <- function(
  object, data = NULL, select = NULL, interaction = FALSE, n = 10,
  intervals = c("uniform", "quantile"), distr = character(),
  method = character(),
  stats = MachineShop::settings("stats.PartialDependence"), na.rm = TRUE
) {

  stopifnot(is(object, "MLModelFit"))

  input <- as.MLInput(object)
  if (is.null(data)) data <- input
  data <- as.data.frame(data)
  pred_names <- all.vars(predictors(terms(input, original = TRUE)))
  pred_names <- do.call(
    subset_names, list(pred_names, substitute(select)), envir = parent.frame()
  )

  data_select <- data[, pred_names, drop = FALSE]

  intervals <- match.arg(intervals)

  stats <- check_stats(stats, convert = TRUE)
  throw(check_assignment(stats))

  select_values <- function(x) {
    if (is(x, "factor")) {
      unique(x)
    } else if (is(x, "logical")) {
      intersect(c(FALSE, TRUE), x)
    } else if (is(x, "numeric") || is(x, "Date")) {
      x <- sort(x)
      n <- min(n, length(x))
      unique(switch(intervals,
        "quantile" = x[round(seq(1, length(x), length = n))],
        "uniform" = {
          y <- seq(x[1], x[length(x)], length = n)
          indices <- findInterval(y, x, all.inside = TRUE)
          x_lower <- x[indices]
          x_upper <- x[indices + 1]
          ifelse(y - x_lower < x_upper - y, x_lower, x_upper)
        }
      ))
    } else {
      throw(TypeError(
        x, c("Date", "factor", "logical", "numeric"), "selected variable"
      ))
    }
  }

  predict_stats <- function(data) {
    stats_list <- map(
      function(x) stats(if (na.rm) na.omit(x) else x),
      as.data.frame(
        predict(
          object, newdata = data, type = "raw", distr = distr, method = method
        )
      )
    )
    x <- do.call(cbind, stats_list)
    if (is.null(rownames(x))) rownames(x) <- make_unique(rep("stat", nrow(x)))
    if (is.null(colnames(x))) colnames(x) <- make_unique(rep("y", ncol(x)))
    names(dimnames(x)) <- c("Statistic", "Response")
    as.data.frame(TabularArray(x))
  }

  grid_list <- map(select_values, data_select)

  data_select_grid <- if (interaction) {
    expand.grid(grid_list, KEEP.OUT.ATTRS = FALSE, stringsAsFactors = FALSE)
  } else {
    df <- data.frame(row.names = seq_len(sum(lengths(grid_list))))
    start <- 1
    for (name in names(grid_list)) {
      n <- length(grid_list[[name]])
      df[[name]] <- rep_len(grid_list[[name]], nrow(df))
      df[[name]][-seq(start, length = n)] <- NA
      start <- start + n
    }
    df
  }

  dependence_list <- map(function(i) {
    for (varname in names(data_select_grid)) {
      x <- data_select_grid[i, varname]
      if (!is.na(x)) data[[varname]] <- x
    }
    df <- predict_stats(data)
    df$Predictors <- data_select_grid[rep(i, nrow(df)), , drop = FALSE]
    df
  }, seq_len(nrow(data_select_grid)))

  PartialDependence(do.call(rbind, dependence_list))

}

#' Boosting with Classification Trees
#'
#' Fits the AdaBoost.M1 (Freund and Schapire, 1996) and SAMME (Zhu et al., 2009)
#' algorithms using classification trees as single classifiers.
#'
#' @param boos if \code{TRUE}, then bootstrap samples are drawn from the
#'   training set using the observation weights at each iteration.  If
#'   \code{FALSE}, then all observations are used with their weights.
#' @param mfinal number of iterations for which boosting is run.
#' @param coeflearn learning algorithm.
#' @param minsplit minimum number of observations that must exist in a node in
#'   order for a split to be attempted.
#' @param minbucket minimum number of observations in any terminal node.
#' @param cp complexity parameter.
#' @param maxcompete number of competitor splits retained in the output.
#' @param maxsurrogate number of surrogate splits retained in the output.
#' @param usesurrogate how to use surrogates in the splitting process.
#' @param xval number of cross-validations.
#' @param surrogatestyle controls the selection of a best surrogate.
#' @param maxdepth maximum depth of any node of the final tree, with the root
#'   node counted as depth 0.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \code{mfinal}, \code{maxdepth}, \code{coeflearn}*
#'   }
#' }
#' * excluded from grids by default
#'
#' Further model details can be found in the source link below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[adabag]{boosting}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package adabag to run
#'
#' fit(Species ~ ., data = iris, model = AdaBoostModel(mfinal = 5))
#' }
#'
AdaBoostModel <- function(
  boos = TRUE, mfinal = 100, coeflearn = c("Breiman", "Freund", "Zhu"),
  minsplit = 20, minbucket = round(minsplit/3), cp = 0.01, maxcompete = 4,
  maxsurrogate = 5, usesurrogate = 2, xval = 10, surrogatestyle = 0,
  maxdepth = 30
) {

  coeflearn <- match.arg(coeflearn)

  MLModel(

    name = "AdaBoostModel",
    label = "Boosting with Classification Trees",
    packages = "adabag",
    response_types = "factor",
    predictor_encoding = "model.frame",
    na.rm = "response",
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = c("mfinal", "maxdepth", "coeflearn"),
      get_values = c(
        function(n, ...) round_int(seq_range(0, 25, c(1, 200), n + 1)),
        function(n, ...) seq_len(min(n, 30)),
        function(n, ...) head(c("Breiman", "Freund", "Zhu"), n)
      ),
      default = c(TRUE, TRUE, FALSE)
    ),

    fit = function(formula, data, weights, boos, mfinal, coeflearn, ...) {
      adabag::boosting(
        formula, data = as.data.frame(formula, data = data), boos = boos,
        mfinal = mfinal, coeflearn = coeflearn, control = list(...)
      )
    },

    predict = function(object, newdata, ...) {
      newdata <- as.data.frame(newdata)
      predict(object, newdata = newdata)$prob
    },

    varimp = function(object, ...) {
      structure(object$importance, metric = "gini")
    }

  )

}

MLModelFunction(AdaBoostModel) <- NULL

#' Parametric Survival Model
#'
#' Fits the accelerated failure time family of parametric survival models.
#'
#' @rdname SurvRegModel
#'
#' @param dist assumed distribution for y variable.
#' @param scale optional fixed value for the scale.
#' @param parms list of fixed parameters.
#' @param ... arguments passed to \code{\link[survival]{survreg.control}}.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{Surv}}
#' }
#'
#' Default argument values and further model details can be found in the source
#' See Also links below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[rms]{psm}}, \code{\link[survival]{survreg}},
#' \code{\link[survival]{survreg.control}}, \code{\link[MASS]{stepAIC}},
#' \code{\link{fit}}, \code{\link{resample}}
#'
SurvRegModel <- function(
  dist = c("weibull", "exponential", "gaussian", "logistic", "lognormal",
           "logloglogistic"),
  scale = 0, parms = list(), ...
) {

  dist <- match.arg(dist)

  MLModel(

    name = "SurvRegModel",
    label = "Parametric Survival",
    packages = c("rms", "Hmisc"),
    response_types = "Surv",
    weights = TRUE,
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment(), ...),

    fit = function(formula, data, weights, dist, scale, parms = NULL, ...) {
      rms::psm(
        formula, data = as.data.frame(formula, data = data), weights = weights,
        na.action = na.pass, dist = dist, scale = scale, parms = parms,
        control = survival::survreg.control(...)
      )
    },

    predict = function(object, newdata, times, ...) {
      newdata <- as.data.frame(newdata)
      pred <- if (length(times)) {
        rms::survest(object, newdata = newdata, times = times, conf.int = FALSE)
      } else {
        Hmisc::Mean(object)(predict(object, newdata = newdata, type = "lp"))
      }
      if (is(pred, "survest.psm")) pred <- as.matrix(pred$surv)
      SurvPrediction(pred, times = times, distr = object$dist)
    },

    varimp = function(object, base = exp(1), ...) {
      varimp_pval(object, base = base)
    }

  )

}

MLModelFunction(SurvRegModel) <- NULL


#' @rdname SurvRegModel
#'
#' @param direction mode of stepwise search, can be one of \code{"both"}
#'   (default), \code{"backward"}, or \code{"forward"}.
#' @param scope defines the range of models examined in the stepwise search.
#'   This should be a list containing components \code{upper} and \code{lower},
#'   both formulae.
#' @param k multiple of the number of degrees of freedom used for the penalty.
#'   Only \code{k = 2} gives the genuine AIC; \code{k = .(log(nobs))} is
#'   sometimes referred to as BIC or SBC.
#' @param trace if positive, information is printed during the running of
#'   \code{stepAIC}. Larger values may give more information on the fitting
#'   process.
#' @param steps maximum number of steps to be considered.
#'
#' @seealso \code{\link[MASS]{stepAIC}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested packages rms and Hmisc to run
#'
#' library(survival)
#'
#' fit(Surv(time, status) ~ ., data = veteran, model = SurvRegModel)
#' }
#'
SurvRegStepAICModel <- function(
  dist = c(
    "weibull", "exponential", "gaussian", "logistic", "lognormal",
    "logloglogistic"
  ),
  scale = 0, parms = list(), ...,
  direction = c("both", "backward", "forward"), scope = list(), k = 2,
  trace = FALSE, steps = 1000
) {

  direction <- match.arg(direction)

  params <- new_params(environment())
  stepmodel <- SurvRegModel(dist = dist, scale = scale, parms = parms, ...)
  params <- params[setdiff(names(params), names(stepmodel@params))]

  MLModel(

    name = "SurvRegStepAICModel",
    label = "Parametric Survival (Stepwise)",
    packages = c(stepmodel@packages, "MASS"),
    response_types = stepmodel@response_types,
    weights = stepmodel@weights,
    predictor_encoding = stepmodel@predictor_encoding,
    na.rm = stepmodel@na.rm,
    params = c(stepmodel@params, params),

    fit = function(
      formula, data, weights, dist, scale, parms = NULL, ...,
      direction, scope = list(), k, trace, steps
    ) {
      data <- as.data.frame(formula, data = data)
      stepargs <- stepAIC_args(formula, direction, scope)
      MASS::stepAIC(
        rms::psm(
          stepargs$formula, data = data, weights = weights,
          na.action = na.pass, dist = dist, scale = scale, parms = parms,
          control = survival::survreg.control(...)
        ),
        direction = direction, scope = stepargs$scope, k = k, trace = trace,
        steps = steps
      )
    },

    predict = stepmodel@predict,

    varimp = stepmodel@varimp

  )

}

MLModelFunction(SurvRegStepAICModel) <- NULL

#' Model Prediction
#'
#' Predict outcomes with a fitted model.
#'
#' @name predict
#'
#' @param object model \link{fit} result.
#' @param newdata optional \link[=data.frame]{data frame} with which to obtain
#'   predictions.  If not specified, the training data will be used by default.
#' @param times numeric vector of follow-up times at which to predict
#'   survival events/probabilities or \code{NULL} for predicted survival means.
#' @param type specifies prediction on the original outcome (\code{"response"}),
#'   numeric (\code{"numeric"}), or probability (\code{"prob"}) scale; or
#'   the \code{"raw"} predictions returned by the model.  Option
#'   \code{"default"} is deprecated and will be removed in the future; use
#'   \code{"raw"} instead.
#' @param cutoff numeric (0, 1) threshold above which binary factor
#'   probabilities are classified as events and below which survival
#'   probabilities are classified.
#' @param distr character string specifying distributional approximations to
#'   estimated survival curves.  Possible values are \code{"empirical"},
#'   \code{"exponential"}, \code{"rayleigh"}, or \code{"weibull"}; with defaults
#'   of \code{"empirical"} for predicted survival events/probabilities and
#'   \code{"weibull"} for predicted survival means.
#' @param method character string specifying the empirical method of estimating
#'   baseline survival curves for Cox proportional hazards-based models.
#'   Choices are \code{"breslow"} or \code{"efron"} (default).
#' @param verbose logical indicating whether to display printed output generated
#'   by some model-specific predict functions to aid in monitoring progress and
#'   diagnosing errors.
#' @param ... arguments passed from the S4 to the S3 method.
#'
#' @seealso \code{\link{confusion}}, \code{\link{performance}},
#' \code{\link{metrics}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' ## Survival response example
#' library(survival)
#'
#' gbm_fit <- fit(Surv(time, status) ~ ., data = veteran, model = GBMModel)
#' predict(gbm_fit, newdata = veteran, times = c(90, 180, 360), type = "prob")
#' }
#'
NULL


#' @name predict
#' @method predict MLModelFit
#'
predict.MLModelFit <- function(
  object, newdata = NULL, times = numeric(),
  type = c("response", "raw", "numeric", "prob", "default"),
  cutoff = MachineShop::settings("cutoff"), distr = character(),
  method = character(), verbose = FALSE, ...
) {
  object <- update(object)
  model <- as.MLModel(object)
  throw(check_packages(model@packages))

  newdata <- PredictorFrame(as.MLInput(object), newdata)
  newnames <- rownames(newdata)
  newdata <- ModelFrame(newdata, na.rm = model@na.rm)
  if (!nrow(newdata)) throw(Error("No case observations to predict."))
  subset <- newnames %in% rownames(newdata)

  times <- check_numeric(
    times, bounds = c(0, Inf), include = FALSE, size = NA, nonempty = FALSE
  )
  throw(check_assignment(times))

  type <- match.arg(type)
  if (type == "default") {
    throw(DeprecatedCondition("type = \"default\"", "type = \"raw\""))
    type <- "raw"
  }

  .MachineShop <- attr(object, ".MachineShop")
  (if (verbose) identity else capture.output)(
    pred <- .predict(
      model, model_fit = unMLModelFit(object), newdata = newdata, times = times,
      distr = distr, method = method, .MachineShop = .MachineShop, ...
    )
  )
  obs <- response(object)
  pred <- convert_predicted(obs, pred)
  pred <- switch(type,
    "raw" = pred,
    "numeric" = convert_numeric(pred),
    "prob" = convert_numeric(pred, bounds = c(0, 1)),
    "response" = convert_response(obs, pred, cutoff = cutoff)
  )
  throw(check_assignment(pred))
  na.add(pred, !subset)
}


#' @name predict
#' @aliases predict,MLModelFit-method
#'
setMethod("predict", "MLModelFit",
  function(object, ...) predict.MLModelFit(object, ...)
)


.predict <- function(object, ...) {
  UseMethod(".predict")
}


.predict.MLModel <- function(object, model_fit, ...) {
  object@predict(model_fit, ...)
}


PredictorFrame <- function(input, newdata = NULL) {
  if (!is(newdata, "PredictorFrame")) {
    new("PredictorFrame",
      ModelFrame(
        delete.response(terms(input)), predictors(input, newdata), na.rm = FALSE
      )
    )
  } else {
    newdata
  }
}


predictors <- function(object, ...) {
  UseMethod("predictors")
}


predictors.formula <- function(object, data = NULL, ...) {
  if (is.null(data)) {
    object[[length(object)]]
  } else {
    data[, all.vars(predictors(object)), drop = FALSE]
  }
}


predictors.ModelFrame <- function(object, newdata = NULL, ...) {
  data <- as.data.frame(if (is.null(newdata)) object else newdata)
  predictors(terms(object), data)
}


predictors.recipe <- function(object, newdata = NULL, ...) {
  object <- prep(object, retain = FALSE)
  data <- bake(object, newdata = newdata)
  info <- summary(object)
  pred_names <- info$variable[info$role %in% c("predictor", "pred_offset")]
  data[, pred_names, drop = FALSE]
}

#' Print MachineShop Objects
#'
#' Print methods for objects defined in the \pkg{MachineShop} package.
#'
#' @name print
#' @rdname print-methods
#'
#' @param x object to print.
#' @param n integer number of models or data frame rows to show.
#' @param id logical indicating whether to show object identifiers.
#' @param data logical indicating whether to show model data.
#' @param ... arguments passed to other methods, including the one described
#'   below.
#'   \describe{
#'     \item{\code{level} = 0}{current nesting level of the corresponding
#'     object in recursive calls to \code{print}.  The amount of information
#'     displayed decreases and increases with positive and negative levels,
#'     respectively.}
#'   }
#'
NULL


setShowDefault <- function(signature) {
  setMethod("show", signature, function(object) {
    print(object)
    invisible()
  })
}


#' @rdname print-methods
#'
print.BinomialVariate <- function(
  x, n = MachineShop::settings("print_max"), ...
) {
  title(x, pad = FALSE, ...)
  print_items(as(x, "matrix"), n = n)
  invisible(x)
}


#' @rdname print-methods
#'
print.Calibration <- function(x, n = MachineShop::settings("print_max"), ...) {
  title(x, ...)
  print_items(as(x, "data.frame"), n = n)
  invisible(x)
}


setShowDefault("Calibration")


print.ConfusionList <- function(
  x, n = MachineShop::settings("print_max"), ...
) {
  title(x, ...)
  level <- nesting_level(...)
  NextMethod(level = level + 1)
}


print.ConfusionMatrix <- function(x, ...) {
  title(x, pad = FALSE, ...)
  print(as(x, "table"))
  invisible(x)
}


setShowDefault("ConfusionMatrix")


print.ConfusionSummary <- function(x, ...) {
  acc <- x@accuracy
  acc_se <- sqrt(acc * (1 - acc) / x@total)
  print_fields(list(
    "Number of responses: " = x@total,
    "Accuracy (SE): " = paste0(format(acc), " (", format(acc_se), ")"),
    "Majority class: " = x@majority,
    "Kappa: " = x@kappa2
  ))
  newline()
  print(as(x, "matrix"))
  invisible(x)
}


setShowDefault("ConfusionSummary")


#' @rdname print-methods
#'
print.DiscreteVariate <- function(
  x, n = MachineShop::settings("print_max"), ...
) {
  title(x, pad = FALSE, ...)
  print_items(as(x, "numeric"), n = n)
  print_fields(list("Range: " = c(x@min, x@max)))
  invisible(x)
}


setShowDefault("DiscreteVariate")


print.EnsembleModel <- function(
  x, n = MachineShop::settings("print_max"), id = FALSE, ...
) {
  title(x, ...)
  level <- nesting_level(...)
  nextlevel <- level + 1
  trained <- is_trained(x)
  print_modelinfo(x, trained = trained, ...)
  if (level < 1) {
    hline(level)
    print_train_label("Models:")
    newline()
    print(x@candidates, n = n, level = nextlevel, id = id)
    hline(level)
    print(x@params, n = n, level = nextlevel)
    if (trained) {
      hline(level)
      print(x@steps, n = n, level = nextlevel)
    }
  } else if (level == 1) {
    newline()
    print_train_label("Models:")
    print(x@candidates, n = n, level = nextlevel, id = id)
  }
  invisible(x)
}


#' @rdname print-methods
#'
print.ListOf <- function(x, n = MachineShop::settings("print_max"), ...) {
  print_items(x, n = n, ...)
  invisible(x)
}


setShowDefault("ListOf")


setShowDefault("MLControl")


#' @rdname print-methods
#'
print.MLControl <- function(x, n = MachineShop::settings("print_max"), ...) {
  print_control(x, n = n, ...)
}


print.BootControl <- function(x, ...) {
  fields <- list("Samples: " = x@samples)
  NextMethod(fields = fields)
}


print.CVControl <- function(x, ...) {
  fields <- list("Folds: " = x@folds, "Repeats: " = x@repeats)
  NextMethod(fields = fields)
}


print.NullControl <- function(x, ...) {
  title(x, ...)
  print_fields(list("Label: " = x@label))
  invisible(x)
}


print.OOBControl <- function(x, ...) {
  fields <- list("Samples: " = x@samples)
  NextMethod(fields = fields)
}


print.SplitControl <- function(x, ...) {
  fields = list("Training proportion: " = x@prop)
  NextMethod(fields = fields)
}


#' @rdname print-methods
#'
print.MLMetric <- function(x, ...) {
  title(x, ...)
  level <- nesting_level(...)
  print_fields(list(
    "Metric name: " = x@name,
    "Label: " = x@label,
    "Maximize: " = x@maximize
  ))
  if (level < 1) {
    info <- metricinfo(x)[[1]]
    newline()
    print_label("Arguments:")
    print(info$arguments)
    newline()
    print_label("Types:")
    print(info$response_types)
  }
  invisible(x)
}


setShowDefault("MLMetric")


#' @rdname print-methods
#'
print.MLModel <- function(
  x, n = MachineShop::settings("print_max"), id = FALSE, ...
) {
  title(x, ...)
  level <- nesting_level(...)
  trained <- is_trained(x)
  print_modelinfo(x, trained = trained, id = id, ...)
  if (level < 1) {
    newline()
    print_str(x@params, label = "Parameter{?s}:", n = n)
    if (trained) {
      newline()
      print(x@steps, n = n, level = level + 1)
    }
  }
  invisible(x)
}


setShowDefault("MLModel")


print.MLModelFit <- function(x, ...) {
  print(unMLModelFit(x), ...)
  invisible(x)
}


setShowDefault("MLModelFit")


#' @rdname print-methods
#'
print.MLModelFunction <- function(x, ...) {
  title(x, ...)
  level <- nesting_level(...)
  print_modelinfo(x, ...)
  if (level < 1) {
    newline()
    print_label("Arguments:")
    print(args(x))
  }
  invisible(x)
}


setShowDefault("MLModelFunction")


print.MLOptimization <- function(
  x, n = MachineShop::settings("print_max"), ...
) {
  title(x, ...)
  level <- nesting_level(...)
  print_fields(list("Label: " = x@label))
  if (level < 1) {
    print_fields(list(
      "Package{?s}: " = unname(x@packages),
      "Monitor: " = x@monitor
    ), add_names = TRUE)
    if (length(x@params)) print_str(x@params, label = "Parameter{?s}:", n = n)
  }
  invisible(x)
}


setShowDefault("MLOptimization")


#' @rdname print-methods
#'
print.ModelFrame <- function(
  x, n = MachineShop::settings("print_max"), id = FALSE, data = TRUE, ...
) {
  title(x, ...)
  level <- nesting_level(...)
  if (id) print_id(x)
  label <- if (level < 2) style_label("Terms: ")
  print_items(formula(x), n = n, label = label)
  if (data) print_data(as(x, "data.frame"), n = n, level = level)
  invisible(x)
}


setShowDefault("ModelFrame")


#' @rdname print-methods
#'
print.ModelRecipe <- function(
  x, n = MachineShop::settings("print_max"), id = FALSE, data = TRUE, ...
) {
  title(x, ...)
  level <- nesting_level(...)
  if (id) print_id(x)
  tbl <- summary(x, original = TRUE)
  print_fields(list(
    "Outcome{?s}: " = tbl$variable[tbl$role == "outcome"],
    "Predictor{?s}: " = tbl$variable[tbl$role == "predictor"]
  ), n = n)
  if (level < 2) {
    steps <- map("char", class1, x$steps)
    step_label <- paste(if (is_trained(x)) "Trained" else "Untrained", "step")
    if (id) {
      steps <- structure(map("char", getElement, x$steps, "id"), names = steps)
      step_label <- paste(step_label, "ID")
    }
    print_fields(
      list(steps = steps), labels = c(steps = paste0(step_label, "{?s}: ")),
      add_names = TRUE, n = n
    )
  }
  if (level < 1) print_fields(list("Tuning grid: " = has_grid(x)))
  if (data) print_data(as_tibble(as.data.frame(x)), n = n, level = level)
  invisible(x)
}


setShowDefault("ModelRecipe")


#' @rdname print-methods
#'
print.ModelSpecification <- function(
  x, n = MachineShop::settings("print_max"), id = FALSE, ...
) {
  title(x, ...)
  level <- nesting_level(...)
  nextlevel <- level + 1
  if (id) {
    print_id(x)
    newline()
  }
  show_all <- length(x@grid) && level < 1
  print(x@input, n = n, level = nextlevel, id = show_all)
  if (!is(x@model, "NullModel")) {
    newline()
    print(x@model, n = n, level = nextlevel, id = show_all)
  }
  if (show_all) {
    newline()
    heading("Grid", level = nextlevel)
    print_items(x@grid, n = n)
    newline()
    print(x@params, n = n, level = nextlevel)
  }
}


setShowDefault("ModelSpecification")


print.ModelTerms <- function(x, n = MachineShop::settings("print_max"), ...) {
  print_items(formula(x), n = n)
  invisible(x)
}


print.ParameterGrid <- function(x, ...) {
  NextMethod()
  newline()
  print(as(x, "parameters"))
  invisible(x)
}


setShowDefault("ParameterGrid")


print.ParsnipModel <- function(x, ...) {
  x@params <- x@params$object$args
  NextMethod()
}


#' @rdname print-methods
#'
print.Performance <- function(x, n = MachineShop::settings("print_max"), ...) {
  title(x, ...)
  level <- nesting_level(...)
  dn <- dimnames(x)
  print_fields(list("Metric{?s}: " = dn[[2]]))
  if (length(dn) > 2) {
    print_fields(list("Model{?s}: " = dn[[3]]), n = n)
  }
  newline()
  print(x@control, level = level + 1)

  invisible(x)
}


#' @rdname print-methods
#'
print.PerformanceCurve <- function(
  x, n = MachineShop::settings("print_max"), ...
) {
  title(x, ...)
  level <- nesting_level(...)
  xy_labels <- c(x@metrics$x@label, x@metrics$y@label)
  print_fields(list("Metrics: " = paste(c("x", "y"), "=", xy_labels)))
  newline()
  print_items(as(x, "data.frame"), n = n)
  if (!is(x@control, "NullControl")) {
    newline()
    print(x@control, level = level + 1)
  }
  invisible(x)
}


setShowDefault("PerformanceCurve")


print.PerformanceDiffTest <- function(x, ...) {
  title(x, ...)
  print_fields(list(
    "Upper diagonal: " = "mean differences (Model1 - Model2)",
    "Lower diagonal: " = "p-values",
    "P-value adjustment method: " = x@adjust
  ))
  newline()
  print(as(x, "array"))
  invisible(x)
}


print.RandomGridSearch <- function(x, ...) {
  NextMethod()
  print_fields(list("Samples: " = x@size))
  invisible(x)
}


#' @rdname print-methods
#'
print.RecipeGrid <- function(x, n = MachineShop::settings("print_max"), ...) {
  title(x, ...)
  print_items(as(x, "tbl_df"), n = n)
  invisible(x)
}


setShowDefault("RecipeGrid")


#' @rdname print-methods
#'
print.Resample <- function(x, n = MachineShop::settings("print_max"), ...) {
  title(x, ...)
  level <- nesting_level(...)
  print_fields(list("Model{?s}: " = levels(x$Model)), n = n)
  if (level >= 1) {
    print_fields(list(
      "Grouping variable: " = names(x@vars[["Grouping"]]),
      "Stratification variable: " = names(x@vars[["Stratification"]])
    ))
  } else if (length(x@vars) > 1) {
    print_label("Sampling variables:")
    print_items(x@vars, n = n)
  }
  newline()
  print(x@control, n = n, level = level + 1)
  invisible(x)
}


setShowDefault("Resample")


print.SelectedInput <- function(
  x, n = MachineShop::settings("print_max"), id = FALSE, ...
) {
  NextMethod()
  level <- nesting_level(...)
  nextlevel <- level + 1
  if (level < 1) {
    hline(level)
    print_train_label("Selection set:")
    newline()
    print(x@candidates, n = n, level = nextlevel, id = id, data = FALSE)
    hline(level)
    print(x@params, n = n, level = nextlevel)
  } else if (level == 1) {
    newline()
    print_train_label("Selection set:")
    print(x@candidates, n = -n, level = nextlevel, id = id, data = FALSE)
  }
  invisible(x)
}


print.SequentialOptimization <- function(x, ...) {
  NextMethod()
  level <- nesting_level(...)
  if (level < 1) {
    print_fields(list("Fallback grid samples: " = if (x@random) x@random))
  }
  invisible(x)
}


format.SurvMatrix <- function(x, ...) {
  format(as(x, "matrix"), ...)
}


#' @rdname print-methods
#'
print.SurvMatrix <- function(x, n = MachineShop::settings("print_max"), ...) {
  title(x, pad = FALSE, ...)
  print_items(as(x, "matrix"), n = n)
  print_fields(list(
    "Time{?s}: " = x@times,
    "Distribution: " = x@distr
  ), n = n)
  invisible(x)
}


setShowDefault("SurvMatrix")


#' @rdname print-methods
#'
print.SurvTimes <- function(x, n = MachineShop::settings("print_max"), ...) {
  title(x, pad = FALSE, ...)
  print_items(as(x, "numeric"), n = n)
  print_fields(list("Distribution: " = x@distr))
  invisible(x)
}


setShowDefault("SurvTimes")


print.TabularArray <- function(x, ...) {
  print(as(x, "array"))
  invisible(x)
}


setShowDefault("TabularArray")


print.TrainingParams <- function(
  x, n = MachineShop::settings("print_max"), ...
) {
  title(x, ...)
  level <- nesting_level(...)
  nextlevel <- level + 1
  if (length(x@options)) {
    print_str(x@options, label = "Option{?s}:", n = n)
    newline()
  }
  if (is_optim_method(x)) {
    print(x@optim, n = n, level = nextlevel)
    newline()
    print(x@control, n = n, level = nextlevel)
    if (nextlevel < 0) {
      newline()
      print_label("Metrics:")
      print(x@metrics)
      newline()
      print_fields(list("Cutoff: " = x@cutoff))
      newline()
      print_label("Stat:")
      print(x@stat)
    }
  } else {
    print(x@control, n = n, level = nextlevel)
  }
  invisible(x)
}


setShowDefault("TrainingParams")


#' @rdname print-methods
#'
print.TrainingStep <- function(x, n = MachineShop::settings("print_max"), ...) {
  title(x, ...)
  level <- nesting_level(...)
  selected <- which(x@log$selected)
  print_fields(list("Optimization method: " = x@method))
  print_label(x@name, " log:")
  print_items(x@log, n = n)
  if (ndim(x@performance) > 2) {
    newline()
    print_fields(list(
      "Selected row: " = selected,
      "Metric: " = unlist(x@log$metrics[selected, 1])
    ), add_name = TRUE)
  }
  invisible(x)
}


setShowDefault("TrainingStep")


print.TunedInput <- function(x, n = MachineShop::settings("print_max"), ...) {
  NextMethod()
  level <- nesting_level(...)
  nextlevel <- level + 1
  if (level < 1) {
    newline()
    grid <- x@grid
    if (isS4(grid)) {
      print(grid, level = nextlevel)
    } else {
      heading("Grid", level = nextlevel)
      print_items(grid, n = n)
    }
    newline()
    print(x@params, n = n, level = nextlevel)
  }
  invisible(x)
}


print.TunedModel <- function(x, n = MachineShop::settings("print_max"), ...) {
  title(x, ...)
  level <- nesting_level(...)
  nextlevel <- level + 1
  trained <- is_trained(x)
  print_modelinfo(x, trained = trained, ...)
  if (level < 2) {
    newline()
    print(x@model, n = n, level = nextlevel)
    if (level < 1) {
      newline()
      grid <- x@grid
      if (isS4(grid)) {
        print(grid, level = nextlevel)
      } else {
        heading("Grid", level = nextlevel)
        print_items(grid, n = n)
      }
      newline()
      print(x@params, n = n, level = nextlevel)
      if (trained) {
        newline()
        print(x@steps, n = n, level = nextlevel)
      }
    }
  }
  invisible(x)
}


print.TuningGrid <- function(x, ...) {
  title(x, ...)
  print_fields(list("Grid size{?s}: " = x@size), add_names = TRUE)
  print_fields(list("Random samples: " = x@random))
  invisible(x)
}


setShowDefault("TuningGrid")


#' @rdname print-methods
#'
print.VariableImportance <- function(
  x, n = MachineShop::settings("print_max"), ...
) {
  x <- update(x)
  title(x, pad = FALSE, ...)
  print_fields(list(
    "Method: " = x@method,
    "Metric: " = x@metric,
    "Scale: " = x@scale
  ))
  print_label("Values:")
  print_items(as(x, "data.frame"), n = n)
  invisible(x)
}


setShowDefault("VariableImportance")


#################### Print Utility Functions ####################


format_len <- function(x) format(x, big.mark = ",")


heading <- function(x, level = 0, pad = TRUE, ...) {
  style <- if (level < 0) {
    make_ansi_style("magenta")
  } else if (level == 0) {
    make_ansi_style("green")
  }
  hline(level, label = style_bold(x), trim = level, style = style)
  if (level < 2 && pad) newline()
}


hline <- function(symbol = 0, label = character(), trim = 0, style = NULL) {
  if (is.numeric(symbol)) {
    symbols <- c("-", "=", ".", "_", "~")
    symbol <- symbols[symbol %% length(symbols) + 1]
  }
  freespace <- max(console_width(), 30)
  if (length(label)) {
    nsymbol <- 3
    str <- strrep(symbol, nsymbol)
    freespace <- freespace - nsymbol - 2
    str[2] <- style_reset(ansi_strtrim(label, freespace - nsymbol))
    if (trim < 1) {
      freespace <- freespace - ansi_nchar(str[2])
      str[3] <- strrep(symbol, freespace)
    } else if (trim == 1) {
      str[3] <- str[1]
    }
    str <- paste(str, collapse = " ")
  } else {
    str <- strrep(symbol, freespace)
  }
  if (!is.null(style)) str <- style(str)
  cat(str, "\n", sep = "")
}


nesting_level <- function(level = 0, ...) {
  level
}


new_print_progress <- function() {
  res <- function(...) NULL
  body(res) <- bquote(print_progress(..., start_time = .(Sys.time())))
  res
}


newline <- function() {
  cat("\n")
}


print_control <- function(x, fields = NULL, n = Inf, ...) {
  level <- nesting_level(...)
  title(x, ...)
  print_fields(c(
    list("Label: " = x@label),
    fields
  ))
  if (level < 1) {
    print_fields(list("Case-weighted metrics: " = x@weights))
    print_fields(x@predict, labels = c(
      times = "Survival time{?s}: ",
      distr = "Distribution: ",
      method = "Method: "
    ), "Prediction parameters", n = n)
    print_fields(x@strata, labels = c(
      breaks = "Breaks: ",
      nunique = "Unique numeric threshold: ",
      prop = "Minimum proportion: ",
      size = "Minimum size: "
    ), "Stratification parameters")
    print_fields(list(
      "Monitor: " = x@monitor,
      "Seed: " = unname(x@seed)
    ), add_names = TRUE)
  }
  invisible(x)
}


print_data <- function(x, n = Inf, level = 0) {
  if (level < 1) {
    print_label("Data:")
    x[["(names)"]] <- NULL
    print_items(x, n = n)
  } else if (level == 1) {
    print_fields(list("Number of observations: " = nrow(x)))
  }
}


print_default <- function(x, max = NULL, ...) {
  if (identical(max, Inf)) max <- .Machine$integer.max
  print(x, max = max, ...)
}


print_fields <- function(x, labels = character(), heading = character(), ...) {
  if (is_empty(labels)) {
    labels <- names(x)
    names(labels) <- labels
  }
  label_names <- names(labels)
  nzlengths <- lengths(x[label_names]) > 0
  if (any(nzlengths)) {
    indent <- 2 * (length(heading) > 0)
    if (indent) print_label(heading)
    for (name in label_names[nzlengths]) {
      print_items(
        as_string(x[[name]], sep = NULL), label = style_label(labels[name]),
        indent = indent, exdent = indent + 2, ...
      )
    }
  }
}


print_id <- function(x) {
  print_fields(list("ID: " = x@id))
}


print_items <- function(x, ...) {
  UseMethod("print_items")
}


print_items.default <- function(x, ...) {
  print(x)
}


print_items.character <- function(
  x, n = Inf, label = "", indent = 0, exdent = 0, add_names = FALSE,
  add_size = FALSE, sep = ", ", style = NULL, ...
) {
  lines <- ansi_strwrap(
    note_items(label, x, add_names = add_names, add_size = add_size, sep = sep),
    indent = indent, exdent = exdent
  )
  max_lines <- ceiling(log(n))
  if (length(lines) > max_lines && length(x) > 1) {
    nlines <- max(max_lines, 1)
    lines <- head(lines, nlines)
    key <- trimws(sep)
    pattern <- paste0("[", key, "](?=([^`]*`[^`]*`)*[^`]*$)[^\\S]")
    last_pattern <- paste0(pattern, "(?:.(?!", pattern, "))+$")
    last_line <- gsub(last_pattern, key, lines[nlines], perl = TRUE)
    last_line <- ansi_strtrim(last_line, console_width() - 4)
    lines[nlines] <- paste0(last_line, style_extra(" ..."))
  }
  if (length(style)) lines <- style(lines)
  writeLines(lines)
}


print_items.data.frame <- function(x, ...) {
  print_items.matrix(format(x), ...)
}


print_items.formula <- function(x, exdent = 2, ...) {
  x <- deparse1(x)
  split <- regexpr("[+]", x)
  if (split > 0) x <- c(substring(x, 1, split - 1), substring(x, split + 1))
  print_items(x, exdent = exdent, sep = " + ", ...)
}


print_items.ListOf <- function(x, n = Inf, level = 0, ...) {
  if (is_empty(x)) {
    hline(level, label = "ListOf()")
  } else {
    labels <- map("char", function(name, i) {
      if (nzchar(name)) {
        prefix <- "$"
        label <- deparse(as.name(name), backtick = TRUE)
      } else {
        prefix <- NULL
        label <- paste0("[[", i, "]]")
      }
      if (i <= n) {
        hline(level, label = paste0(prefix, label))
        print(x[[i]], n = n, level = level, na.print = NULL, ...)
        if (level < 2) newline()
      }
      label
    }, make_names_along(x, "", unique = FALSE), seq_along(x))
    n_more <- max(length(x) - n, 0)
    if (n_more) {
      extra <- tail(labels, n_more)
      msg <- pluralize(
        "... with ", format_len(n_more), " more element{?s}: ", "{qty(n_more)}"
      )
      print_items(extra, label = msg, n = n, style = style_extra)
      if (level < 2) newline()
    }
  }
}


print_items.matrix <- function(x, n = Inf, ...) {
  row_inds <- head(seq_len(nrow(x)), n)
  col_inds <- head(seq_len(ncol(x)), n)
  num_items <- length(row_inds) * length(col_inds)
  if (num_items) {
    print_default(x[row_inds, col_inds, drop = FALSE], max = num_items)
  } else {
    cat(
      "<", format_len(nrow(x)), " x ", format_len(ncol(x)), " ", class1(x),
      ">\n", sep = ""
    )
  }
  diff_rows <- nrow(x) - length(row_inds)
  diff_cols <- ncol(x) - length(col_inds)
  msg <- NULL
  if (diff_rows) {
    msg <- pluralize(format_len(diff_rows), " more row{?s}", "{qty(diff_rows)}")
  }
  if (diff_cols) {
    msg <- c(
      msg,
      pluralize(format_len(diff_cols), " more column{?s}: ", "{qty(diff_cols)}")
    )
    print_extra <- function(msg) {
      extra <- tail(colnames(x), diff_cols)
      print_items(extra, label = msg, n = n, style = style_extra)
    }
  } else {
    print_extra <- function(msg) cat(style_extra(msg), "\n", sep = "")
  }
  if (length(msg)) {
    print_extra(paste("... with", as_string(msg, conj = "and")))
  }
}


print_items.numeric <- function(x, n = Inf, ...) {
  diff <- length(x) - n
  if (diff > 0) {
    print_default(head(x, n), max = n)
    msg <- pluralize(
      "... with ", format_len(diff), " more value{?s}", "{qty(diff)}"
    )
    cat(style_extra(msg), "\n", sep = "")
  } else {
    print_default(x, max = n)
  }
}


print_items.tbl <- function(x, n = Inf, ...) {
  print(x, n = n, max_footer_lines = ceiling(log(n)))
}


print_label <- function(...) {
  cat(style_label(...), "\n", sep = "")
}


print_modelinfo <- function(x, trained = FALSE, level = 0, id = FALSE, ...) {
  info_list <- modelinfo(x)
  if (is_empty(info_list)) return(NULL)
  info <- info_list[[1]]
  print_fields(list(
    "Model name: " =  names(info_list),
    "Label: " = paste0(if (trained) "Trained ", info$label)
  ))
  if (id) print_id(x)
  if (level < 1) {
    print_fields(info, labels = c(
      packages = "Package{?s}: ",
      response_types = "Response type{?s}: ",
      weights = "Case weights support: ",
      na.rm = "Missing case removal: ",
      grid = "Tuning grid: ",
      varimp = "Variable importance: "
    ))
  }
}


print_progress <- function(
  scores, max_iter = Inf, method = character(), items = list(),
  n = MachineShop::settings("print_max"), metric = numeric(),
  start_time = Sys.time(), pad = TRUE
) {
  iter <- length(scores)
  hline(label = paste0(
    "Iteration ", iter, if (is.finite(max_iter)) paste0(" of ", max_iter)
  ))
  newline()
  selected <- which.max(scores)
  if (selected == iter) selected <- paste0(">", selected, "<")
  print_fields(list("Optimization method: " = method))
  print_fields(items, add_size = TRUE, add_names = TRUE, n = n)
  print_fields(list(
    "Metric: " = metric,
    "Current maximum score (iteration): " = as_string(list(
      max(scores), paste0("(", selected, ")")
    ), sep = " "),
    "Elapsed time: " = round(difftime(Sys.time(), start_time), 1)
  ), add_names = TRUE)
  if (pad && iter < max_iter) newline()
}


print_str <- function(x, label, n = Inf) {
  print_label(pluralize(label, "{qty(length(x))}"))
  cat(str(x, give.attr = FALSE, list.len = n))
}


print_train_label <- function(...) {
  print_label(style_bold(...))
}


style_extra <- make_ansi_style("grey")
style_label <- make_ansi_style("blue")


title <- function(x, ...) {
  heading(paste(class1(x), "object"), ...)
}

#' Performance Metrics
#'
#' Compute measures of agreement between observed and predicted responses.
#'
#' @name metrics
#'
#' @param observed \link[=response]{observed responses}; or
#'   \link{confusion}, \link[=curves]{performance curve}, or \link{resample}
#'   result containing observed and predicted responses.
#' @param predicted \link[=predict]{predicted responses} if not contained in
#'   \code{observed}.
#' @param weights numeric vector of non-negative
#'   \link[=case_weights]{case weights} for the observed responses [default:
#'   equal weights].
#' @param beta relative importance of recall to precision in the calculation of
#'   \code{f_score} [default: F1 score].
#' @param cutoff numeric (0, 1) threshold above which binary factor
#'   probabilities are classified as events and below which survival
#'   probabilities are classified.  If \code{NULL}, then confusion matrix-based
#'   metrics are computed on predicted class probabilities if given.
#' @param distr character string specifying a distribution with which to
#'   estimate the observed survival mean in the total sum of square component of
#'   \code{r2}.  Possible values are \code{"empirical"} for the Kaplan-Meier
#'   estimator, \code{"exponential"}, \code{"extreme"}, \code{"gaussian"},
#'   \code{"loggaussian"}, \code{"logistic"}, \code{"loglogistic"},
#'   \code{"lognormal"}, \code{"rayleigh"}, \code{"t"}, or \code{"weibull"}.
#'   Defaults to the distribution that was used in predicting mean survival
#'   times.
#' @param fun function to calculate a desired sensitivity-specificity tradeoff.
#' @param method character string specifying whether to compute \code{r2} as
#'   the coefficient of determination (\code{"mse"}) or as the square of
#'   \code{"pearson"} or \code{"spearman"} correlation.
#' @param metrics vector of two metric functions or function names that define a
#'   curve under which to calculate area [default: ROC metrics].
#' @param multiclass character string specifying the method for computing
#'   generalized area under the performance curve for multiclass factor
#'   responses.  Options are to average over areas for each pair of classes
#'   (\code{"pairs"}) or for each class versus all others (\code{"all"}).
#' @param power power to which positional distances of off-diagonals from the
#'   main diagonal in confusion matrices are raised to calculate
#'   \code{weighted_kappa2}.
#' @param stat function or character string naming a function to compute a
#'   summary statistic at each cutoff value of resampled metrics in performance
#'   curves, or \code{NULL} for resample-specific metrics.
#' @param ... arguments passed to or from other methods.
#'
#' @references
#' Hand, D. J., & Till, R. J. (2001). A simple generalisation of the area under
#' the ROC curve for multiple class classification problems. \emph{Machine
#' Learning}, \emph{45}, 171-186.
#'
#' @seealso \code{\link{metricinfo}}, \code{\link{performance}}
#'
NULL


#################### Metric Method Functions ####################


setMetric_auc <- function(f, metrics) {
  definition <- function(observed, predicted, ...) {
    auc(observed, predicted, metrics = metrics, ...)
  }
  setMetricGeneric(f)
  setMetricMethod(f, c("factor", "factor"))
  setMetricMethod(f, c("factor", "matrix"), definition)
  setMetricMethod(f, c("factor", "numeric"), definition)
  setMetricMethod_Resample(f)
  setMetricMethod(f, c("Surv", "SurvProbs"), definition)
}


setMetric_BinaryConfusionMatrix <- function(f, definition) {
  setMetricGeneric(f)
  setMetricMethod(f, c("BinaryConfusionMatrix", "NULL"), definition)
  setMetricMethod_factor(f, "factor")
  setMetricMethod_factor(f, "numeric")
  setMetricMethod_Resample(f)
  setMetricMethod_Surv(f, "SurvEvents")
  setMetricMethod_Surv(f, "SurvProbs")
}


setMetric_ConfusionMatrix <- function(f, definition) {
  setMetricGeneric(f)
  setMetricMethod(f, c("ConfusionMatrix", "NULL"), definition)
  setMetricMethod_factor(f, "factor")
  setMetricMethod_factor(f, "matrix")
  setMetricMethod_factor(f, "numeric")
  setMetricMethod_Resample(f)
  setMetricMethod_Surv(f, "SurvEvents")
  setMetricMethod_Surv(f, "SurvProbs")
}


setMetric_OrderedConfusionMatrix <- function(f, definition) {
  setMetricGeneric(f)
  setMetricMethod(f, c("OrderedConfusionMatrix", "NULL"), definition)
  definition_ordered <- function(observed, predicted, weights, ...) {
    get(f)(confusion(observed, predicted, weights), ...)
  }
  setMetricMethod(f, c("ordered", "ordered"), definition_ordered)
  setMetricMethod(f, c("ordered", "matrix"), definition_ordered)
  setMetricMethod_Resample(f)
}


setMetric_numeric <- function(f, definition) {
  setMetricGeneric(f)
  setMetricMethod(f, c("numeric", "numeric"), definition)
  setMetricMethod_BinomialVariate(f)
  setMetricMethod_matrix(f)
  setMetricMethod_Resample(f)
  setMetricMethod_Surv(f, "numeric")
}


setMetricGeneric <- function(f) {
  eval(substitute(
    setGeneric(name, function(observed, predicted, ...) standardGeneric(name)),
    list(name = metric_method_name(f))
  ))
  setMetricMethod(f, c("ANY", "ANY"))
}


setMetricMethod <- function(
  f, signature, definition = function(observed, predicted, ...) numeric()
) {
  setMethod(metric_method_name(f), signature, definition)
}


setMetricMethod_BinomialVariate <- function(f) {
  setMetricMethod(f, c("BinomialVariate", "numeric"),
    function(observed, predicted, ...) {
      get(f)(as.numeric(observed), predicted, ...)
    }
  )
}


setMetricMethod_factor <- function(f, type) {
  setMetricMethod(f, c("factor", type),
    function(observed, predicted, weights, cutoff, ...) {
      get(f)(confusion(observed, predicted, weights, cutoff = cutoff), ...)
    }
  )
}


setMetricMethod_matrix <- function(f) {
  setMetricMethod(f, c("matrix", "matrix"),
    function(observed, predicted, ...) {
      metric_matrix(get(f), observed, predicted, ...)
    }
  )
}


setMetricMethod_Resample <- function(f) {
  setMetricMethod(f, c("Resample", "NULL"),
    function(observed, predicted, weights, ...) {
      performance(observed, metrics = get(f), ...)
    }
  )
}


setMetricMethod_Surv <- function(f, type) {
  definition <- switch(type,
    "numeric" = function(observed, predicted, ...) {
      metric_SurvTimes(get(f), observed, predicted, ...)
    },
    function(observed, predicted, ...) {
      metric_SurvMatrix(get(f), observed, predicted, ...)
    }
  )
  setMetricMethod(f, c("Surv", type), definition)
}


#################### Utility Functions ####################


call_metric_method <- function(f, envir) {
  do.call(metric_method_name(f), as.list(envir))
}


metric_method_name <- function(f) {
  paste0(".", f)
}


metric_matrix <- function(.fun, observed, predicted, ...) {
  mean(map("num", function(i) {
    .fun(observed[, i], predicted[, i], ...)
  }, seq_len(ncol(observed))))
}


metric_SurvMatrix <- function(
  .fun, observed, predicted, weights, cutoff = numeric(), ...
) {
  conf_list <- confusion(observed, predicted, weights, cutoff = cutoff)
  x <- map("num", function(conf) .fun(conf, ...), conf_list)
  survmetric_mean(x, predicted@times)
}


metric_SurvTimes <- function(.fun, observed, predicted, weights = NULL, ...) {
  events <- observed[, "status"] == 1
  .fun(time(observed[events]), predicted[events], weights = weights[events],
       ...)
}


weighted_mean <- function(x, weights = NULL) {
  if (is.null(weights)) {
    sum(x) / length(x)
  } else {
    weights <- check_weights(weights, x)
    throw(check_assignment(weights))
    sum(weights * x) / sum(weights)
  }
}


weighted_sd <- function(x, weights = NULL) {
  n <- length(x)
  sqrt(mse(x, weighted_mean(x, weights), weights) * n / (n - 1))
}

#' Sparse Principal Components Analysis Variable Reduction
#'
#' Creates a \emph{specification} of a recipe step that will derive sparse
#' principal components from one or more numeric variables.
#'
#' @inheritParams step_lincomp
#' @param sparsity,num_var sparsity (L1 norm) penalty for each component or
#'   number of variables with non-zero component loadings.  Larger sparsity
#'   values produce more zero loadings.  Argument \code{sparsity} is ignored if
#'   \code{num_var} is given.  The argument value may be a single number
#'   applied to all components or a vector of component-specific numbers.
#' @param shrinkage numeric shrinkage (quadratic) penalty for the components to
#'   improve conditioning; larger values produce more shrinkage of component
#'   loadings toward zero.
#' @param center,scale logicals indicating whether to mean center and standard
#'   deviation scale the original variables prior to deriving components, or
#'   functions or names of functions for the centering and scaling.
#' @param max_iter maximum number of algorithm iterations allowed.
#' @param tol numeric tolerance for the convergence criterion.
#' @param x \code{step_spca} object.
#'
#' @return Function \code{step_spca} creates a new step whose class is of
#' the same name and inherits from \code{\link{step_lincomp}}, adds it to the
#' sequence of existing steps (if any) in the recipe, and returns the updated
#' recipe.  For the \code{tidy} method, a tibble with columns \code{terms}
#' (selectors or variables selected), \code{weight} of each variable loading in
#' the components, and \code{name} of the new variable names; and with
#' attribute \code{pev} containing the proportions of explained variation.
#'
#' @details
#' Sparse principal components analysis (SPCA) is a variant of PCA in which
#' the original variables may have zero loadings in the linear combinations
#' that form the components.
#'
#' @references
#' Zou, H., Hastie, T., & Tibshirani, R. (2006). Sparse principal component
#' analysis. \emph{Journal of Computational and Graphical Statistics},
#' \emph{15}(2), 265-286.
#'
#' @seealso \code{\link[elasticnet]{spca}}, \code{\link[recipes]{recipe}},
#' \code{\link[recipes]{prep}}, \code{\link[recipes]{bake}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package elasticnet to run
#'
#' library(recipes)
#'
#' rec <- recipe(rating ~ ., data = attitude)
#' spca_rec <- rec %>%
#'   step_spca(all_predictors(), num_comp = 5, sparsity = 1)
#' spca_prep <- prep(spca_rec, training = attitude)
#' spca_data <- bake(spca_prep, attitude)
#'
#' pairs(spca_data, lower.panel = NULL)
#'
#' tidy(spca_rec, number = 1)
#' tidy(spca_prep, number = 1)
#' }
#'
step_spca <- function(
  recipe, ..., num_comp = 5, sparsity = 0, num_var = integer(),
  shrinkage = 1e-6, center = TRUE, scale = TRUE, max_iter = 200, tol = 1e-3,
  replace = TRUE, prefix = "SPCA", role = "predictor", skip = FALSE,
  id = recipes::rand_id("spca")
) {

  recipes::add_step(recipe, new_step_spca(
    terms = recipes::ellipse_check(...),
    num_comp = num_comp,
    sparsity = sparsity,
    num_var = num_var,
    shrinkage = shrinkage,
    center = center,
    scale = scale,
    max_iter = max_iter,
    tol = tol,
    replace = replace,
    prefix = prefix,
    role = role,
    skip = skip,
    id = id
  ))

}


new_step_spca <- function(..., sparsity, num_var, shrinkage, max_iter, tol) {

  throw(check_packages("elasticnet"))

  transform <- function(x, step) {

    throw(check_packages("elasticnet"))

    num_comp <- min(step$num_comp, nrow(x))
    if (is_empty(step$num_var)) {
      para <- step$sparsity
      sparse <- "penalty"
    } else {
      para <- step$num_var
      sparse <- "varnum"
    }
    res <- elasticnet::spca(
      x, K = num_comp, para = rep_len(para, num_comp), sparse = sparse,
      lambda = step$shrinkage, max.iter = step$max_iter, eps.conv = step$tol
    )

    list(weights = res$loadings, pev = res$pev)

  }

  options <- list(
    sparsity = sparsity,
    num_var = num_var,
    shrinkage = shrinkage,
    max_iter = max_iter,
    tol = tol
  )

  object <- new_step_lincomp(..., transform = transform, options = options)

  structure(object, class = c("step_spca", "step"))

}


prep.step_spca <- function(x, ...) prep.step_lincomp(x, ...)
bake.step_spca <- function(object, ...) bake.step_lincomp(object, ...)
print.step_spca <- function(x, ...) print.step_lincomp(x, ...)


#' @rdname step_spca
#'
tidy.step_spca <- function(x, ...) tidy.step_lincomp(x, ...)


#' @rdname step_spca
#'
tunable.step_spca <- function(x, ...) {
  tibble(
    name = c("num_comp", "sparsity"),
    call_info = list(
      list(pkg = "dials", fun = "num_comp", range = c(1, 4)),
      list(pkg = "dials", fun = "penalty", range = c(-10, 0))
    ),
    source = "MachineShop",
    component = "step_spca",
    component_id = x$id
  )
}

#' MachineShop Settings
#'
#' Allow the user to view or change global settings which affect default
#' behaviors of functions in the \pkg{MachineShop} package.
#'
#' @param ... character names of settings to view, \code{name = value} pairs
#' giving the values of settings to change, a vector of these, \code{"reset"}
#' to restore all package defaults, or no arguments to view all settings.
#' Partial matching of setting names is supported.
#'
#' @return The setting value if only one is specified to view.  Otherwise, a
#' list of the values of specified settings as they existed prior to any
#' requested changes.  Such a list can be passed as an argument to
#' \code{settings} to restore their values.
#'
#' @section Settings:
#'
#' \describe{
#'   \item{\code{\link[=controls]{control}}}{function, function name, or object
#'     defining a default resampling method [default: \code{"CVControl"}].}
#'   \item{\code{cutoff}}{numeric (0, 1) threshold above which binary factor
#'     probabilities are classified as events and below which survival
#'     probabilities are classified [default: 0.5].}
#'   \item{\code{distr.SurvMeans}}{character string specifying distributional
#'     approximations to estimated survival curves for predicting survival
#'     means.  Choices are \code{"empirical"} for the Kaplan-Meier estimator,
#'     \code{"exponential"}, \code{"rayleigh"}, or \code{"weibull"} (default).}
#'   \item{\code{distr.SurvProbs}}{character string specifying distributional
#'     approximations to estimated survival curves for predicting survival
#'     events/probabilities.  Choices are \code{"empirical"} (default) for the
#'     Kaplan-Meier estimator, \code{"exponential"}, \code{"rayleigh"}, or
#'     \code{"weibull"}.}
#'   \item{\code{grid}}{\code{size} argument to \code{\link{TuningGrid}}
#'     indicating the number of parameter-specific values to generate
#'     automatically for \link[=TunedModel]{tuning} of models that have
#'     pre-defined grids or a \code{\link{TuningGrid}} function, function name,
#'     or object [default: 3].}
#'   \item{\code{method.EmpiricalSurv}}{character string specifying the
#'     empirical method of estimating baseline survival curves for Cox
#'     proportional hazards-based models.  Choices are \code{"breslow"} or
#'     \code{"efron"} (default).}
#'   \item{\code{metrics.ConfusionMatrix}}{function, function name, or vector of
#'     these with which to calculate \link{performance} \link{metrics} for
#'     confusion matrices [default: \code{c(Accuracy = "accuracy", Kappa =
#'     "kappa2", `Weighted Kappa` = "weighted_kappa2", Sensitivity =
#'     "sensitivity", Specificity = "specificity")}].}
#'   \item{\code{metrics.factor}}{function, function name, or vector of these
#'     with which to calculate \link{performance} \link{metrics} for factor
#'     responses [default: \code{c(Brier = "brier", Accuracy = "accuracy",
#'     Kappa = "kappa2", `Weighted Kappa` = "weighted_kappa2", `ROC AUC` =
#'     "roc_auc", Sensitivity = "sensitivity", Specificity = "specificity")}].}
#'   \item{\code{metrics.matrix}}{function, function name, or vector of these
#'     with which to calculate \link{performance} \link{metrics} for matrix
#'     responses [default: \code{c(RMSE = "rmse", R2 = "r2", MAE = "mae")}].}
#'   \item{\code{metrics.numeric}}{function, function name, or vector of these
#'     with which to calculate \link{performance} \link{metrics} for numeric
#'     responses [default: \code{c(RMSE = "rmse", R2 = "r2", MAE = "mae")}].}
#'   \item{\code{metrics.Surv}}{function, function name, or vector of these with
#'     which to calculate \link{performance} \link{metrics} for survival
#'     responses [default: \code{c(`C-Index` = "cindex", Brier = "brier",
#'     `ROC AUC` = "roc_auc", Accuracy = "accuracy")}].}
#'   \item{\code{print_max}}{number of models or data rows to show with print
#'     methods or \code{Inf} to show all [default: 10].}
#'   \item{\code{require}}{names of installed packages to load during parallel
#'     execution of resampling algorithms [default: \code{"MachineShop"}].}
#'   \item{\code{reset}}{character names of settings to reset to their default
#'     values.}
#'   \item{\code{RHS.formula}}{non-modifiable character vector of operators and
#'     functions allowed in traditional formula specifications.}
#'   \item{\code{stat.Curve}}{function or character string naming a function
#'     to compute one \link{summary} statistic at each cutoff value of resampled
#'     metrics in performance curves, or \code{NULL} for resample-specific
#'     metrics [default: \code{"base::mean"}].}
#'   \item{\code{stat.Resample}}{function or character string naming a function
#'     to compute one summary statistic to control the ordering of models in
#'     \link[=plot]{plots} [default: \code{"base::mean"}].}
#'   \item{\code{stat.TrainingParams}}{function or character string naming a function
#'     to compute one summary statistic on resampled performance metrics for
#'     input \link[=SelectedInput]{selection} or \link[=TunedInput]{tuning} or
#'     for model \link[=SelectedModel]{selection} or \link[=TunedModel]{tuning}
#'     [default: \code{"base::mean"}].}
#'   \item{\code{stats.PartialDependence}}{function, function name, or vector of
#'     these with which to compute \link[=dependence]{partial dependence}
#'     summary statistics [default: \code{c(Mean = "base::mean")}].}
#'   \item{\code{stats.Resample}}{function, function name, or vector of these
#'     with which to compute \link{summary} statistics on resampled performance
#'     metrics [default: \code{c(Mean = "base::mean", Median = "stats::median",
#'     SD = "stats::sd", Min = "base::min", Max = "base::max")}].}
#' }
#'
#' @examples
#' ## View all current settings
#' settings()
#'
#' ## Change settings
#' presets <- settings(control = "BootControl", grid = 10)
#'
#' ## View one setting
#' settings("control")
#'
#' ## View multiple settings
#' settings("control", "grid")
#'
#' ## Restore the previous settings
#' settings(presets)
#'
settings <- function(...) {

  args <- list(...)
  if (length(args) == 1 && is.null(names(args)) && is.vector(args[[1]])) {
    args <- as.list(args[[1]])
  }

  global_settings <- MachineShop_global$settings
  global_values <- lapply(global_settings, getElement, name = "value")
  global_checks <- lapply(global_settings, getElement, name = "check")

  if (!length(args)) {
    return(global_values)
  } else if (identical(args, list("reset"))) {
    settings(reset = names(.global_defaults))
    return(invisible(global_values))
  }

  arg_names <- names(args)
  if (is.null(arg_names)) arg_names <- character(length(args))
  arg_names_nzchar <- nzchar(arg_names)

  is_get_args <- !arg_names_nzchar & vapply(args, is.character, logical(1))
  arg_names[is_get_args] <- unlist(args[is_get_args])

  settings_pmatch <- pmatch(arg_names, names(global_settings))
  valid_settings <- !is.na(settings_pmatch)
  for (name in arg_names[!valid_settings]) {
    throw(Warning("Argument `", name, "` is not a MachineShop setting."))
  }

  presets <- global_values[settings_pmatch[valid_settings]]

  which_set_args <- which(arg_names_nzchar & valid_settings)
  for (index in which_set_args) {
    global_name <- as.name(names(global_settings)[settings_pmatch[index]])
    value <- global_checks[[global_name]](args[[index]])
    eval(substitute(
      throw(check_assignment(global_name, value), call = sys.call(-2))
    ))
    MachineShop_global$settings[[global_name]]$value <- value
  }

  if (any(arg_names_nzchar)) {
    invisible(presets)
  } else if (length(args) > 1) {
    presets
  } else if (length(presets)) {
    presets[[1]]
  } else {
    NULL
  }

}


#################### Global Environment ####################


MachineShop_global <- as.environment(list(

  settings = list(

    control = list(
      value = "CVControl",
      check = function(x) {
        tryCatch(
          {
            as.MLControl(x)
            x
          },
          error = function(cond) {
            DomainError(
              x, "must be an MLControl object, function, or function name"
            )
          }
        )
      }
    ),

    cutoff = list(
      value = 0.5,
      check = function(x) {
        check_numeric(x, bounds = c(0, 1), include = FALSE, size = 1)
      }
    ),

    distr.SurvMeans = list(
      value = "weibull",
      check = function(x) {
        check_match(x, c("weibull", "exponential", "rayleigh", "empirical"))
      }
    ),

    distr.SurvProbs = list(
      value = "empirical",
      check = function(x) {
        check_match(x, c("empirical", "weibull", "exponential", "rayleigh"))
      }
    ),

    grid = list(
      value = 3,
      check = check_grid
    ),

    method.EmpiricalSurv = list(
      value = "efron",
      check = function(x) {
        check_match(x, c("efron", "breslow"))
      }
    ),

    metrics = list(
      value = c(
        "accuracy",
        "auc",
        "brier",
        "cindex",
        "cross_entropy",
        "f_score",
        "fnr",
        "fpr",
        "gini",
        "kappa2",
        "mae",
        "mse",
        "msle",
        "npv",
        "ppr",
        "ppv",
        "pr_auc",
        "precision",
        "r2",
        "recall",
        "rmse",
        "rmsle",
        "roc_auc",
        "roc_index",
        "sensitivity",
        "specificity",
        "tnr",
        "tpr",
        "weighted_kappa2"
      ),
      check = function(x) throw(check_const_setting(x, "metrics"))
    ),

    metrics.ConfusionMatrix = list(
      value = c(
        "Accuracy" = "accuracy",
        "Kappa" = "kappa2",
        "Weighted Kappa" = "weighted_kappa2",
        "Sensitivity" = "sensitivity",
        "Specificity" = "specificity"
      ),
      check = check_metrics
    ),

    metrics.factor = list(
      value = c(
        "Brier" = "brier",
        "Accuracy" = "accuracy",
        "Kappa" = "kappa2",
        "Weighted Kappa" = "weighted_kappa2",
        "ROC AUC" = "roc_auc",
        "Sensitivity" = "sensitivity",
        "Specificity" = "specificity"
      ),
      check = check_metrics
    ),

    metrics.matrix = list(
      value = c(
        "RMSE" = "rmse",
        "R2" = "r2",
        "MAE" = "mae"
      ),
      check = check_metrics
    ),

    metrics.numeric = list(
      value = c(
        "RMSE" = "rmse",
        "R2" = "r2",
        "MAE" = "mae"
      ),
      check = check_metrics
    ),

    metrics.Surv = list(
      value = c(
        "C-Index" = "cindex",
        "Brier" = "brier",
        "ROC AUC" = "roc_auc",
        "Accuracy" = "accuracy"
      ),
      check = check_metrics
    ),

    models = list(
      value = c(
        "AdaBagModel",
        "AdaBoostModel",
        "BARTMachineModel",
        "BARTModel",
        "BlackBoostModel",
        "C50Model",
        "CForestModel",
        "CoxModel",
        "CoxStepAICModel",
        "EarthModel",
        "FDAModel",
        "GAMBoostModel",
        "GBMModel",
        "GLMBoostModel",
        "GLMModel",
        "GLMStepAICModel",
        "GLMNetModel",
        "KNNModel",
        "LARSModel",
        "LDAModel",
        "LMModel",
        "MDAModel",
        "NaiveBayesModel",
        "NNetModel",
        "ParsnipModel",
        "PDAModel",
        "PLSModel",
        "POLRModel",
        "QDAModel",
        "RandomForestModel",
        "RangerModel",
        "RFSRCModel",
        "RPartModel",
        "SelectedModel",
        "StackedModel",
        "SuperModel",
        "SurvRegModel",
        "SurvRegStepAICModel",
        "SVMModel",
        "SVMANOVAModel",
        "SVMBesselModel",
        "SVMLaplaceModel",
        "SVMLinearModel",
        "SVMPolyModel",
        "SVMRadialModel",
        "SVMSplineModel",
        "SVMTanhModel",
        "TreeModel",
        "TunedModel",
        "XGBModel",
        "XGBDARTModel",
        "XGBLinearModel",
        "XGBTreeModel"
      ),
      check = function(x) throw(check_const_setting(x, "models"))
    ),

    print_max = list(
      value = 10,
      check = function(x) {
        result <- check_numeric(x, bounds = c(1, Inf), size = 1)
        if (is.numeric(result)) result <- floor(result)
        result
      }
    ),

    require = list(
      value = "MachineShop",
      check = function(x) {
        x <- setdiff(x, .global_defaults$require)
        unavailable <- !vapply(x, requireNamespace, logical(1), quietly = TRUE)
        if (any(unavailable)) {
          DomainError(
            x, note_items("includes unavailable package{?s}: ", x[unavailable])
          )
        } else c(x, .global_defaults$require)
      }
    ),

    reset = list(
      value = character(),
      check = function(x) {
        if (is.character(x)) {
          setting_names <- names(.global_defaults)
          reset_names <- setting_names[charmatch(x, setting_names, nomatch = 0)]
          for (name in reset_names) {
            value <- .global_defaults[[name]]
            MachineShop_global$settings[[name]]$value <- value
          }
          if ("reset" %in% reset_names) .global_defaults$reset else reset_names
        } else {
          DomainError(x, "must be one or more character strings")
        }
      }
    ),

    response_types = list(
      value = c(
        "binary", "BinomialVariate", "DiscreteVariate", "factor", "matrix",
        "NegBinomialVariate", "numeric", "ordered", "PoissonVariate", "Surv"
      ),
      check = function(x) throw(check_const_setting(x, "response_types"))
    ),

    RHS.formula = list(
      value = sort(c(
        ".", "(", ":", "%in%", "I", "offset",
        "+", "-", "*", "/", "^", "%%", "%/%",
        "&", "|", "!",
        "==", "!=", "<", "<=", ">=", ">",
        "abs", "sign", "sqrt", "floor", "ceiling", "trunc", "round", "signif",
        "exp", "log", "expm1", "log1p", "cos", "sin", "tan", "cospi", "sinpi",
        "tanpi", "acos", "asin", "atan",
        "cosh", "sinh", "tanh", "acosh", "asinh", "atanh",
        "lgamma", "gamma", "digamma", "trigamma"
      )),
      check = function(x) throw(check_const_setting(x, "RHS.formula"))
    ),

    stat.Curve = list(
      value = c("Mean" = "base::mean"),
      check = function(x) {
        if (!is.null(x) && is(check_stat(x), "error")) {
          DomainError(
            x, "must be a statistics function, function name, vector of ",
            "these, or NULL"
          )
        } else x
      }
    ),

    stat.Resample = list(
      value = c("Mean" = "base::mean"),
      check = check_stat
    ),

    stat.TrainingParams = list(
      value = c("Mean" = "base::mean"),
      check = check_stat
    ),

    stats.PartialDependence = list(
      value = c("Mean" = "base::mean"),
      check = check_stats

    ),

    stats.Resample = list(
      value = c(
        "Mean" = "base::mean",
        "Median" = "stats::median",
        "SD" = "stats::sd",
        "Min" = "base::min",
        "Max" = "base::max"
      ),
      check = check_stats
    )

  ),

  throw_times = list()

))


.global_defaults <- settings()

## Global Settings

test_that("settings changes and views", {
  skip_if_not(TEST_ALL)

  presets <- settings()

  expect_identical(settings("control", cut = 0.25, print_max = 5),
                   presets[c("control", "cutoff", "print_max")])
  expect_identical(settings("cutoff"), 0.25)
  expect_identical(settings("print_max"), 5)

  new_value <- "BootControl"
  expect_type(settings(control = new_value), "list")
  expect_identical(settings("control"), new_value)
  expect_error(settings(control = "MLModel"))

  new_value <- 0.05
  expect_type(settings(cutoff = new_value), "list")
  expect_identical(settings("cutoff"), new_value)
  expect_error(settings(cutoff = Inf))
  expect_error(settings(cutoff = character()))

  new_value <- "exponential"
  expect_type(settings(distr.SurvMeans = substr(new_value, 1, 3)), "list")
  expect_identical(settings("distr.SurvMeans"), new_value)
  expect_error(settings(distr.SurvMeans = character()))

  new_value <- "exponential"
  expect_type(settings(distr.SurvProbs = substr(new_value, 1, 3)), "list")
  expect_identical(settings("distr.SurvProbs"), new_value)
  expect_error(settings(distr.SurvProbs = character()))

  expect_type(settings(grid = 5), "list")
  expect_s4_class(settings("grid"), "TuningGrid")
  expect_type(settings(grid = TuningGrid), "list")
  expect_s4_class(settings("grid"), "TuningGrid")
  expect_error(settings(grid = character()))

  new_value <- "breslow"
  expect_type(settings(method.EmpiricalSurv = substr(new_value, 1, 3)), "list")
  expect_identical(settings("method.EmpiricalSurv"), new_value)
  expect_error(settings(method.EmpiricalSurv = character()))

  new_value <- c(mse, "mse")
  expect_type(settings(metrics.ConfusionMatrix = new_value), "list")
  expect_identical(settings("metrics.ConfusionMatrix"), new_value)
  expect_error(settings(metrics.ConfusionMatrix = "mean"))

  new_value <- c(mse, "mse")
  expect_type(settings(metrics.factor = new_value), "list")
  expect_identical(settings("metrics.factor"), new_value)
  expect_error(settings(metrics.factor = "mean"))

  new_value <- c(mse, "mse")
  expect_type(settings(metrics.matrix = new_value), "list")
  expect_identical(settings("metrics.matrix"), new_value)
  expect_error(settings(metrics.matrix = "mean"))

  new_value <- c(mse, "mse")
  expect_type(settings(metrics.numeric = new_value), "list")
  expect_identical(settings("metrics.numeric"), new_value)
  expect_error(settings(metrics.numeric = "mean"))

  new_value <- c(mse, "mse")
  expect_type(settings(metrics.Surv = new_value), "list")
  expect_identical(settings("metrics.Surv"), new_value)
  expect_error(settings(metrics.Surv = "mean"))

  expect_type(settings(print_max = 5), "list")
  expect_identical(settings("print_max"), 5)
  expect_type(settings(print_max = Inf), "list")
  expect_identical(settings("print_max"), Inf)
  expect_error(settings(print_max = character()))

  old_values <- settings("require")
  expect_type(settings(require = c("MachineShop", "stats")), "list")
  expect_identical(settings("require"), c("stats", old_values))
  expect_error(settings(require = 1))

  settings("reset")
  values <- c("cut", "grid")
  old_values <- settings(values)
  settings(control = "BootControl", cutoff = 0.1, grid = 10)
  settings(reset = values)
  expect_identical(settings("control"), "BootControl")
  expect_identical(settings(values), old_values)

  new_value <- "median"
  expect_type(settings(stat.Curve = new_value), "list")
  expect_identical(settings("stat.Curve"), new_value)
  expect_error(settings(stat.Curve = "character"))

  new_value <- "median"
  expect_type(settings(stat.Resample = new_value), "list")
  expect_identical(settings("stat.Resample"), new_value)
  expect_error(settings(stat.Resample = "character"))

  new_value <- "median"
  expect_type(settings(stat.TrainingParams = new_value), "list")
  expect_identical(settings("stat.TrainingParams"), new_value)
  expect_error(settings(stat.TrainingParams = "character"))

  new_value <- c("median", sd)
  expect_type(settings(stats.PartialDependence = new_value), "list")
  expect_identical(settings("stats.PartialDependence"), new_value)
  expect_error(settings(stats.PartialDependence = "character"))

  new_value <- c("median", sd)
  expect_type(settings(stats.Resample = new_value), "list")
  expect_identical(settings("stats.Resample"), new_value)
  expect_error(settings(stats.Resample = "character"))

  settings("reset")
  expect_identical(settings(), presets)

})

#' Recursive Partitioning and Regression Tree Models
#'
#' Fit an \code{rpart} model.
#'
#' @param minsplit minimum number of observations that must exist in a node in
#'   order for a split to be attempted.
#' @param minbucket minimum number of observations in any terminal node.
#' @param cp complexity parameter.
#' @param maxcompete number of competitor splits retained in the output.
#' @param maxsurrogate number of surrogate splits retained in the output.
#' @param usesurrogate how to use surrogates in the splitting process.
#' @param xval number of cross-validations.
#' @param surrogatestyle controls the selection of a best surrogate.
#' @param maxdepth maximum depth of any node of the final tree, with the root
#'   node counted as depth 0.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric}, \code{Surv}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameter:}{
#'     \code{cp}
#'   }
#' }
#'
#' Further model details can be found in the source link below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[rpart]{rpart}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested packages rpart and partykit to run
#'
#' fit(Species ~ ., data = iris, model = RPartModel)
#' }
#'
RPartModel <- function(
  minsplit = 20, minbucket = round(minsplit / 3), cp = 0.01, maxcompete = 4,
  maxsurrogate = 5, usesurrogate = 2, xval = 10, surrogatestyle = 0,
  maxdepth = 30
) {

  MLModel(

    name = "RPartModel",
    label = "Recursive Partitioning and Regression Trees",
    packages = c("rpart", "partykit"),
    response_types = c("factor", "numeric", "Surv"),
    weights = TRUE,
    predictor_encoding = "model.frame",
    na.rm = "response",
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = "cp",
      get_values = c(
        function(n, data, ...) {
          cptable <- fit(data, model = RPartModel(cp = 0))$cptable
          xerror_order <- order(cptable[, "xerror"] + cptable[, "xstd"])
          sort(head(cptable[xerror_order, "CP"], n))
        }
      )
    ),

    fit = function(formula, data, weights, ...) {
      method <- switch_class(response(data),
        "factor" = "class",
        "numeric" = "anova",
        "Surv" = "exp"
      )
      rpart::rpart(
        formula, data = as.data.frame(formula, data = data), weights = weights,
        na.action = na.pass, method = method, control = list(...)
      )
    },

    predict = function(object, newdata, .MachineShop, ...) {
      y <- response(.MachineShop$input)
      newdata <- as.data.frame(newdata)
      if (is.Surv(y)) {
        object <- partykit::as.party(object)
        fits <- predict(object, newdata = newdata, type = "prob")
        predict(y, fits, ...)
      } else {
        predict(object, newdata = newdata)
      }
    },

    varimp = function(object, .MachineShop, ...) {
      y <- response(.MachineShop$input)
      structure(
        object$variable.importance,
        metric = switch_class(y,
          "factor" = "gini",
          "numeric" = "mse",
          "Surv" = "deviance"
        )
      )
    }

  )

}

MLModelFunction(RPartModel) <- NULL

#' Random Forest Model
#'
#' Implementation of Breiman's random forest algorithm (based on Breiman and
#' Cutler's original Fortran code) for classification and regression.
#'
#' @param ntree number of trees to grow.
#' @param mtry number of variables randomly sampled as candidates at each split.
#' @param replace should sampling of cases be done with or without replacement?
#' @param nodesize minimum size of terminal nodes.
#' @param maxnodes maximum number of terminal nodes trees in the forest can
#'   have.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \code{mtry}, \code{nodesize}*
#'   }
#' }
#' * excluded from grids by default
#'
#' Default argument values and further model details can be found in the source
#' See Also link below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[randomForest]{randomForest}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package randomForest to run
#'
#' fit(sale_amount ~ ., data = ICHomes, model = RandomForestModel)
#' }
#'
RandomForestModel <- function(
  ntree = 500,
  mtry = .(if (is.factor(y)) floor(sqrt(nvars)) else max(floor(nvars / 3), 1)),
  replace = TRUE, nodesize = .(if (is.factor(y)) 1 else 5), maxnodes = integer()
) {

  MLModel(

    name = "RandomForestModel",
    label = "Random Forests",
    packages = "randomForest",
    response_types = c("factor", "numeric"),
    predictor_encoding = "model.frame",
    na.rm = TRUE,
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = c("mtry", "nodesize"),
      get_values = c(
        function(n, data, ...) seq_nvars(data, RandomForestModel, n),
        function(n, data, ...) {
          round_int(seq(1, min(20, nrow(data)), length = n))
        }
      ),
      default = c(TRUE, FALSE)
    ),

    fit = function(formula, data, weights, ...) {
      eval_fit(
        data,
        formula = randomForest::randomForest(
          formula, data = data, na.action = na.pass, ...
        ),
        matrix = randomForest::randomForest(x, y, ...)
      )
    },

    predict = function(object, newdata, .MachineShop, ...) {
      y <- response(.MachineShop$input)
      newdata <- as.data.frame(newdata)
      predict(
        object, newdata = newdata,
        type = if (is.factor(y)) "prob" else "response"
      )
    },

    varimp = function(object, .MachineShop, ...) {
      y <- response(.MachineShop$input)
      vi <- randomForest::importance(object)
      structure(
        vi[, 1], names = rownames(vi),
        metric = if (is.factor(y)) "gini" else "rss/sse"
      )
    }

  )

}

MLModelFunction(RandomForestModel) <- NULL

#' Model Calibration
#'
#' Calculate calibration estimates from observed and predicted responses.
#'
#' @name calibration
#' @rdname calibration
#'
#' @param x \link[=response]{observed responses} or \link{resample} result
#'   containing observed and predicted responses.
#' @param y \link[=predict]{predicted responses} if not contained in \code{x}.
#' @param weights numeric vector of non-negative
#'   \link[=case_weights]{case weights} for the observed \code{x} responses
#'   [default: equal weights].
#' @param breaks value defining the response variable bins within which to
#'   calculate observed mean values.  May be specified as a number of bins, a
#'   vector of breakpoints, or \code{NULL} to fit smooth curves with splines for
#'   predicted survival probabilities and with \link[stats:loess]{loess} for
#'   others.
#' @param span numeric parameter controlling the degree of loess smoothing.
#' @param distr character string specifying a distribution with which to
#'   estimate the observed survival mean.  Possible values are
#'   \code{"empirical"} for the Kaplan-Meier estimator, \code{"exponential"},
#'   \code{"extreme"}, \code{"gaussian"}, \code{"loggaussian"},
#'   \code{"logistic"}, \code{"loglogistic"}, \code{"lognormal"},
#'   \code{"rayleigh"}, \code{"t"}, or \code{"weibull"}.  Defaults to the
#'   distribution that was used in predicting mean survival times.
#' @param pool logical indicating whether to compute a single calibration curve
#'   on predictions pooled over all resampling iterations or to compute them for
#'   each iteration individually and return the mean calibration curve. Pooling
#'   can result in large memory allocation errors when fitting smooth curves
#'   with \code{breaks = NULL}. The current default is changed from versions
#'   <= 3.8.0 of the package which only implemented \code{pool = TRUE}.
#' @param na.rm logical indicating whether to remove observed or predicted
#'   responses that are \code{NA} when calculating metrics.
#' @param ... arguments passed to other methods.
#'
#' @return \code{Calibration} class object that inherits from \code{data.frame}.
#'
#' @seealso \code{\link{c}}, \code{\link{plot}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' library(survival)
#'
#' control <- CVControl() %>% set_predict(times = c(90, 180, 360))
#' res <- resample(Surv(time, status) ~ ., data = veteran, model = GBMModel,
#'                 control = control)
#' cal <- calibration(res)
#' plot(cal)
#' }
#'
calibration <- function(
  x, y = NULL, weights = NULL, breaks = 10, span = 0.75, distr = character(),
  pool = FALSE, na.rm = TRUE, ...
) {
  if (pool) throw(Warning("pool = TRUE is deprecated"))
  if (na.rm) {
    complete <- complete_subset(x = x, y = y, weights = weights)
    x <- complete$x
    y <- complete$y
    weights <- complete$weights
  }
  Calibration(
    .calibration(
      x, y, weights, breaks = breaks, span = span, distr = distr, pool = pool,
      ...
    ),
    smoothed = is_empty(breaks)
  )
}


Calibration <- function(object, ..., .check = TRUE) {
  if (.check) {
    if (is.null(object$Model)) object$Model <- factor("Model")
    missing <- missing_names(c("Response", "Predicted", "Observed"), object)
    if (length(missing)) {
      throw(Error(note_items(
        "Missing calibration variable{?s}: ", missing, "."
      )))
    }
  }
  rownames(object) <- NULL
  new("Calibration", object, ...)
}


setGeneric(".calibration",
  function(observed, predicted, ...) standardGeneric(".calibration")
)


setMethod(".calibration", c("ANY", "ANY"),
  function(observed, predicted, ...) {
    throw(Error("Calibration unavailable for response type."))
  }
)


setMethod(".calibration", c("factor", "matrix"),
  function(observed, predicted, ...) {
    cal <- .calibration(model.matrix(~ observed - 1), predicted, ...)
    bounds <- c("Lower", "Upper")
    cal$Observed[, bounds] <- pmin(pmax(cal$Observed[, bounds], 0), 1)
    cal
  }
)


setMethod(".calibration", c("factor", "numeric"),
  function(observed, predicted, ...) {
    observed <- as.numeric(observed == levels(observed)[2])
    cal <- .calibration(observed, predicted, ...)
    bounds <- c("Lower", "Upper")
    cal$Observed[, bounds] <- pmin(pmax(cal$Observed[, bounds], 0), 1)
    cal
  }
)


setMethod(".calibration", c("matrix", "matrix"),
  function(observed, predicted, weights, breaks, span, xlim = NULL, ...) {
    weights <- check_weights(weights, observed[, 1])
    throw(check_assignment(weights))
    if (is_empty(breaks)) {
      x <- predicted
      newx <- if (!is_empty(xlim)) {
        matrix(seq(xlim[1], xlim[2], length = 101), 101, ncol(x))
      } else x
      res <- data.frame(
        Response = rep(factor(colnames(x)), each = nrow(newx)),
        Predicted = as.numeric(newx)
      )
      loess_fits <- map(function(col) {
        data <- data.frame(y = observed[, col], x = x[, col])
        newdata <- data.frame(x = newx[, col])
        predict(
          loess(y ~ x, data = data, weights = weights, span = span),
          newdata = newdata, se = TRUE
        )
      }, seq_len(ncol(x)))
      Mean <- c(map("num", getElement, loess_fits, "fit"))
      SE <- c(map("num", getElement, loess_fits, "se.fit"))
      res$Observed <- cbind(
        Mean = Mean, SE = SE, Lower = Mean - SE, Upper = Mean + SE
      )
      res
    } else {
      df <- data.frame(
        Response = rep(factor(colnames(predicted)), each = nrow(predicted)),
        Predicted = midpoints(as.numeric(predicted), breaks, xlim = xlim),
        Observed = c(observed),
        Weight = weights
      )
      res <- rev(expand.grid(
        Predicted = levels(df$Predicted),
        Response = unique(df$Response)
      ))
      res$Observed <- matrix(NA, nrow(res), 4)
      colnames(res$Observed) <- c("Mean", "SE", "Lower", "Upper")
      data_splits <- split(df, df[c("Response", "Predicted")], drop = TRUE)
      for (data in data_splits) {
        Mean <- weighted_mean(data$Observed, data$Weight)
        SE <- weighted_sd(data$Observed, data$Weight) / sqrt(nrow(data))
        ind <- res$Response == data$Response[1] &
          res$Predicted == data$Predicted[1]
        res[ind, "Observed"] <- cbind(Mean, SE, Mean - SE, Mean + SE)
      }
      res
    }
  }
)


setMethod(".calibration", c("numeric", "numeric"),
  function(observed, predicted, ...) {
    .calibration(cbind(y = observed), cbind(y = predicted), ...)
  }
)


setMethod(".calibration", c("Resample", "ANY"),
  function(observed, predicted, weights, pool, ...) {
    cal_models <- by(observed, observed$Model, function(resample) {
      if (pool) {
        calibration(
          resample$Observed, resample$Predicted, resample$Weight, na.rm = FALSE,
          ...
        )
      } else {
        xlim <- range(resample$Predicted, finite = TRUE)
        cal_iters <- by(resample, resample$Iteration, function(resample_iter) {
          calibration(
            resample_iter$Observed, resample_iter$Predicted,
            resample_iter$Weight, na.rm = FALSE, xlim = xlim, ...
          )
        })
        cal_iter1 <- cal_iters[[1]]
        if (length(cal_iters) > 1) {
          get_stats <- function(fun_map, fun_stat) {
            num_mat <- map("num", fun_map, cal_iters)
            apply(num_mat, 1, function(x) fun_stat(na.omit(x)))
          }
          cal_iter1$Predicted <- get_stats(function(x) x$Predicted, mean)
          Mean <- get_stats(function(x) x$Observed[, "Mean"], mean)
          SE <- get_stats(function(x) x$Observed[, "Mean"], sd)
          cal_iter1$Observed <- cbind(
            Mean = Mean, SE = SE, Lower = Mean - SE, Upper = Mean + SE
          )
        }
        cal_iter1
      }
    }, simplify = FALSE)
    do.call(c, cal_models)
  }
)


setMethod(".calibration", c("Surv", "SurvProbs"),
  function(observed, predicted, weights, breaks, xlim = NULL, ...) {
    weights <- check_weights(weights, observed)
    throw(check_assignment(weights))
    times <- predicted@times
    if (is_empty(breaks)) {
      throw(check_censoring(observed, "right"))
      throw(check_equal_weights(weights))
      x <- predicted
      newx <- if (!is_empty(xlim)) {
        matrix(seq(xlim[1], xlim[2], length = 101), ncol = ncol(x))
      } else x
      res <- data.frame(
        Response = rep(factor(colnames(x)), each = nrow(newx)),
        Predicted = as.numeric(newx)
      )
      Mean <- c(map("num", function(col) tryCatch(
        {
          harefit <- polspline::hare(
            observed[, "time"], observed[, "status"], x[, col]
          )
          1 - polspline::phare(times[col], newx[, col], harefit)
        },
        error = function(cond) {
          throw(LocalWarning(conditionMessage(cond)))
          rep(NA, nrow(newx))
        }
      ), seq_len(ncol(x))))
      res$Observed <- cbind(Mean = Mean, SE = NA, Lower = NA, Upper = NA)
      res
    } else {
      df <- data.frame(
        Response = rep(factor(colnames(predicted)), each = nrow(predicted)),
        Predicted = midpoints(as.numeric(predicted), breaks, xlim = xlim),
        Observed = rep(observed, times = length(times)),
        Weight = weights,
        Time = rep(times, each = nrow(predicted))
      )
      res <- rev(expand.grid(
        Predicted = levels(df$Predicted),
        Response = unique(df$Response)
      ))
      res$Observed <- matrix(NA, nrow(res), 4)
      colnames(res$Observed) <- c("Mean", "SE", "Lower", "Upper")
      data_splits <- split(df, df[c("Response", "Predicted")], drop = TRUE)
      for (data in data_splits) {
        km <- survfit(Observed ~ 1, data = data, weights = data$Weight)
        inds <- findInterval(data$Time[1], c(0, km$time))
        Mean <- c(1, km$surv)[inds]
        SE <- c(0, km$std.err)[inds]
        ind <- res$Response == data$Response[1] &
          res$Predicted == data$Predicted[1]
        res[ind, "Observed"] <- cbind(
          Mean, SE, max(Mean - SE, 0), min(Mean + SE, 1)
        )
      }
      res
    }
  }
)


setMethod(".calibration", c("Surv", "numeric"),
  function(observed, predicted, weights, breaks, distr, span, xlim = NULL, ...)
  {
    weights <- check_weights(weights, observed)
    throw(check_assignment(weights))
    max_time <- max(time(observed))
    distr <- get_surv_distr(distr, observed, predicted)
    nparams <- if (distr %in% c("exponential", "rayleigh")) 1 else 2

    survfit_est <- function(observed, weights = NULL) {
      km <- survfit(observed ~ 1, weights = weights, se.fit = FALSE)
      est <- survival:::survmean(km, rmean = max_time)
      list(Mean = est$matrix[["*rmean"]], SE = est$matrix[["*se(rmean)"]])
    }

    survreg_est <- function(observed, distr, weights = NULL) {
      regfit <- survreg(observed ~ 1, weights = weights, dist = distr)
      est <- predict(regfit, data.frame(row.names = 1), se.fit = TRUE)
      list(Mean = est$fit[[1]], SE = est$se.fit[[1]])
    }

    if (is_empty(breaks)) {
      res <- data.frame(
        Response = factor("Mean"),
        Predicted = if (is_empty(xlim)) {
          unique(predicted)
        } else {
          seq(xlim[1], xlim[2], length = 101)
        }
      )
      tricubic <- function(x, span = 1, min_weight = 0) {
        x <- abs(x)
        x_range <- span * diff(range(x))
        (1 - min_weight) * pmax((1 - (x / x_range)^3)^3, 0) + min_weight
      }
      surv_ests <- map(function(value) {
        weights <- weights *
          tricubic(predicted - value, span = span, min_weight = 0.01)
        est <- if (distr == "empirical") {
          survfit_est(observed, weights)
        } else {
          survreg_est(observed, distr, weights)
        }
        with(est, {
          c(Mean = Mean, SE = SE, Lower = max(Mean - SE, 0), Upper = Mean + SE)
        })
      }, res$Predicted)
      res$Observed <- do.call(rbind, surv_ests)
      res
    } else {
      df <- data.frame(
        Response = factor("Mean"),
        Predicted = midpoints(predicted, breaks, xlim = xlim),
        Observed = observed,
        Weight = weights
      )
      res <- rev(expand.grid(
        Predicted = levels(df$Predicted),
        Response = unique(df$Response)
      ))
      res$Observed <- matrix(NA, nrow(res), 4)
      colnames(res$Observed) <- c("Mean", "SE", "Lower", "Upper")
      data_splits <- split(df, df[c("Response", "Predicted")], drop = TRUE)
      for (data in data_splits) {
        observed <- data$Observed
        weights <- data$Weight
        est <- if (distr == "empirical") {
          survfit_est(observed, weights)
        } else if (length(event_time(observed)) >= nparams) {
          survreg_est(observed, distr, weights)
        } else {
          list(Mean = NA_real_, SE = NA_real_)
        }
        ind <- res$Response == data$Response[1] &
          res$Predicted == data$Predicted[1]
        res[ind, "Observed"] <- with(est, {
          cbind(Mean, SE, max(Mean - SE, 0), min(Mean + SE, 1))
        })
      }
      res
    }
  }
)


midpoints <- function(x, breaks, xlim = NULL) {
  breaks <- if (length(breaks) == 1) {
    break_range <- if (is_empty(xlim)) range(x, finite = TRUE) else xlim
    num_breaks <- max(as.integer(breaks), 1) + 1
    seq(break_range[1], break_range[2], length = num_breaks)
  } else {
    sort(breaks)
  }
  levels <- breaks[-length(breaks)] + diff(breaks) / 2
  structure(
    levels[.bincode(x, breaks, include.lowest = TRUE)],
    levels = levels
  )
}

#' Extract Response Variable
#'
#' Extract the response variable from an object.
#'
#' @rdname response-methods
#'
#' @param object model \link{fit}, \link[=inputs]{input}, or
#'   \link[=ModelSpecification]{specification} containing predictor and response
#'   variables.
#' @param newdata \link[=data.frame]{data frame} from which to extract the
#'   response variable values if given; otherwise, \code{object} is used.
#' @param ... arguments passed to other methods.
#'
#' @examples
#' ## Survival response example
#' library(survival)
#'
#' mf <- ModelFrame(Surv(time, status) ~ ., data = veteran)
#' response(mf)
#'
response <- function(object, ...) {
  UseMethod("response")
}


response.formula <- function(object, data = NULL, template = NULL, ...) {
  if (!is.null(data)) {
    expr <- response(object)
    y <- eval(expr, as.data.frame(data[, all.vars(expr), drop = FALSE]))
    if (is.factor(template)) {
      y_levels <- as.character(unique(y))
      template_levels <- levels(template)
      new_levels <- y_levels[is.na(match(y_levels, template_levels))]
      if (length(new_levels)) {
        throw(Error(note_items(
          "Response factor has new level{?s}: ", new_levels, "."
        )))
      }
      y <- factor(y, levels = template_levels, ordered = is.ordered(template),
                  exclude = NULL)
    }
    template_class <- class1(template)
    if (!(is.null(template) || is(y, template_class))) {
      throw(TypeError(y, template_class, "response variable"))
    }
    y
  } else if (length(object) > 2) object[[2]]
}


#' @rdname response-methods
#'
response.MLModelFit <- function(object, newdata = NULL, ...) {
  response(as.MLInput(object), newdata)
}


#' @rdname response-methods
#'
response.ModelFrame <- function(object, newdata = NULL, ...) {
  y <- model.response(object)
  if (!is.null(newdata)) {
    f <- if (is(newdata, "ModelFrame")) formula else terms
    y <- response(f(object), newdata, y)
  }
  y
}


#' @rdname response-methods
#'
response.ModelSpecification <- function(object, newdata = NULL, ...) {
  response(as.MLInput(object), newdata = newdata, ...)
}


#' @rdname response-methods
#'
response.recipe <- function(object, newdata = NULL, ...) {
  object <- prep(object, retain = FALSE)
  response(terms(object), bake(object, newdata = newdata))
}


#################### Discrete Variate Classes ####################


#' Discrete Variate Constructors
#'
#' Create a variate of binomial counts, discrete numbers, negative binomial
#' counts, or Poisson counts.
#'
#' @name DiscreteVariate
#' @rdname DiscreteVariate
#'
#' @param x numeric vector.
#' @param size number or numeric vector of binomial trials.
#' @param min,max minimum and maximum bounds for discrete numbers.
#'
#' @return \code{BinomialVariate} object class, \code{DiscreteVariate} that
#' inherits from \code{numeric}, or \code{NegBinomialVariate} or
#' \code{PoissonVariate} that inherit from \code{DiscreteVariate}.
#'
#' @seealso \code{\link{role_binom}}
#'
#' @examples
#' BinomialVariate(rbinom(25, 10, 0.5), size = 10)
#' PoissonVariate(rpois(25, 10))
#'
BinomialVariate <- function(x = integer(), size = integer()) {
  stopifnot(is.finite(size))
  stopifnot(length(size) <= 1 || length(size) == length(x))
  x <- as.integer(x)
  size <- as.integer(size)
  stopifnot(x >= 0)
  stopifnot(size >= x)
  structure(cbind(Success = x, Failure = size - x), class = "BinomialVariate")
}


#' @rdname DiscreteVariate
#'
DiscreteVariate <- function(x = integer(), min = -Inf, max = Inf) {
  x <- as.integer(x)
  min <- floor(min[[1]])
  max <- floor(max[[1]])
  stopifnot(min <= x)
  stopifnot(max >= x)
  new("DiscreteVariate", x, min = min, max = max)
}


#' @rdname DiscreteVariate
#'
NegBinomialVariate <- function(x = integer()) {
  as(DiscreteVariate(x, min = 0), "NegBinomialVariate")
}


#' @rdname DiscreteVariate
#'
PoissonVariate <- function(x = integer()) {
  as(DiscreteVariate(x, min = 0), "PoissonVariate")
}


#################### SurvMatrix Classes ####################


#' SurvMatrix Class Constructors
#'
#' Create a matrix of survival events or probabilites.
#'
#' @name SurvMatrix
#' @rdname SurvMatrix
#'
#' @param data matrix, or object that can be coerced to one, with survival
#'   events or probabilities at points in time in the columns and cases in the
#'   rows.
#' @param times numeric vector of survival times for the columns.
#' @param distr character string specifying the survival distribution from which
#'   the matrix values were derived.
#'
#' @return Object that is of the same class as the constructor name and inherits
#' from \code{SurvMatrix}.  Examples of these are predicted survival events and
#' probabilities returned by the \link{predict} function.
#'
#' @seealso \code{\link{performance}}, \code{\link{metrics}}
#'
NULL


SurvMatrix <- function(data = NA, times = numeric(), distr = character()) {
  data <- as.matrix(data)

  if (is_empty(times)) times <- rep(NA_real_, ncol(data))
  if (is_empty(distr)) distr <- NA_character_

  if (length(times) != ncol(data)) {
    throw(Error("Unequal number of survival times and predictions."))
  }

  rownames(data) <- NULL
  colnames(data) <- make_names_len(length(times), "Time")

  new("SurvMatrix", data, times = times, distr = distr)
}


#' @rdname SurvMatrix
#'
SurvEvents <- function(data = NA, times = numeric(), distr = character()) {
  as(SurvMatrix(data, times, distr), "SurvEvents")
}


#' @rdname SurvMatrix
#'
SurvProbs <- function(data = NA, times = numeric(), distr = character()) {
  as(SurvMatrix(data, times, distr), "SurvProbs")
}


SurvMeans <- function(...) {
  new("SurvMeans", SurvTimes(...))
}


SurvTimes <- function(x = numeric(), distr = character(), ...) {
  if (is_empty(distr)) distr <- NA_character_
  new("SurvTimes", x, distr = distr)
}


SurvPrediction <- function(object, ...) {
  (if (is.matrix(object)) SurvProbs else SurvMeans)(object, ...)
}

## Discrete Variables

test_that("discrete variate construction and model fitting", {
  skip_if_not(TEST_ALL)
  with_parallel({

    library(MASS)
    library(recipes)

    test_fit <- function(..., model, type) {
      get_fit_type <- switch(model@name,
                             "BlackBoostModel" = function(x) x$family@name,
                             "GAMBoostModel" = function(x) x$family@name,
                             "GBMModel" = function(x) x$distribution$name,
                             "GLMBoostModel" = function(x) x$family@name,
                             "GLMModel" = function(x) x$family$family,
                             "GLMNetModel" = function(x) class(x)[3],
                             "XGBModel" = function(x) attr(x, "params")$objective
      )

      model_fit <- fit(..., model = model)
      obs <- response(model_fit)
      pred <- predict(model_fit)

      metrics <- c(mae, mse, msle, r2, rmse, rmsle)
      perf <- performance(obs, pred, metrics = metrics)

      if (is(obs, "BinomialVariate")) {
        class <- "numeric"
        perf_mse <- mean((obs[, 1] / (obs[, 1] + obs[, 2]) - pred)^2)
      } else {
        class <- class(obs)
        perf_mse <- mean((obs - pred)^2)
      }
      startsWith(get_fit_type(model_fit), type) && class(pred) == class &&
        length(perf) == length(metrics) &&
        all.equal(perf["mse"], perf_mse, check.attributes = FALSE)
    }


    ## BinomialVariate

    expect_s3_class(BinomialVariate(), "BinomialVariate")
    expect_s3_class(BinomialVariate(1:10, 10), "BinomialVariate")
    expect_s3_class(BinomialVariate(1:10, 1:10), "BinomialVariate")
    expect_error(BinomialVariate(-1:10, 10))
    expect_error(BinomialVariate(1:10, Inf))
    expect_error(BinomialVariate(1:10, 1:5))
    expect_error(BinomialVariate(1:10, 10:1))

    df <- esoph
    fo <- BinomialVariate(ncases, size = ncases + ncontrols) ~ .
    expect_true(test_fit(fo, df, model = BlackBoostModel(), type = "Binomial"))
    expect_true(test_fit(fo, df, model = GAMBoostModel(baselearner = "bols"),
                         type = "Binomial"))
    expect_true(test_fit(fo, df, model = GLMBoostModel(), type = "Binomial"))
    expect_true(test_fit(fo, df, model = GLMModel(), type = "binomial"))
    expect_true(test_fit(fo, df, model = GLMNetModel(), type = "lognet"))
    expect_s4_class(resample(fo, df, model = GLMModel), "Resample")

    rec <- recipe(ncases + n ~ ., data = within(df, n <- ncases + ncontrols)) %>%
      role_binom(x = ncases, size = n) %>%
      role_case(stratum = n) %>%
      step_rm(ncontrols)
    expect_true(test_fit(rec, model = BlackBoostModel(), type = "Binomial"))
    expect_true(test_fit(rec, model = GAMBoostModel(baselearner = "bols"),
                         type = "Binomial"))
    expect_true(test_fit(rec, model = GLMBoostModel(), type = "Binomial"))
    expect_true(test_fit(rec, model = GLMModel(), type = "binomial"))
    expect_true(test_fit(rec, model = GLMNetModel(), type = "lognet"))
    expect_s4_class(resample(rec, model = GLMModel), "Resample")


    ## DiscreteVariate

    expect_s4_class(DiscreteVariate(), "DiscreteVariate")
    expect_s4_class(DiscreteVariate(1:10), "DiscreteVariate")
    expect_s4_class(DiscreteVariate(1:10, 1, 10), "DiscreteVariate")
    expect_error(DiscreteVariate(1:10, 5, 10))
    expect_error(DiscreteVariate(1:10, 1, 5))

    df <- ICHomes
    fo <- DiscreteVariate(sale_amount, min = 0) ~ .
    expect_true(test_fit(fo, df, model = GLMModel(), type = "gaussian"))
    expect_s4_class(resample(fo, df, model = GLMModel), "Resample")

    rec <- recipe(
      sale_amount ~ .,
      data = within(df, sale_amount <- DiscreteVariate(sale_amount, min = 0))
    )
    expect_true(test_fit(rec, model = GLMModel(), type = "gaussian"))
    expect_s4_class(resample(rec, model = GLMModel), "Resample")


    ## NegBinomialVariate

    expect_s4_class(NegBinomialVariate(), "NegBinomialVariate")
    expect_s4_class(NegBinomialVariate(0:10), "NegBinomialVariate")

    df <- quine
    fo <- NegBinomialVariate(Days) ~ .
    expect_true(test_fit(fo, df, model = BlackBoostModel(), type = "Negative"))
    expect_true(test_fit(fo, df, model = GAMBoostModel(baselearner = "bols"),
                         type = "Negative"))
    expect_true(test_fit(fo, df, model = GLMBoostModel(), type = "Negative"))
    expect_true(test_fit(fo, df, model = GLMModel(),
                         type = "Negative Binomial"))
    expect_s4_class(resample(fo, df, model = GLMModel), "Resample")

    rec <- recipe(
      Days ~ .,
      data = within(df, Days <- NegBinomialVariate(Days))
    )
    expect_true(test_fit(rec, model = BlackBoostModel(), type = "Negative"))
    expect_true(test_fit(rec, model = GAMBoostModel(baselearner = "bols"),
                         type = "Negative"))
    expect_true(test_fit(rec, model = GLMBoostModel(), type = "Negative"))
    expect_true(test_fit(rec, model = GLMModel(), type = "Negative Binomial"))
    expect_s4_class(resample(rec, model = GLMModel), "Resample")


    ## PoissonVariate

    expect_s4_class(PoissonVariate(), "PoissonVariate")
    expect_s4_class(PoissonVariate(0:10), "PoissonVariate")

    df <- quine
    fo <- PoissonVariate(Days) ~ .
    expect_true(test_fit(fo, df, model = BlackBoostModel(), type = "Poisson"))
    expect_true(test_fit(fo, df, model = GAMBoostModel(baselearner = "bols"),
                         type = "Poisson"))
    expect_true(test_fit(fo, df, model = GLMBoostModel(), type = "Poisson"))
    expect_true(test_fit(fo, df, model = GLMModel(), type = "poisson"))
    expect_true(test_fit(fo, df, model = GBMModel(), type = "poisson"))
    expect_true(test_fit(fo, df, model = GLMNetModel(), type = "fishnet"))
    expect_true(test_fit(fo, df, model = XGBModel(), type = "count:poisson"))
    expect_s4_class(resample(fo, df, model = GLMModel), "Resample")

    rec <- recipe(
      Days ~ .,
      data = within(df, Days <- PoissonVariate(Days))
    )
    expect_true(test_fit(rec, model = BlackBoostModel(), type = "Poisson"))
    expect_true(test_fit(rec, model = GAMBoostModel(baselearner = "bols"),
                         type = "Poisson"))
    expect_true(test_fit(rec, model = GLMBoostModel(), type = "Poisson"))
    expect_true(test_fit(rec, model = GLMModel(), type = "poisson"))
    expect_true(test_fit(rec, model = GBMModel(), type = "poisson"))
    expect_true(test_fit(rec, model = GLMNetModel(), type = "fishnet"))
    expect_true(test_fit(rec, model = XGBModel(), type = "count:poisson"))
    expect_s4_class(resample(rec, model = GLMModel), "Resample")

  })
})

## Case Weighted Metrics

test_that("performance comparisons with and without weights", {
  skip_if_not(TEST_ALL)

  library(survival)

  n <- 100
  weights <- 1:n

  test_metrics <- function(obs, pred, weights) {
    skip_if_not(TEST_ALL)
    metrics <- names(metricinfo(obs, pred))
    perf1 <- performance(obs, pred, metrics = metrics)
    perf2 <- performance(obs, pred, weights, metrics = metrics)
    pass <- perf1 != perf2
    list(pass = all(pass), values = cbind(perf1, perf2, pass))
  }

  ## Observed factor, Predicted factor
  obs <- factor(rbinom(n, 2, 0.5))
  pred <- factor(rbinom(n, 2, 0.5))
  res <- test_metrics(obs, pred, weights)
  expect_true(res$pass)

  ## Observed factor, Predicted matrix
  obs <- factor(rbinom(n, 2, 0.5))
  pred <- matrix(runif(3 * n), ncol = 3)
  pred <- pred / rowSums(pred)
  res <- test_metrics(obs, pred, weights)
  expect_true(res$pass)

  ## Observed factor, Predicted matrix
  obs <- factor(rbinom(n, 1, 0.5))
  pred <- runif(n)
  res <- test_metrics(obs, pred, weights)
  expect_true(res$pass)

  ## Observed matrix, Predicted matrix
  obs <- matrix(rgamma(2 * n, 1, 1), ncol = 2)
  pred <- matrix(rgamma(2 * n, 1, 1), ncol = 2)
  res <- test_metrics(obs, pred, weights)
  expect_true(res$pass)

  ## Observed numeric, Predicted numeric
  obs <- rgamma(n, 1, 1)
  pred <- rgamma(n, 1, 1)
  res <- test_metrics(obs, pred, weights)
  expect_true(res$pass)

  ## Observed Surv, Predicted numeric
  obs <- Surv(rgamma(n, 1, 1), rbinom(n, 1, 0.5))
  pred <- rgamma(n, 1, 1)
  res <- test_metrics(obs, pred, weights)
  expect_true(res$pass)

  ## Observed Surv, Predicted SurvEvents
  obs <- Surv(rgamma(n, 1, 1), rbinom(n, 1, 0.5))
  pred <- SurvEvents(matrix(runif(2 * n), ncol = 2) > 0.5, times = 1:2)
  res <- test_metrics(obs, pred, weights)
  expect_true(res$pass)

  ## Observed Surv, Predicted SurvProbs
  obs <- Surv(rgamma(n, 1, 1), rbinom(n, 1, 0.5))
  pred <- SurvProbs(matrix(runif(2 * n), ncol = 2), times = 1:2)
  res <- test_metrics(obs, pred, weights)
  expect_true(res$pass)

})

## Combine Methods

test_that("object combinations", {
  skip_if_not(TEST_ALL)
  with_parallel({

    library(MASS)

    df <- Pima.tr
    fo <- type ~ .

    model <- GLMModel

    model_fit <- fit(fo, data = df, model = model)
    res <- resample(fo, data = df, model = model)

    cal <- calibration(res)
    conf <- confusion(res)
    conf_summary <- summary(conf)
    conf_perf <- performance(conf)
    curve <- performance_curve(res)
    lift_curve <- lift(res)

    expect_s4_class(c(cal), "Calibration")
    expect_s4_class(c(cal, cal), "Calibration")
    expect_type(c(cal, res), "list")

    expect_s4_class(c(conf), "ConfusionList")
    expect_s4_class(c(conf, conf), "ConfusionList")
    expect_s4_class(c(conf, conf[[1]]), "ConfusionList")
    expect_type(c(conf, res), "list")
    expect_s4_class(c(conf[[1]]), "ConfusionList")
    expect_s4_class(c(conf[[1]], conf[[1]]), "ConfusionList")
    expect_s4_class(c(conf[[1]], conf), "ConfusionList")
    expect_type(c(conf[[1]], res), "list")

    expect_s4_class(c(conf_summary), "ListOf")
    expect_s4_class(c(conf_summary, conf_summary), "ListOf")
    expect_type(c(conf_summary, conf_perf), "list")
    expect_type(c(conf_summary, res), "list")
    expect_s4_class(c(conf_perf), "ListOf")
    expect_s4_class(c(conf_perf, conf_perf), "ListOf")
    expect_type(c(conf_perf, conf_summary), "list")
    expect_type(c(conf_perf, res), "list")

    expect_s4_class(c(curve), "PerformanceCurve")
    expect_s4_class(c(curve, curve), "PerformanceCurve")
    expect_type(c(curve, res), "list")
    expect_error(c(curve, lift_curve))
    expect_s4_class(c(lift_curve), "PerformanceCurve")
    expect_s4_class(c(lift_curve, lift_curve), "PerformanceCurve")
    expect_type(c(lift_curve, curve), "list")

    expect_s4_class(c(res), "Resample")
    expect_s4_class(c(res, res), "Resample")
    expect_type(c(res, cal), "list")
    expect_type(c(res, conf), "list")
    expect_type(c(res, conf[[1]]), "list")
    expect_type(c(res, conf_summary), "list")
    expect_type(c(res, conf_perf), "list")
    expect_type(c(res, curve), "list")
    expect_type(c(res, lift_curve), "list")

    x <- runif(10, 0, 100)
    binom <- BinomialVariate(x, size = 100)
    disc <- DiscreteVariate(x)
    negbinom <- NegBinomialVariate(x)
    pois <- PoissonVariate(x)

    expect_s3_class(c(binom), "BinomialVariate")
    expect_s3_class(c(binom, binom), "BinomialVariate")
    expect_type(c(binom, disc), "integer")
    expect_type(c(binom, negbinom), "integer")
    expect_type(c(binom, pois), "integer")
    expect_type(c(binom, res), "list")
    expect_s4_class(c(disc), "DiscreteVariate")
    expect_type(c(disc, binom), "integer")
    expect_s4_class(c(disc, disc), "DiscreteVariate")
    expect_s4_class(c(disc, negbinom), "DiscreteVariate")
    expect_s4_class(c(disc, pois), "DiscreteVariate")
    expect_type(c(disc, res), "list")
    expect_s4_class(c(negbinom), "NegBinomialVariate")
    expect_type(c(negbinom, binom), "integer")
    expect_type(c(negbinom, disc), "integer")
    expect_s4_class(c(negbinom, negbinom), "NegBinomialVariate")
    expect_type(c(negbinom, pois), "integer")
    expect_type(c(negbinom, res), "list")
    expect_s4_class(c(pois), "PoissonVariate")
    expect_type(c(pois, binom), "integer")
    expect_type(c(pois, disc), "integer")
    expect_type(c(pois, negbinom), "integer")
    expect_s4_class(c(pois, pois), "PoissonVariate")
    expect_type(c(pois, res), "list")

    x <- DiscreteVariate(1:10, 1, 10)
    y <- DiscreteVariate(-(10:1), -10, -1)
    z <- expect_s4_class(c(x, y), "DiscreteVariate")
    expect_equal(length(z), length(x) + length(y))
    expect_equal(z@min, min(x@min, y@min))
    expect_equal(z@max, max(x@max, y@max))

    x <- matrix(1, 10, 2)
    surv_events <- SurvEvents(x, times = 1:2)
    surv_probs <- SurvProbs(x, times = 1:2)
    expect_s4_class(c(surv_events, surv_events), "SurvEvents")
    expect_s4_class(c(surv_probs, surv_probs), "SurvProbs")
    expect_is(c(surv_events, surv_probs), "numeric")
    expect_is(c(surv_probs, surv_events), "numeric")
    expect_error(c(surv_events, SurvEvents(x, times = 2:3)))
    expect_error(c(surv_probs, SurvProbs(x, times = 2:3)))

  })
})

#' Generalized Boosted Regression Model
#'
#' Fits generalized boosted regression models.
#'
#' @param distribution optional character string specifying the name of the
#'   distribution to use or list with a component \code{name} specifying the
#'   distribution and any additional parameters needed.  Set automatically
#'   according to the class type of the response variable.
#' @param n.trees total number of trees to fit.
#' @param interaction.depth maximum depth of variable interactions.
#' @param n.minobsinnode minimum number of observations in the trees terminal
#'   nodes.
#' @param shrinkage shrinkage parameter applied to each tree in the expansion.
#' @param bag.fraction fraction of the training set observations randomly
#'   selected to propose the next tree in the expansion.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric},
#'     \code{PoissonVariate}, \code{Surv}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \code{n.trees}, \code{interaction.depth}, \code{shrinkage}*,
#'     \code{n.minobsinnode}*
#'   }
#' }
#' * excluded from grids by default
#'
#' Default argument values and further model details can be found in the source
#' See Also link below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[gbm]{gbm}}, \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' fit(Species ~ ., data = iris, model = GBMModel)
#' }
#'
GBMModel <- function(
  distribution = character(), n.trees = 100, interaction.depth = 1,
  n.minobsinnode = 10, shrinkage = 0.1, bag.fraction = 0.5
) {

  MLModel(

    name = "GBMModel",
    label = "Generalized Boosted Regression",
    packages = "gbm",
    response_types = c("factor", "numeric", "PoissonVariate", "Surv"),
    weights = TRUE,
    predictor_encoding = "model.frame",
    na.rm = "response",
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = c("n.trees", "interaction.depth", "shrinkage", "n.minobsinnode"),
      get_values = c(
        function(n, ...) round_int(seq_range(0, 50, c(1, 1000), n + 1)),
        function(n, ...) seq_len(min(n, 10)),
        function(n, ...) seq(0.001, 0.1, length = n),
        function(n, data, ...) {
          round_int(seq(1, min(20, nrow(data)), length = n))
        }
      ),
      default = c(TRUE, TRUE, FALSE, FALSE)
    ),

    fit = function(formula, data, weights, distribution = NULL, ...) {
      if (is.null(distribution)) {
        y <- response(data)
        distribution <- switch_class(y,
          "factor" = if (nlevels(y) <= 2) {
            data[[response(formula)]] <- as.numeric(y) - 1
            "bernoulli"
          } else {
            data[[response(formula)]] <- structure(y, class = "factor")
            "multinomial"
          },
          "numeric" = "gaussian",
          "PoissonVariate" = "poisson",
          "Surv" = "coxph"
        )
      }
      cond <- NULL
      suppressWarnings(withCallingHandlers(
        res <- eval_fit(
          data,
          formula = gbm::gbm(
            formula, data = data, weights = weights,
            distribution = distribution, ...
          ),
          matrix = gbm::gbm.fit(
            x, y, offset = model.offset(data), w = weights,
            distribution = distribution, verbose = FALSE, ...
          )
        ),
        warning = function(cond) cond <<- cond
      ))
      if (is(cond, "warning")) {
        if (distribution == "multinomial") {
          throw(Warning(conditionMessage(cond)), call = call("gbm"), times = 3)
        } else {
          warning(cond)
        }
      }
      res
    },

    predict = function(object, newdata, .MachineShop, ...) {
      input <- .MachineShop$input
      newdata <- as.data.frame(newdata)
      n <- object$n.trees
      if (object$distribution$name == "coxph") {
        y <- response(input)
        data <- as.data.frame(PredictorFrame(input))
        lp <- predict(object, newdata = data, n.trees = n, type = "link")
        new_lp <- predict(object, newdata = newdata, n.trees = n, type = "link")
        predict(y, lp, new_lp, weights = case_weights(input), ...)
      } else {
        predict(object, newdata = newdata, n.trees = n, type = "response")
      }
    },

    varimp = function(object, ...) {
      structure(
        gbm::relative.influence(object, n.trees = object$n.trees),
        metric = "influence"
      )
    }

  )

}

MLModelFunction(GBMModel) <- NULL

## Offset Terms

test_that("test model offsets", {
  skip_if_not(TEST_ALL)

  library(dials)
  library(recipes)

  df <- esoph
  fo <- PoissonVariate(ncases) ~ agegp + alcgp + tobgp +
    offset(log(ncases + ncontrols))

  rec_df <- within(esoph, {
    offset <- log(ncases + ncontrols)
    ncases <- PoissonVariate(ncases)
  })
  rec <- recipe(ncases ~ agegp + alcgp + tobgp + offset, data = rec_df) %>%
    role_pred(offset = offset, replace = TRUE)

  test_fit <- function(model) {
    has_offset <- switch(model@name,
                         "GBMModel" = {
                           function(x) !is.null(attr(x$Terms, "offset"))
                         },
                         "GLMModel" = function(x) !is.null(x$offset),
                         "GLMNetModel" = function(x) !is.null(x$offset))


    model_fit <- expect_is(fit(fo, df, model = model), "MLModelFit")
    expect_true(has_offset(model_fit))
    expect_s4_class(predict(model_fit), "PoissonVariate")

    model_fit <- expect_is(fit(rec, model = model), "MLModelFit")
    expect_true(has_offset(model_fit))
    expect_s4_class(predict(model_fit), "PoissonVariate")
  }

  ## GBMModel fitting and prediction
  test_fit(GBMModel())

  ## GLMModel fitting and prediction
  test_fit(GLMModel())

  ## GLMMetModel fitting and prediction
  test_fit(GLMNetModel())

})

## Model Comparisons


test_that("Resample differences", {
  skip_if_not(TEST_MODEL_COMPARISONS)
  with_parallel({

    test_Resample <- function() {
      fo <- factor(Species) ~ .
      control <- CVControl()

      gbmperf1 <- resample(fo, iris, GBMModel(n.trees = 25), control)
      gbmperf2 <- resample(fo, iris, GBMModel(n.trees = 50), control)
      gbmperf3 <- resample(fo, iris, GBMModel(n.trees = 100), control)

      perf <- c(GBM1 = gbmperf1, GBM2 = gbmperf2, GBM3 = gbmperf3)
      summary(perf)
      plot(perf)

      perfdiff <- diff(perf)
      summary(perfdiff)
      plot(perfdiff)
      t.test(perfdiff)
    }

    expect_s4_class(test_Resample(), "PerformanceDiffTest")

  })
})


test_that("TunedModel differences", {
  skip_if_not(TEST_MODEL_COMPARISONS)

  test_TunedModel <- function() {
    library(MASS)
    fo <- medv ~ .

    gbmfit <- fit(fo, data = Boston,
                  model = TunedModel(
                    GBMModel,
                    grid = expand.grid(n.trees = c(25, 50, 100),
                                       interaction.depth = 1:3,
                                       n.minobsinnode = c(5, 10)),
                    control = CVControl(folds = 10, repeats = 5)
                  ))
    gbmtune <- as.MLModel(gbmfit)
    summary(gbmtune)
    plot(gbmtune, type = "line")

    gbmtunediff <- diff(gbmtune)
    summary(gbmtunediff)
    t.test(gbmtunediff[[1]])
  }

  with_parallel({
    expect_s4_class(test_TunedModel(), "PerformanceDiffTest")
  })
})

#' Naive Bayes Classifier Model
#'
#' Computes the conditional a-posterior probabilities of a categorical class
#' variable given independent predictor variables using Bayes rule.
#'
#' @param laplace positive numeric controlling Laplace smoothing.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}}
#' }
#'
#' Further model details can be found in the source link below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[e1071]{naiveBayes}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package e1071 to run
#'
#' fit(Species ~ ., data = iris, model = NaiveBayesModel)
#' }
#'
NaiveBayesModel <- function(laplace = 0) {

  MLModel(

    name = "NaiveBayesModel",
    label = "Naive Bayes Classifier",
    packages = "e1071",
    response_types = "factor",
    predictor_encoding = "model.frame",
    na.rm = FALSE,
    params = new_params(environment()),

    fit = function(formula, data, weights, ...) {
      eval_fit(
        data,
        formula = e1071::naiveBayes(
          formula, data = data, na.action = na.pass, ...
        ),
        matrix = e1071::naiveBayes(x, y, ...)
      )
    },

    predict = function(object, newdata, ...) {
      newdata <- as.data.frame(newdata)
      predict(object, newdata = newdata, type = "raw")
    }

  )

}

MLModelFunction(NaiveBayesModel) <- NULL

#' @rdname metrics
#'
gini <- function(observed, predicted = NULL, weights = NULL, ...) {
  call_metric_method("gini", environment())
}

MLMetric(gini) <- list("gini", "Gini Coefficient", FALSE)


setMetric_numeric("gini",
  function(observed, predicted, weights, ...) {
    weights <- check_weights(weights, observed)
    throw(check_assignment(weights))
    f <- function(x) {
      sort_order <- order(x, weights)
      y <- observed[sort_order]
      w <- weights[sort_order]
      n <- length(x)

      y <- w * y
      y <- cumsum(y)
      y <- y / y[n]

      w <- cumsum(w)
      w <- w / w[n]

      c(y[-1] %*% w[-n] - y[-n] %*% w[-1])
    }
    f(predicted) / f(observed)
  }
)


#' @rdname metrics
#'
mae <- function(observed, predicted = NULL, weights = NULL, ...) {
  call_metric_method("mae", environment())
}

MLMetric(mae) <- list("mae", "Mean Absolute Error", FALSE)


setMetric_numeric("mae",
  function(observed, predicted, weights, ...) {
    weighted_mean(abs(observed - predicted), weights)
  }
)


#' @rdname metrics
#'
mse <- function(observed, predicted = NULL, weights = NULL, ...) {
  call_metric_method("mse", environment())
}

MLMetric(mse) <- list("mse", "Mean Squared Error", FALSE)


setMetric_numeric("mse",
  function(observed, predicted, weights, ...) {
    weighted_mean((observed - predicted)^2, weights)
  }
)


#' @rdname metrics
#'
msle <- function(observed, predicted = NULL, weights = NULL, ...) {
  call_metric_method("msle", environment())
}

MLMetric(msle) <- list("msle", "Mean Squared Log Error", FALSE)


setMetric_numeric("msle",
  function(observed, predicted, weights, ...) {
    mse(log(1 + observed), log(1 + predicted), weights)
  }
)


#' @rdname metrics
#'
r2 <- function(
  observed, predicted = NULL, weights = NULL,
  method = c("mse", "pearson", "spearman"), distr = character(), ...
) {
  method <- match.arg(method)
  call_metric_method("r2", environment())
}

MLMetric(r2) <- list("r2", "Coefficient of Determination", TRUE)


setMetricGeneric("r2")


setMetricMethod_BinomialVariate("r2")


setMetricMethod_matrix("r2")


setMetricMethod("r2", c("numeric", "numeric"),
  function(observed, predicted, weights, method, ...) {
    if (method == "mse") {
      obs_mean <- weighted_mean(observed, weights)
      1 - mse(observed, predicted, weights) / mse(observed, obs_mean, weights)
    } else {
      if (method == "spearman") {
        observed <- rank(observed)
        predicted <- rank(predicted)
      }
      weights <- check_weights(weights, observed)
      throw(check_assignment(weights))
      cov.wt(cbind(observed, predicted), wt = weights, cor = TRUE)$cor[2]^2
    }
  }
)


setMetricMethod_Resample("r2")


setMetricMethod("r2", c("Surv", "numeric"),
  function(observed, predicted, weights, method, distr, ...) {
    if (method == "mse") {
      distr <- get_surv_distr(distr, observed, predicted)
      nparams <- if (distr %in% c("exponential", "rayleigh")) 1 else 2
      obs_mean <- if (distr == "empirical") {
        surv <- survfit(observed ~ 1, weights = weights, se.fit = FALSE)
        rep(mean(surv), length(observed))
      } else if (length(event_time(observed)) >= nparams) {
        predict(survreg(observed ~ 1, weights = weights, dist = distr))
      } else {
        rep(NA_real_, length(observed))
      }
      1 - mse(observed, predicted, weights) / mse(observed, obs_mean, weights)
    } else {
      metric_SurvTimes(r2, observed, predicted, weights, method = method)
    }
  }
)


#' @rdname metrics
#'
rmse <- function(observed, predicted = NULL, weights = NULL, ...) {
  call_metric_method("rmse", environment())
}

MLMetric(rmse) <- list("rmse", "Root Mean Squared Error", FALSE)


setMetric_numeric("rmse",
  function(observed, predicted, weights, ...) {
    sqrt(mse(observed, predicted, weights))
  }
)


#' @rdname metrics
#'
rmsle <- function(observed, predicted = NULL, weights = NULL, ...) {
  call_metric_method("rmsle", environment())
}

MLMetric(rmsle) <- list("rmsle", "Root Mean Squared Log Error", FALSE)


setMetric_numeric("rmsle",
  function(observed, predicted, weights, ...) {
    sqrt(msle(observed, predicted, weights))
  }
)

#################### Constructors ####################


DeprecatedCondition <- function(
  old, new, value = NULL, expired = FALSE, call = FALSE
) {
  make_cond <- function(f, class) {
    f(
      message = paste0(old, " is deprecated; use ", new, " instead."),
      value = value, call = select_call(call, sys.call(-2)), class = class
    )
  }
  if (expired) {
    make_cond(errorCondition, "DeprecatedError")
  } else {
    make_cond(warningCondition, "DeprecatedWarning")
  }
}


DomainError <- function(value, ..., call = FALSE) {
  errorCondition(
    message = paste0("Value ", ..., "."),
    call = select_call(call, sys.call(-1)),
    value = value,
    class = "DomainError"
  )
}


Error <- function(..., call = FALSE) {
  errorCondition(
    message = paste0(...),
    call = select_call(call, sys.call(-1))
  )
}


LocalError <- function(...) {
  errorCondition(
    message = paste0(...),
    class = "LocalError"
  )
}


LocalWarning <- function(..., value = NULL) {
  warningCondition(
    message = paste0(...),
    value = value,
    class = "LocalWarning"
  )
}


TypeError <- function(value, expected, subject = character(), call = FALSE) {
  errorCondition(
    message = paste0(
      "Expected ", subject, if (length(subject)) " to be ",
      as_string(expected, conj = "or"), "; got ", class1(value), " instead."
    ),
    call = select_call(call, sys.call(-1)),
    class = "TypeError"
  )
}


Warning <- function(..., value = NULL, call = FALSE) {
  warningCondition(
    message = paste0(...),
    value = value,
    call = select_call(call, sys.call(-1))
  )
}


#################### Throw ####################


throw <- function(x, call = TRUE, ...) {
  .throw(x, call = call, parent_call = sys.call(-1), ...)
}


.throw <- function(x, ...) {
  UseMethod(".throw")
}


.throw.default <- function(x, ...) {
  x
}


.throw.error <- function(x, call, parent_call, ...) {
  if (match("error", class(x)) > 1) {
    x$message <- paste0(class1(x), ": ", x$message)
  }
  x$call <- select_call(call, parent_call, conditionCall(x))
  stop(x)
}


.throw.LocalError <- function(x, ...) {
  stop(x$message, call. = FALSE)
}


.throw.LocalWarning <- function(x, immediate = FALSE, ...) {
  warning(x$message, call. = FALSE, immediate. = immediate)
  x$value
}


.throw.warning <- function(x, call, parent_call, times = Inf, ...) {
  x$call <- select_call(call, parent_call, conditionCall(x))
  if (is.call(x$call) && is.finite(times)) {
    name <- x$call[[1]]
    times <- min(times, MachineShop_global$throw_times[[name]])
    MachineShop_global$throw_times[[name]] <- times - 1
  }
  if (times > 0) {
    if (match("warning", class(x)) > 1) {
      x$message <- paste0(class1(x), ": ", x$message)
    }
    warning(x)
  }
  x$value
}


select_call <- function(x, parent_call = NULL, last_call = NULL) {
  if (isTRUE(x)) {
    if (is.null(last_call)) parent_call else last_call
  } else if (is.call(x)) {
    x
  }
}


#################### Checks ####################


check_array <- function(
  x, type, size = integer(), nonempty = TRUE, na.fail = TRUE
) {
  tryCatch(
    {
      storage.mode(x) <- type
      result <- x
      n <- length(size)
      if (n > 0 && (ndim(x) != n || any(na.omit(size(x) != size)))) {
        msg <- if (identical(size, 1)) {
          "a scalar"
        } else if (n == 1) {
          paste("a vector of length", size)
        } else {
          paste(if (n == 2) "a matrix" else "an array", "of dimension",
                paste(size, collapse = "x"))
        }
        DomainError(x, paste("must be", msg))
      } else if (nonempty && is_empty(x)) {
        DomainError(x, "must be non-empty")
      } else if (na.fail && anyNA(result)) {
        msg <- paste0(
          c("", "must be ", paste("non-missing", type)),
          if (length(result) == 1) c("", "a ", "") else c("elements ", "", "s"),
          collapse = ""
        )
        DomainError(result, msg)
      } else {
        result
      }
    },
    error = function(cond) TypeError(x, type)
  )
}


check_assignment <- function(x, value = x) {
  msg <- paste0(" `", deparse1(substitute(x)), "` value.\n")
  if (is(value, "error")) {
    value$message <- paste0("Failed to assign", msg, value$message)
  } else if (is(value, "condition")) {
    value$message <- paste0("Issue in assigning", msg, value$message)
  }
  value
}


check_censoring <- function(x, types, ...) {
  type <- attr(x, "type")
  if (!(type %in% types)) {
    types <- paste0("\"", types, "\"")
    Error(
      "Expected survival data censoring type to be ",
      as_string(types, conj = "or"), "; got \"", type, "\" instead."
    )
  } else x
}


check_character <- function(x, ...) {
  check_array(x, type = "character", ...)
}


check_const_setting <- function(x, name) {
  if (!identical(x, x <- .global_defaults[[name]])) {
    LocalWarning(
      "MachineShop `", name, "` setting cannot be changed.", value = x
    )
  } else x
}


check_equal_weights <- function(x) {
  if (length(x) && any(diff(x) != 0)) {
    Warning("Model weights are not supported and will be ignored.", value = x)
  } else x
}


check_grid <- function(x) {
  if (is(x, "numeric")) {
    TuningGrid(x)
  } else if (identical(x, "TuningGrid") || identical(x, TuningGrid)) {
    TuningGrid()
  } else if (is(x, "TuningGrid")) {
    x
  } else {
    DomainError(
      x, "must be one or more positive integers or a TuningGrid function, ",
      "TuningGrid function, function name, or object"
    )
  }
}


check_integer <- function(x, bounds = c(-Inf, Inf), include = TRUE, ...) {
  include <- include & is.finite(bounds)
  check_numeric(x, bounds = bounds, include = include, type = "integer", ...)
}


check_logical <- function(x, ...) {
  check_array(x, type = "logical", ...)
}


check_match <- function(x, choices) {
  tryCatch(
    match.arg(x, choices),
    error = function(cond) {
      choices <- paste0("\"", choices, "\"")
      DomainError(x, "must be one of ", as_string(choices, conj = "or"))
    }
  )
}


check_metric <- function(x, convert = FALSE) {
  tryCatch(
    {
      result <- as.MLMetric(x)
      if (convert) result else x
    },
    error = function(cond) {
      DomainError(x, "must be a metrics function or function name")
    }
  )
}


check_metrics <- function(x, convert = FALSE) {
  tryCatch(
    {
      map(as.MLMetric, c(x))
      if (convert) vector_to_function(x, "metric") else x
    },
    error = function(cond) {
      DomainError(
        x, "must be a metrics function, function name, or vector of these"
      )
    }
  )
}


check_na.rm <- function(x) {
  if (is.logical(x)) {
    if (x) "all" else "none"
  } else {
    na.rm <- check_match(x, c("none", "all", "response"))
    check_assignment(na.rm)
  }
}


check_numeric <- function(
  x, bounds = c(-Inf, Inf), include = TRUE,
  type = c("numeric", "double", "integer"), ...
) {
  result <- check_array(x, type = match.arg(type), ...)
  if (is(result, "error")) return(result)

  include <- rep_len(include, 2)
  ops <- paste0(c(">", "<"), ifelse(include, "=", ""))
  values <- na.omit(c(result))
  inbounds <- function(op, bound) all(do.call(op, list(values, bound)))
  if (!(inbounds(ops[1], bounds[1]) && inbounds(ops[2], bounds[2]))) {
    msg <- paste0(
      if (length(result) > 1) "elements ", "must be ",
      paste(ops, bounds, collapse = " and ")
    )
    DomainError(result, msg)
  } else {
    result
  }
}


check_optim_bounds <- function(x) {
  if (is_empty(x)) {
    Error("No set of bounded tuning parameters to optimize.")
  } else x
}


check_packages <- function(x) {
  paren_pos <- regexpr("\\(([^)]*)\\)", x)
  paren_len <- attr(paren_pos, "match.length")

  end_pos <- ifelse(paren_pos > 0, paren_pos - 1, nchar(x))
  pkg_names <- trimws(substr(x, 1, end_pos))

  check <- function(failures, msg, pkgs_fun) {
    if (any(failures)) {
      x <- x[failures]
      pkg_names <- pkg_names[failures]
      Error(
        "Call ", note_items(msg, x), ".\n",
        "To address this issue, try running ", pkgs_fun, "(",
        deparse1(pkg_names), ")."
      )
    }
  }

  installed <- map("logi", requireNamespace, pkg_names, quietly = TRUE)
  result <- check(
    !installed, "requires prior installation of package{?s}: ",
    "install.packages"
  )
  if (is(result, "error")) return(result)

  end_pos <- paren_pos + paren_len - 2
  compat_versions <- strsplit(substr(x, paren_pos + 1, end_pos), " ")

  compatible <- map("logi", function(pkg_name, compat_version) {
    if (length(compat_version) == 2) {
      version <- packageVersion(pkg_name)
      eval(call(compat_version[1], version, compat_version[2]))
    } else TRUE
  }, pkg_names, compat_versions)
  result <- check(
    !compatible, "requires updated package version{?s}: ", "update.packages"
  )
  if (is(result, "error")) return(result)

  x
}


check_stat <- function(x, convert = FALSE) {
  result <- try({
    stat <- fget(x)
    stat(1:5)
  }, silent = TRUE)
  if (!is.numeric(result) || length(result) != 1) {
    DomainError(x, "must be a statistic function or function name")
  } else if (convert) {
    stat
  } else {
    x
  }
}


check_stats <- function(x, convert = FALSE) {
  result <- try({
    stats <- vector_to_function(x, "stat")
    stats(1:5)
  }, silent = TRUE)
  if (!is.numeric(result)) {
    DomainError(
      x, "must be a statistics function, function name, or vector of these"
    )
  } else if (convert) {
    stats
  } else {
    x
  }
}


check_weights <- function(x, along) {
  n <- length(along)
  if (is.null(x)) {
    rep(1, n)
  } else {
    check_numeric(x, bounds = c(0, Inf), include = c(TRUE, FALSE), size = n)
  }
}

VariableImportance <- function(object, ...) {
  UseMethod("VariableImportance")
}


VariableImportance.data.frame <- function(
  object, method = character(), metric = character(), scale = TRUE,
  sort = "decreasing", ...
) {
  stopifnot(nrow(object) == 0 || is.character(rownames(object)))

  object <- object[rownames(object) != "(Intercept)", , drop = FALSE]
  scale <- rep_len(scale, ncol(object))
  if (any(scale)) {
    scale_num <- max(object[scale], 0) / 100
    object[scale] <- object[scale] / scale_num
    if (!all(scale)) scale_num <- ifelse(scale, scale_num, NA_real_)
    sort_vals <- if (scale[1]) rowSums(object[scale]) else object[[1]]
  } else {
    scale_num <- NA_real_
    sort_vals <- object[[1]]
  }
  if (sort != "asis") {
    sort_order <- order(sort_vals, decreasing = sort == "decreasing")
    object <- object[sort_order, , drop = FALSE]
  }

  new("VariableImportance",
    object, method = method, metric = as.character(metric), scale = scale_num
  )
}


VariableImportance.matrix <- function(object, ...) {
  VariableImportance(as.data.frame(object), ...)
}


VariableImportance.numeric <- function(object, ...) {
  VariableImportance(cbind(Overall = object), ...)
}


VariableImportance.NULL <- function(object, ...) {
  throw(Error("No variables for which to compute importance."))
}


update.VariableImportance <- function(object, ...) {
  if (!.hasSlot(object, "method")) {
    names_split <- strsplit(names(object), ".", fixed = TRUE)
    if (names_split[[1]][1] == "Permute") {
      method <- "permute"
      metric <- names_split[[1]][3]
      stats <- sapply(names_split, function(x) x[2])
      stats[nchar(stats) == 0] <- "stat"
      names(object) <- make_unique(stats, sep = "")
    } else{
      method <- "model"
      metric <- character()
    }
    slot(object, "method", check = FALSE) <- method
    slot(object, "metric", check = FALSE) <- metric
  }
  object
}


#' Variable Importance
#'
#' Calculate measures of relative importance for model predictor variables.
#'
#' @rdname varimp
#'
#' @param object model \link{fit} result.
#' @param method character string specifying the calculation of variable
#'   importance as permutation-base (\code{"permute"}) or model-specific
#'   (\code{"model"}).  If model-specific importance is specified but not
#'   defined, the permutation-based method will be used instead with its default
#'   values (below).  Permutation-based variable importance is defined as the
#'   relative change in model predictive performances between datasets with and
#'   without permuted values for the associated variable (Fisher et al. 2019).
#' @param scale logical value or vector indicating whether importance values are
#'   scaled to a maximum of 100.
#' @param sort character string specifying the sort order of importance values
#'   to be \code{"decreasing"}, \code{"increasing"}, or as predictors appear in
#'   the model formula (\code{"asis"}).
#' @param ... arguments passed to model-specific or permutation-based variable
#'   importance functions.  These include the following arguments and default
#'   values for \code{method = "permute"}.
#'   \describe{
#'     \item{\code{select = NULL}}{expression indicating predictor variables for
#'       which to compute variable importance (see \code{\link[base]{subset}}
#'       for syntax) [default: all].}
#'     \item{\code{samples = 1}}{number of times to permute the values of each
#'       variable.  Larger numbers of samples decrease variability in the
#'       estimates at the expense of increased computation time.}
#'     \item{\code{prop = numeric()}}{proportion of observations to sample
#'       without replacement at each round of variable permutations [default:
#'       all].  Subsampling of observations can decrease computation time.}
#'     \item{\code{size = integer()}}{number of observations to sample at each
#'       round of permutations [default: all].}
#'     \item{\code{times = numeric()}}{numeric vector of follow-up times at
#'       which to predict survival probabilities or \code{NULL} for predicted
#'       survival means.}
#'     \item{\code{metric = NULL}}{\link[=metrics]{metric} function or function
#'       name with which to calculate performance.  If not specified, the first
#'       applicable default metric from the \link{performance} functions is
#'       used.}
#'     \item{\code{compare = c("-", "/")}}{character specifying the relative
#'       change to compute in comparing model predictive performances between
#'       datasets with and without permuted values.  The choices are difference
#'       (\code{"-"}) and ratio (\code{"/"}).}
#'     \item{\code{stats = MachineShop::settings("stat.TrainingParams")}}{
#'       function, function name, or vector of these with which to compute
#'       summary statistics on the set of variable importance values from the
#'       permuted datasets.}
#'     \item{\code{na.rm = TRUE}}{logical indicating whether to exclude missing
#'       variable importance values from the calculation of summary statistics.}
#'     \item{\code{progress = TRUE}}{logical indicating whether to display
#'       iterative progress during computation.}
#'   }
#'
#' @details
#' The \code{varimp} function supports calculation of variable importance with
#' the permutation-based method of Fisher et al. (2019) or with model-based
#' methods where defined.  Permutation-based importance is the default and has
#' the advantages of being available for any model, any performance metric
#' defined for the associated response variable type, and any predictor variable
#' in the original training dataset.  Conversely, model-specific importance is
#' not defined for some models and will fall back to the permutation method in
#' such cases; is generally limited to metrics implemented in the source
#' packages of models; and may be computed on derived, rather than original,
#' predictor variables.  These disadvantages can make comparisons of
#' model-specific importance across different classes of models infeasible.  A
#' downside of the permutation-based approach is increased computation time.  To
#' counter this, the permutation algorithm can be run in parallel simply by
#' loading a parallel backend for the \pkg{foreach} package \code{\%dopar\%}
#' function, such as \pkg{doParallel} or \pkg{doSNOW}.
#'
#' Permutation variable importance is interpreted as the contribution of a
#' predictor variable to the predictive performance of a model as measured by
#' the performance metric used in the calculation.  Importance of a predictor is
#' conditional on and, with the default scaling, relative to the values of all
#' other predictors in the analysis.
#'
#' @return \code{VariableImportance} class object.
#'
#' @references
#' Fisher, A., Rudin, C., & Dominici, F. (2019). All models are wrong, but many
#' are useful: Learning a variable's importance by studying an entire class of
#' prediction models simultaneously. \emph{Journal of Machine Learning
#' Research}, \emph{20}, 1-81.
#'
#' @seealso \code{\link{plot}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' ## Survival response example
#' library(survival)
#'
#' gbm_fit <- fit(Surv(time, status) ~ ., data = veteran, model = GBMModel)
#' (vi <- varimp(gbm_fit))
#' plot(vi)
#' }
#'
varimp <- function(
  object, method = c("permute", "model"), scale = TRUE,
  sort = c("decreasing", "increasing", "asis"), ...
) {
  stopifnot(is(object, "MLModelFit"))
  object <- update(object)
  model <- as.MLModel(object)
  throw(check_packages(model@packages))

  method <- match.arg(method)
  sort <- match.arg(sort)

  switch(method,
    "model" = {
      .MachineShop <- attr(object, ".MachineShop")
      vi <- model@varimp(unMLModelFit(object), .MachineShop = .MachineShop, ...)
      if (is.null(vi)) {
        throw(Warning(
          "Model-specific variable importance is not available; ",
          "computing permutation-based importance instead."
        ))
        method <- "permute"
        vi <- varimp_permute(object)
      }
    },
    "permute" = {
      args <- eval(substitute(alist(object, ...)))
      vi <- do.call(varimp_permute, args, envir = parent.frame())
    }
  )
  metric <- attr(vi, "metric")
  attr(vi, "metric") <- NULL
  
  VariableImportance(
    vi, method = method, metric = metric, scale = scale, sort = sort
  )
}


varimp_permute <- function(
  object, select = NULL, samples = 1, prop = numeric(), size = integer(),
  times = numeric(), metric = NULL, compare = c("-", "/"),
  stats = MachineShop::settings("stat.TrainingParams"), na.rm = TRUE,
  progress = TRUE
) {
  input <- as.MLInput(object)
  data <- if (is.data.frame(input)) input else as.data.frame(input)
  pred_names <- all.vars(predictors(terms(input, original = TRUE)))
  pred_names <- do.call(
    subset_names, list(pred_names, substitute(select)), envir = parent.frame()
  )

  samples <- check_integer(samples, bounds = c(1, Inf), size = 1)
  throw(check_assignment(samples))

  n <- nrow(data)
  size <- if (is_empty(c(prop, size))) {
    n
  } else if (is_empty(size)) {
    prop <- check_numeric(prop, bounds = c(0, 1), include = 0:1, size = 1)
    throw(check_assignment(prop))
    min(max(prop * n, 2), n)
  } else if (is_empty(prop)) {
    size <- check_numeric(size, bounds = c(2, Inf), size = 1)
    throw(check_assignment(size))
    min(size, n)
  } else {
    throw(Error("Arguments `prop` and `size` cannot be specified together."))
  }

  if (!is.null(metric)) {
    metric <- check_metric(metric, convert = TRUE)
    throw(check_assignment(metric))
  }
  compare <- match.arg(compare)
  stats <- check_stats(stats, convert = TRUE)
  throw(check_assignment(stats))

  varimp <- function(x, baseline, maximize = FALSE) {
    compare_args <- if (maximize) list(baseline, x) else list(x, baseline)
    do.call(rbind,
      apply(do.call(compare, compare_args), 2, function(col) {
        stats(if (na.rm) na.omit(col) else col)
      }, simplify = FALSE)
    )
  }

  progress <- if (throw(check_logical(progress))) {
    pb <- progress_bar$new(
      format = "varimp permute [:bar] :percent | :eta",
      total = samples * length(pred_names),
      show_after = 1
    )
    on.exit(pb$terminate())
    function() pb$tick()
  } else {
    function() NULL
  }

  workers <- rep(1:getDoParWorkers(), length = length(pred_names))
  jobs <- split(pred_names, sort(workers))
  seeds <- rand_int(samples)

  presets <- settings()
  foreach(
    pred_names = jobs[lengths(jobs) > 0],
    .multicombine = TRUE,
    .combine = function(...) {
      structure(rbind(...), metric = attr(..1, "metric"))
    },
    .inorder = TRUE,
    .export = c("as.MLMetric", "get_perf_metrics", "map", "permute_int"),
    .packages = "MachineShop"
  ) %dopar% {
    settings(presets)
    subset <- size < n
    base_perf <- rep(NA_real_, samples)
    perf <- matrix(NA_real_, samples, length(pred_names))
    colnames(perf) <- pred_names
    for (s in seq_len(samples)) {
      set.seed(seeds[s])
      inds <- permute_int(n, size)
      base_perf[s] <- if (s == 1 || subset) {
        newdata <- if (subset) data[inds$i, ] else data
        obs <- response(object, newdata)
        pred <- predict(object, newdata, times = times, type = "raw")
        if (is.null(metric)) {
          metric <- as.MLMetric(get_perf_metrics(obs, pred)[[1]])
        }
        metric(obs, pred)[1]
      } else {
        base_perf[1]
      }
      for (name in pred_names) {
        x <- newdata[[name]]
        newdata[[name]] <- data[[name]][inds$j]
        pred <- predict(object, newdata, times = times, type = "raw")
        perf[s, name] <- metric(obs, pred)[1]
        newdata[[name]] <- x
        progress()
      }
    }
    structure(varimp(perf, base_perf, metric@maximize), metric = metric@name)
  }
}


varimp_pval <- function(object, ...) {
  UseMethod("varimp_pval")
}


varimp_pval.default <- function(object, test = "Chisq", base = exp(1), ...) {
  res <- rev(drop1(object, test = test))[-1, 1, drop = FALSE]
  structure(
    -log(res[[1]], base = base),
    names = rownames(res),
    metric = paste0(
      "-log", if (base != exp(1)) paste0("(base = ", as_string(base), ")"),
      " ", names(res)
    )
  )
}


varimp_pval.mlm <- function(object, ...) {
  check <- check_equal_weights(object$weights)
  if (is(check, "warning")) {
    throw(LocalWarning(
      "Model-specific variable importance not defined for ", class1(object),
      " with case weights."
    ))
    NULL
  } else {
    object$weights <- NULL
    varimp_pval(coef(object), diag(vcov(object)), ...)
  }
}


varimp_pval.multinom <- function(object, ...) {
  varimp_pval(t(coef(object)), diag(vcov(object)), ...)
}


varimp_pval.numeric <- function(object, var, base = exp(1), ...) {
  structure(
    -log(pchisq(object^2 / var, 1, lower.tail = FALSE), base = base),
    metric = paste0(
      "-log", if (base != exp(1)) paste0("(base = ", as_string(base), ")"),
      " Pr(>Chi)"
    )
  )
}

utils::globalVariables(c(".data", "x", "y"))


.onLoad <- function(libname, pkgname) {
  registerDoSEQ()
  invisible()
}


#' Quote Operator
#'
#' Shorthand notation for the \code{\link[base:substitute]{quote}} function.
#' The quote operator simply returns its argument unevaluated and can be applied
#' to any \R expression.
#'
#' @name quote
#'
#' @param expr any syntactically valid \R expression.
#'
#' @details
#' Useful for calling \link[=models]{model functions} with quoted parameter
#' values defined in terms of one or more of the following variables.
#' \describe{
#'   \item{\code{nobs}}{number of observations in data to be \link{fit}.}
#'   \item{\code{nvars}}{number of predictor variables.}
#'   \item{\code{y}}{the response variable.}
#' }
#'
#' @return
#' The quoted (unevaluated) expression.
#'
#' @seealso \code{\link[base:substitute]{quote}}
#'
#' @examples
#' ## Stepwise variable selection with BIC
#' glm_fit <- fit(sale_amount ~ ., ICHomes, GLMStepAICModel(k = .(log(nobs))))
#' varimp(glm_fit)
#'
. <- function(expr) {
  eval(substitute(quote(expr)))
}


as_string <- function(x, ...) {
  UseMethod("as_string")
}


as_string.default <- function(x, ...) {
  x <- format(x, trim = TRUE, justify = "none", drop0trailing = TRUE)
  as_string(structure(as.character(x), names = names(x)), ...)
}


as_string.character <- function(x, sep = ", ", conj = character(), ...) {
  x <- trimws(x)
  if (length(conj)) {
    n <- length(x)
    if (n > 1) x[n] <- paste(conj, x[n])
    if (n == 2) x <- paste(x[1], x[2])
  }
  if (length(sep)) paste(x, collapse = sep) else x
}


attach_objects <- function(
  what, pos = 2L, name = deparse1(substitute(what), backtick = FALSE)
) {
  make_attach <- attach
  make_attach(what, pos, name, warn.conflicts = FALSE)
  do.call(on.exit, list(substitute(detach(name))), envir = parent.frame())
}


class1 <- function(x) {
  class(x)[1]
}


combine_inputs <- function(x) {
  names(x) <- make_names_along(x, map("char", class, x))
  all_data <- NULL
  for (i in seq_along(x)) {
    data <- as.data.frame(x[[i]])
    if (is.null(all_data)) all_data <- data[NULL]
    same_cols <- intersect(names(all_data), names(data))
    if (!identical(all_data[same_cols], data[same_cols])) {
      throw(Error(
        "Cannot combine ", names(x)[i], " with previous data fames due to ",
        "differences in column values or attributes."
      ))
    }
    diff_cols <- setdiff(names(data), same_cols)
    all_data[diff_cols] <- data[diff_cols]
    x[[i]] <- update(x[[i]], data = data[NULL, , drop = FALSE])
  }
  list(data = all_data, candidates = ListOf(x))
}


combine_model_slots <- function(models, types) {
  init <- data.frame(type = types, weights = FALSE)
  any_ordered <- any(map("logi", function(model) {
    "ordered" %in% model@response_types
  }, models))
  info_list <- map(function(model) {
    types <- model@response_types
    res <- data.frame(type = types, weights = model@weights)
    factor_match <- match("factor", types, nomatch = 0)
    if (any_ordered && !("ordered" %in% types) && factor_match) {
      res <- rbind(
        res,
        data.frame(type = "ordered", weights = res$weights[factor_match])
      )
    }
    res
  }, models)
  info <- Reduce(function(x, y) {
    res <- merge(x, y, by = "type")
    res$weights <- if (nrow(res)) res$weights.x | res$weights.y else logical()
    res[c("weights.x", "weights.y")] <- NULL
    res
  }, info_list, init)
  list(
    response_types = info$type,
    weights = switch(length(unique(info$weights)) + 1,
      FALSE, info$weights[1], info@weights
    )
  )
}


complete_subset <- function(...) {
  is_complete <- complete.cases(...)
  map(function(x) subset(x, is_complete), list(...))
}


fget <- function(x, package = character()) {
  throw(check_packages(package))
  if (is.character(x)) {
    x <- paste(c(package, x), collapse = "::")
    err_msg <- paste0("Function `", x, "` not found.")
  } else {
    err_msg <- "Invalid function."
  }
  res <- get0(x, mode = "function")
  if (is.null(res)) throw(Error(err_msg))
  if (length(package)) {
    envir <- environment(res)
    if (!isNamespace(envir) || getNamespaceName(envir) != package) {
      throw(Error("Function is not from the ", package, " package."))
    }
  }
  res
}


get0 <- function(x, mode = "any") {
  if (is.character(x) && length(x) == 1) {
    expr <- str2lang(x)
    if (is.name(expr)) {
      base::get0(x, mode = mode)
    } else if (is.call(expr) && any(expr[[1]] == c("::", ":::"))) {
      base::get0(
        as.character(expr[[3]]), mode = mode, envir = asNamespace(expr[[2]])
      )
    }
  } else if (mode %in% c("any", mode(x))) {
    x
  }
}


get_optim_field <- function(object, name = NULL) {
  object <- object@params@optim
  if (is.null(name)) object else slot(object, name)
}


get_perf_metrics <- function(x, y) {
  generic_name <- "performance"
  classes <- substring(methods(generic_name), nchar(generic_name) + 2)
  class <- match_class(x, classes)
  method <- fget(paste0(generic_name, ".", class))
  metrics <- c(eval(formals(method)$metrics))
  is_defined <- map("logi", function(metric) {
    types <- metricinfo(metric)[[1]]$response_types
    any(
      map("logi", is, list(x), types$observed) &
        map("logi", is, list(y), types$predicted)
    )
  }, metrics)
  metrics[is_defined]
}


has_grid <- function(object) {
  is(object, "TunedInput") || is(object, "TunedModel") ||
    (is(object, "MLModel") && nrow(object@gridinfo))
}


has_varimp <- function(object) {
  !is.null(body(object@varimp))
}


identical_elements <- function(x, transform = identity, ...) {
  target <- transform(x[[1]])
  compare <- function(current) identical(transform(current), target, ...)
  all(map("logi", compare, x[-1]))
}


is_counting <- function(x) {
  isTRUE(attr(x, "type") == "counting")
}


is_empty <- function(x) {
  prod(size(x)) == 0
}


is_one_element <- function(x, class = "ANY") {
  is.vector(x) && length(x) == 1 && is(x[[1]], class)
}


is_optim_method <- function(x, ...) {
  UseMethod("is_optim_method")
}


is_optim_method.default <- function(x, ...) {
  FALSE
}


is_optim_method.MLInput <- function(x, ...) {
  is_optim_method(x@params, ...)
}


is_optim_method.MLModel <- function(x, ...) {
  is_optim_method(x@params, ...)
}


is_optim_method.MLOptimization <- function(x, type = "ANY", ...) {
  is(x, type)
}


is_optim_method.ModelSpecification <- function(x, ...) {
  is_optim_method(x@params, ...)
}


is_optim_method.NullOptimization <- function(x, ...) {
  is_optim_method.default(x, ...)
}


is_optim_method.TrainingParams <- function(x, ...) {
  is_optim_method(x@optim, ...)
}


is_response <- function(x, types) {
  map("logi", function(type) {
    if (type == "binary") {
      is(x, "factor") && nlevels(x) == 2
    } else {
      is(x, type)
    }
  }, types)
}


is_trained <- function(x, ...) {
  UseMethod("is_trained")
}


is_trained.MLModel <- function(x, ...) {
  length(x@steps) > 0
}


is_trained.ModelRecipe <- function(x, ...) {
  recipes::fully_trained(x)
}


is_trained.step <- function(x, ...) {
  recipes::is_trained(x)
}


make_id <- function(prefix = character(), n = 4, sep = ".") {
  suffix <- sample(c(letters, LETTERS, 0:9), n, replace = TRUE)
  paste(c(prefix, paste(suffix, collapse = "")), collapse = sep)
}


make_names_along <- function(x, default = "..", unique = TRUE, sep = ".") {
  if (length(default) == 1) default <- rep_len(default, length(x))
  old_names <- names(x)
  names(x) <- default
  if (!is.null(old_names)) {
    keep <- nzchar(old_names) & !is.na(old_names)
    names(x)[keep] <- old_names[keep]
  }
  if (unique) make_unique(names(x), sep = sep) else names(x)
}


make_names_len <- function(n, prefix) {
  paste0(rep_len(prefix, n), seq_len(n))
}


make_unique <- function(names, sep = ".") {
  sort_order <- order(names)
  counts <- table(names)
  unames <- map(function(count, name) {
    paste0(name, if (count > 1) paste0(sep, seq_len(count)))
  }, counts, names(counts))
  names[sort_order] <- unlist(unames)
  names
}


map <- function(.object, ...) {
  UseMethod("map")
}


map.character <- function(.object, .fun, ...) {
  type <- match.arg(
    .object, c(
      "character", "complex", "double", "integer", "list", "logical", "numeric",
      "raw"
    )
  )
  res <- simplify(map(.fun, ...))
  storage.mode(res) <- type
  res
}


map.function <- function(.object, ...) {
  nonempty <- all(lengths(list(...)))
  if (nonempty) mapply(FUN = .object, ..., SIMPLIFY = FALSE) else list()
}


match_class <- function(object, choices) {
  f <- function(x, ...) UseMethod("f")
  f.default <- function(x, ...) NA_character_
  for (choice in choices) {
    assign(paste0("f.", choice), eval(substitute(function(x, ...) choice)))
  }
  f(object)
}


match_indices <- function(indices, choices) {
  lookup <- structure(seq_along(choices), names = choices)
  indices <- na.omit(names(lookup)[lookup[indices]])
  if (is_empty(indices)) {
    indices <- names(lookup)[1]
    throw(LocalWarning(
      "Specified indices not found; using ", indices, " instead."
    ))
  }
  indices
}


max.progress_index <- function(x, ...) {
  attr(x, "max")
}


missing_names <- function(x, data) {
  x[!(x %in% names(data))]
}


na.add <- function(object, at) {
  if (any(at)) {
    df <- tibble(object = object)
    df <- df[pmax(cumsum(!at), 1), ]
    df[at, ] <- NA
    object <- df$object
  }
  object
}


ndim <- function(x) {
  length(size(x))
}


new_params <- function(envir, ...) {
  args <- c(as.list(envir), ...)
  missing <- map("logi", function(x) is.name(x) && !nzchar(x), args)
  if (any(missing)) {
    throw(Error(note_items(
      "Missing values for required argument{?s}: ", names(args)[missing], "."
    )))
  }
  args[lengths(args) > 0]
}


new_progress_bar <- function(total, object, index = 0) {
  backend <- getDoParName()
  progress <- function(...) NULL
  snow_opts <- list()

  if (backend == "doSEQ") index <- as.numeric(index)
  width <- max(round(0.25 * console_width()), 10)
  input <- substr(class(as.MLInput(object)), 1, width)
  model <- substr(as.MLModel(object)@name, 1, width)
  format <- paste(input, "|", model)
  if (index > 0) format <- paste0(index, ": ", format)
  if (backend %in% c("doSEQ", "doSNOW")) {
    format <- paste(format, "[:bar] :percent | :eta")
  }
  pb <- progress_bar$new(
    format = format, total = total, clear = TRUE, show_after = 0
  )
  if (is(index, "progress_index") && index == 1) {
    pb$message(paste0(
      names(index), if (is.finite(max(index))) paste0("(", max(index), ")")
    ))
  }

  if (total) {
    pb$tick(0)
    body(progress) <- quote(pb$tick())
    if (backend == "doSNOW") snow_opts$progress <- progress
  }

  list(pb = pb, progress = progress, snow_opts = snow_opts)
}


new_progress_index <- function(init = 0, max = Inf, name = character()) {
  structure(init, max = max, names = name, class = "progress_index")
}


note_items <- function(
  begin, values, end = character(), add_names = FALSE, add_size = FALSE,
  sep = ", ", conj = character()
) {
  if (add_names && length(names(values))) {
    values <- paste(names(values), values, sep = " = ")
  }
  size <- length(values)
  if (add_size && size) {
    values <- c(
      paste0("[1", if (size > 1) paste0(":", size), "] ", values[1]), values[-1]
    )
  }
  end <- paste0(end, "{qty(size)}")
  pluralize(begin, as_string(values, sep = sep, conj = conj), end)
}


nvars <- function(x, model) {
  stopifnot(is(x, "ModelFrame"))
  model <- as.MLModel(model)
  res <- switch(model@predictor_encoding,
    "model.frame" = {
      x_terms <- attributes(terms(x))
      nrow(x_terms$factors) - x_terms$response - length(x_terms$offset)
    },
    "model.matrix" = ncol(model.matrix(x[1, , drop = FALSE], intercept = FALSE))
  )
  if (is.null(res)) NA else res
}


permute_int <- function(n, size = n) {
  inds <- sample.int(n, size)
  half_size <- size / 2
  inds1 <- head(inds, half_size)
  inds2 <- tail(inds, -half_size)
  res <- data.frame(i = c(inds1, inds2), j = c(inds2, inds1))
  res[order(res$i), ]
}


push <- function(x, object, ...) {
  UseMethod("push")
}


push.TrainingStep <- function(x, object, ...) {
  stopifnot(is(object, "MLModelFit"))
  steps <- ListOf(c(x, attr(object, ".MachineShop")$model@steps))
  names(steps) <- make_names_len(length(steps), class(x))
  attr(object, ".MachineShop")$model@steps <- steps
  object
}


rand_int <- function(n = 1) {
  sample.int(.Machine$integer.max, n)
}


required_packages <- function(object) {
  input <- as.MLInput(object)
  require <- if (is(input, "ModelRecipe")) {
    c("survival", "recipes")
  } else if (is(response(input), "Surv")) {
    "survival"
  }
  union(settings("require"), require)
}


round_int <- function(...) {
  as.integer(round(...))
}


sample_replace <- function(x, inds) {
  if (!is.logical(inds)) {
    old_inds <- inds
    inds <- structure(logical(length(x)), names = names(x))
    inds[old_inds] <- TRUE
  }
  x[inds] <- sample(x[!inds], sum(inds), replace = TRUE)
  x
}


seq_boot <- function(src, dest) {
  indices <- seq_len(nrow(src))
  pad_size <- nrow(dest) - nrow(src)
  if (pad_size >= 0) {
    c(indices, sample(indices, pad_size, replace = TRUE))
  } else {
    sample(indices, nrow(dest))
  }
}


seq_inner <- function(from, to, length) {
  x <- seq(from, to, length = length + 2)
  x[-c(1, length + 2)]
}


seq_int <- function(...) {
  as.integer(seq(...))
}


seq_nvars <- function(x, model, length) {
  nvars <- nvars(x, model)
  length <- min(length, nvars)
  vals <- if (length > 1) {
    if (nvars < 500) {
      seq(2, nvars, length = length)
    } else {
      2^seq(1, log2(nvars), length = length)
    }
  } else {
    numeric()
  }
  round_int(vals)
}


seq_range <- function(from, by, bounds, length) {
  if (length > 0) {
    to <- min(from + by * (length - 1), bounds[2])
    x <- seq(from, to, length = length)
    x[x >= bounds[1]]
  } else {
    seq(from, length = length)
  }
}


set_model_names <- function(x) {
  name <- "Model"
  level_names <- list()
  for (i in seq_along(x)) {
    if (is.null(x[[i]][[name]])) x[[i]][[name]] <- rep(name, nrow(x[[i]]))
    x[[i]][[name]] <- as.factor(x[[i]][[name]])
    arg_name <- names(x)[i]
    level_names[[i]] <- if (!is.null(arg_name) && nzchar(arg_name)) {
      rep(arg_name, nlevels(x[[i]][[name]]))
    } else {
      levels(x[[i]][[name]])
    }
  }
  level_names <- level_names %>% unlist %>% make_unique %>% relist(level_names)

  for (i in seq_along(x)) levels(x[[i]][[name]]) <- level_names[[i]]

  x
}


simplify <- function(x) {
  if (length(x)) simplify2array(x, higher = TRUE) else x
}


size <- function(x, dim = integer()) {
  res <- dim(x)
  if (is.null(res)) res <- length(x)
  if (length(dim)) res[dim] else res
}


stepAIC_args <- function(formula, direction, scope) {
  if (is.null(scope$lower)) scope$lower <- ~ 1
  if (is.null(scope$upper)) scope$upper <- formula[-2]
  formula[-2] <- if (direction == "backward") scope$upper else scope$lower
  list(formula = formula, scope = scope)
}


subset_selected <- function(object, slot_name, id = character()) {
  if (length(id)) {
    keep <- map("char", slot, slot(object, slot_name), "id") %in% id
    slot(object, slot_name) <- slot(object, slot_name)[keep]
  }
  object
}


subset_names <- function(x, select = NULL) {
  indices <- seq_along(x)
  names(indices) <- x
  select <- eval(substitute(select), as.list(indices), parent.frame())
  if (is.character(select)) {
    x <- intersect(x, select)
  } else if (is.numeric(select)) {
    x <- x[select]
  } else if (!is.null(select)) {
    throw(Warning("Invalid `select` value of class ", class1(select), "."))
  }
  x
}


switch_class <- function(EXPR, ...) {
  blocks <- eval(substitute(alist(...)))
  eval.parent(blocks[[match_class(EXPR, names(blocks))]])
}


unnest <- function(data) {
  stopifnot(is(data, "data.frame"))
  df <- data.frame(row.names = seq_len(nrow(data)))
  for (name in names(data)) {
    x <- data[[name]]
    if (ndim(x) > 1) {
      x <- if (is.data.frame(x)) unnest(x) else as.data.frame(as(x, "matrix"))
      name <- paste0(name, "$", names(x))
    }
    df[name] <- x
  }
  df
}


vector_to_function <- function(x, type) {
  err_msg <- paste0("Value of '", deparse1(substitute(x)),
                    "' must be a function, function name, or vector of these.")
  if (is(x, "MLMetric")) x <- list(x)
  if (is(x, "vector")) {
    x <- as.list(x)
    types <- make_names_len(length(x), type)
    x_names <- character()
    for (i in seq_along(x)) {
      if (is(x[[i]], "character")) {
        x_name <- gsub("^.*[:]{2}(?!.*[:]{2})", "", x[[i]], perl = TRUE)
        x[[i]] <- fget(x[[i]])
      } else if (is(x[[i]], "MLMetric")) {
        x_name <- x[[i]]@name
      } else if (is(x[[i]], "function")) {
        x_name <- types[i]
      } else {
        throw(Error(err_msg))
      }
      name <- names(x)[i]
      x_names[i] <- if (is.null(name) || !nzchar(name)) x_name else name
    }
    names(x) <- make_unique(x_names)
    function(...) unlist(map(function(fun) fun(...), x))
  } else if (is(x, "function")) {
    x
  } else {
    throw(Error(err_msg))
  }
}

#' K-Medoids Clustering Variable Selection
#'
#' Creates a \emph{specification} of a recipe step that will partition numeric
#' variables according to k-medoids clustering and select the cluster medoids.
#'
#' @inheritParams step_sbf
#' @param k number of k-medoids clusterings of the variables.  The value of
#'   \code{k} is constrained to be between 1 and one less than the number of
#'   original variables.
#' @param center,scale logicals indicating whether to mean center and median
#'   absolute deviation scale the original variables prior to cluster
#'   partitioning, or functions or names of functions for the centering and
#'   scaling; not applied to selected variables.
#' @param method character string specifying one of the clustering methods
#'   provided by the \pkg{cluster} package.  The \code{clara} (clustering
#'   large applications) method is an extension of \code{pam} (partitioning
#'   around medoids) designed to handle large datasets.
#' @param metric character string specifying the distance metric for calculating
#'   dissimilarities between observations as \code{"euclidean"},
#'   \code{"manhattan"}, or \code{"jaccard"} (\code{clara} only).
#' @param optimize logical indicator or 0:5 integer level specifying
#'   optimization for the \code{\link[cluster]{pam}} clustering method.
#' @param num_samp number of sub-datasets to sample for the
#'   \code{\link[cluster]{clara}} clustering method.
#' @param samp_size number of cases to include in each sub-dataset.
#' @param x \code{step_kmedoids} object.
#'
#' @return Function \code{step_kmedoids} creates a new step whose class is of
#' the same name and inherits from \code{\link{step_sbf}}, adds it to the
#' sequence of existing steps (if any) in the recipe, and returns the updated
#' recipe.  For the \code{tidy} method, a tibble with columns \code{terms}
#' (selectors or variables selected), \code{cluster} assignments,
#' \code{selected} (logical indicator of selected cluster medoids),
#' \code{silhouette} (silhouette values), and \code{name} of the selected
#' variable names.
#'
#' @details
#' K-medoids clustering partitions variables into k groups such that the
#' dissimilarity between the variables and their assigned cluster medoids is
#' minimized.  Cluster medoids are then returned as a set of k variables.
#'
#' @references
#' Kaufman, L., & Rousseeuw, P. J. (1990). \emph{Finding groups in data: An
#' introduction to cluster analysis}. Wiley.
#'
#' Reynolds, A., Richards, G., de la Iglesia, B., & Rayward-Smith, V. (1992).
#' Clustering rules: A comparison of partitioning and hierarchical clustering
#' algorithms. \emph{Journal of Mathematical Modelling and Algorithms},
#' \emph{5}, 475-504.
#'
#' @seealso \code{\link[cluster]{pam}}, \code{\link[cluster]{clara}},
#' \code{\link[recipes]{recipe}}, \code{\link[recipes]{prep}},
#' \code{\link[recipes]{bake}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package cluster to run
#'
#' library(recipes)
#'
#' rec <- recipe(rating ~ ., data = attitude)
#' kmedoids_rec <- rec %>%
#'   step_kmedoids(all_predictors(), k = 3)
#' kmedoids_prep <- prep(kmedoids_rec, training = attitude)
#' kmedoids_data <- bake(kmedoids_prep, attitude)
#'
#' pairs(kmedoids_data, lower.panel = NULL)
#'
#' tidy(kmedoids_rec, number = 1)
#' tidy(kmedoids_prep, number = 1)
#' }
#'
step_kmedoids <- function(
  recipe, ..., k = 5, center = TRUE, scale = TRUE, method = c("pam", "clara"),
  metric = "euclidean", optimize = FALSE, num_samp = 50, samp_size = 40 + 2 * k,
  replace = TRUE, prefix = "KMedoids", role = "predictor", skip = FALSE,
  id = recipes::rand_id("kmedoids")
) {

  recipes::add_step(recipe, new_step_kmedoids(
    terms = recipes::ellipse_check(...),
    k = k,
    center = center,
    scale = scale,
    method = match.arg(method),
    metric = metric,
    optimize = optimize,
    num_samp = num_samp,
    samp_size = samp_size,
    replace = replace,
    prefix = prefix,
    role = role,
    skip = skip,
    id = id
  ))

}


new_step_kmedoids <- function(
  ..., k, center, scale, method, metric, optimize, num_samp, samp_size
) {

  throw(check_packages("cluster"))

  filter <- function(x, y, step) {

    throw(check_packages("cluster"))

    if (ncol(x) < 2) {
      throw(LocalError("Step_kmedoids requires 2 or more variables."))
    }
    recipes::check_type(x)

    stats <- map(function(stat) {
      if (is.function(stat)) apply(x, 2, stat) else stat
    }, step[c("center", "scale")])
    x <- t(base::scale(x, center = stats$center, scale = stats$scale))
    k <- min(max(step$k, 1), nrow(x) - 1)
    switch(step$method,
      "pam" = {
        res <- cluster::pam(
          x, k, metric = step$metric, pamonce = step$optimize,
          keep.diss = FALSE, keep.data = FALSE
        )
      },
      "clara" = {
        samp_size <- min(step$samp_size, nrow(x))
        res <- cluster::clara(
          x, k, metric = step$metric, samples = step$num_samp,
          sampsize = samp_size, medoids.x = FALSE, rngR = TRUE
        )
        names(res)[names(res) == "i.med"] <- "id.med"
      }
    )

    tibble(
      cluster = res$clustering,
      selected = seq_along(res$clustering) %in% res$id.med,
      silhoutte = if (k == 1) NA_real_ else
        res$silinfo$widths[names(res$clustering), "sil_width"]
    )

  }

  options <- list(
    k = k,
    center = if (!is.logical(center)) fget(center) else
      if (center) base::mean else FALSE,
    scale = if (!is.logical(scale)) fget(scale) else
      if (scale) stats::mad else FALSE,
    method = method
  )
  switch(options$method,
    "pam" = {
      options$metric <- match.arg(metric, c("euclidean", "manhattan"))
      options$optimize <- optimize
    },
    "clara" = {
      options$metric <- match.arg(
        metric, c("euclidean", "manhattan", "jaccard")
      )
      options$num_samp <- num_samp
      options$samp_size <- samp_size
    },
    {
      method <- Error("Value must be \"pam\" or \"clara\".")
      throw(check_assignment(method))
    }
  )

  object <- new_step_sbf(
    ..., filter = filter, multivariate = TRUE, options = options
  )
  object$res <- tibble(
    terms = recipes::sel2char(object$terms),
    cluster = NA_integer_,
    selected = NA,
    silhouette = NA_real_,
    name = NA_character_
  )

  structure(object, class = c("step_kmedoids", "step"))

}


prep.step_kmedoids <- function(x, ...) prep.step_sbf(x, ...)
bake.step_kmedoids <- function(object, ...) bake.step_sbf(object, ...)
print.step_kmedoids <- function(x, ...) print.step_sbf(x, ...)


#' @rdname step_kmedoids
#'
tidy.step_kmedoids <- function(x, ...) tidy.step_sbf(x, ...)


#' @rdname step_kmedoids
#'
tunable.step_kmedoids <- function(x, ...) {
  tibble(
    name = "k",
    call_info = list(list(pkg = "dials", fun = "num_comp", range = c(1, 10))),
    source = "MachineShop",
    component = "step_kmedoids",
    component_id = x$id
  )
}

#' Stacked Regression Model
#'
#' Fit a stacked regression model from multiple base learners.
#'
#' @param ... \link[=models]{model} functions, function names, objects; other
#'   objects that can be \link[=as.MLModel]{coerced} to models; or vector of
#'   these to serve as base learners.
#' @param control \link[=controls]{control} function, function name, or object
#'   defining the resampling method to be employed for the estimation of base
#'   learner weights.
#' @param weights optional fixed base learner weights.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric}, \code{ordered},
#'     \code{Surv}}
#' }
#'
#' @return \code{StackedModel} class object that inherits from \code{MLModel}.
#'
#' @references
#' Breiman, L. (1996). Stacked regression. \emph{Machine Learning}, \emph{24},
#' 49-64.
#'
#' @seealso \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested packages gbm and glmnet to run
#'
#' model <- StackedModel(GBMModel, SVMRadialModel, GLMNetModel(lambda = 0.01))
#' model_fit <- fit(sale_amount ~ ., data = ICHomes, model = model)
#' predict(model_fit, newdata = ICHomes)
#' }
#'
StackedModel <- function(
  ..., control = MachineShop::settings("control"), weights = numeric()
) {

  models <- map(as.MLModel, unlist(list(...)))
  names(models) <- make_names_len(length(models), "BaseLearner")

  options <- list()
  if (length(weights)) {
    stopifnot(length(weights) == length(models))
    options$weights <- weights
  }

  slots <- combine_model_slots(models, settings("response_types"))
  new("StackedModel",
    MLModel(
      name = "StackedModel",
      label = "Stacked Regression",
      response_types = slots$response_types,
      weights = slots$weights,
      params = TrainingParams(
        optim = NullOptimization(),
        control = control,
        options = options
      )
    ),
    candidates = ListOf(models)
  )

}

MLModelFunction(StackedModel) <- NULL


.fit.StackedModel <- function(object, input, ...) {
  base_learners <- object@candidates
  control <-  object@params@control
  weights <- object@params@options$weights

  if (is.null(weights)) {
    num_learners <- length(base_learners)
    stack <- list()
    complete_cases <- TRUE
    ind <- new_progress_index(max = num_learners, name = object@name)
    while (ind < max(ind)) {
      ind <- ind + 1
      stack[[ind]] <- resample(
        input, model = base_learners[[ind]], control = control,
        progress_index = ind
      )
      complete_cases <- complete_cases & complete.cases(stack[[ind]])
    }
    stack <- map(function(res) res[complete_cases, ], stack)

    weights <- Rsolnp::solnp(
      rep(1 / num_learners, num_learners),
      function(weights) mean_stack_list(stack, weights),
      eqfun = function(x) sum(x), eqB = 1, LB = numeric(num_learners),
      control = list(trace = FALSE)
    )$pars
  }

  list(
    base_fits = map(
      function(learner) fit(input, model = learner), base_learners
    ),
    weights = weights,
    times = control@predict$times
  ) %>% MLModelFit("StackedModelFit", input = input, model = object)
}


.predict.StackedModel <- function(object, model_fit, newdata, ...) {
  pred <- 0
  for (i in seq_along(model_fit$base_fits)) {
    base_pred <- predict(
      model_fit$base_fits[[i]], newdata = newdata, times = model_fit$times,
      type = "raw"
    )
    pred <- pred + model_fit$weights[i] * base_pred
  }
  pred
}


mean_stack_list <- function(x, weights) {
  pred <- 0
  for (i in seq_along(x)) pred <- pred + weights[i] * x[[i]]$Predicted
  res <- x[[1]]
  res$Predicted <- pred
  mean(stack_loss(res$Observed, res$Predicted, res), na.rm = TRUE)
}


setGeneric("stack_loss",
  function(observed, predicted, x, ...) standardGeneric("stack_loss")
)


setMethod("stack_loss", c("ANY", "ANY", "Resample"),
  function(observed, predicted, x, ...) mse(x)
)


setMethod("stack_loss", c("factor", "ANY", "Resample"),
  function(observed, predicted, x, ...) brier(x)
)


setMethod("stack_loss", c("Surv", "numeric", "Resample"),
  function(observed, predicted, x, ...) -cindex(x)
)


setMethod("stack_loss", c("Surv", "SurvProbs", "Resample"),
  function(observed, predicted, x, ...) brier(x)[, 1]
)

TrainingParams <- function(
  ..., optim = NULL, control = NULL, cutoff = NULL, stat = NULL
) {
  control <- as.MLControl(control)
  if (is(control, "NullControl")) {
    optim <- NullOptimization()
  } else if (is.null(optim)) {
    optim <- GridSearch()
  }

  if (length(cutoff)) {
    cutoff <- check_numeric(cutoff, bounds = c(0, 1), include = FALSE, size = 1)
    throw(check_assignment(cutoff))
  } else {
    cutoff <- 0.5
  }

  if (length(stat)) {
    stat <- check_stat(stat, convert = TRUE)
    throw(check_assignment(stat))
  } else {
    stat <- function(x) NA_real_
  }

  new("TrainingParams",
    ..., control = control, optim = optim, cutoff = cutoff, stat = stat
  )
}


TrainingStep <- function(
  object, method = character(), names = character(), items = tibble(),
  params = tibble(), metrics = tibble(), selected = 0,
  performance = Performance()
) {
  nlog <- length(names)
  new("TrainingStep",
    id = object@id,
    name = class(object),
    method = method,
    log = as_tibble(c(
      list(name = names),
      if (is_tibble(items) && nrow(items) == nlog) items,
      list(
        selected = seq_len(nlog) == selected,
        params = as_tibble(params, .rows = nlog),
        metrics = as_tibble(metrics, .rows = nlog)
      )
    )),
    performance = performance
  )
}


#' Training Parameters Monitoring Control
#'
#' Set parameters that control the monitoring of resample estimation of model
#' performance and of tuning parameter optimization.
#'
#' @rdname set_monitor-methods
#'
#' @param object resampling \link[=controls]{control},
#'   tuning parameter \link[=set_optim]{optimization}, or model
#'   \link[=ModelSpecification]{specification} object.
#' @param progress logical indicating whether to display iterative progress
#'   during resampling or optimization.  In the case of resampling, a progress
#'   bar will be displayed if a computing cluster is not registered or is
#'   registered with the \pkg{doSNOW} package.
#' @param verbose numeric or logical value specifying the level of progress
#'   detail to print, with 0 (\code{FALSE}) indicating none and 1 (\code{TRUE})
#'   or higher indicating increasing amounts of detail.
#' @param which character string specifying the monitoring parameters to set as
#'   \code{"all"}, \code{"control"}, or optimization (\code{"optim"}).
#' @param ... arguments passed from the \code{ModelSpecification} method to the
#'   others.
#'
#' @return Argument \code{object} updated with the supplied parameters.
#'
#' @seealso \code{\link{resample}}, \code{\link{set_optim}},
#' \code{\link{set_predict}}, \code{\link{set_strata}}
#'
#' @examples
#' CVControl() %>% set_monitor(verbose = TRUE)
#'
set_monitor <- function(object, ...) {
  UseMethod("set_monitor")
}


set_monitor.list <- function(object, progress, verbose, ...) {
  progress <- check_logical(progress, size = 1)
  object$progress <- throw(check_assignment(progress))

  verbose <- if (is.logical(verbose)) {
    check_logical(verbose, size = 1)
  } else {
    check_integer(verbose, bounds = c(0, Inf), size = 1)
  }
  object$verbose <- throw(check_assignment(verbose))

  object
}


#' @rdname set_monitor-methods
#'
set_monitor.MLControl <- function(
  object, progress = TRUE, verbose = FALSE, ...
) {
  verbose <- check_logical(verbose)
  object@monitor <- set_monitor(
    object@monitor, progress = progress, verbose = verbose
  )
  object
}


#' @rdname set_monitor-methods
#'
set_monitor.MLOptimization <- function(
  object, progress = FALSE, verbose = FALSE, ...
) {
  object@monitor <- set_monitor(
    object@monitor, progress = progress, verbose = verbose
  )
  object
}


set_monitor.SplitControl <- function(object, ...) {
  set_monitor.MLControl(object, ...)
}


set_monitor.TrainingParams <- function(object, which = "all", ...) {
  if (which == "all") which <- c("control", "optim")
  for (name in which) {
    slot(object, name) <- set_monitor(slot(object, name), ...)
  }
  object
}


#' @rdname set_monitor-methods
#'
set_monitor.ModelSpecification <- function(
  object, which = c("all", "control", "optim"), ...
) {
  object@params <- set_monitor(object@params, which = match.arg(which), ...)
  object
}


#' Tuning Parameter Optimization
#'
#' Set the optimization method and control parameters for tuning of model
#' parameters.
#'
#' @name set_optim
#' @rdname set_optim-methods
#'
#' @param object \link[=inputs]{input} or \link[=models]{model} object.
#' @param num_init number of grid points to sample for the initialization of
#'   Bayesian optimization.
#' @param times maximum number of times to repeat the optimization step.
#'   Multiple sets of model parameters are evaluated automatically at each step
#'   of the BFGS algorithm to compute a finite-difference approximation to the
#'   gradient.
#' @param each number of times to sample and evaluate model parameters at each
#'   optimization step.  This is the swarm size in particle swarm optimization,
#'   which defaults to \code{floor(10 + 2 * sqrt(length(bounds)))}.
#' @param acquisition character string specifying the acquisition function as
#'   \code{"ucb"} (upper confidence bound), \code{"ei"} (expected improvement),
#'   \code{"eips"} (expected improvement per second), or \code{"poi"}
#'   (probability of improvement).
#' @param kappa,conf upper confidence bound (\code{"ucb"}) quantile or its
#'   probability to balance exploitation against exploration.  Argument
#'   \code{kappa} takes precedence if both are given and multiplies the
#'   predictive standard deviation added to the predictive mean in the
#'   acquisition function.  Larger values encourage exploration of the model
#'   parameter space.
#' @param epsilon improvement methods (\code{"ei"}, \code{"eips"}, and
#'   \code{"poi"}) parameter to balance exploitation against exploration.
#'   Values should be between -0.1 and 0.1 with larger ones encouraging
#'   exploration.
#' @param random number of points to sample for a random grid search, or
#'   \code{FALSE} for an exhaustive grid search.  Used when a grid search is
#'   specified or as the fallback method for non-numeric model parameters
#'   present during other optimization methods.
#' @param packages R package or packages to use for the optimization method, or
#'   an empty vector if none are needed.  The first package in
#'   \code{set_optim_bayes} is used unless otherwise specified by the user.
#' @param control list of control parameters passed to
#'   \code{bayesOpt} by \code{set_optim_bayes} with package
#'   \code{"ParBayesianOptimization"}, to
#'   \code{\link[rBayesianOptimization]{BayesianOptimization}} by
#'   \code{set_optim_bayes} with package \code{"rBayesianOptimization"}, to
#'   \code{\link[stats]{optim}} by \code{set_optim_bfgs} and
#'   \code{set_optim_sann}, and to \code{\link[pso]{psoptim}} by
#'   \code{set_optim_pso}.
#' @param progress logical indicating whether to display iterative progress
#'   during optimization.
#' @param verbose numeric or logical value specifying the level of progress
#'   detail to print, with 0 (\code{FALSE}) indicating none and 1 (\code{TRUE})
#'   or higher indicating increasing amounts of detail.
#' @param fun user-defined optimization function to which the arguments below
#'   are passed in order.  An ellipsis can be included in the function
#'   definition when using only a subset of the arguments and ignoring others.
#'   A tibble returned by the function with the same number of rows as model
#'   evaluations will be included in a \code{TrainingStep} summary of
#'   optimization results; other types of return values will be ignored.
#'   \describe{
#'     \item{optim}{function that takes a numeric vector or list of named
#'       model parameters as the first argument, optionally accepts the maximum
#'       number of iterations as argument \code{max_iter}, and returns a scalar
#'       measure of performance to be maximized.  Parameter names are available
#'       from the \code{grid} and \code{bounds} arguments described below.  If
#'       the function cannot be evaluated at a given set of parameter values,
#'       then \code{-Inf} is returned.}
#'     \item{grid}{data frame containing a tuning grid of all model parameters.}
#'     \item{bounds}{named list of lower and upper bounds for each finite
#'       numeric model parameter in \code{grid}.  The types (integer or double)
#'       of the original parameter values are preserved in the bounds.}
#'     \item{params}{list of optimization parameters as supplied to
#'       \code{set_optim_method}.}
#'     \item{monitor}{list of the \code{progress} and \code{verbose} values.}
#'   }
#' @param label character descriptor for the optimization method.
#' @param params list of user-specified model parameters to be passed to
#'   \code{fun}.
#' @param ... arguments passed to the \code{TrainingParams} method of
#'   \code{set_optim_grid} from its other methods.
#'
#' @details
#' The optimization functions implement the following methods.
#' \describe{
#'   \item{\code{set_optim_bayes}}{Bayesian optimization with a Gaussian process
#'     model (Snoek et al. 2012).}
#'   \item{\code{set_optim_bfgs}}{limited-memory modification of quasi-Newton
#'     BFGS optimization (Byrd et al. 1995).}
#'   \item{\code{set_optim_grid}}{exhaustive or random grid search.}
#'   \item{\code{set_optim_pso}}{particle swarm optimization (Bratton and
#'     Kennedy 2007, Zambrano-Bigiarini et al. 2013).}
#'   \item{\code{set_optim_sann}}{simulated annealing (Belisle 1992).  This
#'     method depends critically on the control parameter settings.  It is
#'     not a general-purpose method but can be very useful in getting to good
#'     parameter values on a very rough optimization surface.}
#'   \item{\code{set_optim_method}}{user-defined optimization function.}
#' }
#' The package-defined optimization functions evaluate and return values of the
#' tuning parameters that are of same type (e.g. integer, double, character) as
#' given in the \code{object} grid.  Sequential optimization of numeric tuning
#' parameters is performed over a hypercube defined by their minimum and maximum
#' grid values.  Non-numeric parameters are optimized with grid searches.
#'
#' @return Argument \code{object} updated with the specified optimization method
#' and control parameters.
#'
#' @references
#' Belisle, C. J. P. (1992). Convergence theorems for a class of simulated
#' annealing algorithms on Rd. \emph{Journal of Applied Probability},
#' \emph{29}, 885–895.
#'
#' Bratton, D. & Kennedy, J. (2007), Defining a standard for particle swarm
#' optimization. In \emph{IEEE Swarm Intelligence Symposium, 2007} (pp.
#' 120-127).
#'
#' Byrd, R. H., Lu, P., Nocedal, J., & Zhu, C. (1995). A limited memory
#' algorithm for bound constrained optimization. \emph{SIAM Journal on
#' Scientific Computing}, \emph{16}, 1190–1208.
#'
#' Snoek, J., Larochelle, H., & Adams, R.P. (2012). Practical Bayesian
#' Optimization of Machine Learning Algorithms. arXiv:1206.2944 [stat.ML].
#'
#' Zambrano-Bigiarini, M., Clerc, M., & Rojas, R. (2013). Standard particle
#' swarm optimisation 2011 at CEC-2013: A baseline for future PSO improvements.
#' In \emph{IEEE Congress on Evolutionary Computation, 2013} (pp. 2337-2344).
#'
#' @seealso \code{\link[rBayesianOptimization]{BayesianOptimization}},
#' \code{bayesOpt}, \code{\link[stats]{optim}}, \code{\link[pso]{psoptim}},
#' \code{\link{set_monitor}}, \code{\link{set_predict}},
#' \code{\link{set_strata}}
#'
#' @examples
#' ## Install ParBayesianOptimization package as needed
#' ## pak::repo_add(smith = "https://brian-j-smith.github.io/drat")
#' ## pak::pak("ParBayesianOptimization")
#' 
#' ModelSpecification(
#'   sale_amount ~ ., data = ICHomes,
#'   model = TunedModel(GBMModel)
#' ) %>% set_optim_bayes
#'
NULL


#' @rdname set_optim-methods
#'
set_optim_bayes <- function(object, ...) {
  UseMethod("set_optim_bayes")
}


#' @rdname set_optim-methods
#'
set_optim_bayes.ModelSpecification <- function(
  object, num_init = 5, times = 10, each = 1,
  acquisition = c("ucb", "ei", "eips", "poi"), kappa = stats::qnorm(conf),
  conf = 0.995, epsilon = 0, control = list(),
  packages = c("ParBayesianOptimization", "rBayesianOptimization"),
  random = FALSE, progress = verbose, verbose = 0, ...
) {
  if (!is_optim_method(object)) return(object)

  params <- list()

  packages <- match.arg(packages)
  switch(packages,
    "ParBayesianOptimization" = {
      each_upper <- Inf
      verbose_progress <- FALSE
      fun <- function(optim, bounds, params, monitor, ...) {
        throw(check_optim_bounds(bounds), call = call("set_optim_bayes"))
        num_init <- max(params$num_init, length(bounds) + 1)
        max_iter <- num_init + params$times * params$each
        args <- c(list(
          function(...) {
            score <- optim(list(...), max_iter = max_iter)
            list(Score = max(score, -.Machine$double.xmax))
          },
          bounds = bounds,
          initPoints = num_init,
          iters.n = max_iter - num_init,
          iters.k = params$each,
          acq = params$acquisition,
          kappa = params$kappa,
          eps = params$epsilon,
          verbose = monitor$verbose),
          params$control
        )
        res <- do.call(ParBayesianOptimization::bayesOpt, args)$scoreSummary
        tibble(epoch = as.integer(res$Epoch))
      }
    },
    "rBayesianOptimization" = {
      each_upper <- 1
      verbose_progress <- TRUE && !missing(progress)
      fun <- function(optim, bounds, params, monitor, ...) {
        throw(check_optim_bounds(bounds), call = call("set_optim_bayes"))
        num_init <- max(params$num_init, 2)
        max_iter <- num_init + params$times
        args <- c(list(
          function(...) {
            score <- optim(list(...), max_iter = max_iter)
            list(Score = max(score, -.Machine$double.xmax), Pred = NA)
          },
          bounds = bounds,
          init_points = num_init,
          n_iter = max_iter - num_init,
          acq = switch(params$acquisition,
            "eips" = "ei",
            params$acquisition
          ),
          kappa = params$kappa,
          eps = params$epsilon,
          verbose = as.logical(monitor$verbose)),
          params$control
        )
        do.call(rBayesianOptimization::BayesianOptimization, args)
        tibble(epoch = c(integer(params$num_init), seq_len(params$times)))
      }
    }
  )

  num_init <- check_integer(num_init, bounds = c(1, Inf), size = 1)
  params$num_init <- throw(check_assignment(num_init))

  times <- check_integer(times, bounds = c(1, Inf), size = 1)
  params$times <- throw(check_assignment(times))

  each <- check_integer(each, bounds = c(1, each_upper), size = 1)
  params$each <- throw(check_assignment(each))

  params$acquisition <- match.arg(acquisition)

  if (params$acquisition == "ucb") {
    kappa <- check_numeric(kappa, bounds = c(0, Inf), include = 1:0, size = 1)
    params$kappa <- throw(check_assignment(kappa))
  } else {
    epsilon <- check_numeric(epsilon, include = FALSE, size = 1)
    params$epsilon <- throw(check_assignment(epsilon))
  }

  mloptim <- SequentialOptimization(
    label = "Bayesian Optimization",
    packages = packages,
    params = params,
    fun = fun,
    random = random
  ) %>% set_monitor(progress = progress, verbose = verbose)

  if (verbose_progress && mloptim@monitor$progress) {
    mloptim@monitor$verbose <- verbose_progress
  }

  object@params@optim <- mloptim
  object
}


#' @rdname set_optim-methods
#'
set_optim_bfgs <- function(object, ...) {
  UseMethod("set_optim_bfgs")
}


#' @rdname set_optim-methods
#'
set_optim_bfgs.ModelSpecification <- function(
  object, times = 10, control = list(), random = FALSE, progress = FALSE,
  verbose = 0, ...
) {
  if (!is_optim_method(object)) return(object)

  params <- list()

  times <- check_integer(times, bounds = c(1, Inf), size = 1)
  params$times <- throw(check_assignment(times))

  params$control <- control

  object@params@optim <- SequentialOptimization(
    label = "L-BFGS-B Optimization",
    packages = "stats",
    params = params,
    fun = function(optim, bounds, params, monitor, ...) {
      throw(check_optim_bounds(bounds), call = call("set_optim_bfgs"))
      bounds_mat <- do.call(rbind, bounds)
      lower <- bounds_mat[, 1]
      upper <- bounds_mat[, 2]
      convert <- function(bound, param) {
        if (is.integer(bound)) round_int(param) else param
      }
      stats::optim(
        runif(length(bounds), lower, upper),
        fn = function(x) {
          x <- map("num", convert, bounds, as.list(x))
          min(-optim(x), .Machine$double.xmax)
        },
        method = "L-BFGS-B",
        lower = lower,
        upper = upper,
        control = within(params$control, {
          trace <- monitor$verbose
          fnscale <- 1
          maxit <- params$times
        })
      )
      return(NULL)
    },
    random = random
  ) %>% set_monitor(progress = progress, verbose = verbose)

  object
}


#' @rdname set_optim-methods
#'
set_optim_grid <- function(object, ...) {
  UseMethod("set_optim_grid")
}


#' @rdname set_optim-methods
#'
set_optim_grid.TrainingParams <- function(
  object, random = FALSE, progress = FALSE, ...
) {
  if (!is_optim_method(object)) return(object)

  object@optim <- if (isFALSE(random)) {
    GridSearch()
  } else {
    RandomGridSearch(size = random)
  }

  object %>% set_monitor(which = "optim", progress = progress)
}


#' @rdname set_optim-methods
#'
set_optim_grid.ModelSpecification <- function(object, ...) {
  object@params <- set_optim_grid(object@params, ...)
  object
}


#' @rdname set_optim-methods
#'
set_optim_grid.TunedInput <- function(object, ...) {
  object@params <- set_optim_grid(object@params, ...)
  object
}


#' @rdname set_optim-methods
#'
set_optim_grid.TunedModel <- function(object, ...) {
  object@params <- set_optim_grid(object@params, ...)
  object
}


#' @rdname set_optim-methods
#'
set_optim_pso <- function(object, ...) {
  UseMethod("set_optim_pso")
}


#' @rdname set_optim-methods
#'
set_optim_pso.ModelSpecification <- function(
  object, times = 10, each = NULL, control = list(), random = FALSE,
  progress = FALSE, verbose = 0, ...
) {
  if (!is_optim_method(object)) return(object)

  params <- list()

  times <- check_integer(times, bounds = c(1, Inf), size = 1)
  params$times <- throw(check_assignment(times))

  if (!is.null(each)) {
    each <- check_integer(each, bounds = c(1, Inf), size = 1)
    params$each <- throw(check_assignment(each))
  }

  control$type <- match.arg(control$type, c("SPSO2007", "SPSO2011"))
  params$control <- control

  object@params@optim <- SequentialOptimization(
    label = "Particle Swarm Optimization",
    packages = "pso",
    params = params,
    fun = function(optim, bounds, params, monitor, ...) {
      throw(check_optim_bounds(bounds), call = call("set_optim_pso"))
      bounds_mat <- do.call(rbind, bounds)
      convert <- function(bound, param) {
        if (is.integer(bound)) round_int(param) else param
      }
      if (is.null(params$each)) {
        params$each <- switch(params$control$type,
          "SPSO2011" = 40,
          floor(10 + 2 * sqrt(length(bounds)))
        )
      }
      max_iter <- params$times * params$each
      res <- pso::psoptim(
        rep(NA, length(bounds)),
        fn = function(x) {
          x <- map("num", convert, bounds, as.list(x))
          min(-optim(x, max_iter = max_iter), .Machine$double.xmax)
        },
        lower = bounds_mat[, 1],
        upper = bounds_mat[, 2],
        control = within(params$control, {
          trace <- monitor$verbose
          fnscale <- 1
          maxit <- params$times
          s <- params$each
        })
      )$counts

      times <- res["iteration"]
      tibble(epoch = rep(seq_len(times), each = res["function"] / times))
    },
    random = random
  ) %>% set_monitor(progress = progress, verbose = verbose)

  object
}


#' @rdname set_optim-methods
#'
set_optim_sann <- function(object, ...) {
  UseMethod("set_optim_sann")
}


#' @rdname set_optim-methods
#'
set_optim_sann.ModelSpecification <- function(
  object, times = 10, control = list(), random = FALSE, progress = FALSE,
  verbose = 0, ...
) {
  if (!is_optim_method(object)) return(object)

  params <- list()

  times <- check_integer(times, bounds = c(1, Inf), size = 1)
  params$times <- throw(check_assignment(times))

  params$control <- control

  object@params@optim <- SequentialOptimization(
    label = "Simulated Annealing",
    packages = "stats",
    params = params,
    fun = function(optim, bounds, params, monitor, ...) {
      throw(check_optim_bounds(bounds), call = call("set_optim_sann"))
      bounds_mat <- do.call(rbind, bounds)
      lower <- bounds_mat[, 1]
      upper <- bounds_mat[, 2]
      convert <- function(bound, param) {
        if (is.integer(bound)) round_int(param) else param
      }
      stats::optim(
        runif(length(bounds), lower, upper),
        fn = function(x) {
          x <- map("num", convert, bounds, as.list(x))
          if (all(x >= lower & x <= upper)) {
            -optim(x, max_iter = params$times)
          } else Inf
        },
        method = "SANN",
        control = within(params$control, {
          trace <- monitor$verbose
          fnscale <- 1
          maxit <- params$times
        })
      )
      return(NULL)
    },
    random = random
  ) %>% set_monitor(progress = progress, verbose = verbose)

  object
}


#' @rdname set_optim-methods
#'
set_optim_method <- function(object, ...) {
  UseMethod("set_optim_method")
}


#' @rdname set_optim-methods
#'
set_optim_method.ModelSpecification <- function(
  object, fun, label = "Optimization Function", packages = character(),
  params = list(), random = FALSE, progress = FALSE, verbose = FALSE, ...
) {
  if (!is_optim_method(object)) return(object)

  object@params@optim <- SequentialOptimization(
    label = label,
    packages = packages,
    params = params,
    fun = function(...) fun(...),
    random = random
  ) %>% set_monitor(progress = progress, verbose = verbose)

  object
}


#' Resampling Prediction Control
#'
#' Set parameters that control prediction during resample estimation of model
#' performance.
#'
#' @param object \link[=controls]{control} object.
#' @param times,distr,method arguments passed to \code{\link{predict}}.
#' @param ... arguments passed to other methods.
#'
#' @return Argument \code{object} updated with the supplied parameters.
#'
#' @seealso \code{\link{resample}}, \code{\link{set_monitor}},
#' \code{\link{set_optim}}, \code{\link{set_strata}}
#'
#' @examples
#' CVControl() %>% set_predict(times = 1:3)
#'
set_predict <- function(
  object, times = numeric(), distr = character(), method = character(), ...
) {
  stopifnot(is(object, "MLControl"))

  predict <- list()

  times <- check_numeric(
    times, bounds = c(0, Inf), include = FALSE, size = NA, nonempty = FALSE
  )
  predict$times <- throw(check_assignment(times))

  if (length(distr)) distr <- check_character(distr, size = 1)
  predict$distr <- throw(check_assignment(distr))

  if (length(method)) method <- check_character(method, size = 1)
  predict$method <- throw(check_assignment(method))

  object@predict <- predict
  object
}


#' Resampling Stratification Control
#'
#' Set parameters that control the construction of strata during resample
#' estimation of model performance.
#'
#' @param object \link[=controls]{control} object.
#' @param breaks number of quantile bins desired for stratification of numeric
#'   data during resampling.
#' @param nunique number of unique values at or below which numeric data are
#'   stratified as categorical.
#' @param prop minimum proportion of data in each strata.
#' @param size minimum number of values in each strata.
#' @param ... arguments passed to other methods.
#'
#' @details
#' The arguments control resampling strata which are constructed from numeric
#' proportions for \code{\link{BinomialVariate}}; original values for
#' \code{character}, \code{factor}, \code{logical}, \code{numeric}, and
#' \code{ordered}; first columns of values for \code{matrix}; and numeric times
#' within event statuses for \code{Surv}.  Stratification of survival data by
#' event status only can be achieved by setting \code{breaks = 1}.  Numeric
#' values are stratified into quantile bins and categorical values into factor
#' levels.  The number of bins will be the largest integer less than or equal to
#' \code{breaks} satisfying the \code{prop} and \code{size} control argument
#' thresholds.  Categorical levels below the thresholds will be pooled
#' iteratively by reassigning values in the smallest nominal level to the
#' remaining ones at random and by combining the smallest adjacent ordinal
#' levels.  Missing values are replaced with non-missing values sampled at
#' random with replacement.
#'
#' @return Argument \code{object} updated with the supplied parameters.
#'
#' @seealso \code{\link{resample}}, \code{\link{set_monitor}},
#' \code{\link{set_optim}}, \code{\link{set_predict}}
#'
#' @examples
#' CVControl() %>% set_strata(breaks = 3)
#'
set_strata <- function(
  object, breaks = 4, nunique = 5, prop = 0.1, size = 20, ...
) {
  stopifnot(is(object, "MLControl"))

  strata <- list()

  breaks <- check_integer(breaks, bounds = c(1, Inf), size = 1)
  strata$breaks <- throw(check_assignment(breaks))

  nunique <- check_integer(nunique, bounds = c(1, Inf), size = 1)
  strata$nunique <- throw(check_assignment(nunique))

  prop <- check_numeric(prop, bounds = c(0, 1), size = 1)
  strata$prop <- throw(check_assignment(prop))

  size <- check_integer(size, bounds = c(1, Inf), size = 1)
  strata$size <- throw(check_assignment(size))

  object@strata <- strata
  object
}

## Recursive Feature Elimination

test_that("test rfe", {
  skip_if_not(TEST_ALL)
  with_parallel({

    library(recipes)
    library(survival)

    model <- GBMModel

    expect_output <- function(x, nrow = NULL) {
      x <- summary(x)
      pass <- is.data.frame(x) &&
        nrow(x) > 0 &&
        is.numeric(x$params$size) &&
        all(diff(x$params$size) < 0) &&
        is.list(x$terms) &&
        all(mapply(is.character, x$terms)) &&
        all(lengths(x$terms) == x$params$size) &&
        is.logical(x$selected) &&
        sum(x$selected) == 1 &&
        all(sapply(x$metrics, is.numeric))
      if (!is.null(nrow)) pass <- pass && nrow(x) == nrow
      expect_true(pass)
      x
    }

    ## test matrix
    x <- model.matrix(sale_amount ~ . - 1, data = ICHomes)
    y <- ICHomes$sale_amount
    rfe(x, y, model = model) %>% expect_output()

    ## test recipe
    rfe(recipe(sale_amount ~ ., data = ICHomes), model = model) %>%
      expect_output()

    ## test SelectedModelFrame
    sel_mf <- SelectedInput(
      sale_amount ~ sale_year + built + style + construction,
      sale_amount ~ sale_year + base_size + bedrooms + basement,
      data = ICHomes
    )
    rfe(sel_mf, model = model) %>% expect_output()

    ## test SelectedModelRecipe
    sel_rec <- SelectedInput(
      recipe(sale_amount ~ sale_year + built + style + construction,
             data = ICHomes),
      recipe(sale_amount ~ sale_year + base_size + bedrooms + basement,
             data = ICHomes)
    )
    rfe(sel_rec, model = model) %>% expect_output()

    ## test survival
    rfe(Surv(time, status) ~ ., data = veteran, model = model) %>%
      expect_output()
    rfe(Surv(time, status) ~ ., data = veteran, model = model,
        control = CVControl() %>%
          set_predict(times = quantile(veteran$time, 1:2 / 3))) %>%
      expect_output()
    rfe(Surv(time, status) ~ ., data = veteran, model = model,
        control = CVControl() %>% set_predict(times = median(veteran$time))) %>%
      expect_output()

    ## test defaults
    rfe_default <- function(...) {
      rfe(sale_amount ~ ., data = ICHomes, model = model, ...)
    }
    res <- rfe_default() %>% expect_output()
    n <- nrow(res)

    ## test props
    rfe_default(props = c(1, 1)) %>% expect_output(nrow = 1)
    rfe_default(props = 1:2) %>% expect_error()
    rfe_default(props = 1:3 / 3) %>% expect_output(nrow = 3)
    rfe_default(props = 0:3 / 3) %>% expect_error()

    ## test sizes
    rfe_default(sizes = 0) %>% expect_error()
    rfe_default(sizes = c(8, 8, 12, 100)) %>% expect_output(nrow = 3)

    ## test random
    rfe_default(random = TRUE) %>% expect_output(nrow = n)

    ## test recompute
    rfe_default(recompute = TRUE) %>% expect_output(nrow = n)

    ## test optimize
    rfe_default(optimize = "local") %>% expect_output()

    ## test samples
    rfe_default(samples = 2) %>% expect_output(nrow = n)
    rfe_default(samples = c(2, 2)) %>% expect_output(nrow = n)
    rfe_default(samples = c(rfe = 0)) %>% expect_error()
    rfe_default(samples = c(varimp = 0)) %>% expect_error()

    ## test metrics
    rfe_default(metrics = mse) %>% expect_output(nrow = n)

    ## test stat
    rfe_default(
      samples = c(2, 2),
      stat = function(x) mean(x) + sd(x) / sqrt(length(x))
    ) %>% expect_output(nrow = n)
    rfe_default(stat = identity) %>% expect_error()

  })
})

with_parallel <- function(code) {

  if (utils::packageVersion("bartMachine") >= "1.4") {
    options(
      java.parameters = c(
        "-Xmx20g", "--add-modules=jdk.incubator.vector", "-XX:+UseZGC"
      )
    )
  } else {
    options(java.parameters = "-Xmx20g")
  }

  cores <- as.numeric(Sys.getenv("TEST_CORES"))
  if (is.na(cores)) cores <- 1
  if (require("doParallel", quietly = TRUE) && cores > 1) {
    doParallel::registerDoParallel(cores = cores)
    on.exit(doParallel::stopImplicitCluster())
  }

  code
}

#' GLM Lasso or Elasticnet Model
#'
#' Fit a generalized linear model via penalized maximum likelihood.
#'
#' @param family optional response type.  Set automatically according to the
#'   class type of the response variable.
#' @param alpha elasticnet mixing parameter.
#' @param lambda regularization parameter.  The default value \code{lambda = 0}
#'   performs no regularization and should be increased to avoid model fitting
#'   issues if the number of predictor variables is greater than the number of
#'   observations.
#' @param standardize logical flag for predictor variable standardization, prior
#'   to model fitting.
#' @param intercept logical indicating whether to fit intercepts.
#' @param penalty.factor vector of penalty factors to be applied to each
#'   coefficient.
#' @param standardize.response logical indicating whether to standardize
#'   \code{"mgaussian"} response variables.
#' @param thresh convergence threshold for coordinate descent [deprecated].
#' @param maxit maximum number of passes over the data for all lambda values
#'   [deprecated].
#' @param type.gaussian algorithm type for guassian models.
#' @param type.logistic algorithm type for logistic models.
#' @param type.multinomial algorithm type for multinomial models.
#' @param cox.ties character string of \code{"efron"} or \code{"breslow"} for
#'   the method of handling ties in Cox survival models.
#' @param control named list of algorithm control parameters, providing
#'   per-call overrides of session defaults set by
#'   \code{\link[glmnet]{glmnet.control}}.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{BinomialVariate}, \code{factor},
#'     \code{matrix}, \code{numeric}, \code{PoissonVariate}, \code{Surv}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \code{lambda}, \code{alpha}
#'   }
#' }
#'
#' Default argument values and further model details can be found in the source
#' See Also link below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[glmnet]{glmnet}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package glmnet to run
#'
#' fit(sale_amount ~ ., data = ICHomes, model = GLMNetModel(lambda = 0.01))
#' }
#'
GLMNetModel <- function(
  family = NULL, alpha = 1, lambda = 0, standardize = TRUE,
  intercept = logical(), penalty.factor = .(rep(1, nvars)),
  standardize.response = FALSE, thresh = 1e-7, maxit = 100000,
  type.gaussian = .(if (nvars < 500) "covariance" else "naive"),
  type.logistic = c("Newton", "modified.Newton"),
  type.multinomial = c("ungrouped", "grouped"), cox.ties = NULL,
  control = list()
) {

  type.logistic <- match.arg(type.logistic)
  type.multinomial <- match.arg(type.multinomial)

  if (!missing(thresh)) {
    throw(DeprecatedCondition(
      "Argument 'thresh'", "'control = list(thresh = ...)'"
    ))
  }
  if (is.null(control$thresh)) control$thresh <- thresh
  remove(thresh)

  if (!missing(maxit)) {
    throw(DeprecatedCondition(
      "Argument 'maxit'", "'control = list(maxit = ...)'"
    ))
  }
  if (is.null(control$maxit)) control$maxit <- maxit
  remove(maxit)

  MLModel(

    name = "GLMNetModel",
    label = "Lasso and Elastic-Net",
    packages = "glmnet",
    response_types = c(
      "BinomialVariate", "factor", "matrix", "numeric", "PoissonVariate", "Surv"
    ),
    weights = TRUE,
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = c("lambda", "alpha"),
      get_values = c(
        function(n, data, ...) {
          lambda <- numeric()
          model <- GLMNetModel(lambda = NULL)
          nlambda <- 0
          while ((length(lambda) < 2) && (nlambda < 100)) {
            nlambda <- nlambda + 10
            model@params$nlambda <- nlambda
            lambda <- suppressWarnings(fit(data, model = model)$lambda)
          }
          if (length(lambda) >= 2) {
            exp(seq(log(min(lambda)), log(max(lambda)), length = n))
          } else {
            throw(LocalWarning(
              "GLMNetModel grid values for lambda could not be generated ",
              "automatically."
            ))
            numeric()
          }
        },
        function(n, ...) seq(0.1, 1, length = n)
      )
    ),

    fit = function(formula, data, weights, family = NULL, nlambda = 1, ...) {
      x <- model.matrix(data, intercept = FALSE)
      offset <- model.offset(data)
      y <- response(data)
      if (is.null(family)) {
        family <- switch_class(y,
          "BinomialVariate" = {
            y <- y[, c("Failure", "Success")]
            "binomial"
          },
          "factor" = if (nlevels(y) <= 2) "binomial" else "multinomial",
          "matrix" = "mgaussian",
          "numeric" = "gaussian",
          "PoissonVariate" = "poisson",
          "Surv" = "cox"
        )
      }
      args <- list(
        x, y, family = family, weights = weights, offset = offset,
        nlambda = nlambda, ...
      )
      if (packageVersion("glmnet") < "5.0") {
        args$cox.ties <- NULL
        for (name in c("thresh", "dfmax", "pmax", "maxit")) {
          args[[name]] <- args$control[[name]]
        }
        args$control <- NULL
      }
      do.call(glmnet::glmnet, args)
    },

    predict = function(object, newdata, .MachineShop, ...) {
      input <- .MachineShop$input
      y <- response(input)
      newx <- model.matrix(newdata, intercept = FALSE)
      newoffset <- model.offset(newdata)
      if (is.Surv(y)) {
        data <- PredictorFrame(input)
        lp <- predict(
          object, newx = model.matrix(data, intercept = FALSE),
          newoffset = model.offset(data), type = "link"
        )[, 1]
        new_lp <- predict(
          object, newx = newx, newoffset = newoffset, type = "link"
        )[, 1]
        predict(y, lp, new_lp, ...)
      } else {
        predict(
          object, newx = newx, newoffset = newoffset, s = object$lambda[1],
          type = "response"
        )
      }
    },

    varimp = function(object, ...) {
      convert <- function(x) abs(drop(as.matrix(x)))
      beta <- object$beta
      structure(
        if (is.list(beta)) {
          as.data.frame(map(convert, beta), check.names = FALSE)
        } else {
          convert(beta)
        },
        metric = "coefficient"
      )
    }

  )

}

MLModelFunction(GLMNetModel) <- NULL

#' Extract Elements of an Object
#'
#' Operators acting on data structures to extract elements.
#'
#' @name extract
#' @rdname extract-methods
#'
#' @param x object from which to extract elements.
#' @param i,j,... indices specifying elements to extract.
#' @param drop logical indicating that the result be returned as an object
#'   coerced to the lowest dimension possible if \code{TRUE} or
#'   with the original dimensions and class otherwise.
#'
NULL


#' @rdname extract-methods
#'
"[.BinomialVariate" <- function(x, i, j, ..., drop = FALSE) {
  y <- if (missing(j)) unclass(x)[i, , drop = drop] else NextMethod(drop = drop)
  if (identical(colnames(x), colnames(y))) {
    structure(y, class = "BinomialVariate")
  } else y
}


#' @rdname extract-methods
#'
setMethod("[",
  c(x = "DiscreteVariate", i = "ANY", j = "missing", drop = "missing"),
  function(x, i) {
    new(class(x), asS3(x)[i], min = x@min, max = x@max)
  }
)


#' @rdname extract-methods
#'
setMethod("[", c(x = "ListOf", i = "ANY", j = "missing", drop = "missing"),
  function(x, i) {
    ListOf(asS3(x)[i])
  }
)


#' @rdname extract-methods
#'
setMethod("[", c(x = "ModelFrame", i = "ANY", j = "ANY", drop = "ANY"),
  function(x, i, j, ..., drop = FALSE) {
    y <- as.data.frame(x)[i, j, drop = drop]
    if (identical(colnames(x), colnames(y))) {
      y <- ModelFrame(terms(x), y, na.rm = FALSE)
      if (is(x, "SelectedInput")) {
        y <- new(class(x), y, candidates = x@candidates, params = x@params)
      }
    }
    y
  }
)


#' @rdname extract-methods
#'
setMethod("[", c(x = "ModelFrame", i = "ANY", j = "missing", drop = "ANY"),
  function(x, i, j, ..., drop = FALSE) {
    ninds <- nargs() - 1 - !missing(drop)
    if (ninds > 1) {
      if (!drop) {
        as(asS3(x)[i, , drop = FALSE], class(x))
      } else x[i, TRUE, drop = TRUE]
    } else x[, i, drop = FALSE]
  }
)


#' @rdname extract-methods
#'
setMethod("[", c(x = "ModelFrame", i = "missing", j = "ANY", drop = "ANY"),
  function(x, i, j, ..., drop = FALSE) {
    x[TRUE, j, drop = drop]
  }
)


#' @rdname extract-methods
#'
setMethod("[", c(x = "ModelFrame", i = "missing", j = "missing", drop = "ANY"),
  function(x, i, j, ..., drop = FALSE) {
    x[TRUE, , drop = drop]
  }
)


#' @rdname extract-methods
#'
setMethod("[", c(x = "RecipeGrid", i = "ANY", j = "ANY", drop = "ANY"),
  function(x, i, j, ..., drop = FALSE) {
    asS3(x)[i, j, drop = drop]
  }
)


#' @rdname extract-methods
#'
setMethod("[", c(x = "Resample", i = "ANY", j = "ANY", drop = "ANY"),
  function(x, i, j, ..., drop = FALSE) {
    y <- as.data.frame(x)[i, j, drop = drop]
    if (identical(colnames(x), colnames(y))) {
      Resample(y, control = x@control, vars = x@vars)
    } else y
  }
)


#' @rdname extract-methods
#'
setMethod("[", c(x = "Resample", i = "ANY", j = "missing", drop = "ANY"),
  function(x, i, j, ..., drop = FALSE) {
    ninds <- nargs() - 1 - !missing(drop)
    if (ninds > 1) x[i, TRUE, drop = drop] else x[, i, drop = FALSE]
  }
)


#' @rdname extract-methods
#'
setMethod("[", c(x = "Resample", i = "missing", j = "missing", drop = "ANY"),
  function(x, i, j, ..., drop = FALSE) {
    x[, TRUE, drop = drop]
  }
)


#' @rdname extract-methods
#'
setMethod("[", c(x = "SurvMatrix", i = "ANY", j = "ANY", drop = "ANY"),
  function(x, i, j, ..., drop = FALSE) {
    y <- asS3(x)[i, j, drop = drop]
    if (is.matrix(y)) {
      times <- x@times[match(colnames(y), colnames(x))]
      new(class(x), y, times = times, distr = x@distr)
    } else y
  }
)


#' @rdname extract-methods
#'
setMethod("[", c(x = "SurvTimes", i = "ANY", j = "missing", drop = "missing"),
  function(x, i) {
    new(class(x), asS3(x)[i], distr = x@distr)
  }
)

#' Bayesian Additive Regression Trees Model
#'
#' Builds a BART model for regression or classification.
#'
#' @param num_trees number of trees to be grown in the sum-of-trees model.
#' @param num_burn number of MCMC samples to be discarded as "burn-in".
#' @param num_iter number of MCMC samples to draw from the posterior
#'   distribution.
#' @param alpha,beta base and power hyperparameters in tree prior for whether a
#'   node is nonterminal or not.
#' @param k regression prior probability that \eqn{E(Y|X)} is
#'   contained in the interval \eqn{(y_{min}, y_{max})}, based on a normal
#'   distribution.
#' @param q quantile of the prior on the error variance at which the data-based
#'   estimate is placed.
#' @param nu regression degrees of freedom for the inverse \eqn{sigma^2} prior.
#' @param mh_prob_steps vector of prior probabilities for proposing changes to
#'   the tree structures: (GROW, PRUNE, CHANGE).
#' @param verbose logical indicating whether to print progress information about
#'   the algorithm.
#' @param ... additional arguments to \code{\link[bartMachine]{bartMachine}}.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{binary factor}, \code{numeric}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \code{alpha}, \code{beta}, \code{k}, \code{nu}
#'   }
#' }
#'
#' Further model details can be found in the source link below.
#'
#' In calls to \code{\link{varimp}} for \code{BARTMachineModel}, argument
#' \code{type} may be specified as \code{"splits"} (default) for the
#' proportion of time each predictor is chosen for a splitting rule or as
#' \code{"trees"} for the proportion of times each predictor appears in a tree.
#' Argument \code{num_replicates} is also available to control the number of
#' BART replicates used in estimating the inclusion proportions [default: 5].
#' Variable importance is automatically scaled to range from 0 to 100.  To
#' obtain unscaled importance values, set \code{scale = FALSE}.  See example
#' below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[bartMachine]{bartMachine}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' \dontrun{
#' ## Requires prior installation of suggested package bartMachine, java, and
#' ## setting of java options as shown below to run
#' 
#' if (packageVersion("bartMachine") >= "1.4") {
#'   options(
#'     java.parameters = c(
#'       "-Xmx20g", "--add-modules=jdk.incubator.vector", "-XX:+UseZGC"
#'     )
#'   )
#' } else {
#'   options(java.parameters = "-Xmx20g")
#' }
#'
#' model_fit <- fit(sale_amount ~ ., data = ICHomes, model = BARTMachineModel)
#' varimp(model_fit, method = "model", type = "splits", num_replicates = 20,
#'        scale = FALSE)
#' }
#'
BARTMachineModel <- function(
  num_trees = 50, num_burn = 250, num_iter = 1000, alpha = 0.95, beta = 2,
  k = 2, q = 0.9, nu = 3, mh_prob_steps = c(2.5, 2.5, 4) / 9, verbose = FALSE,
  ...
) {

  MLModel(

    name = "BARTMachineModel",
    label = "Bayesian Additive Regression Trees",
    packages = "bartMachine",
    response_types = c("binary", "numeric"),
    predictor_encoding = "model.matrix",
    na.rm = "response",
    params = new_params(environment(), ...),

    gridinfo = new_gridinfo(
      param = c("alpha", "beta", "k", "nu"),
      get_values = c(
        function(n, ...) seq(0.9, 0.99, length = n),
        function(n, ...) seq(1, 3, length = n),
        function(n, ...) seq(2, length = min(n, 10)),
        function(n, data, ...) {
          if (is.numeric(response(data))) seq(2, length = min(n, 10))
        }
      )
    ),

    fit = function(formula, data, weights, use_missing_data, ...) {
      x <- as.data.frame(model.matrix(data, intercept = FALSE))
      y <- response(data)
      if (is_response(y, "binary")) y <- factor(y, levels = rev(levels(y)))
      bartMachine::build_bart_machine(x, y, use_missing_data = TRUE, ...)
    },

    predict = function(object, newdata, ...) {
      newx <- model.matrix(newdata, intercept = FALSE)
      predict(object, new_data = as.data.frame(newx))
    },

    varimp = function(
      object, type = c("splits", "trees"), num_replicates = 5, ...
    ) {
      type <- match.arg(type)
      structure(
        bartMachine::investigate_var_importance(
          object, type = type, num_replicates = num_replicates, plot = FALSE
        )$avg_var_props,
        metric = type
      )
    }

  )

}

MLModelFunction(BARTMachineModel) <- NULL

#' Weighted k-Nearest Neighbor Model
#'
#' Fit a k-nearest neighbor model for which the k nearest training set vectors
#' (according to Minkowski distance) are found for each row of the test set, and
#' prediction is done via the maximum of summed kernel densities.
#'
#' @param k numer of neigbors considered.
#' @param distance Minkowski distance parameter.
#' @param scale logical indicating whether to scale predictors to have equal
#'   standard deviations.
#' @param kernel kernel to use.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric}, \code{ordinal}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \code{k}, \code{distance}*, \code{kernel}*
#'   }
#' }
#' * excluded from grids by default
#'
#' Further model details can be found in the source link below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[kknn]{kknn}}, \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package kknn to run
#'
#' fit(Species ~ ., data = iris, model = KNNModel)
#' }
#'
KNNModel <- function(
  k = 7, distance = 2, scale = TRUE,
  kernel = c(
    "optimal", "biweight", "cos", "epanechnikov", "gaussian", "inv", "rank",
    "rectangular", "triangular", "triweight"
  )
) {

  kernel <- match.arg(kernel)

  MLModel(

    name = "KNNModel",
    label = "K-Nearest Neighbors Model",
    packages = "kknn",
    response_types = c("factor", "numeric", "ordered"),
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = c("k", "distance", "kernel"),
      get_values = c(
        function(n, data, ...) {
          round_int(seq_range(0, 5, c(1, nrow(data) / 3), n + 1))
        },
        function(n, ...) seq_inner(0, 4, length = n),
        function(n, ...) {
          kernels <- c(
            "optimal", "rectangular", "triangular", "epanechnikov", "biweight",
            "triweight", "cos", "inv", "gaussian", "rank"
          )
          head(kernels, n)
        }
      ),
      default = c(TRUE, FALSE, FALSE)
    ),

    fit = function(formula, data, weights, ...) {
      list(formula = formula, train = as.data.frame(formula, data = data), ...)
    },

    predict = function(object, newdata, ...) {
      opts <- options("contrasts")

      attach_objects(list(
        contr.dummy = kknn::contr.dummy,
        contr.ordinal = kknn::contr.ordinal,
        contr.metrics = kknn::contr.metric
      ), name = "kknn_exports")

      object$test <- as.data.frame(newdata)
      pred <- tryCatch(
        do.call(kknn::kknn, object),
        error = function(cond) {
          options(opts)
          stop(cond)
        }
      )
      if (pred$response == "continuous") pred$fitted.values else pred$prob
    }

  )

}

MLModelFunction(KNNModel) <- NULL

#' Least Angle Regression, Lasso and Infinitesimal Forward Stagewise Models
#'
#' Fit variants of Lasso, and provide the entire sequence of coefficients and
#' fits, starting from zero to the least squares fit.
#'
#' @param type model type.
#' @param trace logical indicating whether status information is printed during
#'   the fitting process.
#' @param normalize whether to standardize each variable to have unit L2 norm.
#' @param intercept whether to include an intercept in the model.
#' @param step algorithm step number to use for prediction.  May be a decimal
#'   number indicating a fractional distance between steps.  If specified, the
#'   maximum number of algorithm steps will be \code{ceiling(step)}; otherwise,
#'   \code{step} will be set equal to the source package default maximum
#'   [default: \code{max.steps}].
#' @param use.Gram whether to precompute the Gram matrix.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{numeric}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameter:}{
#'     \code{step}
#'   }
#' }
#'
#' Default argument values and further model details can be found in the source
#' See Also link below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[lars]{lars}}, \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package lars to run
#'
#' fit(sale_amount ~ ., data = ICHomes, model = LARSModel)
#' }
#'
LARSModel <- function(
  type = c("lasso", "lar", "forward.stagewise", "stepwise"), trace = FALSE,
  normalize = TRUE, intercept = TRUE, step = numeric(), use.Gram = TRUE
) {

  type <- match.arg(type)

  MLModel(

    name = "LARSModel",
    label = "Least Angle Regression",
    packages = "lars",
    response_types = "numeric",
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = "step",
      get_values = c(
        function(n, data, ...) seq_nvars(data, LARSModel, n)
      )
    ),

    fit = function(formula, data, weights, step = NULL, ...) {
      x <- model.matrix(data, intercept = FALSE)
      y <- response(data)
      if (is.null(step)) {
        res <- lars::lars(x, y, ...)
        step <- length(res$df)
      } else {
        res <- lars::lars(x, y, max.steps = ceiling(step), ...)
        step <- step
      }
      attr(res, ".MachineShop") <- list(step = step)
      res
    },

    predict = function(object, newdata, .MachineShop, ...) {
      newx <- model.matrix(newdata, intercept = FALSE)
      predict(object, newx = newx, s = .MachineShop$step, type = "fit")$fit
    }

  )

}

MLModelFunction(LARSModel) <- NULL

case_comps <- function(object, ...) {
  UseMethod("case_comps")
}


case_comps.MLModelFit <- function(object, ...) {
  case_comps(as.MLInput(object), ...)
}


case_comps.ModelFrame <- function(
  object, newdata = NULL, types = c("names", "weights"), response = FALSE,
  original = FALSE, ...
) {
  stopifnot(!original)
  data <- if (is.null(newdata)) object else newdata
  res <- map(function(type) {
    name <- case_comp_name(object, type)
    if (length(name)) data[[name]]
  }, types)
  if (response) res$response <- response(object, newdata)
  res
}


case_comps.recipe <- function(
  object, newdata = NULL, types = c("names", "weights"), response = FALSE,
  original = FALSE, ...
) {
  stopifnot(!original)
  names <- map(function(type) case_comp_name(object, type), types)
  data <- if (any(lengths(names)) || response) {
    bake(prep(object, retain = FALSE), newdata = newdata)
  }
  res <- map(function(type) {
    name <- names[[type]]
    if (length(name)) data[[name]]
  }, types)
  if (response) res$response <- response(terms(object), data)
  res
}


case_comp_name <- function(object, ...) {
  UseMethod("case_comp_name")
}


case_comp_name.data.frame <- function(object, type, ...) {
  name <- paste0("(", type, ")")
  if (name %in% names(object)) name
}


case_comp_name.recipe <- function(object, type, ...) {
  type <- switch(type,
    "groups" = "group",
    "names" = "name",
    "strata" = "stratum",
    "weights" = "weight",
    return(NULL)
  )
  role <- paste0("case_", type)
  info <- summary(object)
  name <- info$variable[info$role == role]
  if (length(name) == 1) {
    name
  } else if (length(name) > 1) {
    throw(Error("Multiple ", role, " variables."))
  }
}


case_groups <- function(object, ...) {
  name <- case_comp_name(object, type = "groups")
  if (length(name)) as.data.frame(object)[[name]]
}


case_strata <- function(object, ...) {
  UseMethod("case_strata")
}


case_strata.default <- function(object, ...) {
  types <- c(
    "BinomialVariate", "character", "factor", "logical", "matrix", "numeric",
    "Surv"
  )
  throw(TypeError(object, types, "`object`"))
}


case_strata.BinomialVariate <- function(object, ...) {
  case_strata(as.numeric(object), ...)
}


case_strata.character <- function(object, ...) {
  case_strata(as.factor(object), ...)
}


case_strata.factor <- function(object, prop = 0.1, size = 20, ...) {
  object <- sample_replace(object, is.na(object))
  min_count <- max(prop * length(object), size, 1)
  while (nlevels(object) > 1 && any((counts <- table(object)) < min_count)) {
    if (is.ordered(object)) {
      i <- which.min(head(counts, -1) + tail(counts, -1))
      inds <- c(i, i + 1)
      levels(object)[inds] <- paste(levels(object)[inds], collapse = "+")
    } else {
      i <- which.min(counts)
      object <- sample_replace(object, object == levels(object)[i])
      levels(object)[i] <- NA
    }
  }
  object
}


case_strata.logical <- function(object, ...) {
  case_strata(as.factor(object), ...)
}


case_strata.matrix <- function(object, ...) {
  case_strata(object[, 1], ...)
}


case_strata.ModelFrame <- function(object, ...) {
  name <- case_comp_name(object, "strata")
  if (length(name)) case_strata(object[[name]], ...)
}


case_strata.recipe <- function(object, ...) {
  name <- case_comp_name(object, "strata")
  if (length(name)) case_strata(as.data.frame(object)[[name]], ...)
}


case_strata.numeric <- function(
  object, breaks = 4, nunique = 5, prop = 0.1, size = 20, ...
) {
  object <- sample_replace(object, is.na(object))
  if (length(unique(object)) <= max(nunique, 1)) {
    object <- ordered(object)
    levels(object) <- make_names_len(nlevels(object), "level")
    res <- case_strata(object, prop = prop, size = size)
  } else {
    min_count <- max(prop * length(object), 1, size)
    breaks <- max(breaks, 1)
    while (breaks > 0) {
      quants <- quantile(object, 0:breaks / breaks)
      res <- cut(object, unique(quants), include.lowest = TRUE)
      if (all(table(res) >= min_count)) break
      breaks <- breaks - 1
    }
  }
  res
}


case_strata.Surv <- function(
  object, breaks = 4, nunique = 5, prop = 0.1, size = 20, ...
) {
  status <- object[, "status"]
  status <- sample_replace(status, is.na(status))

  time_split <- split(time(object), status)
  strata_split <- list()
  strata_levels <- character()

  for (name in names(time_split)) {
    time <- time_split[[name]]
    res <- case_strata(
      time, breaks = breaks, nunique = nunique,
      prop = prop * length(status) / length(time), size = size
    )
    levels(res) <- paste(name, levels(res), sep = ":")
    strata_levels <- c(strata_levels, levels(res))
    strata_split[[name]] <- as.character(res)
  }

  factor(unsplit(strata_split, status), levels = strata_levels)
}


#' Extract Case Weights
#'
#' Extract the case weights from an object.
#'
#' @param object model \link{fit} result, \code{\link{ModelFrame}}, or
#'   \code{\link[recipes]{recipe}}.
#' @param newdata dataset from which to extract the weights if given; otherwise,
#'   \code{object} is used.  The dataset should be given as a \code{ModelFrame}
#'   or as a \link[=data.frame]{data frame} if \code{object} contains a
#'   \code{ModelFrame} or a \code{recipe}, respectively.
#'
#' @examples
#' ## Training and test sets
#' inds <- sample(nrow(ICHomes), nrow(ICHomes) * 2 / 3)
#' trainset <- ICHomes[inds, ]
#' testset <- ICHomes[-inds, ]
#'
#' ## ModelFrame case weights
#' trainmf <- ModelFrame(sale_amount ~ . - built, data = trainset, weights = built)
#' testmf <- ModelFrame(formula(trainmf), data = testset, weights = built)
#' mf_fit <- fit(trainmf, model = GLMModel)
#' rmse(response(mf_fit, testmf), predict(mf_fit, testmf),
#'      case_weights(mf_fit, testmf))
#'
#' ## Recipe case weights
#' library(recipes)
#' rec <- recipe(sale_amount ~ ., data = trainset) %>%
#'   role_case(weight = built, replace = TRUE)
#' rec_fit <- fit(rec, model = GLMModel)
#' rmse(response(rec_fit, testset), predict(rec_fit, testset),
#'      case_weights(rec_fit, testset))
#'
case_weights <- function(object, newdata = NULL) {
  case_comps(object, newdata, "weights")[[1]]
}

#' Partial Least Squares Model
#'
#' Function to perform partial least squares regression.
#'
#' @param ncomp number of components to include in the model.
#' @param scale logical indicating whether to scale the predictors by the
#'   sample standard deviation.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \code{ncomp}
#'   }
#' }
#'
#' Further model details can be found in the source link below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[pls]{mvr}}, \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package pls to run
#'
#' fit(sale_amount ~ ., data = ICHomes, model = PLSModel)
#' }
#'
PLSModel <- function(ncomp = 1, scale = FALSE) {

  MLModel(

    name = "PLSModel",
    label = "Partial Least Squares",
    packages = "pls",
    response_types = c("factor", "numeric"),
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = "ncomp",
      get_values = c(
        function(n, data, ...) {
          seq_len(min(n, nrow(data), nvars(data, PLSModel) - 1))
        }
      )
    ),

    fit = function(formula, data, weights, ...) {
      y <- response(data)
      data <- as.data.frame(formula, data = data)
      if (is.factor(y)) {
        mm <- model.matrix(~ y - 1)
        colnames(mm) <- levels(y)
        data[[response(formula)]] <- mm
      }
      pls::plsr(formula, data = data, na.action = na.pass, ...)
    },

    predict = function(object, newdata, ...) {
      newdata <- as.data.frame(newdata)
      predict(
        object, newdata = newdata, ncomp = object$ncomp, type = "response"
      )
    },

    varimp = function(object, ...) {
      beta <- coef(object, comps = seq_len(object$ncomp))
      perf <- quote(MSEP.mvr(x)$val[1, , , drop = FALSE]) %>%
        eval(list(x = object), asNamespace("pls"))
      vi <- map("num", function(i) {
        drop(as.matrix(abs(beta[, i, ])) %*% proportions(-diff(perf[, i, ])))
      }, seq_len(ncol(beta)))
      dimnames(vi) <- dimnames(beta)[1:2]
      if (ncol(vi) <= 2) vi <- vi[, 1]
      structure(vi, metric = "coefficient")
    }

  )

}

MLModelFunction(PLSModel) <- NULL

#' Flexible and Penalized Discriminant Analysis Models
#'
#' Performs flexible discriminant analysis.
#'
#' @rdname FDAModel
#'
#' @param theta optional matrix of class scores, typically with number of
#'   columns less than one minus the number of classes.
#' @param dimension dimension of the discriminant subspace, less than the number
#'   of classes, to use for prediction.
#' @param eps numeric threshold for small singular values for excluding
#'   discriminant variables.
#' @param method regression function used in optimal scaling.  The default of
#'   linear regression is provided by \code{\link[mda]{polyreg}} from the
#'   \pkg{mda} package.  For penalized discriminant analysis,
#'   \code{\link[mda]{gen.ridge}} is appropriate.  Other possibilities are
#'   \code{\link[mda]{mars}} for multivariate adaptive regression splines and
#'   \code{\link[mda]{bruto}} for adaptive backfitting of additive splines.  Use
#'   the \code{\link[=quote]{.}} operator to quote specified functions.
#' @param ... additional arguments to \code{method} for \code{FDAModel} and to
#'   \code{FDAModel} for \code{PDAModel}.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \itemize{
#'       \item FDAModel: \code{nprune}, \code{degree}*
#'       \item PDAModel: \code{lambda}
#'     }
#'   }
#' }
#' * excluded from grids by default
#'
#' The \code{\link{predict}} function for this model additionally accepts the
#' following argument.
#' \describe{
#'   \item{\code{prior}}{prior class membership probabilities for prediction
#'     data if different from the training set.}
#' }
#'
#' Default argument values and further model details can be found in the source
#' See Also links below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[mda]{fda}}, \code{\link[mda]{predict.fda}},
#' \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package mda to run
#'
#' fit(Species ~ ., data = iris, model = FDAModel)
#' }
#'
FDAModel <- function(
  theta = matrix(NA, 0, 0), dimension = integer(), eps = .Machine$double.eps,
  method = .(mda::polyreg), ...
) {

  MLModel(

    name = "FDAModel",
    label = "Flexible Discriminant Analysis",
    packages = "mda",
    response_types = "factor",
    weights = TRUE,
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment(), ...),

    gridinfo = new_gridinfo(
      param = c("nprune", "degree"),
      get_values = c(
        function(n, data, ...) {
          model_fit <- fit(data, model = EarthModel(pmethod = "none"))
          max_terms <- min(2 + 0.75 * nrow(model_fit$dirs), 200)
          round_int(seq(2, max_terms, length = n))
        },
        function(n, ...) head(1:2, n)
      ),
      default = c(TRUE, FALSE)
    ),

    fit = function(formula, data, weights, ...) {
      mda::fda(
        formula, data = as.data.frame(formula, data = data), weights = weights,
        ...
      )
    },

    predict = function(object, newdata, prior = object$prior, ...) {
      newdata <- as.data.frame(newdata)
      predict(object, newdata = newdata, type = "posterior", prior = prior)
    }

  )

}

MLModelFunction(FDAModel) <- NULL


#' @rdname FDAModel
#'
#' @param lambda shrinkage penalty coefficient.
#' @param df alternative specification of \code{lambda} in terms of equivalent
#'   degrees of freedom.
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package mda to run
#'
#' fit(Species ~ ., data = iris, model = PDAModel)
#' }
#'
PDAModel <- function(lambda = 1, df = numeric(), ...) {
  params <- new_params(environment(), ...)
  params$method <- .(mda::gen.ridge)
  model <- do.call(FDAModel, params, quote = TRUE)
  model@name <- "PDAModel"
  model@label <- "Penalized Discriminant Analysis"
  model@gridinfo <- new_gridinfo(
    param = "lambda",
    get_values = c(
      function(n, ...) c(0, 10^seq_inner(-5, 1, n - 1))
    )
  )
  model
}

MLModelFunction(PDAModel) <- NULL

#' Model Performance Differences
#'
#' Pairwise model differences in resampled performance metrics.
#'
#' @name diff
#' @rdname diff-methods
#'
#' @param x model \link{performance} or \link{resample} result.
#' @param ... arguments passed to other methods.
#'
#' @return \code{PerformanceDiff} class object that inherits from
#' \code{Performance}.
#'
#' @seealso \code{\link{t.test}}, \code{\link{plot}}, \code{\link{summary}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' ## Survival response example
#' library(survival)
#'
#' fo <- Surv(time, status) ~ .
#' control <- CVControl()
#'
#' gbm_res1 <- resample(fo, data = veteran, GBMModel(n.trees = 25), control)
#' gbm_res2 <- resample(fo, data = veteran, GBMModel(n.trees = 50), control)
#' gbm_res3 <- resample(fo, data = veteran, GBMModel(n.trees = 100), control)
#'
#' res <- c(GBM1 = gbm_res1, GBM2 = gbm_res2, GBM3 = gbm_res3)
#' res_diff <- diff(res)
#' summary(res_diff)
#' plot(res_diff)
#' }
#'
NULL


#' @rdname diff-methods
#'
diff.MLModel <- function(x, ...) {
  if (!is_trained(x)) throw(Error("No training results to diff."))
  map(function(step) diff(performance(step)), x@steps)
}


#' @rdname diff-methods
#'
diff.Performance <- function(x, ...) {
  if (ndim(x) <= 2) throw(Error("More than one model needed to diff."))
  indices <- combn(size(x, 3), 2)
  indices1 <- indices[1, ]
  indices2 <- indices[2, ]
  xdiff <- x[, , indices1, drop = FALSE] - x[, , indices2, drop = FALSE]
  model_names <- dimnames(x)[[3]]
  dimnames(xdiff)[[3]] <- paste(
    model_names[indices1], "-", model_names[indices2]
  )
  PerformanceDiff(xdiff, model_names = model_names, control = x@control)
}


#' @rdname diff-methods
#'
diff.Resample <- function(x, ...) {
  diff(performance(x))
}


#' Paired t-Tests for Model Comparisons
#'
#' Paired t-test comparisons of resampled performance metrics from different
#' models.
#'
#' @name t.test
#'
#' @method t.test PerformanceDiff
#'
#' @param x performance \link[=diff]{difference} result.
#' @param adjust method of p-value adjustment for multiple statistical
#'   comparisons as implemented by \code{\link[stats]{p.adjust}}.
#' @param ... arguments passed to other methods.
#'
#' @details
#' The t-test statistic for pairwise model differences of \eqn{R} resampled
#' performance metric values is calculated as
#' \deqn{
#'   t = \frac{\bar{x}_R}{\sqrt{F s^2_R / R}},
#' }
#' where \eqn{\bar{x}_R} and \eqn{s^2_R} are the sample mean and variance.
#' Statistical testing for a mean difference is then performed by comparing
#' \eqn{t} to a \eqn{t_{R-1}} null distribution.  The sample variance in the
#' t statistic is known to underestimate the true variances of cross-validation
#' mean estimators.  Underestimation of these variances will lead to increased
#' probabilities of false-positive statistical conclusions.  Thus, an additional
#' factor \eqn{F} is included in the t statistic to allow for variance
#' corrections.  A correction of \eqn{F = 1 + K / (K - 1)} was found by
#' Nadeau and Bengio (2003) to be a good choice for cross-validation with
#' \eqn{K} folds and is thus used for that resampling method.  The extension of
#' this correction by Bouchaert and Frank (2004) to \eqn{F = 1 + T K / (K - 1)}
#' is used for cross-validation with \eqn{K} folds repeated \eqn{T} times.  For
#' other resampling methods \eqn{F = 1}.
#'
#' @return \code{PerformanceDiffTest} class object that inherits from
#' \code{array}.  p-values and mean differences are contained in the lower and
#' upper triangular portions, respectively, of the first two dimensions.  Model
#' pairs are contained in the third dimension.
#'
#' @references
#' Nadeau, C., & Bengio, Y. (2003). Inference for the generalization error.
#' \emph{Machine Learning}, \emph{52}, 239–81.
#'
#' Bouckaert, R. R., & Frank, E. (2004). Evaluating the replicability of
#' significance tests for comparing learning algorithms. In H. Dai, R. Srikant,
#' & C. Zhang (Eds.), \emph{Advances in knowledge discovery and data mining}
#' (pp. 3–12). Springer.
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' ## Numeric response example
#' fo <- sale_amount ~ .
#' control <- CVControl()
#'
#' gbm_res1 <- resample(fo, ICHomes, GBMModel(n.trees = 25), control)
#' gbm_res2 <- resample(fo, ICHomes, GBMModel(n.trees = 50), control)
#' gbm_res3 <- resample(fo, ICHomes, GBMModel(n.trees = 100), control)
#'
#' res <- c(GBM1 = gbm_res1, GBM2 = gbm_res2, GBM3 = gbm_res3)
#' res_diff <- diff(res)
#' t.test(res_diff)
#' }
#'
t.test.PerformanceDiff <- function(x, adjust = "holm", ...)
{
  vadj <- if (is(x@control, "CVControl")) 1 / (x@control@folds - 1) else 0

  t_test <- function(x) {
    x <- na.omit(x)
    n <- length(x)
    stat <- mean(x) / sqrt(var(x) * (1 / n + vadj))
    2 * pt(abs(stat), n - 1, lower.tail = FALSE)
  }

  pvalues <- x %>%
    apply(c(3, 2), t_test) %>%
    apply(2, p.adjust, method = adjust)
  meandiffs <- apply(x, c(3, 2), mean, na.rm = TRUE)

  model_names <- x@model_names
  num_models <- length(model_names)
  results <- array(
    NA, dim = c(num_models, num_models, size(x, 2)),
    dimnames = list(
      Model2 = model_names, Model1 = model_names, Metric = dimnames(x)[[2]]
    )
  )
  indices <- lower.tri(results[, , 1])
  results[indices] <- meandiffs
  results <- aperm(results, perm = c(2, 1, 3))
  results[indices] <- pvalues

  PerformanceDiffTest(results, adjust = adjust)
}

#' Recursive Feature Elimination
#'
#' A wrapper method of backward feature selection in which a given model is fit
#' to nested subsets of most important predictor variables in order to select
#' the subset whose resampled predictive performance is optimal.
#'
#' @name rfe
#' @rdname rfe-methods
#'
#' @param ... arguments passed from the generic function to its methods, from
#'   the \code{MLModel} and \code{MLModelFunction} methods to first arguments of
#'   others, and from others to the \code{ModelSpecification} method.  The
#'   first argument of each \code{fit} method is positional and, as such, must
#'   be given first in calls to them.
#' @param object model \link[=inputs]{input} or
#'   \link[=ModelSpecification]{specification}.
#' @param formula,data \link[=formula]{formula} defining the model predictor and
#'   response variables and a \link[=data.frame]{data frame} containing them.
#' @param x,y \link{matrix} and object containing predictor and response
#'   variables.
#' @param input \link[=inputs]{input} object defining and containing the model
#'   predictor and response variables.
#' @param model \link[=models]{model} function, function name, or object; or
#'   another object that can be \link[=as.MLModel]{coerced} to a model.  A model
#'   can be given first followed by any of the variable specifications.
#' @param select expression indicating predictor variables that can be
#'   eliminated (see \code{\link[base]{subset}} for syntax) [default: all].
#' @param control \link[=controls]{control} function, function name, or object
#'   defining the resampling method to be employed.
#' @param props numeric vector of the proportions of most important predictor
#'   variables to retain in fitted models or an integer number of equal spaced
#'   proportions to generate automatically; ignored if \code{sizes} are given.
#' @param sizes integer vector of the set sizes of most important predictor
#'   variables to retain.
#' @param random logical indicating whether to eliminate variables at random
#'   with probabilities proportional to their importance.
#' @param recompute logical indicating whether to recompute variable importance
#'   after eliminating each set of variables.
#' @param optimize character string specifying a search through all \code{props}
#'   to identify the globally optimal model (\code{"global"}) or a search that
#'   stops after identifying the first locally optimal model (\code{"local"}).
#' @param samples numeric vector or list giving the number of permutation
#'   samples for each of the \code{rfe} and \code{\link{varimp}} algorithms.
#'   One or both of the values may be specified as named arguments or in the
#'   order in which their defaults appear.  Larger numbers of samples decrease
#'   variability in estimated model performances and variable importances at the
#'   expense of increased computation time.  Samples are more expensive
#'   computationally for \code{rfe} than for \code{varimp}.
#' @param metrics \link[=metrics]{metric} function, function name, or vector of
#'   these with which to calculate performance.  If not specified, default
#'   metrics defined in the \link{performance} functions are used.
#' @param stat functions or character strings naming functions to compute
#'   summary statistics on resampled metric values and permuted samples.  One or
#'   both of the values may be specified as named arguments or in the order in
#'   which their defaults appear.
#' @param progress logical indicating whether to display iterative progress
#'   during elimination.
#'
#' @return \code{TrainingStep} class object containing a summary of the numbers
#' of predictor variables retained (size), their names (terms), logical
#' indicators for the optimal model selected (selected), and associated
#' performance metrics (metrics).
#'
#' @seealso \code{\link{performance}}, \code{\link{plot}},
#' \code{\link{summary}}, \code{\link{varimp}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' (res <- rfe(sale_amount ~ ., data = ICHomes, model = GBMModel))
#' summary(res)
#' summary(performance(res))
#' plot(res, type = "line")
#' }
#'
rfe <- function(...) {
  UseMethod("rfe")
}


#' @rdname rfe-methods
#'
rfe.formula <- function(formula, data, model, ...) {
  modelspec <- ModelSpecification(formula, data, model = model, control = NULL)
  rfe(modelspec, ..., .envir = parent.frame())
}


#' @rdname rfe-methods
#'
rfe.matrix <- function(x, y, model, ...) {
  modelspec <- ModelSpecification(x, y, model = model, control = NULL)
  rfe(modelspec, ..., .envir = parent.frame())
}


#' @rdname rfe-methods
#'
rfe.ModelFrame <- function(input, model, ...) {
  modelspec <- ModelSpecification(input, model = model, control = NULL)
  rfe(modelspec, ..., .envir = parent.frame())
}


#' @rdname rfe-methods
#'
rfe.recipe <- function(input, model, ...) {
  modelspec <- ModelSpecification(input, model = model, control = NULL)
  rfe(modelspec, ..., .envir = parent.frame())
}


#' @rdname rfe-methods
#'
rfe.ModelSpecification <- function(
  object, select = NULL, control = MachineShop::settings("control"), props = 4,
  sizes = integer(), random = FALSE, recompute = TRUE,
  optimize = c("global", "local"), samples = c(rfe = 1, varimp = 1),
  metrics = NULL,
  stat = c(
    resample = MachineShop::settings("stat.Resample"),
    permute = MachineShop::settings("stat.TrainingParams")
  ), progress = FALSE, ...
) {

  data <- as.data.frame(object)
  model_fit <- fit(object)
  control <- as.MLControl(control)
  optimize <- match.arg(optimize)

  method <- paste(optimize, "Recursive Feature Elimination")
  substring(method, 1, 1) <- toupper(substring(method, 1, 1))
  progress <- if (throw(check_logical(progress))) {
    heading(paste(class(object), "feature selection"))
    new_print_progress()
  } else {
    function(...) NULL
  }

  get_samples_args <- function() as.list(environment())
  formals(get_samples_args) <- eval(formals(rfe.ModelSpecification)$samples)
  samples <- do.call("get_samples_args", as.list(samples))
  samples$rfe <- check_integer(samples$rfe, bounds = c(1, Inf), size = 1)
  throw(check_assignment(samples$rfe))
  inds <- replicate(samples$rfe, permute_int(nrow(data))$j)

  times <- control@predict$times
  if (is.null(metrics)) {
    obs <- response(model_fit)
    pred <- predict(model_fit, times = times, type = "raw")
    metrics <- get_perf_metrics(obs, pred)
  }
  metric <- check_metric(c(metrics)[[1]], convert = TRUE)
  throw(check_assignment(metrics, metric))
  loss <- function(x) if (metric@maximize) -x[, 1] else x[, 1]

  get_stat_args <- function() as.list(environment())
  formals(get_stat_args) <- eval(formals(rfe.ModelSpecification)$stat)
  stat <- do.call("get_stat_args", as.list(c(stat)))
  stat$resample <- check_stat(stat$resample, convert = TRUE)
  throw(check_assignment(stat$resample))
  stat$permute <- check_stat(stat$permute, convert = TRUE)
  throw(check_assignment(stat$permute))
  apply_stat_permute <- function(x, along) {
    x <- do.call(abind, c(x, along = along))
    margins <- setdiff(seq_len(ndim(x)), along)
    apply(x, margins, function(x) stat$permute(na.omit(x)))
  }

  varimp <- function(object, select = NULL) {
    res <- do.call(MachineShop::varimp, list(
      object, scale = FALSE, method = "permute", select = select,
      samples = samples$varimp, times = times, metric = metric, compare = "-",
      stats = stat, progress = !is.null(body(progress))
    ))
    res <- structure(res[[1]], names = rownames(res))
    res[setdiff(select, names(res))] <- 0
    res
  }
  scale <- function(x) {
    scale <- max(x, 0)
    if (scale > 0) x <- x / scale
    pmin(pmax(sort(x, decreasing = TRUE), 0.01), 0.99)
  }

  keep <- character()
  superset <- all.vars(predictors(terms(as.MLInput(object), original = TRUE)))
  if (!missing(select)) {
    .envir <- list(...)$.envir
    if (is.null(.envir)) .envir <- parent.frame()
    select <- do.call(
      subset_names, list(superset, substitute(select)), envir = .envir
    )
    keep <- setdiff(superset, select)
    superset <- select
  }
  vi <- scale(varimp(model_fit, select = superset))

  if (length(sizes)) {
    sizes <- check_integer(sizes, bounds = c(1, Inf), size = NA)
    throw(check_assignment(sizes))
    sizes <- pmin(sizes, length(superset))
  } else {
    nprops <- check_integer(props, bounds = c(1, Inf), size = 1)
    if (is(nprops, "error")) {
      props <- check_numeric(props, bounds = c(0, 1), include = 0:1, size = NA)
      throw(check_assignment(props))
    } else {
      props <- seq(0, 1, length = nprops + 1)[-1]
    }
    sizes <- pmax(round(props * length(superset)), 1)
  }
  sizes <- sort(unique(sizes), decreasing = TRUE)

  subsets <- list()
  perf_list <- list()
  perf_stats <- NULL
  scores <- numeric()

  max_iter <- length(sizes)
  for (i in seq_len(max_iter)) {
    size <- sizes[i]
    if (random) {
      subset <- sample(names(vi), size, prob = vi)
      subset <- subset[order(vi[subset], decreasing = TRUE)]
    } else {
      subset <- head(names(vi), size)
    }
    drop <- setdiff(superset, subset)

    perf_samples <- list()
    vi_samples <- list()
    for (s in seq_len(samples$rfe)) {
      data[drop] <- data[inds[, s], drop]
      object <- update(object, data = data)
      res <- resample(object, control = control)
      perf_samples[[s]] <- performance(res, metrics = metrics)

      if (recompute && size > tail(sizes, 1)) {
        vi <- varimp(fit(object), select = subset)
        vi_samples[[s]] <- vi[subset]
      }
    }

    vi <- if (length(vi_samples)) {
      scale(apply_stat_permute(vi_samples, 2))
    } else {
      vi[subset]
    }

    subsets[[i]] <- c(keep, subset)
    perf <- perf_samples[[1]]
    if (s > 1) S3Part(perf) <- apply_stat_permute(perf_samples, 3)
    perf_list[[i]] <- perf
    perf_stats <- rbind(
      perf_stats,
      t(summary(perf, stat = stat$resample)[, 1, drop = FALSE])
    )
    scores[i] <- ifelse(metric@maximize, 1, -1) * perf_stats[i, 1]

    progress(
      scores, max_iter = max_iter, method = method,
      items = list("Term{?s}: " = subset),
      metric = structure(perf_stats[i, 1], names = metric@label)
    )

    check_local_perfs <- optimize == "local" && i > 1
    if (check_local_perfs && diff(tail(scores, 2)) < 0) break
  }

  names(perf_list) <- make_unique(rep_len(class(object), length(perf_list)))

  TrainingStep(
    object,
    method = method,
    names = names(perf_list),
    items = tibble(terms = subsets),
    params = tibble(size = lengths(subsets)),
    metrics = perf_stats,
    selected = length(scores) - which.max(rev(scores)) + 1,
    performance = do.call(c, perf_list)
  )

}


#' @rdname rfe-methods
#'
rfe.MLModel <- function(model, ...) {
  rfe(..., model = model)
}


#' @rdname rfe-methods
#'
rfe.MLModelFunction <- function(model, ...) {
  rfe(as.MLModel(model), ...)
}

#' Support Vector Machine Models
#'
#' Fits the well known C-svc, nu-svc, (classification) one-class-svc (novelty)
#' eps-svr, nu-svr (regression) formulations along with native multi-class
#' classification formulations and the bound-constraint SVM formulations.
#'
#' @rdname SVMModel
#'
#' @param scaled logical vector indicating the variables to be scaled.
#' @param type type of support vector machine.
#' @param kernel kernel function used in training and predicting.
#' @param kpar list of hyper-parameters (kernel parameters).
#' @param C cost of constraints violation defined as the regularization term in
#'   the Lagrange formulation.
#' @param nu parameter needed for nu-svc, one-svc, and nu-svr.
#' @param epsilon parameter in the insensitive-loss function used for eps-svr,
#'   nu-svr and eps-bsvm.
#' @param prob.model logical indicating whether to calculate the scaling
#'   parameter of the Laplacian distribution fitted on the residuals of numeric
#'   response variables.  Ignored in the case of a factor response variable.
#' @param cache cache memory in MB.
#' @param tol tolerance of termination criterion.
#' @param shrinking whether to use the shrinking-heuristics.
#' @param sigma inverse kernel width used by the ANOVA, Bessel, and Laplacian
#'   kernels.
#' @param degree degree of the ANOVA, Bessel, and polynomial kernel functions.
#' @param order order of the Bessel function to be used as a kernel.
#' @param scale scaling parameter of the polynomial and hyperbolic tangent
#'   kernels as a convenient way of normalizing patterns without the need to
#'   modify the data itself.
#' @param offset offset used in polynomial and hyperbolic tangent kernels.
#' @param ... arguments passed to \code{SVMModel} from the other constructors.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \itemize{
#'       \item SVMModel: \code{NULL}
#'       \item SVMANOVAModel: \code{C}, \code{degree}
#'       \item SVMBesselModel: \code{C}, \code{order}, \code{degree}
#'       \item SVMLaplaceModel: \code{C}, \code{sigma}
#'       \item SVMLinearModel: \code{C}
#'       \item SVMPolyModel: \code{C}, \code{degree}, \code{scale}
#'       \item SVMRadialModel: \code{C}, \code{sigma}
#'     }
#'   }
#' }
#'
#' The kernel-specific constructor functions \code{SVMANOVAModel},
#' \code{SVMBesselModel}, \code{SVMLaplaceModel}, \code{SVMLinearModel},
#' \code{SVMPolyModel}, \code{SVMRadialModel}, \code{SVMSplineModel}, and
#' \code{SVMTanhModel} are special cases of \code{SVMModel} which automatically
#' set its \code{kernel} and \code{kpar} arguments.  These are called directly
#' in typical usage unless \code{SVMModel} is needed to specify a more general
#' model.
#' 
#' Default argument values and further model details can be found in the source
#' See Also link below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[kernlab]{ksvm}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' fit(sale_amount ~ ., data = ICHomes, model = SVMRadialModel)
#'
SVMModel <- function(
  scaled = TRUE, type = character(),
  kernel = c(
    "rbfdot", "polydot", "vanilladot", "tanhdot", "laplacedot", "besseldot",
    "anovadot", "splinedot"
  ),
  kpar = "automatic", C = 1, nu = 0.2, epsilon = 0.1, prob.model = FALSE,
  cache = 40, tol = 0.001, shrinking = TRUE
) {

  kernel <- match.arg(kernel)

  MLModel(

    name = "SVMModel",
    label = "Support Vector Machines",
    packages = "kernlab",
    response_types = c("factor", "numeric"),
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment()),

    fit = function(formula, data, weights, prob.model, ...) {
      if (is.factor(response(data))) prob.model <- TRUE
      eval_fit(
        data,
        formula = kernlab::ksvm(
          formula, data = data, prob.model = prob.model, na.action = na.pass,
          ...
        ),
        matrix = kernlab::ksvm(
          x, y, prob.model = prob.model, na.action = na.pass, ...
        )
      )
    },

    predict = function(object, newdata, .MachineShop, ...) {
      y <- response(.MachineShop$input)
      newdata <- as.data.frame(newdata)
      kernlab::predict(
        object, newdata = newdata,
        type = if (is.factor(y)) "probabilities" else "response"
      )
    }

  )

}

MLModelFunction(SVMModel) <- NULL


#' @rdname SVMModel
#'
SVMANOVAModel <- function(sigma = 1, degree = 1, ...) {
  .SVMModel(
    name = "SVMANOVAModel", label = "Support Vector Machines (ANOVA)",
    model = "anovadot", envir = environment(), ...
  )
}

MLModelFunction(SVMANOVAModel) <- NULL


#' @rdname SVMModel
#'
SVMBesselModel <- function(sigma = 1, order = 1, degree = 1, ...) {
  .SVMModel(
    name = "SVMBesselModel", label = "Support Vector Machines (Bessel)",
    model = "besseldot", envir = environment(), ...
  )
}

MLModelFunction(SVMBesselModel) <- NULL


#' @rdname SVMModel
#'
SVMLaplaceModel <- function(sigma = numeric(), ...) {
  .SVMModel(
    name = "SVMLaplaceModel", label = "Support Vector Machines (Laplace)",
    model = "laplacedot", envir = environment(), ...
  )
}

MLModelFunction(SVMLaplaceModel) <- NULL


#' @rdname SVMModel
#'
SVMLinearModel <- function(...) {
  .SVMModel(
    name = "SVMLinearModel", label = "Support Vector Machines (Linear)",
    model = "vanilladot", envir = environment(), ...
  )
}

MLModelFunction(SVMLinearModel) <- NULL


#' @rdname SVMModel
#'
SVMPolyModel <- function(degree = 1, scale = 1, offset = 1, ...) {
  .SVMModel(
    name = "SVMPolyModel", label = "Support Vector Machines (Poly)",
    model = "polydot", envir = environment(), ...
  )
}

MLModelFunction(SVMPolyModel) <- NULL


#' @rdname SVMModel
#'
SVMRadialModel <- function(sigma = numeric(), ...) {
  .SVMModel(
    name = "SVMRadialModel", label = "Support Vector Machines (Radial)",
    model = "rbfdot", envir = environment(), ...
  )
}

MLModelFunction(SVMRadialModel) <- NULL


#' @rdname SVMModel
#'
SVMSplineModel <- function(...) {
  .SVMModel(
    name = "SVMSplineModel", label = "Support Vector Machines (Spline)",
    model = "splinedot", envir = environment(), ...
  )
}

MLModelFunction(SVMSplineModel) <- NULL


#' @rdname SVMModel
#'
SVMTanhModel <- function(scale = 1, offset = 1, ...) {
  .SVMModel(
    name = "SVMTanhModel", label = "Support Vector Machines (Tanh)",
    model = "tanhdot", envir = environment(), ...
  )
}

MLModelFunction(SVMTanhModel) <- NULL


.SVMModel <- function(name, label, model, envir, ...) {
  params <- list(...)
  params$kernel <- model
  kpar <- new_params(envir)
  params$kpar <- if (
    params$kernel %in% c("laplacedot", "rbfdot") && is_empty(kpar)
  ) {
    "automatic"
  } else {
    as.call(c(.(list), kpar))
  }
  model <- do.call(SVMModel, params, quote = TRUE)
  model@name <- name
  model@label <- label

  scaled <- model@params$scaled
  if (!is.logical(scaled)) scaled <- TRUE

  gridinfo <- new_gridinfo(
    param = c("C", "degree", "order", "scale", "sigma"),
    get_values = c(
      function(n, ...) 2^seq_range(-4, 2, c(-4, 10), n),
      function(n, ...) seq_len(min(n, 3)),
      function(n, ...) seq_len(min(n, 3)),
      function(n, ...) 10^seq_range(-4, 2, c(-4, log10(2)), n),
      function(n, data, ...) {
        sigmas <- kernlab::sigest(
          model.matrix(data, intercept = FALSE), scaled = scaled
        )
        exp(seq(log(min(sigmas)), log(max(sigmas)), length = n))
      }
    )
  )
  grid_params <- switch(params$kernel,
    "anovadot" = c("C", "degree"),
    "besseldot" = c("C", "order", "degree"),
    "laplacedot" = c("C", "sigma"),
    "polydot" = c("C", "degree", "scale"),
    "rbfdot" = c("C", "sigma"),
    "vanilladot" = "C"
  )
  model@gridinfo <- gridinfo[gridinfo$param %in% grid_params, ]

  model
}

#' Generalized Linear Model
#'
#' Fits generalized linear models, specified by giving a symbolic description of
#' the linear predictor and a description of the error distribution.
#'
#' @rdname GLMModel
#'
#' @param family optional error distribution and link function to be used in the
#'   model.  Set automatically according to the class type of the response
#'   variable.
#' @param quasi logical indicator for over-dispersion of binomial and Poisson
#'   families; i.e., dispersion parameters not fixed at one.
#' @param ... arguments passed to \code{\link[stats]{glm.control}}.
#'
#' @details
#' \describe{
#'   \item{\code{GLMModel} Response types:}{\code{BinomialVariate},
#'     \code{factor}, \code{matrix}, \code{NegBinomialVariate},
#'     \code{numeric}, \code{PoissonVariate}}
#'   \item{\code{GLMStepAICModel} Response types:}{\code{binary factor},
#'     \code{BinomialVariate}, \code{NegBinomialVariate}, \code{numeric},
#'     \code{PoissonVariate}}
#' }
#'
#' Default argument values and further model details can be found in the source
#' See Also links below.
#'
#' In calls to \code{\link{varimp}} for \code{GLMModel} and
#' \code{GLMStepAICModel}, numeric argument \code{base} may be specified for the
#' (negative) logarithmic transformation of p-values [defaul: \code{exp(1)}].
#' Transformed p-values are automatically scaled in the calculation of variable
#' importance to range from 0 to 100.  To obtain unscaled importance values, set
#' \code{scale = FALSE}.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[stats]{glm}}, \code{\link[stats]{glm.control}},
#' \code{\link[MASS]{stepAIC}}, \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' fit(sale_amount ~ ., data = ICHomes, model = GLMModel)
#'
GLMModel <- function(family = NULL, quasi = FALSE, ...) {

  MLModel(

    name = "GLMModel",
    label = "Generalized Linear Models",
    packages = c("MASS", "nnet", "stats"),
    response_types = c(
      "BinomialVariate", "factor", "matrix", "NegBinomialVariate",
      "numeric", "PoissonVariate"
    ),
    weights = TRUE,
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment(), ...),

    fit = function(formula, data, weights, family = NULL, quasi, ...) {
      if (is.null(family)) {
        quasi_prefix <- function(x) if (quasi) paste0("quasi", x) else x
        y <- response(data)
        family <- switch_class(y,
          "BinomialVariate" = quasi_prefix("binomial"),
          "factor" = if (nlevels(y) <= 2) {
            quasi_prefix("binomial")
          } else {
            "multinom"
          },
          "matrix" = "mgaussian",
          "NegBinomialVariate" = "negbin",
          "numeric" = "gaussian",
          "PoissonVariate" = quasi_prefix("poisson")
        )
      }
      data <- as.data.frame(formula, data = data)
      control <- stats::glm.control(...)
      if (identical(family, "mgaussian")) {
        stats::lm(formula, data = data, weights = weights)
      } else if (identical(family, "multinom")) {
        nnet::multinom(
          formula, data = data, weights = weights, na.action = na.pass,
          maxit = 4 * control$maxit,
          trace = control$trace, reltol = control$epsilons
        )
      } else if (identical(family, "negbin")) {
        res <- MASS::glm.nb(
          formula, data = data, weights = weights, na.action = na.pass,
          control = control
        )
        MASS::glm.convert(res)
      } else {
        stats::glm(
          formula, family = family, data = data, weights = weights,
          na.action = na.pass, control = control
        )
      }
    },

    predict = function(object, newdata, ...) {
      newdata <- as.data.frame(newdata)
      if (is(object, "multinom")) {
        rbind(predict(object, newdata = newdata, type = "probs"))
      } else {
        predict(object, newdata = newdata, type = "response")
      }
    },

    varimp = function(object, base = exp(1), ...) {
      varimp_pval(object, base = base)
    }

  )

}

MLModelFunction(GLMModel) <- NULL


#' @rdname GLMModel
#'
#' @param direction mode of stepwise search, can be one of \code{"both"}
#'   (default), \code{"backward"}, or \code{"forward"}.
#' @param scope defines the range of models examined in the stepwise search.
#'   This should be a list containing components \code{upper} and \code{lower},
#'   both formulae.
#' @param k multiple of the number of degrees of freedom used for the penalty.
#'   Only \code{k = 2} gives the genuine AIC; \code{k = .(log(nobs))} is
#'   sometimes referred to as BIC or SBC.
#' @param trace if positive, information is printed during the running of
#'   \code{stepAIC}. Larger values may give more information on the fitting
#'   process.
#' @param steps maximum number of steps to be considered.
#'
GLMStepAICModel <- function(
  family = NULL, quasi = FALSE, ...,
  direction = c("both", "backward", "forward"), scope = list(), k = 2,
  trace = FALSE, steps = 1000
) {

  direction <- match.arg(direction)

  params <- new_params(environment())
  stepmodel <- GLMModel(family = family, quasi = quasi, ...)
  params <- params[setdiff(names(params), names(stepmodel@params))]

  MLModel(

    name = "GLMStepAICModel",
    label = "Generalized Linear Models (Stepwise)",
    packages = stepmodel@packages,
    response_types = c(
      "binary", "BinomialVariate","NegBinomialVariate", "numeric",
      "PoissonVariate"
    ),
    weights = TRUE,
    predictor_encoding = stepmodel@predictor_encoding,
    na.rm = stepmodel@na.rm,
    params = c(stepmodel@params, params),

    fit = function(
      formula, data, weights, family = NULL, quasi, ...,
      direction, scope = list(), k, trace, steps
    ) {
      if (is.null(family)) {
        quasi_prefix <- function(x) if (quasi) paste0("quasi", x) else x
        family <- switch_class(response(data),
          "BinomialVariate" = quasi_prefix("binomial"),
          "factor" = quasi_prefix("binomial"),
          "NegBinomialVariate" = "negbin",
          "numeric" = "gaussian",
          "PoissonVariate" = quasi_prefix("poisson")
        )
      }
      data <- as.data.frame(formula, data = data)
      stepargs <- stepAIC_args(formula, direction, scope)
      control <- stats::glm.control(...)
      if (family == "negbin") {
        res <- MASS::stepAIC(
          MASS::glm.nb(
            stepargs$formula, data = data, weights = weights,
            na.action = na.pass, control = control
          ),
          direction = direction, scope = stepargs$scope, k = k, trace = trace,
          steps = steps
        )
        MASS::glm.convert(res)
      } else {
        MASS::stepAIC(
          stats::glm(
            stepargs$formula, family = family, data = data, weights = weights,
            na.action = na.pass, control = control
          ),
          direction = direction, scope = stepargs$scope, k = k, trace = trace,
          steps = steps
        )
      }
    },

    predict = stepmodel@predict,

    varimp = stepmodel@varimp

  )

}

MLModelFunction(GLMStepAICModel) <- NULL

#' Variable Selection by Filtering
#'
#' Creates a \emph{specification} of a recipe step that will select variables
#' from a candidate set according to a user-specified filtering function.
#'
#' @inheritParams step_lincomp
#' @param filter function whose first argument \code{x} is a univariate vector
#'   or a \code{multivariate} data frame of candidate variables from which to
#'   select, second argument \code{y} is the response variable as
#'   defined in preceding recipe steps, and third argument \code{step} is the
#'   current step.  The function should return a logical value or vector of
#'   length equal the number of variables in \code{x} indicating whether to
#'   select the corresponding variable, or return a list or data frame with
#'   element \code{`selected`} containing the logical(s) and possibly with other
#'   elements of the same length to be included in output from the \code{tidy}
#'   method.
#' @param multivariate logical indicating that candidate variables be passed to
#'   the \code{x} argument of the \code{filter} function separately as
#'   univariate vectors if \code{FALSE}, or altogether in one multivariate data
#'   frame if \code{TRUE}.
#' @param options list of elements to be added to the step object for use in the
#'   \code{filter} function.
#' @param prefix if the original variables are not replaced, the selected
#'   variables are added to the dataset with the character string prefix added
#'   to their names; otherwise, the original variable names are retained.
#' @param x \code{step_sbf} object.
#'
#' @return An updated version of \code{recipe} with the new step added to the
#' sequence of existing steps (if any).  For the \code{tidy} method, a tibble
#' with columns \code{terms} (selectors or variables selected), \code{selected}
#' (logical indicator of selected variables), and \code{name} of the selected
#' variable names.
#'
#' @seealso \code{\link[recipes]{recipe}}, \code{\link[recipes]{prep}},
#' \code{\link[recipes]{bake}}
#'
#' @examples
#' library(recipes)
#'
#' glm_filter <- function(x, y, step) {
#'   model_fit <- glm(y ~ ., data = data.frame(y, x))
#'   p_value <- drop1(model_fit, test = "F")[-1, "Pr(>F)"]
#'   p_value < step$threshold
#' }
#'
#' rec <- recipe(rating ~ ., data = attitude)
#' sbf_rec <- rec %>%
#'   step_sbf(all_numeric_predictors(),
#'            filter = glm_filter, options = list(threshold = 0.05))
#'
#' sbf_prep <- prep(sbf_rec, training = attitude)
#' sbf_data <- bake(sbf_prep, attitude)
#'
#' pairs(sbf_data, lower.panel = NULL)
#'
#' tidy(sbf_rec, number = 1)
#' tidy(sbf_prep, number = 1)
#'
step_sbf <- function(
  recipe, ..., filter, multivariate = FALSE, options = list(), replace = TRUE,
  prefix = "SBF", role = "predictor", skip = FALSE, id = recipes::rand_id("sbf")
) {

  recipes::add_step(recipe, new_step_sbf(
    terms = recipes::ellipse_check(...),
    filter = filter,
    multivariate = multivariate,
    options = options,
    replace = replace,
    prefix = prefix,
    role = role,
    skip = skip,
    id = id
  ))

}


new_step_sbf <- function(
  terms, filter, multivariate, options, replace, prefix, role, skip, id
) {
  stopifnot(is.function(filter))
  stopifnot(is.list(options))
  step_args <- list(
    subclass = "sbf",
    terms = terms,
    filter = filter,
    multivariate = multivariate,
    res = tibble(
      terms = recipes::sel2char(terms),
      selected = NA,
      name = NA_character_
    ),
    trained = FALSE,
    replace = replace,
    prefix = prefix,
    role = role,
    skip = skip,
    id = id
  )
  invalid_names <- intersect(names(options), names(step_args))
  if (length(invalid_names)) {
    throw(Error(note_items(
      "Options list contains reserved step name{?s}: ", invalid_names, "."
    )))
  }
  do.call(recipes::step, c(step_args, options))
}


prep.step_sbf <- function(x, training, info = NULL, ...) {
  col_names <- recipes::recipes_eval_select(x$terms, training, info)
  x_vars <- training[col_names]
  y_var <- response(terms.recipe_info(info), training)
  res <- if (x$multivariate) {
    x$filter(x = x_vars, y = y_var, step = x)
  } else {
    res_list <- map(x$filter, x = x_vars, y = list(y_var), step = list(x))
    if (all(map("logi", is.list, res_list))) {
      do.call(rbind, map(as_tibble, res_list))
    } else unlist(res_list)
  }
  if (!is.list(res)) res <- list(selected = res)
  if (!is.logical(res$selected)) {
    throw(Error("Filter function should return logical values."))
  }
  if (length(res$selected) != length(x_vars)) {
    throw(Error(
      "Filter function should return one logical value per variable."
    ))
  }
  x$res <- tibble(
    terms = col_names,
    as_tibble(res),
    name = ifelse(
      res$selected,
      paste0(if (!x$replace) paste0(x$prefix, "."), col_names),
      NA_character_
    )
  )
  x$trained <- TRUE
  x
}


bake.step_sbf <- function(object, new_data, ...) {
  res <- object$res
  if (object$replace) {
    names_drop <- res$terms[!res$selected]
    new_data[!(names(new_data) %in% names_drop)]
  } else {
    res <- res[res$selected, c("terms", "name")]
    selected_vars <- recipes::check_name(
      new_data[res$terms], new_data, object, newname = res$name
    )
    as_tibble(c(new_data, selected_vars))
  }
}


print.step_sbf <- function(x, width = console_width(), ...) {
  title <- paste(x$prefix, "selection by filtering for ")
  recipes::print_step(
    x$res$terms, x$terms, x$trained, title = title,
    width = max(width - nchar(title), 20)
  )
  invisible(x)
}


#' @rdname step_sbf
#'
tidy.step_sbf <- function(x, ...) {
  res <- x$res
  res$id <- x$id
  res
}

#' Selected Model
#'
#' Model selection from a candidate set.
#'
#' @rdname SelectedModel
#'
#' @param ... \link[=models]{model} functions, function names, objects; other
#'   objects that can be \link[=as.MLModel]{coerced} to models; vectors of
#'   these to serve as the candidate set from which to select, such as that
#'   returned by \code{\link{expand_model}}; or model
#'   \link[=ModelSpecification]{specifications}.
#' @param x list of models followed by arguments passed to their method
#'   function.
#' @param control \link[=controls]{control} function, function name, or object
#'   defining the resampling method to be employed.
#' @param metrics \link[=metrics]{metric} function, function name, or vector of
#'   these with which to calculate performance.  If not specified, default
#'   metrics defined in the \link{performance} functions are used.  Model
#'   selection is based on the first calculated metric.
#' @param cutoff argument passed to the \code{metrics} functions.
#' @param stat function or character string naming a function to compute a
#'   summary statistic on resampled metric values for model selection.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric}, \code{ordered},
#'     \code{Surv}}
#' }
#'
#' @return \code{SelectedModel} or \code{SelectedModelSpecification} class
#' object that inherits from \code{MLModel} or \code{ModelSpecification},
#' respectively.
#'
#' @seealso \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm and glmnet to run
#'
#' model_fit <- fit(
#'   sale_amount ~ ., data = ICHomes,
#'   model = SelectedModel(GBMModel, GLMNetModel, SVMRadialModel)
#' )
#' (selected_model <- as.MLModel(model_fit))
#' summary(selected_model)
#' }
#'
SelectedModel <- function(...) {
  UseMethod("SelectedModel")
}

MLModelFunction(SelectedModel) <- NULL


#' @rdname SelectedModel
#'
SelectedModel.default <- function(
  ..., control = MachineShop::settings("control"), metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
) {

  models <- as.list(unlist(list(...)))
  for (i in seq_along(models)) models[[i]] <- as.MLModel(models[[i]])

  if (length(models) == 1) return(models[[1]])

  default_names <- map("char", slot, models, "name")
  names(models) <- make_names_along(models, default_names)

  slots <- combine_model_slots(models, settings("response_types"))
  new("SelectedModel",
    MLModel(
      name = "SelectedModel",
      label = "Selected Model",
      response_types = slots$response_types,
      weights = slots$weights,
      params = TrainingParams(
        control = control,
        metrics = metrics,
        cutoff = cutoff,
        stat = stat
      )
    ),
    candidates = ListOf(models)
  )

}


#' @rdname SelectedModel
#'
SelectedModel.ModelSpecification <- function(
  ..., control = MachineShop::settings("control"), metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
) {
  do.call(SelectedInput, c(list(...), as.list(environment())))
}


#' @rdname SelectedModel
#'
SelectedModel.list <- function(x, ...) {
  do.call(SelectedModel, c(x, list(...)))
}


.fit.SelectedModel <- function(object, ...) {
  .fit_optim(object, ...)
}


update.SelectedModel <- function(object, params = NULL, ...) {
  object <- subset_selected(object, "candidates", params$id)
  params$id <- NULL
  NextMethod()
}


#' Tuned Model
#'
#' Model tuning over a grid of parameter values.
#'
#' @param object \link[=models]{model} function, function name, or object
#'   defining the model to be tuned.
#' @param grid single integer or vector of integers whose positions or names
#'   match the parameters in the model's pre-defined tuning grid if one exists
#'   and which specify the number of values used to construct the grid;
#'   \code{\link{TuningGrid}} function, function name, or object;
#'   \code{\link{ParameterGrid}} object; or \link[=data.frame]{data frame}
#'   containing parameter values at which to evaluate the model, such as that
#'   returned by \code{\link{expand_params}}.
#' @param control \link[=controls]{control} function, function name, or object
#'   defining the resampling method to be employed.
#' @param metrics \link[=metrics]{metric} function, function name, or vector of
#'   these with which to calculate performance.  If not specified, default
#'   metrics defined in the \link{performance} functions are used.  Model
#'   selection is based on the first calculated metric.
#' @param stat function or character string naming a function to compute a
#'   summary statistic on resampled metric values for model tuning.
#' @param cutoff argument passed to the \code{metrics} functions.
#'
#' @details
#' The \code{\link{expand_modelgrid}} function enables manual extraction and
#' viewing of grids created automatically when a \code{TunedModel} is fit.
#'
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric}, \code{ordered},
#'     \code{Surv}}
#' }
#'
#' @return \code{TunedModel} class object that inherits from \code{MLModel}.
#'
#' @seealso \code{\link{fit}}, \code{\link{resample}}, \code{\link{set_optim}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#' ## May require a long runtime
#'
#' # Automatically generated grid
#' model_fit <- fit(sale_amount ~ ., data = ICHomes,
#'                  model = TunedModel(GBMModel))
#' varimp(model_fit)
#' (tuned_model <- as.MLModel(model_fit))
#' summary(tuned_model)
#' plot(tuned_model, type = "l")
#'
#' # Randomly sampled grid points
#' fit(sale_amount ~ ., data = ICHomes,
#'     model = TunedModel(
#'       GBMModel,
#'       grid = TuningGrid(size = 1000, random = 5)
#'     ))
#'
#' # User-specified grid
#' fit(sale_amount ~ ., data = ICHomes,
#'     model = TunedModel(
#'       GBMModel,
#'       grid = expand_params(
#'         n.trees = c(50, 100),
#'         interaction.depth = 1:2,
#'         n.minobsinnode = c(5, 10)
#'       )
#'     ))
#' }
#'
TunedModel <- function(
  object, grid = MachineShop::settings("grid"),
  control = MachineShop::settings("control"), metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
) {

  if (missing(object)) {
    object <- NullModel()
    response_types <- settings("response_types")
    weights <- FALSE
  } else {
    object <- as.MLModel(object)
    response_types <- object@response_types
    weights <- object@weights
  }

  params <- names(formals(fget(object@name)))
  params_type <- "function"
  grid <- check_grid(grid)
  if (is(grid, "DomainError")) {
    value <- grid$value
    if (is(value, "data.frame")) {
      grid <- as_tibble(value)
      grid_params <- names(grid)
    } else {
      grid$message <- paste0(
        grid$message, "; a ParameterGrid object; or a data frame"
      )
      throw(check_assignment(grid))
    }
  } else if (is(grid, "ParameterGrid")) {
    grid_params <- grid$id
  } else {
    params <- object@gridinfo$param
    params_type <- "grid"
    grid_params <- names(grid@size)
    if (is.null(grid_params)) {
      if (length(grid@size) == 1) {
        if (!grid@random) params <- params[object@gridinfo$default]
        grid_params <- params
        grid@size <- rep(grid@size, max(length(grid_params), 1))
        names(grid@size) <- grid_params
      } else {
        grid_params <- params[seq_along(grid@size)]
        names(grid@size) <- grid_params
      }
    }
  }
  if (!("..." %in% params || all(grid_params %in% params))) {
    begin <- function(subject) paste0("x ", subject, " has parameter{?s}: ")
    throw(Error(
      "Unmatched tuning parameters.\n",
      note_items(begin(paste(object@name, params_type)), params, ".\n"),
      note_items(begin("Argument `grid`"), grid_params, ".")
    ))
  }

  new("TunedModel",
    MLModel(
      name = "TunedModel",
      label = paste("Grid Tuned", object@name),
      response_types = response_types,
      weights = weights,
      params = TrainingParams(
        control = control,
        metrics = metrics,
        cutoff = cutoff,
        stat = stat
      )
    ),
    model = object,
    grid = grid
  )

}

MLModelFunction(TunedModel) <- NULL


.fit.TunedModel <- function(object, ...) {
  .fit_optim(object, ...)
}


update.TunedModel <- function(
  object, params = NULL, ...
) {
  if (is.list(params)) {
    update(object@model, params = params, id = object@id)
  } else {
    object
  }
}

## Trained Inputs

test_that("testing of trained inputs", {
  skip_if_not(TEST_TRAINING)
  with_parallel({

    library(recipes)

    df <- ICHomes
    fo1 <- sale_amount ~ sale_year + built + style + construction
    fo2 <- sale_amount ~ sale_year + base_size + bedrooms + basement

    rec1 <- recipe(sale_amount ~ ., data = ICHomes)
    rec2 <- rec1 %>%
      step_center(all_numeric_predictors()) %>%
      step_scale(all_numeric_predictors()) %>%
      step_pca(all_numeric_predictors(), id = "pca")

    sel_fo <- SelectedInput(fo1, fo2, data = df)
    sel_mat <- SelectedInput(
      model.matrix(~ sale_year + built + style + construction, ICHomes)[, -1],
      model.matrix(~ sale_year + base_size + bedrooms + basement, ICHomes)[, -1],
      y = ICHomes$sale_amount
    )
    sel_mf <- SelectedInput(
      ModelFrame(fo1, data = df),
      ModelFrame(fo2, data = df)
    )
    sel_mo_mf <- SelectedInput(
      ModelSpecification(fo1, data = df, model = GLMModel),
      ModelSpecification(fo2, data = df, model = GBMModel)
    )
    sel_rec <- SelectedInput(rec1, rec2)
    sel_mo_rec <- SelectedInput(
      ModelSpecification(rec1, model = GLMModel),
      ModelSpecification(rec2, model = GBMModel)
    )
    tun_rec <- TunedInput(rec2, grid = expand_steps(pca = list(num_comp = 1:3)))


    models <- c(
      "GLMModel" = GLMModel,
      "TunedModel(GLMNet)" = TunedModel(GLMNetModel),
      "SelectedModel(SVMModel, RangerModel)" = SelectedModel(SVMModel, RangerModel)
    )


    for (name in names(models)) {

      model <- models[[name]]

      ## model fitting of selected model frames
      expect_is(fit(sel_fo, model = model), "MLModelFit")
      expect_is(fit(sel_mat, model = model), "MLModelFit")
      expect_is(fit(sel_mf, model = model), "MLModelFit")

      ## model fitting of selected recipes
      expect_is(fit(sel_rec, model = model), "MLModelFit")

      ## model fitting of tuned recipe
      expect_is(fit(tun_rec, model = model), "MLModelFit")

    }


    ## model fitting of selected inputs
    model <- models[[1]]
    expect_is(fit(sel_mo_mf), "MLModelFit")
    expect_is(fit(sel_mo_rec), "MLModelFit")


    ## resampling of selected inputs
    model <- models[[1]]
    expect_is(resample(sel_mf, model = model), "Resample")
    expect_is(resample(sel_mo_mf), "Resample")
    expect_is(resample(sel_rec, model = model), "Resample")
    expect_is(resample(sel_mo_rec), "Resample")


    ## resampling of tuned inputs
    model <- models[[1]]
    expect_is(resample(tun_rec, model = model), "Resample")

  })
})

#' Gradient Boosting with Regression Trees
#'
#' Gradient boosting for optimizing arbitrary loss functions where regression
#' trees are utilized as base-learners.
#'
#' @param family optional \code{\link[mboost]{Family}} object.  Set
#'   automatically according to the class type of the response variable.
#' @param mstop number of initial boosting iterations.
#' @param nu step size or shrinkage parameter between 0 and 1.
#' @param risk method to use in computing the empirical risk for each boosting
#'   iteration.
#' @param stopintern logical inidicating whether the boosting algorithm stops
#'   internally when the out-of-bag risk increases at a subsequent iteration.
#' @param trace logical indicating whether status information is printed during
#'   the fitting process.
#' @param teststat type of the test statistic to be applied for variable
#'   selection.
#' @param testtype how to compute the distribution of the test statistic.
#' @param mincriterion value of the test statistic or 1 - p-value that must be
#'   exceeded in order to implement a split.
#' @param minsplit minimum sum of weights in a node in order to be considered
#'   for splitting.
#' @param minbucket minimum sum of weights in a terminal node.
#' @param maxdepth maximum depth of the tree.
#' @param saveinfo logical indicating whether to store information about
#'   variable selection in \code{info} slot of each \code{partynode}.
#' @param ... additional arguments to \code{\link[partykit]{ctree_control}}.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{binary factor}, \code{BinomialVariate},
#'     \code{NegBinomialVariate}, \code{numeric}, \code{PoissonVariate},
#'     \code{Surv}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \code{mstop}, \code{maxdepth}
#'   }
#' }
#'
#' Default argument values and further model details can be found in the source
#' See Also links below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[mboost]{blackboost}}, \code{\link[mboost]{Family}},
#' \code{\link[partykit]{ctree_control}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested packages mboost and partykit to run
#'
#' data(Pima.tr, package = "MASS")
#'
#' fit(type ~ ., data = Pima.tr, model = BlackBoostModel)
#' }
#'
BlackBoostModel <- function(
  family = NULL, mstop = 100, nu = 0.1, risk = c("inbag", "oobag", "none"),
  stopintern = FALSE, trace = FALSE, teststat = c("quadratic", "maximum"),
  testtype = c("Teststatistic", "Univariate", "Bonferroni", "MonteCarlo"),
  mincriterion = 0, minsplit = 10, minbucket = 4, maxdepth = 2,
  saveinfo = FALSE, ...
) {

  risk <- match.arg(risk)
  teststat <- match.arg(teststat)
  testtype <- match.arg(testtype)

  MLModel(

    name = "BlackBoostModel",
    label = "Gradient Boosting with Regression Trees",
    packages = c("mboost", "partykit"),
    response_types = c(
      "binary", "BinomialVariate", "NegBinomialVariate", "numeric",
      "PoissonVariate", "Surv"
    ),
    weights = TRUE,
    predictor_encoding = "model.frame",
    na.rm = "response",
    params = new_params(environment(), ...),

    gridinfo = new_gridinfo(
      param = c("mstop", "maxdepth"),
      get_values = c(
        function(n, ...) round_int(seq_range(0, 50, c(1, 1000), n + 1)),
        function(n, ...) seq_len(min(n, 10))
      )
    ),

    fit = function(
      formula, data, weights, family = NULL, mstop, nu, risk, stopintern,
      trace, ...
    ) {
      if (is.null(family)) {
        family <- switch_class(response(data),
          "BinomialVariate" = mboost::Binomial(type = "glm"),
          "factor" = mboost::Binomial(),
          "NegBinomialVariate" = mboost::NBinomial(),
          "numeric" = mboost::Gaussian(),
          "PoissonVariate" = mboost::Poisson(),
          "Surv" = mboost::CoxPH()
        )
      }
      mboost::blackboost(
        formula, data = as.data.frame(formula, data = data), weights = weights,
        na.action = na.pass, family = family,
        control = mboost::boost_control(
          mstop = mstop, nu = nu, risk = risk, stopintern = stopintern,
          trace = trace
        ),
        tree_controls = partykit::ctree_control(...)
      )
    },

    predict = function(object, newdata, .MachineShop, ...) {
      weights <- case_weights(.MachineShop$input)
      newdata <- as.data.frame(newdata)
      if (object$family@name == "Cox Partial Likelihood") {
        lp <- drop(predict(object, type = "link"))
        new_lp <- drop(predict(object, newdata = newdata, type = "link"))
        predict(object$response, lp, new_lp, weights = weights, ...)
      } else {
        predict(object, newdata = newdata, type = "response")
      }
    },

    varimp = function(object, ...) {
      structure(mboost::varimp(object), class = "numeric", metric = "risk")
    }

  )

}

MLModelFunction(BlackBoostModel) <- NULL

knitr::opts_chunk$set(
  strip.white = FALSE,
  collapse = TRUE,
  comment = "#>",
  warning = FALSE,
  message = FALSE,
  fig.width = 7,
  fig.height = 4,
  fig.align = "center"
)


options(knitr.table.format = "html")


library(kableExtra)


rdoc_url <- function(names, rdoc_names = NULL) {
  version <- packageVersion("MachineShop")
  url <- paste0("https://www.rdocumentation.org/packages/MachineShop/versions",
              "/", version$major, ".", version$minor, ".0",
              "/topics/")

  if (is.null(rdoc_names)) {
    rdoc_names <- sapply(names, function(x) as.character(str2lang(x))[1])
  }

  paste0("[`", names, "`](", url, rdoc_names, ")")
}


median_range <- function(x, prefix = NULL, ...) {
  formatted_median <- format(median(x), trim = TRUE, ...)
  formatted_range <- format(range(x), trim = TRUE, ...)
  paste0(prefix, formatted_median,
         " (", toString(paste0(prefix, formatted_range)), ")")
}


n_perc <- function(x, digits = 2) {
  paste0(sum(x), " (", round(100 * mean(x), digits), "%)")
}


summary_kbl <- function(x, data) {
  vals <- sapply(unlist(unname(x), recursive = FALSE), function(el) {
    eval(el[[2]], envir = data)
  })

  kbl <- data.frame(Characteristic = names(vals), Value = vals) %>%
    kable(row.names = FALSE,
          align = c("l", "c")) %>%
    kable_styling(c("striped", "condensed"), full_width = FALSE,
                  position = "center")

  start_row <- 1
  for (i in seq(x)) {
    group_label <- names(x)[i]
    group_length <- length(x[[i]])
    if (nzchar(group_label)) {
      kbl <- kableExtra::group_rows(kbl, group_label, start_row,
                                    start_row + group_length - 1)
    }
    start_row <- start_row + group_length
  }

  kbl
}

#' @rdname metrics
#'
accuracy <- function(
  observed, predicted = NULL, weights = NULL,
  cutoff = MachineShop::settings("cutoff"), ...
) {
  call_metric_method("accuracy", environment())
}

MLMetric(accuracy) <- list("accuracy", "Accuracy", TRUE)


setMetric_ConfusionMatrix("accuracy",
  function(observed, predicted, ...) {
    sum(diag(observed)) / sum(observed)
  }
)


#' @rdname metrics
#'
auc <- function(
  observed, predicted = NULL, weights = NULL, multiclass = c("pairs", "all"),
  metrics = c(MachineShop::tpr, MachineShop::fpr),
  stat = MachineShop::settings("stat.Curve"), ...
) {
  multiclass <- match.arg(multiclass)
  metrics <- map(as.MLMetric, metrics)
  call_metric_method("auc", environment())
}

MLMetric(auc) <- list("auc", "Area Under Performance Curve", TRUE)


setMetricGeneric("auc")


setMetricMethod("auc", c("factor", "factor"))


setMetricMethod("auc", c("factor", "matrix"),
  function(observed, predicted, weights, multiclass, ...) {
    nzcounts <- table(observed) > 0
    auc_all <- function(observed, predicted, weights, select) {
      labels <- levels(observed)
      levels(observed) <- list("0" = labels[-select], "1" = labels[select])
      auc(observed, predicted[, select], weights, ...)
    }
    fun <- switch(multiclass,
      "all" = function(i) if (nzcounts[i]) {
        auc_all(observed, predicted, weights, i)
      },
      "pairs" = function(i) {
        res <- numeric()
        for (j in seq_len(i - 1)) {
          if (nzcounts[i] && nzcounts[j]) {
            subset <- observed %in% levels(observed)[c(j, i)]
            obs <- droplevels(observed[subset])
            pred <- predicted[subset, c(j, i)]
            wgt <- weights[subset]
            res <- c(res,
              auc_all(obs, pred, wgt, 1),
              auc_all(obs, pred, wgt, 2)
            )
          }
        }
        res
      }
    )
    mean(unlist(map(fun, seq(nlevels(observed)))))
  }
)


setMetricMethod("auc", c("factor", "numeric"),
  function(observed, predicted, weights, metrics, ...) {
    if (all(map("logi", identical, metrics[1:2], c(tpr, fpr)))) {
      cindex(observed, predicted, weights)
    } else {
      unname(
        auc(performance_curve(observed, predicted, weights, metrics = metrics))
      )
    }
  }
)


setMetricMethod("auc", c("PerformanceCurve", "NULL"),
  function(observed, predicted, stat, ...) {
    observed <- summary(observed, stat = stat)
    by(observed, observed["Model"], function(curve) {
      curve <- na.omit(curve)
      n <- nrow(curve)
      if (n) sum(diff(curve$x) * (curve$y[-n] + curve$y[-1]) / 2) else NA
    })
  }
)


setMetricMethod("auc", c("Resample", "NULL"),
  function(observed, predicted, weights, metrics, ...) {
    auc@.Data <- function(...) MachineShop::auc(..., metrics = metrics)
    performance(observed, metrics = auc, ...)
  }
)


setMetricMethod("auc", c("Surv", "SurvProbs"),
  function(observed, predicted, weights, metrics, ...) {
    x <- auc(performance_curve(observed, predicted, weights, metrics = metrics))
    names(x) <- colnames(predicted)
    survmetric_mean(x, predicted@times)
  }
)


#' @rdname metrics
#'
brier <- function(observed, predicted = NULL, weights = NULL, ...) {
  call_metric_method("brier", environment())
}

MLMetric(brier) <- list("brier", "Brier Score", FALSE)


setMetricGeneric("brier")


setMetricMethod("brier", c("factor", "factor"))


setMetricMethod("brier", c("factor", "matrix"),
  function(observed, predicted, weights, ...) {
    observed <- model.matrix(~ observed - 1)
    mse(observed, predicted, weights) * ncol(observed)
  }
)


setMetricMethod("brier", c("factor", "numeric"),
  function(observed, predicted, weights, ...) {
    mse(as.numeric(observed == levels(observed)[2]), predicted, weights)
  }
)


setMetricMethod_Resample("brier")


setMetricMethod("brier", c("Surv", "SurvProbs"),
  function(observed, predicted, weights, ...) {
    weights <- check_weights(weights, observed)
    throw(check_assignment(weights))
    times <- predicted@times

    observed[, "status"] <- 1 - observed[, "status"]
    cens_fit <- survfit(
      observed ~ 1, weights = weights, stype = 2, se.fit = FALSE
    )

    start_time <- function(x) if (is_counting(x)) x[, "start"] else -Inf
    x <- map("num", function(i) {
      time <- times[i]
      start_by <- start_time(observed) <= time
      stop_after <- time(observed) > time
      known_status <- start_by & (observed[, "status"] == 0 | stop_after)
      cens <- predict(cens_fit, pmin(time, time(observed)))
      cens_weights <- proportions(ifelse(known_status, weights / cens, 0))
      sum(cens_weights * (stop_after - predicted[, i, drop = TRUE])^2)
    }, seq_along(times))

    names(x) <- colnames(predicted)
    survmetric_mean(x, times)
  }
)


#' @rdname metrics
#'
cindex <- function(observed, predicted = NULL, weights = NULL, ...) {
  call_metric_method("cindex", environment())
}

MLMetric(cindex) <- list("cindex", "Concordance Index", TRUE)


setMetricGeneric("cindex")


setMetricMethod("cindex", c("factor", "factor"))


setMetricMethod("cindex", c("factor", "numeric"),
  function(observed, predicted, weights, ...) {
    concordance(observed ~ predicted, weights = weights)$concordance
  }
)


setMetricMethod_Resample("cindex")


setMetricMethod("cindex", c("Surv", "numeric"),
  function(observed, predicted, weights, ...) {
    concordance(observed ~ predicted, weights = weights)$concordance
  }
)


#' @rdname metrics
#'
cross_entropy <- function(observed, predicted = NULL, weights = NULL, ...) {
  call_metric_method("cross_entropy", environment())
}

MLMetric(cross_entropy) <- list("cross_entropy", "Cross Entropy", FALSE)


setMetricGeneric("cross_entropy")


setMetricMethod("cross_entropy", c("factor", "factor"))


setMetricMethod("cross_entropy", c("factor", "matrix"),
  function(observed, predicted, weights, ...) {
    observed <- model.matrix(~ observed - 1)
    eps <- 1e-15
    predicted <- pmin(pmax(predicted, eps), 1 - eps)
    n <- ncol(observed)
    -weighted_mean(observed * log(predicted), rep(weights, n)) * n
  }
)


setMetricMethod("cross_entropy", c("factor", "numeric"),
  function(observed, predicted, ...) {
    cross_entropy(observed, cbind(1 - predicted, predicted), ...)
  }
)


setMetricMethod_Resample("cross_entropy")


#' @rdname metrics
#'
f_score <- function(
  observed, predicted = NULL, weights = NULL,
  cutoff = MachineShop::settings("cutoff"), beta = 1, ...
) {
  call_metric_method("f_score", environment())
}

MLMetric(f_score) <- list("f_score", "F Score", TRUE)


setMetric_BinaryConfusionMatrix("f_score",
  function(observed, predicted, beta, ...) {
    beta2 <- beta^2
    (1 + beta2) * observed[2, 2] /
      ((1 + beta2) * observed[2, 2] + beta2 * observed[1, 2] + observed[2, 1])
  }
)


#' @rdname metrics
#'
fnr <- function(
  observed, predicted = NULL, weights = NULL,
  cutoff = MachineShop::settings("cutoff"), ...
) {
  call_metric_method("fnr", environment())
}

MLMetric(fnr) <- list("fnr", "False Negative Rate", FALSE)


setMetric_BinaryConfusionMatrix("fnr",
  function(observed, predicted, ...) {
    1 - tpr(observed)
  }
)


#' @rdname metrics
#'
fpr <- function(
  observed, predicted = NULL, weights = NULL,
  cutoff = MachineShop::settings("cutoff"), ...
) {
  call_metric_method("fpr", environment())
}

MLMetric(fpr) <- list("fpr", "False Positive Rate", FALSE)


setMetric_BinaryConfusionMatrix("fpr",
  function(observed, predicted, ...) {
    1 - tnr(observed)
  }
)


#' @rdname metrics
#'
kappa2 <- function(
  observed, predicted = NULL, weights = NULL,
  cutoff = MachineShop::settings("cutoff"), ...
) {
  call_metric_method("kappa2", environment())
}

MLMetric(kappa2) <- list("kappa2", "Cohen's Kappa", TRUE)


setMetric_ConfusionMatrix("kappa2",
  function(observed, predicted, ...) {
    p <- proportions(observed)
    1 - (1 - sum(diag(p))) / (1 - sum(rowSums(p) * colSums(p)))
  }
)


#' @rdname metrics
#'
npv <- function(
  observed, predicted = NULL, weights = NULL,
  cutoff = MachineShop::settings("cutoff"), ...
) {
  call_metric_method("npv", environment())
}

MLMetric(npv) <- list("npv", "Negative Predictive Value", TRUE)


setMetric_BinaryConfusionMatrix("npv",
  function(observed, predicted, ...) {
    observed[1, 1] / (observed[1, 1] + observed[1, 2])
  }
)


#' @rdname metrics
#'
ppr <- function(
  observed, predicted = NULL, weights = NULL,
  cutoff = MachineShop::settings("cutoff"), ...
) {
  call_metric_method("ppr", environment())
}

MLMetric(ppr) <- list("ppr", "Positive Prediction Rate", FALSE)


setMetric_BinaryConfusionMatrix("ppr",
  function(observed, predicted, ...) {
    (observed[2, 1] + observed[2, 2]) / sum(observed)
  }
)


#' @rdname metrics
#'
ppv <- function(
  observed, predicted = NULL, weights = NULL,
  cutoff = MachineShop::settings("cutoff"), ...
) {
  call_metric_method("ppv", environment())
}

MLMetric(ppv) <- list("ppv", "Positive Predictive Value", TRUE)


setMetric_BinaryConfusionMatrix("ppv",
  function(observed, predicted, ...) {
    observed[2, 2] / (observed[2, 1] + observed[2, 2])
  }
)


#' @rdname metrics
#'
pr_auc <- function(
  observed, predicted = NULL, weights = NULL, multiclass = c("pairs", "all"),
  ...
) {
  call_metric_method("pr_auc", environment())
}

MLMetric(pr_auc) <- list("pr_auc", "Area Under Precision-Recall Curve", TRUE)


setMetric_auc("pr_auc", c(precision, recall))


#' @rdname metrics
#'
precision <- function(
  observed, predicted = NULL, weights = NULL,
  cutoff = MachineShop::settings("cutoff"), ...
) {
  call_metric_method("precision", environment())
}

MLMetric(precision) <- list("precision", "Precision", TRUE)


setMetric_BinaryConfusionMatrix("precision",
  function(observed, predicted, ...) {
    ppv(observed)
  }
)


#' @rdname metrics
#'
recall <- function(
  observed, predicted = NULL, weights = NULL,
  cutoff = MachineShop::settings("cutoff"), ...
) {
  call_metric_method("recall", environment())
}

MLMetric(recall) <- list("recall", "Recall", TRUE)


setMetric_BinaryConfusionMatrix("recall",
  function(observed, predicted, ...) {
    tpr(observed)
  }
)


#' @rdname metrics
#'
roc_auc <- function(
  observed, predicted = NULL, weights = NULL, multiclass = c("pairs", "all"),
  ...
) {
  call_metric_method("roc_auc", environment())
}

MLMetric(roc_auc) <- list("roc_auc", "Area Under ROC Curve", TRUE)


setMetric_auc("roc_auc", c(tpr, fpr))


#' @rdname metrics
#'
roc_index <- function(
  observed, predicted = NULL, weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  fun = function(sensitivity, specificity) (sensitivity + specificity) / 2, ...
) {
  call_metric_method("roc_index", environment())
}

MLMetric(roc_index) <- list("roc_index", "ROC Index", TRUE)


setMetric_BinaryConfusionMatrix("roc_index",
  function(observed, predicted, fun, ...) {
    fun(sens = sensitivity(observed), spec = specificity(observed))
  }
)


#' @rdname metrics
#'
sensitivity <- function(
  observed, predicted = NULL, weights = NULL,
  cutoff = MachineShop::settings("cutoff"), ...
) {
  call_metric_method("sensitivity", environment())
}

MLMetric(sensitivity) <- list("sensitivity", "Sensitivity", TRUE)


setMetric_BinaryConfusionMatrix("sensitivity",
  function(observed, predicted, ...) {
    tpr(observed)
  }
)


#' @rdname metrics
#'
specificity <- function(
  observed, predicted = NULL, weights = NULL,
  cutoff = MachineShop::settings("cutoff"), ...
) {
  call_metric_method("specificity", environment())
}

MLMetric(specificity) <- list("specificity", "Specificity", TRUE)


setMetric_BinaryConfusionMatrix("specificity",
  function(observed, predicted, ...) {
    tnr(observed)
  }
)


#' @rdname metrics
#'
tnr <- function(
  observed, predicted = NULL, weights = NULL,
  cutoff = MachineShop::settings("cutoff"), ...
) {
  call_metric_method("tnr", environment())
}

MLMetric(tnr) <- list("tnr", "True Negative Rate", TRUE)


setMetric_BinaryConfusionMatrix("tnr",
  function(observed, predicted, ...) {
    observed[1, 1] / (observed[1, 1] + observed[2, 1])
  }
)


#' @rdname metrics
#'
tpr <- function(
  observed, predicted = NULL, weights = NULL,
  cutoff = MachineShop::settings("cutoff"), ...
) {
  call_metric_method("tpr", environment())
}

MLMetric(tpr) <- list("tpr", "True Positive Rate", TRUE)


setMetric_BinaryConfusionMatrix("tpr",
  function(observed, predicted, ...) {
    observed[2, 2] / (observed[1, 2] + observed[2, 2])
  }
)


#' @rdname metrics
#'
weighted_kappa2 <- function(
  observed, predicted = NULL, weights = NULL, power = 1, ...
) {
  call_metric_method("weighted_kappa2", environment())
}

MLMetric(weighted_kappa2) <-
  list("weighted_kappa2", "Weighted Cohen's Kappa", TRUE)


setMetric_OrderedConfusionMatrix("weighted_kappa2",
  function(observed, predicted, power, ...) {
    expected <- (rowSums(observed) %o% colSums(observed)) / sum(observed)
    weights <- abs(row(observed) - col(observed))^power
    1 - sum(weights * observed) / sum(weights * expected)
  }
)

#' Model Lift Curves
#'
#' Calculate lift curves from observed and predicted responses.
#'
#' @name lift
#' @rdname lift
#'
#' @param x \link[=response]{observed responses} or \link{resample} result
#'   containing observed and predicted responses.
#' @param y \link[=predict]{predicted responses} if not contained in \code{x}.
#' @param weights numeric vector of non-negative
#'   \link[=case_weights]{case weights} for the observed \code{x} responses
#'   [default: equal weights].
#' @param na.rm logical indicating whether to remove observed or predicted
#'   responses that are \code{NA} when calculating metrics.
#' @param ... arguments passed to other methods.
#'
#' @return \code{LiftCurve} class object that inherits from
#' \code{PerformanceCurve}.
#'
#' @seealso \code{\link{c}}, \code{\link{plot}}, \code{\link{summary}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' data(Pima.tr, package = "MASS")
#'
#' res <- resample(type ~ ., data = Pima.tr, model = GBMModel)
#' lf <- lift(res)
#' plot(lf)
#' }
#'
lift <- function(x, y = NULL, weights = NULL, na.rm = TRUE, ...) {
  as(
    performance_curve(x, y, weights, metrics = c(tpr, ppr), na.rm = na.rm),
    "LiftCurve"
  )
}


LiftCurve <- function(...) {
  object <- as(PerformanceCurve(...), "LiftCurve")
  if (!all(map("logi", identical, object@metrics, c(tpr, ppr)))) {
    throw(Error("Incorrect LiftCurve metrics."))
  }
  object
}


#' Model Performance Curves
#'
#' Calculate curves for the analysis of tradeoffs between metrics for assessing
#' performance in classifying binary outcomes over the range of possible
#' cutoff probabilities.  Available curves include receiver operating
#' characteristic (ROC) and precision recall.
#'
#' @name performance_curve
#' @aliases curves
#'
#' @param x \link[=response]{observed responses} or \link{resample} result
#'   containing observed and predicted responses.
#' @param y \link[=predict]{predicted responses} if not contained in \code{x}.
#' @param weights numeric vector of non-negative
#'   \link[=case_weights]{case weights} for the observed \code{x} responses
#'   [default: equal weights].
#' @param metrics list of two performance \link{metrics} for the analysis
#'   [default: ROC metrics].  Precision recall curves can be obtained with
#'   \code{c(precision, recall)}.
#' @param na.rm logical indicating whether to remove observed or predicted
#'   responses that are \code{NA} when calculating metrics.
#' @param ... arguments passed to other methods.
#'
#' @return \code{PerformanceCurve} class object that inherits from
#' \code{data.frame}.
#'
#' @seealso \code{\link{auc}}, \code{\link{c}}, \code{\link{plot}},
#' \code{\link{summary}}
#'
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' data(Pima.tr, package = "MASS")
#'
#' res <- resample(type ~ ., data = Pima.tr, model = GBMModel)
#'
#' ## ROC curve
#' roc <- performance_curve(res)
#' plot(roc)
#' auc(roc)
#' }
#'
performance_curve <- function(x, ...) {
  UseMethod("performance_curve")
}


#' @rdname performance_curve
#'
performance_curve.default <- function(
  x, y, weights = NULL, metrics = c(MachineShop::tpr, MachineShop::fpr),
  na.rm = TRUE, ...
) {
  if (na.rm) {
    complete <- complete_subset(x = x, y = y, weights = weights)
    x <- complete$x
    y <- complete$y
    weights <- complete$weights
  }
  metrics <- get_curve_metrics(metrics)
  curve <- .performance_curve(x, y, weights, metrics = metrics)
  if (is(curve, "listof")) do.call(c, curve) else curve
}


#' @rdname performance_curve
#'
performance_curve.Resample <- function(
  x, metrics = c(MachineShop::tpr, MachineShop::fpr), na.rm = TRUE, ...
) {
  if (na.rm) x <- na.omit(x)
  metrics <- get_curve_metrics(metrics)

  curves <- NULL
  for (model in unique(x$Model)) {
    for (iter in unique(x$Iteration)) {
      df <- x[x$Model == model & x$Iteration == iter, ]
      curve <- .performance_curve(
        df$Observed, df$Predicted, df$Weight, metrics = metrics
      )
      curve <- if (is(curve, "listof")) {
        structure(curve, names = paste0(model, ".", names(curve)))
      } else {
        structure(list(curve), names = model)
      }
      curve <- do.call(c, curve)
      curve$Iteration <- iter
      curves <- rbind(curves, curve)
    }
  }

  PerformanceCurve(curves, metrics = curve@metrics, control = x@control)
}


get_curve_metrics <- function(metrics) {
  metrics <- map(fget, metrics)
  if (length(metrics) != 2 || !all(map("logi", is, metrics, "MLMetric"))) {
    metrics <- Error("Value must be a list of two performance metrics.")
    throw(check_assignment(metrics))
  }
  metrics
}


PerformanceCurve <- function(
  object, ..., metrics, control = NullControl(), .check = TRUE
) {
  if (.check) {
    if (is.null(object$Model)) object$Model <- factor("Model")
    missing <- missing_names(c("Cutoff", "x", "y"), object)
    if (length(missing)) {
      throw(Error(note_items(
        "Missing performance curve variable{?s}: ", missing, "."
      )))
    }

    if (!all(map("logi", is, metrics[1:2], "MLMetric"))) {
      throw(Error(
        "Missing performance metrics in PerformanceCurve constructor."
      ))
    }
    metrics <- c(y = metrics[[1]], x = metrics[[2]])

    decreasing <- !xor(metrics$x@maximize, metrics$y@maximize)
    sort_order <- order(
      object$Model, object$x, object$y,
      decreasing = c(FALSE, FALSE, decreasing), method = "radix"
    )
    object <- object[sort_order, , drop = FALSE]
  }

  rownames(object) <- NULL
  new("PerformanceCurve", object, metrics = metrics, control = control, ...)
}


setGeneric(".performance_curve",
  function(observed, predicted, ...) standardGeneric(".performance_curve")
)


setMethod(".performance_curve", c("ANY", "ANY"),
  function(observed, predicted, ...) {
    throw(Error(
      "Performance_curve requires a predicted binary factor or survival ",
      "probabilities."
    ))
  }
)


setMethod(".performance_curve", c("factor", "numeric"),
  function(observed, predicted, weights, metrics, ...) {
    cutoffs <- c(-Inf, unique(predicted))
    x <- y <- numeric(length(cutoffs))
    for (i in seq_along(cutoffs)) {
      conf <- confusion(observed, predicted, weights, cutoff = cutoffs[i])
      x[i] <- metrics[[2]](conf)
      y[i] <- metrics[[1]](conf)
    }
    PerformanceCurve(
      data.frame(Cutoff = cutoffs, x = x, y = y), metrics = metrics
    )
  }
)


setMethod(".performance_curve", c("Surv", "SurvProbs"),
  function(observed, predicted, weights, metrics, ...) {

    weights <- check_weights(weights, observed)
    throw(check_assignment(weights))

    times <- predicted@times
    conf <- ConfusionMatrix(table(Predicted = 0:1, Observed = 0:1))

    structure(
      map(function(i) {
        time <- times[i]
        pred <- predicted[, i, drop = TRUE]
        cutoffs <- c(-Inf, unique(pred))
        x <- y <- numeric(length(cutoffs))

        for (j in seq_along(cutoffs)) {
          pos <- pred <= cutoffs[j]
          pos_pred <- surv_subset(observed, weights, pos, time)
          neg_pred <- surv_subset(observed, weights, !pos, time)

          conf[1, 1] <- neg_pred$surv * neg_pred$p
          conf[2, 1] <- pos_pred$surv * pos_pred$p
          conf[1, 2] <- (1 - neg_pred$surv) * neg_pred$p
          conf[2, 2] <- (1 - pos_pred$surv) * pos_pred$p

          x[j] <- metrics[[2]](conf)
          y[j] <- metrics[[1]](conf)
        }

        PerformanceCurve(
          data.frame(Cutoff = cutoffs, x = x, y = y), metrics = metrics
        )
      }, seq_along(times)),
      names = colnames(predicted),
      class = "listof"
    )

  }
)

#' Quadratic Discriminant Analysis Model
#'
#' Performs quadratic discriminant analysis.
#'
#' @param prior prior probabilities of class membership if specified or the
#'   class proportions in the training set otherwise.
#' @param method type of mean and variance estimator.
#' @param nu degrees of freedom for \code{method = "t"}.
#' @param use type of parameter estimation to use for prediction.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}}
#' }
#'
#' The \code{\link{predict}} function for this model additionally accepts the
#' following argument.
#' \describe{
#'   \item{\code{prior}}{prior class membership probabilities for prediction data
#'     if different from the training set.}
#' }
#'
#' Default argument values and further model details can be found in the source
#' See Also links below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[MASS]{qda}}, \code{\link[MASS]{predict.qda}},
#' \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' fit(Species ~ ., data = iris, model = QDAModel)
#'
QDAModel <- function(
  prior = numeric(), method = c("moment", "mle", "mve", "t"), nu = 5,
  use = c("plug-in", "predictive", "debiased", "looCV")
) {

  method <- match.arg(method)
  use <- match.arg(use)

  MLModel(

    name = "QDAModel",
    label = "Quadratic Discriminant Analysis",
    packages = "MASS",
    response_types = "factor",
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment()),

    fit = function(formula, data, weights, use, ...) {
      res <- eval_fit(
        data,
        formula = MASS::qda(formula, data = data, na.action = na.pass, ...),
        matrix = MASS::qda(x, y, ...)
      )
      attr(res, ".MachineShop") <- list(use = use)
      res
    },

    predict = function(
      object, newdata, prior = object$prior, .MachineShop, ...
    ) {
      newdata <- as.data.frame(newdata)
      predict(
        object, newdata = newdata, prior = prior, method = .MachineShop$use
      )$posterior
    }

  )

}

MLModelFunction(QDAModel) <- NULL

#' Model Specification
#'
#' Specification of a relationship between response and predictor variables and
#' a model to define a relationship between them.
#'
#' @name ModelSpecification
#' @rdname ModelSpecification-methods
#'
#' @param ... arguments passed from the generic function to its methods.  The
#'   first argument of each \code{ModelSpecification} method is positional and,
#'   as such, must be given first in calls to them.
#' @param formula,data \link[=formula]{formula} defining the model predictor and
#'   response variables and a \link[=data.frame]{data frame} containing them.
#' @param x,y \link{matrix} and object containing predictor and response
#'   variables.
#' @param input \link[=inputs]{input} object defining and containing the model
#'   predictor and response variables.
#' @param model \link[=models]{model} function, function name, or object; or
#'   another object that can be \link[=as.MLModel]{coerced} to a model.
#' @param control \link[=controls]{control} function, function name, or object
#'   defining the resampling method to be employed.  If \code{NULL} or if
#'   the model specification contains any \code{SelectedInput} or
#'   \code{SelectedModel} objects, then object-specific control structures and
#'   training parameters are used for selection and tuning, as usual, and
#'   objects are trained sequentially with nested resampling.  Otherwise,
#'   \itemize{
#'     \item tuning of input and model objects is performed simultaneously over
#'       a global grid of their parameter values, and
#'     \item the specified \code{control} method and training parameters below
#'       override those of any included \code{TunedInput} or \code{TunedModel}.
#'   }
#' @param metrics \link[=metrics]{metric} function, function name, or vector of
#'   these with which to calculate performance.  If not specified, default
#'   metrics defined in the \link{performance} functions are used.  Model
#'   selection is based on the first calculated metric.
#' @param cutoff argument passed to the \code{metrics} functions.
#' @param stat function or character string naming a function to compute a
#'   summary statistic on resampled metric values for model tuning.
#'
#' @return \code{ModelSpecification} class object.
#'
#' @seealso \code{\link{fit}}, \code{\link{resample}},
#' \code{\link{set_monitor}}, \code{\link{set_optim}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' modelspec <- ModelSpecification(
#'   sale_amount ~ ., data = ICHomes, model = GBMModel
#' )
#' fit(modelspec)
#' }
#'
ModelSpecification <- function(...) {
  UseMethod("ModelSpecification")
}


#' @rdname ModelSpecification-methods
#'
ModelSpecification.default <- function(
  input, model, control = MachineShop::settings("control"), metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams"), ...
) {
  object <- new("ModelSpecification",
    input = as.MLInput(input),
    model = as.MLModel(model),
    params = TrainingParams(
      control = control,
      metrics = metrics,
      cutoff = cutoff,
      stat = stat
    ),
    grid = tibble()
  )

  if (is_optim_method(object)) {
    grid <- expand_modelgrid(object)
    object <- if (is_tibble(grid)) {
      object@grid <- grid
      update(object, params = list())
    } else {
      update(object)
    }
  }

  object
}


#' @rdname ModelSpecification-methods
#'
ModelSpecification.formula <- function(formula, data, model, ...) {
  ModelSpecification(as.MLInput(formula, data), model = model, ...)
}


#' @rdname ModelSpecification-methods
#'
ModelSpecification.matrix <- function(x, y, model, ...) {
  ModelSpecification(as.MLInput(x, y), model = model, ...)
}


#' @rdname ModelSpecification-methods
#'
ModelSpecification.ModelFrame <- function(input, model, ...) {
  ModelSpecification.default(input, model = model, ...)
}


#' @rdname ModelSpecification-methods
#'
ModelSpecification.recipe <- function(input, model, ...) {
  ModelSpecification.default(input, model = model, ...)
}


setMethod("initialize", "ModelSpecification",
  function(.Object, ..., id = make_id("mspec")) {
    callNextMethod(.Object, ..., id = id)
  }
)


#' Update a Model Specification
#'
#' Update the components of a \pkg{MachineShop} model specification.
#'
#' @rdname update
#'
#' @param object model \link[=ModelSpecification]{specification}.
#' @param params list of values to update the corresponding named input and
#'   model objects and their parameters.  See the \code{object} grid for the
#'   names of updatable objects.
#' @param check_grid logical indicating whether to check the tuning grid and
#'   to set the optimization and control methods to null if empty.
#' @param data data frame with which to update data internal to input objects.
#'   The internal data is a processed version of the original and is not
#'   intended to be supplied by users.
#' @param ... arguments passed to other methods.
#'
#' @return Updated \code{ModelSpecification} class object.
#'
#' @noRd
#'
update.ModelSpecification <- function(
  object, params = NULL, check_grid = TRUE, data = NULL, ...
) {
  if (is.list(params)) {
    object@input <- update_slots(object@input, params = params)
    object@model <- update_slots(object@model, params = params)
    object@grid <- unique_grid(grid_diff(object@grid, params))
  }
  if (check_grid && is_empty(object@grid)) {
    object@params@optim <- NullOptimization()
    object@params@control <- NullControl()
  }
  object@input <- update(object@input, data = data)
  object
}


map_slots <- function(
  fun, object, names = c("candidates", "input", "model"), update = FALSE
) {
  slots_res <- NULL
  for (name in intersect(names, slotNames(object))) {
    slot <- slot(object, name)
    if (is(slot, "ListOf")) {
      convert <- function(x) {
        names(x) <- names(slot)
        ListOf(x)
      }
      slot_res <- unlist(map(function(slot_item) {
        map_slots(fun, slot_item, names = names, update = update)
      }, unname(slot)), recursive = FALSE)
    } else {
      convert <- identity
      slot_res <- map_slots(fun, slot, names = names, update = update)
    }
    if (update) {
      slot(object, name) <- convert(slot_res)
    } else {
      slots_res <- c(slots_res, slot_res)
    }
  }
  res <- fun(object)
  if (!update) c(structure(list(res), names = object@id), slots_res) else res
}


update_slots <- function(object, params = NULL, ...) {
  convert <- if (is.null(params)) identity else as.list
  map_slots(function(slot) {
    update(slot, params = convert(params[[slot@id]]))
  }, object, update = TRUE, ...)
}

append <- function(...) {
  Reduce(.append, list(...))
}


setGeneric(".append", function(x, y, ...) standardGeneric(".append"))


setMethod(".append", c("ANY", "ANY"),
  function(x, y) c(x, y)
)


setMethod(".append", c("ANY", "missing"),
  function(x, y) x
)


setMethod(".append", c("data.frame", "data.frame"),
  function(x, y) {
    stopifnot(names(x) == names(y))
    df <- data.frame(row.names = seq_len(nrow(x) + nrow(y)))
    for (varname in names(x)) {
      df[[varname]] <- .append(x[[varname]], y[[varname]])
    }
    df
  }
)


setMethod(".append", c("matrix", "matrix"),
  function(x, y) rbind(x, y)
)


setMethod(".append", c("SurvMatrix", "SurvMatrix"),
  function(x, y) c(x, y)
)


setMethod(".append", c("tbl_df", "tbl_df"),
  function(x, y) {
    stopifnot(names(x) == names(y))
    tbl <- tibble(.rows = nrow(x) + nrow(y))
    for (varname in names(x)) {
      tbl[[varname]] <- .append(x[[varname]], y[[varname]])
    }
    tbl
  }
)

convert_numeric <- function(x, ...) {
  check_numeric(x, ..., nonempty = FALSE, na.fail = FALSE)
}


setGeneric("convert_predicted",
  function(object, x, ...) standardGeneric("convert_predicted")
)


setMethod("convert_predicted", c("ANY", "ANY"),
  function(object, x, ...) x
)


setMethod("convert_predicted", c("ANY", "numeric"),
  function(object, x, ...) unname(x)
)


setMethod("convert_predicted", c("BinomialVariate", "ANY"),
  function(object, x, ...) unname(drop(x))
)


setMethod("convert_predicted", c("factor", "array"),
  function(object, x, ...) {
    convert_predicted(object, adrop(x, ndim(x)))
  }
)


setMethod("convert_predicted", c("factor", "matrix"),
  function(object, x, ...) {
    stopifnot(ncol(x) > 0)
    if (nlevels(object) == 2) {
      convert_predicted(object, x[, ncol(x)])
    } else {
      structure(x, dimnames = list(NULL, levels(object)))
    }
  }
)


setMethod("convert_predicted", c("matrix", "array"),
  function(object, x, ...) {
    convert_predicted(object, adrop(x, ndim(x)))
  }
)


setMethod("convert_predicted", c("matrix", "matrix"),
  function(object, x, ...) {
    stopifnot(ncol(object) == ncol(x))
    var_names <- colnames(x)
    if (is.null(var_names)) var_names <- colnames(object)
    if (is.null(var_names)) var_names <- make_names_len(ncol(x), "y")
    structure(x, dimnames = list(NULL, var_names))
  }
)


setMethod("convert_predicted", c("numeric", "array"),
  function(object, x, ...) {
    n <- ndim(x)
    x <- if (n == 1) c(x) else adrop(x, n)
    convert_predicted(object, x)
  }
)


setMethod("convert_predicted", c("numeric", "matrix"),
  function(object, x, ...) {
    stopifnot(ncol(x) == 1)
    convert_predicted(object, x[, 1])
  }
)


setGeneric("convert_response",
  function(object, x, ...) standardGeneric("convert_response")
)


setMethod("convert_response", c("ANY", "ANY"),
  function(object, x, ...) x
)


setMethod("convert_response", c("DiscreteVariate", "numeric"),
  function(object, x, ...) {
    x <- round(x)
    if (object@min > -Inf) x <- pmax(x, object@min)
    if (object@max < Inf) x <- pmin(x, object@max)
    new(class(object), x, min = object@min, max = object@max)
  }
)


setMethod("convert_response", c("factor", "factor"),
  function(object, x, ...) x
)


setMethod("convert_response", c("factor", "matrix"),
  function(object, x, ...) {
    factor(
      max.col(x), levels = seq_len(nlevels(object)), labels = levels(object)
    )
  }
)


setMethod("convert_response", c("factor", "numeric"),
  function(object, x, cutoff, ...) {
    factor(x > cutoff, levels = c(FALSE, TRUE), labels = levels(object))
  }
)


setMethod("convert_response", c("integer", "numeric"),
  function(object, x, ...) {
    round(x)
  }
)


setMethod("convert_response", c("ordered", "matrix"),
  function(object, x, ...) {
    ordered(
      max.col(x), levels = seq_len(nlevels(object)), labels = levels(object)
    )
  }
)


setMethod("convert_response", c("Surv", "SurvProbs"),
  function(object, x, cutoff, ...) {
    events <- x < cutoff
    storage.mode(events) <- "integer"
    SurvEvents(events, x@times, distr = x@distr)
  }
)

#' Ordered Logistic or Probit Regression Model
#'
#' Fit a logistic or probit regression model to an ordered factor response.
#'
#' @param method logistic or probit or (complementary) log-log or cauchit
#' (corresponding to a Cauchy latent variable).
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{ordered}}
#' }
#'
#' Further model details can be found in the source link below.
#'
#' In calls to \code{\link{varimp}} for \code{POLRModel}, numeric argument
#' \code{base} may be specified for the (negative) logarithmic transformation of
#' p-values [defaul: \code{exp(1)}].  Transformed p-values are automatically
#' scaled in the calculation of variable importance to range from 0 to 100.  To
#' obtain unscaled importance values, set \code{scale = FALSE}.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[MASS]{polr}}, \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' data(Boston, package = "MASS")
#'
#' df <- within(Boston,
#'              medv <- cut(medv,
#'                          breaks = c(0, 10, 15, 20, 25, 50),
#'                          ordered = TRUE))
#' fit(medv ~ ., data = df, model = POLRModel)
#'
POLRModel <- function(
  method = c("logistic", "probit", "loglog", "cloglog", "cauchit")
) {

  method <- match.arg(method)

  MLModel(

    name = "POLRModel",
    label = "Ordered Logistic Regression",
    packages = "MASS",
    response_types = "ordered",
    weights = TRUE,
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment()),

    fit = function(formula, data, weights, ...) {
      MASS::polr(
        formula, data = as.data.frame(formula, data = data), weights = weights,
        na.action = na.pass, Hess = TRUE, ...)
    },

    predict = function(object, newdata, ...) {
      newdata <- as.data.frame(newdata)
      predict(object, newdata = newdata, type = "probs")
    },

    varimp = function(object, base = exp(1), ...) {
      varimp_pval(object, base = base)
    }

  )

}

MLModelFunction(POLRModel) <- NULL

#' Fast Random Forest (SRC) Model
#'
#' Fast OpenMP computing of Breiman's random forest for a variety of data
#' settings including right-censored survival, regression, and classification.
#'
#' @rdname RFSRCModel
#'
#' @param ntree number of trees.
#' @param mtry number of variables randomly selected as candidates for splitting
#'   a node.
#' @param nodesize minumum size of terminal nodes.
#' @param nodedepth maximum depth to which a tree should be grown.
#' @param splitrule splitting rule (see \code{\link[randomForestSRC]{rfsrc}}).
#' @param nsplit non-negative integer value for number of random splits to
#'   consider for each candidate splitting variable.
#' @param block.size interval number of trees at which to compute the cumulative
#'   error rate.
#' @param samptype whether bootstrap sampling is with or without replacement.
#' @param membership logical indicating whether to return terminal node
#'   membership.
#' @param sampsize function specifying the bootstrap size.
#' @param nimpute number of iterations of the missing data imputation algorithm.
#' @param ntime integer number of time points to constrain ensemble calculations
#'   for survival outcomes.
#' @param proximity whether and how to return proximity of cases as measured by
#'   the frequency of sharing the same terminal nodes.
#' @param distance whether and how to return distance between cases as measured
#'   by the ratio of the sum of edges from each case to the root node.
#' @param forest.wt whether and how to return the forest weight matrix.
#' @param xvar.wt vector of non-negative weights representing the probability of
#'   selecting a variable for splitting.
#' @param split.wt vector of non-negative weights used for multiplying the split
#'   statistic for a variable.
#' @param var.used whether and how to return variables used for splitting.
#' @param split.depth whether and how to return minimal depth for each variable.
#' @param do.trace number of seconds between updates to the user on approximate
#'   time to completion.
#' @param statistics logical indicating whether to return split statistics.
#' @param terminal.qualts logical indicating whether to return terminal node
#'   membership information.
#' @param ... arguments passed to \code{RFSRCModel}.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{matrix}, \code{numeric},
#'     \code{Surv}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \code{mtry}, \code{nodesize}
#'   }
#' }
#'
#' Default argument values and further model details can be found in the source
#' See Also links below.
#'
#' In calls to \code{\link{varimp}} for \code{RFSRCModel}, argument
#' \code{type} may be specified as \code{"anti"} (default) for cases assigned to
#' the split opposite of the random assignments, as \code{"permute"} for
#' permutation of OOB cases, or as \code{"random"} for permutation replaced with
#' random assignment.  Variable importance is automatically scaled to range from
#' 0 to 100.  To obtain unscaled importance values, set \code{scale = FALSE}.
#' See example below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[randomForestSRC]{rfsrc}},
#' \code{\link[randomForestSRC]{rfsrc.fast}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package randomForestSRC to run
#'
#' model_fit <- fit(sale_amount ~ ., data = ICHomes, model = RFSRCModel)
#' varimp(model_fit, method = "model", type = "random", scale = TRUE)
#' }
#'
RFSRCModel <- function(
  ntree = 1000, mtry = integer(), nodesize = integer(), nodedepth = integer(),
  splitrule = character(), nsplit = 10, block.size = integer(),
  samptype = c("swor", "swr"), membership = FALSE,
  sampsize = if (samptype == "swor") function(x) 0.632 * x else function(x) x,
  nimpute = 1, ntime = integer(),
  proximity = c(FALSE, TRUE, "inbag", "oob", "all"),
  distance = c(FALSE, TRUE, "inbag", "oob", "all"),
  forest.wt = c(FALSE, TRUE, "inbag", "oob", "all"),
  xvar.wt = numeric(), split.wt = numeric(),
  var.used = c(FALSE, "all.trees", "by.tree"),
  split.depth = c(FALSE, "all.trees", "by.tree"),
  do.trace = FALSE, statistics = FALSE
) {

  samptype <- match.arg(samptype)

  proximity <- match.arg(as.character(proximity), proximity)
  distance <- match.arg(as.character(distance), distance)
  forest.wt <- match.arg(as.character(forest.wt), forest.wt)
  var.used <- match.arg(as.character(var.used), var.used)
  split.depth <- match.arg(as.character(split.depth), split.depth)

  MLModel(

    name = "RFSRCModel",
    label = "Random Forest (SRC)",
    packages = "randomForestSRC",
    response_types = c("factor", "matrix", "numeric", "Surv"),
    weights = TRUE,
    predictor_encoding = "model.frame",
    na.rm = TRUE,
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = c("mtry", "nodesize"),
      get_values = c(
        function(n, data, ...) seq_nvars(data, RFSRCModel, n),
        function(n, data, ...) {
          round_int(seq(1, min(20, nrow(data)), length = n))
        }
      )
    ),

    fit = function(formula, data, weights, ...) {
      y <- response(data)
      data <- as.data.frame(formula, data = data)
      family <- switch_class(y,
        "matrix" = {
          colnames(y) <- make_names_len(ncol(y), "y")
          "Multivar"
        },
        "Surv" = "Surv"
      )
      if (!is.null(family)) {
        y_names <- map("char", recipes::rand_id, colnames(y))
        data[[response(formula)]] <- NULL
        data[y_names] <- y
        formula[[2]] <- as.call(map(as.name, c(family, y_names)))
      }
      randomForestSRC::rfsrc(
        formula, data = data, case.wt = weights, ...
      )
    },

    predict = function(object, newdata, .MachineShop, ...) {
      newdata <- as.data.frame(newdata)
      pred <- randomForestSRC::predict.rfsrc(object, newdata = newdata)
      if (pred$family == "regr+") {
        y_names <- pred$yvar.names
        names(y_names) <- sub("_.*", "", y_names)
        simplify(map(function(name) pred$regrOutput[[name]]$predicted, y_names))
      } else if (pred$family == "surv") {
        predict(Surv(pred$time.interest), pred$survival, ...)
      } else {
        pred$predicted
      }
    },

    varimp = function(object, type = c("anti", "permute", "random"), ...) {
      vi <- randomForestSRC::vimp(object, importance = match.arg(type))
      structure(
        if (vi$family == "regr+") {
          y_names <- vi$yvar.names
          names(y_names) <- sub("_.*", "", y_names)
          simplify(map(function(name) {
            vi$regrOutput[[name]]$importance
          }, y_names))
        } else {
          vi$importance
        },
        metric = switch(vi$family,
          "class" = "accuracy",
          "regr" = "mse",
          "regr+" = "mse",
          "surv" = "cindex"
        )
      )
    }

  )

}

MLModelFunction(RFSRCModel) <- NULL


#' @rdname RFSRCModel
#'
RFSRCFastModel <- function(
  ntree = 500, sampsize = function(x) min(0.632 * x, max(x^0.75, 150)),
  ntime = 50, terminal.qualts = FALSE, ...
) {
  model <- RFSRCModel(ntree = ntree, sampsize = sampsize, ntime = ntime, ...)
  model@params$terminal.qualts <- terminal.qualts
  model
}

MLModelFunction(RFSRCFastModel) <- NULL

#' Bagging with Classification Trees
#'
#' Fits the Bagging algorithm proposed by Breiman in 1996 using classification
#' trees as single classifiers.
#'
#' @param mfinal number of trees to use.
#' @param minsplit minimum number of observations that must exist in a node in
#'   order for a split to be attempted.
#' @param minbucket minimum number of observations in any terminal node.
#' @param cp complexity parameter.
#' @param maxcompete number of competitor splits retained in the output.
#' @param maxsurrogate number of surrogate splits retained in the output.
#' @param usesurrogate how to use surrogates in the splitting process.
#' @param xval number of cross-validations.
#' @param surrogatestyle controls the selection of a best surrogate.
#' @param maxdepth maximum depth of any node of the final tree, with the root
#'   node counted as depth 0.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \code{mfinal}, \code{maxdepth}
#'   }
#' }
#'
#' Further model details can be found in the source link below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[adabag]{bagging}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package adabag to run
#'
#' fit(Species ~ ., data = iris, model = AdaBagModel(mfinal = 5))
#' }
#'
AdaBagModel <- function(
  mfinal = 100, minsplit = 20, minbucket = round(minsplit/3), cp = 0.01,
  maxcompete = 4, maxsurrogate = 5, usesurrogate = 2, xval = 10,
  surrogatestyle = 0, maxdepth = 30
) {

  MLModel(

    name = "AdaBagModel",
    label = "Bagging with Classification Trees",
    packages = "adabag",
    response_types = "factor",
    predictor_encoding = "model.frame",
    na.rm = FALSE,
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = c("mfinal", "maxdepth"),
      get_values = c(
        function(n, ...) round_int(seq_range(0, 25, c(1, 200), n + 1)),
        function(n, ...) seq_len(min(n, 30))
      )
    ),

    fit = function(formula, data, weights, mfinal, ...) {
      adabag::bagging(
        formula, data = as.data.frame(formula, data = data), mfinal = mfinal,
        control = list(...)
      )
    },

    predict = function(object, newdata, ...) {
      newdata <- as.data.frame(newdata)
      predict(object, newdata = newdata)$prob
    },

    varimp = function(object, ...) {
      structure(object$importance, metric = "gini")
    }

  )

}

MLModelFunction(AdaBagModel) <- NULL

#' Selected Model Inputs
#'
#' Formula, design matrix, model frame, or recipe selection from a candidate
#' set.
#'
#' @aliases SelectedModelFrame
#' @aliases SelectedModelRecipe
#' @aliases SelectedModelSpecification
#' @rdname SelectedInput
#'
#' @param ... \link{inputs} defining relationships between model predictor
#'   and response variables.  Supplied inputs must all be of the same type and
#'   may be named or unnamed.
#' @param x list of inputs followed by arguments passed to their method
#'   function.
#' @param y response variable.
#' @param data \link[=data.frame]{data frame} containing predictor and response
#'   variables.
#' @param control \link[=controls]{control} function, function name, or object
#'   defining the resampling method to be employed.
#' @param metrics \link[=metrics]{metric} function, function name, or vector of
#'   these with which to calculate performance.  If not specified, default
#'   metrics defined in the \link{performance} functions are used.  Recipe
#'   selection is based on the first calculated metric.
#' @param cutoff argument passed to the \code{metrics} functions.
#' @param stat function or character string naming a function to compute a
#'   summary statistic on resampled metric values for recipe selection.
#'
#' @return \code{SelectedModelFrame}, \code{SelectedModelRecipe}, or
#' \code{SelectedModelSpecification} class object that inherits from
#' \code{SelectedInput} and \code{ModelFrame}, \code{recipe}, or
#' \code{ModelSpecification}, respectively.
#'
#' @seealso \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' ## Selected model frame
#' sel_mf <- SelectedInput(
#'   sale_amount ~ sale_year + built + style + construction,
#'   sale_amount ~ sale_year + base_size + bedrooms + basement,
#'   data = ICHomes
#' )
#'
#' fit(sel_mf, model = GLMModel)
#'
#' ## Selected recipe
#' library(recipes)
#' data(Boston, package = "MASS")
#'
#' rec1 <- recipe(medv ~ crim + zn + indus + chas + nox + rm, data = Boston)
#' rec2 <- recipe(medv ~ chas + nox + rm + age + dis + rad + tax, data = Boston)
#' sel_rec <- SelectedInput(rec1, rec2)
#'
#' fit(sel_rec, model = GLMModel)
#'
SelectedInput <- function(...) {
  UseMethod("SelectedInput")
}


#' @rdname SelectedInput
#'
SelectedInput.formula <- function(
  ..., data, control = MachineShop::settings("control"), metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
) {
  inputs <- list(...)
  if (!all(map("logi", is, inputs, "formula"))) {
    throw(Error("Inputs must all be formulas."))
  }
  mf_list <- map(function(x) {
    do.call(ModelFrame, list(x, data, strata = response(x), na.rm = FALSE))
  }, inputs)
  SelectedInput(
    mf_list, control = control, metrics = metrics, stat = stat, cutoff = cutoff
  )
}


#' @rdname SelectedInput
#'
SelectedInput.matrix <- function(
  ..., y, control = MachineShop::settings("control"), metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
) {
  inputs <- list(...)
  if (!all(map("logi", is, inputs, "matrix"))) {
    throw(Error("Inputs must all be matrices."))
  }
  mf_list <- map(ModelFrame, inputs, list(y), strata = list(y), na.rm = FALSE)
  SelectedInput(
    mf_list, control = control, metrics = metrics, stat = stat, cutoff = cutoff
  )
}


#' @rdname SelectedInput
#'
SelectedInput.ModelFrame <- function(
  ..., control = MachineShop::settings("control"), metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
) {

  if (...length() == 1) return(..1)

  params <- do.call(TrainingParams, as.list(environment()))
  inputs <- list(...)

  if (!all(map("char", class1, inputs) == "ModelFrame")) {
    throw(Error("Inputs must all be ModelFrames."))
  }

  if (!identical_elements(inputs, function(x) x[[1]])) {
    throw(Error("ModelFrames have different response variables."))
  }

  combined <- combine_inputs(inputs)

  new("SelectedModelFrame",
    ModelFrame(combined$data), candidates = combined$candidates, params = params
  )

}


#' @rdname SelectedInput
#'
SelectedInput.recipe <- function(
  ..., control = MachineShop::settings("control"), metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
) {

  if (...length() == 1) return(..1)

  params <- do.call(TrainingParams, as.list(environment()))
  inputs <- list(...)
  for (i in seq_along(inputs)) inputs[[i]] <- ModelRecipe(inputs[[i]])

  get_info <- function(x) {
    info <- summary(x)
    info <- info[info$role != "predictor", ]
    info[do.call(order, info[c("variable", "role", "source")]), ]
  }
  if (!identical_elements(inputs, get_info)) {
    throw(Error("Recipes have different non-predictor variables."))
  }

  combined <- combine_inputs(inputs)

  info <- summary(combined$candidates[[1]])
  outcomes <- info$variable[info$role == "outcome"]
  others <- info$variable[!(info$role %in% c("outcome", "predictor"))]
  fo <- reformulate(".", paste(outcomes, collapse = "+"))
  rec <- recipe(fo, data = combined$data)
  if (length(others)) {
    args <- list(rec, others, new_role = "other")
    rec <- do.call(recipes::update_role, args)
  }

  new("SelectedModelRecipe",
    new("ModelRecipe", rec), candidates = combined$candidates, params = params
  )

}


#' @rdname SelectedInput
#'
SelectedInput.ModelSpecification <- function(
  ..., control = MachineShop::settings("control"), metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
) {

  if (...length() == 1) return(..1)

  params <- do.call(TrainingParams, as.list(environment()))
  modelspecs <- list(...)
  default_names <- map("char", class, modelspecs)
  names(modelspecs) <- make_names_along(modelspecs, default_names)

  if (any(map("char", class1, modelspecs) != "ModelSpecification")) {
    throw(Error("Inputs must all be ModelSpecifications."))
  }

  sel_input <- SelectedInput(map(slot, modelspecs, "input"))
  classes <- c("ModelFrame", "ModelRecipe")
  input <- as(sel_input, classes[map("logi", is, list(sel_input), classes)])
  modelspecs <- map(function(modelspec, input) {
    update(modelspec, data = as.data.frame(input))
  }, modelspecs, sel_input@candidates)

  new("SelectedModelSpecification",
    ModelSpecification(input, NullModel()),
    candidates = ListOf(modelspecs),
    params = params
  )

}


#' @rdname SelectedInput
#'
SelectedInput.list <- function(x, ...) {
  do.call(SelectedInput, c(x, list(...)))
}


.fit.SelectedInput <- function(object, ...) {
  .fit_optim(object, ...)
}


update.SelectedInput <- function(object, params = NULL, ...) {
  if (is.list(params)) {
    id <- object@id
    object <- subset_selected(object, "candidates", params$id)
    data <- as.data.frame(object)
    object@candidates <- ListOf(map(function(input) {
      res <- update(input, data = data[names(as.data.frame(input))])
      res@id <- input@id
      res
    }, object@candidates))
    new_params <- as(as(object, "SelectedInput"), "list")
    new_params[names(params)] <- params
    candidates <- new_params$candidates
    new_params[c("candidates", "id")] <- NULL
    object <- do.call(SelectedInput, c(candidates, new_params))
    object@id <- id
  }
  NextMethod(params = NULL, check_grid = FALSE)
}


#' Tuned Model Inputs
#'
#' Recipe tuning over a grid of parameter values.
#'
#' @aliases TunedModelRecipe
#'
#' @rdname TunedInput
#'
#' @param object untrained \code{\link[recipes]{recipe}}.
#' @param grid \code{RecipeGrid} containing parameter values at which to
#'   evaluate a recipe, such as those returned by \code{\link{expand_steps}}.
#' @param control \link[=controls]{control} function, function name, or object
#'   defining the resampling method to be employed.
#' @param metrics \link[=metrics]{metric} function, function name, or vector of
#'   these with which to calculate performance.  If not specified, default
#'   metrics defined in the \link{performance} functions are used.  Recipe
#'   selection is based on the first calculated metric.
#' @param cutoff argument passed to the \code{metrics} functions.
#' @param stat function or character string naming a function to compute a
#'   summary statistic on resampled metric values for recipe tuning.
#' @param ... arguments passed to other methods.
#'
#' @return \code{TunedModelRecipe} class object that inherits from
#' \code{TunedInput} and \code{recipe}.
#'
#' @seealso \code{\link{fit}}, \code{\link{resample}}, \code{\link{set_optim}}
#'
#' @examples
#' library(recipes)
#' data(Boston, package = "MASS")
#'
#' rec <- recipe(medv ~ ., data = Boston) %>%
#'   step_pca(all_numeric_predictors(), id = "pca")
#'
#' grid <- expand_steps(
#'   pca = list(num_comp = 1:2)
#' )
#'
#' fit(TunedInput(rec, grid = grid), model = GLMModel)
#'
TunedInput <- function(object, ...) {
  UseMethod("TunedInput")
}


#' @rdname TunedInput
#'
TunedInput.recipe <- function(
  object, grid = expand_steps(), control = MachineShop::settings("control"),
  metrics = NULL, cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams"), ...
) {

  object <- new("TunedModelRecipe",
    ModelRecipe(object),
    grid = grid,
    params = TrainingParams(
      control = control,
      metrics = metrics,
      cutoff = cutoff,
      stat = stat
    )
  )

  grid_names <- names(object@grid)
  step_ids <- map("char", getElement, object$steps, "id")
  missing <- !(grid_names %in% step_ids)
  if (any(missing)) {
    throw(Error(
      note_items("Grid step name{?s}: ", grid_names[missing], "; "),
      note_items("not found in recipe step id{?s}: ", step_ids, ".")
    ))
  }

  object

}


.fit.TunedModelRecipe <- function(object, ...) {
  .fit_optim(object, ...)
}


update.TunedModelRecipe <- function(object, params = NULL, ...) {
  if (is.list(params)) {
    update(as(object, "ModelRecipe"), params = params)
  } else {
    object
  }
}

#' Linear Models
#'
#' Fits linear models.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{matrix}, \code{numeric}}
#' }
#'
#' Further model details can be found in the source link below.
#'
#' In calls to \code{\link{varimp}} for \code{LModel}, numeric argument
#' \code{base} may be specified for the (negative) logarithmic transformation of
#' p-values [defaul: \code{exp(1)}].  Transformed p-values are automatically
#' scaled in the calculation of variable importance to range from 0 to 100.  To
#' obtain unscaled importance values, set \code{scale = FALSE}.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[stats]{lm}}, \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' fit(sale_amount ~ ., data = ICHomes, model = LMModel)
#'
LMModel <- function() {

  MLModel(

    name = "LMModel",
    label = "Linear Model",
    packages = "stats",
    response_types = c("factor", "matrix", "numeric"),
    weights = TRUE,
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment()),

    fit = function(formula, data, weights, ...) {
      y <- response(data)
      data <- as.data.frame(formula, data = data)
      if (is.factor(y)) {
        data[[response(formula)]] <- if (nlevels(y) == 2) {
          as.numeric(y) - 1
        } else {
          mat <- model.matrix(~ y - 1)
          colnames(mat) <- levels(y)
          mat
        }
      }
      stats::lm(
        formula, data = data, weights = weights, na.action = na.pass, ...
      )
    },

    predict = function(object, newdata, ...) {
      newdata <- as.data.frame(newdata)
      predict(object, newdata = newdata)
    },

    varimp = function(object, base = exp(1), ...) {
      varimp_pval(object, test = "F", base = base)
    }

  )

}

MLModelFunction(LMModel) <- NULL

#' Iowa City Home Sales Dataset
#'
#' Characteristics of homes sold in Iowa City, IA from 2005 to 2008 as reported
#' by the county assessor's office.
#'
#' @format A data frame with 753 observations of 17 variables:
#' \describe{
#'   \item{sale_amount}{sale amount in dollars.}
#'   \item{sale_year}{sale year.}
#'   \item{sale_month}{sale month.}
#'   \item{built}{year in which the home was built.}
#'   \item{style}{home stlye (Home/Condo)}
#'   \item{construction}{home construction type.}
#'   \item{base_size}{base foundation size in sq ft.}
#'   \item{add_size}{size of additions made to the base foundation in sq ft.}
#'   \item{garage1_size}{attached garage size in sq ft.}
#'   \item{garage2_size}{detached garage size in sq ft.}
#'   \item{lot_size}{total lot size in sq ft.}
#'   \item{bedrooms}{number of bedrooms.}
#'   \item{basement}{presence of a basement (No/Yes).}
#'   \item{ac}{presence of central air conditioning (No/Yes).}
#'   \item{attic}{presence of a finished attic (No/Yes).}
#'   \item{lon,lat}{home longitude/latitude coordinates.}
#' }
#'
"ICHomes"

#' Display Performance Metric Information
#'
#' Display information about metrics provided by the \pkg{MachineShop} package.
#'
#' @param ... \link[=metrics]{metric} functions or function names;
#' \link[=response]{observed responses}; \link[=response]{observed} and
#' \link[=predict]{predicted} responses; \link{confusion} or \link{resample}
#' results for which to display information.  If none are specified, information
#' is returned on all available metrics by default.
#'
#' @return List of named metric elements each containing the following
#' components:
#' \describe{
#'   \item{label}{character descriptor for the metric.}
#'   \item{maximize}{logical indicating whether higher values of the metric
#'     correspond to better predictive performance.}
#'   \item{arguments}{closure with the argument names and corresponding default
#'     values of the metric function.}
#'   \item{response_types}{data frame of the observed and predicted response
#'     variable types supported by the metric.}
#' }
#'
#' @examples
#' ## All metrics
#' metricinfo()
#'
#' ## Metrics by observed and predicted response types
#' names(metricinfo(factor(0)))
#' names(metricinfo(factor(0), factor(0)))
#' names(metricinfo(factor(0), matrix(0)))
#' names(metricinfo(factor(0), numeric(0)))
#'
#' ## Metric-specific information
#' metricinfo(auc)
#'
metricinfo <- function(...) {
  args <- list(...)
  args <- if (length(args)) unname(args) else as.list(settings("metrics"))
  info <- do.call(.metricinfo, args)

  is_type <- if (length(info)) !map("logi", is, info, "list") else NULL
  if (any(is_type)) {
    info_metrics <- if (all(is_type)) metricinfo() else info[!is_type]
    info_types <- do.call(.metricinfo_types, info[is_type])
    info <- c(info_metrics, info_types)
    info <- info[intersect(names(info_metrics), names(info_types))]
  }

  info[unique(names(info))]
}


.metricinfo <- function(x, ...) {
  UseMethod(".metricinfo")
}


.metricinfo.default <- function(x, ...) {
  info <- list(x)
  if (...length()) c(info, .metricinfo(...)) else info
}


.metricinfo.character <- function(x, ...) {
  metric <- tryCatch(
    as.MLMetric(x),
    error = function(cond) list()
  )
  .metricinfo(metric, ...)
}


.metricinfo.ConfusionList <- function(x, ...) {
  .metricinfo(x[[1]], ...)
}


.metricinfo.ConfusionMatrix <- function(x, ...) {
  c(list(x), .metricinfo(NULL, ...))
}


.metricinfo.function <- function(x, ...) {
  .metricinfo(list(), ...)
}


.metricinfo.list <- function(x, ...) {
  if (...length()) .metricinfo(...) else list()
}


.metricinfo.MLMetric <- function(x, ...) {
  generic_name <- x@name
  if (!isGeneric(generic_name)) generic_name <- paste0(".", x@name)
  if (isGeneric(generic_name)) {
    methods <- findMethods(generic_name)
    is_defined <- map("logi", function(method) {
      body(method) != .(numeric())
    }, methods)
    types <- as.data.frame(do.call(rbind, methods@signatures[is_defined]))
    names(types) <- methods@arguments
  } else {
    types <- NULL
  }

  info <- structure(list(list(
    label = x@label,
    maximize = x@maximize,
    arguments = args(x),
    response_types = types
  )), names = x@name)

  if (...length()) c(info, .metricinfo(...)) else info
}


.metricinfo.Resample <- function(x, ...) {
  .metricinfo(x$Observed, x$Predicted, ...)
}


.metricinfo_types <- function(x, y, ...) {
  not_missing_y <- !missing(y)
  if (not_missing_y && is.factor(x) && is.factor(y)) {
    x <- ConfusionMatrix(table(x, x), ordered = is.ordered(x))
    y <- NULL
  }
  info <- metricinfo()
  is_supported <- map("logi", function(this) {
    types <- this$response_types
    is_types <- map("logi", is, list(x), types$observed)
    if (not_missing_y) {
      is_types <- is_types & map("logi", is, list(y), types$predicted)
    }
    any(is_types)
  }, info)
  info[is_supported]
}

#' Set Recipe Roles
#'
#' Add to or replace the roles of variables in a preprocessing recipe.
#'
#' @name recipe_roles
#' @rdname recipe_roles
#'
#' @param recipe existing \link[recipes]{recipe} object.
#' @param x,size number of counts and trials for the specification of a
#'   \code{\link{BinomialVariate}} outcome.
#' @param group variable defining groupings of case observations, such as
#'   repeated measurements, to keep together during resampling [default: none].
#' @param stratum variable to use in conducting stratified \link{resample}
#'   estimation of model performance.
#' @param weight numeric variable of case weights for model
#'   \link[=fit]{fitting}.
#' @param replace logical indicating whether to replace existing roles.
#' @param offset numeric variable to be added to a linear predictor, such as in
#'   a generalized linear model, with known coefficient 1 rather than an
#'   estimated coefficient.
#' @param time,event numeric follow up time and 0-1 numeric or logical event
#'   indicator for specification of a \code{\link[survival]{Surv}} outcome.  If
#'   the event indicator is omitted, all cases are assumed to have events.
#'
#' @return An updated recipe object.
#'
#' @seealso \code{\link[recipes]{recipe}}
#'
#' @examples
#' library(survival)
#' library(recipes)
#'
#' df <- within(veteran, {
#'   y <- Surv(time, status)
#'   remove(time, status)
#' })
#' rec <- recipe(y ~ ., data = df) %>%
#'   role_case(stratum = y)
#'
#' (res <- resample(rec, model = CoxModel))
#' summary(res)
#'
NULL


#' @rdname recipe_roles
#'
role_binom <- function(recipe, x, size) {
  if (!(missing(x) || missing(size))) {
    recipe <- set_recipe_role(recipe, substitute(x), "binom_x")
    recipe <- set_recipe_role(recipe, substitute(size), "binom_size")
  } else {
    throw(Error("Binomial `x` and `size` variables must be specified."))
  }
}


#' @rdname recipe_roles
#'
role_case <- function(recipe, group, stratum, weight, replace = FALSE) {
  comp_names <- eval(substitute(alist(
    group = group,
    stratum = stratum,
    weight = weight
  )))
  for (type in names(comp_names)[nzchar(comp_names)]) {
    role <- paste0("case_", type)
    recipe <- set_recipe_role(recipe, comp_names[[type]], role, replace)
  }
  recipe
}


#' @rdname recipe_roles
#'
role_pred <- function(recipe, offset, replace = FALSE) {
  if (!missing(offset)) {
    recipe <- set_recipe_role(
      recipe, substitute(offset), "pred_offset", replace, TRUE
    )
  }
  recipe
}


#' @rdname recipe_roles
#'
role_surv <- function(recipe, time, event) {
  if (!missing(time)) {
    recipe <- set_recipe_role(recipe, substitute(time), "surv_time")
    if (!missing(event)) {
      recipe <- set_recipe_role(recipe, substitute(event), "surv_event")
    }
    recipe
  } else {
    throw(Error("A survival `time` variable must be specified."))
  }
}


set_recipe_role <- function(
  recipe, x, role, replace = FALSE, required = FALSE
) {
  set_role <- if (replace) recipes::update_role else recipes::add_role
  recipe <- do.call(set_role, list(recipe, x, new_role = role))
  if (!required) {
    recipes::update_role_requirements(recipe, role = role, bake = FALSE)
  } else recipe
}

## Resampling Controls

test_that("formula resampling", {
  skip_if_not(TEST_ALL)
  with_parallel({

    library(MASS)
    library(recipes)

    df1 <- ICHomes
    fo1 <- sale_amount ~ .
    rec1 <- recipe(fo1, data = df1)

    df2 <- iris
    fo2 <- Species ~ .
    rec2 <- recipe(fo2, data = df2)

    df3 <- within(Melanoma, status <- status != 2)
    fo3 <- Surv(time, status) ~ .
    rec3 <- recipe(time + status ~ ., data = df3) %>%
      role_surv(time = time, event = status) %>%
      role_case(stratum = status)

    model <- GBMModel

    controls <- c(
      "BootControl",
      "BootOptimismControl",
      "CVControl",
      "CVOptimismControl",
      "OOBControl",
      "SplitControl",
      "TrainControl"
    )

    for (control in controls) {
      ## formula resampling
      expect_s4_class(resample(fo1, df1, model, control = control), "Resample")
      expect_s4_class(resample(fo2, df2, model, control = control), "Resample")
      expect_s4_class(resample(fo3, df3, model, control = control), "Resample")
      ## recipe resampling
      expect_s4_class(resample(rec1, model, control = control), "Resample")
      expect_s4_class(resample(rec2, model, control = control), "Resample")
      expect_s4_class(resample(rec3, model, control = control), "Resample")
    }

  })
})

#' Extreme Gradient Boosting Models
#'
#' Fits models with an efficient implementation of the gradient boosting
#' framework from Chen & Guestrin.
#'
#' @rdname XGBModel
#'
#' @param nrounds number of boosting iterations.
#' @param ... model parameters as described below and in the XGBoost
#'   \href{https://xgboost.readthedocs.io/en/latest/parameter.html}{documentation}
#'   and arguments passed to \code{XGBModel} from the other constructors.
#' @param verbose numeric value controlling the amount of output printed
#'   during model fitting, such that 0 = none, 1 = performance information, and
#'   2 = additional information.
#' @param print_every_n numeric value designating the fitting iterations at
#'   at which to print output when \code{verbose > 0}.
#' @param eta shrinkage of variable weights at each iteration to prevent
#'   overfitting.
#' @param gamma minimum loss reduction required to split a tree node.
#' @param max_depth maximum tree depth.
#' @param min_child_weight minimum sum of observation weights required of nodes.
#' @param subsample subsample ratio of the training observations.
#' @param colsample_bytree,colsample_bylevel,colsample_bynode subsample ratio of
#'   variables for each tree, level, or split.
#' @param rate_drop rate at which to drop trees during the dropout procedure.
#' @param one_drop integer indicating whether to drop at least one tree during
#'   the dropout procedure.
#' @param skip_drop probability of skipping the dropout procedure during a
#'   boosting iteration.
#' @param alpha,lambda L1 and L2 regularization terms for variable weights.
#' @param max_delta_step,tree_method,scale_pos_weight,updater,refresh_leaf,grow_policy,max_leaves,max_bin,num_parallel_tree
#'   other tree booster parameters.
#' @param sample_type,normalize_type type of sampling and normalization
#'   algorithms.
#' @param feature_selector,top_k character string specifying the feature
#'   selection and ordering method, and number of top variables to select in the
#'   \code{"greedy"} and \code{"thrifty"} feature selectors.
#' @param objective optional character string defining the learning task and
#'   objective.  Set automatically if not specified according to the following
#'   values available for supported response variable types.
#'   \describe{
#'     \item{\code{factor}:}{\code{"multi:softprob"}, \code{"binary:logistic"}
#'       (2 levels only)}
#'     \item{\code{numeric}:}{\code{"reg:squarederror"}, \code{"reg:logistic"},
#'       \code{"reg:gamma"}, \code{"reg:tweedie"}, \code{"rank:pairwise"},
#'       \code{"rank:ndcg"}, \code{"rank:map"}}
#'     \item{\code{PoissonVariate}:}{\code{"count:poisson"}}
#'     \item{\code{Surv}:}{\code{"survival:aft"}, \code{"survival:cox"}}
#'   }
#'   The first values listed are the defaults for the corresponding response
#'   types.
#' @param aft_loss_distribution character string specifying a distribution for
#'   the accelerated failure time objective (\code{"survival:aft"}) as
#'   \code{"extreme"}, \code{"logistic"}, or \code{"normal"}.
#' @param base_score initial prediction score of all observations, global bias.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric},
#'     \code{PoissonVariate}, \code{Surv}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \itemize{
#'       \item XGBModel: \code{NULL}
#'       \item XGBDARTModel: \code{nrounds}, \code{eta}*, \code{gamma}*,
#'         \code{max_depth}, \code{min_child_weight}*, \code{subsample}*,
#'         \code{colsample_bytree}*, \code{rate_drop}*, \code{skip_drop}*
#'       \item XGBLinearModel: \code{nrounds}, \code{alpha}, \code{lambda}
#'       \item XGBTreeModel: \code{nrounds}, \code{eta}*, \code{gamma}*,
#'         \code{max_depth}, \code{min_child_weight}*, \code{subsample}*,
#'         \code{colsample_bytree}*
#'     }
#'   }
#' }
#' * excluded from grids by default
#'
#' The booster-specific constructor functions \code{XGBDARTModel},
#' \code{XGBLinearModel}, and \code{XGBTreeModel} are special cases of
#' \code{XGBModel} which automatically set the XGBoost \code{booster}
#' \href{https://xgboost.readthedocs.io/en/latest/parameter.html}{parameter}.
#' These are called directly in typical usage unless \code{XGBModel} is needed
#' to specify a more general model.
#' 
#' Default argument values and further model details can be found in the source
#' See Also link below.
#'
#' In calls to \code{\link{varimp}} for \code{XGBTreeModel}, argument
#' \code{type} may be specified as \code{"Gain"} (default) for the fractional
#' contribution of each predictor to the total gain of its splits, as
#' \code{"Cover"} for the number of observations related to each predictor, or
#' as \code{"Frequency"} for the percentage of times each predictor is used in
#' the trees.  Variable importance is automatically scaled to range from 0 to
#' 100.  To obtain unscaled importance values, set \code{scale = FALSE}.  See
#' example below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[xgboost:xgb.train]{xgboost}}, \code{\link{fit}},
#' \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package xgboost to run
#'
#' model_fit <- fit(
#'   Species ~ ., data = iris, model = XGBTreeModel(nthread = 1)
#' )
#' varimp(model_fit, method = "model", type = "Frequency", scale = FALSE)
#' }
#'
XGBModel <- function(
  nrounds = 100, ..., objective = character(), aft_loss_distribution = "normal",
  base_score = 0.5, verbose = 0, print_every_n = 1
) {

  params <- as.list(environment())

  MLModel(

    name = "XGBModel",
    label = "Extreme Gradient Boosting",
    packages = "xgboost (>= 3.1.2)",
    response_types = c("factor", "numeric", "PoissonVariate", "Surv"),
    weights = TRUE,
    predictor_encoding = "model.matrix",
    na.rm = "response",
    params = new_params(c(params[1], ..., params[-1])),

    fit = function(
      formula, data, weights, nrounds, verbose, print_every_n, ...
    ) {
      dmat <- xgboost::xgb.DMatrix(
        model.matrix(data, intercept = FALSE),
        weight = weights,
        base_margin = model.offset(data)
      )
      y <- response(data)
      y_levels <- levels(y)

      params <- list(...)

      choices <- switch_class(y,
        "factor" = {
          y <- as.numeric(y) - 1
          c("multi:softprob", if (length(y_levels) <= 2) "binary:logistic")
        },
        "numeric" = c(
          "reg:squarederror", "reg:logistic", "reg:gamma", "reg:tweedie",
          "rank:pairwise", "rank:ndcg", "rank:map"
        ),
        "PoissonVariate" = "count:poisson",
        "Surv" = {
          throw(check_censoring(y, "right"))
          y_time <- y[, "time"]
          y_event <- y[, "status"] == 1
          c("survival:aft", "survival:cox")
        }
      )
      params$objective <- match.arg(params$objective, choices)

      if (params$objective == "survival:aft") {
        y_lower <- y_upper <- y_time
        y_upper[!y_event] <- Inf
        xgboost::setinfo(dmat, "label_lower_bound", y_lower)
        xgboost::setinfo(dmat, "label_upper_bound", y_upper)
      } else {
        params$aft_loss_distribution <- NULL
        if (params$objective == "survival:cox") {
          y <- ifelse(y_event, y_time, -y_time)
        }
        xgboost::setinfo(dmat, "label", y)
      }

      params$num_class <- if (params$objective == "multi:softprob") {
        length(y_levels)
      }

      scalable <- c("binary:logistic", "reg:logistic", "reg:squarederror")
      if (all(params$objective != scalable)) {
        params$scale_pos_weight <- NULL
      }
      
      if (all(params$feature_selector != c("greedy", "thrifty"))) {
        params$top_k <- NULL
      }

      res <- xgboost::xgb.train(
        data = dmat, params = params, nrounds = nrounds, verbose = verbose,
        print_every_n = print_every_n
      )
      attr(res, ".MachineShop") <- list(y_levels = y_levels)
      res
    },

    predict = function(object, newdata, times, .MachineShop, ...) {
      input <- .MachineShop$input
      xgb_predict <- function(newdata, outputmargin = FALSE) {
        newdmat <- xgboost::xgb.DMatrix(
          model.matrix(newdata, intercept = FALSE),
          base_margin = model.offset(newdata)
        )
        predict(object, newdata = newdmat, outputmargin = outputmargin)
      }
      params <- attr(object, "params")
      switch(params$objective,
        "multi:softprob" = {
          matrix(xgb_predict(newdata), nrow(newdata), byrow = TRUE)
        },
        "survival:aft" = {
          distr <- params$aft_loss_distribution
          if (distr == "normal") distr <- "gaussian"
          if (length(times)) {
            pred <- xgb_predict(newdata, outputmargin = TRUE)
            log_times <- matrix(
              log(times), length(pred), length(times), byrow = TRUE
            )
            quants <- log_times - pred
            surv_probs <- switch(distr,
              "extreme" = exp(-exp(quants)),
              "gaussian" = 1 - pnorm(quants),
              "logistic" = 1 / (1 + exp(quants)),
              throw(Error("Unsupported aft loss distribution \"", distr, "\"."))
            )
            SurvProbs(surv_probs, times = times, distr = distr)
          } else {
            SurvTimes(xgb_predict(newdata), distr = distr)
          }
        },
        "survival:cox" = {
          lp <- xgb_predict(PredictorFrame(input), outputmargin = TRUE)
          new_lp <- xgb_predict(newdata, outputmargin = TRUE)
          predict(
            response(input), lp, new_lp, times = times,
            weights = case_weights(input), ...
          )
        },
        xgb_predict(newdata)
      )
    },

    varimp = function(
      object, type = c("Gain", "Cover", "Frequency"), .MachineShop, ...
    ) {
      vi <- xgboost::xgb.importance(model = object, ...)
      structure(
        if (!is.null(vi$Weight)) {
          metric <- "coefficient"
          if (!is.null(vi$Class)) {
            vi_wide <- reshape(
              vi, idvar = "Feature", timevar = "Class", v.names = "Weight",
              varying = list(.MachineShop$y_levels), direction = "wide"
            )
            rownames(vi_wide) <- vi_wide[[1]]
            vi_wide[-1]
          } else {
            structure(vi$Weight, names = vi$Feature)
          }
        } else {
          type <- match.arg(type)
          metric <- tolower(type)
          structure(vi[[type]], names = vi$Feature)
        },
        metric = metric
      )
    }

  )

}

MLModelFunction(XGBModel) <- NULL


#' @rdname XGBModel
#'
XGBDARTModel <- function(
  eta = 0.3, gamma = 0, max_depth = 6, min_child_weight = 1,
  max_delta_step = .(0.7 * is(y, "PoissonVariate")), subsample = 1,
  colsample_bytree = 1, colsample_bylevel = 1, colsample_bynode = 1,
  alpha = 0, lambda = 1, tree_method = "auto", scale_pos_weight = 1,
  refresh_leaf = 1, grow_policy = "depthwise", max_leaves = 0, max_bin = 256,
  num_parallel_tree = 1, sample_type = "uniform", normalize_type = "tree",
  rate_drop = 0, one_drop = 0, skip_drop = 0, ...
) {
  .XGBModel(name = "XGBDARTModel", label = "Extreme Gradient Boosting (DART)",
            model = "dart", envir = environment(), ...)
}

MLModelFunction(XGBDARTModel) <- NULL


#' @rdname XGBModel
#'
XGBLinearModel <- function(
  alpha = 0, lambda = 0, updater = "shotgun", feature_selector = "cyclic",
  top_k = 0, ...
) {
  .XGBModel(
    name = "XGBLinearModel", label = "Extreme Gradient Boosting (Linear)",
    model = "gblinear", envir = environment(), ...
  )
}

MLModelFunction(XGBLinearModel) <- NULL


#' @rdname XGBModel
#'
XGBTreeModel <- function(
  eta = 0.3, gamma = 0, max_depth = 6, min_child_weight = 1,
  max_delta_step = .(0.7 * is(y, "PoissonVariate")), subsample = 1,
  colsample_bytree = 1, colsample_bylevel = 1, colsample_bynode = 1,
  alpha = 0, lambda = 1, tree_method = "auto", scale_pos_weight = 1,
  refresh_leaf = 1, grow_policy = "depthwise", max_leaves = 0, max_bin = 256,
  num_parallel_tree = 1, ...
) {
  .XGBModel(
    name = "XGBTreeModel", label = "Extreme Gradient Boosting (Tree)",
    model = "gbtree", envir = environment(), ...
  )
}

MLModelFunction(XGBTreeModel) <- NULL


.XGBModel <- function(name, label, model, envir, ...) {
  params <- c(as.list(envir), ...)
  params$booster <- model
  model <- do.call(XGBModel, params, quote = TRUE)
  model@name <- name
  model@label <- label

  gridinfo <- new_gridinfo(
    param = c(
      "nrounds", "eta", "gamma", "max_depth", "min_child_weight", "subsample",
      "colsample_bytree", "rate_drop", "skip_drop", "alpha", "lambda"
    ),
    get_values = c(
      function(n, ...) round_int(seq_range(0, 50, c(1, 1000), n + 1)),
      function(n, ...) seq(0.001, 0.6, length = n),
      function(n, ...) seq(0, 10, length = n),
      function(n, ...) seq_len(min(n, 10)),
      function(n, data, ...) seq(0, min(20, nrow(data)), length = n),
      function(n, ...) seq(0.25, 1, length = n),
      function(n, ...) seq(0.3, 0.8, length = n),
      function(n, ...) seq(0.01, 0.50, length = n),
      function(n, ...) seq(0.05, 0.95, length = n),
      function(n, ...) c(10^-seq_inner(0, 5, n - 1), 0),
      function(n, ...) c(10^-seq_inner(0, 5, n - 1), 0)
    ),
    default = c(TRUE, rep(FALSE, 2), TRUE, rep(FALSE, 5), rep(TRUE, 2))
  )
  grid_params <- switch(params$booster,
    "dart" = c(
      "nrounds", "eta", "gamma", "max_depth", "min_child_weight", "subsample",
      "colsample_bytree", "rate_drop", "skip_drop"
    ),
    "gblinear" = c("nrounds", "alpha", "lambda"),
    "gbtree" = c(
      "nrounds", "eta", "gamma", "max_depth", "min_child_weight", "subsample",
      "colsample_bytree"
    )
  )
  model@gridinfo <- gridinfo[gridinfo$param %in% grid_params, ]

  model
}

## Survival Analysis

test_that("test survival models output", {
  skip_if_not(TEST_ALL)

  library(MASS)
  library(recipes)
  library(survival)

  df <- within(Melanoma, status <- status != 2)
  fo <- Surv(time, status) ~ sex + age + year + thickness + ulcer
  rec <- recipe(time + status ~ sex + age + year + thickness + ulcer, data = df) %>%
    role_surv(time = time, event = status) %>%
    role_case(stratum = status)
  times <- 365 * c(2, 5, 10)

  output <- function(obs, pred) {
    print(pred)
    print(performance(obs, pred))
  }

  test_predict_all <- function(..., model) {

    cat("Model:", model@name, "\n")
    invisible(capture.output(model_fit <- fit(..., model = model)))
    obs <- response(model_fit)

    for (distr in c("emp", "exp", "ray", "wei")) {
      cat("\nPredicted means (distr = ", distr, ")\n", sep = "")
      output(obs, predict(model_fit, distr = distr))
      cat("\nPredicted probabilities (distr = ", distr, ")\n", sep = "")
      output(obs, predict(model_fit, times = times, type = "prob", distr = distr))
    }

    for (method in c("bre", "efr")) {
      cat("\nPredicted means (method = ", method, ")\n", sep = "")
      output(obs, predict(model_fit, method = method))
      cat("\nPredicted probabilities (method = ", method, ")\n", sep = "")
      output(obs, predict(model_fit, times = times, type = "prob", method = method))
    }

    cat("\nPredicted events:", "\n")
    output(obs, predict(model_fit, times = times))

  }

  test_predict_defaults <- function(..., model) {
    cat("Model:", model@name, "\n")
    invisible(capture.output(model_fit <- fit(..., model = model)))
    output(response(model_fit), predict(model_fit))
  }

  verify_output(test_path("test-survival.txt"), {
    skip_if_not(TEST_ALL)
    set.seed(123)
    cat("Recipe Specification", "\n")
    test_predict_all(rec, model = CoxModel())
    cat("Formula Specification", "\n")
    models <- c(CoxModel(), GBMModel(), CForestModel(), SurvRegModel(),
                XGBDARTModel(), XGBLinearModel(updater = "coord_descent"),
                XGBTreeModel(), XGBTreeModel(objective = "survival:aft"))
    for (model in models) {
      test_predict_all(fo, df, model = model)
    }
    models <- c(BARTModel(), BlackBoostModel(),
                GAMBoostModel(baselearner = "bols"), GLMBoostModel(),
                GLMNetModel(lambda = 0.05), RangerModel(), RPartModel())
    for (model in models) {
      test_predict_defaults(fo, df, model = model)
    }
  })

})

#' Resampling Controls
#'
#' Structures to define and control sampling methods for estimation of model
#' predictive performance in the \pkg{MachineShop} package.
#'
#' @name MLControl
#' @rdname MLControl
#' @aliases controls
#'
#' @param folds number of cross-validation folds (K).
#' @param prop proportion of cases to include in the training set
#'   (\code{0 < prop < 1}).
#' @param repeats number of repeats of the K-fold partitioning.
#' @param samples number of bootstrap samples.
#' @param weights logical indicating whether to return case weights in resampled
#'   output for the calculation of performance \link{metrics}.
#' @param seed integer to set the seed at the start of resampling.
#'
#' @return Object that inherits from the \code{MLControl} class.
#'
#' @seealso \code{\link{set_monitor}}, \code{\link{set_predict}},
#' \code{\link{set_strata}},
#' \code{\link{resample}}, \code{\link{SelectedInput}},
#' \code{\link{SelectedModel}}, \code{\link{TunedInput}},
#' \code{\link{TunedModel}}
#'
NULL


#' @rdname MLControl
#'
#' @details
#' \code{BootControl} constructs an \code{MLControl} object for simple bootstrap
#' resampling in which models are fit with bootstrap resampled training sets and
#' used to predict the full data set (Efron and Tibshirani 1993).
#'
#' @references
#' Efron, B., & Tibshirani, R. J. (1993). \emph{An introduction to the
#' bootstrap}. Chapman & Hall/CRC.
#'
#' @examples
#' ## Bootstrapping with 100 samples
#' BootControl(samples = 100)
#'
BootControl <- function(
  samples = 25, weights = TRUE, seed = sample(.Machine$integer.max, 1)
) {
  samples <- check_integer(samples, bounds = c(1, Inf), size = 1)
  throw(check_assignment(samples))

  new_control("BootControl",
    label = "Bootstrap Resampling",
    samples = samples, weights = weights, seed = seed
  ) %>% set_strata
}


#' @rdname MLControl
#'
#' @details
#' \code{BootOptimismControl} constructs an \code{MLControl} object for
#' optimism-corrected bootstrap resampling (Efron and Gong 1983, Harrell et al.
#' 1996).
#'
#' @references
#' Efron, B., & Gong, G. (1983). A leisurely look at the bootstrap, the
#' jackknife, and cross-validation. \emph{The American Statistician},
#' \emph{37}(1), 36-48.
#'
#' Harrell, F. E., Lee, K. L., & Mark, D. B. (1996). Multivariable prognostic
#' models: Issues in developing models, evaluating assumptions and adequacy, and
#' measuring and reducing errors. \emph{Statistics in Medicine}, \emph{15}(4),
#' 361-387.
#'
#' @examples
#' ## Optimism-corrected bootstrapping with 100 samples
#' BootOptimismControl(samples = 100)
#'
BootOptimismControl <- function(
  samples = 25, weights = TRUE, seed = sample(.Machine$integer.max, 1)
) {
  new("BootOptimismControl",
    BootControl(samples = samples, weights = weights, seed = seed),
    label = "Optimism-Corrected Bootstrap Resampling"
  )
}


#' @rdname MLControl
#'
#' @details
#' \code{CVControl} constructs an \code{MLControl} object for repeated K-fold
#' cross-validation (Kohavi 1995).  In this procedure, the full data set is
#' repeatedly partitioned into K-folds.  Within a partitioning, prediction is
#' performed on each of the K folds with models fit on all remaining folds.
#'
#' @references
#' Kohavi, R. (1995). A study of cross-validation and bootstrap for accuracy
#' estimation and model selection. In \emph{IJCAI'95: Proceedings of the 14th
#' International Joint Conference on Artificial Intelligence} (vol. 2, pp.
#' 1137-1143). Morgan Kaufmann Publishers Inc.
#'
#' @examples
#' ## Cross-validation with 5 repeats of 10 folds
#' CVControl(folds = 10, repeats = 5)
#'
CVControl <- function(
  folds = 10, repeats = 1, weights = TRUE,
  seed = sample(.Machine$integer.max, 1)
) {
  folds <- check_integer(folds, bounds = c(2, Inf), size = 1)
  throw(check_assignment(folds))

  repeats <- check_integer(repeats, bounds = c(1, Inf), size = 1)
  throw(check_assignment(repeats))

  new_control("CVControl",
    label = "K-Fold Cross-Validation",
    folds = folds, repeats = repeats, weights = weights, seed = seed
  ) %>% set_strata
}


#' @rdname MLControl
#'
#' @details
#' \code{CVOptimismControl} constructs an \code{MLControl} object for
#' optimism-corrected cross-validation resampling (Davison and Hinkley 1997,
#' eq. 6.48).
#'
#' @references
#' Davison, A. C., & Hinkley, D. V. (1997). \emph{Bootstrap methods and their
#' application}. Cambridge University Press.
#'
#' @examples
#' ## Optimism-corrected cross-validation with 5 repeats of 10 folds
#' CVOptimismControl(folds = 10, repeats = 5)
#'
CVOptimismControl <- function(
  folds = 10, repeats = 1, weights = TRUE,
  seed = sample(.Machine$integer.max, 1)
) {
  new("CVOptimismControl",
    CVControl(folds = folds, repeats = repeats, weights = weights, seed = seed),
    label = "Optimism-Corrected K-Fold Cross-Validation"
  )
}


NullControl <- function() {
  new("NullControl", label = "Null Control")
}


#' @rdname MLControl
#'
#' @details
#' \code{OOBControl} constructs an \code{MLControl} object for out-of-bootstrap
#' resampling in which models are fit with bootstrap resampled training sets and
#' used to predict the unsampled cases.
#'
#' @examples
#' ## Out-of-bootstrap validation with 100 samples
#' OOBControl(samples = 100)
#'
OOBControl <- function(
  samples = 25, weights = TRUE, seed = sample(.Machine$integer.max, 1)
) {
  samples <- check_integer(samples, bounds = c(1, Inf), size = 1)
  throw(check_assignment(samples))

  new_control("OOBControl",
    label = "Out-of-Bootstrap Resampling",
    samples = samples, weights = weights, seed = seed
  ) %>% set_strata
}


#' @rdname MLControl
#'
#' @details
#' \code{SplitControl} constructs an \code{MLControl} object for splitting data
#' into a separate training and test set (Hastie et al. 2009).
#'
#' @references
#' Hastie, T., Tibshirani, R., & Friedman, J. (2009). \emph{The elements of
#' statistical learning: data mining, inference, and prediction} (2nd ed.).
#' Springer.
#'
#' @examples
#' ## Split sample validation with 2/3 training and 1/3 testing
#' SplitControl(prop = 2/3)
#'
SplitControl <- function(
  prop = 2/3, weights = TRUE, seed = sample(.Machine$integer.max, 1)
) {
  prop <- check_numeric(prop, bounds = c(0, 1), include = FALSE, size = 1)
  throw(check_assignment(prop))

  new_control("SplitControl",
    label = "Split Training and Test Samples",
    prop = prop, weights = weights, seed = seed
  ) %>% set_strata
}


#' @rdname MLControl
#'
#' @details
#' \code{TrainControl} constructs an \code{MLControl} object for training and
#' performance evaluation to be performed on the same training set (Efron 1986).
#'
#' @references
#' Efron, B. (1986). How biased is the apparent error rate of a prediction rule?
#' \emph{Journal of the American Statistical Association}, \emph{81}(394),
#' 461-70.
#'
#' @examples
#' ## Training set evaluation
#' TrainControl()
#'
TrainControl <- function(
  weights = TRUE, seed = sample(.Machine$integer.max, 1)
) {
  new_control("TrainControl",
    label = "Training Resubstitution",
    weights = weights, seed = seed
  )
}


new_control <- function(class, ..., weights, seed) {
  weights <- check_logical(weights, size = 1)
  throw(check_assignment(weights))

  seed <- check_integer(seed, size = 1)
  throw(check_assignment(seed))

  new(class, ..., weights = weights, seed = seed) %>%
    set_monitor %>% set_predict
}

#' Model Fitting
#'
#' Fit a model to estimate its parameters from a data set.
#'
#' @name fit
#' @rdname fit-methods
#'
#' @param ... arguments passed from the generic function to its methods, from
#'   the \code{MLModel} and \code{MLModelFunction} methods to first arguments of
#'   others, and from others to the \code{ModelSpecification} method.  The
#'   first argument of each \code{fit} method is positional and, as such, must
#'   be given first in calls to them.
#' @param formula,data \link[=formula]{formula} defining the model predictor and
#'   response variables and a \link[=data.frame]{data frame} containing them.
#' @param x,y \link{matrix} and object containing predictor and response
#'   variables.
#' @param input \link[=inputs]{input} object defining and containing the model
#'   predictor and response variables.
#' @param object model \link[=ModelSpecification]{specification}.
#' @param model \link[=models]{model} function, function name, or object; or
#'   another object that can be \link[=as.MLModel]{coerced} to a model.  A model
#'   can be given first followed by any of the variable specifications.
#' @param verbose logical indicating whether to display printed output generated
#'   by some model-specific fit functions to aid in monitoring progress and
#'   diagnosing errors.
#'
#' @return \code{MLModelFit} class object.
#'
#' @seealso \code{\link{as.MLModel}}, \code{\link{response}},
#' \code{\link{predict}}, \code{\link{varimp}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' ## Survival response example
#' library(survival)
#'
#' gbm_fit <- fit(Surv(time, status) ~ ., data = veteran, model = GBMModel)
#' varimp(gbm_fit)
#' }
#'
fit <- function(...) {
  UseMethod("fit")
}


#' @rdname fit-methods
#'
fit.formula <- function(formula, data, model, ...) {
  fit(ModelSpecification(formula, data, model = model, control = NULL), ...)
}


#' @rdname fit-methods
#'
fit.matrix <- function(x, y, model, ...) {
  fit(ModelSpecification(x, y, model = model, control = NULL), ...)
}


#' @rdname fit-methods
#'
#' @details
#' User-specified case weights may be specified for \code{ModelFrames} upon
#' creation with the \code{\link[=ModelFrame]{weights}} argument in its
#' constructor.
#'
fit.ModelFrame <- function(input, model, ...) {
  fit(ModelSpecification(input, model = model, control = NULL), ...)
}


#' @rdname fit-methods
#'
#' @details
#' Variables in \code{recipe} specifications may be designated as case weights
#' with the \code{\link{role_case}} function.
#'
fit.recipe <- function(input, model, ...) {
  fit(ModelSpecification(input, model = model, control = NULL), ...)
}


#' @rdname fit-methods
#'
fit.ModelSpecification <- function(object, verbose = FALSE, ...) {
  if (is_optim_method(object)) {
    .fit_optim(object)
  } else {
    (if (verbose) identity else capture.output)(
      res <- .fit(as.MLInput(object), model = as.MLModel(object))
    )
    res
  }
}


#' @rdname fit-methods
#'
fit.MLModel <- function(model, ...) {
  fit(..., model = model)
}


#' @rdname fit-methods
#'
fit.MLModelFunction <- function(model, ...) {
  fit(as.MLModel(model), ...)
}


.fit <- function(object, ...) {
  UseMethod(".fit")
}


.fit.MLInput <- function(object, model, ...) {
  .fit(model, input = object)
}


.fit.MLModel <- function(object, input, ...) {
  input_prep <- prep(input)
  mf <- ModelFrame(input_prep, na.rm = object@na.rm)
  if (!nrow(mf)) {
    throw(Error("No case observations to fit."), call = call("fit"))
  }
  y <- response(mf)

  info <- data.frame(type = object@response_types, weights = object@weights)
  is_response_types <- is_response(y, info$type)
  if (!any(is_response_types)) {
    throw(TypeError(y, info$type, paste(object@name, "response variable")))
  }
  weights <- case_weights(mf)
  if (!all(info$weights[is_response_types]) || is.null(weights)) {
    throw(check_equal_weights(weights))
    mf <- ModelFrame(mf, weights = 1, na.rm = FALSE)
  }

  throw(check_packages(object@packages))

  params_env <- list2env(list(
    formula = formula(mf),
    data = mf,
    weights = case_weights(mf),
    y = y,
    nobs = nrow(mf),
    nvars = nvars(mf, object)
  ), parent = new.env(parent = asNamespace("MachineShop")))
  environment(params_env$formula) <- params_env

  args <- c(mget(c("formula", "data", "weights"), params_env), object@params)

  do.call(object@fit, args, envir = params_env) %>%
    MLModelFit(paste0(object@name, "Fit"), input = input_prep, model = object)
}


.fit_optim <- function(object, ...) {
  mloptim <- get_optim_field(object)
  throw(check_packages(mloptim@packages))
  tryCatch(
    optim(mloptim@fun, object, ...),
    error = function(cond) {
      msg <- conditionMessage(cond)
      if (is(mloptim, "SequentialOptimization")) {
        object <- set_optim_grid(
          object, random = mloptim@random, progress = mloptim@monitor$progress
        )
        throw(LocalWarning(
          mloptim@label, " failed: ", msg, "\n",
          "Performing a ", tolower(get_optim_field(object, "label")),
          " instead."
        ))
        fit(object, ...)
      } else {
        throw(Error(msg), call = sys.call(-4))
      }
    }
  )
}


eval_fit <- function(data = NULL, formula, matrix) {
  expr <- if (is(data, "ModelFrame") && is(terms(data), "ModelDesignTerms")) {
    bquote({
      x <- model.matrix(data, intercept = FALSE)
      y <- response(data)
      .(substitute(matrix))
    })
  } else {
    bquote({
      data <- as.data.frame(formula, data = data)
      .(substitute(formula))
    })
  }
  eval.parent(expr)
}

#' Neural Network Model
#'
#' Fit single-hidden-layer neural network, possibly with skip-layer connections.
#'
#' @param size number of units in the hidden layer.
#' @param linout switch for linear output units.  Set automatically according to
#'   the class type of the response variable [numeric: \code{TRUE}, other:
#'   \code{FALSE}].
#' @param entropy switch for entropy (= maximum conditional likelihood) fitting.
#' @param softmax switch for softmax (log-linear model) and maximum conditional
#'   likelihood fitting.
#' @param censored a variant on softmax, in which non-zero targets mean possible
#'   classes.
#' @param skip switch to add skip-layer connections from input to output.
#' @param rang Initial random weights on [\code{-rang}, \code{rang}].
#' @param decay parameter for weight decay.
#' @param maxit maximum number of iterations.
#' @param trace switch for tracing optimization.
#' @param MaxNWts maximum allowable number of weights.
#' @param abstol stop if the fit criterion falls below \code{abstol}, indicating
#'   an essentially perfect fit.
#' @param reltol stop if the optimizer is unable to reduce the fit criterion by
#'   a factor of at least \code{1 - reltol}.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
#'     \code{size}, \code{decay}
#'   }
#' }
#'
#' Default argument values and further model details can be found in the source
#' See Also link below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[nnet]{nnet}}, \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' fit(sale_amount ~ ., data = ICHomes, model = NNetModel)
#'
NNetModel <- function(
  size = 1, linout = logical(), entropy = logical(), softmax = logical(),
  censored = FALSE, skip = FALSE, rang = 0.7, decay = 0, maxit = 100,
  trace = FALSE, MaxNWts = 1000, abstol = 1e-4, reltol = 1e-8
) {

  MLModel(

    name = "NNetModel",
    label = "Feed-Forward Neural Networks",
    packages = "nnet",
    response_types = c("factor", "numeric"),
    weights = TRUE,
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = c("size", "decay"),
      get_values = c(
        function(n, ...) round_int(seq_range(1, 2, c(1, 20), n)),
        function(n, ...) c(0, 10^seq_inner(-5, 1, n - 1))
      )
    ),

    fit = function(formula, data, weights, linout = NULL, ...) {
      y <- response(data)
      if (is.null(linout)) linout <- is_response(y, "numeric")
      if (is(terms(data), "ModelDesignTerms")) {
        x <- model.matrix(data, intercept = FALSE)
        if (is_response(y, "binary")) {
          y <- as.numeric(y) - 1
        } else if (is_response(y, "factor")) {
          y <- structure(
            model.matrix(~ y - 1),
            dimnames = list(names(y), levels(y))
          )
        }
        res <- nnet::nnet(x, y, weights = weights, linout = linout, ...)
        res$terms <- terms(data)
        res$coefnames <- colnames(x)
        res$xlevels <- list()
        res
      } else {
        nnet::nnet(
          formula, data = as.data.frame(formula, data = data),
          weights = weights, na.action = na.pass, linout = linout, ...
        )
      }
    },

    predict = function(object, newdata, ...) {
      newdata <- as.data.frame(newdata)
      predict(object, newdata = newdata, type = "raw")
    },

    varimp = function(object, ...) {
      ninputs <- object$n[1]
      size <- object$n[2]
      noutputs <- object$n[3]

      beta <- abs(coef(object))

      inds <- expand.grid(input = seq_len(ninputs), hidden = seq_len(size))
      labels <- paste0("i", inds$input, "->h", inds$hidden)
      i2h <- matrix(beta[match(labels, names(beta))], ninputs, size)
      rownames(i2h) <- object$coefnames

      inds <- expand.grid(
        hidden = seq_len(size),
        output = if (noutputs == 1) "" else seq_len(noutputs)
      )
      labels <- paste0("h", inds$hidden, "->o", inds$output)
      h2o <- matrix(beta[match(labels, names(beta))], size, noutputs)
      colnames(h2o) <- colnames(object$residuals)

      structure(drop(i2h %*% h2o), metric = "coefficient")
    }

  )

}

MLModelFunction(NNetModel) <- NULL

#' Classification and Regression Tree Models
#'
#' A tree is grown by binary recursive partitioning using the response in the
#' specified formula and choosing splits from the terms of the right-hand-side.
#'
#' @param mincut minimum number of observations to include in either child node.
#' @param minsize smallest allowed node size: a weighted quantity.
#' @param mindev within-node deviance must be at least this times that of the
#'   root node for the node to be split.
#' @param split splitting criterion to use.
#' @param k scalar cost-complexity parameter defining a subtree to return.
#' @param best integer alternative to \code{k} requesting the number of terminal
#'   nodes of a subtree in the cost-complexity sequence to return.
#' @param method character string denoting the measure of node heterogeneity
#'   used to guide cost-complexity pruning.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric}}
#' }
#'
#' Further model details can be found in the source link below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[tree]{tree}}, \code{\link[tree]{prune.tree}},
#' \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package tree to run
#'
#' fit(Species ~ ., data = iris, model = TreeModel)
#' }
#'
TreeModel <- function(
  mincut = 5, minsize = 10, mindev = 0.01, split = c("deviance", "gini"),
  k = numeric(), best = integer(), method = c("deviance", "misclass")
) {

  split <- match.arg(split)
  method <- if (length(k) || length(best)) match.arg(method)

  MLModel(

    name = "TreeModel",
    label = "Regression and Classification Trees",
    packages = "tree",
    response_types = c("factor", "numeric"),
    weights = TRUE,
    predictor_encoding = "model.frame",
    na.rm = FALSE,
    params = new_params(environment()),

    fit = function(
      formula, data, weights, split, k = NULL, best = NULL, method = NULL, ...
    ) {
      res <- tree::tree(
        formula, data = as.data.frame(formula, data = data), weights = weights,
        split = split, ...
      )
      if (!is.null(method)) {
        tree::prune.tree(res, k = k, best = best, method = method)
      } else res
    },

    predict = function(object, newdata, ...) {
      newdata <- as.data.frame(newdata)
      predict(object, newdata = newdata)
    }

  )

}

MLModelFunction(TreeModel) <- NULL

TEST_ALL <- Sys.getenv("TEST_ALL") == "true"
TEST_MODEL_FITTING <- TEST_ALL || Sys.getenv("TEST_MODEL_FITTING") == "true"
TEST_MODEL_COMPARISONS <-
  TEST_ALL || Sys.getenv("TEST_MODEL_COMPARISONS") == "true"
TEST_TRAINING <- TEST_ALL || Sys.getenv("TEST_MODEL_TUNING") == "true"


data(Boston, package = "MASS", envir = environment())
data(Pima.tr, package = "MASS", envir = environment())
data(Melanoma, package = "MASS", envir = environment())


test_model <- function(formula, data, model, times = numeric()) {

  modelfit <- fit(formula, data, model)
  print(modelfit)

  vi <- varimp(modelfit)
  print(vi)
  print(plot(vi))

  pred <- head(predict(modelfit, data))
  print(pred)
  pred_prob <- head(predict(modelfit, data, type = "raw"))
  print(pred_prob)

  pred_times <- head(predict(modelfit, data, times = times))
  print(pred_times)
  pred_times_prob <- head(predict(modelfit, data, times = times, type = "raw"))
  print(pred_times_prob)

  perf_boot <- resample(formula, data, model, BootOptimismControl)
  print(perf_boot)
  perf_cv <- resample(formula, data, model, CVControl)
  print(perf_cv)
  perf_oob <- resample(formula, data, model, OOBControl)
  print(perf_oob)

  print(summary(perf_cv))
  print(plot(perf_cv))

  control <- BootOptimismControl() %>% set_predict(times = times)
  perf_boot_times <- resample(formula, data, model, control)
  print(perf_boot_times)
  control <- CVControl() %>% set_predict(times = times)
  perf_cv_times <- resample(formula, data, model, control)
  print(perf_cv_times)
  control <- OOBControl() %>% set_predict(times = times)
  perf_oob_times <- resample(formula, data, model, control)
  print(perf_oob_times)

  print(summary(perf_cv_times))
  print(plot(perf_cv_times))

}


test_model_binary <- function(model) {
  test_model(factor(type) ~ ., data = Pima.tr, model = model)
}


test_model_factor <- function(model) {
  test_model(factor(Species) ~ ., data = iris, model = model)
}


test_model_numeric <- function(model) {
  test_model(medv ~ ., data = Boston, model = model)
}


test_model_ordered <- function(model) {
  df <- Boston
  df$medv <- cut(Boston$medv, breaks = c(0, 15, 20, 25, 50), ordered = TRUE)
  test_model(medv ~ ., data = df, model = model)
}


test_model_Surv <- function(model) {
  test_model(survival::Surv(time, status != 2) ~ sex + age + year + thickness + ulcer,
             data = Melanoma, model = model, times = 365 * c(2, 5, 10))
}

test_that("model fitting", {
  skip_if_not(TEST_MODEL_FITTING)
  with_parallel({
    model <- GLMNetModel(lambda = 0.01)
    expect_output(test_model_binary(model))
    expect_output(test_model_factor(model))
    expect_output(test_model_numeric(model))
    expect_output(test_model_ordered(model))
    expect_output(test_model_Surv(model))
  })
})

test_that("model fitting", {
  skip_if_not(TEST_MODEL_FITTING)
  with_parallel({
    models <- c("RFSRCModel", "RFSRCFastModel")
    for (model in models) {
      expect_output(test_model_binary(model))
      expect_output(test_model_factor(model))
      expect_output(test_model_numeric(model))
      expect_output(test_model_ordered(model))
      expect_output(test_model_Surv(model))
    }
  })
})

test_that("model fitting", {
  skip_if_not(TEST_MODEL_FITTING)
  with_parallel({

    model <- SuperModel(GBMModel, SVMRadialModel, KNNModel)
    expect_output(test_model_binary(model))
    expect_output(test_model_factor(model))
    expect_output(test_model_numeric(model))

    model <- SuperModel(RandomForestModel, POLRModel)
    expect_output(test_model_ordered(model))

    model <- SuperModel(GBMModel, GLMNetModel(lambda = 0.01))
    expect_output(test_model_Surv(model))

  })
})

test_that("model fitting", {
  skip_if_not(TEST_MODEL_FITTING)
  with_parallel({
    model <- "BlackBoostModel"
    expect_output(test_model_binary(model))
    expect_error(test_model_factor(model))
    expect_output(test_model_numeric(model))
    expect_error(test_model_ordered(model))
    expect_output(test_model_Surv(model))
  })
})

#' Bayesian Additive Regression Trees Model
#'
#' Flexible nonparametric modeling of covariates for continuous, binary,
#' categorical and time-to-event outcomes.
#'
#' @param K if provided, then coarsen the times of survival responses per the
#'   quantiles \eqn{1/K, 2/K, ..., K/K} to reduce computational burdern.
#' @param sparse logical indicating whether to perform variable selection based
#'   on a sparse Dirichlet prior rather than simply uniform; see Linero 2016.
#' @param theta,omega \eqn{theta} and \eqn{omega} parameters; zero means
#'   random.
#' @param a,b sparse parameters for \eqn{Beta(a, b)} prior:
#'   \eqn{0.5 <= a <= 1} where lower values induce more sparsity and typically
#'   \eqn{b = 1}.
#' @param rho sparse parameter: typically \eqn{rho = p} where \eqn{p} is the
#'   number of covariates under consideration.
#' @param augment whether data augmentation is to be performed in sparse
#'   variable selection.
#' @param xinfo optional matrix whose rows are the covariates and columns their
#'   cutpoints.
#' @param usequants whether covariate cutpoints are defined by uniform quantiles
#'   or generated uniformly.
#' @param sigest normal error variance prior for numeric response variables.
#' @param sigdf degrees of freedom for error variance prior.
#' @param sigquant quantile at which a rough estimate of the error standard
#'   deviation is placed.
#' @param lambda scale of the prior error variance.
#' @param k number of standard deviations \eqn{f(x)}  is away from +/-3 for
#'   categorical response variables.
#' @param power,base power and base parameters for tree prior.
#' @param tau.num numerator in the \eqn{tau} definition, i.e.,
#'   \eqn{tau = tau.num / (k * sqrt(ntree))}.
#' @param offset override for the default \eqn{offset} of \eqn{F^-1(mean(y))}
#'   in the multivariate response probability
#'   \eqn{P(y[j] = 1 | x) = F(f(x)[j] + offset[j])}.
#' @param ntree number of trees in the sum.
#' @param numcut number of possible covariate cutoff values.
#' @param ndpost number of posterior draws returned.
#' @param nskip number of MCMC iterations to be treated as burn in.
#' @param keepevery interval at which to keep posterior draws.
#' @param printevery interval at which to print MCMC progress.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}, \code{numeric}, \code{Surv}}
#' }
#'
#' Default argument values and further model details can be found in the source
#' See Also links below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[BART]{gbart}}, \code{\link[BART]{mbart}},
#' \code{\link[BART]{surv.bart}}, \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package BART to run
#'
#' fit(sale_amount ~ ., data = ICHomes, model = BARTModel)
#' }
#'
BARTModel <- function(
  K = integer(), sparse = FALSE, theta = 0, omega = 1, a = 0.5, b = 1,
  rho = numeric(), augment = FALSE, xinfo = matrix(NA, 0, 0), usequants = FALSE,
  sigest = NA, sigdf = 3, sigquant = 0.90, lambda = NA, k = 2, power = 2,
  base = 0.95, tau.num = numeric(), offset = numeric(), ntree = integer(),
  numcut = 100, ndpost = 1000, nskip = integer(), keepevery = integer(),
  printevery = 1000
) {

  MLModel(

    name = "BARTModel",
    label = "Bayesian Additive Regression Trees",
    packages = "BART",
    response_types = c("factor", "numeric", "Surv"),
    weights = c(FALSE, TRUE, FALSE),
    predictor_encoding = "model.matrix",
    na.rm = "response",
    params = new_params(environment()),

    fit = function(
      formula, data, weights, K = NULL, sigest, sigdf, sigquant, lambda, ...
    ) {
      x <- model.matrix(data, intercept = FALSE)
      y <- response(data)
      switch_class(y,
        "factor" = {
          if (nlevels(y) <= 2) {
            f <- BART::gbart
            y <- as.numeric(y) - 1
          } else {
            f <- BART::mbart
            y <- as.numeric(y)
          }
          f(x.train = x, y.train = y, type = "pbart", ...)
        },
        "numeric" = {
          BART::gbart(
            x.train = x, y.train = y, w = weights, sigest = sigest,
            sigdf = sigdf, sigquant = sigquant, lambda = lambda, ...
          )
        },
        "Surv" = {
          throw(check_censoring(y, "right"))
          BART::surv.bart(
            x.train = x, times = y[, "time"], delta = y[, "status"], K = K, ...
          )
        }
      )
    },

    predict = function(object, newdata, ...) {
      newx <- model.matrix(newdata, intercept = FALSE)
      if (is(object, "mbart")) {
        predict(object, newdata = newx)$prob.test.mean %>%
          matrix(nrow = nrow(newx), ncol = object$K, byrow = TRUE)
      } else if (is(object, "pbart")) {
        predict(object, newdata = newx)$prob.test.mean
      } else if (is(object, "survbart")) {
        N <- nrow(newx)
        K <- object$K
        newx <- cbind(t = object$times, newx[rep(seq_len(N), each = K), ])
        pred <- predict(object, newdata = newx)$surv.test.mean %>%
          matrix(nrow = N, ncol = K, byrow = TRUE)
        predict(Surv(object$times), pred, ...)
      } else if (is(object, "wbart")) {
        colMeans(predict(object, newdata = newx))
      }
    }

  )

}

MLModelFunction(BARTModel) <- NULL

#' Model Performance Plots
#'
#' Plot measures of model performance and predictor variable importance.
#'
#' @name plot
#' @rdname plot-methods
#'
#' @param x \link{calibration}, \link{confusion}, \link{lift},
#'   trained model \link{fit}, partial \link{dependence}, \link{performance},
#'   \link[=curves]{performance curve}, \link{resample}, \link{rfe}, or
#'   \link[=varimp]{variable importance} result.
#' @param diagonal logical indicating whether to include a diagonal reference
#'   line.
#' @param find numeric true positive rate at which to display reference lines
#'   identifying the corresponding rates of positive predictions.
#' @param metrics vector of numeric indexes or character names of performance
#'   metrics to plot.
#' @param n number of most important variables to include in the plot.
#' @param se logical indicating whether to include standard error bars.
#' @param stat function or character string naming a function to compute a
#'   summary statistic on resampled metrics for trained \code{MLModel} line
#'   plots and \code{Resample} model ordering.  The original ordering is
#'   preserved if a value of \code{NULL} is given.  For \code{LiftCurve} and
#'   \code{PerformanceCurve} classes, plots are of resampled metrics aggregated
#'   by the statistic if given or of resample-specific metrics if \code{NULL}.
#' @param stats vector of numeric indexes or character names of partial
#'   dependence summary statistics to plot.
#' @param type type of plot to construct.
#' @param ... arguments passed to other methods.
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' ## Factor response example
#'
#' fo <- Species ~ .
#' control <- CVControl()
#'
#' gbm_fit <- fit(fo, data = iris, model = GBMModel, control = control)
#' plot(varimp(gbm_fit))
#'
#' gbm_res1 <- resample(fo, iris, GBMModel(n.trees = 25), control)
#' gbm_res2 <- resample(fo, iris, GBMModel(n.trees = 50), control)
#' gbm_res3 <- resample(fo, iris, GBMModel(n.trees = 100), control)
#' plot(gbm_res3)
#'
#' res <- c(GBM1 = gbm_res1, GBM2 = gbm_res2, GBM3 = gbm_res3)
#' plot(res)
#' }
#'
NULL


#' @rdname plot-methods
#'
plot.Calibration <- function(x, type = c("line", "point"), se = FALSE, ...) {
  type <- match.arg(type)

  args <- alist(.data[["Predicted"]], .data[["Mean"]])
  if (nlevels(x$Response) > 1) {
    args$color <- args$fill <- .(.data[["Response"]])
  }
  mapping <- do.call(aes, args)

  position <- "identity"

  pl <- by(x, x$Model, function(cal) {

    df <- data.frame(
      Response = cal$Response,
      Predicted = cal$Predicted,
      cal$Observed
    )
    Predicted_width <- diff(range(df$Predicted, finite = TRUE))

    p <- ggplot(df, mapping) +
      geom_abline(intercept = 0, slope = 1, color = "gray") +
      labs(title = cal$Model[1], x = "Predicted", y = "Observed Mean")

    if (se) if (x@smoothed) {
      p <- p +
        geom_ribbon(
          aes(ymin = .data[["Lower"]], ymax = .data[["Upper"]]),
          linetype = "blank", alpha = 0.2, na.rm = TRUE
        )
    } else {
      position <- position_dodge(width = 0.025 * Predicted_width)
      p <- p +
        geom_errorbar(
          aes(ymin = .data[["Lower"]], ymax = .data[["Upper"]]),
          width = 0.05 * Predicted_width, position = position, na.rm = TRUE
        )
    }

    switch(type,
      "line" = p + geom_line(position = position, na.rm = TRUE),
      "point" = p + geom_point(position = position, na.rm = TRUE)
    )

  }, simplify = FALSE)

  structure(as(pl, "list"), names = levels(x$Model))
}


#' @rdname plot-methods
#'
plot.ConfusionList <- function(x, ...) {
  pl <- list()
  for (name in names(x)) {
    pl[[name]] <- plot(x[[name]]) + labs(title = name)
  }
  pl
}


#' @rdname plot-methods
#'
plot.ConfusionMatrix <- function(x, ...) {
  df <- as.data.frame(proportions(as(x, "table")), responseName = "Value")
  df$Predicted <- factor(df$Predicted, rev(levels(df$Predicted)))
  ggplot(
    df, aes(.data[["Observed"]], .data[["Predicted"]], fill = .data[["Value"]])
  ) +
    geom_raster() +
    labs(fill = "Probability") +
    scale_fill_gradient(trans = "reverse") +
    coord_fixed()
}


#' @rdname plot-methods
#'
plot.LiftCurve <- function(
  x, find = numeric(), diagonal = TRUE,
  stat = MachineShop::settings("stat.Curve"), ...
) {
  x <- summary(x, stat = stat)
  p <- plot(as(x, "PerformanceCurve"), diagonal = diagonal, stat = NULL)

  if (length(find)) {
    find <- check_numeric(find, bounds = c(0, 1), include = FALSE, size = 1)
    throw(check_assignment(find))

    indices <- x["Model"]
    indices$Iteration <- x$Iteration
    tested <- by(x, indices, function(data) {
      approx(data$y, data$x, find, ties = "ordered")$y
    })

    df <- data.frame(
      x = as.numeric(tested),
      y = find,
      Model = dimnames(tested)$Model
    )
    df$Iteration <- rep(dimnames(tested)$Iteration, each = size(tested, 1))

    p <- p +
      geom_segment(
        aes(x = .data[["x"]], y = 0, xend = .data[["x"]], yend = .data[["y"]]),
        df
      ) +
      geom_segment(
        aes(x = 0, y = .data[["y"]], xend = .data[["x"]], yend = .data[["y"]]),
        df
      )
  }

  p
}


#' @rdname plot-methods
#'
plot.MLModel <- function(
  x, metrics = NULL, stat = MachineShop::settings("stat.TrainingParams"),
  type = c("boxplot", "density", "errorbar", "line", "violin"), ...
) {
  if (!is_trained(x)) throw(Error("No training results to plot."))
  map(function(step) {
    plot(step, metrics = metrics, stat = stat, type = type, ...)
  }, x@steps)
}


#' @rdname plot-methods
#'
plot.PartialDependence <- function(x, stats = NULL, ...) {
  if (any(rowSums(!is.na(x$Predictors)) > 1)) {
    throw(Error(
      "Partial dependence plots not available for interaction efffects."
    ))
  }

  if (!is.null(stats)) {
    stats <- match_indices(stats, levels(x$Statistic))
    x <- x[x$Statistic %in% stats, , drop = FALSE]
  }

  df <- x[c("Statistic", "Response", "Value")]

  args <- alist(.data[["Predictor"]], .data[["Value"]])
  if (nlevels(x$Response) > 1) args$color <- .(.data[["Response"]])
  mapping <- do.call(aes, args)

  pl <- list()
  for (varname in names(x$Predictors)) {
    df$Predictor <- x$Predictors[[varname]]
    p <- ggplot(na.omit(df), mapping)
    p <- switch_class(df$Predictor,
      "factor" = p +
        geom_crossbar(aes(ymin = after_stat(y), ymax = after_stat(y))),
      "numeric" = p + geom_line() + geom_point()
    ) + labs(x = varname) + facet_wrap(~ Statistic, scales = "free")
    pl[[varname]] <- p
  }
  pl
}


#' @rdname plot-methods
#'
plot.Performance <- function(
  x, metrics = NULL, stat = MachineShop::settings("stat.Resample"),
  type = c("boxplot", "density", "errorbar", "violin"), ...
) {
  df <- as.data.frame(x)
  if (is.null(df$Model)) df$Model <- factor("Model")

  metric_levels <- levels(df$Metric)
  if (is.null(metrics)) {
    metrics <- metric_levels
  } else {
    metrics <- match_indices(metrics, metric_levels)
    df <- df[df$Metric %in% metrics, , drop = FALSE]
  }
  df$Metric <- factor(df$Metric, metrics)

  if (is.null(stat)) stat <- function(x) 0
  stat <- check_stat(stat, convert = TRUE)
  throw(check_assignment(stat))

  df_metric <- df[df$Metric == metrics[1], , drop = FALSE]
  sorted_levels <- tapply(
    df_metric$Value, df_metric$Model, function(x) stat(na.omit(x))
  ) %>% sort %>% names
  df$Model <- factor(df$Model, sorted_levels)

  p <- ggplot(df)
  mapping <- aes(.data[["Model"]], .data[["Value"]])
  switch(match.arg(type),
    "boxplot" = p +
      geom_boxplot(mapping) +
      stat_summary(mapping, fun = mean, geom = "point") +
      labs(x = "") +
      coord_flip(),
    "density" = p +
      geom_density(aes(.data[["Value"]], color = .data[["Model"]])) +
      labs(y = "Density", color = ""),
    "errorbar" = p +
      stat_summary(mapping, fun.data = mean_se, geom = "errorbar") +
      stat_summary(mapping, fun = mean, geom = "point") +
      labs(x = "") +
      coord_flip(),
    "violin" = p +
      geom_violin(mapping) +
      stat_summary(mapping, fun = mean, geom = "point") +
      labs(x = "") +
      coord_flip()
  ) + facet_wrap(~ Metric, scales = "free")
}


#' @rdname plot-methods
#'
plot.PerformanceCurve <- function(
  x, type = c("tradeoffs", "cutoffs"), diagonal = FALSE,
  stat = MachineShop::settings("stat.Curve"), ...
) {
  x <- summary(x, stat = stat)

  args <- alist(.data[["x"]], .data[["y"]])
  if (nlevels(x$Model) > 1) args$color <- .(.data[["Model"]])
  if (!is.null(x$Iteration)) {
    args$group <- .(interaction(.data[["Model"]], .data[["Iteration"]]))
  }
  mapping <- do.call(aes, args)

  labels <- c(x = x@metrics$x@label, y = x@metrics$y@label)

  switch(match.arg(type),
    "tradeoffs" = {
      x$Cutoff <- NULL
      p <- ggplot(na.omit(x), mapping) +
        geom_path() +
        labs(x = labels["x"], y = labels["y"])

      if (diagonal) {
        p <- p + geom_abline(intercept = 0, slope = 1, color = "gray")
      }

      p
    },
    "cutoffs" = {
      df <- reshape(
        x, varying = c("x", "y"), v.names = "y", times = labels,
        timevar = "Metric", direction = "long"
      )
      names(df)[names(df) == "Cutoff"] <- "x"

      ggplot(na.omit(df), mapping) +
        geom_line() +
        labs(x = "Cutoff", y = "Performance") +
        facet_wrap(~ Metric)
    }
  )
}


#' @rdname plot-methods
#'
plot.Resample <- function(
  x, metrics = NULL, stat = MachineShop::settings("stat.Resample"),
  type = c("boxplot", "density", "errorbar", "violin"), ...
) {
  plot(performance(x), metrics = metrics, stat = stat, type = type)
}


#' @rdname plot-methods
#'
plot.TrainingStep <- function(
  x, metrics = NULL, stat = MachineShop::settings("stat.TrainingParams"),
  type = c("boxplot", "density", "errorbar", "line", "violin"), ...
) {
  type <- match.arg(type)
  if (type == "line") {
    stat <- check_stat(stat, convert = TRUE)
    throw(check_assignment(stat))

    params <- unnest_params(summary(x)$params)
    stats <- performance(x) %>%
      apply(c(3, 2), function(x) stat(na.omit(x))) %>%
      TabularArray %>%
      as.data.frame
    df <- data.frame(
      x = params[[1]],
      Value = stats$Value,
      Metric = stats$Metric
    )

    metric_levels <- levels(df$Metric)
    if (is.null(metrics)) {
      metrics <- metric_levels
    } else {
      metrics <- match_indices(metrics, metric_levels)
      df <- df[df$Metric %in% metrics, , drop = FALSE]
    }
    df$Metric <- factor(df$Metric, metrics)

    indices <- map("logi", function(x) length(unique(x)), params[-1])
    args <- alist(.data[["x"]], .data[["Value"]])
    if (any(indices)) {
      df$Group <- interaction(params[-1][indices], sep = ":")
      args$color <- args$shape <- .(.data[["Group"]])
    } else {
      args$group <- 1
    }
    mapping <- do.call(aes, args)

    ggplot(df, mapping) +
      geom_line(stat = "summary", fun = mean) +
      geom_point(stat = "summary", fun = mean) +
      labs(x = names(params)[1]) +
      facet_wrap(~ Metric, scales = "free")
  } else {
    plot(performance(x), metrics = metrics, stat = stat, type = type, ...)
  }
}


#' @rdname plot-methods
#'
plot.VariableImportance <- function(x, n = Inf, ...) {
  x <- update(x)
  scales <- if (any(is.na(x@scale))) "free" else "fixed"
  x <- head(x, n)
  var_names <- rownames(x)
  df <- cbind(stack(x), variables = factor(var_names, rev(var_names)))
  ggplot(df, aes(.data[["variables"]], .data[["values"]])) +
    geom_bar(stat = "identity") +
    labs(x = "Variable", y = "Importance") +
    coord_flip() +
    facet_wrap(~ ind, scales = scales)
}

#' Linear Components Variable Reduction
#'
#' Creates a \emph{specification} of a recipe step that will compute one or more
#' linear combinations of a set of numeric variables according to a
#' user-specified transformation matrix.
#'
#' @param recipe \link[recipes]{recipe} object to which the step will be added.
#' @param ... one or more selector functions to choose which variables will be
#'   used to compute the components.  See \code{\link[recipes]{selections}} for
#'   more details.  These are not currently used by the \code{tidy} method.
#' @param transform function whose first argument \code{x} is a matrix of
#'   variables with which to compute linear combinations and second argument
#'   \code{step} is the current step.  The function should return a
#'   transformation \code{\link{matrix}} or \code{\link[Matrix]{Matrix}} of
#'   variable weights in its columns, or return a list with element
#'   \code{`weights`} containing the transformation matrix and possibly with
#'   other elements to be included as attributes in output from the \code{tidy}
#'   method.
#' @param num_comp number of components to derive.  The value of \code{num_comp}
#'   will be constrained to a minimum of 1 and maximum of the number of original
#'   variables when \code{\link[recipes]{prep}} is run.
#' @param options list of elements to be added to the step object for use in the
#'   \code{transform} function.
#' @param center,scale logicals indicating whether to mean center and standard
#'   deviation scale the original variables prior to deriving components, or
#'   functions or names of functions for the centering and scaling.
#' @param replace logical indicating whether to replace the original variables.
#' @param prefix character string prefix added to a sequence of zero-padded
#'   integers to generate names for the resulting new variables.
#' @param role analysis role that added step variables should be assigned.  By
#'   default, they are designated as model predictors.
#' @param skip logical indicating whether to skip the step when the recipe is
#'   baked.  While all operations are baked when \code{\link[recipes]{prep}} is
#'   run, some operations may not be applicable to new data (e.g. processing
#'   outcome variables).  Care should be taken when using \code{skip = TRUE} as
#'   it may affect the computations for subsequent operations.
#' @param id unique character string to identify the step.
#' @param x \code{step_lincomp} object.
#'
#' @return An updated version of \code{recipe} with the new step added to the
#' sequence of existing steps (if any).  For the \code{tidy} method, a tibble
#' with columns \code{terms} (selectors or variables selected), \code{weight}
#' of each variable in the linear transformations, and \code{name} of the new
#' variable names.
#'
#' @seealso \code{\link[recipes]{recipe}}, \code{\link[recipes]{prep}},
#' \code{\link[recipes]{bake}}
#'
#' @examples
#' library(recipes)
#'
#' pca_mat <- function(x, step) {
#'   prcomp(x)$rotation[, 1:step$num_comp, drop = FALSE]
#' }
#'
#' rec <- recipe(rating ~ ., data = attitude)
#' lincomp_rec <- rec %>%
#'   step_lincomp(all_numeric_predictors(),
#'                transform = pca_mat, num_comp = 3, prefix = "PCA")
#'
#' lincomp_prep <- prep(lincomp_rec, training = attitude)
#' lincomp_data <- bake(lincomp_prep, attitude)
#'
#' pairs(lincomp_data, lower.panel = NULL)
#'
#' tidy(lincomp_rec, number = 1)
#' tidy(lincomp_prep, number = 1)
#'
step_lincomp <- function(
  recipe, ..., transform, num_comp = 5, options = list(), center = TRUE,
  scale = TRUE, replace = TRUE, prefix = "LinComp", role = "predictor",
  skip = FALSE, id = recipes::rand_id("lincomp")
) {

  recipes::add_step(recipe, new_step_lincomp(
    terms = recipes::ellipse_check(...),
    transform = transform,
    num_comp = num_comp,
    options = options,
    center = center,
    scale = scale,
    replace = replace,
    prefix = prefix,
    role = role,
    skip = skip,
    id = id
  ))

}


new_step_lincomp <- function(
  terms, transform, num_comp, options, center, scale, replace, prefix, role,
  skip, id
) {
  stopifnot(is.function(transform))
  stopifnot(is.list(options))
  step_args <- list(
    subclass = "lincomp",
    terms = terms,
    transform = transform,
    num_comp = num_comp,
    center = if (!is.logical(center)) fget(center) else
      if (center) base::mean else FALSE,
    scale = if (!is.logical(scale)) fget(scale) else
      if (scale) stats::sd else FALSE,
    res = tibble(
      terms = recipes::sel2char(terms),
      weight = NA_real_,
      name = NA_character_
    ),
    trained = FALSE,
    replace = replace,
    prefix = prefix,
    role = role,
    skip = skip,
    id = id
  )
  invalid_names <- intersect(names(options), names(step_args))
  if (length(invalid_names)) {
    throw(Error(note_items(
      "Options list contains reserved step name{?s}: ", invalid_names, "."
    )))
  }
  do.call(recipes::step, c(step_args, options))
}


prep.step_lincomp <- function(x, training, info = NULL, ...) {

  col_names <- recipes::recipes_eval_select(x$terms, training, info)
  training <- training[col_names]
  recipes::check_type(training)

  if (is.function(x$center)) x$center <- apply(training, 2, x$center)
  if (is.function(x$scale)) x$scale <- apply(training, 2, x$scale)
  training <- scale(training, center = x$center, scale = x$scale)

  x$num_comp <- min(max(x$num_comp, 1), ncol(training))

  res <- x$transform(x = training, step = x)
  if (!is.list(res)) res <- list(weights = res)
  if (!(is(res$weights, "matrix") || is(res$weights, "Matrix"))) {
    throw(Error("Transform matrix should return a matrix or Matrix object."))
  }
  if (nrow(res$weights) != ncol(training)) {
    throw(Error(
      "Transform matrix row length should equal the number of variables."
    ))
  }
  dimnames(res$weights) <- list(
    terms = colnames(training),
    names = recipes::names0(ncol(res$weights), x$prefix)
  )
  x$res <- res
  x$trained <- TRUE
  x

}


bake.step_lincomp <- function(object, new_data, ...) {

  weights <- object$res$weights
  is_lincomp_var <- names(new_data) %in% rownames(weights)

  lincomp_data <- scale(
    new_data[is_lincomp_var], center = object$center, scale = object$scale
  )
  comps <- as_tibble(as.matrix(lincomp_data %*% weights))

  if (object$replace) new_data <- new_data[!is_lincomp_var]
  comps <- recipes::check_name(
    comps, new_data, object, newname = colnames(weights)
  )

  as_tibble(c(new_data, comps))

}


print.step_lincomp <- function(x, width = console_width(), ...) {
  title <- paste(x$prefix, "variable reduction for ")
  recipes::print_step(
    rownames(x$res$weights), x$terms, x$trained, title = title,
    width = max(width - nchar(title), 20)
  )
  invisible(x)
}


#' @rdname step_lincomp
#'
tidy.step_lincomp <- function(x, ...) {
  res <- x$res
  if (is_trained(x)) {
    res_attrs <- res[names(res) != "weights"]
    res <- res$weights %>% as.matrix %>% as.table %>%
      as.data.frame(responseName = "weight")
    res <- tibble(res[c("terms", "weight")], name = res$names)
    attributes(res) <- c(attributes(res), res_attrs)
  }
  res$id <- x$id
  res
}


#' @rdname step_lincomp
#'
tunable.step_lincomp <- function(x, ...) {
  tibble(
    name = c("num_comp"),
    call_info = list(
      list(pkg = "dials", fun = "num_comp", range = c(1, 4))
    ),
    source = "MachineShop",
    component = "step_lincomp",
    component_id = x$id
  )
}

#' Display Model Information
#'
#' Display information about models supplied by the \pkg{MachineShop} package.
#'
#' @param ... \link[=models]{model} functions, function names, or objects;
#' \link[=response]{observed responses} for which to display information.  If
#' none are specified, information is returned on all available models by
#' default.
#'
#' @return List of named model elements each containing the following
#' components:
#' \describe{
#'   \item{label}{character descriptor for the model.}
#'   \item{packages}{character vector of source packages required to use the
#'     model.  These need only be installed with the
#'     \code{\link{install.packages}} function or by equivalent means; but need
#'     not be loaded with, for example, the \code{\link{library}} function.}
#'   \item{response_types}{character vector of response variable types supported
#'     by the model.}
#'  \item{weights}{logical value or vector of the same length as
#'    \code{response_types} indicating whether case weights are supported for
#'    the responses.}
#'   \item{arguments}{closure with the argument names and corresponding default
#'     values of the model function.}
#'   \item{grid}{logical indicating whether automatic generation of tuning
#'     parameter grids is implemented for the model.}
#'   \item{varimp}{logical indicating whether model-specific variable importance
#'     is defined.}
#' }
#'
#' @examples
#' ## All models
#' modelinfo()
#'
#' ## Models by response types
#' names(modelinfo(factor(0)))
#' names(modelinfo(factor(0), numeric(0)))
#'
#' ## Model-specific information
#' modelinfo(GBMModel)
#'
modelinfo <- function(...) {
  args <- list(...)
  args <- if (length(args)) unname(args) else as.list(settings("models"))
  info <- do.call(.modelinfo, args)

  is_type <- if (length(info)) !map("logi", is, info, "list") else NULL
  if (any(is_type)) {
    info_models <- if (all(is_type)) modelinfo() else info[!is_type]
    info_types <- do.call(.modelinfo_types, info[is_type])
    info <- c(info_models, info_types)
    info <- info[intersect(names(info_models), names(info_types))]
  }

  info[unique(names(info))]
}


.modelinfo <- function(x, ...) {
  UseMethod(".modelinfo")
}


.modelinfo.default <- function(x, ...) {
  info <- list(x)
  if (...length()) c(info, .modelinfo(...)) else info
}


.modelinfo.character <- function(x, ...) {
  model <- tryCatch(
    as.MLModel(x),
    error = function(cond) list()
  )
  .modelinfo(model, ...)
}


.modelinfo.function <- function(x, ...) {
  model <- tryCatch(
    as.MLModel(x),
    error = function(cond) list()
  )
  .modelinfo(model, ...)
}


.modelinfo.list <- function(x, ...) {
  if (...length()) .modelinfo(...) else list()
}


.modelinfo.MLModel <- function(x, ...) {
  info <- structure(list(list(
    label = x@label,
    packages = x@packages,
    response_types = x@response_types,
    weights = x@weights,
    na.rm = x@na.rm,
    arguments = args(get0(x@name, mode = "function")),
    grid = has_grid(x),
    varimp = has_varimp(x)
  )), names = x@name)
  if (...length()) c(info, .modelinfo(...)) else info
}


.modelinfo_types <- function(...) {
  info <- modelinfo()
  check_model <- function(model) {
    check_response <- function(y) any(is_response(y, model$response_types))
    all(map("logi", check_response, list(...)))
  }
  info[map("logi", check_model, info)]
}

#' Model Expansion Over Tuning Parameters
#'
#' Expand a model over all combinations of a grid of tuning parameters.
#'
#' @param object \link[=models]{model} function, function name, or object; or
#'   another object that can be \link[=as.MLModel]{coerced} to a model.
#' @param ... named vectors or factors or a list of these containing the
#'   parameter values over which to expand \code{object}.
#' @param random number of points to be randomly sampled from the parameter grid
#'   or \code{FALSE} if all points are to be returned.
#'
#' @return \code{list} of expanded models.
#'
#' @seealso \code{\link{SelectedModel}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' data(Boston, package = "MASS")
#'
#' models <- expand_model(GBMModel, n.trees = c(50, 100),
#'                                  interaction.depth = 1:2)
#'
#' fit(medv ~ ., data = Boston, model = SelectedModel(models))
#' }
#'
expand_model <- function(object, ..., random = FALSE) {
  .expand_model(object, ..., random = random)
}


.expand_model <- function(object, ...) {
  UseMethod(".expand_model")
}


.expand_model.default <- function(object, ..., random) {
  expand_model(as.MLModel(object), ..., random = random)
}


.expand_model.tbl_df <- function(object, model, ...) {
  models <- map(function(params) {
    new_id <- make_id("model")
    do.call(update, list(model, params = params, id = new_id), quote = TRUE)
  }, split(object, seq_len(max(nrow(object), 1))))
  names(models) <- paste0(models[[1]]@name, ".", names(models))
  models
}


.expand_model.MLModel <- function(object, ..., random) {
  expand_model(expand_params(..., random = random), object)
}


#' Model Tuning Grid Expansion
#'
#' Expand a model grid of tuning parameter values.
#'
#' @name expand_modelgrid
#' @rdname expand_modelgrid-methods
#'
#' @param ... arguments passed from the generic function to its methods and from
#'   the \code{MLModel} and \code{MLModelFunction} methods to others.  The
#'   first argument of each \code{expand_modelgrid} method is positional and, as
#'   such, must be given first in calls to them.
#' @param formula,data \link[=formula]{formula} defining the model predictor and
#'   response variables and a \link[=data.frame]{data frame} containing them.
#' @param x,y \link{matrix} and object containing predictor and response
#'   variables.
#' @param input \link[=inputs]{input} object defining and containing the model
#'   predictor and response variables.
#' @param object model \link[=ModelSpecification]{specification}.
#' @param model \link[=models]{model} function, function name, or object; or
#'   another object that can be \link[=as.MLModel]{coerced} to a model.  A model
#'   can be given first followed by any of the variable specifications.
#' @param info logical indicating whether to return model-defined grid
#'   construction information rather than the grid values.
#'
#' @details
#' The \code{expand_modelgrid} function enables manual extraction and viewing of
#' grids created automatically when a \code{\link{TunedModel}} is fit.
#'
#' @return A data frame of parameter values or \code{NULL} if data are required
#' for construction of the grid but not supplied.
#'
#' @seealso \code{\link{TunedModel}}
#'
#' @examples
#' expand_modelgrid(TunedModel(GBMModel, grid = 5))
#'
#' \donttest{
#' ## Requires prior installation of suggested package glmnet to run
#' expand_modelgrid(TunedModel(GLMNetModel, grid = c(alpha = 5, lambda = 10)),
#'                  sale_amount ~ ., data = ICHomes)
#' }
#'
#' gbm_grid <- ParameterGrid(
#'   n.trees = dials::trees(),
#'   interaction.depth = dials::tree_depth(),
#'   size = 5
#' )
#' expand_modelgrid(TunedModel(GBMModel, grid = gbm_grid))
#'
#' rf_grid <- ParameterGrid(
#'   mtry = dials::mtry(),
#'   nodesize = dials::max_nodes(),
#'   size = c(3, 5)
#' )
#' expand_modelgrid(TunedModel(RandomForestModel, grid = rf_grid),
#'                  sale_amount ~ ., data = ICHomes)
#'
expand_modelgrid <- function(...) {
  UseMethod("expand_modelgrid")
}


#' @rdname expand_modelgrid-methods
#'
expand_modelgrid.formula <- function(formula, data, model, info = FALSE, ...) {
  expand_modelgrid(model, formula, data, info = info)
}


#' @rdname expand_modelgrid-methods
#'
expand_modelgrid.matrix <- function(x, y, model, info = FALSE, ...) {
  .expand_modelgrid(as.MLModel(model), x, y, info = info)
}


#' @rdname expand_modelgrid-methods
#'
expand_modelgrid.ModelFrame <- function(input, model, info = FALSE, ...) {
  .expand_modelgrid(as.MLModel(model), input, info = info)
}


#' @rdname expand_modelgrid-methods
#'
expand_modelgrid.recipe <- function(input, model, info = FALSE, ...) {
  .expand_modelgrid(as.MLModel(model), input, info = info)
}


#' @rdname expand_modelgrid-methods
#'
expand_modelgrid.ModelSpecification <- function(object, ...) {
  grid <- object@grid

  if (is_empty(grid)) {

    types <- c("SelectedInput", "SelectedModel")
    found <- unique(unlist(map_slots(function(slot) {
      types[map("logi", is, list(slot), types)]
    }, object)))
    if (length(found)) {
      throw(Warning(
        "Cannot expand the tuning grids of a ModelSpecification containing ",
        as_string(found, conj = "or"), "."
      ))
      return(NULL)
    }

    get_modelgrids <- function(input) {
      map_slots(function(model) {
        get_grid(model, input)$params
      }, object, names = c("model", "candidates"))
    }

    make_grid <- function(grids) {
      expand_params(grids[!map("logi", is_empty, grids)])
    }

    input_grid <- make_grid(map_slots(function(input) {
      get_grid(input)$params
    }, object, names = c("input", "candidates")))

    if (nrow(input_grid)) {
      grid <- NULL
      for (i in seq_len(nrow(input_grid))) {
        params <- input_grid[i, ]
        input <- update_slots(object@input, params = params)
        grid <- rbind(grid, make_grid(c(params, get_modelgrids(input))))
      }
    } else {
      grid <- make_grid(get_modelgrids(object@input))
    }

  }

  grid
}


#' @rdname expand_modelgrid-methods
#'
expand_modelgrid.MLModel <- function(model, ...) {
  .expand_modelgrid(model, ...)
}


#' @rdname expand_modelgrid-methods
#'
expand_modelgrid.MLModelFunction <- function(model, ...) {
  expand_modelgrid(as.MLModel(model), ...)
}


.expand_modelgrid <- function(object, ...) {
  UseMethod(".expand_modelgrid")
}


.expand_modelgrid.MLModel <- function(object, ..., info = FALSE) {
  if (info) object@gridinfo else tibble()
}


.expand_modelgrid.TunedModel <- function(object, ..., info = FALSE) {
  if (info) {
    NextMethod()
  } else {
    random <- if (is_optim_method(object, "RandomGridSearch")) {
      get_optim_field(object, "size")
    } else FALSE
    .expand_modelgrid(object@grid, ..., model = object@model, random = random)
  }
}


.expand_modelgrid.TuningGrid <- function(object, input, ..., model, random) {
  gridinfo <- model@gridinfo
  size <- object@size
  if (object@random) random <- object@random
  mf <- NULL

  not_dup <- function(x) !duplicated(x, fromLast = TRUE)
  if (!is.null(names(size))) {
    size <- size[gridinfo$param] * not_dup(gridinfo$param)
    size[is.na(size)] <- 0L
  } else if (length(size) == 1) {
    if (!random) gridinfo <- gridinfo[gridinfo$default, ]
    size <- size * not_dup(gridinfo$param)
  }
  gridinfo$size <- size
  gridinfo <- gridinfo[gridinfo$size >= 1, ]

  has_data_arg <- function(fun) "data" %in% names(formals(fun))
  needs_data <- any(map("logi", has_data_arg, gridinfo$get_values))
  if (needs_data && has_grid(model)) {
    if (!missing(input)) mf <- ModelFrame(input, ..., na.rm = FALSE)
    if (is.null(mf)) return(NULL)
    y <- response(mf)
    if (!any(is_response(y, model@response_types))) {
      throw(LocalWarning(
        "Invalid model response type in expand_modelgrid().\n",
        "x Exising ", model@name, " supports ",
        note_items("type{?s}: ", model@response_types), ".\n",
        "x Supplied response is of ", note_items("type{?s}: ", class(y)), "."
      ))
      return(NULL)
    }
  }

  param_names <- unique(gridinfo$param)
  params <- map(
    function(fun, n) unique(fun(n = n, data = mf)),
    gridinfo$get_values, gridinfo$size
  )
  params <- map(function(name) {
    unlist(params[gridinfo$param == name], use.names = FALSE)
  }, param_names)
  names(params) <- param_names
  params[lengths(params) == 0] <- NULL
  grid <- expand_params(params, random = random)

  .expand_modelgrid(grid)
}


.expand_modelgrid.ParameterGrid <- function(object, input, ..., model, random) {
  grid <- if (nrow(object)) {
    needs_data <- any(dials::has_unknowns(object$object))
    if (needs_data) {
      if (missing(input)) return(NULL)
      mf <- ModelFrame(input, ..., na.rm = FALSE)
      data <- switch(model@predictor_encoding,
        "model.frame" = {
          mf_terms <- attributes(terms(mf))
          var_list <- eval(mf_terms$variables, mf)
          names(var_list) <- rownames(mf_terms$factors)
          as.data.frame(var_list[-c(1, mf_terms$offset)])
        },
        "model.matrix" = model.matrix(mf, intercept = FALSE)
      )
      object <- dials::finalize(object, x = data)
    }
    params <- map(dials::value_seq, object$object, object@size)
    expand_params(params, random = if (object@random) object@random else random)
  } else {
    tibble()
  }

  .expand_modelgrid(grid)
}


.expand_modelgrid.tbl_df <- function(object, ...) {
  if (ncol(object)) object[!duplicated(object), ] else object
}


#' Model Parameters Expansion
#'
#' Create a grid of parameter values from all combinations of supplied inputs.
#'
#' @param ... named data frames or vectors or a list of these containing the
#'   parameter values over which to create the grid.
#' @param random number of points to be randomly sampled from the parameter grid
#'   or \code{FALSE} if all points are to be returned.
#'
#' @return A data frame containing one row for each combination of the supplied
#' inputs.
#'
#' @seealso \code{\link{TunedModel}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' data(Boston, package = "MASS")
#'
#' grid <- expand_params(
#'   n.trees = c(50, 100),
#'   interaction.depth = 1:2
#' )
#'
#' fit(medv ~ ., data = Boston, model = TunedModel(GBMModel, grid = grid))
#' }
#'
expand_params <- function(..., random = FALSE) {
  x <- list(...)
  if (is_one_element(x, "list") && is.null(names(x))) x <- x[[1]]
  if (random) {
    random_grid(x, size = random)
  } else {
    regular_grid(x)
  }
}


#' Recipe Step Parameters Expansion
#'
#' Create a grid of parameter values from all combinations of lists supplied for
#' steps of a preprocessing recipe.
#'
#' @param ... one or more lists containing parameter values over which to create
#'   the grid.  For each list an argument name should be given as the \code{id}
#'   of the \link[recipes]{recipe} step to which it corresponds.
#' @param random number of points to be randomly sampled from the parameter grid
#'   or \code{FALSE} if all points are to be returned.
#'
#' @return \code{RecipeGrid} class object that inherits from \code{data.frame}.
#'
#' @seealso \code{\link{TunedInput}}
#'
#' @examples
#' library(recipes)
#' data(Boston, package = "MASS")
#'
#' rec <- recipe(medv ~ ., data = Boston) %>%
#'   step_corr(all_numeric_predictors(), id = "corr") %>%
#'   step_pca(all_numeric_predictors(), id = "pca")
#'
#' expand_steps(
#'   corr = list(threshold = c(0.8, 0.9),
#'               method = c("pearson", "spearman")),
#'   pca = list(num_comp = 1:3)
#' )
#'
expand_steps <- function(..., random = FALSE) {

  steps <- list(...)
  step_names <- names(steps)
  if (is_one_element(steps, "list") && is.null(step_names)) {
    steps <- steps[[1]]
    step_names <- names(steps)
  }

  if (!all(map("logi", is.list, steps))) {
    throw(Error("Step arguments must be lists."))
  }

  get_names <- function(x) {
    res <- NULL
    if (is.list(x)) {
      for (i in seq_along(x)) {
        name <- names(x[i])
        if (is.null(name)) name <- ""
        res <- c(res, name, get_names(x[[i]]))
      }
    }
    res
  }

  if (!all(nzchar(get_names(steps)))) {
    throw(Error("All steps and their parameters must be named."))
  } else if (any(duplicated(step_names))) {
    throw(Error("Step names must be unique."))
  }

  steps <- map(expand_params, steps, random = random)
  RecipeGrid(expand_params(steps, random = random))

}

#' ModelFrame Class
#'
#' Class for storing data, formulas, and other attributes for \pkg{MachineShop}
#' model fitting.
#'
#' @name ModelFrame
#' @rdname ModelFrame-methods
#'
#' @param ... arguments passed from the generic function to its methods.  The
#'   first argument of each \code{ModelFrame} method is positional and, as such,
#'   must be given first in calls to them.
#' @param formula,data \link[=formula]{formula} defining the model predictor and
#'   response variables and a \link[=data.frame]{data frame} containing them.
#'   In the associated method, arguments \code{groups}, \code{strata}, and
#'   \code{weights} will be evaluated as expressions, whose objects are searched
#'   for first in the accompanying \code{data} environment and, if not found
#'   there, next in the calling environment.
#' @param x,y \link{matrix} and object containing predictor and response
#'   variables.
#' @param offsets numeric vector, matrix, or data frame of values to be added
#'   with a fixed coefficient of 1 to linear predictors in compatible regression
#'   models.
#' @param groups vector of values defining groupings of case observations, such
#'   as repeated measurements, to keep together during resampling [default:
#'   none].
#' @param strata vector of values to use in conducting stratified
#'   \link{resample} estimation of model performance [default: none].
#' @param weights numeric vector of non-negative case weights for the \code{y}
#'   response variable [default: equal weights].
#' @param na.rm character string or logical specifying removal of \code{"all"}
#'   (\code{TRUE}) cases with missing values, \code{"none"} (\code{FALSE}), or
#'   only those whose missing values are in the \code{"response"} variable.
#'
#' @return \code{ModelFrame} class object that inherits from \code{data.frame}.
#'
#' @seealso \code{\link{fit}}, \code{\link{resample}}, \code{\link{response}},
#' \code{\link{SelectedInput}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' mf <- ModelFrame(ncases / (ncases + ncontrols) ~ agegp + tobgp + alcgp,
#'                  data = esoph, weights = ncases + ncontrols)
#' gbm_fit <- fit(mf, model = GBMModel)
#' varimp(gbm_fit)
#' }
#'
ModelFrame <- function(...) {
  UseMethod("ModelFrame")
}


ModelFrame.data.frame <- function(data, ...) {
  args <- list()
  for (type in c("groups", "names", "strata", "weights")) {
    name <- paste0("(", type, ")")
    args[[type]] <- data[[name]]
    data[[name]] <- NULL
  }
  if (is.null(args$names)) args$names <- rownames(data)
  args$na.rm <- FALSE

  model_terms <- terms(
    map(as.name, names(data)[-1]), names(data)[1],
    all_numeric = all(map("logi", is.numeric, data[-1]))
  )

  do.call(ModelFrame, c(list(model_terms, data), args))
}


#' @rdname ModelFrame-methods
#'
ModelFrame.formula <- function(
  formula, data, groups = NULL, strata = NULL, weights = NULL, na.rm = TRUE, ...
) {
  invalid_calls <- setdiff(
    inline_calls(predictors(formula)), settings("RHS.formula")
  )
  if (length(invalid_calls)) {
    throw(Error(note_items(
      "Unsupported predictor variable function{?s}: ", invalid_calls,
      "; use a recipe or include transformed predictors in the data frame."
    )))
  }

  data <- as.data.frame(data)

  args <- map(
    function(x) eval(x, data, parent.frame()),
    eval(substitute(alist(
      groups = groups,
      strata = strata,
      weights = weights,
      ...
    )))
  )
  args$names <- rownames(data)
  args$na.rm <- na.rm

  model_terms <- terms(formula, data = data)
  data[[deparse1(response(model_terms))]] <- response(model_terms, data)
  data <- data[all.vars(model_formula(model_terms))]

  do.call(ModelFrame, c(list(model_terms, data), args))
}


#' @rdname ModelFrame-methods
#'
ModelFrame.matrix <- function(
  x, y = NULL, offsets = NULL, groups = NULL, strata = NULL, weights = NULL,
  na.rm = TRUE, ...
) {
  data <- as.data.frame(x)
  colnames(x) <- names(data)
  if (!is.null(offsets)) {
    offsets <- if (is.vector(offsets, "numeric")) {
      data.frame(offset = offsets)
    } else if (is.matrix(offsets)) {
      if (is.null(colnames(offsets))) {
        colnames(offsets) <- rep("offset", ncol(offsets))
      }
      data.frame(offsets)
    } else if (!is.data.frame(offsets)) {
      throw(Error(
        "Value of `offsets` must be a numeric vector, matrix, or data frame."
      ))
    }
    data <- data.frame(data, offsets)
    offsets <- tail(names(data), length(offsets))
  }
  model_terms <- terms(x, y, offsets = offsets)
  data[[deparse1(response(model_terms))]] <- y
  end <- length(data)
  data <- data[c(end, seq_len(end - 1))]

  ModelFrame(
    model_terms, data, names = rownames(data), groups = groups, strata = strata,
    weights = weights, na.rm = na.rm, ...
  )
}


ModelFrame.ModelFrame <- function(input, na.rm = TRUE, ...) {
  comps <- list(...)
  for (type in names(comps)) {
    input[[paste0("(", type, ")")]] <- comps[[type]]
  }
  na.rm <- throw(check_na.rm(na.rm))
  switch(na.rm,
    "all" = input <- na.omit(input),
    "response" = {
      y <- response(input)
      if (!is.null(y)) input <- input[complete.cases(y), ]
    }
  )
  input
}


ModelFrame.ModelTerms <- function(object, data, ...) {
  mf <- new("ModelFrame", as.data.frame(data))
  object@id <- mf@id
  attr(mf, "terms") <- object
  ModelFrame(mf, ...)
}


ModelFrame.recipe <- function(input, ...) {
  input <- prep(input)
  data <- bake(input, NULL)

  args <- list()
  for (type in c("groups", "names", "strata", "weights")) {
    name <- case_comp_name(input, type)
    args[[type]] <- if (length(name)) data[[name]]
  }
  if (is.null(args$names)) args$names <- rownames(data)
  args$na.rm <- FALSE

  model_terms <- terms(input)
  data[[deparse1(response(model_terms))]] <- response(model_terms, data)
  data <- data[all.vars(model_formula(model_terms))]

  do.call(ModelFrame, c(list(model_terms, data), args))
}


update.ModelFrame <- function(object, data = NULL, ...) {
  if (is.data.frame(data)) {
    attr(data, "terms") <- terms(object)
    S3Part(object) <- data
  }
  object
}


#################### ModelFrame Formulas ####################


formula.ModelFrame <- function(x, ...) {
  model_formula(terms(x))
}


formula.ModelTerms <- function(x, ...) {
  formula(asS3(x))
}


inline_calls <- function(x) {
  if (is.call(x)) {
    call_name <- as.character(x[[1]])
    unique(c(call_name, unlist(map(inline_calls, x[-1]))))
  }
}


model_formula <- function(x) {
  x <- formula(x)
  response <- response(x)
  if (!is.null(response)) x[[2]] <- as.name(deparse1(response))
  x
}


#################### ModelFrame Terms ####################


terms.formula <- function(x, ...) {
  ModelFormulaTerms(structure(
    stats::terms.formula(x, ...),
    .Environment = asNamespace("MachineShop")
  ))
}


terms.list <- function(
  x, y = NULL, intercept = TRUE, all_numeric = FALSE, ...
) {
  if (is.character(y)) y <- as.name(y)
  has_y <- 1L - is.null(y)

  is_names <- map("logi", is.name, x)
  name_inds <- which(is_names)
  noname_inds <- which(!is_names)

  x_char <- character(length(x))
  x_char[name_inds] <- as.character(x[name_inds])
  x_char[noname_inds] <- map("char", deparse, x[noname_inds])

  valid_calls <- map("logi", function(var) {
    is.call(var) && var[[1]] == "offset"
  }, x[noname_inds])
  if (!all(valid_calls)) {
    throw(Error("Non-offset calls in variable specifications."))
  }
  offsets <- has_y + noname_inds

  fo <- str2lang(
    paste(
      "y ~",
      if (length(x)) paste(x_char, collapse = "+") else if (intercept) "1",
      if (!intercept) "- 1"
    )
  )
  fo[[2]] <- y

  class <- if (all_numeric) {
    x_char[name_inds] <- map("char", as.character, x[name_inds])
    "ModelDesignTerms"
  } else {
    "ModelFormulaTerms"
  }

  var_names <- c(if (has_y) deparse1(y), x_char)
  label_names <- x_char[name_inds]

  if (class == "ModelDesignTerms") {
    factors <- cbind(as.integer(var_names %in% label_names))
    rownames(factors) <- var_names
  } else {
    factors <- matrix(
      0L, length(var_names), length(label_names),
      dimnames = list(var_names, label_names)
    )
    term_match <- match(var_names, label_names, nomatch = 0L)
    factors[cbind(var_names[term_match > 0], label_names[term_match])] <- 1L
  }

  new(class,
    structure(
      fo,
      variables = as.call(c(.(list), y, x)),
      offset = if (length(offsets)) offsets,
      factors = factors,
      term.labels = label_names,
      order = rep(1L, length(label_names)),
      intercept = as.integer(intercept),
      response = has_y,
      .Environment = asNamespace("MachineShop"),
      class = c("terms", "formula")
    )
  )
}


terms.matrix <- function(x, y = NULL, offsets = NULL, ...) {
  stopifnot(is.character(colnames(x)))
  stopifnot(!anyDuplicated(colnames(x)))

  labels <- colnames(x)
  response <- if (!is.null(y)) make_unique(c(labels, "y"))[length(labels) + 1]
  labels <- map(as.name, labels)

  if (length(offsets)) {
    offsets <- paste0("offset(", offsets, ")")
    labels <- c(labels, map(str2lang, offsets))
  }

  terms(labels, response, all_numeric = is.numeric(x))
}


terms.ModelFrame <- function(x, ...) {
  attr(x, "terms")
}


terms.ModelTerms <- function(x, ...) {
  x
}


terms.recipe <- function(x, original = FALSE, ...) {
  terms.recipe_info(summary(x, original = original))
}


terms.recipe_info <- function(x, ...) {

  first <- function(x) head(x, 1)
  get_vars <- function(roles = character(), types = character()) {
    is_match <- by(x, x$variable, function(split) {
      valid_types <- is_empty(types) || any(types %in% unlist(split$type))
      all(roles %in% split$role) && valid_types
    })
    names(is_match)[is_match]
  }

  binom <- c(
    count = first(get_vars(c("binom_x", "outcome"), "numeric")),
    size = first(get_vars(c("binom_size", "outcome"), "numeric"))
  )
  surv <- c(
    time = first(get_vars(c("surv_time", "outcome"), "numeric")),
    event = first(get_vars(c("surv_event", "outcome"), c("logical", "numeric")))
  )
  matrix <- setdiff(get_vars("outcome", "numeric"), c(binom, surv))
  if (length(matrix) == 1) matrix <- character()
  other <- setdiff(get_vars("outcome"), c(binom, surv, matrix))

  num_roles <- sum(lengths(list(binom, surv, matrix, other)) > 0)
  if (num_roles > 1 || length(other) > 1) {
    throw(Error(
      "Specified outcome is not a single variable, binomial variable with ",
      "roles \"binom_x\" and \"binom_size\", survival variables with roles ",
      "\"surv_time\" and \"surv_event\", or multiple numeric variables."
    ))
  }

  outcome <- if (length(other)) {
    other
  } else if (!is.na(surv["time"])) {
    args <- as.name(surv["time"])
    if (!is.na(surv["event"])) args <- c(args, as.name(surv["event"]))
    as.call(c(.(Surv), args))
  } else if (length(surv)) {
    throw(Error(
      "Survival outcome role \"surv_event\" specified without \"surv_time\"."
    ))
  } else if (all(!is.na(binom[c("count", "size")]))) {
    call("BinomialVariate", as.name(binom["count"]), as.name(binom["size"]))
  } else if (length(binom)) {
    throw(Error(
      "Binomial outcome must have \"binom_x\" and \"binom_size\" roles."
    ))
  } else if (length(matrix)) {
    as.call(c(.(cbind), map(as.name, matrix)))
  }

  is_predictor <- x$role == "predictor"
  predictors <- map(as.name, x$variable[is_predictor])
  all_numeric <- all(x$type[is_predictor] == "numeric")

  offsets <- get_vars("pred_offset", "numeric")
  if (length(offsets)) {
    offsets <- paste0("offset(", offsets, ")")
    predictors <- c(predictors, map(str2lang, offsets))
  }

  terms(predictors, outcome, all_numeric = all_numeric)

}


#################### ModelFrame Design Matrices ####################


model.matrix.ModelDesignTerms <- function(object, data, ...) {
  data <- data[labels(object)]
  assign <- seq_len(ncol(data))
  if (attr(object, "intercept")) {
    data <- cbind("(Intercept)" = 1, data)
    assign <- c(0, assign)
  }
  structure(as.matrix(data), assign = assign)
}


model.matrix.ModelFormulaTerms <- function(object, data, ...) {
  fo <- delete.response(object)
  mf <- model.frame(fo, data, na.action = na.pass)
  model.matrix.default(fo, mf, ...)
}


model.matrix.ModelFrame <- function(object, intercept = logical(), ...) {
  model_terms <- terms(object)
  if (length(intercept)) {
    attr(model_terms, "intercept") <- as.integer(intercept)
  }
  model.matrix(model_terms, object, ...)
}


model.matrix.SelectedInput <- function(object, ...) {
  throw(Error("Cannot create a design matrix from a ", class(object), "."))
}


model.offset <- function(x) {
  stopifnot(is(x, "ModelFrame"))
  keep <- 1 + c(0, attr(terms(x), "offset"))
  offsets <- eval(attr(terms(x), "variables")[keep], x)
  Reduce("+", offsets)
}

#' MLModel and MLModelFunction Class Constructors
#'
#' Create a model or model function for use with the \pkg{MachineShop} package.
#'
#' @param name character name of the object to which the model is assigned.
#' @param label optional character descriptor for the model.
#' @param packages character vector of package names upon which the model
#'   depends.  Each name may be optionally followed by a comment in
#'   parentheses specifying a version requirement.  The comment should contain
#'   a comparison operator, whitespace and a valid version number, e.g.
#'   \code{"xgboost (>= 1.3.0)"}.
#' @param response_types character vector of response variable types to which
#'   the model can be fit.  Supported types are \code{"binary"},
#'   \code{"BinomialVariate"}, \code{"DiscreteVariate"}, \code{"factor"},
#'   \code{"matrix"}, \code{"NegBinomialVariate"}, \code{"numeric"},
#'   \code{"ordered"}, \code{"PoissonVariate"}, and \code{"Surv"}.
#' @param weights logical value or vector of the same length as
#'    \code{response_types} indicating whether case weights are supported for
#'    the responses.
#' @param predictor_encoding character string indicating whether the model is
#'   fit with predictor variables encoded as a \code{"\link{model.frame}"},
#'   a \code{"\link{model.matrix}"}, or unspecified (default).
#' @param na.rm character string or logical specifying removal of \code{"all"}
#'   (\code{TRUE}) cases with missing values from model fitting and prediction,
#'   \code{"none"} (\code{FALSE}), or only those whose missing values are in the
#'   \code{"response"} variable.
#' @param params list of user-specified model parameters to be passed to the
#'   \code{fit} function.
#' @param gridinfo tibble of information for construction of tuning grids
#'   consisting of a character column \code{param} with the names of parameters
#'   in the grid, a list column \code{get_values} with functions to generate grid
#'   points for the corresponding parameters, and an optional logical column
#'   \code{default} indicating which parameters to include by default in regular
#'   grids.  Values functions may optionally include arguments \code{n} and
#'   \code{data} for the number of grid points to generate and a
#'   \code{\link{ModelFrame}} of the model fit data and formula, respectively;
#'   and must include an ellipsis (\code{...}).
#' @param fit model fitting function whose arguments are a \code{formula}, a
#'   \code{\link{ModelFrame}} named \code{data}, case \code{weights}, and an
#'   ellipsis.
#' @param predict model prediction function whose arguments are the
#'   \code{object} returned by \code{fit}, a \code{\link{ModelFrame}} named
#'   \code{newdata} of predictor variables, optional vector of \code{times} at
#'   which to predict survival, and an ellipsis.
#' @param varimp variable importance function whose arguments are the
#'   \code{object} returned by \code{fit}, optional arguments passed from calls
#'   to \code{\link{varimp}}, and an ellipsis.
#' @param ... arguments passed to other methods.
#' @param object function that returns an \code{MLModel} object when called
#'   without any supplied argument values.
#'
#' @details
#' If supplied, the \code{grid} function should return a list whose elements are
#' named after and contain values of parameters to include in a tuning grid to
#' be constructed automatically by the package.
#'
#' Arguments \code{data} and \code{newdata} in the \code{fit} and \code{predict}
#' functions may be converted to data frames with \code{as.data.frame()}
#' if needed for their operation.  The \code{fit} function should return the
#' object resulting from the model fit.  Values returned by the \code{predict}
#' functions should be formatted according to the response variable types below.
#' \describe{
#'   \item{factor}{matrix whose columns contain the probabilities for
#'     multi-level factors or vector of probabilities for the second level of
#'     binary factors.}
#'   \item{matrix}{matrix of predicted responses.}
#'   \item{numeric}{vector or column matrix of predicted responses.}
#'   \item{Surv}{matrix whose columns contain survival probabilities at
#'     \code{times} if supplied or a vector of predicted survival means
#'     otherwise.}
#' }
#'
#' The \code{varimp} function should return a vector of importance values named
#' after the predictor variables or a matrix or data frame whose rows are named
#' after the predictors.
#'
#' The \code{predict} and \code{varimp} functions are additionally passed a list
#' named \code{.MachineShop} containing the \code{\link[=inputs]{input}}
#' and \code{\link[=models]{model}} from \code{\link{fit}}.  This argument may
#' be included in the function definitions as needed for their implementations.
#' Otherwise, it will be captured by the ellipsis.
#'
#' @return An \code{MLModel} or \code{MLModelFunction} class object.
#'
#' @seealso \code{\link{models}}, \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' ## Logistic regression model
#' LogisticModel <- MLModel(
#'   name = "LogisticModel",
#'   response_types = "binary",
#'   weights = TRUE,
#'   fit = function(formula, data, weights, ...) {
#'     glm(formula, data = as.data.frame(data), weights = weights,
#'         family = binomial, ...)
#'   },
#'   predict = function(object, newdata, ...) {
#'     predict(object, newdata = as.data.frame(newdata), type = "response")
#'   },
#'   varimp = function(object, ...) {
#'     pchisq(coef(object)^2 / diag(vcov(object)), 1)
#'   }
#' )
#'
#' data(Pima.tr, package = "MASS")
#' res <- resample(type ~ ., data = Pima.tr, model = LogisticModel)
#' summary(res)
#'
MLModel <- function(
  name = "MLModel", label = name, packages = character(),
  response_types = character(), weights = FALSE,
  predictor_encoding = c(NA, "model.frame", "model.matrix"), na.rm = FALSE,
  params = list(),
  gridinfo = tibble::tibble(
    param = character(), get_values = list(), default = logical()
  ),
  fit = function(formula, data, weights, ...) stop("No fit function."),
  predict = function(object, newdata, times, ...) stop("No predict function."),
  varimp = function(object, ...) NULL, ...
) {

  stopifnot(!any(duplicated(response_types)))
  stopifnot(response_types %in% settings("response_types"))
  stopifnot(length(weights) %in% c(1, length(response_types)))

  na.rm <- throw(check_na.rm(na.rm))

  stopifnot(is_tibble(gridinfo))
  gridinfo <- new_gridinfo(
    param = gridinfo[["param"]],
    get_values = gridinfo[["get_values"]],
    default = gridinfo[["default"]]
  )

  new("MLModel",
    name = name,
    label = label,
    packages = packages,
    response_types = response_types,
    weights = weights,
    predictor_encoding = match.arg(predictor_encoding),
    na.rm = na.rm,
    params = params,
    gridinfo = gridinfo,
    fit = fit,
    predict = predict,
    varimp = varimp,
    ...
  )
}


setMethod("initialize", "MLModel",
  function(.Object, ..., id = make_id("model")) {
    callNextMethod(.Object, ..., id = id)
  }
)


NullModel <- function() {
  new("NullModel",
    MLModel(
      id = "null",
      name = "NullModel",
      label = "Null Model",
      response_types = settings("response_types"),
      fit = function(...) {
        throw(Error("No specified model to fit."), call = call("fit"))
      }
    )
  )
}


update.MLModel <- function(
  object, params = NULL, quote = TRUE, id = object@id, ...
) {
  if (is.list(params)) {
    new_params <- as(object, "list")
    new_params[names(params)] <- params
    object <- do.call(object@name, new_params, quote = quote)
  } else if (!.hasSlot(object, "na.rm")) {
    info <- modelinfo(object@name)
    slot(object, "na.rm", check = FALSE) <- if (length(info)) {
      info[[1]]$na.rm
    } else {
      check_na.rm(formals(MLModel)$na.rm)
    }
  }
  object@id <- id
  object
}


#' @rdname MLModel
#'
MLModelFunction <- function(object, ...) {
  if (!is(try(object(), silent = TRUE), "MLModel")) {
    Error("Call to object() does not return an MLModel.")
  }
  new("MLModelFunction", object, ...)
}


"MLModelFunction<-" <- function(object, value) {
  do.call(MLModelFunction, c(object, value))
}


MLModelFit <- function(object, class, input, model) {
  if (is(object, "MLModelFit") && !is(object, class)) {
    throw(Error("Cannot change MLModelFit class."))
  }

  if (!is(model, "MLModel")) throw(TypeError(model, "MLModel", "`model`"))

  if (isS4(object)) {
    object <- new(class, object)
  } else {
    class <- setdiff(c(class, "MLModelFit"), class(object))
    class(object) <- c(class, class(object))
  }

  .MachineShop <- list(
    input = prep(input, retain = FALSE),
    model = model,
    version = packageVersion("MachineShop")
  )
  attr(object, ".MachineShop")[names(.MachineShop)] <- .MachineShop

  object
}


update.MLModelFit <- function(object, ...) {
  model <- NULL
  if ("mlmodel" %in% names(object)) {
    model <- object$mlmodel
    object$mlmodel <- NULL
  } else if ("mlmodel" %in% names(attributes(object))) {
    model <- attr(object, "mlmodel")
    attr(object, "mlmodel") <- NULL
  }
  if (is(model, "MLModel")) {
    .MachineShop <- list()
    y_levels <- c("y_levels" = "levels")
    add <- switch(model@name,
      "LARSModel" = "step",
      "LDAModel" = c("dimen", "use"),
      "QDAModel" = "use",
      "XGBModel" = y_levels,
      "XGBDARTModel" = y_levels,
      "XGBLinearModel" = y_levels,
      "XGBTreeModel" = y_levels
    )
    if (length(add)) {
      if (is.null(names(add))) names(add) <- add
      .MachineShop[names(add)] <- object[add]
      object[add] <- NULL
    }
    attr(object, ".MachineShop") <- .MachineShop
    input <- model@input
    if (is(input, "ModelRecipe") && is.null(input$orig_template)) {
      input$orig_template <- input$template
      input$template <- NULL
    }
    new_model <- update(model, params = model@params)
    new_model@label <- model@label
    new_model@steps <- model@steps
    MLModelFit(object, class1(object), input, new_model)
  } else {
    model <- attr(object, ".MachineShop")$model
    attr(object, ".MachineShop")$model <- update(model)
    object
  }
}


#' Revert an MLModelFit Object
#'
#' Function to revert an \code{MLModelFit} object to its original class.
#'
#' @param object model \link{fit} result.
#'
#' @return The supplied object with its \code{MLModelFit} classes and fields
#' removed.
#'
unMLModelFit <- function(object) {
  if (is(object, "MLModelFit")) {
    if (isS4(object)) {
      classes <- extends(class(object))
      pos <- match("MLModelFit", classes)
      as(object, classes[pos + 1])
    } else {
      object <- update(object)
      attr(object, ".MachineShop") <- NULL
      classes <- class(object)
      pos <- match("MLModelFit", classes)
      structure(object, class = classes[-seq_len(pos)])
    }
  } else object
}

#' Model Performance Metrics
#'
#' Compute measures of model performance.
#'
#' @name performance
#' @rdname performance
#'
#' @param x \link[=response]{observed responses}; or \link{confusion}, trained
#'   model \link{fit}, \link{resample}, or \link{rfe} result.
#' @param y \link[=predict]{predicted responses} if not contained in \code{x}.
#' @param weights numeric vector of non-negative
#'   \link[=case_weights]{case weights} for the observed \code{x} responses
#'   [default: equal weights].
#' @param metrics \link[=metrics]{metric} function, function name, or vector of
#'   these with which to calculate performance.
#' @param cutoff numeric (0, 1) threshold above which binary factor
#'   probabilities are classified as events and below which survival
#'   probabilities are classified.
#' @param na.rm logical indicating whether to remove observed or predicted
#'   responses that are \code{NA} when calculating metrics.
#' @param ... arguments passed from the \code{Resample} method to the response
#'   type-specific methods or from the method for \code{ConfusionList} to
#'   \code{ConfusionMatrix}.  Elliptical arguments in the response
#'   type-specific methods are passed to \code{metrics} supplied as a single
#'   \code{\link[=metrics]{MLMetric}} function and are ignored otherwise.
#'
#' @seealso \code{\link{plot}}, \code{\link{summary}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' res <- resample(Species ~ ., data = iris, model = GBMModel)
#' (perf <- performance(res))
#' summary(perf)
#' plot(perf)
#'
#' ## Survival response example
#' library(survival)
#'
#' gbm_fit <- fit(Surv(time, status) ~ ., data = veteran, model = GBMModel)
#'
#' obs <- response(gbm_fit, newdata = veteran)
#' pred <- predict(gbm_fit, newdata = veteran)
#' performance(obs, pred)
#' }
#'
performance <- function(x, ...) {
  UseMethod("performance")
}


#' @rdname performance
#'
performance.BinomialVariate <- function(
  x, y, weights = NULL, metrics = MachineShop::settings("metrics.numeric"),
  na.rm = TRUE, ...
) {
  .performance(x, y, weights, metrics, na.rm, dots = list(...))
}


#' @rdname performance
#'
performance.factor <- function(
  x, y, weights = NULL, metrics = MachineShop::settings("metrics.factor"),
  cutoff = MachineShop::settings("cutoff"), na.rm = TRUE, ...
) {
  .performance(x, y, weights, metrics, na.rm, cutoff = cutoff, dots = list(...))
}


#' @rdname performance
#'
performance.matrix <- function(
  x, y, weights = NULL, metrics = MachineShop::settings("metrics.matrix"),
  na.rm = TRUE, ...
) {
  .performance(x, y, weights, metrics, na.rm, dots = list(...))
}


#' @rdname performance
#'
performance.numeric <- function(
  x, y, weights = NULL, metrics = MachineShop::settings("metrics.numeric"),
  na.rm = TRUE, ...
) {
  .performance(x, y, weights, metrics, na.rm, dots = list(...))
}


#' @rdname performance
#'
performance.Surv <- function(
  x, y, weights = NULL, metrics = MachineShop::settings("metrics.Surv"),
  cutoff = MachineShop::settings("cutoff"), na.rm = TRUE, ...
) {
  .performance(x, y, weights, metrics, na.rm, cutoff = cutoff, dots = list(...))
}


.performance <- function(x, y, weights, metrics, na.rm, ..., dots = list()) {
  if (na.rm) {
    complete <- complete_subset(x = x, y = y, weights = weights)
    x <- complete$x
    y <- complete$y
    weights <- complete$weights
  }
  if (length(x)) {
    args <- list(x, y, weights = weights, ...)
    metric <- if (is_one_element(metrics)) metrics[[1]] else metrics
    if (is(get0(metric), "MLMetric")) args <- c(args, dots)
    metrics <- check_metrics(metrics, convert = TRUE)
    throw(check_assignment(metrics))
    do.call(metrics, args)
  } else {
    NA_real_
  }
}


#' @rdname performance
#'
performance.ConfusionList <- function(x, ...) {
  ListOf(map(function(conf) performance(conf, ...), x))
}


#' @rdname performance
#'
performance.ConfusionMatrix <- function(
  x, metrics = MachineShop::settings("metrics.ConfusionMatrix"), ...
) {
  args <- list(x)
  metric <- if (is_one_element(metrics)) metrics[[1]] else metrics
  if (is(get0(metric), "MLMetric")) args <- c(args, ...)
  metrics <- check_metrics(metrics, convert = TRUE)
  throw(check_assignment(metrics))
  do.call(metrics, args)
}


#' @rdname performance
#'
performance.MLModel <- function(x, ...) {
  if (!is_trained(x)) throw(Warning("No training results with performance."))
  ListOf(map(performance, x@steps))
}


#' @rdname performance
#'
performance.Resample <- function(x, ...) {
  perf_list <- by(x, x$Model, function(resamples) {
    Performance(performance(x@control, resamples, ...), control = x@control)
  }, simplify = FALSE)
  do.call(c, perf_list)
}


#' @rdname performance
#'
performance.TrainingStep <- function(x, ...) {
  x@performance
}


performance.MLControl <- function(x, resamples, ...) {
  perf_list <- by(resamples, resamples$Iteration, function(resample) {
    performance(resample$Observed, resample$Predicted, resample$Weight, ...)
  }, simplify = FALSE)
  do.call(rbind, perf_list)
}


performance.BootOptimismControl <- function(x, resamples, ...) {
  test_perf_list <- list()
  boot_perf_list <- list()
  resamples_split <- split(resamples, resamples$Iteration)
  for (name in names(resamples_split)) {
    resample <- resamples_split[[name]]
    test_perf_list[[name]] <- performance(
      resample$Observed, resample$Predicted, resample$Weight, ...
    )
    boot_perf_list[[name]] <- performance(
      resample$Boot.Observed, resample$Boot.Predicted, resample$Weight, ...
    )
  }
  test_perf <- do.call(rbind, test_perf_list)
  boot_perf <- do.call(rbind, boot_perf_list)
  train_perf <- performance(
    resample$Observed, resample$Train.Predicted, resample$Weight, ...
  )

  pessimism <- test_perf - boot_perf
  sweep(pessimism, 2, train_perf, "+")
}


performance.CVOptimismControl <- function(x, resamples, ...) {
  test_perf <- NextMethod()

  resamples_split <- split(resamples, ceiling(resamples$Iteration / x@folds))
  pred_names <- make_names_len(x@folds, "CV.Predicted.")
  perf_list <- map(function(resample) {
    f <- function(prop, pred) {
      prop * performance(resample$Observed, pred, resample$Weight, ...)
    }
    props <- proportions(table(resample$Iteration))
    preds <- resample[pred_names]
    Reduce("+", map(f, props, preds))
  }, resamples_split)
  cv_perf <- do.call(rbind, rep(perf_list, each = x@folds))
  train_perf <- performance(
    resamples_split[[1]]$Observed, resamples_split[[1]]$Train.Predicted,
    resamples_split[[1]]$Weight, ...
  )

  pessimism <- test_perf - cv_perf
  sweep(pessimism, 2, train_perf, "+")
}


Performance <- function(...) {
  object <- new("Performance", ...)
  ndim <- length(dimnames(object))
  if (ndim) {
    names(dimnames(object)) <- head(c("Iteration", "Metric", "Model"), ndim)
  }
  object
}

#' Confusion Matrix
#'
#' Calculate confusion matrices of predicted and observed responses.
#'
#' @name confusion
#' @rdname confusion
#'
#' @param x factor of \link[=response]{observed responses} or \link{resample}
#'   result containing observed and predicted responses.
#' @param y \link[=predict]{predicted responses} if not contained in \code{x}.
#' @param weights numeric vector of non-negative
#'   \link[=case_weights]{case weights} for the observed \code{x} responses
#'   [default: equal weights].
#' @param cutoff numeric (0, 1) threshold above which binary factor
#'   probabilities are classified as events and below which survival
#'   probabilities are classified.  If \code{NULL}, then factor responses are
#'   summed directly over predicted class probabilities, whereas a default
#'   cutoff of 0.5 is used for survival probabilities.  Class probability
#'   summations and survival will appear as decimal numbers that can be
#'   interpreted as expected counts.
#' @param na.rm logical indicating whether to remove observed or predicted
#'   responses that are \code{NA} when calculating metrics.
#' @param ... arguments passed to other methods.
#' @param data square matrix, or object that can be converted to one, of
#'   cross-classified predicted and observed values in the rows and columns,
#'   respectively.
#' @param ordered logical indicating whether the confusion matrix row and
#'   columns should be regarded as ordered.
#'
#' @return
#' The return value is a \code{ConfusionMatrix} class object that inherits from
#' \code{table} if \code{x} and \code{y} responses are specified or a
#' \code{ConfusionList} object that inherits from \code{list} if \code{x} is a
#' \code{Resample} object.
#'
#' @seealso \code{\link{c}}, \code{\link{plot}}, \code{\link{summary}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' res <- resample(Species ~ ., data = iris, model = GBMModel)
#' (conf <- confusion(res))
#' plot(conf)
#' }
#'
confusion <- function(
  x, y = NULL, weights = NULL, cutoff = MachineShop::settings("cutoff"),
  na.rm = TRUE, ...
) {
  if (na.rm) {
    complete <- complete_subset(x = x, y = y, weights = weights)
    x <- complete$x
    y <- complete$y
    weights <- complete$weights
  }
  .confusion(x, y, weights, cutoff = cutoff)
}


#' @rdname confusion
#'
ConfusionMatrix <- function(data = NA, ordered = FALSE) {
  data <- as.matrix(data)

  n <- nrow(data)
  if (n != ncol(data)) throw(Error("Unequal number of rows and columns."))

  data_dimnames <- dimnames(data)
  if (is.null(data_dimnames)) data_dimnames <- list(NULL, NULL)
  names(data_dimnames) <- c("Predicted", "Observed")

  data_class <- "ConfusionMatrix"
  if (n == 2) data_class <- paste0("Binary", data_class)
  if (ordered) data_class <- paste0("Ordered", data_class)

  new(data_class, structure(data, dimnames = data_dimnames))
}


setGeneric(".confusion",
  function(observed, predicted, ...) standardGeneric(".confusion")
)


setMethod(".confusion", c("ANY", "ANY"),
  function(observed, predicted, ...) {
    throw(Error(
      "Confusion requires a predicted factor or survival probabilities."
    ))
  }
)


setMethod(".confusion", c("factor", "factor"),
  function(observed, predicted, weights, ...) {
    weights <- check_weights(weights, observed)
    throw(check_assignment(weights))
    conf <- xtabs(weights ~ predicted + observed)
    ConfusionMatrix(conf, ordered = is.ordered(observed))
  }
)


setMethod(".confusion", c("factor", "matrix"),
  function(observed, predicted, weights, cutoff = NULL, ...) {
    if (is_empty(cutoff)) {
      weights <- check_weights(weights, observed)
      throw(check_assignment(weights))
      df <- aggregate(weights * predicted, list(observed), sum, na.rm = TRUE)
      conf <- as.table(t(df[, -1, drop = FALSE]))
      dimnames(conf) <- df[c(1, 1)]
      ConfusionMatrix(conf, ordered = is.ordered(observed))
    } else {
      predicted <- convert_response(observed, predicted)
      .confusion(observed, predicted, weights = weights, ...)
    }
  }
)


setMethod(".confusion", c("factor", "numeric"),
  function(observed, predicted, cutoff, ...) {
    predicted <- if (is_empty(cutoff)) {
      cbind(1 - predicted, predicted)
    } else {
      convert_response(observed, predicted, cutoff = cutoff)
    }
    .confusion(observed, predicted, ...)
  }
)


setMethod(".confusion", c("Resample", "ANY"),
  function(observed, predicted, weights, ...) {
    conf_list <- by(observed, observed$Model, function(resample) {
      confusion(
        resample$Observed, resample$Predicted, resample$Weight, na.rm = FALSE,
        ...
      )
    }, simplify = FALSE)
    if (all(map("logi", is, conf_list, "ConfusionList"))) {
      conf_list <- unlist(conf_list, recursive = FALSE)
    }
    do.call(c, conf_list)
  }
)


setMethod(".confusion", c("Surv", "SurvProbs"),
  function(observed, predicted, cutoff, ...) {
    if (is_empty(cutoff)) cutoff <- 0.5
    predicted <- convert_response(observed, predicted, cutoff = cutoff)
    .confusion(observed, predicted, ...)
  }
)


setMethod(".confusion", c("Surv", "SurvEvents"),
  function(observed, predicted, weights, ...) {

    weights <- check_weights(weights, observed)
    throw(check_assignment(weights))
    times <- predicted@times
    conf_tbl <- table(Predicted = 0:1, Observed = 0:1)

    conf_list <- map(function(i) {
      pos <- predicted[, i, drop = TRUE] == 1
      pos_pred <- surv_subset(observed, weights, pos, times[i])
      neg_pred <- surv_subset(observed, weights, !pos, times[i])

      conf_tbl[1, 1] <- neg_pred$surv * neg_pred$p
      conf_tbl[2, 1] <- pos_pred$surv * pos_pred$p
      conf_tbl[1, 2] <- (1 - neg_pred$surv) * neg_pred$p
      conf_tbl[2, 2] <- (1 - pos_pred$surv) * pos_pred$p

      ConfusionMatrix(length(observed) * conf_tbl)
    }, seq_along(times))
    names(conf_list) <- colnames(predicted)
    do.call(c, conf_list)

  }
)

#' Resample Estimation of Model Performance
#'
#' Estimation of the predictive performance of a model estimated and evaluated
#' on training and test samples generated from an observed data set.
#'
#' @name resample
#' @rdname resample-methods
#'
#' @param ... arguments passed from the generic function to its methods, from
#'   the \code{MLModel} and \code{MLModelFunction} methods to first arguments of
#'   others, and from others to the \code{ModelSpecification} method.  The
#'   first argument of each \code{fit} method is positional and, as such, must
#'   be given first in calls to them.
#' @param object model \link[=inputs]{input} or
#'   \link[=ModelSpecification]{specification}.
#' @param formula,data \link[=formula]{formula} defining the model predictor and
#'   response variables and a \link[=data.frame]{data frame} containing them.
#' @param x,y \link{matrix} and object containing predictor and response
#'   variables.
#' @param input \link[=inputs]{input} object defining and containing the model
#'   predictor and response variables.
#' @param model \link[=models]{model} function, function name, or object; or
#'   another object that can be \link[=as.MLModel]{coerced} to a model.  A model
#'   can be given first followed by any of the variable specifications.
#' @param control \link[=controls]{control} function, function name, or object
#'   defining the resampling method to be employed.
#'
#' @return \code{Resample} class object.
#'
#' @seealso \code{\link{c}}, \code{\link{metrics}}, \code{\link{performance}},
#' \code{\link{plot}}, \code{\link{summary}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package gbm to run
#'
#' ## Factor response example
#'
#' fo <- Species ~ .
#' control <- CVControl()
#'
#' gbm_res1 <- resample(fo, iris, GBMModel(n.trees = 25), control)
#' gbm_res2 <- resample(fo, iris, GBMModel(n.trees = 50), control)
#' gbm_res3 <- resample(fo, iris, GBMModel(n.trees = 100), control)
#'
#' summary(gbm_res1)
#' plot(gbm_res1)
#'
#' res <- c(GBM1 = gbm_res1, GBM2 = gbm_res2, GBM3 = gbm_res3)
#' summary(res)
#' plot(res)
#' }
#'
resample <- function(...) {
  UseMethod("resample")
}


#' @rdname resample-methods
#'
#' @details
#' Stratified resampling is performed automatically for the \code{formula} and
#' \code{matrix} methods according to the type of response variable.  In
#' general, strata are constructed from numeric proportions for
#' \code{\link{BinomialVariate}}; original values for \code{character},
#' \code{factor}, \code{logical}, and \code{ordered}; first columns of values
#' for \code{matrix}; original values for \code{numeric}; and numeric times
#' within event statuses for \code{Surv}.  Numeric values are stratified into
#' quantile bins and categorical values into factor levels defined by
#' \code{\link{MLControl}}.
#'
resample.formula <- function(formula, data, model, ...) {
  resample(
    ModelSpecification(formula, data, model = model, control = NULL), ...
  )
}


#' @rdname resample-methods
#'
resample.matrix <- function(x, y, model, ...) {
  resample(ModelSpecification(x, y, model = model, control = NULL), ...)
}


#' @rdname resample-methods
#'
#' @details
#' Resampling stratification variables may be specified manually for
#' \code{ModelFrames} upon creation with the \code{\link[=ModelFrame]{strata}}
#' argument in their constructor.  Resampling of this class is unstratified by
#' default.
#'
resample.ModelFrame <- function(input, model, ...) {
  resample(ModelSpecification(input, model = model, control = NULL), ...)
}


#' @rdname resample-methods
#'
#' @details
#' Stratification variables may be designated in \code{recipe} specifications
#' with the \code{\link{role_case}} function.  Resampling will be unstratified
#' otherwise.
#'
resample.recipe <- function(input, model, ...) {
  resample(ModelSpecification(input, model = model, control = NULL), ...)
}


#' @rdname resample-methods
#'
resample.ModelSpecification <- function(
  object, control = MachineShop::settings("control"), ...
) {
  .resample(as.MLControl(control), object = object, ...)
}


#' @rdname resample-methods
#'
resample.MLModel <- function(model, ...) {
  resample(..., model = model)
}


#' @rdname resample-methods
#'
resample.MLModelFunction <- function(model, ...) {
  resample(as.MLModel(model), ...)
}


.resample <- function(control, object, ...) {
  UseMethod(".resample")
}


.resample.BootControl <- function(control, object, progress_index = 0, ...) {
  presets <- settings()
  set.seed(control@seed)
  splits <- rsample_split(
    function(data, group = NULL, strata = NULL, ...) {
      samples <- control@samples
      suppressWarnings(
        if (length(group)) {
          group_bootstraps(data, group = group, times = samples, pool = 0)
        } else {
          bootstraps(data, times = samples, strata = strata, pool = 0)
        }
      )$splits
    }, data = object, control = control
  )
  seeds <- rand_int(length(splits))

  is_optimism_control <- is(control, "BootOptimismControl")
  if (is_optimism_control) {
    train_pred <- subsample(object, object, control)$Predicted
  }

  monitor <- new_progress_bar(
    control@monitor$progress * length(splits),
    object = object,
    index = progress_index
  )
  on.exit(monitor$pb$terminate())

  foreach(
    i = i <- seq_along(splits),
    .export = c("seq_boot", "subsample"),
    .packages = required_packages(object),
    .verbose = control@monitor$verbose,
    .options.snow = monitor$snow_opts
  ) %dopar% {
    monitor$progress()
    settings(presets)
    set.seed(seeds[i])
    train <- update(object, data = rsample::analysis(splits[[i]]))
    if (is_optimism_control) {
      subs <- subsample(train, list(object, train), control, i)
      df <- subs[[1]]
      df_boot <- subs[[2]]
      indices <- seq_boot(df_boot, df)
      df["Boot.Observed"] <- df_boot[indices, "Observed"]
      df["Boot.Predicted"] <- df_boot[indices, "Predicted"]
      df$Train.Predicted <- train_pred
      df
    } else {
      subsample(train, object, control, i)
    }
  } %>% Resample(control = control, vars = attr(splits, "vars"))
}


.resample.CVControl <- function(control, object, progress_index = 0, ...) {
  set.seed(control@seed)
  splits <- rsample_split(
    function(data, group = NULL, strata = NULL, ...) {
      v <- control@folds
      repeats <- control@repeats
      res <- suppressWarnings(
        if (length(group)) {
          method <- "grouping"
          group_vfold_cv(
            data, group = group, v = v, repeats = repeats, pool = 0
          )
        } else {
          method <- "stratification"
          vfold_cv(data, v = v, repeats = repeats, strata = strata, pool = 0)
        }
      )$splits
      control@folds <<- as.integer(length(res) / repeats)
      if (control@folds < v) {
        throw(Warning(
          "Fewer folds constructed (", control@folds, ") than specified in ",
          class(control), " (", v, ") due to the resample ", method, " size.  ",
          "To address the issue, increase the size or use a ModelFrame or ",
          "recipe to resample without ", method, "."
        ), call = call("resample"))
      }
      res
    }, data = object, control = control
  )
  seeds <- rand_int(length(splits))

  is_optimism_control <- is(control, "CVOptimismControl")

  monitor <- new_progress_bar(
    control@monitor$progress * length(splits),
    object = object,
    index = progress_index
  )
  on.exit(monitor$pb$terminate())

  presets <- settings()
  df_list <- foreach(
    i = i <- seq_along(splits),
    .export = "subsample",
    .packages = required_packages(object),
    .verbose = control@monitor$verbose,
    .options.snow = monitor$snow_opts
  ) %dopar% {
    monitor$progress()
    settings(presets)
    set.seed(seeds[i])
    train <- update(object, data = rsample::analysis(splits[[i]]))
    test <- update(object, data = rsample::assessment(splits[[i]]))
    if (is_optimism_control) {
      subs <- subsample(train, list(test, object), control, i)
      structure(subs[[1]], CV.Predicted = subs[[2]]["Predicted"])
    } else {
      subsample(train, test, control, i)
    }
  }
  res <- Resample(df_list, control = control, vars = attr(splits, "vars"))

  if (is_optimism_control) {
    pred_list <- map(attr, df_list, "CV.Predicted")
    split_factor <- rep(seq_len(control@folds), control@repeats)
    df <- split(seq_along(pred_list), split_factor) %>%
      map(function(indices) do.call(append, pred_list[indices]), .) %>%
      as.data.frame
    names(df) <- make_names_len(length(df), "CV.Predicted.")
    pred <- subsample(object, object, control)$Predicted
    df$Train.Predicted <- do.call(append, rep(list(pred), control@repeats))
    res[names(df)] <- df
  }

  res
}


.resample.OOBControl <- function(control, object, progress_index = 0, ...) {
  presets <- settings()
  set.seed(control@seed)
  splits <- rsample_split(
    function(data, group = NULL, strata = NULL, ...) {
      samples <- control@samples
      suppressWarnings(
        if (length(group)) {
          group_bootstraps(data, group = group, times = samples, pool = 0)
        } else {
          bootstraps(data, times = samples, strata = strata, pool = 0)
        }
      )$splits
    }, data = object, control = control
  )
  seeds <- rand_int(length(splits))

  monitor <- new_progress_bar(
    control@monitor$progress * length(splits),
    object = object,
    index = progress_index
  )
  on.exit(monitor$pb$terminate())

  foreach(
    i = i <- seq_along(splits),
    .export = "subsample",
    .packages = required_packages(object),
    .verbose = control@monitor$verbose,
    .options.snow = monitor$snow_opts
  ) %dopar% {
    monitor$progress()
    settings(presets)
    set.seed(seeds[i])
    train <- update(object, data = rsample::analysis(splits[[i]]))
    test <- update(object, data = rsample::assessment(splits[[i]]))
    subsample(train, test, control, i)
  } %>% Resample(control = control, vars = attr(splits, "vars"))
}


.resample.SplitControl <- function(control, object, ...) {
  set.seed(control@seed)
  split <- rsample_split(
    function(data, group = NULL, strata = NULL, ...) {
      prop <- control@prop
      if (length(group)) {
        group_initial_split(data, group = group, prop = prop, pool = 0)
      } else {
        initial_split(data, prop = prop, strata = strata, pool = 0)
      }
    }, data = object, control = control
  )
  train <- update(object, data = rsample::training(split))
  test <- update(object, data = rsample::testing(split))
  subsample(train, test, control) %>%
    Resample(control = control, vars = attr(split, "vars"))
}


.resample.TrainControl <- function(control, object, ...) {
  set.seed(control@seed)
  Resample(subsample(object, object, control), control = control)
}


Resample <- function(object, ...) {
  UseMethod("Resample")
}


Resample.data.frame <- function(
  object, ..., control, vars = NULL, .check = TRUE
) {
  if (.check) {
    var_names <- c("Model", "Iteration", "Case", "Observed", "Predicted")
    missing <- missing_names(var_names, object)
    if (length(missing)) {
      throw(Error(note_items("Missing resample variable{?s}: ", missing, ".")))
    }
    object$Model <- droplevels(object$Model)
  }
  rownames(object) <- NULL

  vars <- as_tibble(vars)
  if (is_empty(vars[["Stratification"]])) control@strata <- list()

  new("Resample", object, control = control, vars = vars, ...)
}


Resample.list <- function(object, ...) {
  Resample(do.call(append, object), ...)
}


#################### Utility Functions ####################


rsample_split <- function(fun, data, control) {
  data <- as.MLInput(data)
  df <- as.data.frame(data)
  formal_names <- names(formals(fun))
  df[["(groups)"]] <- if ("group" %in% formal_names) case_groups(data)
  df[["(strata)"]] <- if ("strata" %in% formal_names) {
    do.call(case_strata, c(list(data), control@strata))
  }
  if (length(df[["(groups)"]]) && length(df[["(strata)"]])) {
    throw(LocalWarning(
      "Case groups and strata are both specified for resampling; ",
      "only the strata will be used."
    ))
    df[["(groups)"]] <- NULL
  }
  res <- fun(
    df,
    group = case_comp_name(df, "groups"),
    strata = case_comp_name(df, "strata")
  )

  vars <- tibble(Case = rownames(df))
  types <- c(Grouping = "groups", Stratification = "strata")
  for (i in seq_along(types)) {
    type <- types[i]
    comp <- df[[paste0("(", type, ")")]]
    vars[[names(type)]] <- if (length(comp)) {
      as_tibble_col(comp, column_name = case_comp_name(data, type))
    }
  }
  if (length(vars)) attr(res, "vars") <- vars

  res
}


subsample <- function(train, test, control, iter = 1) {
  verbose <- control@monitor$verbose
  model_fit <- fit(train, verbose = verbose)
  times <- time(model_fit)
  if (length(times)) control@predict$times <- times

  f <- function(test) {
    weights <- if (control@weights) "weights"
    comps <- case_comps(
      model_fit, newdata = as.MLInput(test), types = c("names", weights),
      response = TRUE
    )
    df <- data.frame(
      Model = factor(as.MLModel(train)@name),
      Iteration = as.integer(iter),
      Case = comps$names
    )
    df$Observed <- comps$response
    df$Predicted <- do.call(predict, c(
      list(model_fit, as.data.frame(test), type = "raw", verbose = verbose),
      control@predict
    ))
    df$Weight <- comps$weights
    df
  }

  if (class1(test) == "list") map(f, test) else f(test)
}

GridSearch <- function(...) {
  new("GridSearch", ...,
    label = "Grid Search",
    fun = function(optim, grid, ...) {
      max_iter <- nrow(grid)
      for (i in seq_len(max_iter)) {
        optim(grid[i, ], max_iter = max_iter)
      }
      return(NULL)
    }
  ) %>% set_monitor
}


NullOptimization <- function() {
  new("NullOptimization", label = "Null Optimization")
}


RandomGridSearch <- function(..., size = numeric()) {
  size <- check_integer(size, bounds = c(1, Inf), size = 1)
  throw(check_assignment(size))

  new("RandomGridSearch",
    GridSearch(...), label = "Random Grid Search", size = size
  )
}


SequentialOptimization <- function(..., random = FALSE) {
  if (!isFALSE(random)) {
    random <- check_integer(random, bounds = c(1, Inf), size = 1)
    throw(check_assignment(random))
  }

  new("SequentialOptimization", ..., random = random) %>% set_monitor
}


optim <- function(.fun, object, ...) {

  method <- get_optim_field(object, "label")
  progress <- if (get_optim_field(object, "monitor")$progress) {
    heading(paste(class(object), "optimization"))
    new_print_progress()
  } else {
    function(...) NULL
  }

  grid <- get_grid(object, ...)
  optim_grid <- unnest_params(grid$params)
  names(optim_grid) <- make_names_len(length(optim_grid), "param")
  optim_params <- as.list(optim_grid[1, ])
  optim_params[] <- list(NULL)

  optim_bounds <- map(function(param) {
    if (all(is.finite(param)) && length(unique(param)) > 1) range(param)
  }, optim_grid)
  optim_bounds <- optim_bounds[!map("logi", is.null, optim_bounds)]

  control <- object@params@control
  metrics <- object@params@metrics
  cutoff <- object@params@cutoff
  stat_fun <- function(x) object@params@stat(na.omit(x))

  err_msgs <- character()
  model_params_list <- list()
  perf <- NULL
  perf_list <- list()
  perf_stats_list <- list()
  scores <- numeric()

  optim_fun <- function(x, max_iter = Inf) {
    optim_params[names(x)] <- x
    model_params <- renest_params(optim_params, grid$params[1, ])
    model_params_list <<- c(model_params_list, list(model_params))

    perf_stats <- NA
    metric <- NULL
    score <- tryCatch(
      {
        res <- resample(
          update(object, params = model_params), ..., control = control,
          progress_index = new_progress_index(
            length(scores) + 1, max = max_iter, name = class(object)
          )
        )
        err_msgs <<- c(err_msgs, NA)
        if (is.null(metrics)) {
          metrics <<- get_perf_metrics(res$Observed, res$Predicted)
        }
        perf <<- performance(res, metrics = metrics, cutoff = cutoff)
        perf_list <<- c(perf_list, list(perf))
        perf_stats <- apply(perf, 2, stat_fun)
        perf_stats_list <<- c(perf_stats_list, list(perf_stats))
        metric <- as.MLMetric(c(metrics)[[1]])
        ifelse(metric@maximize, 1, -1) * perf_stats[1]
      },
      error = function(cond) {
        err_msgs <<- c(err_msgs, conditionMessage(cond))
        perf_list <<- c(perf_list, list(NA))
        perf_stats_list <<- c(perf_stats_list, list(perf_stats))
        -Inf
      }
    )
    scores <<- c(scores, score)

    progress(
      scores, max_iter = max_iter, method = method,
      items = list(
        "Parameter{?s}: " = unlist(
          unnest_params(model_params, compact_names = TRUE)
        )
      ),
      metric = structure(perf_stats[1], names = getElement(metric, "name"))
    )

    score
  }

  optim_res <- .fun(
    optim = optim_fun,
    grid = optim_grid,
    bounds = optim_bounds,
    params = get_optim_field(object, "params"),
    monitor = get_optim_field(object, "monitor")
  )

  names(perf_list) <- if (is_optim_method(object, "GridSearch")) {
    grid$name
  } else {
    grid_name <- sub("[.].*$", "", grid$name[1])
    make_unique(rep_len(grid_name, length(scores)))
  }

  failed <- is.na(perf_list)
  names(err_msgs) <- names(perf_list)
  err_msgs <- na.omit(err_msgs)
  err_msgs <- paste0(names(err_msgs), ": ", err_msgs, collapse = "\n")
  if (all(failed)) {
    throw(LocalError("Resampling failed for all models.\n", err_msgs))
  } else if (any(failed)) {
    throw(LocalWarning("Resampling failed for some models.\n", err_msgs))
    perf[] <- NA
    perf_list[failed] <- list(perf)
    perf_stats_list[failed] <- list(perf[1, ])
  }

  step <- TrainingStep(
    object,
    method = method,
    names = names(perf_list),
    items = optim_res,
    params = do.call(rbind, model_params_list),
    metrics = do.call(rbind, perf_stats_list),
    selected = which.max(scores),
    performance = do.call(c, perf_list)
  )

  model_params <- step@log$params[step@log$selected, ]
  push(step, fit(update(object, params = model_params), ...))

}

#' Combine MachineShop Objects
#'
#' Combine one or more \pkg{MachineShop} objects of the same class.
#'
#' @name combine
#' @aliases c
#' @rdname combine-methods
#'
#' @param ... named or unnamed \link{calibration}, \link{confusion},
#'   \link{lift}, \link[=curves]{performance curve}, \link{summary}, or
#'   \link{resample} results.  Curves must have been generated with the same
#'   performance \link{metrics} and resamples with the same resampling
#'   \link[=controls]{control}.
#' @param e1,e2 objects.
#'
#' @return Object of the same class as the arguments.
#'
NULL


c.BinomialVariate <- function(...) {
  args <- list(...)
  if (all(map("logi", is, args, "BinomialVariate"))) {
    structure(do.call(rbind, args), class = "BinomialVariate")
  } else {
    NextMethod()
  }
}


#' @rdname combine-methods
#'
c.Calibration <- function(...) {
  args <- list(...)
  if (all(map("logi", is, args, "Calibration"))) {

    if (!identical_elements(args, function(x) x@smoothed)) {
      throw(Error(
        "Calibration arguments are a mix of smoothed and binned curves."
      ))
    }

    df <- do.call(append, set_model_names(args))
    Calibration(df, smoothed = args[[1]]@smoothed, .check = FALSE)

  } else {
    NextMethod()
  }
}


#' @rdname combine-methods
#'
c.ConfusionList <- function(...) {
  args <- list(...)
  is_valid <- function(x) is(x, "ConfusionList") || is(x, "ConfusionMatrix")
  if (all(map("logi", is_valid, args))) {

    conf_list <- list()
    for (i in seq_along(args)) {
      x <- args[[i]]
      if (is(x, "ConfusionMatrix")) x <- list("Model" = x)
      arg_name <- names(args)[i]
      if (!is.null(arg_name) && nzchar(arg_name)) {
        names(x) <- rep(arg_name, length(x))
      }
      conf_list <- c(conf_list, x)
    }
    names(conf_list) <- make_unique(names(conf_list))

    ConfusionList(conf_list)

  } else {
    NextMethod()
  }
}


#' @rdname combine-methods
#'
c.ConfusionMatrix <- function(...) {
  args <- list(...)
  args[[1]] <- ConfusionList(ListOf(args[1]))
  if (is.null(names(args)[1])) names(args)[1] <- "Model"
  do.call(c, args)
}


c.DiscreteVariate <- function(...) {
  args <- list(...)
  x <- NextMethod()
  class <- class(args[[1]])
  if (all(map("logi", is, args, class))) {
    new(class,
      x,
      min = min(map("num", slot, args, "min")),
      max = max(map("num", slot, args, "max"))
    )
  } else {
    x
  }
}


#' @rdname combine-methods
#'
c.LiftCurve <- function(...) {
  NextMethod()
}


#' @rdname combine-methods
#'
c.ListOf <- function(...) {
  args <- list(...)
  class <- class1(args[[1]][[1]])
  is_valid <- function(x) {
    is(x, "ListOf") && is(x[[1]], class) && is(x[[1]], "vector")
  }
  if (all(map("logi", is_valid, args))) {
    x <- list()
    for (i in seq_along(args)) {
      name <- names(args)[i]
      if (!is.null(name) && nzchar(name)) {
        names(args[[i]]) <- rep(name, length(args[[i]]))
      }
      x <- c(x, args[[i]])
    }
    if (!is.null(names(x))) names(x) <- make_unique(names(x))
    ListOf(x)
  } else {
    NextMethod()
  }
}


c.Performance <- function(...) {
  args <- list(...)
  if (all(map("logi", is, args, "Performance"))) {
    if (length(args) > 1) {

      if (!identical_elements(args, function(x) dimnames(x)[1:2])) {
        throw(Error("Performance objects have different row or column names."))
      }

      if (!identical_elements(args, function(x) x@control)) {
        throw(Error("Performance arguments have different control structures."))
      }

      Performance(abind(args, along = 3), control = args[[1]]@control)

    } else {
      args[[1]]
    }
  } else {
    NextMethod()
  }
}


#' @rdname combine-methods
#'
c.PerformanceCurve <- function(...) {
  args <- list(...)
  class <- class(args[[1]])
  if (all(map("logi", is, args, class))) {

    if (!identical_elements(args, function(x) x@metrics)) {
      throw(Error(class, " arguments have different metrics."))
    }

    if (!identical_elements(args, function(x) x@control)) {
      throw(Error(class, " arguments have different control structures."))
    }

    do.call(class, list(
      do.call(append, set_model_names(args)), metrics = args[[1]]@metrics,
      control = args[[1]]@control, .check = FALSE
    ))

  } else {
    NextMethod()
  }
}


#' @rdname combine-methods
#'
c.Resample <- function(...) {
  args <- list(...)
  if (all(map("logi", is, args, "Resample"))) {

    if (!identical_elements(args, function(x) x@control)) {
      throw(Error("Resample arguments have different control structures."))
    }

    for (name in names(args[[1]]@vars)) {
      if (!identical_elements(args, function(x) unname(x@vars[[name]]))) {
        throw(Error(
          "Resample arguments have different ", tolower(name), " variables."
        ))
      }
    }

    df <- do.call(append, set_model_names(args))
    Resample(
      df, control = args[[1]]@control, vars = args[[1]]@vars, .check = FALSE
    )

  } else {
    NextMethod()
  }
}


c.SurvMatrix <- function(...) {
  args <- list(...)
  arg1 <- args[[1]]
  class <- class(arg1)
  if (all(map("logi", is, args, class))) {

    if (!identical_elements(args, function(x) x@times)) {
      throw(Error(class, " arguments have different times."))
    }

    if (!identical_elements(args, function(x) x@distr)) {
      throw(Error(class, " arguments have different distributions."))
    }

    new(class, do.call(rbind, args), times = arg1@times, distr = arg1@distr)

  } else {
    NextMethod()
  }
}


c.SurvTimes <- function(...) {
  args <- list(...)
  x <- NextMethod()
  class <- class(args[[1]])
  if (all(map("logi", is, args, class))) {
    new(class, x, distr = args[[1]]@distr)
  } else {
    x
  }
}


#' @rdname combine-methods
#'
setMethod("+", c("SurvMatrix", "SurvMatrix"),
  function(e1, e2) {
    x <- callNextMethod()
    class <- class(e1)
    if (all(class(e2) == class, e1@times == e2@times, e1@distr == e2@distr)) {
      new(class, x, times = e1@times, distr = e1@distr)
    } else x
  }
)

#' @details
#' The following set of model fitting, prediction, and performance assessment
#' functions are available for \pkg{MachineShop} \link{models}.
#'
#' Training:
#' \tabular{ll}{
#'   \code{\link{fit}} \tab Model fitting \cr
#'   \code{\link{resample}} \tab Resample estimation of model performance \cr
#' }
#'
#' Tuning Grids:
#' \tabular{ll}{
#'   \code{\link{expand_model}} \tab Model expansion over tuning parameters \cr
#'   \code{\link{expand_modelgrid}} \tab Model tuning grid expansion \cr
#'   \code{\link{expand_params}} \tab Model parameters expansion \cr
#'   \code{\link{expand_steps}} \tab Recipe step parameters expansion \cr
#' }
#'
#' Response Values:
#' \tabular{ll}{
#'   \code{\link{response}} \tab Observed \cr
#'   \code{\link{predict}} \tab Predicted \cr
#' }
#'
#' Performance Assessment:
#' \tabular{ll}{
#'   \code{\link{calibration}} \tab Model calibration \cr
#'   \code{\link{confusion}} \tab Confusion matrix \cr
#'   \code{\link{dependence}} \tab Parital dependence \cr
#'   \code{\link{diff}} \tab Model performance differences \cr
#'   \code{\link{lift}} \tab Lift curves \cr
#'   \code{\link{performance} \link{metrics}} \tab Model performance metrics \cr
#'   \code{\link{performance_curve}} \tab Model performance curves \cr
#'   \code{\link{rfe}} \tab Recursive feature elimination \cr
#'   \code{\link{varimp}} \tab Variable importance \cr
#' }
#'
#' Methods for resample estimation include
#'
#' \tabular{ll}{
#'   \code{\link{BootControl}} \tab Simple bootstrap \cr
#'   \code{\link{BootOptimismControl}} \tab Optimism-corrected bootstrap \cr
#'   \code{\link{CVControl}} \tab Repeated K-fold cross-validation \cr
#'   \code{\link{CVOptimismControl}} \tab Optimism-corrected cross-validation \cr
#'   \code{\link{OOBControl}} \tab Out-of-bootstrap \cr
#'   \code{\link{SplitControl}} \tab Split training-testing \cr
#'   \code{\link{TrainControl}} \tab Training resubstitution \cr
#' }
#'
#' Graphical and tabular summaries of modeling results can be obtained with
#'
#'\tabular{l}{
#'   \code{\link{plot}} \cr
#'   \code{\link{print}} \cr
#'   \code{\link{summary}} \cr
#'}
#'
#' Further information on package features is available with
#'
#' \tabular{ll}{
#'   \code{\link{metricinfo}} \tab Performance metric information \cr
#'   \code{\link{modelinfo}} \tab Model information \cr
#'   \code{\link{settings}} \tab Global settings \cr
#' }
#'
#' Custom metrics and models can be created with the \code{\link{MLMetric}} and
#' \code{\link{MLModel}} constructors.
#'
"_PACKAGE"

#' Linear Discriminant Analysis Model
#'
#' Performs linear discriminant analysis.
#'
#' @param prior prior probabilities of class membership if specified or the
#'   class proportions in the training set otherwise.
#' @param tol tolerance for the determination of singular matrices.
#' @param method type of mean and variance estimator.
#' @param nu degrees of freedom for \code{method = "t"}.
#' @param dimen dimension of the space to use for prediction.
#' @param use type of parameter estimation to use for prediction.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameter:}{
#'     \code{dimen}
#'   }
#' }
#'
#' The \code{\link{predict}} function for this model additionally accepts the
#' following argument.
#' \describe{
#'   \item{\code{prior}}{prior class membership probabilities for prediction
#'     data if different from the training set.}
#' }
#'
#' Default argument values and further model details can be found in the source
#' See Also links below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[MASS]{lda}}, \code{\link[MASS]{predict.lda}},
#' \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' fit(Species ~ ., data = iris, model = LDAModel)
#'
LDAModel <- function(
  prior = numeric(), tol = 1e-4, method = c("moment", "mle", "mve", "t"),
  nu = 5, dimen = integer(), use = c("plug-in", "debiased", "predictive")
) {

  method <- match.arg(method)
  use <- match.arg(use)

  MLModel(

    name = "LDAModel",
    label = "Linear Discriminant Analysis",
    packages = "MASS",
    response_types = "factor",
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment()),

    gridinfo = new_gridinfo(
      param = "dimen",
      get_values = c(
        function(n, data, ...) {
          seq_len(min(n, nlevels(response(data)) - 1, nvars(data, LDAModel)))
        }
      )
    ),

    fit = function(formula, data, weights, dimen, use, ...) {
      res <- MASS::lda(
        formula, data = as.data.frame(formula, data = data),
        na.action = na.pass, ...
      )
      attr(res, ".MachineShop") <- list(
        dimen = if (missing(dimen)) length(res$svd) else dimen,
        use = use
      )
      res
    },

    predict = function(
      object, newdata, prior = object$prior, .MachineShop, ...
    ) {
      newdata <- as.data.frame(newdata)
      predict(
        object, newdata = newdata, prior = prior, dimen = .MachineShop$dimen,
        method = .MachineShop$use
      )$posterior
    }

  )

}

MLModelFunction(LDAModel) <- NULL

#' Mixture Discriminant Analysis Model
#'
#' Performs mixture discriminant analysis.
#'
#' @param subclasses numeric value or vector of subclasses per class.
#' @param sub.df effective degrees of freedom of the centroids per class if
#'   subclass centroid shrinkage is performed.
#' @param tot.df specification of the total degrees of freedom as an alternative
#'   to \code{sub.df}.
#' @param dimension dimension of the discriminant subspace to use for
#'   prediction.
#' @param eps numeric threshold for automatically truncating the dimension.
#' @param iter limit on the total number of iterations.
#' @param method regression function used in optimal scaling.  The default of
#'   linear regression is provided by \code{\link[mda]{polyreg}} from the
#'   \pkg{mda} package.  For penalized mixture discriminant models,
#'   \code{\link[mda]{gen.ridge}} is appropriate.  Other possibilities are
#'   \code{\link[mda]{mars}} for multivariate adaptive regression splines and
#'   \code{\link[mda]{bruto}} for adaptive backfitting of additive splines.  Use
#'   the \code{\link[=quote]{.}} operator to quote specified functions.
#' @param trace logical indicating whether iteration information is printed.
#' @param ... additional arguments to \code{mda.start} and \code{method}.
#'
#' @details
#' \describe{
#'   \item{Response types:}{\code{factor}}
#'   \item{\link[=TunedModel]{Automatic tuning} of grid parameter:}{
#'     \code{subclasses}
#'   }
#' }
#'
#' The \code{\link{predict}} function for this model additionally accepts the
#' following argument.
#' \describe{
#'   \item{\code{prior}}{prior class membership probabilities for prediction data
#'     if different from the training set.}
#' }
#'
#' Default argument values and further model details can be found in the source
#' See Also links below.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link[mda]{mda}}, \code{\link[mda]{predict.mda}},
#' \code{\link{fit}}, \code{\link{resample}}
#'
#' @examples
#' \donttest{
#' ## Requires prior installation of suggested package mda to run
#'
#' fit(Species ~ ., data = iris, model = MDAModel)
#' }
#'
MDAModel <- function(
  subclasses = 3, sub.df = numeric(), tot.df = numeric(),
  dimension = sum(subclasses) - 1, eps = .Machine$double.eps, iter = 5,
  method = .(mda::polyreg), trace = FALSE, ...
) {

  MLModel(

    name = "MDAModel",
    label = "Mixture Discriminant Analysis",
    packages = "mda",
    response_types = "factor",
    predictor_encoding = "model.matrix",
    na.rm = TRUE,
    params = new_params(environment(), ...),

    gridinfo = new_gridinfo(
      param = "subclasses",
      get_values = c(
        function(n, ...) seq_int(2, length = min(n, 10))
      )
    ),

    fit = function(formula, data, weights, ...) {
      mda::mda(formula, data = as.data.frame(formula, data = data), ...)
    },

    predict = function(object, newdata, prior = object$prior, ...) {
      newdata <- as.data.frame(newdata)
      predict(object, newdata = newdata, type = "posterior", prior = prior)
    }

  )

}

MLModelFunction(MDAModel) <- NULL

setAsMLInput <- function(from, to) {
  setAs(from, to, function(from) {
    new(to, asS3(S3Part(from)), id = from@id, params = from@params)
  })
}


setAsS3Part <- function(from, to) {
  setAs(from, to, function(from) {
    if (!isS4(from)) throw(TypeError(from, "S4 class"))
    asS3(S3Part(from))
  })
}


#' Coerce to a Data Frame
#'
#' Functions to coerce objects to data frames.
#'
#' @name as.data.frame
#'
#' @param x \code{\link{ModelFrame}}, \link{resample} results, resampled
#'   \link{performance} estimates, model performance \link[=diff]{differences},
#'   or \link[=t.test]{t-test} comparisons of the differences.
#' @param ... arguments passed to other methods.
#'
#' @return \code{data.frame} class object.
#'
NULL


as.data.frame.BinomialVariate <- function(x, ...) {
  as.data.frame.model.matrix(x, ...)
}


as.data.frame.formula <- function(x, ..., data) {
  eval.parent(substitute(environment(x) <- environment()))
  as.data.frame(data, ...)
}


#' @rdname as.data.frame
#'
as.data.frame.ModelFrame <- function(x, ...) {
  structure(asS3(S3Part(x)), terms = NULL)
}


setAs("ModelFrame", "data.frame",
  function(from) as.data.frame(from)
)


setAs("SelectedModelFrame", "data.frame",
  function(from) as.data.frame(from)
)


as.data.frame.ModelRecipe <- function(x, ...) {
  as.data.frame(x[[if (is_trained(x)) "orig_template" else "template"]])
}


as.data.frame.ModelSpecification <- function(x, ...) {
  as.data.frame(as.MLInput(x))
}


as.data.frame.PerformanceDiffTest <- function(x, ...) {
  stat_names <- matrix(NA_character_, nrow(x), ncol(x))
  stat_names[upper.tri(stat_names)] <- "Mean"
  stat_names[lower.tri(stat_names)] <- "P-Value"
  df_stat_names <- as.data.frame(TabularArray(stat_names))

  x <- cbind(NextMethod(), Statistic = df_stat_names$Value)
  x <- x[!is.na(x$Statistic), ]
  is_pval <- x$Statistic == "P-Value"
  x[is_pval, c("Model1", "Model2")] <- x[is_pval, c("Model2", "Model1")]
  x$Model <- paste(x$Model1, "-", x$Model2)

  ind <- order(x$Statistic, x$Metric, x$Model)
  x <- x[ind, c("Statistic", "Metric", "Model", "Value")]
  rownames(x) <- NULL
  x
}


#' @rdname as.data.frame
#'
as.data.frame.Resample <- function(x, ...) {
  asS3(as(x, "data.frame"))
}


as.data.frame.SurvMatrix <- function(x, ...) {
  as.data.frame.model.matrix(x, ...)
}


#' @rdname as.data.frame
#'
as.data.frame.TabularArray <- function(x, ...) {
  as.data.frame.table(as(x, "array"), responseName = "Value", ...)
}


as.double.BinomialVariate <- function(x, ...) {
  as.numeric(x[, "Success"] / (x[, "Success"] + x[, "Failure"]))
}


setAs("EnsembleInputOrModel", "list",
  function(from) c(list(from@candidates), as(from@params, "list"))
)


setAs("MLModel", "list",
  function(from) as(from@params, "list")
)


setAs("StackedModel", "list",
  function(from) {
    c(list(
      candidates = from@candidates,
      control = from@params@control),
      from@params@options
    )
  }
)


setAs("SuperModel", "list",
  function(from) {
    res <- callNextMethod()
    res$model <- res$candidates[[1]]
    res$candidates[1] <- NULL
    res
  }
)


setAs("TrainingParams", "list",
  function(from) {
    res <- map(function(name) slot(from, name), slotNames(from))
    options <- res$options
    res$optim <- NULL
    res$options <- NULL
    c(new_params(res), options)
  }
)


as.MLControl <- function(x, ...) {
  UseMethod("as.MLControl")
}


as.MLControl.default <- function(x, ...) {
  throw(Error("Cannot coerce class ", class1(x), " to an MLControl."))
}


as.MLControl.character <- function(x, ...) {
  x <- get0(x)
  as.MLControl(x)
}


as.MLControl.function <- function(x, ...) {
  result <- try(x(), silent = TRUE)
  if (!is(result, "MLControl")) {
    throw(Error("Cannot coerce specified function to an MLControl"))
  } else result
}


as.MLControl.MLControl <- function(x, ...) {
  x
}


as.MLControl.NULL <- function(x, ...) {
  NullControl()
}


#' Coerce to an MLInput
#'
#' Function to coerce an object to \code{MLInput}.
#'
#' @param x model \link{fit} result or \pkg{MachineShop}
#'   \link[=ModelSpecification]{model specification}.
#' @param ... arguments passed to other methods.
#'
#' @return \code{MLInput} class object.
#'
as.MLInput <- function(x, ...) {
  UseMethod("as.MLInput")
}


as.MLInput.default <- function(x, ...) {
  throw(Error("Cannot coerce class ", class1(x), " to an MLInput."))
}


as.MLInput.formula <- function(x, data, ...) {
  args <- list(x, data, strata = response(x), na.rm = FALSE)
  do.call(ModelFrame, args)
}


as.MLInput.matrix <- function(x, y, ...) {
  ModelFrame(x, y, strata = y, na.rm = FALSE)
}


as.MLInput.MLInput <- function(x, ...) {
  x
}


#' @rdname as.MLInput
#'
as.MLInput.MLModelFit <- function(x, ...) {
  attr(update(x), ".MachineShop")$input
}


#' @rdname as.MLInput
#'
as.MLInput.ModelSpecification <- function(x, ...) {
  x@input
}


as.MLInput.recipe <- function(x, ...) {
  ModelRecipe(x)
}


as.MLMetric <- function(x, ...) {
  UseMethod("as.MLMetric")
}


as.MLMetric.default <- function(x, ...) {
  throw(Error("Cannot coerce class ", class1(x), " to an MLMetric."))
}


as.MLMetric.character <- function(x, ...) {
  x <- get0(x)
  if (is.null(x)) throw(Error("Cannot coerce \"", x, "\" to an MLMetric."))
  as.MLMetric(x)
}


as.MLMetric.MLMetric <- function(x, ...) {
  x
}


#' Coerce to an MLModel
#'
#' Function to coerce an object to \code{MLModel}.
#'
#' @param x model \link{fit} result, \pkg{MachineShop}
#'   \link[=ModelSpecification]{model specification}, or
#'   \pkg{parsnip} \link[parsnip:model_spec]{model specification}.
#' @param ... arguments passed to other methods.
#'
#' @return \code{MLModel} class object.
#'
#' @seealso \code{\link{ParsnipModel}}
#'
as.MLModel <- function(x, ...) {
  UseMethod("as.MLModel")
}


as.MLModel.default <- function(x, ...) {
  throw(Error("Cannot coerce class ", class1(x), " to an MLModel."))
}


as.MLModel.character <- function(x, ...) {
  x <- get0(x)
  if (is.null(x)) throw(Error("Cannot coerce \"", x, "\" to an MLModel."))
  as.MLModel(x)
}


as.MLModel.MLModel <- function(x, ...) {
  update(x)
}


#' @rdname as.MLModel
#'
as.MLModel.MLModelFit <- function(x, ...) {
  attr(update(x), ".MachineShop")$model
}


as.MLModel.MLModelFunction <- function(x, ...) {
  x()
}


#' @rdname as.MLModel
#'
as.MLModel.ModelSpecification <- function(x, ...) {
  x@model
}


#' @rdname as.MLModel
#'
as.MLModel.model_spec <- function(x, ...) {
  ParsnipModel(x)
}


as.MLModel.NULL <- function(x, ...) {
  NullModel()
}


setAsMLInput("SelectedModelFrame", "ModelFrame")


setAs("recipe", "ModelRecipe",
  function(from) ModelRecipe(from)
)


setAsMLInput("SelectedModelRecipe", "ModelRecipe")


setAsMLInput("TunedModelRecipe", "ModelRecipe")


setAsS3Part("ParameterGrid", "parameters")


setAsS3Part("ModelRecipe", "recipe")


setAsS3Part("SelectedModelRecipe", "recipe")


setAsS3Part("TunedModelRecipe", "recipe")


setAsS3Part("RecipeGrid", "tbl_df")

#################### Base Classes ####################


setOldClass("listof")
setOldClass(c("tbl_df", "tbl", "data.frame"))
setOldClass(c("xtabs", "table"))


ListOf <- setClass("ListOf",
  contains = c("listof", "list")
)


TabularArray <- setClass("TabularArray",
  contains = "array"
)


setClassUnion("IntegerOrLogical",
  members = c("integer", "logical")
)


#################### Response Variables ####################


setOldClass("BinomialVariate")
setOldClass("Surv")


setClass("DiscreteVariate",
  contains = "numeric",
  slots = c(
    min = "numeric",
    max = "numeric"
  )
)


setClass("NegBinomialVariate",
  contains = "DiscreteVariate"
)


setClass("PoissonVariate",
  contains = "DiscreteVariate"
)


setClass("SurvMatrix",
  contains = "matrix",
  slots = c(
    times = "numeric",
    distr = "character"
  )
)


setClass("SurvEvents",
  contains = "SurvMatrix"
)


setClass("SurvProbs",
  contains = "SurvMatrix"
)


setClass("SurvTimes",
  contains = "numeric",
  slots = c(distr = "character")
)


setClass("SurvMeans", contains = "SurvTimes")


#################### Resampling Controls ####################


setClass("MLControl",
  slots = c(
    label = "character",
    monitor = "list",
    predict = "list",
    strata = "list",
    weights = "logical",
    seed = "numeric"
  )
)


setClass("BootControl",
  contains = "MLControl",
  slots = c(samples = "integer")
)


setClass("BootOptimismControl",
  contains = "BootControl"
)


setClass("CVControl",
  contains = "MLControl",
  slots = c(
    folds = "integer",
    repeats = "integer"
  )
)


setClass("CVOptimismControl",
  contains = "CVControl"
)


setClass("NullControl", contains = "MLControl")


setClass("OOBControl",
  contains = "MLControl",
  slots = c(samples = "integer")
)


setClass("SplitControl",
  contains = "MLControl",
  slots = c(prop = "numeric")
)


setClass("TrainControl",
  contains = "MLControl"
)


#################### Parameters and Grids ####################


setOldClass(c("parameters", "tbl_df"))


setClass("MLOptimization",
  slots = c(
    label = "character",
    packages = "character",
    params = "list",
    monitor = "list",
    fun = "function"
  )
)


setClass("GridSearch", contains = "MLOptimization")
setClass("NullOptimization", contains = "MLOptimization")

setClass("RandomGridSearch",
  contains = "GridSearch",
  slots = c(size = "integer")
)

setClass("SequentialOptimization",
  contains = "MLOptimization",
  slots = c(random = "IntegerOrLogical")
)


setClass("TrainingParams",
  slots = c(
    optim = "MLOptimization",
    control = "MLControl",
    metrics = "ANY",
    cutoff = "numeric",
    stat = "function",
    options = "list"
  )
)


setClassUnion("Params",
  members = c("list", "TrainingParams")
)


setClass("TuningGrid",
  slots = c(
    size = "integer",
    random = "IntegerOrLogical"
  )
)


setClass("ParameterGrid",
  contains = c("TuningGrid", "parameters")
)


RecipeGrid <- setClass("RecipeGrid",
  contains = "tbl_df"
)


setClassUnion("Grid",
  members = c("tbl_df", "TuningGrid")
)


#################### Model Components ####################


setClass("EnsembleInputOrModel",
  contains = "VIRTUAL",
  slots = c(
    candidates = "ListOf",
    params = "TrainingParams"
  )
)


setClass("MLInput",
  slots = c(
    id = "character",
    params = "ANY"
  )
)


setClass("MLModel",
  slots = c(
    id = "character",
    name = "character",
    label = "character",
    packages = "character",
    response_types = "character",
    weights = "logical",
    predictor_encoding = "character",
    na.rm = "character",
    params = "Params",
    gridinfo = "tbl_df",
    fit = "function",
    predict = "function",
    varimp = "function",
    steps = "ListOf"
  )
)


#################### Model Inputs ####################


setOldClass("data.frame")
setOldClass("recipe")
setOldClass(c("terms", "formula"))


setClass("ModelFrame",
  contains = c("data.frame", "MLInput")
)


setClass("ModelTerms",
  contains = "terms",
  slots = c(id = "character")
)


ModelDesignTerms <- setClass("ModelDesignTerms",
  contains = "ModelTerms"
)


ModelFormulaTerms <- setClass("ModelFormulaTerms",
  contains = "ModelTerms"
)


setClass("ModelRecipe",
  contains = c("recipe", "MLInput")
)


setClass("ModelSpecification",
  slots = c(
    id = "character",
    input = "MLInput",
    model = "MLModel",
    params = "TrainingParams",
    grid = "tbl_df"
  )
)


setClass("PredictorFrame", contains = "ModelFrame")


setClass("SelectedInput",
  contains = c("EnsembleInputOrModel", "MLInput")
)


setClass("SelectedModelFrame",
  contains = c("SelectedInput", "EnsembleInputOrModel", "ModelFrame")
)


setClass("SelectedModelRecipe",
  contains = c("SelectedInput", "EnsembleInputOrModel", "ModelRecipe")
)


setClass("SelectedModelSpecification",
  contains = c("SelectedInput", "EnsembleInputOrModel", "ModelSpecification")
)


setClass("TunedInput",
  contains = "VIRTUAL",
  slots = c(
    grid = "Grid",
    params = "TrainingParams"
  )
)


setClass("TunedModelRecipe",
  contains = c("TunedInput", "ModelRecipe"),
  slots = c(grid = "RecipeGrid")
)


#################### Models ####################


setClass("EnsembleModel", contains = c("EnsembleInputOrModel", "MLModel"))
setClass("NullModel", contains = "MLModel")
setClass("ParsnipModel", contains = "MLModel")
setClass("SelectedModel", contains = "EnsembleModel")
setClass("StackedModel", contains = "EnsembleModel")
setClass("SuperModel", contains = "StackedModel")


setClass("TunedModel",
  contains = "MLModel",
  slots = c(
    model = "MLModel",
    grid = "Grid",
    params = "TrainingParams"
  )
)


setClass("MLModelFunction",
  contains = "function"
)


#################### Model Fits ####################


setClass("MLModelFit",
  contains = "VIRTUAL",
  slots = c(.MachineShop = "list")
)


setClass("CForestModelFit", contains = c("MLModelFit", "RandomForest"))
setClass("SVMModelFit", contains = c("MLModelFit", "ksvm"))
setClass("SVMANOVAModelFit", contains = c("MLModelFit", "ksvm"))
setClass("SVMBesselModelFit", contains = c("MLModelFit", "ksvm"))
setClass("SVMLaplaceModelFit", contains = c("MLModelFit", "ksvm"))
setClass("SVMLinearModelFit", contains = c("MLModelFit", "ksvm"))
setClass("SVMPolyModelFit", contains = c("MLModelFit", "ksvm"))
setClass("SVMRadialModelFit", contains = c("MLModelFit", "ksvm"))
setClass("SVMSplineModelFit", contains = c("MLModelFit", "ksvm"))
setClass("SVMTanhModelFit", contains = c("MLModelFit", "ksvm"))


#################### Performance Measures ####################


setClass("Calibration",
  contains = "data.frame",
  slots = c(smoothed = "logical")
)


ConfusionList <- setClass("ConfusionList",
  contains = "ListOf"
)


setClass("ConfusionMatrix",
  contains = c("table", "matrix")
)


setClass("BinaryConfusionMatrix",
  contains = "ConfusionMatrix"
)


setClass("OrderedConfusionMatrix",
  contains = "ConfusionMatrix"
)


setClass("OrderedBinaryConfusionMatrix",
  contains = c("OrderedConfusionMatrix", "BinaryConfusionMatrix")
)


ConfusionSummary <- setClass("ConfusionSummary",
  contains = "matrix",
  slots = c(
    total = "numeric",
    accuracy = "numeric",
    majority = "numeric",
    kappa2 = "numeric"
  )
)


setClass("MLMetric",
  contains = "function",
  slots = c(
    name = "character",
    label = "character",
    maximize = "logical"
  )
)


setClass("Performance",
  contains = "TabularArray",
  slots = c(control = "MLControl")
)


PerformanceDiff <- setClass("PerformanceDiff",
  contains = "Performance",
  slots = c(model_names = "character")
)


PerformanceDiffTest <- setClass("PerformanceDiffTest",
  contains = "TabularArray",
  slots = c(adjust = "character")
)


setClass("PerformanceCurve",
  contains = "data.frame",
  slots = c(
    metrics = "list",
    control = "MLControl"
  )
)


setClass("LiftCurve",
  contains = "PerformanceCurve"
)


setClass("Resample",
  contains = "data.frame",
  slots = c(
    control = "MLControl",
    vars = "tbl_df"
  )
)


setClass("TrainingStep",
  slots = c(
    id = "character",
    name = "character",
    method = "character",
    log = "tbl_df",
    performance = "Performance"
  )
)


setClass("VariableImportance",
  contains = "data.frame",
  slots = c(
    method = "character",
    metric = "character",
    scale = "numeric"
  )
)

---
title: "Conventions for MLModels Implementation"
author: "Brian J Smith"
date: "2021-07-23"
output: rmarkdown::html_vignette
bibliography: bibliography.bib
vignette: >
  %\VignetteIndexEntry{Conventions for MLModels Implementation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```


## Model Constructor Components

- `MLModel` is a function supplied by the **MachineShop** package.  It allows for the integration of statistical and machine learning models supplied by other R packages with the **MachineShop** model fitting, prediction, and performance assessment tools.

- The following are guidelines for writing model constructor functions that are wrappers around the `MLModel` function.

- In this context, the term "constructor" refers to the wrapper function and "source package" to the package supplying the original model implementation.


### Constructor Arguments

- The constructor should produce a valid model if called without any arguments; i.e., not have any required arguments.

- The source package defaults will be used for parameters with `NULL` values.

- Model formula, data, and weights are separate from model parameters and should not be defined as constructor arguments.


### name Slot

- Use the same name as the constructor.


### packages Slot

- Include all external packages whose functions are called directly from within the constructor.

- Use :: to reference source package functions.


### response_types Slot

- Include all response variable types (`"binary"`, `"factor"`, `"matrix"`, `"numeric"`, `"ordered"`, and/or `"Surv"`) that can be analyzed with the model.


### weights Slot

- Logical indicating whether the model supports case weights.


### params Slot

- List of parameter values set by the constructor, typically obtained internally with `new_params(environment())` if all arguments are to be passed to the source package fit function as supplied.  Additional steps may be needed to pass the constructor arguments to the source package in a different format; e.g., when some model parameters must be passed in a control structure, as in `C50Model` and `CForestModel`.


### fit Function

- The first three arguments should be `formula`, `data`, and `weights` followed by an ellipsis (`...`).

- If weights are not supported, the following, or equivalent, should be included in the function:

```{r eval = FALSE}
if(!all(weights == 1)) warning("weights are not supported and will be ignored")
```

- Only add elements to the resulting fit object if they are needed and will be used in the `predict` or `varimp` functions.

- Return the fit object.


### predict Function

- The arguments are a model fit `object`, `newdata` frame, optionally `times` for prediction at survival time points, and an ellipsis.

- The predict function should return a vector or column matrix of probabilities for the second level of binary factors, a matrix whose columns contain the probabilities for factors with more than two levels, a matrix of predicted responses if matrix, a vector or column matrix of predicted responses if numeric, a matrix whose columns contain survival probabilities at `times` if supplied, or a vector of predicted survival means if `times` are not supplied.


### varimp Function

- Should have a single model fit `object` argument followed by an ellipsis.

- Variable importance results should generally be returned as a vector with elements named after the corresponding predictor variables.  The package will handle conversions to a data frame and `VariableImportance` object.  If there is more than one set of relevant variable importance measures, they can be returned as a matrix or data frame with predictor variable names as the row names.


## Documenting an MLModel


### Model Parameters

- Include the first sentences from the source package.

- Start sentences with the parameter value type (logical, numeric, character, etc.).

- Start sentences with lowercase.

- Omit indefinite articles (a, an, etc.) from the starting sentences.


### Details Section

- Include response types (binary, factor, matrix, numeric, ordered, and/or Surv).

- Include the following sentence:
  
> Default values for the \code{NULL} arguments and further model details can be
> found in the source link below.


### Return (Value) Section

- Include the following sentence:

> MLModel class object.


### See Also Section

- Include a link to the source package function and the other method functions shown below.

```
\code{\link[<source package>]{<fit function>}}, \code{\link{fit}},
\code{\link{resample}}
```

## Package Extensions

- If adding a new model to the package, save its source code in a file whose name begins with "ML_" followed by the model name, and ending with a .R extension; e.g., `"R/ML_CustomModel.R"`.

- Export the model in `NAMESPACE`.

- Add any required packages to the "Suggests" section of `DESCRIPTION`.

- Add the model to `R/models.R`.

- Add the model to `R/modelinfo.R`.

- Add a unit testing file to `tests/testthat`.

---
title: "MachineShop User Guide"
subtitle: "R Package Version `r packageVersion('MachineShop')`"
author:
- name: "Brian J Smith"
  affiliation: "University of Iowa"
  email: "brian-j-smith@uiowa.edu"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 3
    number_sections: true
bibliography: bibliography.bib
vignette: >
  %\VignetteIndexEntry{MachineShop User Guide}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
source("setup.R")
rdoc_url <- function(names, ...) names
```


# Description

**MachineShop** is a meta-package for statistical and machine learning with a unified interface for model fitting, prediction, performance assessment, and presentation of results.  Support is provided for predictive modeling of numerical, categorical, and censored time-to-event outcomes and for resample (bootstrap, cross-validation, and split training-test sets) estimation of model performance.  This vignette introduces the package interface with a survival data analysis example, followed by supported methods of variable specification; applications to other response variable types; available performance metrics, resampling techniques, and graphical and tabular summaries; and modeling strategies.


# Features

```{r overview_modelinfo, echo=FALSE, message=FALSE}
library(MachineShop)
info <- modelinfo()
```

* Unified and concise interface for model fitting, prediction, and performance assessment.
* Support for `r length(info)`+ models from `r length(unique(sapply(info, function(x) x$packages[1])))` **R** packages, including model specifications from the [**parsnip**](https://parsnip.tidymodels.org/) package.
* Dynamic model parameters.
* Ensemble modeling with stacked regression and super learners.
* Modeling of response variables types: binary factors, multi-class nominal and ordinal factors, numeric vectors and matrices, and censored time-to-event survival.
* Model specification with traditional formulas, design matrices, and flexible pre-processing [recipes](https://recipes.tidymodels.org/).
* Resample estimation of predictive performance, including cross-validation, bootstrap resampling, and split training-test set validation.
* Parallel execution of resampling algorithms.
* Choices of performance metrics: accuracy, areas under ROC and precision recall curves, Brier score, coefficient of determination (R^2^), concordance index, cross entropy, F score, Gini coefficient, unweighted and weighted Cohen's kappa, mean absolute error, mean squared error, mean squared log error, positive and negative predictive values, precision and recall, and sensitivity and specificity.
* Graphical and tabular performance summaries: calibration curves, confusion matrices, partial dependence plots, performance curves, lift curves, and model-specific and permutation-based variable importance.
* Model tuning over automatically generated grids and with exhaustive and random grid searches, Bayesian optimization, particle swarm optimization, quasi-Newton BFGS optimization, simulated annealing, and support for user-defined optimization functions.
* Model selection and comparisons for any combination of models and model parameter values.
* Recursive feature elimination.
* User-definable models and performance metrics.


# Getting Started


## Installation

```{r overview_install, eval = FALSE}
# Current release from CRAN
install.packages("MachineShop")

# Development version from GitHub
# install.packages("devtools")
devtools::install_github("brian-j-smith/MachineShop")

# Development version with vignettes
devtools::install_github("brian-j-smith/MachineShop", build_vignettes = TRUE)
```


## Documentation

Once installed, the following **R** commands will load the package and display its help system documentation.  Online documentation and examples are available at the [MachineShop website](https://brian-j-smith.github.io/MachineShop/).

```{r overview_docs, eval = FALSE, message = FALSE}
library(MachineShop)

# Package help summary
?MachineShop

# Vignette
RShowDoc("UserGuide", package = "MachineShop")
```


# Melanoma Example

Use of the **MachineShop** package is demonstrated with a survival analysis example in which the response variable is a censored time to event outcome.  Since survival outcomes are a combination of numerical (time to event) and categorical (event) variables, package features for both variable types are illustrated with the example.  Support for outcomes other than survival, including nominal and ordinal factors as well as numeric vectors and matrices, is also discussed.

Survival analysis is performed with the `Melanoma` dataset from the **MASS** package [@andersen:1993:SMB].  This dataset provides survival time, in days, from disease treatment to (1) death from disease, (2) alive at study termination, or (3) death from other causes for 205 Denmark patients with malignant melanomas.  Also provided are potential predictors of the survival outcomes.  The analysis begins by loading required packages **MachineShop**, **survival** [@therneau:2020:SA], and **MASS**.  For the analysis, a binary overall survival outcome is created by combining the two death categories (1 and 3) into one. 

```{r using_example_melanoma}
## Analysis libraries and dataset
library(MachineShop)
library(survival)
data(Melanoma, package = "MASS")

## Malignant melanoma analysis dataset
surv_df <- within(Melanoma, {
  y <- Surv(time, status != 2)
  remove(time, status)
})
```

Descriptive summaries of the study variables are given below in Table 1, followed by a plot of estimated overall survival probabilities and 95% confidence intervals.

<center>
Table 1. Variable summaries for the Melanoma survival analysis example.
</center>

```{r using_example_summary, echo=FALSE}
surv_stats <- list(
  list("Number of subjects" = ~ length(y)),
  "time" = list("Median (Range)" = ~ median_range(y[, "time"])),
  "status" = list("1 = Dead" = ~ n_perc(y[, "status"] == 1),
                  "0 = Alive" = ~ n_perc(y[, "status"] == 0)),
  "sex" = list("1 = Male" = ~ n_perc(sex == 1),
               "0 = Female" = ~ n_perc(sex == 0)),
  "age" = list("Median (Range)" = ~ median_range(age)),
  "year" = list("Median (Range)" = ~ median_range(year)),
  "thickness" = list("Median (Range)" = ~ median_range(thickness)),
  "ulcer" = list("1 = Presence" = ~ n_perc(ulcer == 1),
                 "0 = Absence" = ~ n_perc(ulcer == 0))
)

summary_kbl(surv_stats, surv_df)
```

```{r using_example_survfit, echo=FALSE}
library(ggplot2)

col <- "#F8766D"
survfit(y ~ 1, data = surv_df) %>%
  with(data.frame(time, surv, lower, upper, censor = ifelse(n.censor > 0, time, NA))) %>%
  ggplot(aes(x = time, y = surv)) +
  geom_ribbon(aes(ymin = lower, ymax = upper), fill = col, alpha = 0.2) +
  geom_step(color = col) +
  geom_point(aes(x = censor), shape = 3, color = col) +
  coord_cartesian(ylim = c(0, 1)) +
  labs(x = "Follow-Up Time (Days)", y = "Overall Survival Probability",
       title = "Kaplan-Meier survival plot")
```

For the analyses, the dataset is split into a training set on which survival models will be fit and a test set on which predictions will be made and performance evaluated.  A global formula `surv_fo` is defined to relate the predictors on the right hand side to the overall survival outcome on the left and will be used for all subsequent survival models.

```{r using_example_datasets}
## Training and test sets
set.seed(123)
train_indices <- sample(nrow(surv_df), nrow(surv_df) * 2 / 3)
surv_train <- surv_df[train_indices, ]
surv_test <- surv_df[-train_indices, ]

## Global formula for the analysis
surv_fo <- y ~ sex + age + year + thickness + ulcer
```


# Model Fit and Prediction


## Model Information

Model fitting requires user specification of a **MachineShop** compatible model.  A named list of package-supplied models can be obtained interactively with the `r rdoc_url("modelinfo()")` function, and includes the following components for each.

`label`
  : character descriptor for the model.

`packages`
  : character vector of source packages required to use the model.  These need only be installed with the `install.packages()` function or by equivalent means; but need not be loaded with, for example, the `library()` function.

`response_types`
  : character vector of response variable types supported by the model.

`weights`
  : logical value or vector of the same length as `response_types` indicating whether case weights are supported for the responses.

`na.rm`
  : character sting specifying removal of `"all"` cases with missing values from model fitting, `"none"`, or only those whose missing values are in the `"response"` variable.

`arguments`
  : closure with the argument names and corresponding default values of the model function.

`grid`
  : logical indicating whether automatic generation of tuning parameter grids is implemented for the model.

`varimp`
  : logical indicating whether model-specific variable importance is defined.

Function `r rdoc_url("modelinfo()")` may be called without arguments, with one or more model functions, observed response variables, or vectors representing response variable types; and will return information on all matching models.

```{r using_fit_modelinfo}
## All available models
modelinfo() %>% names
```

Information is displayed below for the `r rdoc_url("GBMModel()")` function corresponding to a generalized boosted regression model, which is applicable to survival outcomes.

```{r using_fit_modelinfo_gbmmodel}
## Model-specific information
modelinfo(GBMModel)
```

Submitting the model function at the console will result in similar information being displayed as formatted text.

```{r using_fit_gbmmodel}
GBMModel
```


### Type-Specific Models

When data objects are supplied as arguments to `r rdoc_url("modelinfo()")`, information is returned on all models applicable to response variables of the same data types.  If model functions are additionally supplied as arguments, information on the subset matching the data types is returned.

```{r using_fit_modelinfo_type}
## All survival response-specific models
modelinfo(Surv(0)) %>% names

## Identify survival response-specific models
modelinfo(Surv(0), CoxModel, GBMModel, SVMModel) %>% names
```


### Response Variable-Specific Models

As a special case of type-specific arguments, existing response variables to be used in analyses may be given as arguments to identify applicable models.

```{r using_fit_modelinfo_response}
## Models for a responses variable
modelinfo(surv_df$y) %>% names
```


## Fit Function

Package models, such as `r rdoc_url("GBMModel")`, can be specified in the `model` argument of the `r rdoc_url("fit()")` function to estimate a relationship (`surv_fo`) between predictors and an outcome based on a set of data (`surv_train`).  Argument specifications may be in terms of a model function, function name, or object.

```{r using_fit_function, results="hide"}
## Generalized boosted regression fit

## Model function
surv_fit <- fit(surv_fo, data = surv_train, model = GBMModel)

## Model function name
fit(surv_fo, data = surv_train, model = "GBMModel")

## Model object
fit(surv_fo, data = surv_train, model = GBMModel(n.trees = 100, interaction.depth = 1))
```

Model function arguments will assume their default values unless otherwise changed in a function call.


## Dynamic Model Parameters

*Dynamic model parameters* are model function arguments defined as expressions to be evaluated at the time of model fitting.  As such, their values can change based on characteristics of the analytic dataset, including the number of observations or predictor variables.  Expressions to dynamic parameters are specified within the package-supplied quote operator `r rdoc_url(".()", "quote")` and can include the following objects:

`nobs`
  : number of observations in `data`.

`nvars`
  : number of predictor variables in `data`.

`y`
  : response variable.

In the example below, Bayesian information criterion (BIC) based stepwise variable selection is performed by creating a `r rdoc_url("CoxStepAICModel", "CoxModel")` with dynamic parameter `k` to be calculated as the log number of observations in the fitted dataset.

```{r using_fit_dynamic, results="hide"}
## Dynamic model parameter k = log number of observations

## Number of observations: nobs
fit(surv_fo, data = surv_train, model = CoxStepAICModel(k = .(log(nobs))))

## Response variable: y
fit(surv_fo, data = surv_train, model = CoxStepAICModel(k = .(log(length(y)))))
```


## Predict Function

A `r rdoc_url("predict()")` function is supplied for application to model fit results to obtain predicted values on a dataset specified with its `newdata` argument or on the original dataset if not specified.  Survival means are predicted for survival outcomes by default.  Estimates of the associated survival distributions are needed to calculate the means.  For models, like `r rdoc_url("GBMModel")`, that perform semi- or non-parametric survival analysis, Weibull approximations to the survival distributions are the default for mean estimation.  Other choices of distributional approximations are exponential, Rayleigh, and empirical.  Empirical distributions are applicable to Cox proportional hazards-based models and can be calculated with the method of Breslow [-@breslow:1972:DPC] or Efron [-@efron:1977:ECL, default].  Note, however, that empirical survival means are undefined mathematically if an event does not occur at the longest follow-up time.  In such situations, a restricted survival mean is calculated by changing the longest follow-up time to an event, as suggested by Efron [-@efron:1967:PFB], which will be negatively biased.

```{r using_predict_function}
## Predicted survival means (default: Weibull distribution)
predict(surv_fit, newdata = surv_test)

## Predicted survival means (empirical distribution)
predict(surv_fit, newdata = surv_test, distr = "empirical")
```

In addition to survival means, predicted survival probabilities (`type = "prob"`) or 0-1 survival events (default: `type = "response"`) can be obtained with the follow-up `times` argument.  The cutoff probability for classification of survival events (or other binary responses) can be set optionally with the `cutoff` argument (default: `cutoff = 0.5`).  As with mean estimation, distributional approximations to the survival functions may be specified for the predictions, with the default for survival probabilities being the empirical distribution.

```{r using_predict_function_times}
## Predict survival probabilities and events at specified follow-up times
surv_times <- 365 * c(5, 10)

predict(surv_fit, newdata = surv_test, times = surv_times, type = "prob")

predict(surv_fit, newdata = surv_test, times = surv_times, cutoff = 0.7)
```

Prediction of other outcome types is more straightforward.  Predicted numeric and factor responses are of the same class as the observed values at the default `type = "response"`; whereas, double (decimal) numeric values and factor level probabilities result when `type = "prob"`.




# Variable Specifications

Variable specification defines the relationship between response and predictor variables as well as the data used to estimate the relationship.  Four main types of specifications are supported by the package's `r rdoc_url("fit()")` and `r rdoc_url("resample()")` functions: traditional formula, design matrix, model frame, and recipe.


## Traditional Formula

Variables may be specified with a traditional formula and data frame pair, as was done at the start of the survival example.  This specification allows for crossing (`*`), interaction (`:`), and removal (`-`) of predictors in the formula; `.` substitution of variables not already appearing in the formula; in-line functions of response variables; and in-lining of operators and functions of predictors.

```{r using_variables_formula}
## Datasets
data(Pima.te, package = "MASS")
data(Pima.tr, package = "MASS")

## Formula specification
model_fit <- fit(type ~ ., data = Pima.tr, model = GBMModel)
predict(model_fit, newdata = Pima.te) %>% head
```

The syntax for traditional formulas is detailed in the **R** help documentation on the `formula()` function.  However, some constraints are placed on the syntax by the **MachineShop** package.  Specifically, in-lining on the right-hand side of formulas is limited to the operators and functions listed in the `"RHS.formula"` package setting.

```{r using_variables_formula_RHS}
settings("RHS.formula")
```

This setting is intended to help avoid the definition of predictor variable encodings that involve dataset-specific parameter calculations.  Such parameters would be calculated separated on training and test sets, and could lead to failed calculations or improper estimates of predictive performance.  For example, the `factor()` function is not allowed because consistency of its (default) encoding requires that all levels be present in every dataset.  Resampled datasets subset the original cases and are thus prone to missing factor levels.  For users wishing to apply factor encodings or other encodings not available with traditional formulas, a more flexible preprocessing recipe syntax is supported, as described later.


## Design Matrix

Variables stored separately in a design matrix of predictors and object of responses can be supplied to the fit functions directly.  Fitting with design matrices has less computational overhead than traditional formulas and allows for greater numbers of predictor variables in some models, including `r rdoc_url("GBMModel")`, `r rdoc_url("GLMNetModel")`, and `r rdoc_url("RandomForestModel")`.

```{r using_variables_matrix}
## Example design matrix and response object
x <- model.matrix(type ~ . - 1, data = Pima.tr)
y <- Pima.tr$type

## Design matrix specification
model_fit <- fit(x, y, model = GBMModel)
predict(model_fit, newdata = Pima.te) %>% head
```


## Model Frame

A `r rdoc_url("ModelFrame")` class is defined by the package for specification of predictor and response variables along with other attributes to control model fitting.  Model frames can be created with calls to the `r rdoc_url("ModelFrame()")` constructor function using a syntax similar to the traditional formula or design matrix.

```{r using_variables_modelframe}
## Model frame specification

## Formula
mf <- ModelFrame(type ~ ., data = Pima.tr)
model_fit <- fit(mf, model = GBMModel)
predict(model_fit, newdata = Pima.te) %>% head

## Design matrix
mf <- ModelFrame(x, y)
model_fit <- fit(mf, model = GBMModel)
predict(model_fit, newdata = Pima.te) %>% head
```

The model frame approach has a few advantages over model fitting directly with a traditional formula.  One is that cases with missing values on any of the response or predictor variables are excluded from the model frame by default.  This is often desirable for models that do not handle missing values.  Conversely, missing values can be retained in the model frame by setting its argument `na.rm = FALSE` for models, like `r rdoc_url("GBMModel")`, that do handle them.  A second advantage is that case weights can be included in the model frame to be passed on to the model fitting functions.

```{r using_variables_modelframe_weights, results="hide"}
## Model frame specification with case weights
mf <- ModelFrame(ncases / (ncases + ncontrols) ~ agegp + tobgp + alcgp, data = esoph,
                 weights = ncases + ncontrols)
fit(mf, model = GBMModel)
```

A third, which will be illustrated later, is user-specification of a variable for stratified resampling via the constructor's `strata` argument.


## Preprocessing Recipe

The **recipes** package [@kuhn:2020:RPT] provides a flexible framework for defining predictor and response variables as well as preprocessing steps to be applied to them prior to model fitting.  Using recipes helps ensure that estimation of predictive performance accounts for all modeling step.  They are also a convenient way of consistently applying preprocessing to new data.  A basic recipe is given below in terms of the formula and data frame ingredients needed for the analysis.

```{r using_variables_recipe}
## Recipe specification
library(recipes)

rec <- recipe(type ~ ., data = Pima.tr)
model_fit <- fit(rec, model = GBMModel)
predict(model_fit, newdata = Pima.te) %>% head
```

As shown, prediction on new data with a model fit to a recipe is done on an unprocessed dataset.  Recipe case weights and stratified resampling are supported with the `r rdoc_url("role_case()", "recipe_roles")` function.  As an example, an initial step is included in the recipe below to replace the original role of variable `weights` with a designation of case weights.  That is followed by a step to convert three ordinal factors to integer scores.

```{r using_variables_recipe_weights, results="hide"}
## Recipe specification with case weights
df <- within(esoph, {
  y <- ncases / (ncases + ncontrols)
  weights <- ncases + ncontrols
  remove(ncases, ncontrols)
})

rec <- recipe(y ~ agegp + tobgp + alcgp + weights, data = df) %>%
  role_case(weight = weights, replace = TRUE) %>%
  step_ordinalscore(agegp, tobgp, alcgp)
fit(rec, model = GBMModel)
```


## Summary

The variable specification approaches differ with respect to support for preprocessing, in-line functions, case weights, resampling strata, and computational overhead, as summarized in the table below.  Only recipes apply preprocessing steps automatically during model fitting and should be used when it is important to account for such steps in the estimation of model predictive performance.  Preprocessing would need to be done manually and separately otherwise.  Design matrices have the lowest computational overhead and can enable analyses involving larger numbers of predictors than the other approaches.  Both recipes and model frames allow for user-defined case weights (default: equal) and resampling strata (default: none).  The remaining approaches are fixed to have equal weights and strata defined by the response variable.  Syntax ranges from simplest to most complex for design matrices, traditional formulas, model frames, and recipes, respectively.  The relative strengths of each approach should be considered within the context of a given analysis when deciding upon which one to use.

```{r using_variables_summary, echo=FALSE}
df <- data.frame(
  "Specification" = c("Traditional Formula", "Design Matrix",
                      "Traditional Formula", "Design Matrix", "Recipe"),
  "Preprocessing" = factor(c("manual", "manual", "manual", "manual",
                             "automatic"), levels = c("manual", "automatic")),
  "In-line Functions" = factor(c("yes", "no", "yes", "no", "no"),
                               levels = c("no", "yes")),
  "Case Weights" = factor(c("equal", "equal", "user", "user", "user"),
                          levels = c("equal", "user")),
  "Resampling Strata" = factor(c("response", "response", "user", "user",
                                 "user"), levels = c("response", "user")),
  "Computational Overhead" = factor(c("medium", "low", "medium", "low", "high"),
                                    levels = c("high", "medium", "low")),
  check.names = FALSE
)

bg_colors <- c("orange", "blue", "green")
df[-1] <- lapply(df[-1], function(x) {
  bg_colors <- if (nlevels(x) == 2) bg_colors[c(1, 3)] else bg_colors
  cell_spec(x, color = "white", background = bg_colors[x])
})

kable(df, align = c("l", rep("c", ncol(df) - 1)), escape = FALSE,
      caption = "Table 2. Characteristics of available variable specification approaches.") %>%
  kable_styling(c("striped", "condensed"), full_width = FALSE,
                position = "center") %>%
  column_spec(1, bold = TRUE) %>%
  kableExtra::group_rows("Model Frame", 3, 4)
```


# Response Variable Types

The **R** class types of response variables play a central role in their analysis with the package.  They determine, for example, the specific models that can be fit, fitting algorithms employed, predicted values produced, and applicable performance metrics and analyses.  As described in the following sections, factors, ordered factors, numeric vectors and matrices, and survival objects are supported by the package.


## Factor

Categorical responses with two or more levels should be coded as a `factor` variable for analysis.  Prediction is of factor levels by default and of level-specific probabilities if `type = "prob"`.

```{r using_responses_factor}
## Iris flowers species (3-level factor)
model_fit <- fit(Species ~ ., data = iris, model = GBMModel)
predict(model_fit) %>% head
predict(model_fit, type = "prob") %>% head
```

In the case of a binary factor, the second factor level is treated as the event and the level for which predicted probabilities are computed. 

```{r using_responses_factor_binary}
## Pima Indians diabetes statuses (binary factor)
data(Pima.te, package = "MASS")
data(Pima.tr, package = "MASS")

model_fit <- fit(type ~ ., data = Pima.tr, model = GBMModel)
predict(model_fit, newdata = Pima.te) %>% head
predict(model_fit, newdata = Pima.te, type = "prob") %>% head
```


## Ordered Factor

Categorical responses can be designated as having ordered levels by storing them as an `ordered` factor variable.  For categorical vectors, this can be accomplished with the `factor()` function and its argument `ordered = TRUE` or more simply with the `ordered()` function.  Numeric vectors can be converted to ordered factors with the `cut()` function.

```{r using_responses_ordered}
## Iowa City housing prices (ordered factor)
df <- within(ICHomes, {
  sale_amount <- cut(sale_amount, breaks = 3,
                     labels = c("Low", "Medium", "High"),
                     ordered_result = TRUE)
})

model_fit <- fit(sale_amount ~ ., data = df, model = GBMModel)
predict(model_fit) %>% head
predict(model_fit, type = "prob") %>% head
```


## Numeric Vector

Code univariate numerical responses as a `numeric` variable.  Predicted numeric values are of the original type (integer or double) by default, and doubles if `type = "numeric"`.

```{r using_responses_numeric}
## Iowa City housing prices
model_fit <- fit(sale_amount ~ ., data = ICHomes, model = GBMModel)
predict(model_fit) %>% head
predict(model_fit, type = "numeric") %>% head
```


## Numeric Matrix

Store multivariate numerical responses as a numeric `matrix` variable for model fitting with traditional formulas and model frames.

```{r using_responses_matrix}
## Anscombe's multiple regression models dataset

## Numeric matrix response formula
model_fit <- fit(cbind(y1, y2, y3) ~ x1, data = anscombe, model = LMModel)
predict(model_fit) %>% head
```

For recipes, the multiple response may be defined on the left hand side of a recipe formula or as a single variable within a data frame.

```{r using_responses_matrix_recipe}
## Numeric matrix response recipe
## Defined in a recipe formula
rec <- recipe(y1 + y2 + y3 ~ x1, data = anscombe)

## Defined within a data frame
df <- within(anscombe, {
  y <- cbind(y1, y2, y3)
  remove(y1, y2, y3)
})
rec <- recipe(y ~ x1, data = df)
model_fit <- fit(rec, model = LMModel)
predict(model_fit) %>% head
```


## Survival Objects

Censored time-to-event survival responses should be stored as a `Surv` variable for model fitting with traditional formulas and model frames.

```{r using_responses_surv, results="hide"}
## Survival response formula
library(survival)

fit(Surv(time, status) ~ ., data = veteran, model = GBMModel)
```

For recipes, survival responses may be defined with the individual survival time and event variables given on the left hand side of a recipe formula and their roles designated with the `r rdoc_url("role_surv()", "recipe_roles")` function or as a single `Surv` variable within a data frame.

```{r using_responses_surv_recipe, results="hide"}
## Survival response recipe
## Defined in a recipe formula
rec <- recipe(time + status ~ ., data = veteran) %>%
  role_surv(time = time, event = status)

## Defined within a data frame
df <- within(veteran, {
  y <- Surv(time, status)
  remove(time, status)
})
rec <- recipe(y ~ ., data = df)
fit(rec, model = GBMModel)
```


# Model Performance Metrics

Performance metrics quantify associations between observed and predicted responses and provide a means of evaluating the predictive performances of models.


## Performance Function

Metrics can be computed with the `r rdoc_url("performance()")` function applied to observed responses and responses predicted with the `r rdoc_url("predict()")` function.  In the case of observed versus predicted survival probabilities or events, metrics will be calculated at user-specified survival times and returned along with their time-integrated mean.

```{r using_performance_function}
## Survival performance metrics

## Observed responses
obs <- response(surv_fit, newdata = surv_test)

## Predicted survival means
pred_means <- predict(surv_fit, newdata = surv_test)
performance(obs, pred_means)

## Predicted survival probabilities
pred_probs <- predict(surv_fit, newdata = surv_test, times = surv_times, type = "prob")
performance(obs, pred_probs)

## Predicted survival events
pred_events <- predict(surv_fit, newdata = surv_test, times = surv_times)
performance(obs, pred_events)
```

Function `r rdoc_url("performance()")` computes a default set of metrics according to the observed and predicted response types, as indicated and in the order given in the table below.


Table 3. Default performance metrics by response types.

Response                 | Default Metrics
:----------------------- | :------------------------------------------------------
Factor                   | Brier Score, Accuracy, Cohen's Kappa
Binary Factor            | Brier Score, Accuracy, Cohen's Kappa, Area Under ROC Curve, Sensitivity, Specificity
Numeric Vector or Matrix | Root Mean Squared Error, R^2^, Mean Absolute Error
Survival Means           | Concordance Index
Survival Probabilities   | Brier Score, Area Under ROC Curve, Accuracy
Survival Events          | Accuracy


These defaults may be changed by specifying one or more package-supplied metric functions to the `metrics` argument of `r rdoc_url("performance()")`.  Specification of the `metrics` argument can be in terms of a single metric function, function name, or list of metric functions.  List names, if specified, will be displayed as metric labels in graphical and tabular summaries; otherwise, the function names will be used as labels for unnamed lists.

```{r using_performance_function_metrics, eval=FALSE}
## Single metric function
performance(obs, pred_means, metrics = cindex)

## Single metric function name
performance(obs, pred_means, metrics = "cindex")

## List of metric functions
performance(obs, pred_means, metrics = c(cindex, rmse, rmsle))

## Named list of metric functions
performance(obs, pred_means,
            metrics = c("CIndex" = cindex, "RMSE" = rmse, "RMSLE" = rmsle))
```

Metrics based on classification of two-level class probabilities, like sensitivity and specificity, optionally allow for specification of the classification cutoff probability (default: `cutoff = 0.5`).

```{r using_performance_function_cutoff}
## User-specified survival probability metrics
performance(obs, pred_probs, metrics = c(sensitivity, specificity), cutoff = 0.7)
```


## Metric Functions

Whereas multiple package-supplied metrics can be calculated simultaneously with the `r rdoc_url("performance()")` function, each exists as a stand-alone function that can be called individually.

```{r using_metrics_functions}
## Metric functions for survival means
cindex(obs, pred_means)

rmse(obs, pred_means)

rmsle(obs, pred_means)

## Metric functions for survival probabilities
sensitivity(obs, pred_probs)

specificity(obs, pred_probs)
```


## Metric Information

A named list of available metrics can be obtained interactively with the `r rdoc_url("metricinfo()")` function, and includes the following components for each one.

`label`
  : character descriptor for the metric.

`maximize`
  : logical indicating whether maximization of the metric leads to better predictive performance.

`arguments`
  : closure with the argument names and corresponding default values of the metric function.

`response_types`
  : data frame of the observed and predicted response variable types supported by the metric.


Function `r rdoc_url("metricinfo()")` may be called without arguments, with one or more metric functions, an observed response variable, an observed and predicted response variable pair, response variable types, or resampled output; and will return information on all matching metrics.

```{r using_metrics_metricinfo}
## All available metrics
metricinfo() %>% names
```

Information is displayed below for the `r rdoc_url("cindex()", "metrics")` function corresponding to a concordance index, which is applicable to observed survival and predicted means.

```{r using_metrics_metricinfo_cindex}
## Metric-specific information
metricinfo(cindex)
```

Submitting the metric function at the console will result in similar information being displayed as formatted text.

```{r using_metrics_cindex}
cindex
```


### Type-Specific Metrics

When data objects are supplied as arguments to `r rdoc_url("metricinfo()")`, information is returned on all metrics applicable to response variables of the same data types.  Observed response variable type is inferred from the first data argument and predicted type from the second, if given.  For survival responses, predicted types may be `numeric` for survival means, `r rdoc_url("SurvEvents", "SurvMatrix")` for 0-1 survival events at specified follow-up times, or `r rdoc_url("SurvProbs", "SurvMatrix")` for follow-up time survival probabilities.  If model functions are additionally supplied as arguments, information on the subset matching the data types is returned.

```{r using_metrics_metricinfo_type}
## Metrics for observed and predicted response variable types
metricinfo(Surv(0)) %>% names

metricinfo(Surv(0), numeric(0)) %>% names

metricinfo(Surv(0), SurvEvents(0)) %>% names

metricinfo(Surv(0), SurvProbs(0)) %>% names

## Identify survival-specific metrics
metricinfo(Surv(0), auc, cross_entropy, cindex) %>% names
```


### Response Variable-Specific Metrics

Existing response variables observed and those obtained from the `r rdoc_url("predict()")` function may be given as arguments to identify metrics that are applicable to them.

```{r using_metrics_metricinfo_response}
## Metrics for observed and predicted responses from model fits
metricinfo(obs, pred_means) %>% names

metricinfo(obs, pred_probs) %>% names
```


## Factors

Metrics applicable to multi-level factor response variables are summarized below.

`r rdoc_url("accuracy()", "metrics")`
  : proportion of correctly classified responses.

`r rdoc_url("brier()", "metrics")`
  : [Brier score](https://en.wikipedia.org/wiki/Brier_score).

`r rdoc_url("cross_entropy()", "metrics")`
  : [cross entropy](https://en.wikipedia.org/wiki/Cross_entropy) loss averaged over the number of cases.

`r rdoc_url("kappa2()", "metrics")`
  : [Cohen's kappa](https://en.wikipedia.org/wiki/Cohen%27s_kappa) statistic measuring relative agreement between observed and predicted classifications.

`r rdoc_url("weighted_kappa2()", "metrics")`
  : [weighted Cohen's kappa](https://en.wikipedia.org/wiki/Cohen%27s_kappa#Weighted_kappa) for ordered factor responses only.

Brier score and cross entropy loss are computed directly on predicted class probabilities.  The other metrics are computed on predicted class membership, defined as the factor level with the highest predicted probability.


## Binary Factors

Metrics for binary factors include those given for multi-level factors as well as the following.

`r rdoc_url("auc()", "metrics")`
  : area under a performance curve.

`r rdoc_url("cindex()", "metrics")`
  : concordance index computed as rank order agreement between predicted probabilities for paired event and non-event cases.  This metric can be interpreted as the probability that a randomly selected event case will have a higher predicted value than a randomly selected non-event case, and is the same as area under the ROC curve.

`r rdoc_url("f_score()", "metrics")`
  : [F score](https://en.wikipedia.org/wiki/Precision_and_recall#F-measure), $F_\beta = (1 + \beta^2) \frac{\text{precision} \times \text{recall}}{\beta^2 \times \text{precision} + \text{recall}}$.  F1 score $(\beta = 1)$ is the package default.

`r rdoc_url("fnr()", "metrics")`
  : false negative rate, $FNR = \frac{FN}{TP + FN} = 1 - TPR$.

```{r using_metrics_conf, echo=FALSE}
conf <- matrix(c("True Negative (TN)", "False Positive (FP)",
                 "False Negative (FN)", "True Positive (TP)"),
               2, 2,
               dimnames = list("Predicted Response" = c("Negative", "Positive"),
                               "Observed Response" = c("Negative", "Positive")))
kable(conf,
      caption = "Table 4. Confusion matrix of observed and predicted response classifications.",
      align = c("c", "c")) %>%
  kable_styling(full_width = FALSE, position = "center") %>%
  add_header_above(c("Predicted Response" = 1, "Observed Response" = 2))
```

`r rdoc_url("fpr()", "metrics")`
  : false positive rate, $FPR = \frac{FP}{TN + FP} = 1 - TNR$.

`r rdoc_url("npv()", "metrics")`
  : negative predictive value, $NPV = \frac{TN}{TN + FN}$.

`r rdoc_url("ppr()", "metrics")`
  : positive prediction rate, $PPR = \frac{TP + FP}{TP + FP + TN + FN}$.

`r rdoc_url("ppv()", "metrics")`, `r rdoc_url("precision()", "metrics")`
  : positive predictive value, $PPV = \frac{TP}{TP + FP}$.

`r rdoc_url("pr_auc()", "metrics")`, `r rdoc_url("auc()", "metrics")`
  : area under a precision recall curve.

`r rdoc_url("roc_auc()", "metrics")`, `r rdoc_url("auc()", "metrics")`
  : [area under an ROC curve](https://en.wikipedia.org/wiki/Receiver_operating_characteristic#Area_under_the_curve).

`r rdoc_url("roc_index()", "metrics")`
  : a tradeoff function of sensitivity and specificity as defined by the `fun` argument in this function (default: (sensitivity + specificity) / 2).  The function allows for specification of tradeoffs [@perkins:2006:IOC] other than the default of Youden's J statistic [@youden:1950:IRD].

`r rdoc_url("sensitivity()", "metrics")`, `r rdoc_url("recall()", "metrics")`, `r rdoc_url("tpr()", "metrics")`
  : true positive rate, $TPR =\frac{TP}{TP + FN} = 1 - FNR$.

`r rdoc_url("specificity()", "metrics")`, `r rdoc_url("tnr()", "metrics")`
  : true negative rate, $TNR = \frac{TN}{TN + FP} = 1 - FPR$.

Area under the ROC and precision-recall curves as well as the concordance index are computed directly on predicted class probabilities.  The other metrics are computed on predicted class membership.  Memberships are defined to be in the second factor level if predicted probabilities are greater than the function default or user-specified cutoff value.


## Numerics

Performance metrics are defined below for numeric vector responses.  If applied to a numeric matrix response, the metrics are computed separately for each column and then averaged to produce a single value.

`r rdoc_url("gini()", "metrics")`
  : [Gini coefficient](https://en.wikipedia.org/wiki/Gini_coefficient).

`r rdoc_url("mae()", "metrics")`
  : mean absolute error, $MAE = \frac{1}{N}\sum_{i=1}^N|y_i - \hat{y}_i|$, where $y_i$ and $\hat{y}_i$ are the $N$ observed and predicted responses.

`r rdoc_url("mse()", "metrics")`
  : mean squared error, $MSE = \frac{1}{N}\sum_{i=1}^N(y_i - \hat{y}_i)^2$.

`r rdoc_url("msle()", "metrics")`
  : mean squared log error, $MSLE = \frac{1}{N}\sum_{i=1}^N(log(1 + y_i) - log(1 + \hat{y}_i))^2$.

`r rdoc_url("r2()", "metrics")`
  : one minus residual divided by total sums of squares, $R^2 = 1 - \sum_{i=1}^N(y_i - \hat{y}_i)^2 / \sum_{i=1}^N(y_i - \bar{y})^2$.

`r rdoc_url("rmse()", "metrics")`
  : square root of mean squared error.

`r rdoc_url("rmsle()", "metrics")`
  : square root of mean squared log error.


## Survival Objects

All previously described metrics for binary factor responses---plus accuracy, Brier score and Cohen's kappa---are applicable to survival probabilities predicted at specified follow-up times.  Metrics are evaluated separately at each follow-up time and reported along with a time-integrated mean.  The survival concordance index is computed with the method of Harrell [-@harrell:1982:EYM] and Brier score according to Graf et al. [-@graf:1999:ACP]; whereas, the others are computed according to the confusion matrix probabilities below, in which term $\hat{S}(t)$ is the predicted survival probability at follow-up time $t$ and $T$ is the survival time [@heagerty:2004:TDR].

```{r using_metrics_conf_surv, echo=FALSE}
conf <- matrix(
  c(
    "$TN = \\Pr(\\hat{S}(t) \\ge \\text{cutoff} \\cap T \\gt t)$",
    "$FP = \\Pr(\\hat{S}(t) \\lt \\text{cutoff} \\cap T \\gt t)$",
    "$FN = \\Pr(\\hat{S}(t) \\ge \\text{cutoff} \\cap T \\le t)$",
    "$TP = \\Pr(\\hat{S}(t) \\lt \\text{cutoff} \\cap T \\le t)$"
  ),
  2, 2,
  dimnames = list(
    "Predicted Response" = c("Non-Event", "Event"),
    "Observed Response" = c("Non-Event", "Event")
  )
)
kable(conf,
      caption = "Table 5. Confusion matrix of observed and predicted survival response classifications.",
      align = c("c", "c"),
      escape = FALSE) %>%
  kable_styling(full_width = FALSE, position = "center") %>%
  add_header_above(c("Predicted Response" = 1, "Observed Response" = 2))
```

In addition, all of the metrics described for numeric vector responses are applicable to predicted survival means and are computed using only those cases with observed (non-censored) events.


# Resampled Performance


## Algorithms

Model performance can be estimated with resampling methods that simulate repeated training and test set fits and predictions.  With these methods, performance metrics are computed on each resample to produce an empirical distribution for inference.  Resampling is controlled in the **MachineShop** with the functions:

`r rdoc_url("BootControl()", "MLControl")`
  : simple bootstrap resampling [@efron:1993:IB] to repeatedly fit a model to bootstrap resampled training sets and predict the full dataset.

`r rdoc_url("BootOptimismControl()", "MLControl")`
  : optimism-corrected bootstrap resampling [@efron:1983:LLB; @harrell:1996:MPM].

`r rdoc_url("CVControl()", "MLControl")`
  : repeated K-fold cross-validation [@kohavi:1995:SCB] to repeatedly partition the full dataset into K folds.  For a given partitioning, prediction is performed on each of the K folds with models fit on all remaining folds.  The package default is 10-fold cross-validation.

`r rdoc_url("CVOptimismControl()", "MLControl")`
  :  optimism-corrected cross-validation [@davison:1997:BMA, eq. 6.48].

`r rdoc_url("OOBControl()", "MLControl")`
  : out-of-bootstrap resampling to repeatedly fit a model to bootstrap resampled training sets and predict the unsampled cases.

`r rdoc_url("SplitControl()", "MLControl")`
  : split training and test sets [@hastie:2009:ESL7] to randomly partition the full dataset for model fitting and prediction, respectively.

`r rdoc_url("TrainControl()", "MLControl")`
  : training resubstitution for both model fitting and prediction on the full dataset in order to estimate training, or apparent, error [@efron:1986:HBA].

For the survival example, repeated cross-validation control structures are defined for the estimation of model performance in predicting survival means and 5 and 10-year survival probabilities.  In addition to arguments controlling the resampling methods, a `seed` can be set to ensure reproducibility of resampling results obtained with the structures.  The default control structure can also be set globally by users and subsequently changed back to its package default as desired.

```{r using_resample_control}
## Control structures for K-fold cross-validation

## Prediction of survival means
surv_means_control <- CVControl(folds = 5, repeats = 3, seed = 123)

## Prediction of survival probabilities
surv_probs_control <- CVControl(folds = 5, repeats = 3, seed = 123) %>%
  set_predict(times = surv_times)

## User-specification of the default control structure
MachineShop::settings(control = CVControl(folds = 5, seed = 123))

## Package default
# MachineShop::settings(reset = "control")
```


## Parallel Processing

Resampling and permutation-based variable importance are implemented with the **foreach** package [@microsoft:2019:FPF] and will run in parallel if a compatible backend is loaded, such as that provided by the **doParallel** [@microsoft:2019:DFP] or **doSNOW** package [@microsoft:2019:DFS].

```{r using_resample_parallel}
## Register multiple cores for parallel computations
library(doParallel)
registerDoParallel(cores = 2)
```


## Resample Function

Resampling is performed by calling the `r rdoc_url("resample()")` function with a variable specification, model, and control structure.  Like the `r rdoc_url("fit()")` function, variables may be specified in terms of a traditional formula, design matrix, model frame, or recipe.

```{r using_resample_function}
## Resample estimation for survival means and probabilities
(res_means <- resample(surv_fo, data = surv_train, model = GBMModel,
                       control = surv_means_control))

(res_probs <- resample(surv_fo, data = surv_train, model = GBMModel,
                       control = surv_probs_control))
```


## Summary Statistics

The `r rdoc_url("summary()")` function when applied directly to output from `r rdoc_url("resample()")` computes summary statistics for the default performance metrics described in the *Performance Function* section.

```{r using_resample_summary}
## Summary of survival means metric
summary(res_means)

## Summary of survival probability metrics
summary(res_probs)
```

Other relevant metrics can be identified with `r rdoc_url("metricinfo()")` and summarized with `r rdoc_url("performance()")`.

```{r using_resample_summary_performance}
## Resample-specific metrics
metricinfo(res_means) %>% names

## User-specified survival means metrics
summary(performance(res_means, metrics = c(cindex, rmse)))
```

Furthermore, summaries can be customized with a user-defined statistics function or list of statistics functions passed to the `stats` argument of `r rdoc_url("summary()")`.

```{r using_resample_summary_stats}
## User-defined statistics function
percentiles <- function(x) quantile(x, probs = c(0.25, 0.50, 0.75))
summary(res_means, stats = percentiles)

## User-defined list of statistics functions
summary(res_means, stats = c(Mean = mean, Percentile = percentiles))
```


## Plots

Summary plots of resample output can be obtained with the `r rdoc_url("plot()")` function.  Boxplots are the default plot type; but density, errorbar, and violin plots are also available.  Plots are generated with the **ggplot2** package [@wickham:2016:GEG] and returned as `ggplot` objects.  As such, annotation and formatting defined for ggplots can be applied to the returned plots.

```{r using_resample_plots}
## Libraries for plot annotation and fomatting
library(ggplot2)
library(gridExtra)

## Individual ggplots
p1 <- plot(res_means)
p2 <- plot(res_means, type = "density")
p3 <- plot(res_means, type = "errorbar")
p4 <- plot(res_means, type = "violin")

## Grid of plots
grid.arrange(p1, p2, p3, p4, nrow = 2)
```


## Stratified Resampling

Stratification of cases for the construction of resampled training and test sets can be employed to help achieve balance across the sets.  Stratified resampling is automatically performed if variable specification is in terms of a traditional formula or design matrix and will be done according to the response variable.  For model frames and recipes, stratification variables must be defined explicitly with the `strata` argument to the `r rdoc_url("ModelFrame()")` constructor or with the `r rdoc_url("role_case()", "recipe_roles")` function.   In general, strata are constructed from numeric proportions for `BinomialVariate`; original values for `character`, `factor`, `logical`, and `ordered`; first columns of values for `matrix`; original values for `numeric`; and numeric times within event statuses for `Surv`.  Numeric values are stratified into quantile bins and categorical values into factor levels defined by the resampling control functions.  Missing values are replaced with non-missing values sampled at random with replacement.

```{r using_resample_strata, results="hide"}
## Model frame with response variable stratification
mf <- ModelFrame(surv_fo, data = surv_train, strata = surv_train$y)
resample(mf, model = GBMModel)

## Recipe with response variable stratification
rec <- recipe(y ~ ., data = surv_train) %>%
  role_case(stratum = y)
resample(rec, model = GBMModel)
```


## Dynamic Model Parameters

As discussed previously in the *Model Fit and Prediction* section, dynamic model parameters are evaluated at the time of model fitting and can depend on the number of observations in the fitted dataset.  In the context of resampling, dynamic parameters are repeatedly evaluated at each fit of the resampled datasets.  As such, their values can change based on the observations selected for training at each iteration of the resampling algorithm.

```{r using_resample_dynamic, results="hide"}
## Dynamic model parameter k = log number of training set observations
resample(surv_fo, data = surv_train, model = CoxStepAICModel(k = .(log(nobs))))
```


## Model Comparisons

Resampled metrics from different models can be combined for comparison with the `r rdoc_url("c()", "combine")` function.  Optional names given on the left hand side of equal operators within `r rdoc_url("c()", "combine")` calls will be used as labels in output from the `r rdoc_url("summary()")` and `r rdoc_url("plot()")` functions.  For comparisons of resampled output, the same control structure must be used in all associated calls to `r rdoc_url("resample()")` to ensure that resulting model metrics are computed on the same resampled training and test sets.  The combined resample output can be summarized and plotted as usual.

```{r using_resample_comparisons}
## Resample estimation
res1 <- resample(surv_fo, data = surv_train, model = GBMModel(n.trees = 25),
                 control = surv_means_control)
res2 <- resample(surv_fo, data = surv_train, model = GBMModel(n.trees = 50),
                 control = surv_means_control)
res3 <- resample(surv_fo, data = surv_train, model = GBMModel(n.trees = 100),
                 control = surv_means_control)

## Combine resample output for comparison
(res <- c(GBM1 = res1, GBM2 = res2, GBM3 = res3))

summary(res)

plot(res)
```

Pairwise model differences for each metric can be calculated with the `r rdoc_url("diff()")` function applied to results from a call to `r rdoc_url("c()", "combine")`.  Resulting differences can be summarized descriptively with the `r rdoc_url("summary()")` and `r rdoc_url("plot()")` functions and assessed for statistical significance with pairwise t-tests performed by the `r rdoc_url("t.test()")` function.  The t-test statistic for a given set of $R$ resampled differences is calculated as
$$
t = \frac{\bar{x}_R}{\sqrt{F s^2_R / R}},
$$
where $\bar{x}_R$ and $s^2_R$ are the sample mean and variance.  Statistical testing for a mean difference is then performed by comparing $t$ to a $t_{R-1}$ null distribution.  The sample variance in the t statistic is known to underestimate the true variances of cross-validation mean estimators.  Underestimation of these variances will lead to increased probabilities of false-positive statistical conclusions.  Thus, an additional factor $F$ is included in the t statistic to allow for variance corrections.  A correction of $F = 1 + K / (K - 1)$ was found by Nadeau and Bengio [-@nadeau:2003:IGE] to be a good choice for cross-validation with $K$ folds and is thus used for that resampling method.  The extension of this correction by Bouchaert and Frank [-@bouckaert:2004:ERS] to $F = 1 + T K / (K - 1)$ is used for cross-validation with $K$ folds repeated $T$ times.  For other resampling methods $F = 1$.  Below are t-test results based on the extended correction factor for 3 repeats of 5-fold cross-validation.

```{r using_resample_diff}
## Pairwise model comparisons
(res_diff <- diff(res))

summary(res_diff)

plot(res_diff)
```

```{r using_resample_diff_test}
t.test(res_diff)
```


# Model Predictor Effects and Diagnostics

Calculation of performance metrics on test sets or by resampling, as discussed previously, is one method of assessing model performance.  Others available include measures of predictor variable importance, partial dependence plots, calibration curves comparing observed and predicted response values, and receiver operating characteristic analysis.


## Variable Importance

The importance of predictor variables in a model fit is estimated with the `r rdoc_url("varimp()")` function and displayed graphically with `r rdoc_url("plot()")`.  Variable importance is a relative measure of the contributions of model predictors and is scaled by default to have a maximum value of 100, where higher values represent more important variables.  Model-specific implementations of variable importance are available in many cases, although their definitions may differ.  In the case of a `r rdoc_url("GBMModel")`, importance of each predictor is based on the sum of squared empirical improvements over all internal tree nodes created by splitting on that variable [@greenwell:2019:GBM].

```{r using_analyses_vi}
## Predictor variable importance
(vi <- varimp(surv_fit, method = "model"))

plot(vi)
```

In contrast, importance is based on negative log-transformed p-values for statistical models, like `r rdoc_url("CoxModel")`, that produce them.  For other models, variable importance may be defined and calculated by their underlying source packages or not defined at all, as is the case for `r rdoc_url("SVMModel")`.  Logical indicators of model-specific variable importance are given in the information displayed by model constructors and returned by `r rdoc_url("modelinfo()")`.

```{r using_analysis_vi_info}
SVMModel

modelinfo(SVMModel)[[1]]$varimp
```

Variable importance can be computed with model-agnostic permutation methods [@fisher:2019:AMW] as an alternative to model-specific methods.  The following algorithm for permutation-based variable importance is implemented and the default method in **MachineShop**.

* Fit a model to a training dataset.

* Compute training performance.

* For $s = 1, \ldots, S$.

  * Optionally sample a subset of the training set without replacement.
    
  * For predictor variable $p = 1, \ldots, P$.
  
    * Randomly permute the variable values.
    
    * Compute performance on the permutation set.
    
    * Compute importance as the difference or ratio between the permutation and training performances.
    
    * Reset the variable to its original values.

* Return the mean or other summary statistics of importance for each variable.

```{r using_analysis_vi_permute}
## Permutation-based variable importance
varimp(surv_fit)
```

There are a number of advantages to permutation-based variable importance.  In particular, it can be computed for any

1. model,
2. performance metric defined for the given response variable type, and
3. predictor variable in the original training set.

Conversely, model-specific methods are not defined for some models, are generally limited to metrics implemented in their source packages, and might be computed on derived, rather than original, predictor variables.  These differences can make comparisons of variable importance across classes of models difficult if not impossible.  The trade-off for the advantages of a permutation-based approach is increased computation time.  To speed up computations, the algorithm will run in parallel if a compatible backend is loaded as described in the *Parallel Processing* section. 


## Recursive Feature Elimination

Recursive feature elimination (RFE) is a wrapper method of variable selection.  In wrapper methods, a given model is fit to subsets of predictor variables in order to select the subset whose fit is optimal.  Forward, backward, and step-wise variable selection are examples of wrapper methods.  RFE is a type of backward selection in which subsets are formed from decreasing numbers of the most important predictor variables.  The RFE algorithm implemented in the package-supplied `rfe()` function is summarized below.

* Compute variable importance for all predictors.

* For predictor subsets of sizes $S = S_n > ... > S_1$.

  * Eliminate predictors whose variable importance is not in the top $S$ by randomly permuting their values.
  
  * Compute a resampled estimate of model predictive performance.
  
  * Optionally recompute variable importance for the top $S$ predictors, and set importance equal to zero for those eliminated.
  
* Select the predictor set with highest predictive performance.

This RFE algorithm differs from others in that variables are "eliminated" by permuting their values rather than by removing them from the dataset.  Using a permutation approach for both the elimination of variables and computation of variable importance enables application of the `rfe()` function to any variable specification (traditional formula, design matrix, model frame, or recipe) and any model available in the package.  The syntax for `rfe()` is similar to `resample()` as illustrated in the following example.

```{r using_analysis_rfe}
## Recursive feature elimination
(surv_rfe <- rfe(surv_fo, data = surv_train, model = GBMModel,
                 control = surv_means_control))
rfe_summary <- summary(surv_rfe)
rfe_summary$terms[rfe_summary$selected]
```


## Partial Dependence Plots

Partial dependence plots show the marginal effects of predictors on a response variable.  Dependence for a select set of one or more predictor variables $X_S$ is computed as
$$
\bar{f}_S(X_S) = \frac{1}{N}\sum_{i=1}^N f(X_S, x_{iS'}),
$$
where $f$ is a fitted prediction function and $x_{iS'}$ are values of the remaining predictors in a dataset of $N$ cases.  The response scale displayed in dependence plots will depend on the response variable type: probability for predicted factors and survival probabilities, original scale for numerics, and survival time for predicted survival means.  By default, dependence is computed for each selected predictor individually over a grid of 10 approximately evenly spaced values and averaged over the dataset on which the prediction function was fit.

```{r using_analyses_pd, results = "hide"}
## Partial dependence plots
pd <- dependence(surv_fit, select = c(thickness, age))
plot(pd)
```

Estimated predictor effects are marginal in that they are averaged over the remaining variables, whose distribution depends on the population represented by the dataset.  Consequently, partial dependence plots for a given model can vary across datasets and populations.  The package allows averaging over different datasets to estimate marginal effects in other case populations, over different numbers of predictor values, and over quantile spacing of the values.

```{r using_analyses_pd_data, results = "hide"}
pd <- dependence(surv_fit, data = surv_test, select = thickness, n = 20,
                 intervals = "quantile")
plot(pd)
```

In addition, dependence may be computed for combinations of multiple predictors to examine interaction effects and for summary statistics other than the mean.


## Calibration Curves

Agreement between model-predicted and observed values can be visualized with calibration curves.  Calibration curves supplement individual performance metrics with information on model fit in different regions of predicted values.  They also provide more direct assessment of agreement than some performance metrics, like ROC AUC, that do not account for scale and location differences.  In the construction of binned calibration curves, cases are partitioned into equal-width intervals according to their (resampled) predicted responses.  Mean observed responses are then calculated within each of the bins and plotted on the vertical axis against the bin midpoints on the horizontal axis. 

```{r using_analyses_cal, results="hide"}
## Binned calibration curves
cal <- calibration(res_probs, breaks = 10)
plot(cal, se = TRUE)
```

As an alternative to discrete bins, curves can be smoothed by setting `breaks = NULL` to compute weighted averages of observed values.  Smoothing has the advantage of producing more precise curves by including more observed values in the calculation at each predicted value.

```{r using_analyses_cal_smoothed, results="hide"}
## Smoothed calibration curves
cal <- calibration(res_probs, breaks = NULL)
plot(cal)
```

Calibration curves close to the 45$^\circ$ line represent agreement between observed and predicted responses and a model that is said to be well calibrated.


## Confusion Matrices

Confusion matrices of cross-classified observed and predicted categorical responses are available with the `r rdoc_url("confusion()")` function.  They can be constructed with predicted class membership or with predicted class probabilities.  In the latter case, predicted class membership is derived from predicted probabilities according to a probability cutoff value for binary factors (default: `cutoff = 0.5`) and according to the class with highest probability for factors with more than two levels.


```{r using_analyses_conf}
## Confusion matrices
(conf <- confusion(res_probs, cutoff = 0.7))
```

```{r using_analyses_conf_plot, results="hide"}
plot(conf)
```

Confusion matrices are the data structure upon which many of the performance metrics described earlier for factor predictor variables are based.  Metrics commonly reported for confusion matrices are generated by the `r rdoc_url("summary()")` function.

```{r using_analyses_conf_summary}
## Summary performance metrics
summary(conf)
```

Summaries can also be obtained with the `r rdoc_url("performance()")` function for default or use-specified metrics.

```{r using_analyses_conf_performance}
## Confusion matrix-specific metrics
metricinfo(conf) %>% names

## User-specified metrics
performance(conf, metrics = c("Accuracy" = accuracy,
                              "Sensitivity" = sensitivity,
                              "Specificity" = specificity))
```


## Performance Curves

Tradeoffs between correct and incorrect classifications of binary responses, across the range of possible cutoff probabilities, can be studied with performance curves.  In general, any two binary response metrics may be specified for the construction of a performance curve.


### ROC Curves

Receiver operating characteristic (ROC) curves are one example in which true positive rates (sensitivity) are plotted against false positive rates (1 - specificity) [@fawcett:2006:IRA].  True positive rate (TPR) and false positive rate (FPR) are defined as
$$
\begin{aligned}
 TPR &= \text{sensitivity} = \Pr(\hat{p} > c \mid D^+) \\
 FPR &= 1 - \text{specificity} = \Pr(\hat{p} > c \mid D^-),
\end{aligned}
$$
where $\hat{p}$ is the model-predicted probability of being positive, $0 \le c \le 1$ is a probability cutoff value for classification as positive or negative, and $D^+/D^-$ is positive/negative case status.  ROC curves show tradeoffs between the two rates over the range of possible cutoff values.  Higher curves are indicative of better predictive performance.

```{r using_analyses_roc}
## ROC curves
roc <- performance_curve(res_probs)
plot(roc, diagonal = TRUE)
```

ROC curves show the relation between the two rates being plotted but not their relationships with specific cutoff values.  The latter may be helpful for the selection of a cutoff to apply in practice.  Accordingly, separate plots of each rate versus the range of possible cutoffs are available with the `type = "cutoffs"` option.

```{r using_analyses_roc_cutoffs}
plot(roc, type = "cutoffs")
```

Area under the ROC curve (ROC AUC) is an overall measure of model predictive performance.  It is interpreted as the probability that a randomly selected positive case will have a higher predicted value than a randomly selected negative case.  AUC values of 0.5 and 1.0 indicate chance and perfect *concordance* between predicted probabilities and observed responses.

```{r using_analyses_roc_auc}
auc(roc)
```


### Precision Recall Curves

Precision recall curves plot precision (positive predictive value) against recall (sensitivity) [@davis:2006:RPR], where
$$
\begin{aligned}
  \text{precision} &= PPV = \Pr(D^+ \mid \hat{p} > c) \\
  \text{recall} &= \text{sensitivity} = \Pr(\hat{p} > c \mid D^+).
\end{aligned}
$$
These curves tend to be used when primary interest lies in detecting positive cases and such cases are rare.

```{r using_analyses_pr}
## Precision recall curves
pr <- performance_curve(res_probs, metrics = c(precision, recall))
plot(pr)
```

```{r using_analyses_pr_auc}
auc(pr)
```


### Lift Curves

Lift curves depict the rate at which positive cases are found as a function of the proportion predicted to be positive in the population.  In particular, they plot true positive rate (sensitivity) against positive prediction rate (PPR) for all possible classification probability cutoffs, where
$$
\begin{aligned}
  TPR &= \Pr(\hat{p} > c \mid D^+) \\
  PPR &= \Pr(\hat{p} > c).
\end{aligned}
$$
Models more efficient (lower cost) at identifying positive cases find them at a higher proportion ($TPR$) while predicting fewer in the overall population to be positive ($PPR$).  In other words, higher lift curves are signs of model efficiency.

```{r using_analyses_lift}
## Lift curves
lf <- lift(res_probs)
plot(lf, find = 0.75)
```


# Modeling Strategies

Model development often involves the comparison of multiple models from a candidate set for the purpose of selecting a final one.  Models in the set may differ with respect to their predictor variables, preprocessing steps and parameters, and model types and parameters.  Complex model selection strategies for sets that involve one or more of these differences can be implemented with the **MachineShop** package.  Implementation is achieved with a straightforward syntax based on the meta-input and meta-model functions listed in the table below and with resampling, including nested resampling, conducted automatically for model selection and predictive performance evaluation.

| Parameter Grid Tuning        | Candidate Set Selection         | Ensemble Learning
|:-----------------------------|:--------------------------------|:------------------------------
| `r rdoc_url("TunedInput()")` | `r rdoc_url("SelectedInput()")` | `r rdoc_url("StackedModel()")`
| `r rdoc_url("TunedModel()")` | `r rdoc_url("SelectedModel()")` | `r rdoc_url("SuperModel()")`

These meta-functions fall into three main categories: 1) tuning of a given input or model over a grid of parameter values, 2) selection from an arbitrary set of different inputs or models, or 3) combining multiple models into an ensemble learner.  In the context of these strategies, an input may be a formula, design matrix, model frame, or preprocessing recipe.  The meta-input and meta-model functions themselves return input and model class objects, respectively.  Combinations and multiple levels of nesting of meta-functions, inputs, and models are allowed.  For example, `r rdoc_url("StackedModel()")` and `r rdoc_url("SuperModel()")` may consist of `r rdoc_url("TunedModel")` and other model objects.  `r rdoc_url("SelectedModel()")` can select among mixes of `r rdoc_url("TunedModel")`, ensemble model, and other model objects.  Likewise, `r rdoc_url("TunedInput")` objects, along with other inputs, may be nested within `r rdoc_url("SelectedInput()")`.  Furthermore, selection and tuning of both inputs and models can be performed simultaneously.  These and other possibilities are illustrated in the following sections.


## Inputs

Inputs to model fitting functions define the predictor and response variables and the dataset containing their values.  These can be specified with traditional formula and dataset pairs, design matrix and response variable pairs, model frames, and preprocessing recipes.  The package supports (1) tuning of an input over a grid of parameter values and (2) selection of inputs from candidate sets that differ with respect to their predictors or their preprocessing steps and parameters.


### Input Tuning

Preprocessing recipes may have step with parameters that affect predictive performance.  Steps can be tuned over a grid of parameter values with `r rdoc_url("TunedInput()")` to select the best performing values.  Calls to `r rdoc_url("TunedInput()")` return an input object that may be trained on data with the `r rdoc_url("fit()")` function or evaluated for predictive performance with `r rdoc_url("resample()")`.  As an example, a principal components analysis (PCA) step could be included in a preprocessing recipe for tuning over the number of components to retain in the final model.  Such a recipe is shown below accompanied by a call to `r rdoc_url("expand_steps()")` to construct a tuning grid.  The grid parameter `num_comp` and name `PCA` correspond to the argument and id of the `r rdoc_url("step_pca()")` function to which the values `1:3` apply.  The recipe and grid may then be passed to `r rdoc_url("TunedInput()")` for model fitting.

```{r using_stategies_TunedInput1, eval=FALSE}
## Preprocessing recipe with PCA steps
pca_rec <- recipe(y ~ ., data = surv_train) %>%
  role_case(stratum = y) %>%
  step_center(all_predictors()) %>%
  step_scale(all_predictors()) %>%
  step_pca(all_predictors(), id = "PCA")

## Tuning grid of number of PCA components
pca_grid <- expand_steps(
  PCA = list(num_comp = 1:3)
)

## Tuning specification
tun_rec <- TunedInput(pca_rec, grid = pca_grid)
```

From the fit, the resulting model can be extracted with `r rdoc_url("as.MLModel()")`.  The output shows that one principal component was selected.  Resample estimation of predictive performance is applied to a `r rdoc_url("TunedInput")` specification for the selection.  The default resampling method is cross-validation.  Other methods, performance metrics, and selection statistics can be supplied to the `r rdoc_url("TunedInput()")` arguments.

```{r using_stategies_TunedInput2, eval=FALSE}
## Input-tuned model fit and final trained model
model_fit <- fit(tun_rec, model = GBMModel)
as.MLModel(model_fit)
#> --- MLModel object ----------------------------------------------------------
#> 
#> Model name: GBMModel
#> Label: Trained Generalized Boosted Regression
#> Package: gbm
#> Response types: factor, numeric, PoissonVariate, Surv
#> Case weights support: TRUE
#> Missing case removal: response
#> Tuning grid: TRUE
#> Variable importance: TRUE
#> 
#> Parameters:
#> List of 5
#>  $ n.trees          : num 100
#>  $ interaction.depth: num 1
#>  $ n.minobsinnode   : num 10
#>  $ shrinkage        : num 0.1
#>  $ bag.fraction     : num 0.5
#> 
#> === $TrainingStep1 ==========================================================
#> === TrainingStep object ===
#> 
#> Optimization method: Grid Search
#> TunedModelRecipe log:
#> # A tibble: 3 × 4
#>   name          selected params$PCA$num_comp metrics$`C-Index`
#>   <chr>         <lgl>                  <int>             <dbl>
#> 1 ModelRecipe.1 TRUE                       1             0.740
#> 2 ModelRecipe.2 FALSE                      2             0.720
#> 3 ModelRecipe.3 FALSE                      3             0.725
#> 
#> Selected row: 1
#> Metric: C-Index = 0.7399641
```


### Input Selection

Selection of recipes with different steps or predictors can be conducted with `r rdoc_url("SelectedInput()")`.

```{r using_strategies_SelectedInput1, eval=FALSE}
## Preprocessing recipe without PCA steps
rec1 <- recipe(y ~ sex + age + year + thickness + ulcer, data = surv_train) %>%
  role_case(stratum = y)
rec2 <- recipe(y ~ sex + age + year, data = surv_train) %>%
  role_case(stratum = y)

## Selection among recipes with and without PCA steps
sel_rec <- SelectedInput(
  rec1,
  rec2,
  TunedInput(pca_rec, grid = pca_grid)
)
```

In this case, the third recipe with PCA steps is selected.

```{r using_strategies_SelectedInput2, eval=FALSE}
## Input-selected model fit and model
model_fit <- fit(sel_rec, model = GBMModel)
as.MLModel(model_fit)
#> --- MLModel object ----------------------------------------------------------
#> 
#> Model name: GBMModel
#> Label: Trained Generalized Boosted Regression
#> Package: gbm
#> Response types: factor, numeric, PoissonVariate, Surv
#> Case weights support: TRUE
#> Missing case removal: response
#> Tuning grid: TRUE
#> Variable importance: TRUE
#> 
#> Parameters:
#> List of 5
#>  $ n.trees          : num 100
#>  $ interaction.depth: num 1
#>  $ n.minobsinnode   : num 10
#>  $ shrinkage        : num 0.1
#>  $ bag.fraction     : num 0.5
#> 
#> === $TrainingStep1 ==========================================================
#> === TrainingStep object ===
#> 
#> Optimization method: Grid Search
#> SelectedModelRecipe log:
#> # A tibble: 3 × 4
#>   name             selected params$id  metrics$`C-Index`
#>   <chr>            <lgl>    <chr>                  <dbl>
#> 1 ModelRecipe.1    FALSE    input.azpi             0.761
#> 2 ModelRecipe.2    FALSE    input.aLHo             0.643
#> 3 TunedModelRecipe TRUE     input.W4KN             0.796
#> 
#> Selected row: 3
#> Metric: C-Index = 0.7960841
#> 
#> === $TrainingStep2 ==========================================================
#> === TrainingStep object ===
#> 
#> Optimization method: Grid Search
#> TunedModelRecipe log:
#> # A tibble: 3 × 4
#>   name          selected params$PCA$num_comp metrics$`C-Index`
#>   <chr>         <lgl>                  <int>             <dbl>
#> 1 ModelRecipe.1 TRUE                       1             0.740
#> 2 ModelRecipe.2 FALSE                      2             0.720
#> 3 ModelRecipe.3 FALSE                      3             0.725
#> 
#> Selected row: 1
#> Metric: C-Index = 0.7399641
```

Selection can also be performed among traditional formulas, design matrices, or model frames.

```{r using_strategies_SelectedInput3, eval=FALSE}
## Traditional formulas
fo1 <- y ~ sex + age + year + thickness + ulcer
fo2 <- y ~ sex + age + year

## Selection among formulas
sel_fo <- SelectedInput(fo1, fo2, data = surv_train)

## Input-selected model fit and final trained model
model_fit <- fit(sel_fo, model = GBMModel)
as.MLModel(model_fit)
```

In the previous examples, selection of different inputs was performed with the same model (`r rdoc_url("GBMModel")`).  Selection among different combinations of inputs and models is supported with the `r rdoc_url("ModelSpecification()")` constructor.

```{r using_strategies_SelectedInput4, eval=FALSE}
## Different combinations of inputs and models
sel_mfo <- SelectedInput(
  ModelSpecification(fo1, data = surv_train, model = CoxModel),
  ModelSpecification(fo2, data = surv_train, model = GBMModel)
)

## Input-selected model fit and final trained model
model_fit <- fit(sel_mfo)
as.MLModel(model_fit)
```


## Models

Models define the functional relationships between predictor and response variables from a given set of inputs.


### Model Tuning

Many of the package-supplied modeling functions have arguments, or tuning parameters, that control aspects of their model fitting algorithms.  For example, `r rdoc_url("GBMModel")` parameters `n.trees` and `interaction.depth` control the number of decision trees to fit and the maximum tree depths.  When called with a `r rdoc_url("TunedModel")`, the `r rdoc_url("fit()")` function performs model fitting over a grid of parameter values and returns the model with the optimal values.  Optimality is determined based on the first performance metric of the `metrics` argument to `r rdoc_url("TunedModel()")` if given or the first default metric of the `r rdoc_url("performance()")` function otherwise.  Argument `grid` additionally controls the construction of grid values and can be a single integer or vector of integers whose positions or names match the parameters in a model's pre-defined tuning grid if one exists and which specify the number of values used to construct the grid.  Pre-defined `r rdoc_url("TunedModel")` grids can be extract and viewed apart from model fitting with `r rdoc_url("expand_modelgrid()")`.  As shown in the output below, `r rdoc_url("as.MLModel()")` will extract a tuned model from fit results for viewing of the tuning parameter grid values, the names of models fit to each, all calculated metrics, the final model selected, the metric upon which its selection was based, and its parameters.

```{r using_strategies_tune, eval=FALSE}
## Tune over automatic grid of model parameters
model_fit <- fit(surv_fo, data = surv_train,
                 model = TunedModel(
                   GBMModel,
                   grid = 3,
                   control = surv_means_control,
                   metrics = c("CIndex" = cindex, "RMSE" = rmse)
                 ))
(trained_model <- as.MLModel(model_fit))
#> --- MLModel object ----------------------------------------------------------
#> 
#> Model name: GBMModel
#> Label: Trained Generalized Boosted Regression
#> Package: gbm
#> Response types: factor, numeric, PoissonVariate, Surv
#> Case weights support: TRUE
#> Missing case removal: response
#> Tuning grid: TRUE
#> Variable importance: TRUE
#> 
#> Parameters:
#> List of 5
#>  $ n.trees          : int 50
#>  $ interaction.depth: int 1
#>  $ n.minobsinnode   : num 10
#>  $ shrinkage        : num 0.1
#>  $ bag.fraction     : num 0.5
#> 
#> === $TrainingStep1 ==========================================================
#> === TrainingStep object ===
#> 
#> Optimization method: Grid Search
#> TunedModel log:
#> # A tibble: 9 × 4
#>   name       selected params$n.trees $interaction.depth metrics$CIndex  $RMSE
#>   <chr>      <lgl>             <int>              <int>          <dbl>  <dbl>
#> 1 GBMModel.1 TRUE                 50                  1          0.765  3869.
#> 2 GBMModel.2 FALSE                50                  2          0.750  5450.
#> 3 GBMModel.3 FALSE                50                  3          0.739  7877.
#> 4 GBMModel.4 FALSE               100                  1          0.746  3919.
#> 5 GBMModel.5 FALSE               100                  2          0.730  5262.
#> 6 GBMModel.6 FALSE               100                  3          0.720 10596.
#> 7 GBMModel.7 FALSE               150                  1          0.726  4167.
#> 8 GBMModel.8 FALSE               150                  2          0.712  5428.
#> 9 GBMModel.9 FALSE               150                  3          0.699 16942.
#> 
#> Selected row: 1
#> Metric: CIndex = 0.7647633
```

Grid values may also be given as a `r rdoc_url("TuningGrid")` function, function name, or object; `r rdoc_url("ParameterGrid")` object; or data frame containing parameter values at which to evaluate the model, such as that returned by `r rdoc_url("expand_params()")`.

```{r using_strategies_tune_grid, eval=FALSE}
## Tune over randomly sampled grid points
fit(surv_fo, data = surv_train,
    model = TunedModel(
      GBMModel,
      grid = TuningGrid(size = 100, random = 10),
      control = surv_means_control
    ))

## Tune over user-specified grid points
fit(surv_fo, data = surv_train,
    model = TunedModel(
      GBMModel,
      grid = expand_params(n.trees = c(25, 50, 100),
                           interaction.depth = 1:3),
      control = surv_means_control
    ))
```

Statistics summarizing the resampled performance metrics across all tuning parameter combinations can be obtained with the `r rdoc_url("summary()")` and `r rdoc_url("performance()")`functions.

```{r using_strategies_tune_summary, eval=FALSE}
summary(trained_model)
#> --- $TrainingStep1 ----------------------------------------------------------
#> # A tibble: 9 × 4
#>   name       selected params$n.trees $interaction.depth metrics$CIndex  $RMSE
#>   <chr>      <lgl>             <int>              <int>          <dbl>  <dbl>
#> 1 GBMModel.1 TRUE                 50                  1          0.765  3869.
#> 2 GBMModel.2 FALSE                50                  2          0.750  5450.
#> 3 GBMModel.3 FALSE                50                  3          0.739  7877.
#> 4 GBMModel.4 FALSE               100                  1          0.746  3919.
#> 5 GBMModel.5 FALSE               100                  2          0.730  5262.
#> 6 GBMModel.6 FALSE               100                  3          0.720 10596.
#> 7 GBMModel.7 FALSE               150                  1          0.726  4167.
#> 8 GBMModel.8 FALSE               150                  2          0.712  5428.
#> 9 GBMModel.9 FALSE               150                  3          0.699 16942.

performance(trained_model) %>% lapply(summary)
#> $TrainingStep1
#> , , Metric = CIndex
#> 
#>             Statistic
#> Model             Mean    Median         SD       Min       Max NA
#>   GBMModel.1 0.7647633 0.7630332 0.05151443 0.6497797 0.8632075  0
#>   GBMModel.2 0.7502765 0.7549020 0.04636802 0.6688742 0.8349057  0
#>   GBMModel.3 0.7388766 0.7500000 0.04594197 0.6621622 0.8197425  0
#>   GBMModel.4 0.7463392 0.7601810 0.05087334 0.6497797 0.8160377  0
#>   GBMModel.5 0.7301734 0.7345133 0.04361416 0.6621622 0.7847222  0
#>   GBMModel.6 0.7199483 0.7122642 0.05168649 0.6216216 0.7889908  0
#>   GBMModel.7 0.7263693 0.7385321 0.05069831 0.6351351 0.7896996  0
#>   GBMModel.8 0.7121314 0.7169811 0.05331152 0.6199095 0.7847222  0
#>   GBMModel.9 0.6994678 0.7156863 0.05314909 0.6081081 0.7639485  0
#> 
#> , , Metric = RMSE
#> 
#>             Statistic
#> Model             Mean    Median        SD      Min       Max NA
#>   GBMModel.1  3868.629  3321.249  1431.839 2123.875  7401.789  0
#>   GBMModel.2  5450.093  4741.129  2438.173 2385.426 11695.273  0
#>   GBMModel.3  7877.487  5827.739  5036.837 4192.463 20009.854  0
#>   GBMModel.4  3919.126  3535.514  1454.999 2373.413  7558.280  0
#>   GBMModel.5  5262.122  5052.969  1900.353 2280.193  8212.727  0
#>   GBMModel.6 10595.888  8308.733  7057.327 4996.420 28610.367  0
#>   GBMModel.7  4167.128  3789.596  1115.371 2587.440  6701.087  0
#>   GBMModel.8  5427.518  6129.309  2313.742 2039.894  9212.369  0
#>   GBMModel.9 16941.696 10865.160 16841.774 4914.743 71150.177  0
```

Line plots of tuning results display the resampled metric means, or another statistic specified with the `stat` argument, versus the first tuning parameter values and with lines grouped according to the remaining parameters, if any.

```{r using_strategies_tune_plot, eval=FALSE}
plot(trained_model, type = "line")
#> $TrainStep1
```

```{r using_strategies_tune_png, echo=FALSE}
knitr::include_graphics("img/using_strategies_tune_plot-1.png")
```


### Model Selection

Model selection can be conducted by calling `r rdoc_url("fit()")` with a `r rdoc_url("SelectedModel")` to automatically choose from any combination of models and model parameters.  Selection has as a special case the just-discussed tuning of a single model over a grid of parameter values.  Combinations of model functions, function names, or objects can be supplied to `r rdoc_url("SelectedModel()")` in order to define sets of candidate models from which to select.  An `r rdoc_url("expand_model()")` helper function is additionally available to expand a model over a grid of tuning parameters for inclusion in the candidate set if so desired.

```{r using_strategies_select, results="hide", eval=FALSE}
## Model interface for model selection
sel_model <- SelectedModel(
  expand_model(GBMModel, n.trees = c(50, 100), interaction.depth = 1:2),
  GLMNetModel(lambda = 0.01),
  CoxModel,
  SurvRegModel
)

## Fit the selected model
fit(surv_fo, data = surv_train, model = sel_model)
```

Selection may also be performed over candidate sets that include tuned models.  For instance, the `r rdoc_url("SelectedModel()")` function is applicable to sets containing different classes of models each individually tuned over a grid of parameters.

```{r using_strategies_select_tune, results="hide", eval=FALSE}
## Model interface for selection among tuned models
sel_tun_model <- SelectedModel(
  TunedModel(GBMModel, control = surv_means_control),
  TunedModel(GLMNetModel, control = surv_means_control),
  TunedModel(CoxModel, control = surv_means_control)
)

## Fit the selected tuned model
fit(surv_fo, data = surv_train, model = sel_tun_model)
```


## Ensemble Learning

Ensemble learning models combine $m = 1, \ldots, M$ base models as a strategy to improve predictive performance.  Two methods implemented in **MachineShop** are *stacked regression* [@breiman:1996:SR] and *super learners* [@vanderLaan:2007:SL].  Stacked regression fits a linear combination of predictions from specified base learners to produce a prediction function of the form
$$
\hat{f}(x) = \sum_{m=1}^M \hat{w}_m \hat{f}_m(x).
$$
Stacking weights $w$ are estimated by (constrained) least squares regression of case responses $y_i$ on predictions $\hat{f}_m^{-\kappa(i)}(x_i)$ from learners fit to data subsamples $-\kappa(i)$ not containing the corresponding cases. In particular, they are obtained as the solution
$$
\hat{w} = \underset{w}{\operatorname{argmin}} \sum_{i=1}^{N}\left(y_i - \sum_{m=1}^{M} w_m \hat{f}_m^{-\kappa(i)}(x_i) \right)^2
$$
subject to the constraints that all $w_m \ge 0$ and $\sum_m w_m = 1$.  K-fold cross-validation is the default subsampling method employed in the estimation, with the other resampling methods provided by the package available as options.  Survival outcomes are handled with a modified version of the stacked regression algorithm in which

* minimization of least squares is replaced by maximization of Harrell's concordance index [-@harrell:1982:EYM] to accommodate censoring, and
* prediction can only be performed on the same response type used for the model fit; i.e., either survival means or survival probabilities at given follow-up times.

Super learners are a generalization of stacked regression that fit a specified model, such as `r rdoc_url("GBMModel")`, to case responses $y_i$, base learner predictions $\hat{f}_m^{-\kappa(i)}(x_i)$, and optionally also to the original predictor variables $x_i$.  Given below are examples of a stacked regression and super learner each fit with gradient boosted, random forest, and Cox regression base learners.  A separate gradient boosted model is used as the super learner in the latter.

```{r using_strategies_ensembles, eval=FALSE}
## Stacked regression
stackedmodel <- StackedModel(CoxModel, CForestModel, GLMBoostModel)
res_stacked <- resample(surv_fo, data = surv_train, model = stackedmodel)
summary(res_stacked)
#>          Statistic
#> Metric         Mean   Median        SD       Min       Max NA
#>   C-Index 0.7294869 0.762963 0.1275484 0.5194805 0.8556701  0

## Super learner
supermodel <- SuperModel(CoxModel, CForestModel, GLMBoostModel,
                         model = GBMModel)
res_super <- resample(surv_fo, data = surv_train, model = supermodel)
summary(res_super)
#>          Statistic
#> Metric         Mean    Median        SD       Min       Max NA
#>   C-Index 0.7534803 0.8325472 0.1243104 0.5748899 0.8454545  0
```


## Methodology

Combinations and multiple levels of nested meta-functions, inputs, and models are possible.  If model fitting involves a single meta-function, performances of the inputs or models under consideration are estimated with standard resampling, and the best performing model is returned.  Nestings of meta-functions are trained with nested resampling.  Consider the example below in which training involves input tuning and model selection.  In particular, a preprocessing recipe is tuned over the number of predictor-derived principal components and model selection is of an untuned `r rdoc_url("GBMModel")`, a tuned `r rdoc_url("GBMModel")`, and a `r rdoc_url("StackedModel")`.

```{r using_strategies_methods, eval=FALSE}
## Preprocessing recipe with PCA steps
pca_rec <- recipe(y ~ ., data = surv_train) %>%
  role_case(stratum = y) %>%
  step_center(all_predictors()) %>%
  step_scale(all_predictors()) %>%
  step_pca(all_predictors(), id = "PCA")

## Tuning grid of number of PCA components
pca_grid <- expand_steps(
  PCA = list(num_comp = 1:3)
)

## Input specification
tun_rec <- TunedInput(pca_rec, grid = pca_grid)

## Model specification
sel_model <- SelectedModel(
  GBMModel,
  TunedModel(GBMModel),
  StackedModel(CoxModel, TunedModel(CForestModel), TunedModel(GBMModel))
)

## Model fit and final trained model
model_fit <- fit(tun_rec, model = sel_model)
as.MLModel(model_fit)
```

Model fitting proceeds with instances of the specified model selection nested within each of the input tuning grid parameter values.  Tuning of `r rdoc_url("GBMModel")` and construction of `r rdoc_url("StackedModel")` are further nested within the model selection, with tuning of `r rdoc_url("CForestModel")` and `r rdoc_url("GBMModel")` nested within `r rdoc_url("StackedModel")`.  Altogether, there are four levels of meta-input and meta-model functions in the hierarchy.

```{r using_strategies_dag, echo = FALSE, out.width = "100%"}
knitr::include_graphics("img/FigModelDAG.png")
```

Each meta-function is fit based on resample estimation (default: cross-validation) of predictive performance.  When one meta-function is nested within another, nested resampling is employed, as illustrated in the figure below.

```{r using_strategies_nestedcv, echo = FALSE, out.width = "100%"}
knitr::include_graphics("img/FigNestedCV.png")
```

Nesting of resampling routines is repeated recursively when a fit involves multiple levels of nested meta-functions.  For example, predictive performance estimation for the training of `r rdoc_url("TunedInput(pca_rec, grid = pca_grid)")` involves up to three nested meta functions: `r rdoc_url("SelectedModel(...)")` &rarr; `r rdoc_url("StackedModel(...)")` &rarr; `r rdoc_url("TunedModel(CForestModel)")`.  For this relationship, an outer and three nested inner resampling loops are executed as follows.  First, `r rdoc_url("CForestModel")` is tuned at the third inner resampling loop.  Second, the tuned model is passed to the second inner loop for construction of `r rdoc_url("StackedModel")`.  Third, the constructed model is passed to the first inner loop for model selection from the candidate set.  Finally, the selected model is passed to the outer loop for tuning of the preprocessing recipe.  Based on resample performance estimation of the entire input/model specification, one principal component is selected for the `GBMModel`.

```{r using_strategies_methods1, eval=FALSE}
#> === $TrainingStep1 ==========================================================
#> === TrainingStep object ===
#> 
#> Optimization method: Grid Search
#> TunedModelRecipe log:
#> # A tibble: 3 × 4
#>   name          selected params$PCA$num_comp metrics$`C-Index`
#>   <chr>         <lgl>                  <int>             <dbl>
#> 1 ModelRecipe.1 TRUE                       1             0.741
#> 2 ModelRecipe.2 FALSE                      2             0.730
#> 3 ModelRecipe.3 FALSE                      3             0.702
#> 
#> Selected row: 1
#> Metric: C-Index = 0.7405534
```

In order to identify and return a final model fitted to the entire input data, the hierarchy is traversed from top to bottom along the path determined by the choice at each node.  Steps along the path are labeled `TrainStep1` and `TrainStep2` in the output.  As seen above in `TrainStep1`, one principal component is first selected for the tuned input.  Using an input recipe with one principal component, the entire dataset is refit at `TrainStep2` to finally select `r rdoc_url("GBMModel")`.

```{r using_strategies_methods2, eval=FALSE}
#> === $TrainingStep2 ==========================================================
#> === TrainingStep object ===
#> 
#> Optimization method: Grid Search
#> SelectedModel log:
#> # A tibble: 3 × 4
#>   name         selected params$id  metrics$`C-Index`
#>   <chr>        <lgl>    <chr>                  <dbl>
#> 1 GBMModel     TRUE     model.1mQk             0.740
#> 2 TunedModel   FALSE    model.hRF5             0.735
#> 3 StackedModel FALSE    model.CYj0             0.653
#> 
#> Selected row: 1
#> Metric: C-Index = 0.7399641
```

After the series of training steps reaches the bottom of its hierarchy, the final model is fitted to the entire dataset and returned.

```{r using_strategies_methods0, eval=FALSE}
#> --- MLModel object ----------------------------------------------------------
#> 
#> Model name: GBMModel
#> Label: Trained Generalized Boosted Regression
#> Package: gbm
#> Response types: factor, numeric, PoissonVariate, Surv
#> Case weights support: TRUE
#> Missing case removal: response
#> Tuning grid: TRUE
#> Variable importance: TRUE
#> 
#> Parameters:
#> List of 5
#>  $ n.trees          : num 100
#>  $ interaction.depth: num 1
#>  $ n.minobsinnode   : num 10
#>  $ shrinkage        : num 0.1
#>  $ bag.fraction     : num 0.5
```

Generalization performance of the entire process can be estimated with a call to `r rdoc_url("resample()")`.

```{r eval=FALSE}
## Generalization performance of the modeling strategy
resample(tun_rec, model = sel_model)
```

There is no conceptual limit to the number of nested inputs and models that can be specified with the package.  However, there are some practical issues to consider.

Computational Expense
  : Computational expense of nested resampling increases exponentially.  For instance, execution of *r* levels of a nested 10-fold cross-validation algorithm is an O(10^*r*^) operation.  Runtimes can be decreased by registering multiple cores to run the resampling algorithms in parallel.  However, the exponential increase in computational complexity quickly outpaces the number of available cores.

Data Reduction
  : Training data is reduced at each subsequent resampling level.  For 10-fold cross-validation and a training set of *N* total cases, there will be 0.9^*r*^ cases available at each fold of the *r*^th^ resampling algorithm.  Bootstrapping could be used, as an alternative to cross-validation, to ensure *N* cases at each resampling level.  However, the number of unique cases at level *r* will be decreased to approximately *N*(2/3)^*r*^.


# Global Settings

Core default behaviors of functions in the package can be viewed or changed globally through the `r rdoc_url("settings()")` function.  The function accepts one or more character names of settings to view, `name = value` pairs giving the values of settings to change, or a vector of these, with available settings summarized below.

`control`
  : function, function name, or object defining a default resampling method [default: ``"CVControl"``].

`cutoff`
  : numeric (0, 1) threshold above which binary factor probabilities are classified as events and below which survival probabilities are classified [default: 0.5].

`distr.SurvMeans`
  : character string specifying distributional approximations to estimated survival curves for predicting survival means.  Choices are `"empirical"` for the Kaplan-Meier estimator, `"exponential"`, `"rayleigh"`, or `"weibull"` (default).

`distr.SurvProbs`
  : character string specifying distributional approximations to estimated survival curves for predicting survival events/probabilities.  Choices are `"empirical"` (default) for the Kaplan-Meier estimator, `"exponential"`, `"rayleigh"`, or `"weibull"`.

`grid`
  : `size` argument to `r rdoc_url("TuningGrid()")` indicating the number of parameter-specific values to generate automatically for tuning of models that have pre-defined grids or a `r rdoc_url("TuningGrid()")` function, function name, or object [default: 3].

`method.EmpiricalSurv`
  : character string specifying the empirical method of estimating baseline survival curves for Cox proportional hazards-based models.  Choices are `"breslow"` or `"efron"` (default).

`metrics.ConfusionMatrix`
  : function, function name, or vector of these with which to calculate performance metrics for confusion matrices [default: ``c(Accuracy = "accuracy", Kappa = "kappa2", `Weighted Kappa` = "weighted_kappa2", Sensitivity = "sensitivity", Specificity = "specificity")``].

`metrics.factor`
  : function, function name, or vector of these with which to calculate performance metrics for factor responses [default: ``c(Brier = "brier", Accuracy = "accuracy", Kappa = "kappa2", `Weighted Kappa` = "weighted_kappa2", `ROC AUC` = "roc_auc", Sensitivity = "sensitivity", Specificity = "specificity")``].

`metrics.matrix`
  : function, function name, or vector of these with which to calculate performance metrics for matrix responses [default: ``c(RMSE = "rmse", R2 = "r2", MAE = "mae")``].

`metrics.numeric`
  : function, function name, or vector of these with which to calculate performance metrics for numeric responses [default: ``c(RMSE = "rmse", R2 = "r2", MAE = "mae")``].

`metrics.Surv`
  : function, function name, or vector of these with which to calculate performance metrics for survival responses [default: ``c(`C-Index` = "cindex", Brier = "brier", `ROC AUC` = "roc_auc", Accuracy = "accuracy")``].

`print_max`
  : number of models or data rows to show with print methods or `Inf` to show all [default: 10].

`require`
  : names of installed packages to load during parallel execution of resampling algorithms [default: ``"MachineShop"``].

`reset`
  : character names of settings to reset to their default values.

`RHS.formula`
  : non-modifiable character vector of operators and functions allowed in traditional formula specifications.

`stat.Curve`
  : function or character string naming a function to compute one summary statistic at each cutoff value of resampled metrics in performance curves, or `NULL` for resample-specific metrics [default: ``"base::mean"``].

`stat.Resample`
  : function or character string naming a function to compute one summary statistic to control the ordering of models in plots [default: ``"base::mean"``].

`stat.TrainingParams`
  : function or character string naming a function to compute one summary statistic on resampled performance metrics for input selection or tuning or for model selection or tuning [default: ``"base::mean"``].

`stats.PartialDependence`
  : function, function name, or vector of these with which to compute partial dependence summary statistics [default: ``c(Mean = "base::mean")``].

`stats.Resample`
  : function, function name, or vector of these with which to compute summary statistics on resampled performance metrics [default: ``c(Mean = "base::mean", Median = "stats::median", SD = "stats::sd", Min = "base::min", Max = "base::max")``].

A call to `r rdoc_url("settings()")` with `"reset"` will restore all package defaults and with no arguments will display the current values of all.  Settings may also be supplied as a single unnamed argument which is a named list.  Partial matching of setting names is supported.  The setting value is returned if only one is specified to view.  Otherwise, a list is returned with the values of specified settings as they existed prior to any requested changes.  Such a list can be passed as an argument to `r rdoc_url("settings()")` to restore their values.

```{r using_settings}
## Change settings
presets <- settings(control = "BootControl", grid = 10)

## View one setting
settings("control")

## View multiple settings
settings("control", "grid")

## Restore the previous settings
settings(presets)
```


# Package Extensions

Custom models and metrics can be defined with `r rdoc_url("MLModel()")` and `r rdoc_url("MLMetric()")` for use with the model fitting, prediction, and performance assessment tools provided by the package.


## Custom Models

The `r rdoc_url("MLModel()")` function creates a model object that can be used with the previously described fitting functions.  It take the following arguments.

`name`
  : character name of the object to which the model is assigned.

`label`
  : optional character descriptor for the model (default: `name`).

`packages`
  : character vector of packages upon which the model depends.  Each name may be optionally followed by a comment in parentheses specifying a version requirement.  The comment should contain a comparison operator, whitespace and a valid version number, e.g. `"xgboost (>= 1.3.0)"`.

`response_types`
  : character vector of response variable types to which the model can be fit.  Supported types are `"binary"`, `"BinomialVariate"`, `"DiscreteVariate"`, `"factor"`, `"matrix"`, `"NegBinomialVariate"`, `"numeric"`, `"ordered"`, `"PoissonVariate"`, and `"Surv"`.

`fit`
  : model fitting function whose arguments are a `formula`, a `r rdoc_url("ModelFrame")` named `data`, case `weights`, and an ellipsis.  Argument `data` may be converted to a data frame with the `as.data.frame()` function as is commonly needed.  The fit function should return the object resulting from the model fit.

`predict`
  : prediction function whose arguments are the `object` returned by `r rdoc_url("fit()")`, a `r rdoc_url("ModelFrame")` named `newdata` of predictor variables, optional vector of `times` at which to predict survival, and an ellipsis.  Argument `data` may be converted to a data frame with the `as.data.frame()` function as needed.  Values returned by the function should be formatted according to the response variable types below.

* `factor` : matrix whose columns contain the probabilities for multi-level factors or vector of probabilities for the second level of binary factors.
* `matrix` : matrix of predicted responses.
* `numeric` : vector or column matrix of predicted responses.
* `Surv` : matrix whose columns contain survival probabilities at `times` if supplied or vector of predicted survival means otherwise.

`varimp`
  : optional variable importance function whose arguments are the `object` returned by `r rdoc_url("fit()")`, optional arguments passed from calls to `r rdoc_url("varimp()")`, and an ellipsis.  The function should return a vector of importance values named after the predictor variables or a matrix or data frame whose rows are named after the predictors.

```{r using_extensions_mlmodel}
## Logistic regression model extension
LogisticModel <- MLModel(
  name = "LogisticModel",
  label = "Logistic Model",
  response_types = "binary",
  weights = TRUE,
  fit = function(formula, data, weights, ...) {
    glm(formula, data = as.data.frame(data), weights = weights,
        family = binomial, ...)
  },
  predict = function(object, newdata, ...) {
    predict(object, newdata = as.data.frame(newdata), type = "response")
  },
  varimp = function(object, ...) {
    pchisq(coef(object)^2 / diag(vcov(object)), 1)
  }
)
```


## Custom Metrics

The `r rdoc_url("MLMetric()")` function creates a metric object that can be used as previously described for the model performance metrics.  Its first argument is a function to compute the metric, defined to accept `observed` and `predicted` as the first two arguments and with an ellipsis to accommodate others.  Its remaining arguments are as follows.

`name`
  :  character name of the object to which the metric is assigned.

`label`
  : optional character descriptor for the metric (default: `name`).

`maximize`
  : logical indicating whether higher values of the metric correspond to better predictive performance.

```{r using_extensions_mlmetric}
## F2 score metric extension
f2_score <- MLMetric(
  function(observed, predicted, ...) {
    f_score(observed, predicted, beta = 2, ...)
  },
  name = "f2_score",
  label = "F2 Score",
  maximize = TRUE
)
```


## Usage

Once created, model and metric extensions can be used with the package-supplied fitting and performance functions.

```{r using_extensions_usage}
## Logistic regression analysis
data(Pima.tr, package = "MASS")
res <- resample(type ~ ., data = Pima.tr, model = LogisticModel)
summary(performance(res, metric = f2_score))
```


# Model Constructor Functions

```{r reference_models, echo = FALSE}
library(MachineShop)

info <- modelinfo()
types <- c("binary" = "b", "factor" = "f", "matrix" = "m", "numeric" = "n",
           "ordered" = "o", "Surv" = "S")
x <- lapply(names(info), function(modelname) {
  c(modelname, ifelse(names(types) %in% info[[modelname]]$response_types, types, NA))
})
df <- as.data.frame(do.call(rbind, x), stringsAsFactors = FALSE)
names(df) <- c("Function", names(types))
rdoc_names <- sapply(df$Function, function(name) {
  switch(name,
    "CoxStepAICModel" = "CoxModel",
    "GLMStepAICModel" = "GLMModel",
    "PDAModel" = "FDAModel",
    "SurvRegStepAICModel" = "SurvRegModel",
    "SVMANOVAModel" = "SVMModel",
    "SVMBesselModel" = "SVMModel",
    "SVMLaplaceModel" = "SVMModel",
    "SVMLinearModel" = "SVMModel",
    "SVMPolyModel" = "SVMModel",
    "SVMRadialModel" = "SVMModel",
    "SVMSplineModel" = "SVMModel",
    "SVMTanhModel" = "SVMModel",
    "XGBDARTModel" = "XGBModel",
    "XGBLinearModel" = "XGBModel",
    "XGBTreeModel" = "XGBModel",
    name
  )
})

toString2 <- function(x) toString(na.omit(x))
df_classes <- data.frame(
  Function = rdoc_url(df$Function, rdoc_names),
  Label = sapply(info, getElement, name = "label"),
  Categorical = apply(df[c("binary", "factor", "ordered")], 1, toString2),
  Continuous = apply(df[c("matrix", "numeric")], 1, toString2),
  Survival = apply(df["Surv"], 1, toString2)
)
names(df_classes)[3:5] <- paste0(names(df_classes)[3:5], footnote_marker_number(1:3))

kable(df_classes,
      caption = "Package-supplied model constructor functions and supported response variable types.",
      align = c("l", "l", "c", "c", "c"), row.names = FALSE,
      escape = FALSE) %>%
  kable_styling(c("striped", "condensed"), full_width = FALSE, position = "center") %>%
  add_header_above(c(" " = 1, " " = 1, "Response Variable Types" = 3)) %>%
  footnote(number = c("b = binary factor, f = factor, o = ordered factor",
                      "m = matrix, n = numeric",
                      "S = Surv"))
```


# Metric Functions

```{r reference_metrics, echo=FALSE}
library(MachineShop)

f <- function(x) {
  types <- x$response_types
  
  is_type <- function(observed, predicted) {
    any(types$observed == observed & types$predicted == predicted)
  }
  
  categorical <- if (is_type("factor", "matrix")) "f" else
    if (is_type("factor", "numeric")) "b" else NULL
  if (is_type("ordered", "ordered")) categorical <- c(categorical, "o")

  continuous <- NULL
  if (is_type("matrix", "matrix")) continuous <- "m"
  if (is_type("numeric", "numeric")) continuous <- c(continuous, "n")
  
  survival <- NULL
  if (any(mapply(is_type, "Surv", c("numeric", "SurvEvents", "SurvProbs")))) {
    survival <- "S"
  }

  data.frame(
    Label = x$label,
    Categorical = toString(categorical),
    Continuous = toString(continuous),
    Survival = toString(survival)
  )
}

info <- metricinfo()
df <- cbind("Function" = rdoc_url(names(info), "metrics"),
            do.call(rbind, lapply(info, f)))
names(df)[3:5] <- paste0(names(df)[3:5], footnote_marker_number(1:3))

kable(df, caption = "Package-supplied performance metric functions and supported response variable types.",
      align = c("l", "l", "c", "c", "c"), row.names = FALSE,
      escape = FALSE) %>%
  kable_styling(c("striped", "condensed"), full_width = FALSE, position = "center") %>%
  add_header_above(c(" " = 1, " " = 1, "Response Variable Types" = 3)) %>%
  footnote(number = c("b = binary factor, f = factor, o = ordered factor",
                      "m = matrix, n = numeric",
                      "S = Surv"))
```


# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_SuperModel.R
\name{SuperModel}
\alias{SuperModel}
\title{Super Learner Model}
\usage{
SuperModel(
  ...,
  model = GBMModel,
  control = MachineShop::settings("control"),
  all_vars = FALSE
)
}
\arguments{
\item{...}{\link[=models]{model} functions, function names, objects; other
objects that can be \link[=as.MLModel]{coerced} to models; or vector of
these to serve as base learners.}

\item{model}{\link[=models]{model} function, function name, or object
defining the super model; or another object that can be
\link[=as.MLModel]{coerced} to the model.}

\item{control}{\link[=controls]{control} function, function name, or object
defining the resampling method to be employed for the estimation of base
learner weights.}

\item{all_vars}{logical indicating whether to include the original
predictor variables in the super model.}
}
\value{
\code{SuperModel} class object that inherits from \code{MLModel}.
}
\description{
Fit a super learner model to predictions from multiple base learners.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{numeric}, \code{ordered},
    \code{Surv}}
}
}
\examples{
\donttest{
## Requires prior installation of suggested packages gbm and glmnet to run

model <- SuperModel(GBMModel, SVMRadialModel, GLMNetModel(lambda = 0.01))
model_fit <- fit(sale_amount ~ ., data = ICHomes, model = model)
predict(model_fit, newdata = ICHomes)
}

}
\references{
van der Laan, M. J., Polley, E. C., & Hubbard, A. E. (2007). Super learner.
\emph{Statistical Applications in Genetics and Molecular Biology},
\emph{6}(1).
}
\seealso{
\code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_SVMModel.R
\name{SVMModel}
\alias{SVMModel}
\alias{SVMANOVAModel}
\alias{SVMBesselModel}
\alias{SVMLaplaceModel}
\alias{SVMLinearModel}
\alias{SVMPolyModel}
\alias{SVMRadialModel}
\alias{SVMSplineModel}
\alias{SVMTanhModel}
\title{Support Vector Machine Models}
\usage{
SVMModel(
  scaled = TRUE,
  type = character(),
  kernel = c("rbfdot", "polydot", "vanilladot", "tanhdot", "laplacedot", "besseldot",
    "anovadot", "splinedot"),
  kpar = "automatic",
  C = 1,
  nu = 0.2,
  epsilon = 0.1,
  prob.model = FALSE,
  cache = 40,
  tol = 0.001,
  shrinking = TRUE
)

SVMANOVAModel(sigma = 1, degree = 1, ...)

SVMBesselModel(sigma = 1, order = 1, degree = 1, ...)

SVMLaplaceModel(sigma = numeric(), ...)

SVMLinearModel(...)

SVMPolyModel(degree = 1, scale = 1, offset = 1, ...)

SVMRadialModel(sigma = numeric(), ...)

SVMSplineModel(...)

SVMTanhModel(scale = 1, offset = 1, ...)
}
\arguments{
\item{scaled}{logical vector indicating the variables to be scaled.}

\item{type}{type of support vector machine.}

\item{kernel}{kernel function used in training and predicting.}

\item{kpar}{list of hyper-parameters (kernel parameters).}

\item{C}{cost of constraints violation defined as the regularization term in
the Lagrange formulation.}

\item{nu}{parameter needed for nu-svc, one-svc, and nu-svr.}

\item{epsilon}{parameter in the insensitive-loss function used for eps-svr,
nu-svr and eps-bsvm.}

\item{prob.model}{logical indicating whether to calculate the scaling
parameter of the Laplacian distribution fitted on the residuals of numeric
response variables.  Ignored in the case of a factor response variable.}

\item{cache}{cache memory in MB.}

\item{tol}{tolerance of termination criterion.}

\item{shrinking}{whether to use the shrinking-heuristics.}

\item{sigma}{inverse kernel width used by the ANOVA, Bessel, and Laplacian
kernels.}

\item{degree}{degree of the ANOVA, Bessel, and polynomial kernel functions.}

\item{...}{arguments passed to \code{SVMModel} from the other constructors.}

\item{order}{order of the Bessel function to be used as a kernel.}

\item{scale}{scaling parameter of the polynomial and hyperbolic tangent
kernels as a convenient way of normalizing patterns without the need to
modify the data itself.}

\item{offset}{offset used in polynomial and hyperbolic tangent kernels.}
}
\value{
\code{MLModel} class object.
}
\description{
Fits the well known C-svc, nu-svc, (classification) one-class-svc (novelty)
eps-svr, nu-svr (regression) formulations along with native multi-class
classification formulations and the bound-constraint SVM formulations.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{numeric}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \itemize{
      \item SVMModel: \code{NULL}
      \item SVMANOVAModel: \code{C}, \code{degree}
      \item SVMBesselModel: \code{C}, \code{order}, \code{degree}
      \item SVMLaplaceModel: \code{C}, \code{sigma}
      \item SVMLinearModel: \code{C}
      \item SVMPolyModel: \code{C}, \code{degree}, \code{scale}
      \item SVMRadialModel: \code{C}, \code{sigma}
    }
  }
}

The kernel-specific constructor functions \code{SVMANOVAModel},
\code{SVMBesselModel}, \code{SVMLaplaceModel}, \code{SVMLinearModel},
\code{SVMPolyModel}, \code{SVMRadialModel}, \code{SVMSplineModel}, and
\code{SVMTanhModel} are special cases of \code{SVMModel} which automatically
set its \code{kernel} and \code{kpar} arguments.  These are called directly
in typical usage unless \code{SVMModel} is needed to specify a more general
model.

Default argument values and further model details can be found in the source
See Also link below.
}
\examples{
fit(sale_amount ~ ., data = ICHomes, model = SVMRadialModel)

}
\seealso{
\code{\link[kernlab]{ksvm}}, \code{\link{fit}},
\code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_BARTMachineModel.R
\name{BARTMachineModel}
\alias{BARTMachineModel}
\title{Bayesian Additive Regression Trees Model}
\usage{
BARTMachineModel(
  num_trees = 50,
  num_burn = 250,
  num_iter = 1000,
  alpha = 0.95,
  beta = 2,
  k = 2,
  q = 0.9,
  nu = 3,
  mh_prob_steps = c(2.5, 2.5, 4)/9,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{num_trees}{number of trees to be grown in the sum-of-trees model.}

\item{num_burn}{number of MCMC samples to be discarded as "burn-in".}

\item{num_iter}{number of MCMC samples to draw from the posterior
distribution.}

\item{alpha, beta}{base and power hyperparameters in tree prior for whether a
node is nonterminal or not.}

\item{k}{regression prior probability that \eqn{E(Y|X)} is
contained in the interval \eqn{(y_{min}, y_{max})}, based on a normal
distribution.}

\item{q}{quantile of the prior on the error variance at which the data-based
estimate is placed.}

\item{nu}{regression degrees of freedom for the inverse \eqn{sigma^2} prior.}

\item{mh_prob_steps}{vector of prior probabilities for proposing changes to
the tree structures: (GROW, PRUNE, CHANGE).}

\item{verbose}{logical indicating whether to print progress information about
the algorithm.}

\item{...}{additional arguments to \code{\link[bartMachine]{bartMachine}}.}
}
\value{
\code{MLModel} class object.
}
\description{
Builds a BART model for regression or classification.
}
\details{
\describe{
  \item{Response types:}{\code{binary factor}, \code{numeric}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \code{alpha}, \code{beta}, \code{k}, \code{nu}
  }
}

Further model details can be found in the source link below.

In calls to \code{\link{varimp}} for \code{BARTMachineModel}, argument
\code{type} may be specified as \code{"splits"} (default) for the
proportion of time each predictor is chosen for a splitting rule or as
\code{"trees"} for the proportion of times each predictor appears in a tree.
Argument \code{num_replicates} is also available to control the number of
BART replicates used in estimating the inclusion proportions [default: 5].
Variable importance is automatically scaled to range from 0 to 100.  To
obtain unscaled importance values, set \code{scale = FALSE}.  See example
below.
}
\examples{
\dontrun{
## Requires prior installation of suggested package bartMachine, java, and
## setting of java options as shown below to run

if (packageVersion("bartMachine") >= "1.4") {
  options(
    java.parameters = c(
      "-Xmx20g", "--add-modules=jdk.incubator.vector", "-XX:+UseZGC"
    )
  )
} else {
  options(java.parameters = "-Xmx20g")
}

model_fit <- fit(sale_amount ~ ., data = ICHomes, model = BARTMachineModel)
varimp(model_fit, method = "model", type = "splits", num_replicates = 20,
       scale = FALSE)
}

}
\seealso{
\code{\link[bartMachine]{bartMachine}}, \code{\link{fit}},
\code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metricinfo.R
\name{metricinfo}
\alias{metricinfo}
\title{Display Performance Metric Information}
\usage{
metricinfo(...)
}
\arguments{
\item{...}{\link[=metrics]{metric} functions or function names;
\link[=response]{observed responses}; \link[=response]{observed} and
\link[=predict]{predicted} responses; \link{confusion} or \link{resample}
results for which to display information.  If none are specified, information
is returned on all available metrics by default.}
}
\value{
List of named metric elements each containing the following
components:
\describe{
  \item{label}{character descriptor for the metric.}
  \item{maximize}{logical indicating whether higher values of the metric
    correspond to better predictive performance.}
  \item{arguments}{closure with the argument names and corresponding default
    values of the metric function.}
  \item{response_types}{data frame of the observed and predicted response
    variable types supported by the metric.}
}
}
\description{
Display information about metrics provided by the \pkg{MachineShop} package.
}
\examples{
## All metrics
metricinfo()

## Metrics by observed and predicted response types
names(metricinfo(factor(0)))
names(metricinfo(factor(0), factor(0)))
names(metricinfo(factor(0), matrix(0)))
names(metricinfo(factor(0), numeric(0)))

## Metric-specific information
metricinfo(auc)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TrainingParams.R
\name{set_strata}
\alias{set_strata}
\title{Resampling Stratification Control}
\usage{
set_strata(object, breaks = 4, nunique = 5, prop = 0.1, size = 20, ...)
}
\arguments{
\item{object}{\link[=controls]{control} object.}

\item{breaks}{number of quantile bins desired for stratification of numeric
data during resampling.}

\item{nunique}{number of unique values at or below which numeric data are
stratified as categorical.}

\item{prop}{minimum proportion of data in each strata.}

\item{size}{minimum number of values in each strata.}

\item{...}{arguments passed to other methods.}
}
\value{
Argument \code{object} updated with the supplied parameters.
}
\description{
Set parameters that control the construction of strata during resample
estimation of model performance.
}
\details{
The arguments control resampling strata which are constructed from numeric
proportions for \code{\link{BinomialVariate}}; original values for
\code{character}, \code{factor}, \code{logical}, \code{numeric}, and
\code{ordered}; first columns of values for \code{matrix}; and numeric times
within event statuses for \code{Surv}.  Stratification of survival data by
event status only can be achieved by setting \code{breaks = 1}.  Numeric
values are stratified into quantile bins and categorical values into factor
levels.  The number of bins will be the largest integer less than or equal to
\code{breaks} satisfying the \code{prop} and \code{size} control argument
thresholds.  Categorical levels below the thresholds will be pooled
iteratively by reassigning values in the smallest nominal level to the
remaining ones at random and by combining the smallest adjacent ordinal
levels.  Missing values are replaced with non-missing values sampled at
random with replacement.
}
\examples{
CVControl() \%>\% set_strata(breaks = 3)

}
\seealso{
\code{\link{resample}}, \code{\link{set_monitor}},
\code{\link{set_optim}}, \code{\link{set_predict}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TrainingParams.R
\name{set_monitor}
\alias{set_monitor}
\alias{set_monitor.MLControl}
\alias{set_monitor.MLOptimization}
\alias{set_monitor.ModelSpecification}
\title{Training Parameters Monitoring Control}
\usage{
set_monitor(object, ...)

\method{set_monitor}{MLControl}(object, progress = TRUE, verbose = FALSE, ...)

\method{set_monitor}{MLOptimization}(object, progress = FALSE, verbose = FALSE, ...)

\method{set_monitor}{ModelSpecification}(object, which = c("all", "control", "optim"), ...)
}
\arguments{
\item{object}{resampling \link[=controls]{control},
tuning parameter \link[=set_optim]{optimization}, or model
\link[=ModelSpecification]{specification} object.}

\item{...}{arguments passed from the \code{ModelSpecification} method to the
others.}

\item{progress}{logical indicating whether to display iterative progress
during resampling or optimization.  In the case of resampling, a progress
bar will be displayed if a computing cluster is not registered or is
registered with the \pkg{doSNOW} package.}

\item{verbose}{numeric or logical value specifying the level of progress
detail to print, with 0 (\code{FALSE}) indicating none and 1 (\code{TRUE})
or higher indicating increasing amounts of detail.}

\item{which}{character string specifying the monitoring parameters to set as
\code{"all"}, \code{"control"}, or optimization (\code{"optim"}).}
}
\value{
Argument \code{object} updated with the supplied parameters.
}
\description{
Set parameters that control the monitoring of resample estimation of model
performance and of tuning parameter optimization.
}
\examples{
CVControl() \%>\% set_monitor(verbose = TRUE)

}
\seealso{
\code{\link{resample}}, \code{\link{set_optim}},
\code{\link{set_predict}}, \code{\link{set_strata}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calibration.R
\name{calibration}
\alias{calibration}
\title{Model Calibration}
\usage{
calibration(
  x,
  y = NULL,
  weights = NULL,
  breaks = 10,
  span = 0.75,
  distr = character(),
  pool = FALSE,
  na.rm = TRUE,
  ...
)
}
\arguments{
\item{x}{\link[=response]{observed responses} or \link{resample} result
containing observed and predicted responses.}

\item{y}{\link[=predict]{predicted responses} if not contained in \code{x}.}

\item{weights}{numeric vector of non-negative
\link[=case_weights]{case weights} for the observed \code{x} responses
[default: equal weights].}

\item{breaks}{value defining the response variable bins within which to
calculate observed mean values.  May be specified as a number of bins, a
vector of breakpoints, or \code{NULL} to fit smooth curves with splines for
predicted survival probabilities and with \link[stats:loess]{loess} for
others.}

\item{span}{numeric parameter controlling the degree of loess smoothing.}

\item{distr}{character string specifying a distribution with which to
estimate the observed survival mean.  Possible values are
\code{"empirical"} for the Kaplan-Meier estimator, \code{"exponential"},
\code{"extreme"}, \code{"gaussian"}, \code{"loggaussian"},
\code{"logistic"}, \code{"loglogistic"}, \code{"lognormal"},
\code{"rayleigh"}, \code{"t"}, or \code{"weibull"}.  Defaults to the
distribution that was used in predicting mean survival times.}

\item{pool}{logical indicating whether to compute a single calibration curve
on predictions pooled over all resampling iterations or to compute them for
each iteration individually and return the mean calibration curve. Pooling
can result in large memory allocation errors when fitting smooth curves
with \code{breaks = NULL}. The current default is changed from versions
<= 3.8.0 of the package which only implemented \code{pool = TRUE}.}

\item{na.rm}{logical indicating whether to remove observed or predicted
responses that are \code{NA} when calculating metrics.}

\item{...}{arguments passed to other methods.}
}
\value{
\code{Calibration} class object that inherits from \code{data.frame}.
}
\description{
Calculate calibration estimates from observed and predicted responses.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

library(survival)

control <- CVControl() \%>\% set_predict(times = c(90, 180, 360))
res <- resample(Surv(time, status) ~ ., data = veteran, model = GBMModel,
                control = control)
cal <- calibration(res)
plot(cal)
}

}
\seealso{
\code{\link{c}}, \code{\link{plot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rfe.R
\name{rfe}
\alias{rfe}
\alias{rfe.formula}
\alias{rfe.matrix}
\alias{rfe.ModelFrame}
\alias{rfe.recipe}
\alias{rfe.ModelSpecification}
\alias{rfe.MLModel}
\alias{rfe.MLModelFunction}
\title{Recursive Feature Elimination}
\usage{
rfe(...)

\method{rfe}{formula}(formula, data, model, ...)

\method{rfe}{matrix}(x, y, model, ...)

\method{rfe}{ModelFrame}(input, model, ...)

\method{rfe}{recipe}(input, model, ...)

\method{rfe}{ModelSpecification}(
  object,
  select = NULL,
  control = MachineShop::settings("control"),
  props = 4,
  sizes = integer(),
  random = FALSE,
  recompute = TRUE,
  optimize = c("global", "local"),
  samples = c(rfe = 1, varimp = 1),
  metrics = NULL,
  stat = c(resample = MachineShop::settings("stat.Resample"), permute =
    MachineShop::settings("stat.TrainingParams")),
  progress = FALSE,
  ...
)

\method{rfe}{MLModel}(model, ...)

\method{rfe}{MLModelFunction}(model, ...)
}
\arguments{
\item{...}{arguments passed from the generic function to its methods, from
the \code{MLModel} and \code{MLModelFunction} methods to first arguments of
others, and from others to the \code{ModelSpecification} method.  The
first argument of each \code{fit} method is positional and, as such, must
be given first in calls to them.}

\item{formula, data}{\link[=formula]{formula} defining the model predictor and
response variables and a \link[=data.frame]{data frame} containing them.}

\item{model}{\link[=models]{model} function, function name, or object; or
another object that can be \link[=as.MLModel]{coerced} to a model.  A model
can be given first followed by any of the variable specifications.}

\item{x, y}{\link{matrix} and object containing predictor and response
variables.}

\item{input}{\link[=inputs]{input} object defining and containing the model
predictor and response variables.}

\item{object}{model \link[=inputs]{input} or
\link[=ModelSpecification]{specification}.}

\item{select}{expression indicating predictor variables that can be
eliminated (see \code{\link[base]{subset}} for syntax) [default: all].}

\item{control}{\link[=controls]{control} function, function name, or object
defining the resampling method to be employed.}

\item{props}{numeric vector of the proportions of most important predictor
variables to retain in fitted models or an integer number of equal spaced
proportions to generate automatically; ignored if \code{sizes} are given.}

\item{sizes}{integer vector of the set sizes of most important predictor
variables to retain.}

\item{random}{logical indicating whether to eliminate variables at random
with probabilities proportional to their importance.}

\item{recompute}{logical indicating whether to recompute variable importance
after eliminating each set of variables.}

\item{optimize}{character string specifying a search through all \code{props}
to identify the globally optimal model (\code{"global"}) or a search that
stops after identifying the first locally optimal model (\code{"local"}).}

\item{samples}{numeric vector or list giving the number of permutation
samples for each of the \code{rfe} and \code{\link{varimp}} algorithms.
One or both of the values may be specified as named arguments or in the
order in which their defaults appear.  Larger numbers of samples decrease
variability in estimated model performances and variable importances at the
expense of increased computation time.  Samples are more expensive
computationally for \code{rfe} than for \code{varimp}.}

\item{metrics}{\link[=metrics]{metric} function, function name, or vector of
these with which to calculate performance.  If not specified, default
metrics defined in the \link{performance} functions are used.}

\item{stat}{functions or character strings naming functions to compute
summary statistics on resampled metric values and permuted samples.  One or
both of the values may be specified as named arguments or in the order in
which their defaults appear.}

\item{progress}{logical indicating whether to display iterative progress
during elimination.}
}
\value{
\code{TrainingStep} class object containing a summary of the numbers
of predictor variables retained (size), their names (terms), logical
indicators for the optimal model selected (selected), and associated
performance metrics (metrics).
}
\description{
A wrapper method of backward feature selection in which a given model is fit
to nested subsets of most important predictor variables in order to select
the subset whose resampled predictive performance is optimal.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

(res <- rfe(sale_amount ~ ., data = ICHomes, model = GBMModel))
summary(res)
summary(performance(res))
plot(res, type = "line")
}

}
\seealso{
\code{\link{performance}}, \code{\link{plot}},
\code{\link{summary}}, \code{\link{varimp}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_GLMModel.R
\name{GLMModel}
\alias{GLMModel}
\alias{GLMStepAICModel}
\title{Generalized Linear Model}
\usage{
GLMModel(family = NULL, quasi = FALSE, ...)

GLMStepAICModel(
  family = NULL,
  quasi = FALSE,
  ...,
  direction = c("both", "backward", "forward"),
  scope = list(),
  k = 2,
  trace = FALSE,
  steps = 1000
)
}
\arguments{
\item{family}{optional error distribution and link function to be used in the
model.  Set automatically according to the class type of the response
variable.}

\item{quasi}{logical indicator for over-dispersion of binomial and Poisson
families; i.e., dispersion parameters not fixed at one.}

\item{...}{arguments passed to \code{\link[stats]{glm.control}}.}

\item{direction}{mode of stepwise search, can be one of \code{"both"}
(default), \code{"backward"}, or \code{"forward"}.}

\item{scope}{defines the range of models examined in the stepwise search.
This should be a list containing components \code{upper} and \code{lower},
both formulae.}

\item{k}{multiple of the number of degrees of freedom used for the penalty.
Only \code{k = 2} gives the genuine AIC; \code{k = .(log(nobs))} is
sometimes referred to as BIC or SBC.}

\item{trace}{if positive, information is printed during the running of
\code{stepAIC}. Larger values may give more information on the fitting
process.}

\item{steps}{maximum number of steps to be considered.}
}
\value{
\code{MLModel} class object.
}
\description{
Fits generalized linear models, specified by giving a symbolic description of
the linear predictor and a description of the error distribution.
}
\details{
\describe{
  \item{\code{GLMModel} Response types:}{\code{BinomialVariate},
    \code{factor}, \code{matrix}, \code{NegBinomialVariate},
    \code{numeric}, \code{PoissonVariate}}
  \item{\code{GLMStepAICModel} Response types:}{\code{binary factor},
    \code{BinomialVariate}, \code{NegBinomialVariate}, \code{numeric},
    \code{PoissonVariate}}
}

Default argument values and further model details can be found in the source
See Also links below.

In calls to \code{\link{varimp}} for \code{GLMModel} and
\code{GLMStepAICModel}, numeric argument \code{base} may be specified for the
(negative) logarithmic transformation of p-values [defaul: \code{exp(1)}].
Transformed p-values are automatically scaled in the calculation of variable
importance to range from 0 to 100.  To obtain unscaled importance values, set
\code{scale = FALSE}.
}
\examples{
fit(sale_amount ~ ., data = ICHomes, model = GLMModel)

}
\seealso{
\code{\link[stats]{glm}}, \code{\link[stats]{glm.control}},
\code{\link[MASS]{stepAIC}}, \code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/case_comps.R
\name{case_weights}
\alias{case_weights}
\title{Extract Case Weights}
\usage{
case_weights(object, newdata = NULL)
}
\arguments{
\item{object}{model \link{fit} result, \code{\link{ModelFrame}}, or
\code{\link[recipes]{recipe}}.}

\item{newdata}{dataset from which to extract the weights if given; otherwise,
\code{object} is used.  The dataset should be given as a \code{ModelFrame}
or as a \link[=data.frame]{data frame} if \code{object} contains a
\code{ModelFrame} or a \code{recipe}, respectively.}
}
\description{
Extract the case weights from an object.
}
\examples{
## Training and test sets
inds <- sample(nrow(ICHomes), nrow(ICHomes) * 2 / 3)
trainset <- ICHomes[inds, ]
testset <- ICHomes[-inds, ]

## ModelFrame case weights
trainmf <- ModelFrame(sale_amount ~ . - built, data = trainset, weights = built)
testmf <- ModelFrame(formula(trainmf), data = testset, weights = built)
mf_fit <- fit(trainmf, model = GLMModel)
rmse(response(mf_fit, testmf), predict(mf_fit, testmf),
     case_weights(mf_fit, testmf))

## Recipe case weights
library(recipes)
rec <- recipe(sale_amount ~ ., data = trainset) \%>\%
  role_case(weight = built, replace = TRUE)
rec_fit <- fit(rec, model = GLMModel)
rmse(response(rec_fit, testset), predict(rec_fit, testset),
     case_weights(rec_fit, testset))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_RangerModel.R
\name{RangerModel}
\alias{RangerModel}
\title{Fast Random Forest Model}
\usage{
RangerModel(
  num.trees = 500,
  mtry = integer(),
  importance = c("impurity", "impurity_corrected", "permutation"),
  min.node.size = integer(),
  replace = TRUE,
  sample.fraction = if (replace) 1 else 0.632,
  splitrule = character(),
  num.random.splits = 1,
  alpha = 0.5,
  minprop = 0.1,
  split.select.weights = numeric(),
  always.split.variables = character(),
  respect.unordered.factors = character(),
  scale.permutation.importance = FALSE,
  verbose = FALSE
)
}
\arguments{
\item{num.trees}{number of trees.}

\item{mtry}{number of variables to possibly split at in each node.}

\item{importance}{variable importance mode.}

\item{min.node.size}{minimum node size.}

\item{replace}{logical indicating whether to sample with replacement.}

\item{sample.fraction}{fraction of observations to sample.}

\item{splitrule}{splitting rule.}

\item{num.random.splits}{number of random splits to consider for each
candidate splitting variable in the \code{"extratrees"} rule.}

\item{alpha}{significance threshold to allow splitting in the
\code{"maxstat"} rule.}

\item{minprop}{lower quantile of covariate distribution to be considered for
splitting in the \code{"maxstat"} rule.}

\item{split.select.weights}{numeric vector with weights between 0 and 1,
representing the probability to select variables for splitting.}

\item{always.split.variables}{character vector with variable names to be
always selected in addition to the \code{mtry} variables tried for
splitting.}

\item{respect.unordered.factors}{handling of unordered factor covariates.}

\item{scale.permutation.importance}{scale permutation importance by
standard error.}

\item{verbose}{show computation status and estimated runtime.}
}
\value{
\code{MLModel} class object.
}
\description{
Fast implementation of random forests or recursive partitioning.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{numeric}, \code{Surv}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \code{mtry}, \code{min.node.size}*, \code{splitrule}*
  }
}
* excluded from grids by default

Default argument values and further model details can be found in the source
See Also link below.
}
\examples{
\donttest{
## Requires prior installation of suggested package ranger to run

fit(Species ~ ., data = iris, model = RangerModel)
}

}
\seealso{
\code{\link[ranger]{ranger}}, \code{\link{fit}},
\code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelFrame.R
\name{ModelFrame}
\alias{ModelFrame}
\alias{ModelFrame.formula}
\alias{ModelFrame.matrix}
\title{ModelFrame Class}
\usage{
ModelFrame(...)

\method{ModelFrame}{formula}(
  formula,
  data,
  groups = NULL,
  strata = NULL,
  weights = NULL,
  na.rm = TRUE,
  ...
)

\method{ModelFrame}{matrix}(
  x,
  y = NULL,
  offsets = NULL,
  groups = NULL,
  strata = NULL,
  weights = NULL,
  na.rm = TRUE,
  ...
)
}
\arguments{
\item{...}{arguments passed from the generic function to its methods.  The
first argument of each \code{ModelFrame} method is positional and, as such,
must be given first in calls to them.}

\item{formula, data}{\link[=formula]{formula} defining the model predictor and
response variables and a \link[=data.frame]{data frame} containing them.
In the associated method, arguments \code{groups}, \code{strata}, and
\code{weights} will be evaluated as expressions, whose objects are searched
for first in the accompanying \code{data} environment and, if not found
there, next in the calling environment.}

\item{groups}{vector of values defining groupings of case observations, such
as repeated measurements, to keep together during resampling [default:
none].}

\item{strata}{vector of values to use in conducting stratified
\link{resample} estimation of model performance [default: none].}

\item{weights}{numeric vector of non-negative case weights for the \code{y}
response variable [default: equal weights].}

\item{na.rm}{character string or logical specifying removal of \code{"all"}
(\code{TRUE}) cases with missing values, \code{"none"} (\code{FALSE}), or
only those whose missing values are in the \code{"response"} variable.}

\item{x, y}{\link{matrix} and object containing predictor and response
variables.}

\item{offsets}{numeric vector, matrix, or data frame of values to be added
with a fixed coefficient of 1 to linear predictors in compatible regression
models.}
}
\value{
\code{ModelFrame} class object that inherits from \code{data.frame}.
}
\description{
Class for storing data, formulas, and other attributes for \pkg{MachineShop}
model fitting.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

mf <- ModelFrame(ncases / (ncases + ncontrols) ~ agegp + tobgp + alcgp,
                 data = esoph, weights = ncases + ncontrols)
gbm_fit <- fit(mf, model = GBMModel)
varimp(gbm_fit)
}

}
\seealso{
\code{\link{fit}}, \code{\link{resample}}, \code{\link{response}},
\code{\link{SelectedInput}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_GAMBoostModel.R
\name{GAMBoostModel}
\alias{GAMBoostModel}
\title{Gradient Boosting with Additive Models}
\usage{
GAMBoostModel(
  family = NULL,
  baselearner = c("bbs", "bols", "btree", "bss", "bns"),
  dfbase = 4,
  mstop = 100,
  nu = 0.1,
  risk = c("inbag", "oobag", "none"),
  stopintern = FALSE,
  trace = FALSE
)
}
\arguments{
\item{family}{optional \code{\link[mboost]{Family}} object.  Set
automatically according to the class type of the response variable.}

\item{baselearner}{character specifying the component-wise
\code{\link[mboost:baselearners]{base learner}} to be used.}

\item{dfbase}{gobal degrees of freedom for P-spline base learners
(\code{"bbs"}).}

\item{mstop}{number of initial boosting iterations.}

\item{nu}{step size or shrinkage parameter between 0 and 1.}

\item{risk}{method to use in computing the empirical risk for each boosting
iteration.}

\item{stopintern}{logical inidicating whether the boosting algorithm stops
internally when the out-of-bag risk increases at a subsequent iteration.}

\item{trace}{logical indicating whether status information is printed during
the fitting process.}
}
\value{
\code{MLModel} class object.
}
\description{
Gradient boosting for optimizing arbitrary loss functions, where
component-wise arbitrary base-learners, e.g., smoothing procedures, are
utilized as additive base-learners.
}
\details{
\describe{
  \item{Response types:}{\code{binary factor}, \code{BinomialVariate},
    \code{NegBinomialVariate}, \code{numeric}, \code{PoissonVariate},
    \code{Surv}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameter:}{
    \code{mstop}
  }
}

Default argument values and further model details can be found in the source
See Also links below.
}
\examples{
\donttest{
## Requires prior installation of suggested package mboost to run

data(Pima.tr, package = "MASS")

fit(type ~ ., data = Pima.tr, model = GAMBoostModel)
}

}
\seealso{
\code{\link[mboost]{gamboost}}, \code{\link[mboost]{Family}},
\code{\link[mboost]{baselearners}}, \code{\link{fit}},
\code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModelSpecification.R
\name{ModelSpecification}
\alias{ModelSpecification}
\alias{ModelSpecification.default}
\alias{ModelSpecification.formula}
\alias{ModelSpecification.matrix}
\alias{ModelSpecification.ModelFrame}
\alias{ModelSpecification.recipe}
\title{Model Specification}
\usage{
ModelSpecification(...)

\method{ModelSpecification}{default}(
  input,
  model,
  control = MachineShop::settings("control"),
  metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams"),
  ...
)

\method{ModelSpecification}{formula}(formula, data, model, ...)

\method{ModelSpecification}{matrix}(x, y, model, ...)

\method{ModelSpecification}{ModelFrame}(input, model, ...)

\method{ModelSpecification}{recipe}(input, model, ...)
}
\arguments{
\item{...}{arguments passed from the generic function to its methods.  The
first argument of each \code{ModelSpecification} method is positional and,
as such, must be given first in calls to them.}

\item{input}{\link[=inputs]{input} object defining and containing the model
predictor and response variables.}

\item{model}{\link[=models]{model} function, function name, or object; or
another object that can be \link[=as.MLModel]{coerced} to a model.}

\item{control}{\link[=controls]{control} function, function name, or object
defining the resampling method to be employed.  If \code{NULL} or if
the model specification contains any \code{SelectedInput} or
\code{SelectedModel} objects, then object-specific control structures and
training parameters are used for selection and tuning, as usual, and
objects are trained sequentially with nested resampling.  Otherwise,
\itemize{
  \item tuning of input and model objects is performed simultaneously over
    a global grid of their parameter values, and
  \item the specified \code{control} method and training parameters below
    override those of any included \code{TunedInput} or \code{TunedModel}.
}}

\item{metrics}{\link[=metrics]{metric} function, function name, or vector of
these with which to calculate performance.  If not specified, default
metrics defined in the \link{performance} functions are used.  Model
selection is based on the first calculated metric.}

\item{cutoff}{argument passed to the \code{metrics} functions.}

\item{stat}{function or character string naming a function to compute a
summary statistic on resampled metric values for model tuning.}

\item{formula, data}{\link[=formula]{formula} defining the model predictor and
response variables and a \link[=data.frame]{data frame} containing them.}

\item{x, y}{\link{matrix} and object containing predictor and response
variables.}
}
\value{
\code{ModelSpecification} class object.
}
\description{
Specification of a relationship between response and predictor variables and
a model to define a relationship between them.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

modelspec <- ModelSpecification(
  sale_amount ~ ., data = ICHomes, model = GBMModel
)
fit(modelspec)
}

}
\seealso{
\code{\link{fit}}, \code{\link{resample}},
\code{\link{set_monitor}}, \code{\link{set_optim}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/performance_curve.R
\name{performance_curve}
\alias{performance_curve}
\alias{curves}
\alias{performance_curve.default}
\alias{performance_curve.Resample}
\title{Model Performance Curves}
\usage{
performance_curve(x, ...)

\method{performance_curve}{default}(
  x,
  y,
  weights = NULL,
  metrics = c(MachineShop::tpr, MachineShop::fpr),
  na.rm = TRUE,
  ...
)

\method{performance_curve}{Resample}(
  x,
  metrics = c(MachineShop::tpr, MachineShop::fpr),
  na.rm = TRUE,
  ...
)
}
\arguments{
\item{x}{\link[=response]{observed responses} or \link{resample} result
containing observed and predicted responses.}

\item{...}{arguments passed to other methods.}

\item{y}{\link[=predict]{predicted responses} if not contained in \code{x}.}

\item{weights}{numeric vector of non-negative
\link[=case_weights]{case weights} for the observed \code{x} responses
[default: equal weights].}

\item{metrics}{list of two performance \link{metrics} for the analysis
[default: ROC metrics].  Precision recall curves can be obtained with
\code{c(precision, recall)}.}

\item{na.rm}{logical indicating whether to remove observed or predicted
responses that are \code{NA} when calculating metrics.}
}
\value{
\code{PerformanceCurve} class object that inherits from
\code{data.frame}.
}
\description{
Calculate curves for the analysis of tradeoffs between metrics for assessing
performance in classifying binary outcomes over the range of possible
cutoff probabilities.  Available curves include receiver operating
characteristic (ROC) and precision recall.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

data(Pima.tr, package = "MASS")

res <- resample(type ~ ., data = Pima.tr, model = GBMModel)

## ROC curve
roc <- performance_curve(res)
plot(roc)
auc(roc)
}

}
\seealso{
\code{\link{auc}}, \code{\link{c}}, \code{\link{plot}},
\code{\link{summary}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_ParsnipModel.R
\name{ParsnipModel}
\alias{ParsnipModel}
\title{Parsnip Model}
\usage{
ParsnipModel(object, ...)
}
\arguments{
\item{object}{\link[parsnip:model_spec]{model specification} from the
\pkg{parsnip} package.}

\item{...}{tuning parameters with which to update \code{object}.}
}
\value{
\code{ParsnipModel} class object that inherits from \code{MLModel}.
}
\description{
Convert a model specification from the \pkg{parsnip} package to one that can
be used with the \pkg{MachineShop} package.
}
\examples{
\donttest{
## Requires prior installation of suggested package parsnip to run

prsp_model <- parsnip::linear_reg(engine = "glmnet")

model <- ParsnipModel(prsp_model, penalty = 1, mixture = 1)
model

model_fit <- fit(sale_amount ~ ., data = ICHomes, model = model)
predict(model_fit)
}

}
\seealso{
\code{\link{as.MLModel}}, \code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diff.R
\name{diff}
\alias{diff}
\alias{diff.MLModel}
\alias{diff.Performance}
\alias{diff.Resample}
\title{Model Performance Differences}
\usage{
\method{diff}{MLModel}(x, ...)

\method{diff}{Performance}(x, ...)

\method{diff}{Resample}(x, ...)
}
\arguments{
\item{x}{model \link{performance} or \link{resample} result.}

\item{...}{arguments passed to other methods.}
}
\value{
\code{PerformanceDiff} class object that inherits from
\code{Performance}.
}
\description{
Pairwise model differences in resampled performance metrics.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

## Survival response example
library(survival)

fo <- Surv(time, status) ~ .
control <- CVControl()

gbm_res1 <- resample(fo, data = veteran, GBMModel(n.trees = 25), control)
gbm_res2 <- resample(fo, data = veteran, GBMModel(n.trees = 50), control)
gbm_res3 <- resample(fo, data = veteran, GBMModel(n.trees = 100), control)

res <- c(GBM1 = gbm_res1, GBM2 = gbm_res2, GBM3 = gbm_res3)
res_diff <- diff(res)
summary(res_diff)
plot(res_diff)
}

}
\seealso{
\code{\link{t.test}}, \code{\link{plot}}, \code{\link{summary}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expand.R
\name{expand_modelgrid}
\alias{expand_modelgrid}
\alias{expand_modelgrid.formula}
\alias{expand_modelgrid.matrix}
\alias{expand_modelgrid.ModelFrame}
\alias{expand_modelgrid.recipe}
\alias{expand_modelgrid.ModelSpecification}
\alias{expand_modelgrid.MLModel}
\alias{expand_modelgrid.MLModelFunction}
\title{Model Tuning Grid Expansion}
\usage{
expand_modelgrid(...)

\method{expand_modelgrid}{formula}(formula, data, model, info = FALSE, ...)

\method{expand_modelgrid}{matrix}(x, y, model, info = FALSE, ...)

\method{expand_modelgrid}{ModelFrame}(input, model, info = FALSE, ...)

\method{expand_modelgrid}{recipe}(input, model, info = FALSE, ...)

\method{expand_modelgrid}{ModelSpecification}(object, ...)

\method{expand_modelgrid}{MLModel}(model, ...)

\method{expand_modelgrid}{MLModelFunction}(model, ...)
}
\arguments{
\item{...}{arguments passed from the generic function to its methods and from
the \code{MLModel} and \code{MLModelFunction} methods to others.  The
first argument of each \code{expand_modelgrid} method is positional and, as
such, must be given first in calls to them.}

\item{formula, data}{\link[=formula]{formula} defining the model predictor and
response variables and a \link[=data.frame]{data frame} containing them.}

\item{model}{\link[=models]{model} function, function name, or object; or
another object that can be \link[=as.MLModel]{coerced} to a model.  A model
can be given first followed by any of the variable specifications.}

\item{info}{logical indicating whether to return model-defined grid
construction information rather than the grid values.}

\item{x, y}{\link{matrix} and object containing predictor and response
variables.}

\item{input}{\link[=inputs]{input} object defining and containing the model
predictor and response variables.}

\item{object}{model \link[=ModelSpecification]{specification}.}
}
\value{
A data frame of parameter values or \code{NULL} if data are required
for construction of the grid but not supplied.
}
\description{
Expand a model grid of tuning parameter values.
}
\details{
The \code{expand_modelgrid} function enables manual extraction and viewing of
grids created automatically when a \code{\link{TunedModel}} is fit.
}
\examples{
expand_modelgrid(TunedModel(GBMModel, grid = 5))

\donttest{
## Requires prior installation of suggested package glmnet to run
expand_modelgrid(TunedModel(GLMNetModel, grid = c(alpha = 5, lambda = 10)),
                 sale_amount ~ ., data = ICHomes)
}

gbm_grid <- ParameterGrid(
  n.trees = dials::trees(),
  interaction.depth = dials::tree_depth(),
  size = 5
)
expand_modelgrid(TunedModel(GBMModel, grid = gbm_grid))

rf_grid <- ParameterGrid(
  mtry = dials::mtry(),
  nodesize = dials::max_nodes(),
  size = c(3, 5)
)
expand_modelgrid(TunedModel(RandomForestModel, grid = rf_grid),
                 sale_amount ~ ., data = ICHomes)

}
\seealso{
\code{\link{TunedModel}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_C50Model.R
\name{C50Model}
\alias{C50Model}
\title{C5.0 Decision Trees and Rule-Based Model}
\usage{
C50Model(
  trials = 1,
  rules = FALSE,
  subset = TRUE,
  bands = 0,
  winnow = FALSE,
  noGlobalPruning = FALSE,
  CF = 0.25,
  minCases = 2,
  fuzzyThreshold = FALSE,
  sample = 0,
  earlyStopping = TRUE
)
}
\arguments{
\item{trials}{integer number of boosting iterations.}

\item{rules}{logical indicating whether to decompose the tree into a
rule-based model.}

\item{subset}{logical indicating whether the model should evaluate groups of
discrete predictors for splits.}

\item{bands}{integer between 2 and 1000 specifying a number of bands into
which to group rules ordered by their affect on the error rate.}

\item{winnow}{logical indicating use of predictor winnowing (i.e. feature
selection).}

\item{noGlobalPruning}{logical indicating a final, global pruning step to
simplify the tree.}

\item{CF}{number in (0, 1) for the confidence factor.}

\item{minCases}{integer for the smallest number of samples that must be put
in at least two of the splits.}

\item{fuzzyThreshold}{logical indicating whether to evaluate possible
advanced splits of the data.}

\item{sample}{value between (0, 0.999) that specifies the random proportion
of data to use in training the model.}

\item{earlyStopping}{logical indicating whether the internal method for
stopping boosting should be used.}
}
\value{
\code{MLModel} class object.
}
\description{
Fit classification tree models or rule-based models using Quinlan's C5.0
algorithm.
}
\details{
\describe{
  \item{Response types:}{\code{factor}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \code{trials}, \code{rules}, \code{winnow}
  }
}

Latter arguments are passed to \code{\link[C50]{C5.0Control}}.
Further model details can be found in the source link below.

In calls to \code{\link{varimp}} for \code{C50Model}, argument \code{type}
may be specified as \code{"usage"} (default) for the percentage of training
set samples that fall into all terminal nodes after the split of each
predictor or as \code{"splits"} for the percentage of splits associated with
each predictor.  Variable importance is automatically scaled to range from 0
to 100.  To obtain unscaled importance values, set \code{scale = FALSE}.  See
example below.
}
\examples{
\donttest{
## Requires prior installation of suggested package C50 to run

model_fit <- fit(Species ~ ., data = iris, model = C50Model)
varimp(model_fit, method = "model", type = "splits", scale = FALSE)
}

}
\seealso{
\code{\link[C50]{C5.0}}, \code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_TreeModel.R
\name{TreeModel}
\alias{TreeModel}
\title{Classification and Regression Tree Models}
\usage{
TreeModel(
  mincut = 5,
  minsize = 10,
  mindev = 0.01,
  split = c("deviance", "gini"),
  k = numeric(),
  best = integer(),
  method = c("deviance", "misclass")
)
}
\arguments{
\item{mincut}{minimum number of observations to include in either child node.}

\item{minsize}{smallest allowed node size: a weighted quantity.}

\item{mindev}{within-node deviance must be at least this times that of the
root node for the node to be split.}

\item{split}{splitting criterion to use.}

\item{k}{scalar cost-complexity parameter defining a subtree to return.}

\item{best}{integer alternative to \code{k} requesting the number of terminal
nodes of a subtree in the cost-complexity sequence to return.}

\item{method}{character string denoting the measure of node heterogeneity
used to guide cost-complexity pruning.}
}
\value{
\code{MLModel} class object.
}
\description{
A tree is grown by binary recursive partitioning using the response in the
specified formula and choosing splits from the terms of the right-hand-side.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{numeric}}
}

Further model details can be found in the source link below.
}
\examples{
\donttest{
## Requires prior installation of suggested package tree to run

fit(Species ~ ., data = iris, model = TreeModel)
}

}
\seealso{
\code{\link[tree]{tree}}, \code{\link[tree]{prune.tree}},
\code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_MDAModel.R
\name{MDAModel}
\alias{MDAModel}
\title{Mixture Discriminant Analysis Model}
\usage{
MDAModel(
  subclasses = 3,
  sub.df = numeric(),
  tot.df = numeric(),
  dimension = sum(subclasses) - 1,
  eps = .Machine$double.eps,
  iter = 5,
  method = .(mda::polyreg),
  trace = FALSE,
  ...
)
}
\arguments{
\item{subclasses}{numeric value or vector of subclasses per class.}

\item{sub.df}{effective degrees of freedom of the centroids per class if
subclass centroid shrinkage is performed.}

\item{tot.df}{specification of the total degrees of freedom as an alternative
to \code{sub.df}.}

\item{dimension}{dimension of the discriminant subspace to use for
prediction.}

\item{eps}{numeric threshold for automatically truncating the dimension.}

\item{iter}{limit on the total number of iterations.}

\item{method}{regression function used in optimal scaling.  The default of
linear regression is provided by \code{\link[mda]{polyreg}} from the
\pkg{mda} package.  For penalized mixture discriminant models,
\code{\link[mda]{gen.ridge}} is appropriate.  Other possibilities are
\code{\link[mda]{mars}} for multivariate adaptive regression splines and
\code{\link[mda]{bruto}} for adaptive backfitting of additive splines.  Use
the \code{\link[=quote]{.}} operator to quote specified functions.}

\item{trace}{logical indicating whether iteration information is printed.}

\item{...}{additional arguments to \code{mda.start} and \code{method}.}
}
\value{
\code{MLModel} class object.
}
\description{
Performs mixture discriminant analysis.
}
\details{
\describe{
  \item{Response types:}{\code{factor}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameter:}{
    \code{subclasses}
  }
}

The \code{\link{predict}} function for this model additionally accepts the
following argument.
\describe{
  \item{\code{prior}}{prior class membership probabilities for prediction data
    if different from the training set.}
}

Default argument values and further model details can be found in the source
See Also links below.
}
\examples{
\donttest{
## Requires prior installation of suggested package mda to run

fit(Species ~ ., data = iris, model = MDAModel)
}

}
\seealso{
\code{\link[mda]{mda}}, \code{\link[mda]{predict.mda}},
\code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_BlackBoostModel.R
\name{BlackBoostModel}
\alias{BlackBoostModel}
\title{Gradient Boosting with Regression Trees}
\usage{
BlackBoostModel(
  family = NULL,
  mstop = 100,
  nu = 0.1,
  risk = c("inbag", "oobag", "none"),
  stopintern = FALSE,
  trace = FALSE,
  teststat = c("quadratic", "maximum"),
  testtype = c("Teststatistic", "Univariate", "Bonferroni", "MonteCarlo"),
  mincriterion = 0,
  minsplit = 10,
  minbucket = 4,
  maxdepth = 2,
  saveinfo = FALSE,
  ...
)
}
\arguments{
\item{family}{optional \code{\link[mboost]{Family}} object.  Set
automatically according to the class type of the response variable.}

\item{mstop}{number of initial boosting iterations.}

\item{nu}{step size or shrinkage parameter between 0 and 1.}

\item{risk}{method to use in computing the empirical risk for each boosting
iteration.}

\item{stopintern}{logical inidicating whether the boosting algorithm stops
internally when the out-of-bag risk increases at a subsequent iteration.}

\item{trace}{logical indicating whether status information is printed during
the fitting process.}

\item{teststat}{type of the test statistic to be applied for variable
selection.}

\item{testtype}{how to compute the distribution of the test statistic.}

\item{mincriterion}{value of the test statistic or 1 - p-value that must be
exceeded in order to implement a split.}

\item{minsplit}{minimum sum of weights in a node in order to be considered
for splitting.}

\item{minbucket}{minimum sum of weights in a terminal node.}

\item{maxdepth}{maximum depth of the tree.}

\item{saveinfo}{logical indicating whether to store information about
variable selection in \code{info} slot of each \code{partynode}.}

\item{...}{additional arguments to \code{\link[partykit]{ctree_control}}.}
}
\value{
\code{MLModel} class object.
}
\description{
Gradient boosting for optimizing arbitrary loss functions where regression
trees are utilized as base-learners.
}
\details{
\describe{
  \item{Response types:}{\code{binary factor}, \code{BinomialVariate},
    \code{NegBinomialVariate}, \code{numeric}, \code{PoissonVariate},
    \code{Surv}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \code{mstop}, \code{maxdepth}
  }
}

Default argument values and further model details can be found in the source
See Also links below.
}
\examples{
\donttest{
## Requires prior installation of suggested packages mboost and partykit to run

data(Pima.tr, package = "MASS")

fit(type ~ ., data = Pima.tr, model = BlackBoostModel)
}

}
\seealso{
\code{\link[mboost]{blackboost}}, \code{\link[mboost]{Family}},
\code{\link[partykit]{ctree_control}}, \code{\link{fit}},
\code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\name{settings}
\alias{settings}
\title{MachineShop Settings}
\usage{
settings(...)
}
\arguments{
\item{...}{character names of settings to view, \code{name = value} pairs
giving the values of settings to change, a vector of these, \code{"reset"}
to restore all package defaults, or no arguments to view all settings.
Partial matching of setting names is supported.}
}
\value{
The setting value if only one is specified to view.  Otherwise, a
list of the values of specified settings as they existed prior to any
requested changes.  Such a list can be passed as an argument to
\code{settings} to restore their values.
}
\description{
Allow the user to view or change global settings which affect default
behaviors of functions in the \pkg{MachineShop} package.
}
\section{Settings}{


\describe{
  \item{\code{\link[=controls]{control}}}{function, function name, or object
    defining a default resampling method [default: \code{"CVControl"}].}
  \item{\code{cutoff}}{numeric (0, 1) threshold above which binary factor
    probabilities are classified as events and below which survival
    probabilities are classified [default: 0.5].}
  \item{\code{distr.SurvMeans}}{character string specifying distributional
    approximations to estimated survival curves for predicting survival
    means.  Choices are \code{"empirical"} for the Kaplan-Meier estimator,
    \code{"exponential"}, \code{"rayleigh"}, or \code{"weibull"} (default).}
  \item{\code{distr.SurvProbs}}{character string specifying distributional
    approximations to estimated survival curves for predicting survival
    events/probabilities.  Choices are \code{"empirical"} (default) for the
    Kaplan-Meier estimator, \code{"exponential"}, \code{"rayleigh"}, or
    \code{"weibull"}.}
  \item{\code{grid}}{\code{size} argument to \code{\link{TuningGrid}}
    indicating the number of parameter-specific values to generate
    automatically for \link[=TunedModel]{tuning} of models that have
    pre-defined grids or a \code{\link{TuningGrid}} function, function name,
    or object [default: 3].}
  \item{\code{method.EmpiricalSurv}}{character string specifying the
    empirical method of estimating baseline survival curves for Cox
    proportional hazards-based models.  Choices are \code{"breslow"} or
    \code{"efron"} (default).}
  \item{\code{metrics.ConfusionMatrix}}{function, function name, or vector of
    these with which to calculate \link{performance} \link{metrics} for
    confusion matrices [default: \code{c(Accuracy = "accuracy", Kappa =
    "kappa2", `Weighted Kappa` = "weighted_kappa2", Sensitivity =
    "sensitivity", Specificity = "specificity")}].}
  \item{\code{metrics.factor}}{function, function name, or vector of these
    with which to calculate \link{performance} \link{metrics} for factor
    responses [default: \code{c(Brier = "brier", Accuracy = "accuracy",
    Kappa = "kappa2", `Weighted Kappa` = "weighted_kappa2", `ROC AUC` =
    "roc_auc", Sensitivity = "sensitivity", Specificity = "specificity")}].}
  \item{\code{metrics.matrix}}{function, function name, or vector of these
    with which to calculate \link{performance} \link{metrics} for matrix
    responses [default: \code{c(RMSE = "rmse", R2 = "r2", MAE = "mae")}].}
  \item{\code{metrics.numeric}}{function, function name, or vector of these
    with which to calculate \link{performance} \link{metrics} for numeric
    responses [default: \code{c(RMSE = "rmse", R2 = "r2", MAE = "mae")}].}
  \item{\code{metrics.Surv}}{function, function name, or vector of these with
    which to calculate \link{performance} \link{metrics} for survival
    responses [default: \code{c(`C-Index` = "cindex", Brier = "brier",
    `ROC AUC` = "roc_auc", Accuracy = "accuracy")}].}
  \item{\code{print_max}}{number of models or data rows to show with print
    methods or \code{Inf} to show all [default: 10].}
  \item{\code{require}}{names of installed packages to load during parallel
    execution of resampling algorithms [default: \code{"MachineShop"}].}
  \item{\code{reset}}{character names of settings to reset to their default
    values.}
  \item{\code{RHS.formula}}{non-modifiable character vector of operators and
    functions allowed in traditional formula specifications.}
  \item{\code{stat.Curve}}{function or character string naming a function
    to compute one \link{summary} statistic at each cutoff value of resampled
    metrics in performance curves, or \code{NULL} for resample-specific
    metrics [default: \code{"base::mean"}].}
  \item{\code{stat.Resample}}{function or character string naming a function
    to compute one summary statistic to control the ordering of models in
    \link[=plot]{plots} [default: \code{"base::mean"}].}
  \item{\code{stat.TrainingParams}}{function or character string naming a function
    to compute one summary statistic on resampled performance metrics for
    input \link[=SelectedInput]{selection} or \link[=TunedInput]{tuning} or
    for model \link[=SelectedModel]{selection} or \link[=TunedModel]{tuning}
    [default: \code{"base::mean"}].}
  \item{\code{stats.PartialDependence}}{function, function name, or vector of
    these with which to compute \link[=dependence]{partial dependence}
    summary statistics [default: \code{c(Mean = "base::mean")}].}
  \item{\code{stats.Resample}}{function, function name, or vector of these
    with which to compute \link{summary} statistics on resampled performance
    metrics [default: \code{c(Mean = "base::mean", Median = "stats::median",
    SD = "stats::sd", Min = "base::min", Max = "base::max")}].}
}
}

\examples{
## View all current settings
settings()

## Change settings
presets <- settings(control = "BootControl", grid = 10)

## View one setting
settings("control")

## View multiple settings
settings("control", "grid")

## Restore the previous settings
settings(presets)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dependence.R
\name{dependence}
\alias{dependence}
\title{Partial Dependence}
\usage{
dependence(
  object,
  data = NULL,
  select = NULL,
  interaction = FALSE,
  n = 10,
  intervals = c("uniform", "quantile"),
  distr = character(),
  method = character(),
  stats = MachineShop::settings("stats.PartialDependence"),
  na.rm = TRUE
)
}
\arguments{
\item{object}{model \link{fit} result.}

\item{data}{\link[=data.frame]{data frame} containing all predictor
variables.  If not specified, the training data will be used by default.}

\item{select}{expression indicating predictor variables for which to compute
partial dependence (see \code{\link[base]{subset}} for syntax)
[default: all].}

\item{interaction}{logical indicating whether to calculate dependence on the
interacted predictors.}

\item{n}{number of predictor values at which to perform calculations.}

\item{intervals}{character string specifying whether the \code{n} values are
spaced uniformly (\code{"uniform"}) or according to variable quantiles
(\code{"quantile"}).}

\item{distr, method}{arguments passed to \code{\link{predict}}.}

\item{stats}{function, function name, or vector of these with which to
compute response variable summary statistics over non-selected predictor
variables.}

\item{na.rm}{logical indicating whether to exclude missing predicted response
values from the calculation of summary statistics.}
}
\value{
\code{PartialDependence} class object that inherits from
\code{data.frame}.
}
\description{
Calculate partial dependence of a response on select predictor variables.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

gbm_fit <- fit(Species ~ ., data = iris, model = GBMModel)
(pd <- dependence(gbm_fit, select = c(Petal.Length, Petal.Width)))
plot(pd)
}

}
\seealso{
\code{\link{plot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TrainedModels.R
\name{TunedModel}
\alias{TunedModel}
\title{Tuned Model}
\usage{
TunedModel(
  object,
  grid = MachineShop::settings("grid"),
  control = MachineShop::settings("control"),
  metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
)
}
\arguments{
\item{object}{\link[=models]{model} function, function name, or object
defining the model to be tuned.}

\item{grid}{single integer or vector of integers whose positions or names
match the parameters in the model's pre-defined tuning grid if one exists
and which specify the number of values used to construct the grid;
\code{\link{TuningGrid}} function, function name, or object;
\code{\link{ParameterGrid}} object; or \link[=data.frame]{data frame}
containing parameter values at which to evaluate the model, such as that
returned by \code{\link{expand_params}}.}

\item{control}{\link[=controls]{control} function, function name, or object
defining the resampling method to be employed.}

\item{metrics}{\link[=metrics]{metric} function, function name, or vector of
these with which to calculate performance.  If not specified, default
metrics defined in the \link{performance} functions are used.  Model
selection is based on the first calculated metric.}

\item{cutoff}{argument passed to the \code{metrics} functions.}

\item{stat}{function or character string naming a function to compute a
summary statistic on resampled metric values for model tuning.}
}
\value{
\code{TunedModel} class object that inherits from \code{MLModel}.
}
\description{
Model tuning over a grid of parameter values.
}
\details{
The \code{\link{expand_modelgrid}} function enables manual extraction and
viewing of grids created automatically when a \code{TunedModel} is fit.

\describe{
  \item{Response types:}{\code{factor}, \code{numeric}, \code{ordered},
    \code{Surv}}
}
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run
## May require a long runtime

# Automatically generated grid
model_fit <- fit(sale_amount ~ ., data = ICHomes,
                 model = TunedModel(GBMModel))
varimp(model_fit)
(tuned_model <- as.MLModel(model_fit))
summary(tuned_model)
plot(tuned_model, type = "l")

# Randomly sampled grid points
fit(sale_amount ~ ., data = ICHomes,
    model = TunedModel(
      GBMModel,
      grid = TuningGrid(size = 1000, random = 5)
    ))

# User-specified grid
fit(sale_amount ~ ., data = ICHomes,
    model = TunedModel(
      GBMModel,
      grid = expand_params(
        n.trees = c(50, 100),
        interaction.depth = 1:2,
        n.minobsinnode = c(5, 10)
      )
    ))
}

}
\seealso{
\code{\link{fit}}, \code{\link{resample}}, \code{\link{set_optim}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print}
\alias{print}
\alias{print.BinomialVariate}
\alias{print.Calibration}
\alias{print.DiscreteVariate}
\alias{print.ListOf}
\alias{print.MLControl}
\alias{print.MLMetric}
\alias{print.MLModel}
\alias{print.MLModelFunction}
\alias{print.ModelFrame}
\alias{print.ModelRecipe}
\alias{print.ModelSpecification}
\alias{print.Performance}
\alias{print.PerformanceCurve}
\alias{print.RecipeGrid}
\alias{print.Resample}
\alias{print.SurvMatrix}
\alias{print.SurvTimes}
\alias{print.TrainingStep}
\alias{print.VariableImportance}
\title{Print MachineShop Objects}
\usage{
\method{print}{BinomialVariate}(x, n = MachineShop::settings("print_max"), ...)

\method{print}{Calibration}(x, n = MachineShop::settings("print_max"), ...)

\method{print}{DiscreteVariate}(x, n = MachineShop::settings("print_max"), ...)

\method{print}{ListOf}(x, n = MachineShop::settings("print_max"), ...)

\method{print}{MLControl}(x, n = MachineShop::settings("print_max"), ...)

\method{print}{MLMetric}(x, ...)

\method{print}{MLModel}(x, n = MachineShop::settings("print_max"), id = FALSE, ...)

\method{print}{MLModelFunction}(x, ...)

\method{print}{ModelFrame}(x, n = MachineShop::settings("print_max"), id = FALSE, data = TRUE, ...)

\method{print}{ModelRecipe}(x, n = MachineShop::settings("print_max"), id = FALSE, data = TRUE, ...)

\method{print}{ModelSpecification}(x, n = MachineShop::settings("print_max"), id = FALSE, ...)

\method{print}{Performance}(x, n = MachineShop::settings("print_max"), ...)

\method{print}{PerformanceCurve}(x, n = MachineShop::settings("print_max"), ...)

\method{print}{RecipeGrid}(x, n = MachineShop::settings("print_max"), ...)

\method{print}{Resample}(x, n = MachineShop::settings("print_max"), ...)

\method{print}{SurvMatrix}(x, n = MachineShop::settings("print_max"), ...)

\method{print}{SurvTimes}(x, n = MachineShop::settings("print_max"), ...)

\method{print}{TrainingStep}(x, n = MachineShop::settings("print_max"), ...)

\method{print}{VariableImportance}(x, n = MachineShop::settings("print_max"), ...)
}
\arguments{
\item{x}{object to print.}

\item{n}{integer number of models or data frame rows to show.}

\item{...}{arguments passed to other methods, including the one described
below.
\describe{
  \item{\code{level} = 0}{current nesting level of the corresponding
  object in recursive calls to \code{print}.  The amount of information
  displayed decreases and increases with positive and negative levels,
  respectively.}
}}

\item{id}{logical indicating whether to show object identifiers.}

\item{data}{logical indicating whether to show model data.}
}
\description{
Print methods for objects defined in the \pkg{MachineShop} package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expand.R
\name{expand_params}
\alias{expand_params}
\title{Model Parameters Expansion}
\usage{
expand_params(..., random = FALSE)
}
\arguments{
\item{...}{named data frames or vectors or a list of these containing the
parameter values over which to create the grid.}

\item{random}{number of points to be randomly sampled from the parameter grid
or \code{FALSE} if all points are to be returned.}
}
\value{
A data frame containing one row for each combination of the supplied
inputs.
}
\description{
Create a grid of parameter values from all combinations of supplied inputs.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

data(Boston, package = "MASS")

grid <- expand_params(
  n.trees = c(50, 100),
  interaction.depth = 1:2
)

fit(medv ~ ., data = Boston, model = TunedModel(GBMModel, grid = grid))
}

}
\seealso{
\code{\link{TunedModel}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TrainingParams.R
\name{set_predict}
\alias{set_predict}
\title{Resampling Prediction Control}
\usage{
set_predict(
  object,
  times = numeric(),
  distr = character(),
  method = character(),
  ...
)
}
\arguments{
\item{object}{\link[=controls]{control} object.}

\item{times, distr, method}{arguments passed to \code{\link{predict}}.}

\item{...}{arguments passed to other methods.}
}
\value{
Argument \code{object} updated with the supplied parameters.
}
\description{
Set parameters that control prediction during resample estimation of model
performance.
}
\examples{
CVControl() \%>\% set_predict(times = 1:3)

}
\seealso{
\code{\link{resample}}, \code{\link{set_monitor}},
\code{\link{set_optim}}, \code{\link{set_strata}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MachineShop-package.R
\docType{package}
\name{MachineShop-package}
\alias{MachineShop}
\alias{MachineShop-package}
\title{MachineShop: Machine Learning Models and Tools}
\description{
Meta-package for statistical and machine learning with a unified interface for model fitting, prediction, performance assessment, and presentation of results. Approaches for model fitting and prediction of numerical, categorical, or censored time-to-event outcomes include traditional regression models, regularization methods, tree-based methods, support vector machines, neural networks, ensembles, data preprocessing, filtering, and model tuning and selection. Performance metrics are provided for model assessment and can be estimated with independent test sets, split sampling, cross-validation, or bootstrap resampling. Resample estimation can be executed in parallel for faster processing and nested in cases of model tuning and selection. Modeling results can be summarized with descriptive statistics; calibration curves; variable importance; partial dependence plots; confusion matrices; and ROC, lift, and other performance curves.
}
\details{
The following set of model fitting, prediction, and performance assessment
functions are available for \pkg{MachineShop} \link{models}.

Training:
\tabular{ll}{
  \code{\link{fit}} \tab Model fitting \cr
  \code{\link{resample}} \tab Resample estimation of model performance \cr
}

Tuning Grids:
\tabular{ll}{
  \code{\link{expand_model}} \tab Model expansion over tuning parameters \cr
  \code{\link{expand_modelgrid}} \tab Model tuning grid expansion \cr
  \code{\link{expand_params}} \tab Model parameters expansion \cr
  \code{\link{expand_steps}} \tab Recipe step parameters expansion \cr
}

Response Values:
\tabular{ll}{
  \code{\link{response}} \tab Observed \cr
  \code{\link{predict}} \tab Predicted \cr
}

Performance Assessment:
\tabular{ll}{
  \code{\link{calibration}} \tab Model calibration \cr
  \code{\link{confusion}} \tab Confusion matrix \cr
  \code{\link{dependence}} \tab Parital dependence \cr
  \code{\link{diff}} \tab Model performance differences \cr
  \code{\link{lift}} \tab Lift curves \cr
  \code{\link{performance} \link{metrics}} \tab Model performance metrics \cr
  \code{\link{performance_curve}} \tab Model performance curves \cr
  \code{\link{rfe}} \tab Recursive feature elimination \cr
  \code{\link{varimp}} \tab Variable importance \cr
}

Methods for resample estimation include

\tabular{ll}{
  \code{\link{BootControl}} \tab Simple bootstrap \cr
  \code{\link{BootOptimismControl}} \tab Optimism-corrected bootstrap \cr
  \code{\link{CVControl}} \tab Repeated K-fold cross-validation \cr
  \code{\link{CVOptimismControl}} \tab Optimism-corrected cross-validation \cr
  \code{\link{OOBControl}} \tab Out-of-bootstrap \cr
  \code{\link{SplitControl}} \tab Split training-testing \cr
  \code{\link{TrainControl}} \tab Training resubstitution \cr
}

Graphical and tabular summaries of modeling results can be obtained with

\tabular{l}{
  \code{\link{plot}} \cr
  \code{\link{print}} \cr
  \code{\link{summary}} \cr
}

Further information on package features is available with

\tabular{ll}{
  \code{\link{metricinfo}} \tab Performance metric information \cr
  \code{\link{modelinfo}} \tab Model information \cr
  \code{\link{settings}} \tab Global settings \cr
}

Custom metrics and models can be created with the \code{\link{MLMetric}} and
\code{\link{MLModel}} constructors.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://brian-j-smith.github.io/MachineShop/}
  \item Report bugs at \url{https://github.com/brian-j-smith/MachineShop/issues}
}

}
\author{
\strong{Maintainer}: Brian J Smith \email{brian-j-smith@uiowa.edu}

Authors:
\itemize{
  \item Brian J Smith \email{brian-j-smith@uiowa.edu}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_AdaBoostModel.R
\name{AdaBoostModel}
\alias{AdaBoostModel}
\title{Boosting with Classification Trees}
\usage{
AdaBoostModel(
  boos = TRUE,
  mfinal = 100,
  coeflearn = c("Breiman", "Freund", "Zhu"),
  minsplit = 20,
  minbucket = round(minsplit/3),
  cp = 0.01,
  maxcompete = 4,
  maxsurrogate = 5,
  usesurrogate = 2,
  xval = 10,
  surrogatestyle = 0,
  maxdepth = 30
)
}
\arguments{
\item{boos}{if \code{TRUE}, then bootstrap samples are drawn from the
training set using the observation weights at each iteration.  If
\code{FALSE}, then all observations are used with their weights.}

\item{mfinal}{number of iterations for which boosting is run.}

\item{coeflearn}{learning algorithm.}

\item{minsplit}{minimum number of observations that must exist in a node in
order for a split to be attempted.}

\item{minbucket}{minimum number of observations in any terminal node.}

\item{cp}{complexity parameter.}

\item{maxcompete}{number of competitor splits retained in the output.}

\item{maxsurrogate}{number of surrogate splits retained in the output.}

\item{usesurrogate}{how to use surrogates in the splitting process.}

\item{xval}{number of cross-validations.}

\item{surrogatestyle}{controls the selection of a best surrogate.}

\item{maxdepth}{maximum depth of any node of the final tree, with the root
node counted as depth 0.}
}
\value{
\code{MLModel} class object.
}
\description{
Fits the AdaBoost.M1 (Freund and Schapire, 1996) and SAMME (Zhu et al., 2009)
algorithms using classification trees as single classifiers.
}
\details{
\describe{
  \item{Response types:}{\code{factor}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \code{mfinal}, \code{maxdepth}, \code{coeflearn}*
  }
}
* excluded from grids by default

Further model details can be found in the source link below.
}
\examples{
\donttest{
## Requires prior installation of suggested package adabag to run

fit(Species ~ ., data = iris, model = AdaBoostModel(mfinal = 5))
}

}
\seealso{
\code{\link[adabag]{boosting}}, \code{\link{fit}},
\code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_QDAModel.R
\name{QDAModel}
\alias{QDAModel}
\title{Quadratic Discriminant Analysis Model}
\usage{
QDAModel(
  prior = numeric(),
  method = c("moment", "mle", "mve", "t"),
  nu = 5,
  use = c("plug-in", "predictive", "debiased", "looCV")
)
}
\arguments{
\item{prior}{prior probabilities of class membership if specified or the
class proportions in the training set otherwise.}

\item{method}{type of mean and variance estimator.}

\item{nu}{degrees of freedom for \code{method = "t"}.}

\item{use}{type of parameter estimation to use for prediction.}
}
\value{
\code{MLModel} class object.
}
\description{
Performs quadratic discriminant analysis.
}
\details{
\describe{
  \item{Response types:}{\code{factor}}
}

The \code{\link{predict}} function for this model additionally accepts the
following argument.
\describe{
  \item{\code{prior}}{prior class membership probabilities for prediction data
    if different from the training set.}
}

Default argument values and further model details can be found in the source
See Also links below.
}
\examples{
fit(Species ~ ., data = iris, model = QDAModel)

}
\seealso{
\code{\link[MASS]{qda}}, \code{\link[MASS]{predict.qda}},
\code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract.R
\name{extract}
\alias{extract}
\alias{[.BinomialVariate}
\alias{[,DiscreteVariate,ANY,missing,missing-method}
\alias{[,ListOf,ANY,missing,missing-method}
\alias{[,ModelFrame,ANY,ANY,ANY-method}
\alias{[,ModelFrame,ANY,missing,ANY-method}
\alias{[,ModelFrame,missing,ANY,ANY-method}
\alias{[,ModelFrame,missing,missing,ANY-method}
\alias{[,RecipeGrid,ANY,ANY,ANY-method}
\alias{[,Resample,ANY,ANY,ANY-method}
\alias{[,Resample,ANY,missing,ANY-method}
\alias{[,Resample,missing,missing,ANY-method}
\alias{[,SurvMatrix,ANY,ANY,ANY-method}
\alias{[,SurvTimes,ANY,missing,missing-method}
\title{Extract Elements of an Object}
\usage{
\method{[}{BinomialVariate}(x, i, j, ..., drop = FALSE)

\S4method{[}{DiscreteVariate,ANY,missing,missing}(x, i)

\S4method{[}{ListOf,ANY,missing,missing}(x, i)

\S4method{[}{ModelFrame,ANY,ANY,ANY}(x, i, j, ..., drop = FALSE)

\S4method{[}{ModelFrame,ANY,missing,ANY}(x, i, j, ..., drop = FALSE)

\S4method{[}{ModelFrame,missing,ANY,ANY}(x, i, j, ..., drop = FALSE)

\S4method{[}{ModelFrame,missing,missing,ANY}(x, i, j, ..., drop = FALSE)

\S4method{[}{RecipeGrid,ANY,ANY,ANY}(x, i, j, ..., drop = FALSE)

\S4method{[}{Resample,ANY,ANY,ANY}(x, i, j, ..., drop = FALSE)

\S4method{[}{Resample,ANY,missing,ANY}(x, i, j, ..., drop = FALSE)

\S4method{[}{Resample,missing,missing,ANY}(x, i, j, ..., drop = FALSE)

\S4method{[}{SurvMatrix,ANY,ANY,ANY}(x, i, j, ..., drop = FALSE)

\S4method{[}{SurvTimes,ANY,missing,missing}(x, i)
}
\arguments{
\item{x}{object from which to extract elements.}

\item{i, j, ...}{indices specifying elements to extract.}

\item{drop}{logical indicating that the result be returned as an object
coerced to the lowest dimension possible if \code{TRUE} or
with the original dimensions and class otherwise.}
}
\description{
Operators acting on data structures to extract elements.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modelinfo.R
\name{modelinfo}
\alias{modelinfo}
\title{Display Model Information}
\usage{
modelinfo(...)
}
\arguments{
\item{...}{\link[=models]{model} functions, function names, or objects;
\link[=response]{observed responses} for which to display information.  If
none are specified, information is returned on all available models by
default.}
}
\value{
List of named model elements each containing the following
components:
\describe{
  \item{label}{character descriptor for the model.}
  \item{packages}{character vector of source packages required to use the
    model.  These need only be installed with the
    \code{\link{install.packages}} function or by equivalent means; but need
    not be loaded with, for example, the \code{\link{library}} function.}
  \item{response_types}{character vector of response variable types supported
    by the model.}
 \item{weights}{logical value or vector of the same length as
   \code{response_types} indicating whether case weights are supported for
   the responses.}
  \item{arguments}{closure with the argument names and corresponding default
    values of the model function.}
  \item{grid}{logical indicating whether automatic generation of tuning
    parameter grids is implemented for the model.}
  \item{varimp}{logical indicating whether model-specific variable importance
    is defined.}
}
}
\description{
Display information about models supplied by the \pkg{MachineShop} package.
}
\examples{
## All models
modelinfo()

## Models by response types
names(modelinfo(factor(0)))
names(modelinfo(factor(0), numeric(0)))

## Model-specific information
modelinfo(GBMModel)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/step_spca.R
\name{step_spca}
\alias{step_spca}
\alias{tidy.step_spca}
\alias{tunable.step_spca}
\title{Sparse Principal Components Analysis Variable Reduction}
\usage{
step_spca(
  recipe,
  ...,
  num_comp = 5,
  sparsity = 0,
  num_var = integer(),
  shrinkage = 1e-06,
  center = TRUE,
  scale = TRUE,
  max_iter = 200,
  tol = 0.001,
  replace = TRUE,
  prefix = "SPCA",
  role = "predictor",
  skip = FALSE,
  id = recipes::rand_id("spca")
)

\method{tidy}{step_spca}(x, ...)

\method{tunable}{step_spca}(x, ...)
}
\arguments{
\item{recipe}{\link[recipes]{recipe} object to which the step will be added.}

\item{...}{one or more selector functions to choose which variables will be
used to compute the components.  See \code{\link[recipes]{selections}} for
more details.  These are not currently used by the \code{tidy} method.}

\item{num_comp}{number of components to derive.  The value of \code{num_comp}
will be constrained to a minimum of 1 and maximum of the number of original
variables when \code{\link[recipes]{prep}} is run.}

\item{sparsity, num_var}{sparsity (L1 norm) penalty for each component or
number of variables with non-zero component loadings.  Larger sparsity
values produce more zero loadings.  Argument \code{sparsity} is ignored if
\code{num_var} is given.  The argument value may be a single number
applied to all components or a vector of component-specific numbers.}

\item{shrinkage}{numeric shrinkage (quadratic) penalty for the components to
improve conditioning; larger values produce more shrinkage of component
loadings toward zero.}

\item{center, scale}{logicals indicating whether to mean center and standard
deviation scale the original variables prior to deriving components, or
functions or names of functions for the centering and scaling.}

\item{max_iter}{maximum number of algorithm iterations allowed.}

\item{tol}{numeric tolerance for the convergence criterion.}

\item{replace}{logical indicating whether to replace the original variables.}

\item{prefix}{character string prefix added to a sequence of zero-padded
integers to generate names for the resulting new variables.}

\item{role}{analysis role that added step variables should be assigned.  By
default, they are designated as model predictors.}

\item{skip}{logical indicating whether to skip the step when the recipe is
baked.  While all operations are baked when \code{\link[recipes]{prep}} is
run, some operations may not be applicable to new data (e.g. processing
outcome variables).  Care should be taken when using \code{skip = TRUE} as
it may affect the computations for subsequent operations.}

\item{id}{unique character string to identify the step.}

\item{x}{\code{step_spca} object.}
}
\value{
Function \code{step_spca} creates a new step whose class is of
the same name and inherits from \code{\link{step_lincomp}}, adds it to the
sequence of existing steps (if any) in the recipe, and returns the updated
recipe.  For the \code{tidy} method, a tibble with columns \code{terms}
(selectors or variables selected), \code{weight} of each variable loading in
the components, and \code{name} of the new variable names; and with
attribute \code{pev} containing the proportions of explained variation.
}
\description{
Creates a \emph{specification} of a recipe step that will derive sparse
principal components from one or more numeric variables.
}
\details{
Sparse principal components analysis (SPCA) is a variant of PCA in which
the original variables may have zero loadings in the linear combinations
that form the components.
}
\examples{
\donttest{
## Requires prior installation of suggested package elasticnet to run

library(recipes)

rec <- recipe(rating ~ ., data = attitude)
spca_rec <- rec \%>\%
  step_spca(all_predictors(), num_comp = 5, sparsity = 1)
spca_prep <- prep(spca_rec, training = attitude)
spca_data <- bake(spca_prep, attitude)

pairs(spca_data, lower.panel = NULL)

tidy(spca_rec, number = 1)
tidy(spca_prep, number = 1)
}

}
\references{
Zou, H., Hastie, T., & Tibshirani, R. (2006). Sparse principal component
analysis. \emph{Journal of Computational and Graphical Statistics},
\emph{15}(2), 265-286.
}
\seealso{
\code{\link[elasticnet]{spca}}, \code{\link[recipes]{recipe}},
\code{\link[recipes]{prep}}, \code{\link[recipes]{bake}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary}
\alias{summary}
\alias{summary.ConfusionList}
\alias{summary.ConfusionMatrix}
\alias{summary.MLModel}
\alias{summary.MLModelFit}
\alias{summary.Performance}
\alias{summary.PerformanceCurve}
\alias{summary.Resample}
\alias{summary.TrainingStep}
\title{Model Performance Summaries}
\usage{
\method{summary}{ConfusionList}(object, ...)

\method{summary}{ConfusionMatrix}(object, ...)

\method{summary}{MLModel}(
  object,
  stats = MachineShop::settings("stats.Resample"),
  na.rm = TRUE,
  ...
)

\method{summary}{MLModelFit}(object, .type = c("default", "glance", "tidy"), ...)

\method{summary}{Performance}(
  object,
  stats = MachineShop::settings("stats.Resample"),
  na.rm = TRUE,
  ...
)

\method{summary}{PerformanceCurve}(object, stat = MachineShop::settings("stat.Curve"), ...)

\method{summary}{Resample}(
  object,
  stats = MachineShop::settings("stats.Resample"),
  na.rm = TRUE,
  ...
)

\method{summary}{TrainingStep}(object, ...)
}
\arguments{
\item{object}{\link{confusion}, \link{lift}, trained model \link{fit},
\link{performance}, \link[=curves]{performance curve}, \link{resample}, or
\link{rfe} result.}

\item{...}{arguments passed to other methods.}

\item{stats}{function, function name, or vector of these with which to
compute summary statistics.}

\item{na.rm}{logical indicating whether to exclude missing values.}

\item{.type}{character string specifying that
\code{\link[=unMLModelFit]{unMLModelFit(object)}} be passed to
\code{\link[base:summary]{summary}} (\code{"default"}),
\code{\link[generics:glance]{glance}}, or
\code{\link[generics:tidy]{tidy}}.}

\item{stat}{function or character string naming a function to compute a
summary statistic at each cutoff value of resampled metrics in
\code{PerformanceCurve}, or \code{NULL} for resample-specific metrics.}
}
\value{
An object of summmary statistics.
}
\description{
Summary statistics for resampled model performance metrics.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

## Factor response example

fo <- Species ~ .
control <- CVControl()

gbm_res1 <- resample(fo, iris, GBMModel(n.trees = 25), control)
gbm_res2 <- resample(fo, iris, GBMModel(n.trees = 50), control)
gbm_res3 <- resample(fo, iris, GBMModel(n.trees = 100), control)
summary(gbm_res3)

res <- c(GBM1 = gbm_res1, GBM2 = gbm_res2, GBM3 = gbm_res3)
summary(res)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_CoxModel.R
\name{CoxModel}
\alias{CoxModel}
\alias{CoxStepAICModel}
\title{Proportional Hazards Regression Model}
\usage{
CoxModel(ties = c("efron", "breslow", "exact"), ...)

CoxStepAICModel(
  ties = c("efron", "breslow", "exact"),
  ...,
  direction = c("both", "backward", "forward"),
  scope = list(),
  k = 2,
  trace = FALSE,
  steps = 1000
)
}
\arguments{
\item{ties}{character string specifying the method for tie handling.}

\item{...}{arguments passed to \code{\link[survival]{coxph.control}}.}

\item{direction}{mode of stepwise search, can be one of \code{"both"}
(default), \code{"backward"}, or \code{"forward"}.}

\item{scope}{defines the range of models examined in the stepwise search.
This should be a list containing components \code{upper} and \code{lower},
both formulae.}

\item{k}{multiple of the number of degrees of freedom used for the penalty.
Only \code{k = 2} gives the genuine AIC; \code{k = .(log(nobs))} is
sometimes referred to as BIC or SBC.}

\item{trace}{if positive, information is printed during the running of
\code{stepAIC}. Larger values may give more information on the fitting
process.}

\item{steps}{maximum number of steps to be considered.}
}
\value{
\code{MLModel} class object.
}
\description{
Fits a Cox proportional hazards regression model. Time dependent variables,
time dependent strata, multiple events per subject, and other extensions are
incorporated using the counting process formulation of Andersen and Gill.
}
\details{
\describe{
  \item{Response types:}{\code{Surv}}
}

Default argument values and further model details can be found in the source
See Also links below.

In calls to \code{\link{varimp}} for \code{CoxModel} and
\code{CoxStepAICModel}, numeric argument \code{base} may be specified for the
(negative) logarithmic transformation of p-values [defaul: \code{exp(1)}].
Transformed p-values are automatically scaled in the calculation of variable
importance to range from 0 to 100.  To obtain unscaled importance values, set
\code{scale = FALSE}.
}
\examples{
library(survival)

fit(Surv(time, status) ~ ., data = veteran, model = CoxModel)

}
\seealso{
\code{\link[survival]{coxph}},
\code{\link[survival]{coxph.control}}, \code{\link[MASS]{stepAIC}},
\code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predict}
\alias{predict}
\alias{predict.MLModelFit}
\alias{predict,MLModelFit-method}
\title{Model Prediction}
\usage{
\method{predict}{MLModelFit}(
  object,
  newdata = NULL,
  times = numeric(),
  type = c("response", "raw", "numeric", "prob", "default"),
  cutoff = MachineShop::settings("cutoff"),
  distr = character(),
  method = character(),
  verbose = FALSE,
  ...
)

\S4method{predict}{MLModelFit}(object, ...)
}
\arguments{
\item{object}{model \link{fit} result.}

\item{newdata}{optional \link[=data.frame]{data frame} with which to obtain
predictions.  If not specified, the training data will be used by default.}

\item{times}{numeric vector of follow-up times at which to predict
survival events/probabilities or \code{NULL} for predicted survival means.}

\item{type}{specifies prediction on the original outcome (\code{"response"}),
numeric (\code{"numeric"}), or probability (\code{"prob"}) scale; or
the \code{"raw"} predictions returned by the model.  Option
\code{"default"} is deprecated and will be removed in the future; use
\code{"raw"} instead.}

\item{cutoff}{numeric (0, 1) threshold above which binary factor
probabilities are classified as events and below which survival
probabilities are classified.}

\item{distr}{character string specifying distributional approximations to
estimated survival curves.  Possible values are \code{"empirical"},
\code{"exponential"}, \code{"rayleigh"}, or \code{"weibull"}; with defaults
of \code{"empirical"} for predicted survival events/probabilities and
\code{"weibull"} for predicted survival means.}

\item{method}{character string specifying the empirical method of estimating
baseline survival curves for Cox proportional hazards-based models.
Choices are \code{"breslow"} or \code{"efron"} (default).}

\item{verbose}{logical indicating whether to display printed output generated
by some model-specific predict functions to aid in monitoring progress and
diagnosing errors.}

\item{...}{arguments passed from the S4 to the S3 method.}
}
\description{
Predict outcomes with a fitted model.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

## Survival response example
library(survival)

gbm_fit <- fit(Surv(time, status) ~ ., data = veteran, model = GBMModel)
predict(gbm_fit, newdata = veteran, times = c(90, 180, 360), type = "prob")
}

}
\seealso{
\code{\link{confusion}}, \code{\link{performance}},
\code{\link{metrics}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_SurvRegModel.R
\name{SurvRegModel}
\alias{SurvRegModel}
\alias{SurvRegStepAICModel}
\title{Parametric Survival Model}
\usage{
SurvRegModel(
  dist = c("weibull", "exponential", "gaussian", "logistic", "lognormal",
    "logloglogistic"),
  scale = 0,
  parms = list(),
  ...
)

SurvRegStepAICModel(
  dist = c("weibull", "exponential", "gaussian", "logistic", "lognormal",
    "logloglogistic"),
  scale = 0,
  parms = list(),
  ...,
  direction = c("both", "backward", "forward"),
  scope = list(),
  k = 2,
  trace = FALSE,
  steps = 1000
)
}
\arguments{
\item{dist}{assumed distribution for y variable.}

\item{scale}{optional fixed value for the scale.}

\item{parms}{list of fixed parameters.}

\item{...}{arguments passed to \code{\link[survival]{survreg.control}}.}

\item{direction}{mode of stepwise search, can be one of \code{"both"}
(default), \code{"backward"}, or \code{"forward"}.}

\item{scope}{defines the range of models examined in the stepwise search.
This should be a list containing components \code{upper} and \code{lower},
both formulae.}

\item{k}{multiple of the number of degrees of freedom used for the penalty.
Only \code{k = 2} gives the genuine AIC; \code{k = .(log(nobs))} is
sometimes referred to as BIC or SBC.}

\item{trace}{if positive, information is printed during the running of
\code{stepAIC}. Larger values may give more information on the fitting
process.}

\item{steps}{maximum number of steps to be considered.}
}
\value{
\code{MLModel} class object.
}
\description{
Fits the accelerated failure time family of parametric survival models.
}
\details{
\describe{
  \item{Response types:}{\code{Surv}}
}

Default argument values and further model details can be found in the source
See Also links below.
}
\examples{
\donttest{
## Requires prior installation of suggested packages rms and Hmisc to run

library(survival)

fit(Surv(time, status) ~ ., data = veteran, model = SurvRegModel)
}

}
\seealso{
\code{\link[rms]{psm}}, \code{\link[survival]{survreg}},
\code{\link[survival]{survreg.control}}, \code{\link[MASS]{stepAIC}},
\code{\link{fit}}, \code{\link{resample}}

\code{\link[MASS]{stepAIC}}, \code{\link{fit}},
\code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/recipe_roles.R
\name{recipe_roles}
\alias{recipe_roles}
\alias{role_binom}
\alias{role_case}
\alias{role_pred}
\alias{role_surv}
\title{Set Recipe Roles}
\usage{
role_binom(recipe, x, size)

role_case(recipe, group, stratum, weight, replace = FALSE)

role_pred(recipe, offset, replace = FALSE)

role_surv(recipe, time, event)
}
\arguments{
\item{recipe}{existing \link[recipes]{recipe} object.}

\item{x, size}{number of counts and trials for the specification of a
\code{\link{BinomialVariate}} outcome.}

\item{group}{variable defining groupings of case observations, such as
repeated measurements, to keep together during resampling [default: none].}

\item{stratum}{variable to use in conducting stratified \link{resample}
estimation of model performance.}

\item{weight}{numeric variable of case weights for model
\link[=fit]{fitting}.}

\item{replace}{logical indicating whether to replace existing roles.}

\item{offset}{numeric variable to be added to a linear predictor, such as in
a generalized linear model, with known coefficient 1 rather than an
estimated coefficient.}

\item{time, event}{numeric follow up time and 0-1 numeric or logical event
indicator for specification of a \code{\link[survival]{Surv}} outcome.  If
the event indicator is omitted, all cases are assumed to have events.}
}
\value{
An updated recipe object.
}
\description{
Add to or replace the roles of variables in a preprocessing recipe.
}
\examples{
library(survival)
library(recipes)

df <- within(veteran, {
  y <- Surv(time, status)
  remove(time, status)
})
rec <- recipe(y ~ ., data = df) \%>\%
  role_case(stratum = y)

(res <- resample(rec, model = CoxModel))
summary(res)

}
\seealso{
\code{\link[recipes]{recipe}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{as.MLInput}
\alias{as.MLInput}
\alias{as.MLInput.MLModelFit}
\alias{as.MLInput.ModelSpecification}
\title{Coerce to an MLInput}
\usage{
as.MLInput(x, ...)

\method{as.MLInput}{MLModelFit}(x, ...)

\method{as.MLInput}{ModelSpecification}(x, ...)
}
\arguments{
\item{x}{model \link{fit} result or \pkg{MachineShop}
\link[=ModelSpecification]{model specification}.}

\item{...}{arguments passed to other methods.}
}
\value{
\code{MLInput} class object.
}
\description{
Function to coerce an object to \code{MLInput}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_PLSModel.R
\name{PLSModel}
\alias{PLSModel}
\title{Partial Least Squares Model}
\usage{
PLSModel(ncomp = 1, scale = FALSE)
}
\arguments{
\item{ncomp}{number of components to include in the model.}

\item{scale}{logical indicating whether to scale the predictors by the
sample standard deviation.}
}
\value{
\code{MLModel} class object.
}
\description{
Function to perform partial least squares regression.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{numeric}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \code{ncomp}
  }
}

Further model details can be found in the source link below.
}
\examples{
\donttest{
## Requires prior installation of suggested package pls to run

fit(sale_amount ~ ., data = ICHomes, model = PLSModel)
}

}
\seealso{
\code{\link[pls]{mvr}}, \code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_StackedModel.R
\name{StackedModel}
\alias{StackedModel}
\title{Stacked Regression Model}
\usage{
StackedModel(
  ...,
  control = MachineShop::settings("control"),
  weights = numeric()
)
}
\arguments{
\item{...}{\link[=models]{model} functions, function names, objects; other
objects that can be \link[=as.MLModel]{coerced} to models; or vector of
these to serve as base learners.}

\item{control}{\link[=controls]{control} function, function name, or object
defining the resampling method to be employed for the estimation of base
learner weights.}

\item{weights}{optional fixed base learner weights.}
}
\value{
\code{StackedModel} class object that inherits from \code{MLModel}.
}
\description{
Fit a stacked regression model from multiple base learners.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{numeric}, \code{ordered},
    \code{Surv}}
}
}
\examples{
\donttest{
## Requires prior installation of suggested packages gbm and glmnet to run

model <- StackedModel(GBMModel, SVMRadialModel, GLMNetModel(lambda = 0.01))
model_fit <- fit(sale_amount ~ ., data = ICHomes, model = model)
predict(model_fit, newdata = ICHomes)
}

}
\references{
Breiman, L. (1996). Stacked regression. \emph{Machine Learning}, \emph{24},
49-64.
}
\seealso{
\code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_RandomForestModel.R
\name{RandomForestModel}
\alias{RandomForestModel}
\title{Random Forest Model}
\usage{
RandomForestModel(
  ntree = 500,
  mtry = .(if (is.factor(y)) floor(sqrt(nvars)) else max(floor(nvars/3), 1)),
  replace = TRUE,
  nodesize = .(if (is.factor(y)) 1 else 5),
  maxnodes = integer()
)
}
\arguments{
\item{ntree}{number of trees to grow.}

\item{mtry}{number of variables randomly sampled as candidates at each split.}

\item{replace}{should sampling of cases be done with or without replacement?}

\item{nodesize}{minimum size of terminal nodes.}

\item{maxnodes}{maximum number of terminal nodes trees in the forest can
have.}
}
\value{
\code{MLModel} class object.
}
\description{
Implementation of Breiman's random forest algorithm (based on Breiman and
Cutler's original Fortran code) for classification and regression.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{numeric}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \code{mtry}, \code{nodesize}*
  }
}
* excluded from grids by default

Default argument values and further model details can be found in the source
See Also link below.
}
\examples{
\donttest{
## Requires prior installation of suggested package randomForest to run

fit(sale_amount ~ ., data = ICHomes, model = RandomForestModel)
}

}
\seealso{
\code{\link[randomForest]{randomForest}}, \code{\link{fit}},
\code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response.R
\name{response}
\alias{response}
\alias{response.MLModelFit}
\alias{response.ModelFrame}
\alias{response.ModelSpecification}
\alias{response.recipe}
\title{Extract Response Variable}
\usage{
response(object, ...)

\method{response}{MLModelFit}(object, newdata = NULL, ...)

\method{response}{ModelFrame}(object, newdata = NULL, ...)

\method{response}{ModelSpecification}(object, newdata = NULL, ...)

\method{response}{recipe}(object, newdata = NULL, ...)
}
\arguments{
\item{object}{model \link{fit}, \link[=inputs]{input}, or
\link[=ModelSpecification]{specification} containing predictor and response
variables.}

\item{...}{arguments passed to other methods.}

\item{newdata}{\link[=data.frame]{data frame} from which to extract the
response variable values if given; otherwise, \code{object} is used.}
}
\description{
Extract the response variable from an object.
}
\examples{
## Survival response example
library(survival)

mf <- ModelFrame(Surv(time, status) ~ ., data = veteran)
response(mf)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_GBMModel.R
\name{GBMModel}
\alias{GBMModel}
\title{Generalized Boosted Regression Model}
\usage{
GBMModel(
  distribution = character(),
  n.trees = 100,
  interaction.depth = 1,
  n.minobsinnode = 10,
  shrinkage = 0.1,
  bag.fraction = 0.5
)
}
\arguments{
\item{distribution}{optional character string specifying the name of the
distribution to use or list with a component \code{name} specifying the
distribution and any additional parameters needed.  Set automatically
according to the class type of the response variable.}

\item{n.trees}{total number of trees to fit.}

\item{interaction.depth}{maximum depth of variable interactions.}

\item{n.minobsinnode}{minimum number of observations in the trees terminal
nodes.}

\item{shrinkage}{shrinkage parameter applied to each tree in the expansion.}

\item{bag.fraction}{fraction of the training set observations randomly
selected to propose the next tree in the expansion.}
}
\value{
\code{MLModel} class object.
}
\description{
Fits generalized boosted regression models.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{numeric},
    \code{PoissonVariate}, \code{Surv}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \code{n.trees}, \code{interaction.depth}, \code{shrinkage}*,
    \code{n.minobsinnode}*
  }
}
* excluded from grids by default

Default argument values and further model details can be found in the source
See Also link below.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

fit(Species ~ ., data = iris, model = GBMModel)
}

}
\seealso{
\code{\link[gbm]{gbm}}, \code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_LDAModel.R
\name{LDAModel}
\alias{LDAModel}
\title{Linear Discriminant Analysis Model}
\usage{
LDAModel(
  prior = numeric(),
  tol = 1e-04,
  method = c("moment", "mle", "mve", "t"),
  nu = 5,
  dimen = integer(),
  use = c("plug-in", "debiased", "predictive")
)
}
\arguments{
\item{prior}{prior probabilities of class membership if specified or the
class proportions in the training set otherwise.}

\item{tol}{tolerance for the determination of singular matrices.}

\item{method}{type of mean and variance estimator.}

\item{nu}{degrees of freedom for \code{method = "t"}.}

\item{dimen}{dimension of the space to use for prediction.}

\item{use}{type of parameter estimation to use for prediction.}
}
\value{
\code{MLModel} class object.
}
\description{
Performs linear discriminant analysis.
}
\details{
\describe{
  \item{Response types:}{\code{factor}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameter:}{
    \code{dimen}
  }
}

The \code{\link{predict}} function for this model additionally accepts the
following argument.
\describe{
  \item{\code{prior}}{prior class membership probabilities for prediction
    data if different from the training set.}
}

Default argument values and further model details can be found in the source
See Also links below.
}
\examples{
fit(Species ~ ., data = iris, model = LDAModel)

}
\seealso{
\code{\link[MASS]{lda}}, \code{\link[MASS]{predict.lda}},
\code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_EarthModel.R
\name{EarthModel}
\alias{EarthModel}
\title{Multivariate Adaptive Regression Splines Model}
\usage{
EarthModel(
  pmethod = c("backward", "none", "exhaustive", "forward", "seqrep", "cv"),
  trace = 0,
  degree = 1,
  nprune = integer(),
  nfold = 0,
  ncross = 1,
  stratify = TRUE
)
}
\arguments{
\item{pmethod}{pruning method.}

\item{trace}{level of execution information to display.}

\item{degree}{maximum degree of interaction.}

\item{nprune}{maximum number of terms (including intercept) in the pruned
model.}

\item{nfold}{number of cross-validation folds.}

\item{ncross}{number of cross-validations if \code{nfold > 1}.}

\item{stratify}{logical indicating whether to stratify cross-validation
samples by the response levels.}
}
\value{
\code{MLModel} class object.
}
\description{
Build a regression model using the techniques in Friedman's papers
"Multivariate Adaptive Regression Splines" and "Fast MARS".
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{numeric}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \code{nprune}, \code{degree}*
  }
}
* excluded from grids by default

Default argument values and further model details can be found in the source
See Also link below.

In calls to \code{\link{varimp}} for \code{EarthModel}, argument
\code{type} may be specified as \code{"nsubsets"} (default) for the number of
model subsets that include each predictor, as \code{"gcv"} for the
generalized cross-validation decrease over all subsets that include each
predictor, or as \code{"rss"} for the residual sums of squares decrease.
Variable importance is automatically scaled to range from 0 to 100.  To
obtain unscaled importance values, set \code{scale = FALSE}.  See example
below.
}
\examples{
\donttest{
## Requires prior installation of suggested package earth to run

model_fit <- fit(Species ~ ., data = iris, model = EarthModel)
varimp(model_fit, method = "model", type = "gcv", scale = FALSE)
}

}
\seealso{
\code{\link[earth]{earth}}, \code{\link{fit}},
\code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_KNNModel.R
\name{KNNModel}
\alias{KNNModel}
\title{Weighted k-Nearest Neighbor Model}
\usage{
KNNModel(
  k = 7,
  distance = 2,
  scale = TRUE,
  kernel = c("optimal", "biweight", "cos", "epanechnikov", "gaussian", "inv", "rank",
    "rectangular", "triangular", "triweight")
)
}
\arguments{
\item{k}{numer of neigbors considered.}

\item{distance}{Minkowski distance parameter.}

\item{scale}{logical indicating whether to scale predictors to have equal
standard deviations.}

\item{kernel}{kernel to use.}
}
\value{
\code{MLModel} class object.
}
\description{
Fit a k-nearest neighbor model for which the k nearest training set vectors
(according to Minkowski distance) are found for each row of the test set, and
prediction is done via the maximum of summed kernel densities.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{numeric}, \code{ordinal}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \code{k}, \code{distance}*, \code{kernel}*
  }
}
* excluded from grids by default

Further model details can be found in the source link below.
}
\examples{
\donttest{
## Requires prior installation of suggested package kknn to run

fit(Species ~ ., data = iris, model = KNNModel)
}

}
\seealso{
\code{\link[kknn]{kknn}}, \code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resample.R
\name{resample}
\alias{resample}
\alias{resample.formula}
\alias{resample.matrix}
\alias{resample.ModelFrame}
\alias{resample.recipe}
\alias{resample.ModelSpecification}
\alias{resample.MLModel}
\alias{resample.MLModelFunction}
\title{Resample Estimation of Model Performance}
\usage{
resample(...)

\method{resample}{formula}(formula, data, model, ...)

\method{resample}{matrix}(x, y, model, ...)

\method{resample}{ModelFrame}(input, model, ...)

\method{resample}{recipe}(input, model, ...)

\method{resample}{ModelSpecification}(object, control = MachineShop::settings("control"), ...)

\method{resample}{MLModel}(model, ...)

\method{resample}{MLModelFunction}(model, ...)
}
\arguments{
\item{...}{arguments passed from the generic function to its methods, from
the \code{MLModel} and \code{MLModelFunction} methods to first arguments of
others, and from others to the \code{ModelSpecification} method.  The
first argument of each \code{fit} method is positional and, as such, must
be given first in calls to them.}

\item{formula, data}{\link[=formula]{formula} defining the model predictor and
response variables and a \link[=data.frame]{data frame} containing them.}

\item{model}{\link[=models]{model} function, function name, or object; or
another object that can be \link[=as.MLModel]{coerced} to a model.  A model
can be given first followed by any of the variable specifications.}

\item{x, y}{\link{matrix} and object containing predictor and response
variables.}

\item{input}{\link[=inputs]{input} object defining and containing the model
predictor and response variables.}

\item{object}{model \link[=inputs]{input} or
\link[=ModelSpecification]{specification}.}

\item{control}{\link[=controls]{control} function, function name, or object
defining the resampling method to be employed.}
}
\value{
\code{Resample} class object.
}
\description{
Estimation of the predictive performance of a model estimated and evaluated
on training and test samples generated from an observed data set.
}
\details{
Stratified resampling is performed automatically for the \code{formula} and
\code{matrix} methods according to the type of response variable.  In
general, strata are constructed from numeric proportions for
\code{\link{BinomialVariate}}; original values for \code{character},
\code{factor}, \code{logical}, and \code{ordered}; first columns of values
for \code{matrix}; original values for \code{numeric}; and numeric times
within event statuses for \code{Surv}.  Numeric values are stratified into
quantile bins and categorical values into factor levels defined by
\code{\link{MLControl}}.

Resampling stratification variables may be specified manually for
\code{ModelFrames} upon creation with the \code{\link[=ModelFrame]{strata}}
argument in their constructor.  Resampling of this class is unstratified by
default.

Stratification variables may be designated in \code{recipe} specifications
with the \code{\link{role_case}} function.  Resampling will be unstratified
otherwise.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

## Factor response example

fo <- Species ~ .
control <- CVControl()

gbm_res1 <- resample(fo, iris, GBMModel(n.trees = 25), control)
gbm_res2 <- resample(fo, iris, GBMModel(n.trees = 50), control)
gbm_res3 <- resample(fo, iris, GBMModel(n.trees = 100), control)

summary(gbm_res1)
plot(gbm_res1)

res <- c(GBM1 = gbm_res1, GBM2 = gbm_res2, GBM3 = gbm_res3)
summary(res)
plot(res)
}

}
\seealso{
\code{\link{c}}, \code{\link{metrics}}, \code{\link{performance}},
\code{\link{plot}}, \code{\link{summary}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MLModel.R
\name{unMLModelFit}
\alias{unMLModelFit}
\title{Revert an MLModelFit Object}
\usage{
unMLModelFit(object)
}
\arguments{
\item{object}{model \link{fit} result.}
}
\value{
The supplied object with its \code{MLModelFit} classes and fields
removed.
}
\description{
Function to revert an \code{MLModelFit} object to its original class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response.R
\name{DiscreteVariate}
\alias{DiscreteVariate}
\alias{BinomialVariate}
\alias{NegBinomialVariate}
\alias{PoissonVariate}
\title{Discrete Variate Constructors}
\usage{
BinomialVariate(x = integer(), size = integer())

DiscreteVariate(x = integer(), min = -Inf, max = Inf)

NegBinomialVariate(x = integer())

PoissonVariate(x = integer())
}
\arguments{
\item{x}{numeric vector.}

\item{size}{number or numeric vector of binomial trials.}

\item{min, max}{minimum and maximum bounds for discrete numbers.}
}
\value{
\code{BinomialVariate} object class, \code{DiscreteVariate} that
inherits from \code{numeric}, or \code{NegBinomialVariate} or
\code{PoissonVariate} that inherit from \code{DiscreteVariate}.
}
\description{
Create a variate of binomial counts, discrete numbers, negative binomial
counts, or Poisson counts.
}
\examples{
BinomialVariate(rbinom(25, 10, 0.5), size = 10)
PoissonVariate(rpois(25, 10))

}
\seealso{
\code{\link{role_binom}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/varimp.R
\name{varimp}
\alias{varimp}
\title{Variable Importance}
\usage{
varimp(
  object,
  method = c("permute", "model"),
  scale = TRUE,
  sort = c("decreasing", "increasing", "asis"),
  ...
)
}
\arguments{
\item{object}{model \link{fit} result.}

\item{method}{character string specifying the calculation of variable
importance as permutation-base (\code{"permute"}) or model-specific
(\code{"model"}).  If model-specific importance is specified but not
defined, the permutation-based method will be used instead with its default
values (below).  Permutation-based variable importance is defined as the
relative change in model predictive performances between datasets with and
without permuted values for the associated variable (Fisher et al. 2019).}

\item{scale}{logical value or vector indicating whether importance values are
scaled to a maximum of 100.}

\item{sort}{character string specifying the sort order of importance values
to be \code{"decreasing"}, \code{"increasing"}, or as predictors appear in
the model formula (\code{"asis"}).}

\item{...}{arguments passed to model-specific or permutation-based variable
importance functions.  These include the following arguments and default
values for \code{method = "permute"}.
\describe{
  \item{\code{select = NULL}}{expression indicating predictor variables for
    which to compute variable importance (see \code{\link[base]{subset}}
    for syntax) [default: all].}
  \item{\code{samples = 1}}{number of times to permute the values of each
    variable.  Larger numbers of samples decrease variability in the
    estimates at the expense of increased computation time.}
  \item{\code{prop = numeric()}}{proportion of observations to sample
    without replacement at each round of variable permutations [default:
    all].  Subsampling of observations can decrease computation time.}
  \item{\code{size = integer()}}{number of observations to sample at each
    round of permutations [default: all].}
  \item{\code{times = numeric()}}{numeric vector of follow-up times at
    which to predict survival probabilities or \code{NULL} for predicted
    survival means.}
  \item{\code{metric = NULL}}{\link[=metrics]{metric} function or function
    name with which to calculate performance.  If not specified, the first
    applicable default metric from the \link{performance} functions is
    used.}
  \item{\code{compare = c("-", "/")}}{character specifying the relative
    change to compute in comparing model predictive performances between
    datasets with and without permuted values.  The choices are difference
    (\code{"-"}) and ratio (\code{"/"}).}
  \item{\code{stats = MachineShop::settings("stat.TrainingParams")}}{
    function, function name, or vector of these with which to compute
    summary statistics on the set of variable importance values from the
    permuted datasets.}
  \item{\code{na.rm = TRUE}}{logical indicating whether to exclude missing
    variable importance values from the calculation of summary statistics.}
  \item{\code{progress = TRUE}}{logical indicating whether to display
    iterative progress during computation.}
}}
}
\value{
\code{VariableImportance} class object.
}
\description{
Calculate measures of relative importance for model predictor variables.
}
\details{
The \code{varimp} function supports calculation of variable importance with
the permutation-based method of Fisher et al. (2019) or with model-based
methods where defined.  Permutation-based importance is the default and has
the advantages of being available for any model, any performance metric
defined for the associated response variable type, and any predictor variable
in the original training dataset.  Conversely, model-specific importance is
not defined for some models and will fall back to the permutation method in
such cases; is generally limited to metrics implemented in the source
packages of models; and may be computed on derived, rather than original,
predictor variables.  These disadvantages can make comparisons of
model-specific importance across different classes of models infeasible.  A
downside of the permutation-based approach is increased computation time.  To
counter this, the permutation algorithm can be run in parallel simply by
loading a parallel backend for the \pkg{foreach} package \code{\%dopar\%}
function, such as \pkg{doParallel} or \pkg{doSNOW}.

Permutation variable importance is interpreted as the contribution of a
predictor variable to the predictive performance of a model as measured by
the performance metric used in the calculation.  Importance of a predictor is
conditional on and, with the default scaling, relative to the values of all
other predictors in the analysis.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

## Survival response example
library(survival)

gbm_fit <- fit(Surv(time, status) ~ ., data = veteran, model = GBMModel)
(vi <- varimp(gbm_fit))
plot(vi)
}

}
\references{
Fisher, A., Rudin, C., & Dominici, F. (2019). All models are wrong, but many
are useful: Learning a variable's importance by studying an entire class of
prediction models simultaneously. \emph{Journal of Machine Learning
Research}, \emph{20}, 1-81.
}
\seealso{
\code{\link{plot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_CForestModel.R
\name{CForestModel}
\alias{CForestModel}
\title{Conditional Random Forest Model}
\usage{
CForestModel(
  teststat = c("quad", "max"),
  testtype = c("Univariate", "Teststatistic", "Bonferroni", "MonteCarlo"),
  mincriterion = 0,
  ntree = 500,
  mtry = 5,
  replace = TRUE,
  fraction = 0.632
)
}
\arguments{
\item{teststat}{character specifying the type of the test statistic to be
applied.}

\item{testtype}{character specifying how to compute the distribution of the
test statistic.}

\item{mincriterion}{value of the test statistic that must be exceeded in
order to implement a split.}

\item{ntree}{number of trees to grow in a forest.}

\item{mtry}{number of input variables randomly sampled as candidates at each
node for random forest like algorithms.}

\item{replace}{logical indicating whether sampling of observations is done
with or without replacement.}

\item{fraction}{fraction of number of observations to draw without
replacement (only relevant if \code{replace = FALSE}).}
}
\value{
\code{MLModel} class object.
}
\description{
An implementation of the random forest and bagging ensemble algorithms
utilizing conditional inference trees as base learners.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{numeric}, \code{Surv}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameter:}{
    \code{mtry}
  }
}

Supplied arguments are passed to \code{\link[party]{cforest_control}}.
Further model details can be found in the source link below.
}
\examples{
fit(sale_amount ~ ., data = ICHomes, model = CForestModel)

}
\seealso{
\code{\link[party]{cforest}}, \code{\link{fit}},
\code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MLModel.R
\name{MLModel}
\alias{MLModel}
\alias{MLModelFunction}
\title{MLModel and MLModelFunction Class Constructors}
\usage{
MLModel(
  name = "MLModel",
  label = name,
  packages = character(),
  response_types = character(),
  weights = FALSE,
  predictor_encoding = c(NA, "model.frame", "model.matrix"),
  na.rm = FALSE,
  params = list(),
  gridinfo = tibble::tibble(param = character(), get_values = list(), default =
    logical()),
  fit = function(formula, data, weights, ...) stop("No fit function."),
  predict = function(object, newdata, times, ...) stop("No predict function."),
  varimp = function(object, ...) NULL,
  ...
)

MLModelFunction(object, ...)
}
\arguments{
\item{name}{character name of the object to which the model is assigned.}

\item{label}{optional character descriptor for the model.}

\item{packages}{character vector of package names upon which the model
depends.  Each name may be optionally followed by a comment in
parentheses specifying a version requirement.  The comment should contain
a comparison operator, whitespace and a valid version number, e.g.
\code{"xgboost (>= 1.3.0)"}.}

\item{response_types}{character vector of response variable types to which
the model can be fit.  Supported types are \code{"binary"},
\code{"BinomialVariate"}, \code{"DiscreteVariate"}, \code{"factor"},
\code{"matrix"}, \code{"NegBinomialVariate"}, \code{"numeric"},
\code{"ordered"}, \code{"PoissonVariate"}, and \code{"Surv"}.}

\item{weights}{logical value or vector of the same length as
\code{response_types} indicating whether case weights are supported for
the responses.}

\item{predictor_encoding}{character string indicating whether the model is
fit with predictor variables encoded as a \code{"\link{model.frame}"},
a \code{"\link{model.matrix}"}, or unspecified (default).}

\item{na.rm}{character string or logical specifying removal of \code{"all"}
(\code{TRUE}) cases with missing values from model fitting and prediction,
\code{"none"} (\code{FALSE}), or only those whose missing values are in the
\code{"response"} variable.}

\item{params}{list of user-specified model parameters to be passed to the
\code{fit} function.}

\item{gridinfo}{tibble of information for construction of tuning grids
consisting of a character column \code{param} with the names of parameters
in the grid, a list column \code{get_values} with functions to generate grid
points for the corresponding parameters, and an optional logical column
\code{default} indicating which parameters to include by default in regular
grids.  Values functions may optionally include arguments \code{n} and
\code{data} for the number of grid points to generate and a
\code{\link{ModelFrame}} of the model fit data and formula, respectively;
and must include an ellipsis (\code{...}).}

\item{fit}{model fitting function whose arguments are a \code{formula}, a
\code{\link{ModelFrame}} named \code{data}, case \code{weights}, and an
ellipsis.}

\item{predict}{model prediction function whose arguments are the
\code{object} returned by \code{fit}, a \code{\link{ModelFrame}} named
\code{newdata} of predictor variables, optional vector of \code{times} at
which to predict survival, and an ellipsis.}

\item{varimp}{variable importance function whose arguments are the
\code{object} returned by \code{fit}, optional arguments passed from calls
to \code{\link{varimp}}, and an ellipsis.}

\item{...}{arguments passed to other methods.}

\item{object}{function that returns an \code{MLModel} object when called
without any supplied argument values.}
}
\value{
An \code{MLModel} or \code{MLModelFunction} class object.
}
\description{
Create a model or model function for use with the \pkg{MachineShop} package.
}
\details{
If supplied, the \code{grid} function should return a list whose elements are
named after and contain values of parameters to include in a tuning grid to
be constructed automatically by the package.

Arguments \code{data} and \code{newdata} in the \code{fit} and \code{predict}
functions may be converted to data frames with \code{as.data.frame()}
if needed for their operation.  The \code{fit} function should return the
object resulting from the model fit.  Values returned by the \code{predict}
functions should be formatted according to the response variable types below.
\describe{
  \item{factor}{matrix whose columns contain the probabilities for
    multi-level factors or vector of probabilities for the second level of
    binary factors.}
  \item{matrix}{matrix of predicted responses.}
  \item{numeric}{vector or column matrix of predicted responses.}
  \item{Surv}{matrix whose columns contain survival probabilities at
    \code{times} if supplied or a vector of predicted survival means
    otherwise.}
}

The \code{varimp} function should return a vector of importance values named
after the predictor variables or a matrix or data frame whose rows are named
after the predictors.

The \code{predict} and \code{varimp} functions are additionally passed a list
named \code{.MachineShop} containing the \code{\link[=inputs]{input}}
and \code{\link[=models]{model}} from \code{\link{fit}}.  This argument may
be included in the function definitions as needed for their implementations.
Otherwise, it will be captured by the ellipsis.
}
\examples{
## Logistic regression model
LogisticModel <- MLModel(
  name = "LogisticModel",
  response_types = "binary",
  weights = TRUE,
  fit = function(formula, data, weights, ...) {
    glm(formula, data = as.data.frame(data), weights = weights,
        family = binomial, ...)
  },
  predict = function(object, newdata, ...) {
    predict(object, newdata = as.data.frame(newdata), type = "response")
  },
  varimp = function(object, ...) {
    pchisq(coef(object)^2 / diag(vcov(object)), 1)
  }
)

data(Pima.tr, package = "MASS")
res <- resample(type ~ ., data = Pima.tr, model = LogisticModel)
summary(res)

}
\seealso{
\code{\link{models}}, \code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expand.R
\name{expand_model}
\alias{expand_model}
\title{Model Expansion Over Tuning Parameters}
\usage{
expand_model(object, ..., random = FALSE)
}
\arguments{
\item{object}{\link[=models]{model} function, function name, or object; or
another object that can be \link[=as.MLModel]{coerced} to a model.}

\item{...}{named vectors or factors or a list of these containing the
parameter values over which to expand \code{object}.}

\item{random}{number of points to be randomly sampled from the parameter grid
or \code{FALSE} if all points are to be returned.}
}
\value{
\code{list} of expanded models.
}
\description{
Expand a model over all combinations of a grid of tuning parameters.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

data(Boston, package = "MASS")

models <- expand_model(GBMModel, n.trees = c(50, 100),
                                 interaction.depth = 1:2)

fit(medv ~ ., data = Boston, model = SelectedModel(models))
}

}
\seealso{
\code{\link{SelectedModel}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid.R
\name{ParameterGrid}
\alias{ParameterGrid}
\alias{ParameterGrid.param}
\alias{ParameterGrid.list}
\alias{ParameterGrid.parameters}
\title{Tuning Parameters Grid}
\usage{
ParameterGrid(...)

\method{ParameterGrid}{param}(..., size = 3, random = FALSE)

\method{ParameterGrid}{list}(object, size = 3, random = FALSE, ...)

\method{ParameterGrid}{parameters}(object, size = 3, random = FALSE, ...)
}
\arguments{
\item{...}{named \code{param} objects as defined in the \pkg{dials} package.}

\item{size}{single integer or vector of integers whose positions or names
match the given parameters and which specify the number of values used to
construct the grid.}

\item{random}{number of unique points to sample at random from the grid
defined by \code{size}, or \code{FALSE} for all points.}

\item{object}{list of named \code{param} objects or a
\code{\link[dials]{parameters}} object.  This is a positional argument
that must be given first in calls to its methods.}
}
\value{
\code{ParameterGrid} class object that inherits from
\code{parameters} and \code{TuningGrid}.
}
\description{
Defines a tuning grid from a set of parameters.
}
\examples{
## GBMModel tuning parameters
grid <- ParameterGrid(
  n.trees = dials::trees(),
  interaction.depth = dials::tree_depth(),
  random = 5
)
TunedModel(GBMModel, grid = grid)

}
\seealso{
\code{\link{TunedModel}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MLMetric.R
\name{MLMetric}
\alias{MLMetric}
\alias{MLMetric<-}
\title{MLMetric Class Constructor}
\usage{
MLMetric(object, name = "MLMetric", label = name, maximize = TRUE)

MLMetric(object) <- value
}
\arguments{
\item{object}{function to compute the metric, defined to accept
\code{observed} and \code{predicted} as the first two arguments and with an
ellipsis (\code{...}) to accommodate others.}

\item{name}{character name of the object to which the metric is assigned.}

\item{label}{optional character descriptor for the model.}

\item{maximize}{logical indicating whether higher values of the metric
correspond to better predictive performance.}

\item{value}{list of arguments to pass to the \code{MLMetric} constructor.}
}
\value{
\code{MLMetric} class object.
}
\description{
Create a performance metric for use with the \pkg{MachineShop} package.
}
\examples{
f2_score <- MLMetric(
  function(observed, predicted, ...) {
    f_score(observed, predicted, beta = 2, ...)
  },
  name = "f2_score",
  label = "F Score (beta = 2)",
  maximize = TRUE
)

}
\seealso{
\code{\link{metrics}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/step_kmedoids.R
\name{step_kmedoids}
\alias{step_kmedoids}
\alias{tidy.step_kmedoids}
\alias{tunable.step_kmedoids}
\title{K-Medoids Clustering Variable Selection}
\usage{
step_kmedoids(
  recipe,
  ...,
  k = 5,
  center = TRUE,
  scale = TRUE,
  method = c("pam", "clara"),
  metric = "euclidean",
  optimize = FALSE,
  num_samp = 50,
  samp_size = 40 + 2 * k,
  replace = TRUE,
  prefix = "KMedoids",
  role = "predictor",
  skip = FALSE,
  id = recipes::rand_id("kmedoids")
)

\method{tidy}{step_kmedoids}(x, ...)

\method{tunable}{step_kmedoids}(x, ...)
}
\arguments{
\item{recipe}{\link[recipes]{recipe} object to which the step will be added.}

\item{...}{one or more selector functions to choose which variables will be
used to compute the components.  See \code{\link[recipes]{selections}} for
more details.  These are not currently used by the \code{tidy} method.}

\item{k}{number of k-medoids clusterings of the variables.  The value of
\code{k} is constrained to be between 1 and one less than the number of
original variables.}

\item{center, scale}{logicals indicating whether to mean center and median
absolute deviation scale the original variables prior to cluster
partitioning, or functions or names of functions for the centering and
scaling; not applied to selected variables.}

\item{method}{character string specifying one of the clustering methods
provided by the \pkg{cluster} package.  The \code{clara} (clustering
large applications) method is an extension of \code{pam} (partitioning
around medoids) designed to handle large datasets.}

\item{metric}{character string specifying the distance metric for calculating
dissimilarities between observations as \code{"euclidean"},
\code{"manhattan"}, or \code{"jaccard"} (\code{clara} only).}

\item{optimize}{logical indicator or 0:5 integer level specifying
optimization for the \code{\link[cluster]{pam}} clustering method.}

\item{num_samp}{number of sub-datasets to sample for the
\code{\link[cluster]{clara}} clustering method.}

\item{samp_size}{number of cases to include in each sub-dataset.}

\item{replace}{logical indicating whether to replace the original variables.}

\item{prefix}{if the original variables are not replaced, the selected
variables are added to the dataset with the character string prefix added
to their names; otherwise, the original variable names are retained.}

\item{role}{analysis role that added step variables should be assigned.  By
default, they are designated as model predictors.}

\item{skip}{logical indicating whether to skip the step when the recipe is
baked.  While all operations are baked when \code{\link[recipes]{prep}} is
run, some operations may not be applicable to new data (e.g. processing
outcome variables).  Care should be taken when using \code{skip = TRUE} as
it may affect the computations for subsequent operations.}

\item{id}{unique character string to identify the step.}

\item{x}{\code{step_kmedoids} object.}
}
\value{
Function \code{step_kmedoids} creates a new step whose class is of
the same name and inherits from \code{\link{step_sbf}}, adds it to the
sequence of existing steps (if any) in the recipe, and returns the updated
recipe.  For the \code{tidy} method, a tibble with columns \code{terms}
(selectors or variables selected), \code{cluster} assignments,
\code{selected} (logical indicator of selected cluster medoids),
\code{silhouette} (silhouette values), and \code{name} of the selected
variable names.
}
\description{
Creates a \emph{specification} of a recipe step that will partition numeric
variables according to k-medoids clustering and select the cluster medoids.
}
\details{
K-medoids clustering partitions variables into k groups such that the
dissimilarity between the variables and their assigned cluster medoids is
minimized.  Cluster medoids are then returned as a set of k variables.
}
\examples{
\donttest{
## Requires prior installation of suggested package cluster to run

library(recipes)

rec <- recipe(rating ~ ., data = attitude)
kmedoids_rec <- rec \%>\%
  step_kmedoids(all_predictors(), k = 3)
kmedoids_prep <- prep(kmedoids_rec, training = attitude)
kmedoids_data <- bake(kmedoids_prep, attitude)

pairs(kmedoids_data, lower.panel = NULL)

tidy(kmedoids_rec, number = 1)
tidy(kmedoids_prep, number = 1)
}

}
\references{
Kaufman, L., & Rousseeuw, P. J. (1990). \emph{Finding groups in data: An
introduction to cluster analysis}. Wiley.

Reynolds, A., Richards, G., de la Iglesia, B., & Rayward-Smith, V. (1992).
Clustering rules: A comparison of partitioning and hierarchical clustering
algorithms. \emph{Journal of Mathematical Modelling and Algorithms},
\emph{5}, 475-504.
}
\seealso{
\code{\link[cluster]{pam}}, \code{\link[cluster]{clara}},
\code{\link[recipes]{recipe}}, \code{\link[recipes]{prep}},
\code{\link[recipes]{bake}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_GLMNetModel.R
\name{GLMNetModel}
\alias{GLMNetModel}
\title{GLM Lasso or Elasticnet Model}
\usage{
GLMNetModel(
  family = NULL,
  alpha = 1,
  lambda = 0,
  standardize = TRUE,
  intercept = logical(),
  penalty.factor = .(rep(1, nvars)),
  standardize.response = FALSE,
  thresh = 1e-07,
  maxit = 1e+05,
  type.gaussian = .(if (nvars < 500) "covariance" else "naive"),
  type.logistic = c("Newton", "modified.Newton"),
  type.multinomial = c("ungrouped", "grouped"),
  cox.ties = NULL,
  control = list()
)
}
\arguments{
\item{family}{optional response type.  Set automatically according to the
class type of the response variable.}

\item{alpha}{elasticnet mixing parameter.}

\item{lambda}{regularization parameter.  The default value \code{lambda = 0}
performs no regularization and should be increased to avoid model fitting
issues if the number of predictor variables is greater than the number of
observations.}

\item{standardize}{logical flag for predictor variable standardization, prior
to model fitting.}

\item{intercept}{logical indicating whether to fit intercepts.}

\item{penalty.factor}{vector of penalty factors to be applied to each
coefficient.}

\item{standardize.response}{logical indicating whether to standardize
\code{"mgaussian"} response variables.}

\item{thresh}{convergence threshold for coordinate descent [deprecated].}

\item{maxit}{maximum number of passes over the data for all lambda values
[deprecated].}

\item{type.gaussian}{algorithm type for guassian models.}

\item{type.logistic}{algorithm type for logistic models.}

\item{type.multinomial}{algorithm type for multinomial models.}

\item{cox.ties}{character string of \code{"efron"} or \code{"breslow"} for
the method of handling ties in Cox survival models.}

\item{control}{named list of algorithm control parameters, providing
per-call overrides of session defaults set by
\code{\link[glmnet]{glmnet.control}}.}
}
\value{
\code{MLModel} class object.
}
\description{
Fit a generalized linear model via penalized maximum likelihood.
}
\details{
\describe{
  \item{Response types:}{\code{BinomialVariate}, \code{factor},
    \code{matrix}, \code{numeric}, \code{PoissonVariate}, \code{Surv}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \code{lambda}, \code{alpha}
  }
}

Default argument values and further model details can be found in the source
See Also link below.
}
\examples{
\donttest{
## Requires prior installation of suggested package glmnet to run

fit(sale_amount ~ ., data = ICHomes, model = GLMNetModel(lambda = 0.01))
}

}
\seealso{
\code{\link[glmnet]{glmnet}}, \code{\link{fit}},
\code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/performance.R
\name{performance}
\alias{performance}
\alias{performance.BinomialVariate}
\alias{performance.factor}
\alias{performance.matrix}
\alias{performance.numeric}
\alias{performance.Surv}
\alias{performance.ConfusionList}
\alias{performance.ConfusionMatrix}
\alias{performance.MLModel}
\alias{performance.Resample}
\alias{performance.TrainingStep}
\title{Model Performance Metrics}
\usage{
performance(x, ...)

\method{performance}{BinomialVariate}(
  x,
  y,
  weights = NULL,
  metrics = MachineShop::settings("metrics.numeric"),
  na.rm = TRUE,
  ...
)

\method{performance}{factor}(
  x,
  y,
  weights = NULL,
  metrics = MachineShop::settings("metrics.factor"),
  cutoff = MachineShop::settings("cutoff"),
  na.rm = TRUE,
  ...
)

\method{performance}{matrix}(
  x,
  y,
  weights = NULL,
  metrics = MachineShop::settings("metrics.matrix"),
  na.rm = TRUE,
  ...
)

\method{performance}{numeric}(
  x,
  y,
  weights = NULL,
  metrics = MachineShop::settings("metrics.numeric"),
  na.rm = TRUE,
  ...
)

\method{performance}{Surv}(
  x,
  y,
  weights = NULL,
  metrics = MachineShop::settings("metrics.Surv"),
  cutoff = MachineShop::settings("cutoff"),
  na.rm = TRUE,
  ...
)

\method{performance}{ConfusionList}(x, ...)

\method{performance}{ConfusionMatrix}(x, metrics = MachineShop::settings("metrics.ConfusionMatrix"), ...)

\method{performance}{MLModel}(x, ...)

\method{performance}{Resample}(x, ...)

\method{performance}{TrainingStep}(x, ...)
}
\arguments{
\item{x}{\link[=response]{observed responses}; or \link{confusion}, trained
model \link{fit}, \link{resample}, or \link{rfe} result.}

\item{...}{arguments passed from the \code{Resample} method to the response
type-specific methods or from the method for \code{ConfusionList} to
\code{ConfusionMatrix}.  Elliptical arguments in the response
type-specific methods are passed to \code{metrics} supplied as a single
\code{\link[=metrics]{MLMetric}} function and are ignored otherwise.}

\item{y}{\link[=predict]{predicted responses} if not contained in \code{x}.}

\item{weights}{numeric vector of non-negative
\link[=case_weights]{case weights} for the observed \code{x} responses
[default: equal weights].}

\item{metrics}{\link[=metrics]{metric} function, function name, or vector of
these with which to calculate performance.}

\item{na.rm}{logical indicating whether to remove observed or predicted
responses that are \code{NA} when calculating metrics.}

\item{cutoff}{numeric (0, 1) threshold above which binary factor
probabilities are classified as events and below which survival
probabilities are classified.}
}
\description{
Compute measures of model performance.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

res <- resample(Species ~ ., data = iris, model = GBMModel)
(perf <- performance(res))
summary(perf)
plot(perf)

## Survival response example
library(survival)

gbm_fit <- fit(Surv(time, status) ~ ., data = veteran, model = GBMModel)

obs <- response(gbm_fit, newdata = veteran)
pred <- predict(gbm_fit, newdata = veteran)
performance(obs, pred)
}

}
\seealso{
\code{\link{plot}}, \code{\link{summary}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_XGBModel.R
\name{XGBModel}
\alias{XGBModel}
\alias{XGBDARTModel}
\alias{XGBLinearModel}
\alias{XGBTreeModel}
\title{Extreme Gradient Boosting Models}
\usage{
XGBModel(
  nrounds = 100,
  ...,
  objective = character(),
  aft_loss_distribution = "normal",
  base_score = 0.5,
  verbose = 0,
  print_every_n = 1
)

XGBDARTModel(
  eta = 0.3,
  gamma = 0,
  max_depth = 6,
  min_child_weight = 1,
  max_delta_step = .(0.7 * is(y, "PoissonVariate")),
  subsample = 1,
  colsample_bytree = 1,
  colsample_bylevel = 1,
  colsample_bynode = 1,
  alpha = 0,
  lambda = 1,
  tree_method = "auto",
  scale_pos_weight = 1,
  refresh_leaf = 1,
  grow_policy = "depthwise",
  max_leaves = 0,
  max_bin = 256,
  num_parallel_tree = 1,
  sample_type = "uniform",
  normalize_type = "tree",
  rate_drop = 0,
  one_drop = 0,
  skip_drop = 0,
  ...
)

XGBLinearModel(
  alpha = 0,
  lambda = 0,
  updater = "shotgun",
  feature_selector = "cyclic",
  top_k = 0,
  ...
)

XGBTreeModel(
  eta = 0.3,
  gamma = 0,
  max_depth = 6,
  min_child_weight = 1,
  max_delta_step = .(0.7 * is(y, "PoissonVariate")),
  subsample = 1,
  colsample_bytree = 1,
  colsample_bylevel = 1,
  colsample_bynode = 1,
  alpha = 0,
  lambda = 1,
  tree_method = "auto",
  scale_pos_weight = 1,
  refresh_leaf = 1,
  grow_policy = "depthwise",
  max_leaves = 0,
  max_bin = 256,
  num_parallel_tree = 1,
  ...
)
}
\arguments{
\item{nrounds}{number of boosting iterations.}

\item{...}{model parameters as described below and in the XGBoost
\href{https://xgboost.readthedocs.io/en/latest/parameter.html}{documentation}
and arguments passed to \code{XGBModel} from the other constructors.}

\item{objective}{optional character string defining the learning task and
objective.  Set automatically if not specified according to the following
values available for supported response variable types.
\describe{
  \item{\code{factor}:}{\code{"multi:softprob"}, \code{"binary:logistic"}
    (2 levels only)}
  \item{\code{numeric}:}{\code{"reg:squarederror"}, \code{"reg:logistic"},
    \code{"reg:gamma"}, \code{"reg:tweedie"}, \code{"rank:pairwise"},
    \code{"rank:ndcg"}, \code{"rank:map"}}
  \item{\code{PoissonVariate}:}{\code{"count:poisson"}}
  \item{\code{Surv}:}{\code{"survival:aft"}, \code{"survival:cox"}}
}
The first values listed are the defaults for the corresponding response
types.}

\item{aft_loss_distribution}{character string specifying a distribution for
the accelerated failure time objective (\code{"survival:aft"}) as
\code{"extreme"}, \code{"logistic"}, or \code{"normal"}.}

\item{base_score}{initial prediction score of all observations, global bias.}

\item{verbose}{numeric value controlling the amount of output printed
during model fitting, such that 0 = none, 1 = performance information, and
2 = additional information.}

\item{print_every_n}{numeric value designating the fitting iterations at
at which to print output when \code{verbose > 0}.}

\item{eta}{shrinkage of variable weights at each iteration to prevent
overfitting.}

\item{gamma}{minimum loss reduction required to split a tree node.}

\item{max_depth}{maximum tree depth.}

\item{min_child_weight}{minimum sum of observation weights required of nodes.}

\item{max_delta_step, tree_method, scale_pos_weight, updater, refresh_leaf, grow_policy, max_leaves, max_bin, num_parallel_tree}{other tree booster parameters.}

\item{subsample}{subsample ratio of the training observations.}

\item{colsample_bytree, colsample_bylevel, colsample_bynode}{subsample ratio of
variables for each tree, level, or split.}

\item{alpha, lambda}{L1 and L2 regularization terms for variable weights.}

\item{sample_type, normalize_type}{type of sampling and normalization
algorithms.}

\item{rate_drop}{rate at which to drop trees during the dropout procedure.}

\item{one_drop}{integer indicating whether to drop at least one tree during
the dropout procedure.}

\item{skip_drop}{probability of skipping the dropout procedure during a
boosting iteration.}

\item{feature_selector, top_k}{character string specifying the feature
selection and ordering method, and number of top variables to select in the
\code{"greedy"} and \code{"thrifty"} feature selectors.}
}
\value{
\code{MLModel} class object.
}
\description{
Fits models with an efficient implementation of the gradient boosting
framework from Chen & Guestrin.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{numeric},
    \code{PoissonVariate}, \code{Surv}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \itemize{
      \item XGBModel: \code{NULL}
      \item XGBDARTModel: \code{nrounds}, \code{eta}*, \code{gamma}*,
        \code{max_depth}, \code{min_child_weight}*, \code{subsample}*,
        \code{colsample_bytree}*, \code{rate_drop}*, \code{skip_drop}*
      \item XGBLinearModel: \code{nrounds}, \code{alpha}, \code{lambda}
      \item XGBTreeModel: \code{nrounds}, \code{eta}*, \code{gamma}*,
        \code{max_depth}, \code{min_child_weight}*, \code{subsample}*,
        \code{colsample_bytree}*
    }
  }
}
* excluded from grids by default

The booster-specific constructor functions \code{XGBDARTModel},
\code{XGBLinearModel}, and \code{XGBTreeModel} are special cases of
\code{XGBModel} which automatically set the XGBoost \code{booster}
\href{https://xgboost.readthedocs.io/en/latest/parameter.html}{parameter}.
These are called directly in typical usage unless \code{XGBModel} is needed
to specify a more general model.

Default argument values and further model details can be found in the source
See Also link below.

In calls to \code{\link{varimp}} for \code{XGBTreeModel}, argument
\code{type} may be specified as \code{"Gain"} (default) for the fractional
contribution of each predictor to the total gain of its splits, as
\code{"Cover"} for the number of observations related to each predictor, or
as \code{"Frequency"} for the percentage of times each predictor is used in
the trees.  Variable importance is automatically scaled to range from 0 to
100.  To obtain unscaled importance values, set \code{scale = FALSE}.  See
example below.
}
\examples{
\donttest{
## Requires prior installation of suggested package xgboost to run

model_fit <- fit(
  Species ~ ., data = iris, model = XGBTreeModel(nthread = 1)
)
varimp(model_fit, method = "model", type = "Frequency", scale = FALSE)
}

}
\seealso{
\code{\link[xgboost:xgb.train]{xgboost}}, \code{\link{fit}},
\code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_NaiveBayesModel.R
\name{NaiveBayesModel}
\alias{NaiveBayesModel}
\title{Naive Bayes Classifier Model}
\usage{
NaiveBayesModel(laplace = 0)
}
\arguments{
\item{laplace}{positive numeric controlling Laplace smoothing.}
}
\value{
\code{MLModel} class object.
}
\description{
Computes the conditional a-posterior probabilities of a categorical class
variable given independent predictor variables using Bayes rule.
}
\details{
\describe{
  \item{Response types:}{\code{factor}}
}

Further model details can be found in the source link below.
}
\examples{
\donttest{
## Requires prior installation of suggested package e1071 to run

fit(Species ~ ., data = iris, model = NaiveBayesModel)
}

}
\seealso{
\code{\link[e1071]{naiveBayes}}, \code{\link{fit}},
\code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_LARSModel.R
\name{LARSModel}
\alias{LARSModel}
\title{Least Angle Regression, Lasso and Infinitesimal Forward Stagewise Models}
\usage{
LARSModel(
  type = c("lasso", "lar", "forward.stagewise", "stepwise"),
  trace = FALSE,
  normalize = TRUE,
  intercept = TRUE,
  step = numeric(),
  use.Gram = TRUE
)
}
\arguments{
\item{type}{model type.}

\item{trace}{logical indicating whether status information is printed during
the fitting process.}

\item{normalize}{whether to standardize each variable to have unit L2 norm.}

\item{intercept}{whether to include an intercept in the model.}

\item{step}{algorithm step number to use for prediction.  May be a decimal
number indicating a fractional distance between steps.  If specified, the
maximum number of algorithm steps will be \code{ceiling(step)}; otherwise,
\code{step} will be set equal to the source package default maximum
[default: \code{max.steps}].}

\item{use.Gram}{whether to precompute the Gram matrix.}
}
\value{
\code{MLModel} class object.
}
\description{
Fit variants of Lasso, and provide the entire sequence of coefficients and
fits, starting from zero to the least squares fit.
}
\details{
\describe{
  \item{Response types:}{\code{numeric}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameter:}{
    \code{step}
  }
}

Default argument values and further model details can be found in the source
See Also link below.
}
\examples{
\donttest{
## Requires prior installation of suggested package lars to run

fit(sale_amount ~ ., data = ICHomes, model = LARSModel)
}

}
\seealso{
\code{\link[lars]{lars}}, \code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TrainingParams.R
\name{set_optim}
\alias{set_optim}
\alias{set_optim_bayes}
\alias{set_optim_bayes.ModelSpecification}
\alias{set_optim_bfgs}
\alias{set_optim_bfgs.ModelSpecification}
\alias{set_optim_grid}
\alias{set_optim_grid.TrainingParams}
\alias{set_optim_grid.ModelSpecification}
\alias{set_optim_grid.TunedInput}
\alias{set_optim_grid.TunedModel}
\alias{set_optim_pso}
\alias{set_optim_pso.ModelSpecification}
\alias{set_optim_sann}
\alias{set_optim_sann.ModelSpecification}
\alias{set_optim_method}
\alias{set_optim_method.ModelSpecification}
\title{Tuning Parameter Optimization}
\usage{
set_optim_bayes(object, ...)

\method{set_optim_bayes}{ModelSpecification}(
  object,
  num_init = 5,
  times = 10,
  each = 1,
  acquisition = c("ucb", "ei", "eips", "poi"),
  kappa = stats::qnorm(conf),
  conf = 0.995,
  epsilon = 0,
  control = list(),
  packages = c("ParBayesianOptimization", "rBayesianOptimization"),
  random = FALSE,
  progress = verbose,
  verbose = 0,
  ...
)

set_optim_bfgs(object, ...)

\method{set_optim_bfgs}{ModelSpecification}(
  object,
  times = 10,
  control = list(),
  random = FALSE,
  progress = FALSE,
  verbose = 0,
  ...
)

set_optim_grid(object, ...)

\method{set_optim_grid}{TrainingParams}(object, random = FALSE, progress = FALSE, ...)

\method{set_optim_grid}{ModelSpecification}(object, ...)

\method{set_optim_grid}{TunedInput}(object, ...)

\method{set_optim_grid}{TunedModel}(object, ...)

set_optim_pso(object, ...)

\method{set_optim_pso}{ModelSpecification}(
  object,
  times = 10,
  each = NULL,
  control = list(),
  random = FALSE,
  progress = FALSE,
  verbose = 0,
  ...
)

set_optim_sann(object, ...)

\method{set_optim_sann}{ModelSpecification}(
  object,
  times = 10,
  control = list(),
  random = FALSE,
  progress = FALSE,
  verbose = 0,
  ...
)

set_optim_method(object, ...)

\method{set_optim_method}{ModelSpecification}(
  object,
  fun,
  label = "Optimization Function",
  packages = character(),
  params = list(),
  random = FALSE,
  progress = FALSE,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{object}{\link[=inputs]{input} or \link[=models]{model} object.}

\item{...}{arguments passed to the \code{TrainingParams} method of
\code{set_optim_grid} from its other methods.}

\item{num_init}{number of grid points to sample for the initialization of
Bayesian optimization.}

\item{times}{maximum number of times to repeat the optimization step.
Multiple sets of model parameters are evaluated automatically at each step
of the BFGS algorithm to compute a finite-difference approximation to the
gradient.}

\item{each}{number of times to sample and evaluate model parameters at each
optimization step.  This is the swarm size in particle swarm optimization,
which defaults to \code{floor(10 + 2 * sqrt(length(bounds)))}.}

\item{acquisition}{character string specifying the acquisition function as
\code{"ucb"} (upper confidence bound), \code{"ei"} (expected improvement),
\code{"eips"} (expected improvement per second), or \code{"poi"}
(probability of improvement).}

\item{kappa, conf}{upper confidence bound (\code{"ucb"}) quantile or its
probability to balance exploitation against exploration.  Argument
\code{kappa} takes precedence if both are given and multiplies the
predictive standard deviation added to the predictive mean in the
acquisition function.  Larger values encourage exploration of the model
parameter space.}

\item{epsilon}{improvement methods (\code{"ei"}, \code{"eips"}, and
\code{"poi"}) parameter to balance exploitation against exploration.
Values should be between -0.1 and 0.1 with larger ones encouraging
exploration.}

\item{control}{list of control parameters passed to
\code{bayesOpt} by \code{set_optim_bayes} with package
\code{"ParBayesianOptimization"}, to
\code{\link[rBayesianOptimization]{BayesianOptimization}} by
\code{set_optim_bayes} with package \code{"rBayesianOptimization"}, to
\code{\link[stats]{optim}} by \code{set_optim_bfgs} and
\code{set_optim_sann}, and to \code{\link[pso]{psoptim}} by
\code{set_optim_pso}.}

\item{packages}{R package or packages to use for the optimization method, or
an empty vector if none are needed.  The first package in
\code{set_optim_bayes} is used unless otherwise specified by the user.}

\item{random}{number of points to sample for a random grid search, or
\code{FALSE} for an exhaustive grid search.  Used when a grid search is
specified or as the fallback method for non-numeric model parameters
present during other optimization methods.}

\item{progress}{logical indicating whether to display iterative progress
during optimization.}

\item{verbose}{numeric or logical value specifying the level of progress
detail to print, with 0 (\code{FALSE}) indicating none and 1 (\code{TRUE})
or higher indicating increasing amounts of detail.}

\item{fun}{user-defined optimization function to which the arguments below
are passed in order.  An ellipsis can be included in the function
definition when using only a subset of the arguments and ignoring others.
A tibble returned by the function with the same number of rows as model
evaluations will be included in a \code{TrainingStep} summary of
optimization results; other types of return values will be ignored.
\describe{
  \item{optim}{function that takes a numeric vector or list of named
    model parameters as the first argument, optionally accepts the maximum
    number of iterations as argument \code{max_iter}, and returns a scalar
    measure of performance to be maximized.  Parameter names are available
    from the \code{grid} and \code{bounds} arguments described below.  If
    the function cannot be evaluated at a given set of parameter values,
    then \code{-Inf} is returned.}
  \item{grid}{data frame containing a tuning grid of all model parameters.}
  \item{bounds}{named list of lower and upper bounds for each finite
    numeric model parameter in \code{grid}.  The types (integer or double)
    of the original parameter values are preserved in the bounds.}
  \item{params}{list of optimization parameters as supplied to
    \code{set_optim_method}.}
  \item{monitor}{list of the \code{progress} and \code{verbose} values.}
}}

\item{label}{character descriptor for the optimization method.}

\item{params}{list of user-specified model parameters to be passed to
\code{fun}.}
}
\value{
Argument \code{object} updated with the specified optimization method
and control parameters.
}
\description{
Set the optimization method and control parameters for tuning of model
parameters.
}
\details{
The optimization functions implement the following methods.
\describe{
  \item{\code{set_optim_bayes}}{Bayesian optimization with a Gaussian process
    model (Snoek et al. 2012).}
  \item{\code{set_optim_bfgs}}{limited-memory modification of quasi-Newton
    BFGS optimization (Byrd et al. 1995).}
  \item{\code{set_optim_grid}}{exhaustive or random grid search.}
  \item{\code{set_optim_pso}}{particle swarm optimization (Bratton and
    Kennedy 2007, Zambrano-Bigiarini et al. 2013).}
  \item{\code{set_optim_sann}}{simulated annealing (Belisle 1992).  This
    method depends critically on the control parameter settings.  It is
    not a general-purpose method but can be very useful in getting to good
    parameter values on a very rough optimization surface.}
  \item{\code{set_optim_method}}{user-defined optimization function.}
}
The package-defined optimization functions evaluate and return values of the
tuning parameters that are of same type (e.g. integer, double, character) as
given in the \code{object} grid.  Sequential optimization of numeric tuning
parameters is performed over a hypercube defined by their minimum and maximum
grid values.  Non-numeric parameters are optimized with grid searches.
}
\examples{
## Install ParBayesianOptimization package as needed
## pak::repo_add(smith = "https://brian-j-smith.github.io/drat")
## pak::pak("ParBayesianOptimization")

ModelSpecification(
  sale_amount ~ ., data = ICHomes,
  model = TunedModel(GBMModel)
) \%>\% set_optim_bayes

}
\references{
Belisle, C. J. P. (1992). Convergence theorems for a class of simulated
annealing algorithms on Rd. \emph{Journal of Applied Probability},
\emph{29}, 885–895.

Bratton, D. & Kennedy, J. (2007), Defining a standard for particle swarm
optimization. In \emph{IEEE Swarm Intelligence Symposium, 2007} (pp.
120-127).

Byrd, R. H., Lu, P., Nocedal, J., & Zhu, C. (1995). A limited memory
algorithm for bound constrained optimization. \emph{SIAM Journal on
Scientific Computing}, \emph{16}, 1190–1208.

Snoek, J., Larochelle, H., & Adams, R.P. (2012). Practical Bayesian
Optimization of Machine Learning Algorithms. arXiv:1206.2944 [stat.ML].

Zambrano-Bigiarini, M., Clerc, M., & Rojas, R. (2013). Standard particle
swarm optimisation 2011 at CEC-2013: A baseline for future PSO improvements.
In \emph{IEEE Congress on Evolutionary Computation, 2013} (pp. 2337-2344).
}
\seealso{
\code{\link[rBayesianOptimization]{BayesianOptimization}},
\code{bayesOpt}, \code{\link[stats]{optim}}, \code{\link[pso]{psoptim}},
\code{\link{set_monitor}}, \code{\link{set_predict}},
\code{\link{set_strata}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_LMModel.R
\name{LMModel}
\alias{LMModel}
\title{Linear Models}
\usage{
LMModel()
}
\value{
\code{MLModel} class object.
}
\description{
Fits linear models.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{matrix}, \code{numeric}}
}

Further model details can be found in the source link below.

In calls to \code{\link{varimp}} for \code{LModel}, numeric argument
\code{base} may be specified for the (negative) logarithmic transformation of
p-values [defaul: \code{exp(1)}].  Transformed p-values are automatically
scaled in the calculation of variable importance to range from 0 to 100.  To
obtain unscaled importance values, set \code{scale = FALSE}.
}
\examples{
fit(sale_amount ~ ., data = ICHomes, model = LMModel)

}
\seealso{
\code{\link[stats]{lm}}, \code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response.R
\name{SurvMatrix}
\alias{SurvMatrix}
\alias{SurvEvents}
\alias{SurvProbs}
\title{SurvMatrix Class Constructors}
\usage{
SurvEvents(data = NA, times = numeric(), distr = character())

SurvProbs(data = NA, times = numeric(), distr = character())
}
\arguments{
\item{data}{matrix, or object that can be coerced to one, with survival
events or probabilities at points in time in the columns and cases in the
rows.}

\item{times}{numeric vector of survival times for the columns.}

\item{distr}{character string specifying the survival distribution from which
the matrix values were derived.}
}
\value{
Object that is of the same class as the constructor name and inherits
from \code{SurvMatrix}.  Examples of these are predicted survival events and
probabilities returned by the \link{predict} function.
}
\description{
Create a matrix of survival events or probabilites.
}
\seealso{
\code{\link{performance}}, \code{\link{metrics}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit.R
\name{fit}
\alias{fit}
\alias{fit.formula}
\alias{fit.matrix}
\alias{fit.ModelFrame}
\alias{fit.recipe}
\alias{fit.ModelSpecification}
\alias{fit.MLModel}
\alias{fit.MLModelFunction}
\title{Model Fitting}
\usage{
fit(...)

\method{fit}{formula}(formula, data, model, ...)

\method{fit}{matrix}(x, y, model, ...)

\method{fit}{ModelFrame}(input, model, ...)

\method{fit}{recipe}(input, model, ...)

\method{fit}{ModelSpecification}(object, verbose = FALSE, ...)

\method{fit}{MLModel}(model, ...)

\method{fit}{MLModelFunction}(model, ...)
}
\arguments{
\item{...}{arguments passed from the generic function to its methods, from
the \code{MLModel} and \code{MLModelFunction} methods to first arguments of
others, and from others to the \code{ModelSpecification} method.  The
first argument of each \code{fit} method is positional and, as such, must
be given first in calls to them.}

\item{formula, data}{\link[=formula]{formula} defining the model predictor and
response variables and a \link[=data.frame]{data frame} containing them.}

\item{model}{\link[=models]{model} function, function name, or object; or
another object that can be \link[=as.MLModel]{coerced} to a model.  A model
can be given first followed by any of the variable specifications.}

\item{x, y}{\link{matrix} and object containing predictor and response
variables.}

\item{input}{\link[=inputs]{input} object defining and containing the model
predictor and response variables.}

\item{object}{model \link[=ModelSpecification]{specification}.}

\item{verbose}{logical indicating whether to display printed output generated
by some model-specific fit functions to aid in monitoring progress and
diagnosing errors.}
}
\value{
\code{MLModelFit} class object.
}
\description{
Fit a model to estimate its parameters from a data set.
}
\details{
User-specified case weights may be specified for \code{ModelFrames} upon
creation with the \code{\link[=ModelFrame]{weights}} argument in its
constructor.

Variables in \code{recipe} specifications may be designated as case weights
with the \code{\link{role_case}} function.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

## Survival response example
library(survival)

gbm_fit <- fit(Surv(time, status) ~ ., data = veteran, model = GBMModel)
varimp(gbm_fit)
}

}
\seealso{
\code{\link{as.MLModel}}, \code{\link{response}},
\code{\link{predict}}, \code{\link{varimp}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot}
\alias{plot}
\alias{plot.Calibration}
\alias{plot.ConfusionList}
\alias{plot.ConfusionMatrix}
\alias{plot.LiftCurve}
\alias{plot.MLModel}
\alias{plot.PartialDependence}
\alias{plot.Performance}
\alias{plot.PerformanceCurve}
\alias{plot.Resample}
\alias{plot.TrainingStep}
\alias{plot.VariableImportance}
\title{Model Performance Plots}
\usage{
\method{plot}{Calibration}(x, type = c("line", "point"), se = FALSE, ...)

\method{plot}{ConfusionList}(x, ...)

\method{plot}{ConfusionMatrix}(x, ...)

\method{plot}{LiftCurve}(
  x,
  find = numeric(),
  diagonal = TRUE,
  stat = MachineShop::settings("stat.Curve"),
  ...
)

\method{plot}{MLModel}(
  x,
  metrics = NULL,
  stat = MachineShop::settings("stat.TrainingParams"),
  type = c("boxplot", "density", "errorbar", "line", "violin"),
  ...
)

\method{plot}{PartialDependence}(x, stats = NULL, ...)

\method{plot}{Performance}(
  x,
  metrics = NULL,
  stat = MachineShop::settings("stat.Resample"),
  type = c("boxplot", "density", "errorbar", "violin"),
  ...
)

\method{plot}{PerformanceCurve}(
  x,
  type = c("tradeoffs", "cutoffs"),
  diagonal = FALSE,
  stat = MachineShop::settings("stat.Curve"),
  ...
)

\method{plot}{Resample}(
  x,
  metrics = NULL,
  stat = MachineShop::settings("stat.Resample"),
  type = c("boxplot", "density", "errorbar", "violin"),
  ...
)

\method{plot}{TrainingStep}(
  x,
  metrics = NULL,
  stat = MachineShop::settings("stat.TrainingParams"),
  type = c("boxplot", "density", "errorbar", "line", "violin"),
  ...
)

\method{plot}{VariableImportance}(x, n = Inf, ...)
}
\arguments{
\item{x}{\link{calibration}, \link{confusion}, \link{lift},
trained model \link{fit}, partial \link{dependence}, \link{performance},
\link[=curves]{performance curve}, \link{resample}, \link{rfe}, or
\link[=varimp]{variable importance} result.}

\item{type}{type of plot to construct.}

\item{se}{logical indicating whether to include standard error bars.}

\item{...}{arguments passed to other methods.}

\item{find}{numeric true positive rate at which to display reference lines
identifying the corresponding rates of positive predictions.}

\item{diagonal}{logical indicating whether to include a diagonal reference
line.}

\item{stat}{function or character string naming a function to compute a
summary statistic on resampled metrics for trained \code{MLModel} line
plots and \code{Resample} model ordering.  The original ordering is
preserved if a value of \code{NULL} is given.  For \code{LiftCurve} and
\code{PerformanceCurve} classes, plots are of resampled metrics aggregated
by the statistic if given or of resample-specific metrics if \code{NULL}.}

\item{metrics}{vector of numeric indexes or character names of performance
metrics to plot.}

\item{stats}{vector of numeric indexes or character names of partial
dependence summary statistics to plot.}

\item{n}{number of most important variables to include in the plot.}
}
\description{
Plot measures of model performance and predictor variable importance.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

## Factor response example

fo <- Species ~ .
control <- CVControl()

gbm_fit <- fit(fo, data = iris, model = GBMModel, control = control)
plot(varimp(gbm_fit))

gbm_res1 <- resample(fo, iris, GBMModel(n.trees = 25), control)
gbm_res2 <- resample(fo, iris, GBMModel(n.trees = 50), control)
gbm_res3 <- resample(fo, iris, GBMModel(n.trees = 100), control)
plot(gbm_res3)

res <- c(GBM1 = gbm_res1, GBM2 = gbm_res2, GBM3 = gbm_res3)
plot(res)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expand.R
\name{expand_steps}
\alias{expand_steps}
\title{Recipe Step Parameters Expansion}
\usage{
expand_steps(..., random = FALSE)
}
\arguments{
\item{...}{one or more lists containing parameter values over which to create
the grid.  For each list an argument name should be given as the \code{id}
of the \link[recipes]{recipe} step to which it corresponds.}

\item{random}{number of points to be randomly sampled from the parameter grid
or \code{FALSE} if all points are to be returned.}
}
\value{
\code{RecipeGrid} class object that inherits from \code{data.frame}.
}
\description{
Create a grid of parameter values from all combinations of lists supplied for
steps of a preprocessing recipe.
}
\examples{
library(recipes)
data(Boston, package = "MASS")

rec <- recipe(medv ~ ., data = Boston) \%>\%
  step_corr(all_numeric_predictors(), id = "corr") \%>\%
  step_pca(all_numeric_predictors(), id = "pca")

expand_steps(
  corr = list(threshold = c(0.8, 0.9),
              method = c("pearson", "spearman")),
  pca = list(num_comp = 1:3)
)

}
\seealso{
\code{\link{TunedInput}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metrics.R, R/metrics_factor.R,
%   R/metrics_numeric.R
\name{metrics}
\alias{metrics}
\alias{accuracy}
\alias{auc}
\alias{brier}
\alias{cindex}
\alias{cross_entropy}
\alias{f_score}
\alias{fnr}
\alias{fpr}
\alias{kappa2}
\alias{npv}
\alias{ppr}
\alias{ppv}
\alias{pr_auc}
\alias{precision}
\alias{recall}
\alias{roc_auc}
\alias{roc_index}
\alias{sensitivity}
\alias{specificity}
\alias{tnr}
\alias{tpr}
\alias{weighted_kappa2}
\alias{gini}
\alias{mae}
\alias{mse}
\alias{msle}
\alias{r2}
\alias{rmse}
\alias{rmsle}
\title{Performance Metrics}
\usage{
accuracy(
  observed,
  predicted = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  ...
)

auc(
  observed,
  predicted = NULL,
  weights = NULL,
  multiclass = c("pairs", "all"),
  metrics = c(MachineShop::tpr, MachineShop::fpr),
  stat = MachineShop::settings("stat.Curve"),
  ...
)

brier(observed, predicted = NULL, weights = NULL, ...)

cindex(observed, predicted = NULL, weights = NULL, ...)

cross_entropy(observed, predicted = NULL, weights = NULL, ...)

f_score(
  observed,
  predicted = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  beta = 1,
  ...
)

fnr(
  observed,
  predicted = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  ...
)

fpr(
  observed,
  predicted = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  ...
)

kappa2(
  observed,
  predicted = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  ...
)

npv(
  observed,
  predicted = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  ...
)

ppr(
  observed,
  predicted = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  ...
)

ppv(
  observed,
  predicted = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  ...
)

pr_auc(
  observed,
  predicted = NULL,
  weights = NULL,
  multiclass = c("pairs", "all"),
  ...
)

precision(
  observed,
  predicted = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  ...
)

recall(
  observed,
  predicted = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  ...
)

roc_auc(
  observed,
  predicted = NULL,
  weights = NULL,
  multiclass = c("pairs", "all"),
  ...
)

roc_index(
  observed,
  predicted = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  fun = function(sensitivity, specificity) (sensitivity + specificity)/2,
  ...
)

sensitivity(
  observed,
  predicted = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  ...
)

specificity(
  observed,
  predicted = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  ...
)

tnr(
  observed,
  predicted = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  ...
)

tpr(
  observed,
  predicted = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  ...
)

weighted_kappa2(observed, predicted = NULL, weights = NULL, power = 1, ...)

gini(observed, predicted = NULL, weights = NULL, ...)

mae(observed, predicted = NULL, weights = NULL, ...)

mse(observed, predicted = NULL, weights = NULL, ...)

msle(observed, predicted = NULL, weights = NULL, ...)

r2(
  observed,
  predicted = NULL,
  weights = NULL,
  method = c("mse", "pearson", "spearman"),
  distr = character(),
  ...
)

rmse(observed, predicted = NULL, weights = NULL, ...)

rmsle(observed, predicted = NULL, weights = NULL, ...)
}
\arguments{
\item{observed}{\link[=response]{observed responses}; or
\link{confusion}, \link[=curves]{performance curve}, or \link{resample}
result containing observed and predicted responses.}

\item{predicted}{\link[=predict]{predicted responses} if not contained in
\code{observed}.}

\item{weights}{numeric vector of non-negative
\link[=case_weights]{case weights} for the observed responses [default:
equal weights].}

\item{cutoff}{numeric (0, 1) threshold above which binary factor
probabilities are classified as events and below which survival
probabilities are classified.  If \code{NULL}, then confusion matrix-based
metrics are computed on predicted class probabilities if given.}

\item{...}{arguments passed to or from other methods.}

\item{multiclass}{character string specifying the method for computing
generalized area under the performance curve for multiclass factor
responses.  Options are to average over areas for each pair of classes
(\code{"pairs"}) or for each class versus all others (\code{"all"}).}

\item{metrics}{vector of two metric functions or function names that define a
curve under which to calculate area [default: ROC metrics].}

\item{stat}{function or character string naming a function to compute a
summary statistic at each cutoff value of resampled metrics in performance
curves, or \code{NULL} for resample-specific metrics.}

\item{beta}{relative importance of recall to precision in the calculation of
\code{f_score} [default: F1 score].}

\item{fun}{function to calculate a desired sensitivity-specificity tradeoff.}

\item{power}{power to which positional distances of off-diagonals from the
main diagonal in confusion matrices are raised to calculate
\code{weighted_kappa2}.}

\item{method}{character string specifying whether to compute \code{r2} as
the coefficient of determination (\code{"mse"}) or as the square of
\code{"pearson"} or \code{"spearman"} correlation.}

\item{distr}{character string specifying a distribution with which to
estimate the observed survival mean in the total sum of square component of
\code{r2}.  Possible values are \code{"empirical"} for the Kaplan-Meier
estimator, \code{"exponential"}, \code{"extreme"}, \code{"gaussian"},
\code{"loggaussian"}, \code{"logistic"}, \code{"loglogistic"},
\code{"lognormal"}, \code{"rayleigh"}, \code{"t"}, or \code{"weibull"}.
Defaults to the distribution that was used in predicting mean survival
times.}
}
\description{
Compute measures of agreement between observed and predicted responses.
}
\references{
Hand, D. J., & Till, R. J. (2001). A simple generalisation of the area under
the ROC curve for multiple class classification problems. \emph{Machine
Learning}, \emph{45}, 171-186.
}
\seealso{
\code{\link{metricinfo}}, \code{\link{performance}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{models}
\alias{models}
\title{Models}
\description{
Model constructor functions supplied by \pkg{MachineShop} are summarized in
the table below according to the types of response variables with which each
can be used.
\tabular{lccc}{
  \strong{Function} \tab \strong{Categorical} \tab \strong{Continuous}
  \tab \strong{Survival} \cr
  \code{\link{AdaBagModel}}         \tab f   \tab     \tab   \cr
  \code{\link{AdaBoostModel}}       \tab f   \tab     \tab   \cr
  \code{\link{BARTModel}}           \tab f   \tab n   \tab S \cr
  \code{\link{BARTMachineModel}}    \tab b   \tab n   \tab   \cr
  \code{\link{BlackBoostModel}}     \tab b   \tab n   \tab S \cr
  \code{\link{C50Model}}            \tab f   \tab     \tab   \cr
  \code{\link{CForestModel}}        \tab f   \tab n   \tab S \cr
  \code{\link{CoxModel}}            \tab     \tab     \tab S \cr
  \code{\link{CoxStepAICModel}}     \tab     \tab     \tab S \cr
  \code{\link{EarthModel}}          \tab f   \tab n   \tab   \cr
  \code{\link{FDAModel}}            \tab f   \tab     \tab   \cr
  \code{\link{GAMBoostModel}}       \tab b   \tab n   \tab S \cr
  \code{\link{GBMModel}}            \tab f   \tab n   \tab S \cr
  \code{\link{GLMBoostModel}}       \tab b   \tab n   \tab S \cr
  \code{\link{GLMModel}}            \tab f   \tab m,n \tab   \cr
  \code{\link{GLMStepAICModel}}     \tab b   \tab n   \tab   \cr
  \code{\link{GLMNetModel}}         \tab f   \tab m,n \tab S \cr
  \code{\link{KNNModel}}            \tab f,o \tab n   \tab   \cr
  \code{\link{LARSModel}}           \tab     \tab n   \tab   \cr
  \code{\link{LDAModel}}            \tab f   \tab     \tab   \cr
  \code{\link{LMModel}}             \tab f   \tab m,n \tab   \cr
  \code{\link{MDAModel}}            \tab f   \tab     \tab   \cr
  \code{\link{NaiveBayesModel}}     \tab f   \tab     \tab   \cr
  \code{\link{NNetModel}}           \tab f   \tab n   \tab   \cr
  \code{\link{ParsnipModel}}        \tab f   \tab m,n \tab S \cr
  \code{\link{PDAModel}}            \tab f   \tab     \tab   \cr
  \code{\link{PLSModel}}            \tab f   \tab n   \tab   \cr
  \code{\link{POLRModel}}           \tab o   \tab     \tab   \cr
  \code{\link{QDAModel}}            \tab f   \tab     \tab   \cr
  \code{\link{RandomForestModel}}   \tab f   \tab n   \tab   \cr
  \code{\link{RangerModel}}         \tab f   \tab n   \tab S \cr
  \code{\link{RFSRCModel}}          \tab f   \tab m,n \tab S \cr
  \code{\link{RFSRCFastModel}}      \tab f   \tab m,n \tab S \cr
  \code{\link{RPartModel}}          \tab f   \tab n   \tab S \cr
  \code{\link{SurvRegModel}}        \tab     \tab     \tab S \cr
  \code{\link{SurvRegStepAICModel}} \tab     \tab     \tab S \cr
  \code{\link{SVMModel}}            \tab f   \tab n   \tab   \cr
  \code{\link{SVMANOVAModel}}       \tab f   \tab n   \tab   \cr
  \code{\link{SVMBesselModel}}      \tab f   \tab n   \tab   \cr
  \code{\link{SVMLaplaceModel}}     \tab f   \tab n   \tab   \cr
  \code{\link{SVMLinearModel}}      \tab f   \tab n   \tab   \cr
  \code{\link{SVMPolyModel}}        \tab f   \tab n   \tab   \cr
  \code{\link{SVMRadialModel}}      \tab f   \tab n   \tab   \cr
  \code{\link{SVMSplineModel}}      \tab f   \tab n   \tab   \cr
  \code{\link{SVMTanhModel}}        \tab f   \tab n   \tab   \cr
  \code{\link{TreeModel}}           \tab f   \tab n   \tab   \cr
  \code{\link{XGBModel}}            \tab f   \tab n   \tab S \cr
  \code{\link{XGBDARTModel}}        \tab f   \tab n   \tab S \cr
  \code{\link{XGBLinearModel}}      \tab f   \tab n   \tab S \cr
  \code{\link{XGBTreeModel}}        \tab f   \tab n   \tab S \cr
}
Categorical: b = binary, f = factor, o = ordered\cr
Continuous: m = matrix, n = numeric\cr
Survival: S = Surv\cr
\cr
Models may be combined, tuned, or selected with the following meta-model
functions.
\tabular{ll}{
  \code{\link{ModelSpecification}} \tab Model specification \cr
  \code{\link{StackedModel}}  \tab Stacked regression \cr
  \code{\link{SuperModel}}    \tab Super learner \cr
  \code{\link{SelectedModel}} \tab Model selection from a candidate set \cr
  \code{\link{TunedModel}}    \tab Model tuning over a parameter grid \cr
}
}
\seealso{
\code{\link{modelinfo}}, \code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_GLMBoostModel.R
\name{GLMBoostModel}
\alias{GLMBoostModel}
\title{Gradient Boosting with Linear Models}
\usage{
GLMBoostModel(
  family = NULL,
  mstop = 100,
  nu = 0.1,
  risk = c("inbag", "oobag", "none"),
  stopintern = FALSE,
  trace = FALSE
)
}
\arguments{
\item{family}{optional \code{\link[mboost]{Family}} object.  Set
automatically according to the class type of the response variable.}

\item{mstop}{number of initial boosting iterations.}

\item{nu}{step size or shrinkage parameter between 0 and 1.}

\item{risk}{method to use in computing the empirical risk for each boosting
iteration.}

\item{stopintern}{logical inidicating whether the boosting algorithm stops
internally when the out-of-bag risk increases at a subsequent iteration.}

\item{trace}{logical indicating whether status information is printed during
the fitting process.}
}
\value{
\code{MLModel} class object.
}
\description{
Gradient boosting for optimizing arbitrary loss functions where
component-wise linear models are utilized as base-learners.
}
\details{
\describe{
  \item{Response types:}{\code{binary factor}, \code{BinomialVariate},
    \code{NegBinomialVariate}, \code{numeric}, \code{PoissonVariate},
    \code{Surv}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameter:}{
    \code{mstop}
  }
}

Default argument values and further model details can be found in the source
See Also links below.
}
\examples{
\donttest{
## Requires prior installation of suggested package mboost to run

data(Pima.tr, package = "MASS")

fit(type ~ ., data = Pima.tr, model = GLMBoostModel)
}

}
\seealso{
\code{\link[mboost]{glmboost}}, \code{\link[mboost]{Family}},
\code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TrainedModels.R
\name{SelectedModel}
\alias{SelectedModel}
\alias{SelectedModel.default}
\alias{SelectedModel.ModelSpecification}
\alias{SelectedModel.list}
\title{Selected Model}
\usage{
SelectedModel(...)

\method{SelectedModel}{default}(
  ...,
  control = MachineShop::settings("control"),
  metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
)

\method{SelectedModel}{ModelSpecification}(
  ...,
  control = MachineShop::settings("control"),
  metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
)

\method{SelectedModel}{list}(x, ...)
}
\arguments{
\item{...}{\link[=models]{model} functions, function names, objects; other
objects that can be \link[=as.MLModel]{coerced} to models; vectors of
these to serve as the candidate set from which to select, such as that
returned by \code{\link{expand_model}}; or model
\link[=ModelSpecification]{specifications}.}

\item{control}{\link[=controls]{control} function, function name, or object
defining the resampling method to be employed.}

\item{metrics}{\link[=metrics]{metric} function, function name, or vector of
these with which to calculate performance.  If not specified, default
metrics defined in the \link{performance} functions are used.  Model
selection is based on the first calculated metric.}

\item{cutoff}{argument passed to the \code{metrics} functions.}

\item{stat}{function or character string naming a function to compute a
summary statistic on resampled metric values for model selection.}

\item{x}{list of models followed by arguments passed to their method
function.}
}
\value{
\code{SelectedModel} or \code{SelectedModelSpecification} class
object that inherits from \code{MLModel} or \code{ModelSpecification},
respectively.
}
\description{
Model selection from a candidate set.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{numeric}, \code{ordered},
    \code{Surv}}
}
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm and glmnet to run

model_fit <- fit(
  sale_amount ~ ., data = ICHomes,
  model = SelectedModel(GBMModel, GLMNetModel, SVMRadialModel)
)
(selected_model <- as.MLModel(model_fit))
summary(selected_model)
}

}
\seealso{
\code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_NNetModel.R
\name{NNetModel}
\alias{NNetModel}
\title{Neural Network Model}
\usage{
NNetModel(
  size = 1,
  linout = logical(),
  entropy = logical(),
  softmax = logical(),
  censored = FALSE,
  skip = FALSE,
  rang = 0.7,
  decay = 0,
  maxit = 100,
  trace = FALSE,
  MaxNWts = 1000,
  abstol = 1e-04,
  reltol = 1e-08
)
}
\arguments{
\item{size}{number of units in the hidden layer.}

\item{linout}{switch for linear output units.  Set automatically according to
the class type of the response variable [numeric: \code{TRUE}, other:
\code{FALSE}].}

\item{entropy}{switch for entropy (= maximum conditional likelihood) fitting.}

\item{softmax}{switch for softmax (log-linear model) and maximum conditional
likelihood fitting.}

\item{censored}{a variant on softmax, in which non-zero targets mean possible
classes.}

\item{skip}{switch to add skip-layer connections from input to output.}

\item{rang}{Initial random weights on [\code{-rang}, \code{rang}].}

\item{decay}{parameter for weight decay.}

\item{maxit}{maximum number of iterations.}

\item{trace}{switch for tracing optimization.}

\item{MaxNWts}{maximum allowable number of weights.}

\item{abstol}{stop if the fit criterion falls below \code{abstol}, indicating
an essentially perfect fit.}

\item{reltol}{stop if the optimizer is unable to reduce the fit criterion by
a factor of at least \code{1 - reltol}.}
}
\value{
\code{MLModel} class object.
}
\description{
Fit single-hidden-layer neural network, possibly with skip-layer connections.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{numeric}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \code{size}, \code{decay}
  }
}

Default argument values and further model details can be found in the source
See Also link below.
}
\examples{
fit(sale_amount ~ ., data = ICHomes, model = NNetModel)

}
\seealso{
\code{\link[nnet]{nnet}}, \code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/step_sbf.R
\name{step_sbf}
\alias{step_sbf}
\alias{tidy.step_sbf}
\title{Variable Selection by Filtering}
\usage{
step_sbf(
  recipe,
  ...,
  filter,
  multivariate = FALSE,
  options = list(),
  replace = TRUE,
  prefix = "SBF",
  role = "predictor",
  skip = FALSE,
  id = recipes::rand_id("sbf")
)

\method{tidy}{step_sbf}(x, ...)
}
\arguments{
\item{recipe}{\link[recipes]{recipe} object to which the step will be added.}

\item{...}{one or more selector functions to choose which variables will be
used to compute the components.  See \code{\link[recipes]{selections}} for
more details.  These are not currently used by the \code{tidy} method.}

\item{filter}{function whose first argument \code{x} is a univariate vector
or a \code{multivariate} data frame of candidate variables from which to
select, second argument \code{y} is the response variable as
defined in preceding recipe steps, and third argument \code{step} is the
current step.  The function should return a logical value or vector of
length equal the number of variables in \code{x} indicating whether to
select the corresponding variable, or return a list or data frame with
element \code{`selected`} containing the logical(s) and possibly with other
elements of the same length to be included in output from the \code{tidy}
method.}

\item{multivariate}{logical indicating that candidate variables be passed to
the \code{x} argument of the \code{filter} function separately as
univariate vectors if \code{FALSE}, or altogether in one multivariate data
frame if \code{TRUE}.}

\item{options}{list of elements to be added to the step object for use in the
\code{filter} function.}

\item{replace}{logical indicating whether to replace the original variables.}

\item{prefix}{if the original variables are not replaced, the selected
variables are added to the dataset with the character string prefix added
to their names; otherwise, the original variable names are retained.}

\item{role}{analysis role that added step variables should be assigned.  By
default, they are designated as model predictors.}

\item{skip}{logical indicating whether to skip the step when the recipe is
baked.  While all operations are baked when \code{\link[recipes]{prep}} is
run, some operations may not be applicable to new data (e.g. processing
outcome variables).  Care should be taken when using \code{skip = TRUE} as
it may affect the computations for subsequent operations.}

\item{id}{unique character string to identify the step.}

\item{x}{\code{step_sbf} object.}
}
\value{
An updated version of \code{recipe} with the new step added to the
sequence of existing steps (if any).  For the \code{tidy} method, a tibble
with columns \code{terms} (selectors or variables selected), \code{selected}
(logical indicator of selected variables), and \code{name} of the selected
variable names.
}
\description{
Creates a \emph{specification} of a recipe step that will select variables
from a candidate set according to a user-specified filtering function.
}
\examples{
library(recipes)

glm_filter <- function(x, y, step) {
  model_fit <- glm(y ~ ., data = data.frame(y, x))
  p_value <- drop1(model_fit, test = "F")[-1, "Pr(>F)"]
  p_value < step$threshold
}

rec <- recipe(rating ~ ., data = attitude)
sbf_rec <- rec \%>\%
  step_sbf(all_numeric_predictors(),
           filter = glm_filter, options = list(threshold = 0.05))

sbf_prep <- prep(sbf_rec, training = attitude)
sbf_data <- bake(sbf_prep, attitude)

pairs(sbf_data, lower.panel = NULL)

tidy(sbf_rec, number = 1)
tidy(sbf_prep, number = 1)

}
\seealso{
\code{\link[recipes]{recipe}}, \code{\link[recipes]{prep}},
\code{\link[recipes]{bake}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diff.R
\name{t.test}
\alias{t.test}
\alias{t.test.PerformanceDiff}
\title{Paired t-Tests for Model Comparisons}
\usage{
\method{t.test}{PerformanceDiff}(x, adjust = "holm", ...)
}
\arguments{
\item{x}{performance \link[=diff]{difference} result.}

\item{adjust}{method of p-value adjustment for multiple statistical
comparisons as implemented by \code{\link[stats]{p.adjust}}.}

\item{...}{arguments passed to other methods.}
}
\value{
\code{PerformanceDiffTest} class object that inherits from
\code{array}.  p-values and mean differences are contained in the lower and
upper triangular portions, respectively, of the first two dimensions.  Model
pairs are contained in the third dimension.
}
\description{
Paired t-test comparisons of resampled performance metrics from different
models.
}
\details{
The t-test statistic for pairwise model differences of \eqn{R} resampled
performance metric values is calculated as
\deqn{
  t = \frac{\bar{x}_R}{\sqrt{F s^2_R / R}},
}
where \eqn{\bar{x}_R} and \eqn{s^2_R} are the sample mean and variance.
Statistical testing for a mean difference is then performed by comparing
\eqn{t} to a \eqn{t_{R-1}} null distribution.  The sample variance in the
t statistic is known to underestimate the true variances of cross-validation
mean estimators.  Underestimation of these variances will lead to increased
probabilities of false-positive statistical conclusions.  Thus, an additional
factor \eqn{F} is included in the t statistic to allow for variance
corrections.  A correction of \eqn{F = 1 + K / (K - 1)} was found by
Nadeau and Bengio (2003) to be a good choice for cross-validation with
\eqn{K} folds and is thus used for that resampling method.  The extension of
this correction by Bouchaert and Frank (2004) to \eqn{F = 1 + T K / (K - 1)}
is used for cross-validation with \eqn{K} folds repeated \eqn{T} times.  For
other resampling methods \eqn{F = 1}.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

## Numeric response example
fo <- sale_amount ~ .
control <- CVControl()

gbm_res1 <- resample(fo, ICHomes, GBMModel(n.trees = 25), control)
gbm_res2 <- resample(fo, ICHomes, GBMModel(n.trees = 50), control)
gbm_res3 <- resample(fo, ICHomes, GBMModel(n.trees = 100), control)

res <- c(GBM1 = gbm_res1, GBM2 = gbm_res2, GBM3 = gbm_res3)
res_diff <- diff(res)
t.test(res_diff)
}

}
\references{
Nadeau, C., & Bengio, Y. (2003). Inference for the generalization error.
\emph{Machine Learning}, \emph{52}, 239–81.

Bouckaert, R. R., & Frank, E. (2004). Evaluating the replicability of
significance tests for comparing learning algorithms. In H. Dai, R. Srikant,
& C. Zhang (Eds.), \emph{Advances in knowledge discovery and data mining}
(pp. 3–12). Springer.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MLControl.R
\name{MLControl}
\alias{MLControl}
\alias{controls}
\alias{BootControl}
\alias{BootOptimismControl}
\alias{CVControl}
\alias{CVOptimismControl}
\alias{OOBControl}
\alias{SplitControl}
\alias{TrainControl}
\title{Resampling Controls}
\usage{
BootControl(
  samples = 25,
  weights = TRUE,
  seed = sample(.Machine$integer.max, 1)
)

BootOptimismControl(
  samples = 25,
  weights = TRUE,
  seed = sample(.Machine$integer.max, 1)
)

CVControl(
  folds = 10,
  repeats = 1,
  weights = TRUE,
  seed = sample(.Machine$integer.max, 1)
)

CVOptimismControl(
  folds = 10,
  repeats = 1,
  weights = TRUE,
  seed = sample(.Machine$integer.max, 1)
)

OOBControl(
  samples = 25,
  weights = TRUE,
  seed = sample(.Machine$integer.max, 1)
)

SplitControl(
  prop = 2/3,
  weights = TRUE,
  seed = sample(.Machine$integer.max, 1)
)

TrainControl(weights = TRUE, seed = sample(.Machine$integer.max, 1))
}
\arguments{
\item{samples}{number of bootstrap samples.}

\item{weights}{logical indicating whether to return case weights in resampled
output for the calculation of performance \link{metrics}.}

\item{seed}{integer to set the seed at the start of resampling.}

\item{folds}{number of cross-validation folds (K).}

\item{repeats}{number of repeats of the K-fold partitioning.}

\item{prop}{proportion of cases to include in the training set
(\code{0 < prop < 1}).}
}
\value{
Object that inherits from the \code{MLControl} class.
}
\description{
Structures to define and control sampling methods for estimation of model
predictive performance in the \pkg{MachineShop} package.
}
\details{
\code{BootControl} constructs an \code{MLControl} object for simple bootstrap
resampling in which models are fit with bootstrap resampled training sets and
used to predict the full data set (Efron and Tibshirani 1993).

\code{BootOptimismControl} constructs an \code{MLControl} object for
optimism-corrected bootstrap resampling (Efron and Gong 1983, Harrell et al.
1996).

\code{CVControl} constructs an \code{MLControl} object for repeated K-fold
cross-validation (Kohavi 1995).  In this procedure, the full data set is
repeatedly partitioned into K-folds.  Within a partitioning, prediction is
performed on each of the K folds with models fit on all remaining folds.

\code{CVOptimismControl} constructs an \code{MLControl} object for
optimism-corrected cross-validation resampling (Davison and Hinkley 1997,
eq. 6.48).

\code{OOBControl} constructs an \code{MLControl} object for out-of-bootstrap
resampling in which models are fit with bootstrap resampled training sets and
used to predict the unsampled cases.

\code{SplitControl} constructs an \code{MLControl} object for splitting data
into a separate training and test set (Hastie et al. 2009).

\code{TrainControl} constructs an \code{MLControl} object for training and
performance evaluation to be performed on the same training set (Efron 1986).
}
\examples{
## Bootstrapping with 100 samples
BootControl(samples = 100)

## Optimism-corrected bootstrapping with 100 samples
BootOptimismControl(samples = 100)

## Cross-validation with 5 repeats of 10 folds
CVControl(folds = 10, repeats = 5)

## Optimism-corrected cross-validation with 5 repeats of 10 folds
CVOptimismControl(folds = 10, repeats = 5)

## Out-of-bootstrap validation with 100 samples
OOBControl(samples = 100)

## Split sample validation with 2/3 training and 1/3 testing
SplitControl(prop = 2/3)

## Training set evaluation
TrainControl()

}
\references{
Efron, B., & Tibshirani, R. J. (1993). \emph{An introduction to the
bootstrap}. Chapman & Hall/CRC.

Efron, B., & Gong, G. (1983). A leisurely look at the bootstrap, the
jackknife, and cross-validation. \emph{The American Statistician},
\emph{37}(1), 36-48.

Harrell, F. E., Lee, K. L., & Mark, D. B. (1996). Multivariable prognostic
models: Issues in developing models, evaluating assumptions and adequacy, and
measuring and reducing errors. \emph{Statistics in Medicine}, \emph{15}(4),
361-387.

Kohavi, R. (1995). A study of cross-validation and bootstrap for accuracy
estimation and model selection. In \emph{IJCAI'95: Proceedings of the 14th
International Joint Conference on Artificial Intelligence} (vol. 2, pp.
1137-1143). Morgan Kaufmann Publishers Inc.

Davison, A. C., & Hinkley, D. V. (1997). \emph{Bootstrap methods and their
application}. Cambridge University Press.

Hastie, T., Tibshirani, R., & Friedman, J. (2009). \emph{The elements of
statistical learning: data mining, inference, and prediction} (2nd ed.).
Springer.

Efron, B. (1986). How biased is the apparent error rate of a prediction rule?
\emph{Journal of the American Statistical Association}, \emph{81}(394),
461-70.
}
\seealso{
\code{\link{set_monitor}}, \code{\link{set_predict}},
\code{\link{set_strata}},
\code{\link{resample}}, \code{\link{SelectedInput}},
\code{\link{SelectedModel}}, \code{\link{TunedInput}},
\code{\link{TunedModel}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TrainedInputs.R
\name{TunedInput}
\alias{TunedInput}
\alias{TunedModelRecipe}
\alias{TunedInput.recipe}
\title{Tuned Model Inputs}
\usage{
TunedInput(object, ...)

\method{TunedInput}{recipe}(
  object,
  grid = expand_steps(),
  control = MachineShop::settings("control"),
  metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams"),
  ...
)
}
\arguments{
\item{object}{untrained \code{\link[recipes]{recipe}}.}

\item{...}{arguments passed to other methods.}

\item{grid}{\code{RecipeGrid} containing parameter values at which to
evaluate a recipe, such as those returned by \code{\link{expand_steps}}.}

\item{control}{\link[=controls]{control} function, function name, or object
defining the resampling method to be employed.}

\item{metrics}{\link[=metrics]{metric} function, function name, or vector of
these with which to calculate performance.  If not specified, default
metrics defined in the \link{performance} functions are used.  Recipe
selection is based on the first calculated metric.}

\item{cutoff}{argument passed to the \code{metrics} functions.}

\item{stat}{function or character string naming a function to compute a
summary statistic on resampled metric values for recipe tuning.}
}
\value{
\code{TunedModelRecipe} class object that inherits from
\code{TunedInput} and \code{recipe}.
}
\description{
Recipe tuning over a grid of parameter values.
}
\examples{
library(recipes)
data(Boston, package = "MASS")

rec <- recipe(medv ~ ., data = Boston) \%>\%
  step_pca(all_numeric_predictors(), id = "pca")

grid <- expand_steps(
  pca = list(num_comp = 1:2)
)

fit(TunedInput(rec, grid = grid), model = GLMModel)

}
\seealso{
\code{\link{fit}}, \code{\link{resample}}, \code{\link{set_optim}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/confusion.R
\name{confusion}
\alias{confusion}
\alias{ConfusionMatrix}
\title{Confusion Matrix}
\usage{
confusion(
  x,
  y = NULL,
  weights = NULL,
  cutoff = MachineShop::settings("cutoff"),
  na.rm = TRUE,
  ...
)

ConfusionMatrix(data = NA, ordered = FALSE)
}
\arguments{
\item{x}{factor of \link[=response]{observed responses} or \link{resample}
result containing observed and predicted responses.}

\item{y}{\link[=predict]{predicted responses} if not contained in \code{x}.}

\item{weights}{numeric vector of non-negative
\link[=case_weights]{case weights} for the observed \code{x} responses
[default: equal weights].}

\item{cutoff}{numeric (0, 1) threshold above which binary factor
probabilities are classified as events and below which survival
probabilities are classified.  If \code{NULL}, then factor responses are
summed directly over predicted class probabilities, whereas a default
cutoff of 0.5 is used for survival probabilities.  Class probability
summations and survival will appear as decimal numbers that can be
interpreted as expected counts.}

\item{na.rm}{logical indicating whether to remove observed or predicted
responses that are \code{NA} when calculating metrics.}

\item{...}{arguments passed to other methods.}

\item{data}{square matrix, or object that can be converted to one, of
cross-classified predicted and observed values in the rows and columns,
respectively.}

\item{ordered}{logical indicating whether the confusion matrix row and
columns should be regarded as ordered.}
}
\value{
The return value is a \code{ConfusionMatrix} class object that inherits from
\code{table} if \code{x} and \code{y} responses are specified or a
\code{ConfusionList} object that inherits from \code{list} if \code{x} is a
\code{Resample} object.
}
\description{
Calculate confusion matrices of predicted and observed responses.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

res <- resample(Species ~ ., data = iris, model = GBMModel)
(conf <- confusion(res))
plot(conf)
}

}
\seealso{
\code{\link{c}}, \code{\link{plot}}, \code{\link{summary}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combine.R
\name{combine}
\alias{combine}
\alias{c}
\alias{c.Calibration}
\alias{c.ConfusionList}
\alias{c.ConfusionMatrix}
\alias{c.LiftCurve}
\alias{c.ListOf}
\alias{c.PerformanceCurve}
\alias{c.Resample}
\alias{+,SurvMatrix,SurvMatrix-method}
\title{Combine MachineShop Objects}
\usage{
\method{c}{Calibration}(...)

\method{c}{ConfusionList}(...)

\method{c}{ConfusionMatrix}(...)

\method{c}{LiftCurve}(...)

\method{c}{ListOf}(...)

\method{c}{PerformanceCurve}(...)

\method{c}{Resample}(...)

\S4method{+}{SurvMatrix,SurvMatrix}(e1, e2)
}
\arguments{
\item{...}{named or unnamed \link{calibration}, \link{confusion},
\link{lift}, \link[=curves]{performance curve}, \link{summary}, or
\link{resample} results.  Curves must have been generated with the same
performance \link{metrics} and resamples with the same resampling
\link[=controls]{control}.}

\item{e1, e2}{objects.}
}
\value{
Object of the same class as the arguments.
}
\description{
Combine one or more \pkg{MachineShop} objects of the same class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grid.R
\name{TuningGrid}
\alias{TuningGrid}
\title{Tuning Grid Control}
\usage{
TuningGrid(size = 3, random = FALSE)
}
\arguments{
\item{size}{single integer or vector of integers whose positions or names
match the parameters in a model's tuning grid and which specify the number
of values used to construct the grid.}

\item{random}{number of unique points to sample at random from the grid
defined by \code{size}.  If \code{size} is a single unnamed integer, then
\code{random = Inf} will include all values of all grid parameters in the
constructed grid, whereas \code{random = FALSE} will include all values of
default grid parameters.}
}
\value{
\code{TuningGrid} class object.
}
\description{
Defines control parameters for a tuning grid.
}
\details{
Returned \code{TuningGrid} objects may be supplied to
\code{\link{TunedModel}} for automated construction of model tuning grids.
These grids can be extracted manually and viewed with the
\code{\link{expand_modelgrid}} function.
}
\examples{
TunedModel(XGBTreeModel, grid = TuningGrid(10, random = 5))

}
\seealso{
\code{\link{TunedModel}}, \code{\link{expand_modelgrid}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_BARTModel.R
\name{BARTModel}
\alias{BARTModel}
\title{Bayesian Additive Regression Trees Model}
\usage{
BARTModel(
  K = integer(),
  sparse = FALSE,
  theta = 0,
  omega = 1,
  a = 0.5,
  b = 1,
  rho = numeric(),
  augment = FALSE,
  xinfo = matrix(NA, 0, 0),
  usequants = FALSE,
  sigest = NA,
  sigdf = 3,
  sigquant = 0.9,
  lambda = NA,
  k = 2,
  power = 2,
  base = 0.95,
  tau.num = numeric(),
  offset = numeric(),
  ntree = integer(),
  numcut = 100,
  ndpost = 1000,
  nskip = integer(),
  keepevery = integer(),
  printevery = 1000
)
}
\arguments{
\item{K}{if provided, then coarsen the times of survival responses per the
quantiles \eqn{1/K, 2/K, ..., K/K} to reduce computational burdern.}

\item{sparse}{logical indicating whether to perform variable selection based
on a sparse Dirichlet prior rather than simply uniform; see Linero 2016.}

\item{theta, omega}{\eqn{theta} and \eqn{omega} parameters; zero means
random.}

\item{a, b}{sparse parameters for \eqn{Beta(a, b)} prior:
\eqn{0.5 <= a <= 1} where lower values induce more sparsity and typically
\eqn{b = 1}.}

\item{rho}{sparse parameter: typically \eqn{rho = p} where \eqn{p} is the
number of covariates under consideration.}

\item{augment}{whether data augmentation is to be performed in sparse
variable selection.}

\item{xinfo}{optional matrix whose rows are the covariates and columns their
cutpoints.}

\item{usequants}{whether covariate cutpoints are defined by uniform quantiles
or generated uniformly.}

\item{sigest}{normal error variance prior for numeric response variables.}

\item{sigdf}{degrees of freedom for error variance prior.}

\item{sigquant}{quantile at which a rough estimate of the error standard
deviation is placed.}

\item{lambda}{scale of the prior error variance.}

\item{k}{number of standard deviations \eqn{f(x)}  is away from +/-3 for
categorical response variables.}

\item{power, base}{power and base parameters for tree prior.}

\item{tau.num}{numerator in the \eqn{tau} definition, i.e.,
\eqn{tau = tau.num / (k * sqrt(ntree))}.}

\item{offset}{override for the default \eqn{offset} of \eqn{F^-1(mean(y))}
in the multivariate response probability
\eqn{P(y[j] = 1 | x) = F(f(x)[j] + offset[j])}.}

\item{ntree}{number of trees in the sum.}

\item{numcut}{number of possible covariate cutoff values.}

\item{ndpost}{number of posterior draws returned.}

\item{nskip}{number of MCMC iterations to be treated as burn in.}

\item{keepevery}{interval at which to keep posterior draws.}

\item{printevery}{interval at which to print MCMC progress.}
}
\value{
\code{MLModel} class object.
}
\description{
Flexible nonparametric modeling of covariates for continuous, binary,
categorical and time-to-event outcomes.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{numeric}, \code{Surv}}
}

Default argument values and further model details can be found in the source
See Also links below.
}
\examples{
\donttest{
## Requires prior installation of suggested package BART to run

fit(sale_amount ~ ., data = ICHomes, model = BARTModel)
}

}
\seealso{
\code{\link[BART]{gbart}}, \code{\link[BART]{mbart}},
\code{\link[BART]{surv.bart}}, \code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{ICHomes}
\alias{ICHomes}
\title{Iowa City Home Sales Dataset}
\format{
A data frame with 753 observations of 17 variables:
\describe{
  \item{sale_amount}{sale amount in dollars.}
  \item{sale_year}{sale year.}
  \item{sale_month}{sale month.}
  \item{built}{year in which the home was built.}
  \item{style}{home stlye (Home/Condo)}
  \item{construction}{home construction type.}
  \item{base_size}{base foundation size in sq ft.}
  \item{add_size}{size of additions made to the base foundation in sq ft.}
  \item{garage1_size}{attached garage size in sq ft.}
  \item{garage2_size}{detached garage size in sq ft.}
  \item{lot_size}{total lot size in sq ft.}
  \item{bedrooms}{number of bedrooms.}
  \item{basement}{presence of a basement (No/Yes).}
  \item{ac}{presence of central air conditioning (No/Yes).}
  \item{attic}{presence of a finished attic (No/Yes).}
  \item{lon,lat}{home longitude/latitude coordinates.}
}
}
\usage{
ICHomes
}
\description{
Characteristics of homes sold in Iowa City, IA from 2005 to 2008 as reported
by the county assessor's office.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_FDAModel.R
\name{FDAModel}
\alias{FDAModel}
\alias{PDAModel}
\title{Flexible and Penalized Discriminant Analysis Models}
\usage{
FDAModel(
  theta = matrix(NA, 0, 0),
  dimension = integer(),
  eps = .Machine$double.eps,
  method = .(mda::polyreg),
  ...
)

PDAModel(lambda = 1, df = numeric(), ...)
}
\arguments{
\item{theta}{optional matrix of class scores, typically with number of
columns less than one minus the number of classes.}

\item{dimension}{dimension of the discriminant subspace, less than the number
of classes, to use for prediction.}

\item{eps}{numeric threshold for small singular values for excluding
discriminant variables.}

\item{method}{regression function used in optimal scaling.  The default of
linear regression is provided by \code{\link[mda]{polyreg}} from the
\pkg{mda} package.  For penalized discriminant analysis,
\code{\link[mda]{gen.ridge}} is appropriate.  Other possibilities are
\code{\link[mda]{mars}} for multivariate adaptive regression splines and
\code{\link[mda]{bruto}} for adaptive backfitting of additive splines.  Use
the \code{\link[=quote]{.}} operator to quote specified functions.}

\item{...}{additional arguments to \code{method} for \code{FDAModel} and to
\code{FDAModel} for \code{PDAModel}.}

\item{lambda}{shrinkage penalty coefficient.}

\item{df}{alternative specification of \code{lambda} in terms of equivalent
degrees of freedom.}
}
\value{
\code{MLModel} class object.
}
\description{
Performs flexible discriminant analysis.
}
\details{
\describe{
  \item{Response types:}{\code{factor}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \itemize{
      \item FDAModel: \code{nprune}, \code{degree}*
      \item PDAModel: \code{lambda}
    }
  }
}
* excluded from grids by default

The \code{\link{predict}} function for this model additionally accepts the
following argument.
\describe{
  \item{\code{prior}}{prior class membership probabilities for prediction
    data if different from the training set.}
}

Default argument values and further model details can be found in the source
See Also links below.
}
\examples{
\donttest{
## Requires prior installation of suggested package mda to run

fit(Species ~ ., data = iris, model = FDAModel)
}

\donttest{
## Requires prior installation of suggested package mda to run

fit(Species ~ ., data = iris, model = PDAModel)
}

}
\seealso{
\code{\link[mda]{fda}}, \code{\link[mda]{predict.fda}},
\code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{as.MLModel}
\alias{as.MLModel}
\alias{as.MLModel.MLModelFit}
\alias{as.MLModel.ModelSpecification}
\alias{as.MLModel.model_spec}
\title{Coerce to an MLModel}
\usage{
as.MLModel(x, ...)

\method{as.MLModel}{MLModelFit}(x, ...)

\method{as.MLModel}{ModelSpecification}(x, ...)

\method{as.MLModel}{model_spec}(x, ...)
}
\arguments{
\item{x}{model \link{fit} result, \pkg{MachineShop}
\link[=ModelSpecification]{model specification}, or
\pkg{parsnip} \link[parsnip:model_spec]{model specification}.}

\item{...}{arguments passed to other methods.}
}
\value{
\code{MLModel} class object.
}
\description{
Function to coerce an object to \code{MLModel}.
}
\seealso{
\code{\link{ParsnipModel}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coerce.R
\name{as.data.frame}
\alias{as.data.frame}
\alias{as.data.frame.ModelFrame}
\alias{as.data.frame.Resample}
\alias{as.data.frame.TabularArray}
\title{Coerce to a Data Frame}
\usage{
\method{as.data.frame}{ModelFrame}(x, ...)

\method{as.data.frame}{Resample}(x, ...)

\method{as.data.frame}{TabularArray}(x, ...)
}
\arguments{
\item{x}{\code{\link{ModelFrame}}, \link{resample} results, resampled
\link{performance} estimates, model performance \link[=diff]{differences},
or \link[=t.test]{t-test} comparisons of the differences.}

\item{...}{arguments passed to other methods.}
}
\value{
\code{data.frame} class object.
}
\description{
Functions to coerce objects to data frames.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_POLRModel.R
\name{POLRModel}
\alias{POLRModel}
\title{Ordered Logistic or Probit Regression Model}
\usage{
POLRModel(method = c("logistic", "probit", "loglog", "cloglog", "cauchit"))
}
\arguments{
\item{method}{logistic or probit or (complementary) log-log or cauchit
(corresponding to a Cauchy latent variable).}
}
\value{
\code{MLModel} class object.
}
\description{
Fit a logistic or probit regression model to an ordered factor response.
}
\details{
\describe{
  \item{Response types:}{\code{ordered}}
}

Further model details can be found in the source link below.

In calls to \code{\link{varimp}} for \code{POLRModel}, numeric argument
\code{base} may be specified for the (negative) logarithmic transformation of
p-values [defaul: \code{exp(1)}].  Transformed p-values are automatically
scaled in the calculation of variable importance to range from 0 to 100.  To
obtain unscaled importance values, set \code{scale = FALSE}.
}
\examples{
data(Boston, package = "MASS")

df <- within(Boston,
             medv <- cut(medv,
                         breaks = c(0, 10, 15, 20, 25, 50),
                         ordered = TRUE))
fit(medv ~ ., data = df, model = POLRModel)

}
\seealso{
\code{\link[MASS]{polr}}, \code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/step_lincomp.R
\name{step_lincomp}
\alias{step_lincomp}
\alias{tidy.step_lincomp}
\alias{tunable.step_lincomp}
\title{Linear Components Variable Reduction}
\usage{
step_lincomp(
  recipe,
  ...,
  transform,
  num_comp = 5,
  options = list(),
  center = TRUE,
  scale = TRUE,
  replace = TRUE,
  prefix = "LinComp",
  role = "predictor",
  skip = FALSE,
  id = recipes::rand_id("lincomp")
)

\method{tidy}{step_lincomp}(x, ...)

\method{tunable}{step_lincomp}(x, ...)
}
\arguments{
\item{recipe}{\link[recipes]{recipe} object to which the step will be added.}

\item{...}{one or more selector functions to choose which variables will be
used to compute the components.  See \code{\link[recipes]{selections}} for
more details.  These are not currently used by the \code{tidy} method.}

\item{transform}{function whose first argument \code{x} is a matrix of
variables with which to compute linear combinations and second argument
\code{step} is the current step.  The function should return a
transformation \code{\link{matrix}} or \code{\link[Matrix]{Matrix}} of
variable weights in its columns, or return a list with element
\code{`weights`} containing the transformation matrix and possibly with
other elements to be included as attributes in output from the \code{tidy}
method.}

\item{num_comp}{number of components to derive.  The value of \code{num_comp}
will be constrained to a minimum of 1 and maximum of the number of original
variables when \code{\link[recipes]{prep}} is run.}

\item{options}{list of elements to be added to the step object for use in the
\code{transform} function.}

\item{center, scale}{logicals indicating whether to mean center and standard
deviation scale the original variables prior to deriving components, or
functions or names of functions for the centering and scaling.}

\item{replace}{logical indicating whether to replace the original variables.}

\item{prefix}{character string prefix added to a sequence of zero-padded
integers to generate names for the resulting new variables.}

\item{role}{analysis role that added step variables should be assigned.  By
default, they are designated as model predictors.}

\item{skip}{logical indicating whether to skip the step when the recipe is
baked.  While all operations are baked when \code{\link[recipes]{prep}} is
run, some operations may not be applicable to new data (e.g. processing
outcome variables).  Care should be taken when using \code{skip = TRUE} as
it may affect the computations for subsequent operations.}

\item{id}{unique character string to identify the step.}

\item{x}{\code{step_lincomp} object.}
}
\value{
An updated version of \code{recipe} with the new step added to the
sequence of existing steps (if any).  For the \code{tidy} method, a tibble
with columns \code{terms} (selectors or variables selected), \code{weight}
of each variable in the linear transformations, and \code{name} of the new
variable names.
}
\description{
Creates a \emph{specification} of a recipe step that will compute one or more
linear combinations of a set of numeric variables according to a
user-specified transformation matrix.
}
\examples{
library(recipes)

pca_mat <- function(x, step) {
  prcomp(x)$rotation[, 1:step$num_comp, drop = FALSE]
}

rec <- recipe(rating ~ ., data = attitude)
lincomp_rec <- rec \%>\%
  step_lincomp(all_numeric_predictors(),
               transform = pca_mat, num_comp = 3, prefix = "PCA")

lincomp_prep <- prep(lincomp_rec, training = attitude)
lincomp_data <- bake(lincomp_prep, attitude)

pairs(lincomp_data, lower.panel = NULL)

tidy(lincomp_rec, number = 1)
tidy(lincomp_prep, number = 1)

}
\seealso{
\code{\link[recipes]{recipe}}, \code{\link[recipes]{prep}},
\code{\link[recipes]{bake}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/step_kmeans.R
\name{step_kmeans}
\alias{step_kmeans}
\alias{tidy.step_kmeans}
\alias{tunable.step_kmeans}
\title{K-Means Clustering Variable Reduction}
\usage{
step_kmeans(
  recipe,
  ...,
  k = 5,
  center = TRUE,
  scale = TRUE,
  algorithm = c("Hartigan-Wong", "Lloyd", "Forgy", "MacQueen"),
  max_iter = 10,
  num_start = 1,
  replace = TRUE,
  prefix = "KMeans",
  role = "predictor",
  skip = FALSE,
  id = recipes::rand_id("kmeans")
)

\method{tidy}{step_kmeans}(x, ...)

\method{tunable}{step_kmeans}(x, ...)
}
\arguments{
\item{recipe}{\link[recipes]{recipe} object to which the step will be added.}

\item{...}{one or more selector functions to choose which variables will be
used to compute the components.  See \code{\link[recipes]{selections}} for
more details.  These are not currently used by the \code{tidy} method.}

\item{k}{number of k-means clusterings of the variables.  The value of
\code{k} is constrained to be between 1 and one less than the number of
original variables.}

\item{center, scale}{logicals indicating whether to mean center and standard
deviation scale the original variables prior to deriving components, or
functions or names of functions for the centering and scaling.}

\item{algorithm}{character string specifying the clustering algorithm to use.}

\item{max_iter}{maximum number of algorithm iterations allowed.}

\item{num_start}{number of random cluster centers generated for starting the
Hartigan-Wong algorithm.}

\item{replace}{logical indicating whether to replace the original variables.}

\item{prefix}{character string prefix added to a sequence of zero-padded
integers to generate names for the resulting new variables.}

\item{role}{analysis role that added step variables should be assigned.  By
default, they are designated as model predictors.}

\item{skip}{logical indicating whether to skip the step when the recipe is
baked.  While all operations are baked when \code{\link[recipes]{prep}} is
run, some operations may not be applicable to new data (e.g. processing
outcome variables).  Care should be taken when using \code{skip = TRUE} as
it may affect the computations for subsequent operations.}

\item{id}{unique character string to identify the step.}

\item{x}{\code{step_kmeans} object.}
}
\value{
Function \code{step_kmeans} creates a new step whose class is of
the same name and inherits from \code{\link{step_lincomp}}, adds it to the
sequence of existing steps (if any) in the recipe, and returns the updated
recipe.  For the \code{tidy} method, a tibble with columns \code{terms}
(selectors or variables selected), \code{cluster} assignments, \code{sqdist}
(squared distance from cluster centers), and \code{name} of the new variable
names.
}
\description{
Creates a \emph{specification} of a recipe step that will convert numeric
variables into one or more by averaging within k-means clusters.
}
\details{
K-means clustering partitions variables into k groups such that the sum of
squares between the variables and their assigned cluster means is minimized.
Variables within each cluster are then averaged to derive a new set of k
variables.
}
\examples{
library(recipes)

rec <- recipe(rating ~ ., data = attitude)
kmeans_rec <- rec \%>\%
  step_kmeans(all_predictors(), k = 3)
kmeans_prep <- prep(kmeans_rec, training = attitude)
kmeans_data <- bake(kmeans_prep, attitude)

pairs(kmeans_data, lower.panel = NULL)

tidy(kmeans_rec, number = 1)
tidy(kmeans_prep, number = 1)

}
\references{
Forgy, E. W. (1965). Cluster analysis of multivariate data: efficiency versus
interpretability of classifications. \emph{Biometrics}, \emph{21}, 768-769.

Hartigan, J. A., & Wong, M. A. (1979). A K-means clustering algorithm.
\emph{Applied Statistics}, \emph{28}, 100-108.

Lloyd, S. P. (1982). Least squares quantization in PCM. \emph{IEEE
Transactions on Information Theory}, \emph{28}(2), 129-137.

MacQueen, J. (1967). Some methods for classification and analysis of
multivariate observations. In L. M. Le Cam & J. Neyman (Eds.),
\emph{Proceedings of the fifth Berkeley Symposium on Mathematical Statistics
and Probability} (vol. 1, pp. 281-297). University of California Press.
}
\seealso{
\code{\link[stats]{kmeans}}, \code{\link[recipes]{recipe}},
\code{\link[recipes]{prep}}, \code{\link[recipes]{bake}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TrainedInputs.R
\name{SelectedInput}
\alias{SelectedInput}
\alias{SelectedModelFrame}
\alias{SelectedModelRecipe}
\alias{SelectedModelSpecification}
\alias{SelectedInput.formula}
\alias{SelectedInput.matrix}
\alias{SelectedInput.ModelFrame}
\alias{SelectedInput.recipe}
\alias{SelectedInput.ModelSpecification}
\alias{SelectedInput.list}
\title{Selected Model Inputs}
\usage{
SelectedInput(...)

\method{SelectedInput}{formula}(
  ...,
  data,
  control = MachineShop::settings("control"),
  metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
)

\method{SelectedInput}{matrix}(
  ...,
  y,
  control = MachineShop::settings("control"),
  metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
)

\method{SelectedInput}{ModelFrame}(
  ...,
  control = MachineShop::settings("control"),
  metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
)

\method{SelectedInput}{recipe}(
  ...,
  control = MachineShop::settings("control"),
  metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
)

\method{SelectedInput}{ModelSpecification}(
  ...,
  control = MachineShop::settings("control"),
  metrics = NULL,
  cutoff = MachineShop::settings("cutoff"),
  stat = MachineShop::settings("stat.TrainingParams")
)

\method{SelectedInput}{list}(x, ...)
}
\arguments{
\item{...}{\link{inputs} defining relationships between model predictor
and response variables.  Supplied inputs must all be of the same type and
may be named or unnamed.}

\item{data}{\link[=data.frame]{data frame} containing predictor and response
variables.}

\item{control}{\link[=controls]{control} function, function name, or object
defining the resampling method to be employed.}

\item{metrics}{\link[=metrics]{metric} function, function name, or vector of
these with which to calculate performance.  If not specified, default
metrics defined in the \link{performance} functions are used.  Recipe
selection is based on the first calculated metric.}

\item{cutoff}{argument passed to the \code{metrics} functions.}

\item{stat}{function or character string naming a function to compute a
summary statistic on resampled metric values for recipe selection.}

\item{y}{response variable.}

\item{x}{list of inputs followed by arguments passed to their method
function.}
}
\value{
\code{SelectedModelFrame}, \code{SelectedModelRecipe}, or
\code{SelectedModelSpecification} class object that inherits from
\code{SelectedInput} and \code{ModelFrame}, \code{recipe}, or
\code{ModelSpecification}, respectively.
}
\description{
Formula, design matrix, model frame, or recipe selection from a candidate
set.
}
\examples{
## Selected model frame
sel_mf <- SelectedInput(
  sale_amount ~ sale_year + built + style + construction,
  sale_amount ~ sale_year + base_size + bedrooms + basement,
  data = ICHomes
)

fit(sel_mf, model = GLMModel)

## Selected recipe
library(recipes)
data(Boston, package = "MASS")

rec1 <- recipe(medv ~ crim + zn + indus + chas + nox + rm, data = Boston)
rec2 <- recipe(medv ~ chas + nox + rm + age + dis + rad + tax, data = Boston)
sel_rec <- SelectedInput(rec1, rec2)

fit(sel_rec, model = GLMModel)

}
\seealso{
\code{\link{fit}}, \code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ML_RFSRCModel.R
\name{RFSRCModel}
\alias{RFSRCModel}
\alias{RFSRCFastModel}
\title{Fast Random Forest (SRC) Model}
\usage{
RFSRCModel(
  ntree = 1000,
  mtry = integer(),
  nodesize = integer(),
  nodedepth = integer(),
  splitrule = character(),
  nsplit = 10,
  block.size = integer(),
  samptype = c("swor", "swr"),
  membership = FALSE,
  sampsize = if (samptype == "swor") function(x) 0.632 * x else function(x) x,
  nimpute = 1,
  ntime = integer(),
  proximity = c(FALSE, TRUE, "inbag", "oob", "all"),
  distance = c(FALSE, TRUE, "inbag", "oob", "all"),
  forest.wt = c(FALSE, TRUE, "inbag", "oob", "all"),
  xvar.wt = numeric(),
  split.wt = numeric(),
  var.used = c(FALSE, "all.trees", "by.tree"),
  split.depth = c(FALSE, "all.trees", "by.tree"),
  do.trace = FALSE,
  statistics = FALSE
)

RFSRCFastModel(
  ntree = 500,
  sampsize = function(x) min(0.632 * x, max(x^0.75, 150)),
  ntime = 50,
  terminal.qualts = FALSE,
  ...
)
}
\arguments{
\item{ntree}{number of trees.}

\item{mtry}{number of variables randomly selected as candidates for splitting
a node.}

\item{nodesize}{minumum size of terminal nodes.}

\item{nodedepth}{maximum depth to which a tree should be grown.}

\item{splitrule}{splitting rule (see \code{\link[randomForestSRC]{rfsrc}}).}

\item{nsplit}{non-negative integer value for number of random splits to
consider for each candidate splitting variable.}

\item{block.size}{interval number of trees at which to compute the cumulative
error rate.}

\item{samptype}{whether bootstrap sampling is with or without replacement.}

\item{membership}{logical indicating whether to return terminal node
membership.}

\item{sampsize}{function specifying the bootstrap size.}

\item{nimpute}{number of iterations of the missing data imputation algorithm.}

\item{ntime}{integer number of time points to constrain ensemble calculations
for survival outcomes.}

\item{proximity}{whether and how to return proximity of cases as measured by
the frequency of sharing the same terminal nodes.}

\item{distance}{whether and how to return distance between cases as measured
by the ratio of the sum of edges from each case to the root node.}

\item{forest.wt}{whether and how to return the forest weight matrix.}

\item{xvar.wt}{vector of non-negative weights representing the probability of
selecting a variable for splitting.}

\item{split.wt}{vector of non-negative weights used for multiplying the split
statistic for a variable.}

\item{var.used}{whether and how to return variables used for splitting.}

\item{split.depth}{whether and how to return minimal depth for each variable.}

\item{do.trace}{number of seconds between updates to the user on approximate
time to completion.}

\item{statistics}{logical indicating whether to return split statistics.}

\item{terminal.qualts}{logical indicating whether to return terminal node
membership information.}

\item{...}{arguments passed to \code{RFSRCModel}.}
}
\value{
\code{MLModel} class object.
}
\description{
Fast OpenMP computing of Breiman's random forest for a variety of data
settings including right-censored survival, regression, and classification.
}
\details{
\describe{
  \item{Response types:}{\code{factor}, \code{matrix}, \code{numeric},
    \code{Surv}}
  \item{\link[=TunedModel]{Automatic tuning} of grid parameters:}{
    \code{mtry}, \code{nodesize}
  }
}

Default argument values and further model details can be found in the source
See Also links below.

In calls to \code{\link{varimp}} for \code{RFSRCModel}, argument
\code{type} may be specified as \code{"anti"} (default) for cases assigned to
the split opposite of the random assignments, as \code{"permute"} for
permutation of OOB cases, or as \code{"random"} for permutation replaced with
random assignment.  Variable importance is automatically scaled to range from
0 to 100.  To obtain unscaled importance values, set \code{scale = FALSE}.
See example below.
}
\examples{
\donttest{
## Requires prior installation of suggested package randomForestSRC to run

model_fit <- fit(sale_amount ~ ., data = ICHomes, model = RFSRCModel)
varimp(model_fit, method = "model", type = "random", scale = TRUE)
}

}
\seealso{
\code{\link[randomForestSRC]{rfsrc}},
\code{\link[randomForestSRC]{rfsrc.fast}}, \code{\link{fit}},
\code{\link{resample}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/performance_curve.R
\name{lift}
\alias{lift}
\title{Model Lift Curves}
\usage{
lift(x, y = NULL, weights = NULL, na.rm = TRUE, ...)
}
\arguments{
\item{x}{\link[=response]{observed responses} or \link{resample} result
containing observed and predicted responses.}

\item{y}{\link[=predict]{predicted responses} if not contained in \code{x}.}

\item{weights}{numeric vector of non-negative
\link[=case_weights]{case weights} for the observed \code{x} responses
[default: equal weights].}

\item{na.rm}{logical indicating whether to remove observed or predicted
responses that are \code{NA} when calculating metrics.}

\item{...}{arguments passed to other methods.}
}
\value{
\code{LiftCurve} class object that inherits from
\code{PerformanceCurve}.
}
\description{
Calculate lift curves from observed and predicted responses.
}
\examples{
\donttest{
## Requires prior installation of suggested package gbm to run

data(Pima.tr, package = "MASS")

res <- resample(type ~ ., data = Pima.tr, model = GBMModel)
lf <- lift(res)
plot(lf)
}

}
\seealso{
\code{\link{c}}, \code{\link{plot}}, \code{\link{summary}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{quote}
\alias{quote}
\alias{.}
\title{Quote Operator}
\usage{
.(expr)
}
\arguments{
\item{expr}{any syntactically valid \R expression.}
}
\value{
The quoted (unevaluated) expression.
}
\description{
Shorthand notation for the \code{\link[base:substitute]{quote}} function.
The quote operator simply returns its argument unevaluated and can be applied
to any \R expression.
}
\details{
Useful for calling \link[=models]{model functions} with quoted parameter
values defined in terms of one or more of the following variables.
\describe{
  \item{\code{nobs}}{number of observations in data to be \link{fit}.}
  \item{\code{nvars}}{number of predictor variables.}
  \item{\code{y}}{the response variable.}
}
}
\examples{
## Stepwise variable selection with BIC
glm_fit <- fit(sale_amount ~ ., ICHomes, GLMStepAICModel(k = .(log(nobs))))
varimp(glm_fit)

}
\seealso{
\code{\link[base:substitute]{quote}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/models.R
\name{inputs}
\alias{inputs}
\title{Model Inputs}
\description{
Model inputs are the predictor and response variables whose relationship is
determined by a model fit.  Input specifications supported by
\pkg{MachineShop} are summarized in the table below.
\tabular{ll}{
  \code{\link{formula}}         \tab Traditional model formula \cr
  \code{\link{matrix}}          \tab Design matrix of predictors \cr
  \code{\link{ModelFrame}}      \tab Model frame \cr
  \code{\link{ModelSpecification}} \tab Model specification \cr
  \code{\link[recipes]{recipe}} \tab Preprocessing recipe roles and steps \cr
}
Response variable types in the input specifications are defined by the user
with the functions and recipe roles:
\tabular{ll}{
  \strong{Response Functions}
    \tab \code{\link{BinomialVariate}} \cr
    \tab \code{\link{DiscreteVariate}} \cr
    \tab \code{\link{factor}} \cr
    \tab \code{\link{matrix}} \cr
    \tab \code{\link{NegBinomialVariate}} \cr
    \tab \code{\link{numeric}} \cr
    \tab \code{\link{ordered}} \cr
    \tab \code{\link{PoissonVariate}} \cr
    \tab \code{\link[survival]{Surv}} \cr
  \strong{Recipe Roles}
    \tab \code{\link{role_binom}} \cr
    \tab \code{\link{role_surv}} \cr
}
Inputs may be combined, selected, or tuned with the following meta-input
functions.
\tabular{ll}{
  \code{\link{ModelSpecification}} \tab Model specification \cr
  \code{\link{SelectedInput}} \tab Input selection from a candidate set \cr
  \code{\link{TunedInput}}    \tab Input tuning over a parameter grid \cr
}
}
\seealso{
\code{\link{fit}}, \code{\link{resample}}
}
